///////////////////////////////////////////////////////////////////////////////
// vim:         ts=4 sw=4
// Name:        xm/encoding/cp20932.cpp
// Description: Define the Double-byte Encoding CP20932
// Copyright:   2013-2023  JiaYanwei   <wxmedit@gmail.com>
// License:     GPLv3
///////////////////////////////////////////////////////////////////////////////

#include "cp20932.h"
#include <unicode/uversion.h>
#include <boost/static_assert.hpp>

#ifdef _DEBUG
#include <crtdbg.h>
#define new new(_NORMAL_BLOCK ,__FILE__, __LINE__)
#endif

namespace xm
{

size_t CP20932Converter::MB2WC(UChar32& ch, const char* src, size_t src_len)
{
	if (src_len == 2 && (src[1] & 0x80) == 0)
	{
		char mbs[3] = { (char)0x8F, src[0], (char)(src[1] | 0x80)};
		return ICUConverter::MB2WC(ch, mbs, 3);
	}

	return ICUConverter::MB2WC(ch, src, src_len);
}

size_t CP20932Converter::WC2MB(char* dest, size_t dest_len, const UChar32& ch)
{
	char mbs[3];
	size_t len = ICUConverter::WC2MB(mbs, 3, ch);
	if (len == 0)
		return 0;

	if (len == 3)
	{
		dest[0] = mbs[1];
		dest[1] = mbs[2] & 0x7f;
		return 2;
	}

	dest[0] = mbs[0];
	if (len == 1)
		return 1;

	dest[1] = mbs[1];
	return 2;
}

void CP20932TableFixer::fix()
{
	ChangeMB2U(0x8F00, 0x00008F);
	ChangeMB2U(0xA000, 0x00F8F0);

	RemoveMB2U(0x8EE0);
	RemoveMB2U(0x8EE1);
	RemoveMB2U(0x8EE2);
	RemoveMB2U(0xA237);
	RemoveMB2U(0xA271);
	RemoveMB2U(0xAC00);
	RemoveMB2U(0xADF0);
	RemoveMB2U(0xADF1);
	RemoveMB2U(0xADF2);
	RemoveMB2U(0xADF5);
	RemoveMB2U(0xADF6);
	RemoveMB2U(0xADF7);
	RemoveMB2U(0xADFA);
	RemoveMB2U(0xADFB);
	RemoveMB2U(0xADFC);
	RemoveMB2U(0xAE00);
	RemoveMB2U(0xAF00);
	RemoveMB2U(0xF32B);
	RemoveMB2U(0xF32C);
	RemoveMB2U(0xF32D);
	RemoveMB2U(0xF32E);
	RemoveMB2U(0xF32F);
	RemoveMB2U(0xF330);
	RemoveMB2U(0xF331);
	RemoveMB2U(0xF332);
	RemoveMB2U(0xF333);
	RemoveMB2U(0xF334);
	RemoveMB2U(0xF337);

	AddLeadByte(0xF5);
	AddLeadByte(0xF6);
	AddLeadByte(0xF7);
	AddLeadByte(0xF8);
	AddLeadByte(0xFD);
	AddLeadByte(0xFE);

	ChangeMB2U(0xA1C1, 0x00301C);
	ChangeMB2U(0xA1C2, 0x002016);
	ChangeMB2U(0xA1DD, 0x002212);
	ChangeMB2U(0xA1F1, 0x0000A2);
	ChangeMB2U(0xA1F2, 0x0000A3);
	ChangeMB2U(0xA2CC, 0x0000AC);
	ChangeMB2U(0xADE0, 0x00301E);
	ChangeMB2U(0xF335, 0x00FF07);
	ChangeMB2U(0xF336, 0x00FF02);
	ChangeMB2U(0xF33A, 0x0070BB);
	ChangeMB2U(0xF33B, 0x004EFC);
	ChangeMB2U(0xF33C, 0x0050F4);
	ChangeMB2U(0xF33D, 0x0051EC);
	ChangeMB2U(0xF33E, 0x005307);
	ChangeMB2U(0xF33F, 0x005324);
	ChangeMB2U(0xF340, 0x00FA0E);
	ChangeMB2U(0xF341, 0x00548A);
	ChangeMB2U(0xF342, 0x005759);
	ChangeMB2U(0xF343, 0x00FA0F);
	ChangeMB2U(0xF344, 0x00FA10);
	ChangeMB2U(0xF345, 0x00589E);
	ChangeMB2U(0xF346, 0x005BEC);
	ChangeMB2U(0xF347, 0x005CF5);
	ChangeMB2U(0xF348, 0x005D53);
	ChangeMB2U(0xF349, 0x00FA11);
	ChangeMB2U(0xF34A, 0x005FB7);
	ChangeMB2U(0xF34B, 0x006085);
	ChangeMB2U(0xF34C, 0x006120);
	ChangeMB2U(0xF34D, 0x00654E);
	ChangeMB2U(0xF34F, 0x006665);
	ChangeMB2U(0xF350, 0x00FA12);
	ChangeMB2U(0xF351, 0x00F929);
	ChangeMB2U(0xF352, 0x006801);
	ChangeMB2U(0xF353, 0x00FA13);
	ChangeMB2U(0xF354, 0x00FA14);
	ChangeMB2U(0xF355, 0x006A6B);
	ChangeMB2U(0xF356, 0x006AE2);
	ChangeMB2U(0xF357, 0x006DF8);
	ChangeMB2U(0xF358, 0x006DF2);
	ChangeMB2U(0xF359, 0x007028);
	ChangeMB2U(0xF35A, 0x00FA15);
	ChangeMB2U(0xF35B, 0x00FA16);
	ChangeMB2U(0xF35C, 0x007501);
	ChangeMB2U(0xF35D, 0x007682);
	ChangeMB2U(0xF35E, 0x00769E);
	ChangeMB2U(0xF35F, 0x00FA17);
	ChangeMB2U(0xF360, 0x007930);
	ChangeMB2U(0xF361, 0x00FA18);
	ChangeMB2U(0xF362, 0x00FA19);
	ChangeMB2U(0xF363, 0x00FA1A);
	ChangeMB2U(0xF364, 0x00FA1B);
	ChangeMB2U(0xF365, 0x007AE7);
	ChangeMB2U(0xF366, 0x00FA1C);
	ChangeMB2U(0xF367, 0x00FA1D);
	ChangeMB2U(0xF368, 0x007DA0);
	ChangeMB2U(0xF369, 0x007DD6);
	ChangeMB2U(0xF36A, 0x00FA1E);
	ChangeMB2U(0xF36B, 0x008362);
	ChangeMB2U(0xF36C, 0x00FA1F);
	ChangeMB2U(0xF36D, 0x0085B0);
	ChangeMB2U(0xF36E, 0x00FA20);
	ChangeMB2U(0xF36F, 0x00FA21);
	ChangeMB2U(0xF370, 0x008807);
	ChangeMB2U(0xF371, 0x00FA22);
	ChangeMB2U(0xF372, 0x008B7F);
	ChangeMB2U(0xF373, 0x008CF4);
	ChangeMB2U(0xF374, 0x008D76);
	ChangeMB2U(0xF375, 0x00FA23);
	ChangeMB2U(0xF376, 0x00FA24);
	ChangeMB2U(0xF377, 0x00FA25);
	ChangeMB2U(0xF378, 0x0090DE);
	ChangeMB2U(0xF379, 0x00FA26);
	ChangeMB2U(0xF37A, 0x009115);
	ChangeMB2U(0xF37B, 0x00FA27);
	ChangeMB2U(0xF37C, 0x00FA28);
	ChangeMB2U(0xF37D, 0x009592);
	ChangeMB2U(0xF37E, 0x00F9DC);
	ChangeMB2U(0xF421, 0x00FA29);
	ChangeMB2U(0xF422, 0x00973B);
	ChangeMB2U(0xF424, 0x009751);
	ChangeMB2U(0xF425, 0x00FA2A);
	ChangeMB2U(0xF426, 0x00FA2B);
	ChangeMB2U(0xF427, 0x00FA2C);
	ChangeMB2U(0xF428, 0x00999E);
	ChangeMB2U(0xF429, 0x009AD9);
	ChangeMB2U(0xF42A, 0x009B72);
	ChangeMB2U(0xF42B, 0x00FA2D);
	ChangeMB2U(0xF42C, 0x009ED1);
	ChangeMB2U(0xF463, 0x00974D);
	ChangeMB2U(0xF521, 0x00E3AC);
	ChangeMB2U(0xF522, 0x00E3AD);
	ChangeMB2U(0xF523, 0x00E3AE);
	ChangeMB2U(0xF524, 0x00E3AF);
	ChangeMB2U(0xF525, 0x00E3B0);
	ChangeMB2U(0xF526, 0x00E3B1);
	ChangeMB2U(0xF527, 0x00E3B2);
	ChangeMB2U(0xF528, 0x00E3B3);
	ChangeMB2U(0xF529, 0x00E3B4);
	ChangeMB2U(0xF52A, 0x00E3B5);
	ChangeMB2U(0xF52B, 0x00E3B6);
	ChangeMB2U(0xF52C, 0x00E3B7);
	ChangeMB2U(0xF52D, 0x00E3B8);
	ChangeMB2U(0xF52E, 0x00E3B9);
	ChangeMB2U(0xF52F, 0x00E3BA);
	ChangeMB2U(0xF530, 0x00E3BB);
	ChangeMB2U(0xF531, 0x00E3BC);
	ChangeMB2U(0xF532, 0x00E3BD);
	ChangeMB2U(0xF533, 0x00E3BE);
	ChangeMB2U(0xF534, 0x00E3BF);
	ChangeMB2U(0xF535, 0x00E3C0);
	ChangeMB2U(0xF536, 0x00E3C1);
	ChangeMB2U(0xF537, 0x00E3C2);
	ChangeMB2U(0xF538, 0x00E3C3);
	ChangeMB2U(0xF539, 0x00E3C4);
	ChangeMB2U(0xF53A, 0x00E3C5);
	ChangeMB2U(0xF53B, 0x00E3C6);
	ChangeMB2U(0xF53C, 0x00E3C7);
	ChangeMB2U(0xF53D, 0x00E3C8);
	ChangeMB2U(0xF53E, 0x00E3C9);
	ChangeMB2U(0xF53F, 0x00E3CA);
	ChangeMB2U(0xF540, 0x00E3CB);
	ChangeMB2U(0xF541, 0x00E3CC);
	ChangeMB2U(0xF542, 0x00E3CD);
	ChangeMB2U(0xF543, 0x00E3CE);
	ChangeMB2U(0xF544, 0x00E3CF);
	ChangeMB2U(0xF545, 0x00E3D0);
	ChangeMB2U(0xF546, 0x00E3D1);
	ChangeMB2U(0xF547, 0x00E3D2);
	ChangeMB2U(0xF548, 0x00E3D3);
	ChangeMB2U(0xF549, 0x00E3D4);
	ChangeMB2U(0xF54A, 0x00E3D5);
	ChangeMB2U(0xF54B, 0x00E3D6);
	ChangeMB2U(0xF54C, 0x00E3D7);
	ChangeMB2U(0xF54D, 0x00E3D8);
	ChangeMB2U(0xF54E, 0x00E3D9);
	ChangeMB2U(0xF54F, 0x00E3DA);
	ChangeMB2U(0xF550, 0x00E3DB);
	ChangeMB2U(0xF551, 0x00E3DC);
	ChangeMB2U(0xF552, 0x00E3DD);
	ChangeMB2U(0xF553, 0x00E3DE);
	ChangeMB2U(0xF554, 0x00E3DF);
	ChangeMB2U(0xF555, 0x00E3E0);
	ChangeMB2U(0xF556, 0x00E3E1);
	ChangeMB2U(0xF557, 0x00E3E2);
	ChangeMB2U(0xF558, 0x00E3E3);
	ChangeMB2U(0xF559, 0x00E3E4);
	ChangeMB2U(0xF55A, 0x00E3E5);
	ChangeMB2U(0xF55B, 0x00E3E6);
	ChangeMB2U(0xF55C, 0x00E3E7);
	ChangeMB2U(0xF55D, 0x00E3E8);
	ChangeMB2U(0xF55E, 0x00E3E9);
	ChangeMB2U(0xF55F, 0x00E3EA);
	ChangeMB2U(0xF560, 0x00E3EB);
	ChangeMB2U(0xF561, 0x00E3EC);
	ChangeMB2U(0xF562, 0x00E3ED);
	ChangeMB2U(0xF563, 0x00E3EE);
	ChangeMB2U(0xF564, 0x00E3EF);
	ChangeMB2U(0xF565, 0x00E3F0);
	ChangeMB2U(0xF566, 0x00E3F1);
	ChangeMB2U(0xF567, 0x00E3F2);
	ChangeMB2U(0xF568, 0x00E3F3);
	ChangeMB2U(0xF569, 0x00E3F4);
	ChangeMB2U(0xF56A, 0x00E3F5);
	ChangeMB2U(0xF56B, 0x00E3F6);
	ChangeMB2U(0xF56C, 0x00E3F7);
	ChangeMB2U(0xF56D, 0x00E3F8);
	ChangeMB2U(0xF56E, 0x00E3F9);
	ChangeMB2U(0xF56F, 0x00E3FA);
	ChangeMB2U(0xF570, 0x00E3FB);
	ChangeMB2U(0xF571, 0x00E3FC);
	ChangeMB2U(0xF572, 0x00E3FD);
	ChangeMB2U(0xF573, 0x00E3FE);
	ChangeMB2U(0xF574, 0x00E3FF);
	ChangeMB2U(0xF575, 0x00E400);
	ChangeMB2U(0xF576, 0x00E401);
	ChangeMB2U(0xF577, 0x00E402);
	ChangeMB2U(0xF578, 0x00E403);
	ChangeMB2U(0xF579, 0x00E404);
	ChangeMB2U(0xF57A, 0x00E405);
	ChangeMB2U(0xF57B, 0x00E406);
	ChangeMB2U(0xF57C, 0x00E407);
	ChangeMB2U(0xF57D, 0x00E408);
	ChangeMB2U(0xF57E, 0x00E409);
	ChangeMB2U(0xF5A1, 0x00E000);
	ChangeMB2U(0xF5A2, 0x00E001);
	ChangeMB2U(0xF5A3, 0x00E002);
	ChangeMB2U(0xF5A4, 0x00E003);
	ChangeMB2U(0xF5A5, 0x00E004);
	ChangeMB2U(0xF5A6, 0x00E005);
	ChangeMB2U(0xF5A7, 0x00E006);
	ChangeMB2U(0xF5A8, 0x00E007);
	ChangeMB2U(0xF5A9, 0x00E008);
	ChangeMB2U(0xF5AA, 0x00E009);
	ChangeMB2U(0xF5AB, 0x00E00A);
	ChangeMB2U(0xF5AC, 0x00E00B);
	ChangeMB2U(0xF5AD, 0x00E00C);
	ChangeMB2U(0xF5AE, 0x00E00D);
	ChangeMB2U(0xF5AF, 0x00E00E);
	ChangeMB2U(0xF5B0, 0x00E00F);
	ChangeMB2U(0xF5B1, 0x00E010);
	ChangeMB2U(0xF5B2, 0x00E011);
	ChangeMB2U(0xF5B3, 0x00E012);
	ChangeMB2U(0xF5B4, 0x00E013);
	ChangeMB2U(0xF5B5, 0x00E014);
	ChangeMB2U(0xF5B6, 0x00E015);
	ChangeMB2U(0xF5B7, 0x00E016);
	ChangeMB2U(0xF5B8, 0x00E017);
	ChangeMB2U(0xF5B9, 0x00E018);
	ChangeMB2U(0xF5BA, 0x00E019);
	ChangeMB2U(0xF5BB, 0x00E01A);
	ChangeMB2U(0xF5BC, 0x00E01B);
	ChangeMB2U(0xF5BD, 0x00E01C);
	ChangeMB2U(0xF5BE, 0x00E01D);
	ChangeMB2U(0xF5BF, 0x00E01E);
	ChangeMB2U(0xF5C0, 0x00E01F);
	ChangeMB2U(0xF5C1, 0x00E020);
	ChangeMB2U(0xF5C2, 0x00E021);
	ChangeMB2U(0xF5C3, 0x00E022);
	ChangeMB2U(0xF5C4, 0x00E023);
	ChangeMB2U(0xF5C5, 0x00E024);
	ChangeMB2U(0xF5C6, 0x00E025);
	ChangeMB2U(0xF5C7, 0x00E026);
	ChangeMB2U(0xF5C8, 0x00E027);
	ChangeMB2U(0xF5C9, 0x00E028);
	ChangeMB2U(0xF5CA, 0x00E029);
	ChangeMB2U(0xF5CB, 0x00E02A);
	ChangeMB2U(0xF5CC, 0x00E02B);
	ChangeMB2U(0xF5CD, 0x00E02C);
	ChangeMB2U(0xF5CE, 0x00E02D);
	ChangeMB2U(0xF5CF, 0x00E02E);
	ChangeMB2U(0xF5D0, 0x00E02F);
	ChangeMB2U(0xF5D1, 0x00E030);
	ChangeMB2U(0xF5D2, 0x00E031);
	ChangeMB2U(0xF5D3, 0x00E032);
	ChangeMB2U(0xF5D4, 0x00E033);
	ChangeMB2U(0xF5D5, 0x00E034);
	ChangeMB2U(0xF5D6, 0x00E035);
	ChangeMB2U(0xF5D7, 0x00E036);
	ChangeMB2U(0xF5D8, 0x00E037);
	ChangeMB2U(0xF5D9, 0x00E038);
	ChangeMB2U(0xF5DA, 0x00E039);
	ChangeMB2U(0xF5DB, 0x00E03A);
	ChangeMB2U(0xF5DC, 0x00E03B);
	ChangeMB2U(0xF5DD, 0x00E03C);
	ChangeMB2U(0xF5DE, 0x00E03D);
	ChangeMB2U(0xF5DF, 0x00E03E);
	ChangeMB2U(0xF5E0, 0x00E03F);
	ChangeMB2U(0xF5E1, 0x00E040);
	ChangeMB2U(0xF5E2, 0x00E041);
	ChangeMB2U(0xF5E3, 0x00E042);
	ChangeMB2U(0xF5E4, 0x00E043);
	ChangeMB2U(0xF5E5, 0x00E044);
	ChangeMB2U(0xF5E6, 0x00E045);
	ChangeMB2U(0xF5E7, 0x00E046);
	ChangeMB2U(0xF5E8, 0x00E047);
	ChangeMB2U(0xF5E9, 0x00E048);
	ChangeMB2U(0xF5EA, 0x00E049);
	ChangeMB2U(0xF5EB, 0x00E04A);
	ChangeMB2U(0xF5EC, 0x00E04B);
	ChangeMB2U(0xF5ED, 0x00E04C);
	ChangeMB2U(0xF5EE, 0x00E04D);
	ChangeMB2U(0xF5EF, 0x00E04E);
	ChangeMB2U(0xF5F0, 0x00E04F);
	ChangeMB2U(0xF5F1, 0x00E050);
	ChangeMB2U(0xF5F2, 0x00E051);
	ChangeMB2U(0xF5F3, 0x00E052);
	ChangeMB2U(0xF5F4, 0x00E053);
	ChangeMB2U(0xF5F5, 0x00E054);
	ChangeMB2U(0xF5F6, 0x00E055);
	ChangeMB2U(0xF5F7, 0x00E056);
	ChangeMB2U(0xF5F8, 0x00E057);
	ChangeMB2U(0xF5F9, 0x00E058);
	ChangeMB2U(0xF5FA, 0x00E059);
	ChangeMB2U(0xF5FB, 0x00E05A);
	ChangeMB2U(0xF5FC, 0x00E05B);
	ChangeMB2U(0xF5FD, 0x00E05C);
	ChangeMB2U(0xF5FE, 0x00E05D);
	ChangeMB2U(0xF621, 0x00E40A);
	ChangeMB2U(0xF622, 0x00E40B);
	ChangeMB2U(0xF623, 0x00E40C);
	ChangeMB2U(0xF624, 0x00E40D);
	ChangeMB2U(0xF625, 0x00E40E);
	ChangeMB2U(0xF626, 0x00E40F);
	ChangeMB2U(0xF627, 0x00E410);
	ChangeMB2U(0xF628, 0x00E411);
	ChangeMB2U(0xF629, 0x00E412);
	ChangeMB2U(0xF62A, 0x00E413);
	ChangeMB2U(0xF62B, 0x00E414);
	ChangeMB2U(0xF62C, 0x00E415);
	ChangeMB2U(0xF62D, 0x00E416);
	ChangeMB2U(0xF62E, 0x00E417);
	ChangeMB2U(0xF62F, 0x00E418);
	ChangeMB2U(0xF630, 0x00E419);
	ChangeMB2U(0xF631, 0x00E41A);
	ChangeMB2U(0xF632, 0x00E41B);
	ChangeMB2U(0xF633, 0x00E41C);
	ChangeMB2U(0xF634, 0x00E41D);
	ChangeMB2U(0xF635, 0x00E41E);
	ChangeMB2U(0xF636, 0x00E41F);
	ChangeMB2U(0xF637, 0x00E420);
	ChangeMB2U(0xF638, 0x00E421);
	ChangeMB2U(0xF639, 0x00E422);
	ChangeMB2U(0xF63A, 0x00E423);
	ChangeMB2U(0xF63B, 0x00E424);
	ChangeMB2U(0xF63C, 0x00E425);
	ChangeMB2U(0xF63D, 0x00E426);
	ChangeMB2U(0xF63E, 0x00E427);
	ChangeMB2U(0xF63F, 0x00E428);
	ChangeMB2U(0xF640, 0x00E429);
	ChangeMB2U(0xF641, 0x00E42A);
	ChangeMB2U(0xF642, 0x00E42B);
	ChangeMB2U(0xF643, 0x00E42C);
	ChangeMB2U(0xF644, 0x00E42D);
	ChangeMB2U(0xF645, 0x00E42E);
	ChangeMB2U(0xF646, 0x00E42F);
	ChangeMB2U(0xF647, 0x00E430);
	ChangeMB2U(0xF648, 0x00E431);
	ChangeMB2U(0xF649, 0x00E432);
	ChangeMB2U(0xF64A, 0x00E433);
	ChangeMB2U(0xF64B, 0x00E434);
	ChangeMB2U(0xF64C, 0x00E435);
	ChangeMB2U(0xF64D, 0x00E436);
	ChangeMB2U(0xF64E, 0x00E437);
	ChangeMB2U(0xF64F, 0x00E438);
	ChangeMB2U(0xF650, 0x00E439);
	ChangeMB2U(0xF651, 0x00E43A);
	ChangeMB2U(0xF652, 0x00E43B);
	ChangeMB2U(0xF653, 0x00E43C);
	ChangeMB2U(0xF654, 0x00E43D);
	ChangeMB2U(0xF655, 0x00E43E);
	ChangeMB2U(0xF656, 0x00E43F);
	ChangeMB2U(0xF657, 0x00E440);
	ChangeMB2U(0xF658, 0x00E441);
	ChangeMB2U(0xF659, 0x00E442);
	ChangeMB2U(0xF65A, 0x00E443);
	ChangeMB2U(0xF65B, 0x00E444);
	ChangeMB2U(0xF65C, 0x00E445);
	ChangeMB2U(0xF65D, 0x00E446);
	ChangeMB2U(0xF65E, 0x00E447);
	ChangeMB2U(0xF65F, 0x00E448);
	ChangeMB2U(0xF660, 0x00E449);
	ChangeMB2U(0xF661, 0x00E44A);
	ChangeMB2U(0xF662, 0x00E44B);
	ChangeMB2U(0xF663, 0x00E44C);
	ChangeMB2U(0xF664, 0x00E44D);
	ChangeMB2U(0xF665, 0x00E44E);
	ChangeMB2U(0xF666, 0x00E44F);
	ChangeMB2U(0xF667, 0x00E450);
	ChangeMB2U(0xF668, 0x00E451);
	ChangeMB2U(0xF669, 0x00E452);
	ChangeMB2U(0xF66A, 0x00E453);
	ChangeMB2U(0xF66B, 0x00E454);
	ChangeMB2U(0xF66C, 0x00E455);
	ChangeMB2U(0xF66D, 0x00E456);
	ChangeMB2U(0xF66E, 0x00E457);
	ChangeMB2U(0xF66F, 0x00E458);
	ChangeMB2U(0xF670, 0x00E459);
	ChangeMB2U(0xF671, 0x00E45A);
	ChangeMB2U(0xF672, 0x00E45B);
	ChangeMB2U(0xF673, 0x00E45C);
	ChangeMB2U(0xF674, 0x00E45D);
	ChangeMB2U(0xF675, 0x00E45E);
	ChangeMB2U(0xF676, 0x00E45F);
	ChangeMB2U(0xF677, 0x00E460);
	ChangeMB2U(0xF678, 0x00E461);
	ChangeMB2U(0xF679, 0x00E462);
	ChangeMB2U(0xF67A, 0x00E463);
	ChangeMB2U(0xF67B, 0x00E464);
	ChangeMB2U(0xF67C, 0x00E465);
	ChangeMB2U(0xF67D, 0x00E466);
	ChangeMB2U(0xF67E, 0x00E467);
	ChangeMB2U(0xF6A1, 0x00E05E);
	ChangeMB2U(0xF6A2, 0x00E05F);
	ChangeMB2U(0xF6A3, 0x00E060);
	ChangeMB2U(0xF6A4, 0x00E061);
	ChangeMB2U(0xF6A5, 0x00E062);
	ChangeMB2U(0xF6A6, 0x00E063);
	ChangeMB2U(0xF6A7, 0x00E064);
	ChangeMB2U(0xF6A8, 0x00E065);
	ChangeMB2U(0xF6A9, 0x00E066);
	ChangeMB2U(0xF6AA, 0x00E067);
	ChangeMB2U(0xF6AB, 0x00E068);
	ChangeMB2U(0xF6AC, 0x00E069);
	ChangeMB2U(0xF6AD, 0x00E06A);
	ChangeMB2U(0xF6AE, 0x00E06B);
	ChangeMB2U(0xF6AF, 0x00E06C);
	ChangeMB2U(0xF6B0, 0x00E06D);
	ChangeMB2U(0xF6B1, 0x00E06E);
	ChangeMB2U(0xF6B2, 0x00E06F);
	ChangeMB2U(0xF6B3, 0x00E070);
	ChangeMB2U(0xF6B4, 0x00E071);
	ChangeMB2U(0xF6B5, 0x00E072);
	ChangeMB2U(0xF6B6, 0x00E073);
	ChangeMB2U(0xF6B7, 0x00E074);
	ChangeMB2U(0xF6B8, 0x00E075);
	ChangeMB2U(0xF6B9, 0x00E076);
	ChangeMB2U(0xF6BA, 0x00E077);
	ChangeMB2U(0xF6BB, 0x00E078);
	ChangeMB2U(0xF6BC, 0x00E079);
	ChangeMB2U(0xF6BD, 0x00E07A);
	ChangeMB2U(0xF6BE, 0x00E07B);
	ChangeMB2U(0xF6BF, 0x00E07C);
	ChangeMB2U(0xF6C0, 0x00E07D);
	ChangeMB2U(0xF6C1, 0x00E07E);
	ChangeMB2U(0xF6C2, 0x00E07F);
	ChangeMB2U(0xF6C3, 0x00E080);
	ChangeMB2U(0xF6C4, 0x00E081);
	ChangeMB2U(0xF6C5, 0x00E082);
	ChangeMB2U(0xF6C6, 0x00E083);
	ChangeMB2U(0xF6C7, 0x00E084);
	ChangeMB2U(0xF6C8, 0x00E085);
	ChangeMB2U(0xF6C9, 0x00E086);
	ChangeMB2U(0xF6CA, 0x00E087);
	ChangeMB2U(0xF6CB, 0x00E088);
	ChangeMB2U(0xF6CC, 0x00E089);
	ChangeMB2U(0xF6CD, 0x00E08A);
	ChangeMB2U(0xF6CE, 0x00E08B);
	ChangeMB2U(0xF6CF, 0x00E08C);
	ChangeMB2U(0xF6D0, 0x00E08D);
	ChangeMB2U(0xF6D1, 0x00E08E);
	ChangeMB2U(0xF6D2, 0x00E08F);
	ChangeMB2U(0xF6D3, 0x00E090);
	ChangeMB2U(0xF6D4, 0x00E091);
	ChangeMB2U(0xF6D5, 0x00E092);
	ChangeMB2U(0xF6D6, 0x00E093);
	ChangeMB2U(0xF6D7, 0x00E094);
	ChangeMB2U(0xF6D8, 0x00E095);
	ChangeMB2U(0xF6D9, 0x00E096);
	ChangeMB2U(0xF6DA, 0x00E097);
	ChangeMB2U(0xF6DB, 0x00E098);
	ChangeMB2U(0xF6DC, 0x00E099);
	ChangeMB2U(0xF6DD, 0x00E09A);
	ChangeMB2U(0xF6DE, 0x00E09B);
	ChangeMB2U(0xF6DF, 0x00E09C);
	ChangeMB2U(0xF6E0, 0x00E09D);
	ChangeMB2U(0xF6E1, 0x00E09E);
	ChangeMB2U(0xF6E2, 0x00E09F);
	ChangeMB2U(0xF6E3, 0x00E0A0);
	ChangeMB2U(0xF6E4, 0x00E0A1);
	ChangeMB2U(0xF6E5, 0x00E0A2);
	ChangeMB2U(0xF6E6, 0x00E0A3);
	ChangeMB2U(0xF6E7, 0x00E0A4);
	ChangeMB2U(0xF6E8, 0x00E0A5);
	ChangeMB2U(0xF6E9, 0x00E0A6);
	ChangeMB2U(0xF6EA, 0x00E0A7);
	ChangeMB2U(0xF6EB, 0x00E0A8);
	ChangeMB2U(0xF6EC, 0x00E0A9);
	ChangeMB2U(0xF6ED, 0x00E0AA);
	ChangeMB2U(0xF6EE, 0x00E0AB);
	ChangeMB2U(0xF6EF, 0x00E0AC);
	ChangeMB2U(0xF6F0, 0x00E0AD);
	ChangeMB2U(0xF6F1, 0x00E0AE);
	ChangeMB2U(0xF6F2, 0x00E0AF);
	ChangeMB2U(0xF6F3, 0x00E0B0);
	ChangeMB2U(0xF6F4, 0x00E0B1);
	ChangeMB2U(0xF6F5, 0x00E0B2);
	ChangeMB2U(0xF6F6, 0x00E0B3);
	ChangeMB2U(0xF6F7, 0x00E0B4);
	ChangeMB2U(0xF6F8, 0x00E0B5);
	ChangeMB2U(0xF6F9, 0x00E0B6);
	ChangeMB2U(0xF6FA, 0x00E0B7);
	ChangeMB2U(0xF6FB, 0x00E0B8);
	ChangeMB2U(0xF6FC, 0x00E0B9);
	ChangeMB2U(0xF6FD, 0x00E0BA);
	ChangeMB2U(0xF6FE, 0x00E0BB);
	ChangeMB2U(0xF721, 0x00E468);
	ChangeMB2U(0xF722, 0x00E469);
	ChangeMB2U(0xF723, 0x00E46A);
	ChangeMB2U(0xF724, 0x00E46B);
	ChangeMB2U(0xF725, 0x00E46C);
	ChangeMB2U(0xF726, 0x00E46D);
	ChangeMB2U(0xF727, 0x00E46E);
	ChangeMB2U(0xF728, 0x00E46F);
	ChangeMB2U(0xF729, 0x00E470);
	ChangeMB2U(0xF72A, 0x00E471);
	ChangeMB2U(0xF72B, 0x00E472);
	ChangeMB2U(0xF72C, 0x00E473);
	ChangeMB2U(0xF72D, 0x00E474);
	ChangeMB2U(0xF72E, 0x00E475);
	ChangeMB2U(0xF72F, 0x00E476);
	ChangeMB2U(0xF730, 0x00E477);
	ChangeMB2U(0xF731, 0x00E478);
	ChangeMB2U(0xF732, 0x00E479);
	ChangeMB2U(0xF733, 0x00E47A);
	ChangeMB2U(0xF734, 0x00E47B);
	ChangeMB2U(0xF735, 0x00E47C);
	ChangeMB2U(0xF736, 0x00E47D);
	ChangeMB2U(0xF737, 0x00E47E);
	ChangeMB2U(0xF738, 0x00E47F);
	ChangeMB2U(0xF739, 0x00E480);
	ChangeMB2U(0xF73A, 0x00E481);
	ChangeMB2U(0xF73B, 0x00E482);
	ChangeMB2U(0xF73C, 0x00E483);
	ChangeMB2U(0xF73D, 0x00E484);
	ChangeMB2U(0xF73E, 0x00E485);
	ChangeMB2U(0xF73F, 0x00E486);
	ChangeMB2U(0xF740, 0x00E487);
	ChangeMB2U(0xF741, 0x00E488);
	ChangeMB2U(0xF742, 0x00E489);
	ChangeMB2U(0xF743, 0x00E48A);
	ChangeMB2U(0xF744, 0x00E48B);
	ChangeMB2U(0xF745, 0x00E48C);
	ChangeMB2U(0xF746, 0x00E48D);
	ChangeMB2U(0xF747, 0x00E48E);
	ChangeMB2U(0xF748, 0x00E48F);
	ChangeMB2U(0xF749, 0x00E490);
	ChangeMB2U(0xF74A, 0x00E491);
	ChangeMB2U(0xF74B, 0x00E492);
	ChangeMB2U(0xF74C, 0x00E493);
	ChangeMB2U(0xF74D, 0x00E494);
	ChangeMB2U(0xF74E, 0x00E495);
	ChangeMB2U(0xF74F, 0x00E496);
	ChangeMB2U(0xF750, 0x00E497);
	ChangeMB2U(0xF751, 0x00E498);
	ChangeMB2U(0xF752, 0x00E499);
	ChangeMB2U(0xF753, 0x00E49A);
	ChangeMB2U(0xF754, 0x00E49B);
	ChangeMB2U(0xF755, 0x00E49C);
	ChangeMB2U(0xF756, 0x00E49D);
	ChangeMB2U(0xF757, 0x00E49E);
	ChangeMB2U(0xF758, 0x00E49F);
	ChangeMB2U(0xF759, 0x00E4A0);
	ChangeMB2U(0xF75A, 0x00E4A1);
	ChangeMB2U(0xF75B, 0x00E4A2);
	ChangeMB2U(0xF75C, 0x00E4A3);
	ChangeMB2U(0xF75D, 0x00E4A4);
	ChangeMB2U(0xF75E, 0x00E4A5);
	ChangeMB2U(0xF75F, 0x00E4A6);
	ChangeMB2U(0xF760, 0x00E4A7);
	ChangeMB2U(0xF761, 0x00E4A8);
	ChangeMB2U(0xF762, 0x00E4A9);
	ChangeMB2U(0xF763, 0x00E4AA);
	ChangeMB2U(0xF764, 0x00E4AB);
	ChangeMB2U(0xF765, 0x00E4AC);
	ChangeMB2U(0xF766, 0x00E4AD);
	ChangeMB2U(0xF767, 0x00E4AE);
	ChangeMB2U(0xF768, 0x00E4AF);
	ChangeMB2U(0xF769, 0x00E4B0);
	ChangeMB2U(0xF76A, 0x00E4B1);
	ChangeMB2U(0xF76B, 0x00E4B2);
	ChangeMB2U(0xF76C, 0x00E4B3);
	ChangeMB2U(0xF76D, 0x00E4B4);
	ChangeMB2U(0xF76E, 0x00E4B5);
	ChangeMB2U(0xF76F, 0x00E4B6);
	ChangeMB2U(0xF770, 0x00E4B7);
	ChangeMB2U(0xF771, 0x00E4B8);
	ChangeMB2U(0xF772, 0x00E4B9);
	ChangeMB2U(0xF773, 0x00E4BA);
	ChangeMB2U(0xF774, 0x00E4BB);
	ChangeMB2U(0xF775, 0x00E4BC);
	ChangeMB2U(0xF776, 0x00E4BD);
	ChangeMB2U(0xF777, 0x00E4BE);
	ChangeMB2U(0xF778, 0x00E4BF);
	ChangeMB2U(0xF779, 0x00E4C0);
	ChangeMB2U(0xF77A, 0x00E4C1);
	ChangeMB2U(0xF77B, 0x00E4C2);
	ChangeMB2U(0xF77C, 0x00E4C3);
	ChangeMB2U(0xF77D, 0x00E4C4);
	ChangeMB2U(0xF77E, 0x00E4C5);
	ChangeMB2U(0xF7A1, 0x00E0BC);
	ChangeMB2U(0xF7A2, 0x00E0BD);
	ChangeMB2U(0xF7A3, 0x00E0BE);
	ChangeMB2U(0xF7A4, 0x00E0BF);
	ChangeMB2U(0xF7A5, 0x00E0C0);
	ChangeMB2U(0xF7A6, 0x00E0C1);
	ChangeMB2U(0xF7A7, 0x00E0C2);
	ChangeMB2U(0xF7A8, 0x00E0C3);
	ChangeMB2U(0xF7A9, 0x00E0C4);
	ChangeMB2U(0xF7AA, 0x00E0C5);
	ChangeMB2U(0xF7AB, 0x00E0C6);
	ChangeMB2U(0xF7AC, 0x00E0C7);
	ChangeMB2U(0xF7AD, 0x00E0C8);
	ChangeMB2U(0xF7AE, 0x00E0C9);
	ChangeMB2U(0xF7AF, 0x00E0CA);
	ChangeMB2U(0xF7B0, 0x00E0CB);
	ChangeMB2U(0xF7B1, 0x00E0CC);
	ChangeMB2U(0xF7B2, 0x00E0CD);
	ChangeMB2U(0xF7B3, 0x00E0CE);
	ChangeMB2U(0xF7B4, 0x00E0CF);
	ChangeMB2U(0xF7B5, 0x00E0D0);
	ChangeMB2U(0xF7B6, 0x00E0D1);
	ChangeMB2U(0xF7B7, 0x00E0D2);
	ChangeMB2U(0xF7B8, 0x00E0D3);
	ChangeMB2U(0xF7B9, 0x00E0D4);
	ChangeMB2U(0xF7BA, 0x00E0D5);
	ChangeMB2U(0xF7BB, 0x00E0D6);
	ChangeMB2U(0xF7BC, 0x00E0D7);
	ChangeMB2U(0xF7BD, 0x00E0D8);
	ChangeMB2U(0xF7BE, 0x00E0D9);
	ChangeMB2U(0xF7BF, 0x00E0DA);
	ChangeMB2U(0xF7C0, 0x00E0DB);
	ChangeMB2U(0xF7C1, 0x00E0DC);
	ChangeMB2U(0xF7C2, 0x00E0DD);
	ChangeMB2U(0xF7C3, 0x00E0DE);
	ChangeMB2U(0xF7C4, 0x00E0DF);
	ChangeMB2U(0xF7C5, 0x00E0E0);
	ChangeMB2U(0xF7C6, 0x00E0E1);
	ChangeMB2U(0xF7C7, 0x00E0E2);
	ChangeMB2U(0xF7C8, 0x00E0E3);
	ChangeMB2U(0xF7C9, 0x00E0E4);
	ChangeMB2U(0xF7CA, 0x00E0E5);
	ChangeMB2U(0xF7CB, 0x00E0E6);
	ChangeMB2U(0xF7CC, 0x00E0E7);
	ChangeMB2U(0xF7CD, 0x00E0E8);
	ChangeMB2U(0xF7CE, 0x00E0E9);
	ChangeMB2U(0xF7CF, 0x00E0EA);
	ChangeMB2U(0xF7D0, 0x00E0EB);
	ChangeMB2U(0xF7D1, 0x00E0EC);
	ChangeMB2U(0xF7D2, 0x00E0ED);
	ChangeMB2U(0xF7D3, 0x00E0EE);
	ChangeMB2U(0xF7D4, 0x00E0EF);
	ChangeMB2U(0xF7D5, 0x00E0F0);
	ChangeMB2U(0xF7D6, 0x00E0F1);
	ChangeMB2U(0xF7D7, 0x00E0F2);
	ChangeMB2U(0xF7D8, 0x00E0F3);
	ChangeMB2U(0xF7D9, 0x00E0F4);
	ChangeMB2U(0xF7DA, 0x00E0F5);
	ChangeMB2U(0xF7DB, 0x00E0F6);
	ChangeMB2U(0xF7DC, 0x00E0F7);
	ChangeMB2U(0xF7DD, 0x00E0F8);
	ChangeMB2U(0xF7DE, 0x00E0F9);
	ChangeMB2U(0xF7DF, 0x00E0FA);
	ChangeMB2U(0xF7E0, 0x00E0FB);
	ChangeMB2U(0xF7E1, 0x00E0FC);
	ChangeMB2U(0xF7E2, 0x00E0FD);
	ChangeMB2U(0xF7E3, 0x00E0FE);
	ChangeMB2U(0xF7E4, 0x00E0FF);
	ChangeMB2U(0xF7E5, 0x00E100);
	ChangeMB2U(0xF7E6, 0x00E101);
	ChangeMB2U(0xF7E7, 0x00E102);
	ChangeMB2U(0xF7E8, 0x00E103);
	ChangeMB2U(0xF7E9, 0x00E104);
	ChangeMB2U(0xF7EA, 0x00E105);
	ChangeMB2U(0xF7EB, 0x00E106);
	ChangeMB2U(0xF7EC, 0x00E107);
	ChangeMB2U(0xF7ED, 0x00E108);
	ChangeMB2U(0xF7EE, 0x00E109);
	ChangeMB2U(0xF7EF, 0x00E10A);
	ChangeMB2U(0xF7F0, 0x00E10B);
	ChangeMB2U(0xF7F1, 0x00E10C);
	ChangeMB2U(0xF7F2, 0x00E10D);
	ChangeMB2U(0xF7F3, 0x00E10E);
	ChangeMB2U(0xF7F4, 0x00E10F);
	ChangeMB2U(0xF7F5, 0x00E110);
	ChangeMB2U(0xF7F6, 0x00E111);
	ChangeMB2U(0xF7F7, 0x00E112);
	ChangeMB2U(0xF7F8, 0x00E113);
	ChangeMB2U(0xF7F9, 0x00E114);
	ChangeMB2U(0xF7FA, 0x00E115);
	ChangeMB2U(0xF7FB, 0x00E116);
	ChangeMB2U(0xF7FC, 0x00E117);
	ChangeMB2U(0xF7FD, 0x00E118);
	ChangeMB2U(0xF7FE, 0x00E119);
	ChangeMB2U(0xF821, 0x00E4C6);
	ChangeMB2U(0xF822, 0x00E4C7);
	ChangeMB2U(0xF823, 0x00E4C8);
	ChangeMB2U(0xF824, 0x00E4C9);
	ChangeMB2U(0xF825, 0x00E4CA);
	ChangeMB2U(0xF826, 0x00E4CB);
	ChangeMB2U(0xF827, 0x00E4CC);
	ChangeMB2U(0xF828, 0x00E4CD);
	ChangeMB2U(0xF829, 0x00E4CE);
	ChangeMB2U(0xF82A, 0x00E4CF);
	ChangeMB2U(0xF82B, 0x00E4D0);
	ChangeMB2U(0xF82C, 0x00E4D1);
	ChangeMB2U(0xF82D, 0x00E4D2);
	ChangeMB2U(0xF82E, 0x00E4D3);
	ChangeMB2U(0xF82F, 0x00E4D4);
	ChangeMB2U(0xF830, 0x00E4D5);
	ChangeMB2U(0xF831, 0x00E4D6);
	ChangeMB2U(0xF832, 0x00E4D7);
	ChangeMB2U(0xF833, 0x00E4D8);
	ChangeMB2U(0xF834, 0x00E4D9);
	ChangeMB2U(0xF835, 0x00E4DA);
	ChangeMB2U(0xF836, 0x00E4DB);
	ChangeMB2U(0xF837, 0x00E4DC);
	ChangeMB2U(0xF838, 0x00E4DD);
	ChangeMB2U(0xF839, 0x00E4DE);
	ChangeMB2U(0xF83A, 0x00E4DF);
	ChangeMB2U(0xF83B, 0x00E4E0);
	ChangeMB2U(0xF83C, 0x00E4E1);
	ChangeMB2U(0xF83D, 0x00E4E2);
	ChangeMB2U(0xF83E, 0x00E4E3);
	ChangeMB2U(0xF83F, 0x00E4E4);
	ChangeMB2U(0xF840, 0x00E4E5);
	ChangeMB2U(0xF841, 0x00E4E6);
	ChangeMB2U(0xF842, 0x00E4E7);
	ChangeMB2U(0xF843, 0x00E4E8);
	ChangeMB2U(0xF844, 0x00E4E9);
	ChangeMB2U(0xF845, 0x00E4EA);
	ChangeMB2U(0xF846, 0x00E4EB);
	ChangeMB2U(0xF847, 0x00E4EC);
	ChangeMB2U(0xF848, 0x00E4ED);
	ChangeMB2U(0xF849, 0x00E4EE);
	ChangeMB2U(0xF84A, 0x00E4EF);
	ChangeMB2U(0xF84B, 0x00E4F0);
	ChangeMB2U(0xF84C, 0x00E4F1);
	ChangeMB2U(0xF84D, 0x00E4F2);
	ChangeMB2U(0xF84E, 0x00E4F3);
	ChangeMB2U(0xF84F, 0x00E4F4);
	ChangeMB2U(0xF850, 0x00E4F5);
	ChangeMB2U(0xF851, 0x00E4F6);
	ChangeMB2U(0xF852, 0x00E4F7);
	ChangeMB2U(0xF853, 0x00E4F8);
	ChangeMB2U(0xF854, 0x00E4F9);
	ChangeMB2U(0xF855, 0x00E4FA);
	ChangeMB2U(0xF856, 0x00E4FB);
	ChangeMB2U(0xF857, 0x00E4FC);
	ChangeMB2U(0xF858, 0x00E4FD);
	ChangeMB2U(0xF859, 0x00E4FE);
	ChangeMB2U(0xF85A, 0x00E4FF);
	ChangeMB2U(0xF85B, 0x00E500);
	ChangeMB2U(0xF85C, 0x00E501);
	ChangeMB2U(0xF85D, 0x00E502);
	ChangeMB2U(0xF85E, 0x00E503);
	ChangeMB2U(0xF85F, 0x00E504);
	ChangeMB2U(0xF860, 0x00E505);
	ChangeMB2U(0xF861, 0x00E506);
	ChangeMB2U(0xF862, 0x00E507);
	ChangeMB2U(0xF863, 0x00E508);
	ChangeMB2U(0xF864, 0x00E509);
	ChangeMB2U(0xF865, 0x00E50A);
	ChangeMB2U(0xF866, 0x00E50B);
	ChangeMB2U(0xF867, 0x00E50C);
	ChangeMB2U(0xF868, 0x00E50D);
	ChangeMB2U(0xF869, 0x00E50E);
	ChangeMB2U(0xF86A, 0x00E50F);
	ChangeMB2U(0xF86B, 0x00E510);
	ChangeMB2U(0xF86C, 0x00E511);
	ChangeMB2U(0xF86D, 0x00E512);
	ChangeMB2U(0xF86E, 0x00E513);
	ChangeMB2U(0xF86F, 0x00E514);
	ChangeMB2U(0xF870, 0x00E515);
	ChangeMB2U(0xF871, 0x00E516);
	ChangeMB2U(0xF872, 0x00E517);
	ChangeMB2U(0xF873, 0x00E518);
	ChangeMB2U(0xF874, 0x00E519);
	ChangeMB2U(0xF875, 0x00E51A);
	ChangeMB2U(0xF876, 0x00E51B);
	ChangeMB2U(0xF877, 0x00E51C);
	ChangeMB2U(0xF878, 0x00E51D);
	ChangeMB2U(0xF879, 0x00E51E);
	ChangeMB2U(0xF87A, 0x00E51F);
	ChangeMB2U(0xF87B, 0x00E520);
	ChangeMB2U(0xF87C, 0x00E521);
	ChangeMB2U(0xF87D, 0x00E522);
	ChangeMB2U(0xF87E, 0x00E523);
	ChangeMB2U(0xF8A1, 0x00E11A);
	ChangeMB2U(0xF8A2, 0x00E11B);
	ChangeMB2U(0xF8A3, 0x00E11C);
	ChangeMB2U(0xF8A4, 0x00E11D);
	ChangeMB2U(0xF8A5, 0x00E11E);
	ChangeMB2U(0xF8A6, 0x00E11F);
	ChangeMB2U(0xF8A7, 0x00E120);
	ChangeMB2U(0xF8A8, 0x00E121);
	ChangeMB2U(0xF8A9, 0x00E122);
	ChangeMB2U(0xF8AA, 0x00E123);
	ChangeMB2U(0xF8AB, 0x00E124);
	ChangeMB2U(0xF8AC, 0x00E125);
	ChangeMB2U(0xF8AD, 0x00E126);
	ChangeMB2U(0xF8AE, 0x00E127);
	ChangeMB2U(0xF8AF, 0x00E128);
	ChangeMB2U(0xF8B0, 0x00E129);
	ChangeMB2U(0xF8B1, 0x00E12A);
	ChangeMB2U(0xF8B2, 0x00E12B);
	ChangeMB2U(0xF8B3, 0x00E12C);
	ChangeMB2U(0xF8B4, 0x00E12D);
	ChangeMB2U(0xF8B5, 0x00E12E);
	ChangeMB2U(0xF8B6, 0x00E12F);
	ChangeMB2U(0xF8B7, 0x00E130);
	ChangeMB2U(0xF8B8, 0x00E131);
	ChangeMB2U(0xF8B9, 0x00E132);
	ChangeMB2U(0xF8BA, 0x00E133);
	ChangeMB2U(0xF8BB, 0x00E134);
	ChangeMB2U(0xF8BC, 0x00E135);
	ChangeMB2U(0xF8BD, 0x00E136);
	ChangeMB2U(0xF8BE, 0x00E137);
	ChangeMB2U(0xF8BF, 0x00E138);
	ChangeMB2U(0xF8C0, 0x00E139);
	ChangeMB2U(0xF8C1, 0x00E13A);
	ChangeMB2U(0xF8C2, 0x00E13B);
	ChangeMB2U(0xF8C3, 0x00E13C);
	ChangeMB2U(0xF8C4, 0x00E13D);
	ChangeMB2U(0xF8C5, 0x00E13E);
	ChangeMB2U(0xF8C6, 0x00E13F);
	ChangeMB2U(0xF8C7, 0x00E140);
	ChangeMB2U(0xF8C8, 0x00E141);
	ChangeMB2U(0xF8C9, 0x00E142);
	ChangeMB2U(0xF8CA, 0x00E143);
	ChangeMB2U(0xF8CB, 0x00E144);
	ChangeMB2U(0xF8CC, 0x00E145);
	ChangeMB2U(0xF8CD, 0x00E146);
	ChangeMB2U(0xF8CE, 0x00E147);
	ChangeMB2U(0xF8CF, 0x00E148);
	ChangeMB2U(0xF8D0, 0x00E149);
	ChangeMB2U(0xF8D1, 0x00E14A);
	ChangeMB2U(0xF8D2, 0x00E14B);
	ChangeMB2U(0xF8D3, 0x00E14C);
	ChangeMB2U(0xF8D4, 0x00E14D);
	ChangeMB2U(0xF8D5, 0x00E14E);
	ChangeMB2U(0xF8D6, 0x00E14F);
	ChangeMB2U(0xF8D7, 0x00E150);
	ChangeMB2U(0xF8D8, 0x00E151);
	ChangeMB2U(0xF8D9, 0x00E152);
	ChangeMB2U(0xF8DA, 0x00E153);
	ChangeMB2U(0xF8DB, 0x00E154);
	ChangeMB2U(0xF8DC, 0x00E155);
	ChangeMB2U(0xF8DD, 0x00E156);
	ChangeMB2U(0xF8DE, 0x00E157);
	ChangeMB2U(0xF8DF, 0x00E158);
	ChangeMB2U(0xF8E0, 0x00E159);
	ChangeMB2U(0xF8E1, 0x00E15A);
	ChangeMB2U(0xF8E2, 0x00E15B);
	ChangeMB2U(0xF8E3, 0x00E15C);
	ChangeMB2U(0xF8E4, 0x00E15D);
	ChangeMB2U(0xF8E5, 0x00E15E);
	ChangeMB2U(0xF8E6, 0x00E15F);
	ChangeMB2U(0xF8E7, 0x00E160);
	ChangeMB2U(0xF8E8, 0x00E161);
	ChangeMB2U(0xF8E9, 0x00E162);
	ChangeMB2U(0xF8EA, 0x00E163);
	ChangeMB2U(0xF8EB, 0x00E164);
	ChangeMB2U(0xF8EC, 0x00E165);
	ChangeMB2U(0xF8ED, 0x00E166);
	ChangeMB2U(0xF8EE, 0x00E167);
	ChangeMB2U(0xF8EF, 0x00E168);
	ChangeMB2U(0xF8F0, 0x00E169);
	ChangeMB2U(0xF8F1, 0x00E16A);
	ChangeMB2U(0xF8F2, 0x00E16B);
	ChangeMB2U(0xF8F3, 0x00E16C);
	ChangeMB2U(0xF8F4, 0x00E16D);
	ChangeMB2U(0xF8F5, 0x00E16E);
	ChangeMB2U(0xF8F6, 0x00E16F);
	ChangeMB2U(0xF8F7, 0x00E170);
	ChangeMB2U(0xF8F8, 0x00E171);
	ChangeMB2U(0xF8F9, 0x00E172);
	ChangeMB2U(0xF8FA, 0x00E173);
	ChangeMB2U(0xF8FB, 0x00E174);
	ChangeMB2U(0xF8FC, 0x00E175);
	ChangeMB2U(0xF8FD, 0x00E176);
	ChangeMB2U(0xF8FE, 0x00E177);
	ChangeMB2U(0xF921, 0x00E524);
	ChangeMB2U(0xF922, 0x00E525);
	ChangeMB2U(0xF923, 0x00E526);
	ChangeMB2U(0xF924, 0x00E527);
	ChangeMB2U(0xF925, 0x00E528);
	ChangeMB2U(0xF926, 0x00E529);
	ChangeMB2U(0xF927, 0x00E52A);
	ChangeMB2U(0xF928, 0x00E52B);
	ChangeMB2U(0xF929, 0x00E52C);
	ChangeMB2U(0xF92A, 0x00E52D);
	ChangeMB2U(0xF92B, 0x00E52E);
	ChangeMB2U(0xF92C, 0x00E52F);
	ChangeMB2U(0xF92D, 0x00E530);
	ChangeMB2U(0xF92E, 0x00E531);
	ChangeMB2U(0xF92F, 0x00E532);
	ChangeMB2U(0xF930, 0x00E533);
	ChangeMB2U(0xF931, 0x00E534);
	ChangeMB2U(0xF932, 0x00E535);
	ChangeMB2U(0xF933, 0x00E536);
	ChangeMB2U(0xF934, 0x00E537);
	ChangeMB2U(0xF935, 0x00E538);
	ChangeMB2U(0xF936, 0x00E539);
	ChangeMB2U(0xF937, 0x00E53A);
	ChangeMB2U(0xF938, 0x00E53B);
	ChangeMB2U(0xF939, 0x00E53C);
	ChangeMB2U(0xF93A, 0x00E53D);
	ChangeMB2U(0xF93B, 0x00E53E);
	ChangeMB2U(0xF93C, 0x00E53F);
	ChangeMB2U(0xF93D, 0x00E540);
	ChangeMB2U(0xF93E, 0x00E541);
	ChangeMB2U(0xF93F, 0x00E542);
	ChangeMB2U(0xF940, 0x00E543);
	ChangeMB2U(0xF941, 0x00E544);
	ChangeMB2U(0xF942, 0x00E545);
	ChangeMB2U(0xF943, 0x00E546);
	ChangeMB2U(0xF944, 0x00E547);
	ChangeMB2U(0xF945, 0x00E548);
	ChangeMB2U(0xF946, 0x00E549);
	ChangeMB2U(0xF947, 0x00E54A);
	ChangeMB2U(0xF948, 0x00E54B);
	ChangeMB2U(0xF949, 0x00E54C);
	ChangeMB2U(0xF94A, 0x00E54D);
	ChangeMB2U(0xF94B, 0x00E54E);
	ChangeMB2U(0xF94C, 0x00E54F);
	ChangeMB2U(0xF94D, 0x00E550);
	ChangeMB2U(0xF94E, 0x00E551);
	ChangeMB2U(0xF94F, 0x00E552);
	ChangeMB2U(0xF950, 0x00E553);
	ChangeMB2U(0xF951, 0x00E554);
	ChangeMB2U(0xF952, 0x00E555);
	ChangeMB2U(0xF953, 0x00E556);
	ChangeMB2U(0xF954, 0x00E557);
	ChangeMB2U(0xF955, 0x00E558);
	ChangeMB2U(0xF956, 0x00E559);
	ChangeMB2U(0xF957, 0x00E55A);
	ChangeMB2U(0xF958, 0x00E55B);
	ChangeMB2U(0xF959, 0x00E55C);
	ChangeMB2U(0xF95A, 0x00E55D);
	ChangeMB2U(0xF95B, 0x00E55E);
	ChangeMB2U(0xF95C, 0x00E55F);
	ChangeMB2U(0xF95D, 0x00E560);
	ChangeMB2U(0xF95E, 0x00E561);
	ChangeMB2U(0xF95F, 0x00E562);
	ChangeMB2U(0xF960, 0x00E563);
	ChangeMB2U(0xF961, 0x00E564);
	ChangeMB2U(0xF962, 0x00E565);
	ChangeMB2U(0xF963, 0x00E566);
	ChangeMB2U(0xF964, 0x00E567);
	ChangeMB2U(0xF965, 0x00E568);
	ChangeMB2U(0xF966, 0x00E569);
	ChangeMB2U(0xF967, 0x00E56A);
	ChangeMB2U(0xF968, 0x00E56B);
	ChangeMB2U(0xF969, 0x00E56C);
	ChangeMB2U(0xF96A, 0x00E56D);
	ChangeMB2U(0xF96B, 0x00E56E);
	ChangeMB2U(0xF96C, 0x00E56F);
	ChangeMB2U(0xF96D, 0x00E570);
	ChangeMB2U(0xF96E, 0x00E571);
	ChangeMB2U(0xF96F, 0x00E572);
	ChangeMB2U(0xF970, 0x00E573);
	ChangeMB2U(0xF971, 0x00E574);
	ChangeMB2U(0xF972, 0x00E575);
	ChangeMB2U(0xF973, 0x00E576);
	ChangeMB2U(0xF974, 0x00E577);
	ChangeMB2U(0xF975, 0x00E578);
	ChangeMB2U(0xF976, 0x00E579);
	ChangeMB2U(0xF977, 0x00E57A);
	ChangeMB2U(0xF978, 0x00E57B);
	ChangeMB2U(0xF979, 0x00E57C);
	ChangeMB2U(0xF97A, 0x00E57D);
	ChangeMB2U(0xF97B, 0x00E57E);
	ChangeMB2U(0xF97C, 0x00E57F);
	ChangeMB2U(0xF97D, 0x00E580);
	ChangeMB2U(0xF97E, 0x00E581);
	ChangeMB2U(0xF9A1, 0x00E178);
	ChangeMB2U(0xF9A2, 0x00E179);
	ChangeMB2U(0xF9A3, 0x00E17A);
	ChangeMB2U(0xF9A4, 0x00E17B);
	ChangeMB2U(0xF9A5, 0x00E17C);
	ChangeMB2U(0xF9A6, 0x00E17D);
	ChangeMB2U(0xF9A7, 0x00E17E);
	ChangeMB2U(0xF9A8, 0x00E17F);
	ChangeMB2U(0xF9A9, 0x00E180);
	ChangeMB2U(0xF9AA, 0x00E181);
	ChangeMB2U(0xF9AB, 0x00E182);
	ChangeMB2U(0xF9AC, 0x00E183);
	ChangeMB2U(0xF9AD, 0x00E184);
	ChangeMB2U(0xF9AE, 0x00E185);
	ChangeMB2U(0xF9AF, 0x00E186);
	ChangeMB2U(0xF9B0, 0x00E187);
	ChangeMB2U(0xF9B1, 0x00E188);
	ChangeMB2U(0xF9B2, 0x00E189);
	ChangeMB2U(0xF9B3, 0x00E18A);
	ChangeMB2U(0xF9B4, 0x00E18B);
	ChangeMB2U(0xF9B5, 0x00E18C);
	ChangeMB2U(0xF9B6, 0x00E18D);
	ChangeMB2U(0xF9B7, 0x00E18E);
	ChangeMB2U(0xF9B8, 0x00E18F);
	ChangeMB2U(0xF9B9, 0x00E190);
	ChangeMB2U(0xF9BA, 0x00E191);
	ChangeMB2U(0xF9BB, 0x00E192);
	ChangeMB2U(0xF9BC, 0x00E193);
	ChangeMB2U(0xF9BD, 0x00E194);
	ChangeMB2U(0xF9BE, 0x00E195);
	ChangeMB2U(0xF9BF, 0x00E196);
	ChangeMB2U(0xF9C0, 0x00E197);
	ChangeMB2U(0xF9C1, 0x00E198);
	ChangeMB2U(0xF9C2, 0x00E199);
	ChangeMB2U(0xF9C3, 0x00E19A);
	ChangeMB2U(0xF9C4, 0x00E19B);
	ChangeMB2U(0xF9C5, 0x00E19C);
	ChangeMB2U(0xF9C6, 0x00E19D);
	ChangeMB2U(0xF9C7, 0x00E19E);
	ChangeMB2U(0xF9C8, 0x00E19F);
	ChangeMB2U(0xF9C9, 0x00E1A0);
	ChangeMB2U(0xF9CA, 0x00E1A1);
	ChangeMB2U(0xF9CB, 0x00E1A2);
	ChangeMB2U(0xF9CC, 0x00E1A3);
	ChangeMB2U(0xF9CD, 0x00E1A4);
	ChangeMB2U(0xF9CE, 0x00E1A5);
	ChangeMB2U(0xF9CF, 0x00E1A6);
	ChangeMB2U(0xF9D0, 0x00E1A7);
	ChangeMB2U(0xF9D1, 0x00E1A8);
	ChangeMB2U(0xF9D2, 0x00E1A9);
	ChangeMB2U(0xF9D3, 0x00E1AA);
	ChangeMB2U(0xF9D4, 0x00E1AB);
	ChangeMB2U(0xF9D5, 0x00E1AC);
	ChangeMB2U(0xF9D6, 0x00E1AD);
	ChangeMB2U(0xF9D7, 0x00E1AE);
	ChangeMB2U(0xF9D8, 0x00E1AF);
	ChangeMB2U(0xF9D9, 0x00E1B0);
	ChangeMB2U(0xF9DA, 0x00E1B1);
	ChangeMB2U(0xF9DB, 0x00E1B2);
	ChangeMB2U(0xF9DC, 0x00E1B3);
	ChangeMB2U(0xF9DD, 0x00E1B4);
	ChangeMB2U(0xF9DE, 0x00E1B5);
	ChangeMB2U(0xF9DF, 0x00E1B6);
	ChangeMB2U(0xF9E0, 0x00E1B7);
	ChangeMB2U(0xF9E1, 0x00E1B8);
	ChangeMB2U(0xF9E2, 0x00E1B9);
	ChangeMB2U(0xF9E3, 0x00E1BA);
	ChangeMB2U(0xF9E4, 0x00E1BB);
	ChangeMB2U(0xF9E5, 0x00E1BC);
	ChangeMB2U(0xF9E6, 0x00E1BD);
	ChangeMB2U(0xF9E7, 0x00E1BE);
	ChangeMB2U(0xF9E8, 0x00E1BF);
	ChangeMB2U(0xF9E9, 0x00E1C0);
	ChangeMB2U(0xF9EA, 0x00E1C1);
	ChangeMB2U(0xF9EB, 0x00E1C2);
	ChangeMB2U(0xF9EC, 0x00E1C3);
	ChangeMB2U(0xF9ED, 0x00E1C4);
	ChangeMB2U(0xF9EE, 0x00E1C5);
	ChangeMB2U(0xF9EF, 0x00E1C6);
	ChangeMB2U(0xF9F0, 0x00E1C7);
	ChangeMB2U(0xF9F1, 0x00E1C8);
	ChangeMB2U(0xF9F2, 0x00E1C9);
	ChangeMB2U(0xF9F3, 0x00E1CA);
	ChangeMB2U(0xF9F4, 0x00E1CB);
	ChangeMB2U(0xF9F5, 0x00E1CC);
	ChangeMB2U(0xF9F6, 0x00E1CD);
	ChangeMB2U(0xF9F7, 0x00E1CE);
	ChangeMB2U(0xF9F8, 0x00E1CF);
	ChangeMB2U(0xF9F9, 0x00E1D0);
	ChangeMB2U(0xF9FA, 0x00E1D1);
	ChangeMB2U(0xF9FB, 0x00E1D2);
	ChangeMB2U(0xF9FC, 0x00E1D3);
	ChangeMB2U(0xF9FD, 0x00E1D4);
	ChangeMB2U(0xF9FE, 0x00E1D5);
	ChangeMB2U(0xFA21, 0x00E582);
	ChangeMB2U(0xFA22, 0x00E583);
	ChangeMB2U(0xFA23, 0x00E584);
	ChangeMB2U(0xFA24, 0x00E585);
	ChangeMB2U(0xFA25, 0x00E586);
	ChangeMB2U(0xFA26, 0x00E587);
	ChangeMB2U(0xFA27, 0x00E588);
	ChangeMB2U(0xFA28, 0x00E589);
	ChangeMB2U(0xFA29, 0x00E58A);
	ChangeMB2U(0xFA2A, 0x00E58B);
	ChangeMB2U(0xFA2B, 0x00E58C);
	ChangeMB2U(0xFA2C, 0x00E58D);
	ChangeMB2U(0xFA2D, 0x00E58E);
	ChangeMB2U(0xFA2E, 0x00E58F);
	ChangeMB2U(0xFA2F, 0x00E590);
	ChangeMB2U(0xFA30, 0x00E591);
	ChangeMB2U(0xFA31, 0x00E592);
	ChangeMB2U(0xFA32, 0x00E593);
	ChangeMB2U(0xFA33, 0x00E594);
	ChangeMB2U(0xFA34, 0x00E595);
	ChangeMB2U(0xFA35, 0x00E596);
	ChangeMB2U(0xFA36, 0x00E597);
	ChangeMB2U(0xFA37, 0x00E598);
	ChangeMB2U(0xFA38, 0x00E599);
	ChangeMB2U(0xFA39, 0x00E59A);
	ChangeMB2U(0xFA3A, 0x00E59B);
	ChangeMB2U(0xFA3B, 0x00E59C);
	ChangeMB2U(0xFA3C, 0x00E59D);
	ChangeMB2U(0xFA3D, 0x00E59E);
	ChangeMB2U(0xFA3E, 0x00E59F);
	ChangeMB2U(0xFA3F, 0x00E5A0);
	ChangeMB2U(0xFA40, 0x00E5A1);
	ChangeMB2U(0xFA41, 0x00E5A2);
	ChangeMB2U(0xFA42, 0x00E5A3);
	ChangeMB2U(0xFA43, 0x00E5A4);
	ChangeMB2U(0xFA44, 0x00E5A5);
	ChangeMB2U(0xFA45, 0x00E5A6);
	ChangeMB2U(0xFA46, 0x00E5A7);
	ChangeMB2U(0xFA47, 0x00E5A8);
	ChangeMB2U(0xFA48, 0x00E5A9);
	ChangeMB2U(0xFA49, 0x00E5AA);
	ChangeMB2U(0xFA4A, 0x00E5AB);
	ChangeMB2U(0xFA4B, 0x00E5AC);
	ChangeMB2U(0xFA4C, 0x00E5AD);
	ChangeMB2U(0xFA4D, 0x00E5AE);
	ChangeMB2U(0xFA4E, 0x00E5AF);
	ChangeMB2U(0xFA4F, 0x00E5B0);
	ChangeMB2U(0xFA50, 0x00E5B1);
	ChangeMB2U(0xFA51, 0x00E5B2);
	ChangeMB2U(0xFA52, 0x00E5B3);
	ChangeMB2U(0xFA53, 0x00E5B4);
	ChangeMB2U(0xFA54, 0x00E5B5);
	ChangeMB2U(0xFA55, 0x00E5B6);
	ChangeMB2U(0xFA56, 0x00E5B7);
	ChangeMB2U(0xFA57, 0x00E5B8);
	ChangeMB2U(0xFA58, 0x00E5B9);
	ChangeMB2U(0xFA59, 0x00E5BA);
	ChangeMB2U(0xFA5A, 0x00E5BB);
	ChangeMB2U(0xFA5B, 0x00E5BC);
	ChangeMB2U(0xFA5C, 0x00E5BD);
	ChangeMB2U(0xFA5D, 0x00E5BE);
	ChangeMB2U(0xFA5E, 0x00E5BF);
	ChangeMB2U(0xFA5F, 0x00E5C0);
	ChangeMB2U(0xFA60, 0x00E5C1);
	ChangeMB2U(0xFA61, 0x00E5C2);
	ChangeMB2U(0xFA62, 0x00E5C3);
	ChangeMB2U(0xFA63, 0x00E5C4);
	ChangeMB2U(0xFA64, 0x00E5C5);
	ChangeMB2U(0xFA65, 0x00E5C6);
	ChangeMB2U(0xFA66, 0x00E5C7);
	ChangeMB2U(0xFA67, 0x00E5C8);
	ChangeMB2U(0xFA68, 0x00E5C9);
	ChangeMB2U(0xFA69, 0x00E5CA);
	ChangeMB2U(0xFA6A, 0x00E5CB);
	ChangeMB2U(0xFA6B, 0x00E5CC);
	ChangeMB2U(0xFA6C, 0x00E5CD);
	ChangeMB2U(0xFA6D, 0x00E5CE);
	ChangeMB2U(0xFA6E, 0x00E5CF);
	ChangeMB2U(0xFA6F, 0x00E5D0);
	ChangeMB2U(0xFA70, 0x00E5D1);
	ChangeMB2U(0xFA71, 0x00E5D2);
	ChangeMB2U(0xFA72, 0x00E5D3);
	ChangeMB2U(0xFA73, 0x00E5D4);
	ChangeMB2U(0xFA74, 0x00E5D5);
	ChangeMB2U(0xFA75, 0x00E5D6);
	ChangeMB2U(0xFA76, 0x00E5D7);
	ChangeMB2U(0xFA77, 0x00E5D8);
	ChangeMB2U(0xFA78, 0x00E5D9);
	ChangeMB2U(0xFA79, 0x00E5DA);
	ChangeMB2U(0xFA7A, 0x00E5DB);
	ChangeMB2U(0xFA7B, 0x00E5DC);
	ChangeMB2U(0xFA7C, 0x00E5DD);
	ChangeMB2U(0xFA7D, 0x00E5DE);
	ChangeMB2U(0xFA7E, 0x00E5DF);
	ChangeMB2U(0xFAA1, 0x00E1D6);
	ChangeMB2U(0xFAA2, 0x00E1D7);
	ChangeMB2U(0xFAA3, 0x00E1D8);
	ChangeMB2U(0xFAA4, 0x00E1D9);
	ChangeMB2U(0xFAA5, 0x00E1DA);
	ChangeMB2U(0xFAA6, 0x00E1DB);
	ChangeMB2U(0xFAA7, 0x00E1DC);
	ChangeMB2U(0xFAA8, 0x00E1DD);
	ChangeMB2U(0xFAA9, 0x00E1DE);
	ChangeMB2U(0xFAAA, 0x00E1DF);
	ChangeMB2U(0xFAAB, 0x00E1E0);
	ChangeMB2U(0xFAAC, 0x00E1E1);
	ChangeMB2U(0xFAAD, 0x00E1E2);
	ChangeMB2U(0xFAAE, 0x00E1E3);
	ChangeMB2U(0xFAAF, 0x00E1E4);
	ChangeMB2U(0xFAB0, 0x00E1E5);
	ChangeMB2U(0xFAB1, 0x00E1E6);
	ChangeMB2U(0xFAB2, 0x00E1E7);
	ChangeMB2U(0xFAB3, 0x00E1E8);
	ChangeMB2U(0xFAB4, 0x00E1E9);
	ChangeMB2U(0xFAB5, 0x00E1EA);
	ChangeMB2U(0xFAB6, 0x00E1EB);
	ChangeMB2U(0xFAB7, 0x00E1EC);
	ChangeMB2U(0xFAB8, 0x00E1ED);
	ChangeMB2U(0xFAB9, 0x00E1EE);
	ChangeMB2U(0xFABA, 0x00E1EF);
	ChangeMB2U(0xFABB, 0x00E1F0);
	ChangeMB2U(0xFABC, 0x00E1F1);
	ChangeMB2U(0xFABD, 0x00E1F2);
	ChangeMB2U(0xFABE, 0x00E1F3);
	ChangeMB2U(0xFABF, 0x00E1F4);
	ChangeMB2U(0xFAC0, 0x00E1F5);
	ChangeMB2U(0xFAC1, 0x00E1F6);
	ChangeMB2U(0xFAC2, 0x00E1F7);
	ChangeMB2U(0xFAC3, 0x00E1F8);
	ChangeMB2U(0xFAC4, 0x00E1F9);
	ChangeMB2U(0xFAC5, 0x00E1FA);
	ChangeMB2U(0xFAC6, 0x00E1FB);
	ChangeMB2U(0xFAC7, 0x00E1FC);
	ChangeMB2U(0xFAC8, 0x00E1FD);
	ChangeMB2U(0xFAC9, 0x00E1FE);
	ChangeMB2U(0xFACA, 0x00E1FF);
	ChangeMB2U(0xFACB, 0x00E200);
	ChangeMB2U(0xFACC, 0x00E201);
	ChangeMB2U(0xFACD, 0x00E202);
	ChangeMB2U(0xFACE, 0x00E203);
	ChangeMB2U(0xFACF, 0x00E204);
	ChangeMB2U(0xFAD0, 0x00E205);
	ChangeMB2U(0xFAD1, 0x00E206);
	ChangeMB2U(0xFAD2, 0x00E207);
	ChangeMB2U(0xFAD3, 0x00E208);
	ChangeMB2U(0xFAD4, 0x00E209);
	ChangeMB2U(0xFAD5, 0x00E20A);
	ChangeMB2U(0xFAD6, 0x00E20B);
	ChangeMB2U(0xFAD7, 0x00E20C);
	ChangeMB2U(0xFAD8, 0x00E20D);
	ChangeMB2U(0xFAD9, 0x00E20E);
	ChangeMB2U(0xFADA, 0x00E20F);
	ChangeMB2U(0xFADB, 0x00E210);
	ChangeMB2U(0xFADC, 0x00E211);
	ChangeMB2U(0xFADD, 0x00E212);
	ChangeMB2U(0xFADE, 0x00E213);
	ChangeMB2U(0xFADF, 0x00E214);
	ChangeMB2U(0xFAE0, 0x00E215);
	ChangeMB2U(0xFAE1, 0x00E216);
	ChangeMB2U(0xFAE2, 0x00E217);
	ChangeMB2U(0xFAE3, 0x00E218);
	ChangeMB2U(0xFAE4, 0x00E219);
	ChangeMB2U(0xFAE5, 0x00E21A);
	ChangeMB2U(0xFAE6, 0x00E21B);
	ChangeMB2U(0xFAE7, 0x00E21C);
	ChangeMB2U(0xFAE8, 0x00E21D);
	ChangeMB2U(0xFAE9, 0x00E21E);
	ChangeMB2U(0xFAEA, 0x00E21F);
	ChangeMB2U(0xFAEB, 0x00E220);
	ChangeMB2U(0xFAEC, 0x00E221);
	ChangeMB2U(0xFAED, 0x00E222);
	ChangeMB2U(0xFAEE, 0x00E223);
	ChangeMB2U(0xFAEF, 0x00E224);
	ChangeMB2U(0xFAF0, 0x00E225);
	ChangeMB2U(0xFAF1, 0x00E226);
	ChangeMB2U(0xFAF2, 0x00E227);
	ChangeMB2U(0xFAF3, 0x00E228);
	ChangeMB2U(0xFAF4, 0x00E229);
	ChangeMB2U(0xFAF5, 0x00E22A);
	ChangeMB2U(0xFAF6, 0x00E22B);
	ChangeMB2U(0xFAF7, 0x00E22C);
	ChangeMB2U(0xFAF8, 0x00E22D);
	ChangeMB2U(0xFAF9, 0x00E22E);
	ChangeMB2U(0xFAFA, 0x00E22F);
	ChangeMB2U(0xFAFB, 0x00E230);
	ChangeMB2U(0xFAFC, 0x00E231);
	ChangeMB2U(0xFAFD, 0x00E232);
	ChangeMB2U(0xFAFE, 0x00E233);
	ChangeMB2U(0xFB21, 0x00E5E0);
	ChangeMB2U(0xFB22, 0x00E5E1);
	ChangeMB2U(0xFB23, 0x00E5E2);
	ChangeMB2U(0xFB24, 0x00E5E3);
	ChangeMB2U(0xFB25, 0x00E5E4);
	ChangeMB2U(0xFB26, 0x00E5E5);
	ChangeMB2U(0xFB27, 0x00E5E6);
	ChangeMB2U(0xFB28, 0x00E5E7);
	ChangeMB2U(0xFB29, 0x00E5E8);
	ChangeMB2U(0xFB2A, 0x00E5E9);
	ChangeMB2U(0xFB2B, 0x00E5EA);
	ChangeMB2U(0xFB2C, 0x00E5EB);
	ChangeMB2U(0xFB2D, 0x00E5EC);
	ChangeMB2U(0xFB2E, 0x00E5ED);
	ChangeMB2U(0xFB2F, 0x00E5EE);
	ChangeMB2U(0xFB30, 0x00E5EF);
	ChangeMB2U(0xFB31, 0x00E5F0);
	ChangeMB2U(0xFB32, 0x00E5F1);
	ChangeMB2U(0xFB33, 0x00E5F2);
	ChangeMB2U(0xFB34, 0x00E5F3);
	ChangeMB2U(0xFB35, 0x00E5F4);
	ChangeMB2U(0xFB36, 0x00E5F5);
	ChangeMB2U(0xFB37, 0x00E5F6);
	ChangeMB2U(0xFB38, 0x00E5F7);
	ChangeMB2U(0xFB39, 0x00E5F8);
	ChangeMB2U(0xFB3A, 0x00E5F9);
	ChangeMB2U(0xFB3B, 0x00E5FA);
	ChangeMB2U(0xFB3C, 0x00E5FB);
	ChangeMB2U(0xFB3D, 0x00E5FC);
	ChangeMB2U(0xFB3E, 0x00E5FD);
	ChangeMB2U(0xFB3F, 0x00E5FE);
	ChangeMB2U(0xFB40, 0x00E5FF);
	ChangeMB2U(0xFB41, 0x00E600);
	ChangeMB2U(0xFB42, 0x00E601);
	ChangeMB2U(0xFB43, 0x00E602);
	ChangeMB2U(0xFB44, 0x00E603);
	ChangeMB2U(0xFB45, 0x00E604);
	ChangeMB2U(0xFB46, 0x00E605);
	ChangeMB2U(0xFB47, 0x00E606);
	ChangeMB2U(0xFB48, 0x00E607);
	ChangeMB2U(0xFB49, 0x00E608);
	ChangeMB2U(0xFB4A, 0x00E609);
	ChangeMB2U(0xFB4B, 0x00E60A);
	ChangeMB2U(0xFB4C, 0x00E60B);
	ChangeMB2U(0xFB4D, 0x00E60C);
	ChangeMB2U(0xFB4E, 0x00E60D);
	ChangeMB2U(0xFB4F, 0x00E60E);
	ChangeMB2U(0xFB50, 0x00E60F);
	ChangeMB2U(0xFB51, 0x00E610);
	ChangeMB2U(0xFB52, 0x00E611);
	ChangeMB2U(0xFB53, 0x00E612);
	ChangeMB2U(0xFB54, 0x00E613);
	ChangeMB2U(0xFB55, 0x00E614);
	ChangeMB2U(0xFB56, 0x00E615);
	ChangeMB2U(0xFB57, 0x00E616);
	ChangeMB2U(0xFB58, 0x00E617);
	ChangeMB2U(0xFB59, 0x00E618);
	ChangeMB2U(0xFB5A, 0x00E619);
	ChangeMB2U(0xFB5B, 0x00E61A);
	ChangeMB2U(0xFB5C, 0x00E61B);
	ChangeMB2U(0xFB5D, 0x00E61C);
	ChangeMB2U(0xFB5E, 0x00E61D);
	ChangeMB2U(0xFB5F, 0x00E61E);
	ChangeMB2U(0xFB60, 0x00E61F);
	ChangeMB2U(0xFB61, 0x00E620);
	ChangeMB2U(0xFB62, 0x00E621);
	ChangeMB2U(0xFB63, 0x00E622);
	ChangeMB2U(0xFB64, 0x00E623);
	ChangeMB2U(0xFB65, 0x00E624);
	ChangeMB2U(0xFB66, 0x00E625);
	ChangeMB2U(0xFB67, 0x00E626);
	ChangeMB2U(0xFB68, 0x00E627);
	ChangeMB2U(0xFB69, 0x00E628);
	ChangeMB2U(0xFB6A, 0x00E629);
	ChangeMB2U(0xFB6B, 0x00E62A);
	ChangeMB2U(0xFB6C, 0x00E62B);
	ChangeMB2U(0xFB6D, 0x00E62C);
	ChangeMB2U(0xFB6E, 0x00E62D);
	ChangeMB2U(0xFB6F, 0x00E62E);
	ChangeMB2U(0xFB70, 0x00E62F);
	ChangeMB2U(0xFB71, 0x00E630);
	ChangeMB2U(0xFB72, 0x00E631);
	ChangeMB2U(0xFB73, 0x00E632);
	ChangeMB2U(0xFB74, 0x00E633);
	ChangeMB2U(0xFB75, 0x00E634);
	ChangeMB2U(0xFB76, 0x00E635);
	ChangeMB2U(0xFB77, 0x00E636);
	ChangeMB2U(0xFB78, 0x00E637);
	ChangeMB2U(0xFB79, 0x00E638);
	ChangeMB2U(0xFB7A, 0x00E639);
	ChangeMB2U(0xFB7B, 0x00E63A);
	ChangeMB2U(0xFB7C, 0x00E63B);
	ChangeMB2U(0xFB7D, 0x00E63C);
	ChangeMB2U(0xFB7E, 0x00E63D);
	ChangeMB2U(0xFBA1, 0x00E234);
	ChangeMB2U(0xFBA2, 0x00E235);
	ChangeMB2U(0xFBA3, 0x00E236);
	ChangeMB2U(0xFBA4, 0x00E237);
	ChangeMB2U(0xFBA5, 0x00E238);
	ChangeMB2U(0xFBA6, 0x00E239);
	ChangeMB2U(0xFBA7, 0x00E23A);
	ChangeMB2U(0xFBA8, 0x00E23B);
	ChangeMB2U(0xFBA9, 0x00E23C);
	ChangeMB2U(0xFBAA, 0x00E23D);
	ChangeMB2U(0xFBAB, 0x00E23E);
	ChangeMB2U(0xFBAC, 0x00E23F);
	ChangeMB2U(0xFBAD, 0x00E240);
	ChangeMB2U(0xFBAE, 0x00E241);
	ChangeMB2U(0xFBAF, 0x00E242);
	ChangeMB2U(0xFBB0, 0x00E243);
	ChangeMB2U(0xFBB1, 0x00E244);
	ChangeMB2U(0xFBB2, 0x00E245);
	ChangeMB2U(0xFBB3, 0x00E246);
	ChangeMB2U(0xFBB4, 0x00E247);
	ChangeMB2U(0xFBB5, 0x00E248);
	ChangeMB2U(0xFBB6, 0x00E249);
	ChangeMB2U(0xFBB7, 0x00E24A);
	ChangeMB2U(0xFBB8, 0x00E24B);
	ChangeMB2U(0xFBB9, 0x00E24C);
	ChangeMB2U(0xFBBA, 0x00E24D);
	ChangeMB2U(0xFBBB, 0x00E24E);
	ChangeMB2U(0xFBBC, 0x00E24F);
	ChangeMB2U(0xFBBD, 0x00E250);
	ChangeMB2U(0xFBBE, 0x00E251);
	ChangeMB2U(0xFBBF, 0x00E252);
	ChangeMB2U(0xFBC0, 0x00E253);
	ChangeMB2U(0xFBC1, 0x00E254);
	ChangeMB2U(0xFBC2, 0x00E255);
	ChangeMB2U(0xFBC3, 0x00E256);
	ChangeMB2U(0xFBC4, 0x00E257);
	ChangeMB2U(0xFBC5, 0x00E258);
	ChangeMB2U(0xFBC6, 0x00E259);
	ChangeMB2U(0xFBC7, 0x00E25A);
	ChangeMB2U(0xFBC8, 0x00E25B);
	ChangeMB2U(0xFBC9, 0x00E25C);
	ChangeMB2U(0xFBCA, 0x00E25D);
	ChangeMB2U(0xFBCB, 0x00E25E);
	ChangeMB2U(0xFBCC, 0x00E25F);
	ChangeMB2U(0xFBCD, 0x00E260);
	ChangeMB2U(0xFBCE, 0x00E261);
	ChangeMB2U(0xFBCF, 0x00E262);
	ChangeMB2U(0xFBD0, 0x00E263);
	ChangeMB2U(0xFBD1, 0x00E264);
	ChangeMB2U(0xFBD2, 0x00E265);
	ChangeMB2U(0xFBD3, 0x00E266);
	ChangeMB2U(0xFBD4, 0x00E267);
	ChangeMB2U(0xFBD5, 0x00E268);
	ChangeMB2U(0xFBD6, 0x00E269);
	ChangeMB2U(0xFBD7, 0x00E26A);
	ChangeMB2U(0xFBD8, 0x00E26B);
	ChangeMB2U(0xFBD9, 0x00E26C);
	ChangeMB2U(0xFBDA, 0x00E26D);
	ChangeMB2U(0xFBDB, 0x00E26E);
	ChangeMB2U(0xFBDC, 0x00E26F);
	ChangeMB2U(0xFBDD, 0x00E270);
	ChangeMB2U(0xFBDE, 0x00E271);
	ChangeMB2U(0xFBDF, 0x00E272);
	ChangeMB2U(0xFBE0, 0x00E273);
	ChangeMB2U(0xFBE1, 0x00E274);
	ChangeMB2U(0xFBE2, 0x00E275);
	ChangeMB2U(0xFBE3, 0x00E276);
	ChangeMB2U(0xFBE4, 0x00E277);
	ChangeMB2U(0xFBE5, 0x00E278);
	ChangeMB2U(0xFBE6, 0x00E279);
	ChangeMB2U(0xFBE7, 0x00E27A);
	ChangeMB2U(0xFBE8, 0x00E27B);
	ChangeMB2U(0xFBE9, 0x00E27C);
	ChangeMB2U(0xFBEA, 0x00E27D);
	ChangeMB2U(0xFBEB, 0x00E27E);
	ChangeMB2U(0xFBEC, 0x00E27F);
	ChangeMB2U(0xFBED, 0x00E280);
	ChangeMB2U(0xFBEE, 0x00E281);
	ChangeMB2U(0xFBEF, 0x00E282);
	ChangeMB2U(0xFBF0, 0x00E283);
	ChangeMB2U(0xFBF1, 0x00E284);
	ChangeMB2U(0xFBF2, 0x00E285);
	ChangeMB2U(0xFBF3, 0x00E286);
	ChangeMB2U(0xFBF4, 0x00E287);
	ChangeMB2U(0xFBF5, 0x00E288);
	ChangeMB2U(0xFBF6, 0x00E289);
	ChangeMB2U(0xFBF7, 0x00E28A);
	ChangeMB2U(0xFBF8, 0x00E28B);
	ChangeMB2U(0xFBF9, 0x00E28C);
	ChangeMB2U(0xFBFA, 0x00E28D);
	ChangeMB2U(0xFBFB, 0x00E28E);
	ChangeMB2U(0xFBFC, 0x00E28F);
	ChangeMB2U(0xFBFD, 0x00E290);
	ChangeMB2U(0xFBFE, 0x00E291);
	ChangeMB2U(0xFC21, 0x00E63E);
	ChangeMB2U(0xFC22, 0x00E63F);
	ChangeMB2U(0xFC23, 0x00E640);
	ChangeMB2U(0xFC24, 0x00E641);
	ChangeMB2U(0xFC25, 0x00E642);
	ChangeMB2U(0xFC26, 0x00E643);
	ChangeMB2U(0xFC27, 0x00E644);
	ChangeMB2U(0xFC28, 0x00E645);
	ChangeMB2U(0xFC29, 0x00E646);
	ChangeMB2U(0xFC2A, 0x00E647);
	ChangeMB2U(0xFC2B, 0x00E648);
	ChangeMB2U(0xFC2C, 0x00E649);
	ChangeMB2U(0xFC2D, 0x00E64A);
	ChangeMB2U(0xFC2E, 0x00E64B);
	ChangeMB2U(0xFC2F, 0x00E64C);
	ChangeMB2U(0xFC30, 0x00E64D);
	ChangeMB2U(0xFC31, 0x00E64E);
	ChangeMB2U(0xFC32, 0x00E64F);
	ChangeMB2U(0xFC33, 0x00E650);
	ChangeMB2U(0xFC34, 0x00E651);
	ChangeMB2U(0xFC35, 0x00E652);
	ChangeMB2U(0xFC36, 0x00E653);
	ChangeMB2U(0xFC37, 0x00E654);
	ChangeMB2U(0xFC38, 0x00E655);
	ChangeMB2U(0xFC39, 0x00E656);
	ChangeMB2U(0xFC3A, 0x00E657);
	ChangeMB2U(0xFC3B, 0x00E658);
	ChangeMB2U(0xFC3C, 0x00E659);
	ChangeMB2U(0xFC3D, 0x00E65A);
	ChangeMB2U(0xFC3E, 0x00E65B);
	ChangeMB2U(0xFC3F, 0x00E65C);
	ChangeMB2U(0xFC40, 0x00E65D);
	ChangeMB2U(0xFC41, 0x00E65E);
	ChangeMB2U(0xFC42, 0x00E65F);
	ChangeMB2U(0xFC43, 0x00E660);
	ChangeMB2U(0xFC44, 0x00E661);
	ChangeMB2U(0xFC45, 0x00E662);
	ChangeMB2U(0xFC46, 0x00E663);
	ChangeMB2U(0xFC47, 0x00E664);
	ChangeMB2U(0xFC48, 0x00E665);
	ChangeMB2U(0xFC49, 0x00E666);
	ChangeMB2U(0xFC4A, 0x00E667);
	ChangeMB2U(0xFC4B, 0x00E668);
	ChangeMB2U(0xFC4C, 0x00E669);
	ChangeMB2U(0xFC4D, 0x00E66A);
	ChangeMB2U(0xFC4E, 0x00E66B);
	ChangeMB2U(0xFC4F, 0x00E66C);
	ChangeMB2U(0xFC50, 0x00E66D);
	ChangeMB2U(0xFC51, 0x00E66E);
	ChangeMB2U(0xFC52, 0x00E66F);
	ChangeMB2U(0xFC53, 0x00E670);
	ChangeMB2U(0xFC54, 0x00E671);
	ChangeMB2U(0xFC55, 0x00E672);
	ChangeMB2U(0xFC56, 0x00E673);
	ChangeMB2U(0xFC57, 0x00E674);
	ChangeMB2U(0xFC58, 0x00E675);
	ChangeMB2U(0xFC59, 0x00E676);
	ChangeMB2U(0xFC5A, 0x00E677);
	ChangeMB2U(0xFC5B, 0x00E678);
	ChangeMB2U(0xFC5C, 0x00E679);
	ChangeMB2U(0xFC5D, 0x00E67A);
	ChangeMB2U(0xFC5E, 0x00E67B);
	ChangeMB2U(0xFC5F, 0x00E67C);
	ChangeMB2U(0xFC60, 0x00E67D);
	ChangeMB2U(0xFC61, 0x00E67E);
	ChangeMB2U(0xFC62, 0x00E67F);
	ChangeMB2U(0xFC63, 0x00E680);
	ChangeMB2U(0xFC64, 0x00E681);
	ChangeMB2U(0xFC65, 0x00E682);
	ChangeMB2U(0xFC66, 0x00E683);
	ChangeMB2U(0xFC67, 0x00E684);
	ChangeMB2U(0xFC68, 0x00E685);
	ChangeMB2U(0xFC69, 0x00E686);
	ChangeMB2U(0xFC6A, 0x00E687);
	ChangeMB2U(0xFC6B, 0x00E688);
	ChangeMB2U(0xFC6C, 0x00E689);
	ChangeMB2U(0xFC6D, 0x00E68A);
	ChangeMB2U(0xFC6E, 0x00E68B);
	ChangeMB2U(0xFC6F, 0x00E68C);
	ChangeMB2U(0xFC70, 0x00E68D);
	ChangeMB2U(0xFC71, 0x00E68E);
	ChangeMB2U(0xFC72, 0x00E68F);
	ChangeMB2U(0xFC73, 0x00E690);
	ChangeMB2U(0xFC74, 0x00E691);
	ChangeMB2U(0xFC75, 0x00E692);
	ChangeMB2U(0xFC76, 0x00E693);
	ChangeMB2U(0xFC77, 0x00E694);
	ChangeMB2U(0xFC78, 0x00E695);
	ChangeMB2U(0xFC79, 0x00E696);
	ChangeMB2U(0xFC7A, 0x00E697);
	ChangeMB2U(0xFC7B, 0x00E698);
	ChangeMB2U(0xFC7C, 0x00E699);
	ChangeMB2U(0xFC7D, 0x00E69A);
	ChangeMB2U(0xFC7E, 0x00E69B);
	ChangeMB2U(0xFCA1, 0x00E292);
	ChangeMB2U(0xFCA2, 0x00E293);
	ChangeMB2U(0xFCA3, 0x00E294);
	ChangeMB2U(0xFCA4, 0x00E295);
	ChangeMB2U(0xFCA5, 0x00E296);
	ChangeMB2U(0xFCA6, 0x00E297);
	ChangeMB2U(0xFCA7, 0x00E298);
	ChangeMB2U(0xFCA8, 0x00E299);
	ChangeMB2U(0xFCA9, 0x00E29A);
	ChangeMB2U(0xFCAA, 0x00E29B);
	ChangeMB2U(0xFCAB, 0x00E29C);
	ChangeMB2U(0xFCAC, 0x00E29D);
	ChangeMB2U(0xFCAD, 0x00E29E);
	ChangeMB2U(0xFCAE, 0x00E29F);
	ChangeMB2U(0xFCAF, 0x00E2A0);
	ChangeMB2U(0xFCB0, 0x00E2A1);
	ChangeMB2U(0xFCB1, 0x00E2A2);
	ChangeMB2U(0xFCB2, 0x00E2A3);
	ChangeMB2U(0xFCB3, 0x00E2A4);
	ChangeMB2U(0xFCB4, 0x00E2A5);
	ChangeMB2U(0xFCB5, 0x00E2A6);
	ChangeMB2U(0xFCB6, 0x00E2A7);
	ChangeMB2U(0xFCB7, 0x00E2A8);
	ChangeMB2U(0xFCB8, 0x00E2A9);
	ChangeMB2U(0xFCB9, 0x00E2AA);
	ChangeMB2U(0xFCBA, 0x00E2AB);
	ChangeMB2U(0xFCBB, 0x00E2AC);
	ChangeMB2U(0xFCBC, 0x00E2AD);
	ChangeMB2U(0xFCBD, 0x00E2AE);
	ChangeMB2U(0xFCBE, 0x00E2AF);
	ChangeMB2U(0xFCBF, 0x00E2B0);
	ChangeMB2U(0xFCC0, 0x00E2B1);
	ChangeMB2U(0xFCC1, 0x00E2B2);
	ChangeMB2U(0xFCC2, 0x00E2B3);
	ChangeMB2U(0xFCC3, 0x00E2B4);
	ChangeMB2U(0xFCC4, 0x00E2B5);
	ChangeMB2U(0xFCC5, 0x00E2B6);
	ChangeMB2U(0xFCC6, 0x00E2B7);
	ChangeMB2U(0xFCC7, 0x00E2B8);
	ChangeMB2U(0xFCC8, 0x00E2B9);
	ChangeMB2U(0xFCC9, 0x00E2BA);
	ChangeMB2U(0xFCCA, 0x00E2BB);
	ChangeMB2U(0xFCCB, 0x00E2BC);
	ChangeMB2U(0xFCCC, 0x00E2BD);
	ChangeMB2U(0xFCCD, 0x00E2BE);
	ChangeMB2U(0xFCCE, 0x00E2BF);
	ChangeMB2U(0xFCCF, 0x00E2C0);
	ChangeMB2U(0xFCD0, 0x00E2C1);
	ChangeMB2U(0xFCD1, 0x00E2C2);
	ChangeMB2U(0xFCD2, 0x00E2C3);
	ChangeMB2U(0xFCD3, 0x00E2C4);
	ChangeMB2U(0xFCD4, 0x00E2C5);
	ChangeMB2U(0xFCD5, 0x00E2C6);
	ChangeMB2U(0xFCD6, 0x00E2C7);
	ChangeMB2U(0xFCD7, 0x00E2C8);
	ChangeMB2U(0xFCD8, 0x00E2C9);
	ChangeMB2U(0xFCD9, 0x00E2CA);
	ChangeMB2U(0xFCDA, 0x00E2CB);
	ChangeMB2U(0xFCDB, 0x00E2CC);
	ChangeMB2U(0xFCDC, 0x00E2CD);
	ChangeMB2U(0xFCDD, 0x00E2CE);
	ChangeMB2U(0xFCDE, 0x00E2CF);
	ChangeMB2U(0xFCDF, 0x00E2D0);
	ChangeMB2U(0xFCE0, 0x00E2D1);
	ChangeMB2U(0xFCE1, 0x00E2D2);
	ChangeMB2U(0xFCE2, 0x00E2D3);
	ChangeMB2U(0xFCE3, 0x00E2D4);
	ChangeMB2U(0xFCE4, 0x00E2D5);
	ChangeMB2U(0xFCE5, 0x00E2D6);
	ChangeMB2U(0xFCE6, 0x00E2D7);
	ChangeMB2U(0xFCE7, 0x00E2D8);
	ChangeMB2U(0xFCE8, 0x00E2D9);
	ChangeMB2U(0xFCE9, 0x00E2DA);
	ChangeMB2U(0xFCEA, 0x00E2DB);
	ChangeMB2U(0xFCEB, 0x00E2DC);
	ChangeMB2U(0xFCEC, 0x00E2DD);
	ChangeMB2U(0xFCED, 0x00E2DE);
	ChangeMB2U(0xFCEE, 0x00E2DF);
	ChangeMB2U(0xFCEF, 0x00E2E0);
	ChangeMB2U(0xFCF0, 0x00E2E1);
	ChangeMB2U(0xFCF1, 0x00E2E2);
	ChangeMB2U(0xFCF2, 0x00E2E3);
	ChangeMB2U(0xFCF3, 0x00E2E4);
	ChangeMB2U(0xFCF4, 0x00E2E5);
	ChangeMB2U(0xFCF5, 0x00E2E6);
	ChangeMB2U(0xFCF6, 0x00E2E7);
	ChangeMB2U(0xFCF7, 0x00E2E8);
	ChangeMB2U(0xFCF8, 0x00E2E9);
	ChangeMB2U(0xFCF9, 0x00E2EA);
	ChangeMB2U(0xFCFA, 0x00E2EB);
	ChangeMB2U(0xFCFB, 0x00E2EC);
	ChangeMB2U(0xFCFC, 0x00E2ED);
	ChangeMB2U(0xFCFD, 0x00E2EE);
	ChangeMB2U(0xFCFE, 0x00E2EF);
	ChangeMB2U(0xFD21, 0x00E69C);
	ChangeMB2U(0xFD22, 0x00E69D);
	ChangeMB2U(0xFD23, 0x00E69E);
	ChangeMB2U(0xFD24, 0x00E69F);
	ChangeMB2U(0xFD25, 0x00E6A0);
	ChangeMB2U(0xFD26, 0x00E6A1);
	ChangeMB2U(0xFD27, 0x00E6A2);
	ChangeMB2U(0xFD28, 0x00E6A3);
	ChangeMB2U(0xFD29, 0x00E6A4);
	ChangeMB2U(0xFD2A, 0x00E6A5);
	ChangeMB2U(0xFD2B, 0x00E6A6);
	ChangeMB2U(0xFD2C, 0x00E6A7);
	ChangeMB2U(0xFD2D, 0x00E6A8);
	ChangeMB2U(0xFD2E, 0x00E6A9);
	ChangeMB2U(0xFD2F, 0x00E6AA);
	ChangeMB2U(0xFD30, 0x00E6AB);
	ChangeMB2U(0xFD31, 0x00E6AC);
	ChangeMB2U(0xFD32, 0x00E6AD);
	ChangeMB2U(0xFD33, 0x00E6AE);
	ChangeMB2U(0xFD34, 0x00E6AF);
	ChangeMB2U(0xFD35, 0x00E6B0);
	ChangeMB2U(0xFD36, 0x00E6B1);
	ChangeMB2U(0xFD37, 0x00E6B2);
	ChangeMB2U(0xFD38, 0x00E6B3);
	ChangeMB2U(0xFD39, 0x00E6B4);
	ChangeMB2U(0xFD3A, 0x00E6B5);
	ChangeMB2U(0xFD3B, 0x00E6B6);
	ChangeMB2U(0xFD3C, 0x00E6B7);
	ChangeMB2U(0xFD3D, 0x00E6B8);
	ChangeMB2U(0xFD3E, 0x00E6B9);
	ChangeMB2U(0xFD3F, 0x00E6BA);
	ChangeMB2U(0xFD40, 0x00E6BB);
	ChangeMB2U(0xFD41, 0x00E6BC);
	ChangeMB2U(0xFD42, 0x00E6BD);
	ChangeMB2U(0xFD43, 0x00E6BE);
	ChangeMB2U(0xFD44, 0x00E6BF);
	ChangeMB2U(0xFD45, 0x00E6C0);
	ChangeMB2U(0xFD46, 0x00E6C1);
	ChangeMB2U(0xFD47, 0x00E6C2);
	ChangeMB2U(0xFD48, 0x00E6C3);
	ChangeMB2U(0xFD49, 0x00E6C4);
	ChangeMB2U(0xFD4A, 0x00E6C5);
	ChangeMB2U(0xFD4B, 0x00E6C6);
	ChangeMB2U(0xFD4C, 0x00E6C7);
	ChangeMB2U(0xFD4D, 0x00E6C8);
	ChangeMB2U(0xFD4E, 0x00E6C9);
	ChangeMB2U(0xFD4F, 0x00E6CA);
	ChangeMB2U(0xFD50, 0x00E6CB);
	ChangeMB2U(0xFD51, 0x00E6CC);
	ChangeMB2U(0xFD52, 0x00E6CD);
	ChangeMB2U(0xFD53, 0x00E6CE);
	ChangeMB2U(0xFD54, 0x00E6CF);
	ChangeMB2U(0xFD55, 0x00E6D0);
	ChangeMB2U(0xFD56, 0x00E6D1);
	ChangeMB2U(0xFD57, 0x00E6D2);
	ChangeMB2U(0xFD58, 0x00E6D3);
	ChangeMB2U(0xFD59, 0x00E6D4);
	ChangeMB2U(0xFD5A, 0x00E6D5);
	ChangeMB2U(0xFD5B, 0x00E6D6);
	ChangeMB2U(0xFD5C, 0x00E6D7);
	ChangeMB2U(0xFD5D, 0x00E6D8);
	ChangeMB2U(0xFD5E, 0x00E6D9);
	ChangeMB2U(0xFD5F, 0x00E6DA);
	ChangeMB2U(0xFD60, 0x00E6DB);
	ChangeMB2U(0xFD61, 0x00E6DC);
	ChangeMB2U(0xFD62, 0x00E6DD);
	ChangeMB2U(0xFD63, 0x00E6DE);
	ChangeMB2U(0xFD64, 0x00E6DF);
	ChangeMB2U(0xFD65, 0x00E6E0);
	ChangeMB2U(0xFD66, 0x00E6E1);
	ChangeMB2U(0xFD67, 0x00E6E2);
	ChangeMB2U(0xFD68, 0x00E6E3);
	ChangeMB2U(0xFD69, 0x00E6E4);
	ChangeMB2U(0xFD6A, 0x00E6E5);
	ChangeMB2U(0xFD6B, 0x00E6E6);
	ChangeMB2U(0xFD6C, 0x00E6E7);
	ChangeMB2U(0xFD6D, 0x00E6E8);
	ChangeMB2U(0xFD6E, 0x00E6E9);
	ChangeMB2U(0xFD6F, 0x00E6EA);
	ChangeMB2U(0xFD70, 0x00E6EB);
	ChangeMB2U(0xFD71, 0x00E6EC);
	ChangeMB2U(0xFD72, 0x00E6ED);
	ChangeMB2U(0xFD73, 0x00E6EE);
	ChangeMB2U(0xFD74, 0x00E6EF);
	ChangeMB2U(0xFD75, 0x00E6F0);
	ChangeMB2U(0xFD76, 0x00E6F1);
	ChangeMB2U(0xFD77, 0x00E6F2);
	ChangeMB2U(0xFD78, 0x00E6F3);
	ChangeMB2U(0xFD79, 0x00E6F4);
	ChangeMB2U(0xFD7A, 0x00E6F5);
	ChangeMB2U(0xFD7B, 0x00E6F6);
	ChangeMB2U(0xFD7C, 0x00E6F7);
	ChangeMB2U(0xFD7D, 0x00E6F8);
	ChangeMB2U(0xFD7E, 0x00E6F9);
	ChangeMB2U(0xFDA1, 0x00E2F0);
	ChangeMB2U(0xFDA2, 0x00E2F1);
	ChangeMB2U(0xFDA3, 0x00E2F2);
	ChangeMB2U(0xFDA4, 0x00E2F3);
	ChangeMB2U(0xFDA5, 0x00E2F4);
	ChangeMB2U(0xFDA6, 0x00E2F5);
	ChangeMB2U(0xFDA7, 0x00E2F6);
	ChangeMB2U(0xFDA8, 0x00E2F7);
	ChangeMB2U(0xFDA9, 0x00E2F8);
	ChangeMB2U(0xFDAA, 0x00E2F9);
	ChangeMB2U(0xFDAB, 0x00E2FA);
	ChangeMB2U(0xFDAC, 0x00E2FB);
	ChangeMB2U(0xFDAD, 0x00E2FC);
	ChangeMB2U(0xFDAE, 0x00E2FD);
	ChangeMB2U(0xFDAF, 0x00E2FE);
	ChangeMB2U(0xFDB0, 0x00E2FF);
	ChangeMB2U(0xFDB1, 0x00E300);
	ChangeMB2U(0xFDB2, 0x00E301);
	ChangeMB2U(0xFDB3, 0x00E302);
	ChangeMB2U(0xFDB4, 0x00E303);
	ChangeMB2U(0xFDB5, 0x00E304);
	ChangeMB2U(0xFDB6, 0x00E305);
	ChangeMB2U(0xFDB7, 0x00E306);
	ChangeMB2U(0xFDB8, 0x00E307);
	ChangeMB2U(0xFDB9, 0x00E308);
	ChangeMB2U(0xFDBA, 0x00E309);
	ChangeMB2U(0xFDBB, 0x00E30A);
	ChangeMB2U(0xFDBC, 0x00E30B);
	ChangeMB2U(0xFDBD, 0x00E30C);
	ChangeMB2U(0xFDBE, 0x00E30D);
	ChangeMB2U(0xFDBF, 0x00E30E);
	ChangeMB2U(0xFDC0, 0x00E30F);
	ChangeMB2U(0xFDC1, 0x00E310);
	ChangeMB2U(0xFDC2, 0x00E311);
	ChangeMB2U(0xFDC3, 0x00E312);
	ChangeMB2U(0xFDC4, 0x00E313);
	ChangeMB2U(0xFDC5, 0x00E314);
	ChangeMB2U(0xFDC6, 0x00E315);
	ChangeMB2U(0xFDC7, 0x00E316);
	ChangeMB2U(0xFDC8, 0x00E317);
	ChangeMB2U(0xFDC9, 0x00E318);
	ChangeMB2U(0xFDCA, 0x00E319);
	ChangeMB2U(0xFDCB, 0x00E31A);
	ChangeMB2U(0xFDCC, 0x00E31B);
	ChangeMB2U(0xFDCD, 0x00E31C);
	ChangeMB2U(0xFDCE, 0x00E31D);
	ChangeMB2U(0xFDCF, 0x00E31E);
	ChangeMB2U(0xFDD0, 0x00E31F);
	ChangeMB2U(0xFDD1, 0x00E320);
	ChangeMB2U(0xFDD2, 0x00E321);
	ChangeMB2U(0xFDD3, 0x00E322);
	ChangeMB2U(0xFDD4, 0x00E323);
	ChangeMB2U(0xFDD5, 0x00E324);
	ChangeMB2U(0xFDD6, 0x00E325);
	ChangeMB2U(0xFDD7, 0x00E326);
	ChangeMB2U(0xFDD8, 0x00E327);
	ChangeMB2U(0xFDD9, 0x00E328);
	ChangeMB2U(0xFDDA, 0x00E329);
	ChangeMB2U(0xFDDB, 0x00E32A);
	ChangeMB2U(0xFDDC, 0x00E32B);
	ChangeMB2U(0xFDDD, 0x00E32C);
	ChangeMB2U(0xFDDE, 0x00E32D);
	ChangeMB2U(0xFDDF, 0x00E32E);
	ChangeMB2U(0xFDE0, 0x00E32F);
	ChangeMB2U(0xFDE1, 0x00E330);
	ChangeMB2U(0xFDE2, 0x00E331);
	ChangeMB2U(0xFDE3, 0x00E332);
	ChangeMB2U(0xFDE4, 0x00E333);
	ChangeMB2U(0xFDE5, 0x00E334);
	ChangeMB2U(0xFDE6, 0x00E335);
	ChangeMB2U(0xFDE7, 0x00E336);
	ChangeMB2U(0xFDE8, 0x00E337);
	ChangeMB2U(0xFDE9, 0x00E338);
	ChangeMB2U(0xFDEA, 0x00E339);
	ChangeMB2U(0xFDEB, 0x00E33A);
	ChangeMB2U(0xFDEC, 0x00E33B);
	ChangeMB2U(0xFDED, 0x00E33C);
	ChangeMB2U(0xFDEE, 0x00E33D);
	ChangeMB2U(0xFDEF, 0x00E33E);
	ChangeMB2U(0xFDF0, 0x00E33F);
	ChangeMB2U(0xFDF1, 0x00E340);
	ChangeMB2U(0xFDF2, 0x00E341);
	ChangeMB2U(0xFDF3, 0x00E342);
	ChangeMB2U(0xFDF4, 0x00E343);
	ChangeMB2U(0xFDF5, 0x00E344);
	ChangeMB2U(0xFDF6, 0x00E345);
	ChangeMB2U(0xFDF7, 0x00E346);
	ChangeMB2U(0xFDF8, 0x00E347);
	ChangeMB2U(0xFDF9, 0x00E348);
	ChangeMB2U(0xFDFA, 0x00E349);
	ChangeMB2U(0xFDFB, 0x00E34A);
	ChangeMB2U(0xFDFC, 0x00E34B);
	ChangeMB2U(0xFDFD, 0x00E34C);
	ChangeMB2U(0xFDFE, 0x00E34D);
	ChangeMB2U(0xFE21, 0x00E6FA);
	ChangeMB2U(0xFE22, 0x00E6FB);
	ChangeMB2U(0xFE23, 0x00E6FC);
	ChangeMB2U(0xFE24, 0x00E6FD);
	ChangeMB2U(0xFE25, 0x00E6FE);
	ChangeMB2U(0xFE26, 0x00E6FF);
	ChangeMB2U(0xFE27, 0x00E700);
	ChangeMB2U(0xFE28, 0x00E701);
	ChangeMB2U(0xFE29, 0x00E702);
	ChangeMB2U(0xFE2A, 0x00E703);
	ChangeMB2U(0xFE2B, 0x00E704);
	ChangeMB2U(0xFE2C, 0x00E705);
	ChangeMB2U(0xFE2D, 0x00E706);
	ChangeMB2U(0xFE2E, 0x00E707);
	ChangeMB2U(0xFE2F, 0x00E708);
	ChangeMB2U(0xFE30, 0x00E709);
	ChangeMB2U(0xFE31, 0x00E70A);
	ChangeMB2U(0xFE32, 0x00E70B);
	ChangeMB2U(0xFE33, 0x00E70C);
	ChangeMB2U(0xFE34, 0x00E70D);
	ChangeMB2U(0xFE35, 0x00E70E);
	ChangeMB2U(0xFE36, 0x00E70F);
	ChangeMB2U(0xFE37, 0x00E710);
	ChangeMB2U(0xFE38, 0x00E711);
	ChangeMB2U(0xFE39, 0x00E712);
	ChangeMB2U(0xFE3A, 0x00E713);
	ChangeMB2U(0xFE3B, 0x00E714);
	ChangeMB2U(0xFE3C, 0x00E715);
	ChangeMB2U(0xFE3D, 0x00E716);
	ChangeMB2U(0xFE3E, 0x00E717);
	ChangeMB2U(0xFE3F, 0x00E718);
	ChangeMB2U(0xFE40, 0x00E719);
	ChangeMB2U(0xFE41, 0x00E71A);
	ChangeMB2U(0xFE42, 0x00E71B);
	ChangeMB2U(0xFE43, 0x00E71C);
	ChangeMB2U(0xFE44, 0x00E71D);
	ChangeMB2U(0xFE45, 0x00E71E);
	ChangeMB2U(0xFE46, 0x00E71F);
	ChangeMB2U(0xFE47, 0x00E720);
	ChangeMB2U(0xFE48, 0x00E721);
	ChangeMB2U(0xFE49, 0x00E722);
	ChangeMB2U(0xFE4A, 0x00E723);
	ChangeMB2U(0xFE4B, 0x00E724);
	ChangeMB2U(0xFE4C, 0x00E725);
	ChangeMB2U(0xFE4D, 0x00E726);
	ChangeMB2U(0xFE4E, 0x00E727);
	ChangeMB2U(0xFE4F, 0x00E728);
	ChangeMB2U(0xFE50, 0x00E729);
	ChangeMB2U(0xFE51, 0x00E72A);
	ChangeMB2U(0xFE52, 0x00E72B);
	ChangeMB2U(0xFE53, 0x00E72C);
	ChangeMB2U(0xFE54, 0x00E72D);
	ChangeMB2U(0xFE55, 0x00E72E);
	ChangeMB2U(0xFE56, 0x00E72F);
	ChangeMB2U(0xFE57, 0x00E730);
	ChangeMB2U(0xFE58, 0x00E731);
	ChangeMB2U(0xFE59, 0x00E732);
	ChangeMB2U(0xFE5A, 0x00E733);
	ChangeMB2U(0xFE5B, 0x00E734);
	ChangeMB2U(0xFE5C, 0x00E735);
	ChangeMB2U(0xFE5D, 0x00E736);
	ChangeMB2U(0xFE5E, 0x00E737);
	ChangeMB2U(0xFE5F, 0x00E738);
	ChangeMB2U(0xFE60, 0x00E739);
	ChangeMB2U(0xFE61, 0x00E73A);
	ChangeMB2U(0xFE62, 0x00E73B);
	ChangeMB2U(0xFE63, 0x00E73C);
	ChangeMB2U(0xFE64, 0x00E73D);
	ChangeMB2U(0xFE65, 0x00E73E);
	ChangeMB2U(0xFE66, 0x00E73F);
	ChangeMB2U(0xFE67, 0x00E740);
	ChangeMB2U(0xFE68, 0x00E741);
	ChangeMB2U(0xFE69, 0x00E742);
	ChangeMB2U(0xFE6A, 0x00E743);
	ChangeMB2U(0xFE6B, 0x00E744);
	ChangeMB2U(0xFE6C, 0x00E745);
	ChangeMB2U(0xFE6D, 0x00E746);
	ChangeMB2U(0xFE6E, 0x00E747);
	ChangeMB2U(0xFE6F, 0x00E748);
	ChangeMB2U(0xFE70, 0x00E749);
	ChangeMB2U(0xFE71, 0x00E74A);
	ChangeMB2U(0xFE72, 0x00E74B);
	ChangeMB2U(0xFE73, 0x00E74C);
	ChangeMB2U(0xFE74, 0x00E74D);
	ChangeMB2U(0xFE75, 0x00E74E);
	ChangeMB2U(0xFE76, 0x00E74F);
	ChangeMB2U(0xFE77, 0x00E750);
	ChangeMB2U(0xFE78, 0x00E751);
	ChangeMB2U(0xFE79, 0x00E752);
	ChangeMB2U(0xFE7A, 0x00E753);
	ChangeMB2U(0xFE7B, 0x00E754);
	ChangeMB2U(0xFE7C, 0x00E755);
	ChangeMB2U(0xFE7D, 0x00E756);
	ChangeMB2U(0xFE7E, 0x00E757);
	ChangeMB2U(0xFEA1, 0x00E34E);
	ChangeMB2U(0xFEA2, 0x00E34F);
	ChangeMB2U(0xFEA3, 0x00E350);
	ChangeMB2U(0xFEA4, 0x00E351);
	ChangeMB2U(0xFEA5, 0x00E352);
	ChangeMB2U(0xFEA6, 0x00E353);
	ChangeMB2U(0xFEA7, 0x00E354);
	ChangeMB2U(0xFEA8, 0x00E355);
	ChangeMB2U(0xFEA9, 0x00E356);
	ChangeMB2U(0xFEAA, 0x00E357);
	ChangeMB2U(0xFEAB, 0x00E358);
	ChangeMB2U(0xFEAC, 0x00E359);
	ChangeMB2U(0xFEAD, 0x00E35A);
	ChangeMB2U(0xFEAE, 0x00E35B);
	ChangeMB2U(0xFEAF, 0x00E35C);
	ChangeMB2U(0xFEB0, 0x00E35D);
	ChangeMB2U(0xFEB1, 0x00E35E);
	ChangeMB2U(0xFEB2, 0x00E35F);
	ChangeMB2U(0xFEB3, 0x00E360);
	ChangeMB2U(0xFEB4, 0x00E361);
	ChangeMB2U(0xFEB5, 0x00E362);
	ChangeMB2U(0xFEB6, 0x00E363);
	ChangeMB2U(0xFEB7, 0x00E364);
	ChangeMB2U(0xFEB8, 0x00E365);
	ChangeMB2U(0xFEB9, 0x00E366);
	ChangeMB2U(0xFEBA, 0x00E367);
	ChangeMB2U(0xFEBB, 0x00E368);
	ChangeMB2U(0xFEBC, 0x00E369);
	ChangeMB2U(0xFEBD, 0x00E36A);
	ChangeMB2U(0xFEBE, 0x00E36B);
	ChangeMB2U(0xFEBF, 0x00E36C);
	ChangeMB2U(0xFEC0, 0x00E36D);
	ChangeMB2U(0xFEC1, 0x00E36E);
	ChangeMB2U(0xFEC2, 0x00E36F);
	ChangeMB2U(0xFEC3, 0x00E370);
	ChangeMB2U(0xFEC4, 0x00E371);
	ChangeMB2U(0xFEC5, 0x00E372);
	ChangeMB2U(0xFEC6, 0x00E373);
	ChangeMB2U(0xFEC7, 0x00E374);
	ChangeMB2U(0xFEC8, 0x00E375);
	ChangeMB2U(0xFEC9, 0x00E376);
	ChangeMB2U(0xFECA, 0x00E377);
	ChangeMB2U(0xFECB, 0x00E378);
	ChangeMB2U(0xFECC, 0x00E379);
	ChangeMB2U(0xFECD, 0x00E37A);
	ChangeMB2U(0xFECE, 0x00E37B);
	ChangeMB2U(0xFECF, 0x00E37C);
	ChangeMB2U(0xFED0, 0x00E37D);
	ChangeMB2U(0xFED1, 0x00E37E);
	ChangeMB2U(0xFED2, 0x00E37F);
	ChangeMB2U(0xFED3, 0x00E380);
	ChangeMB2U(0xFED4, 0x00E381);
	ChangeMB2U(0xFED5, 0x00E382);
	ChangeMB2U(0xFED6, 0x00E383);
	ChangeMB2U(0xFED7, 0x00E384);
	ChangeMB2U(0xFED8, 0x00E385);
	ChangeMB2U(0xFED9, 0x00E386);
	ChangeMB2U(0xFEDA, 0x00E387);
	ChangeMB2U(0xFEDB, 0x00E388);
	ChangeMB2U(0xFEDC, 0x00E389);
	ChangeMB2U(0xFEDD, 0x00E38A);
	ChangeMB2U(0xFEDE, 0x00E38B);
	ChangeMB2U(0xFEDF, 0x00E38C);
	ChangeMB2U(0xFEE0, 0x00E38D);
	ChangeMB2U(0xFEE1, 0x00E38E);
	ChangeMB2U(0xFEE2, 0x00E38F);
	ChangeMB2U(0xFEE3, 0x00E390);
	ChangeMB2U(0xFEE4, 0x00E391);
	ChangeMB2U(0xFEE5, 0x00E392);
	ChangeMB2U(0xFEE6, 0x00E393);
	ChangeMB2U(0xFEE7, 0x00E394);
	ChangeMB2U(0xFEE8, 0x00E395);
	ChangeMB2U(0xFEE9, 0x00E396);
	ChangeMB2U(0xFEEA, 0x00E397);
	ChangeMB2U(0xFEEB, 0x00E398);
	ChangeMB2U(0xFEEC, 0x00E399);
	ChangeMB2U(0xFEED, 0x00E39A);
	ChangeMB2U(0xFEEE, 0x00E39B);
	ChangeMB2U(0xFEEF, 0x00E39C);
	ChangeMB2U(0xFEF0, 0x00E39D);
	ChangeMB2U(0xFEF1, 0x00E39E);
	ChangeMB2U(0xFEF2, 0x00E39F);
	ChangeMB2U(0xFEF3, 0x00E3A0);
	ChangeMB2U(0xFEF4, 0x00E3A1);
	ChangeMB2U(0xFEF5, 0x00E3A2);
	ChangeMB2U(0xFEF6, 0x00E3A3);
	ChangeMB2U(0xFEF7, 0x00E3A4);
	ChangeMB2U(0xFEF8, 0x00E3A5);
	ChangeMB2U(0xFEF9, 0x00E3A6);
	ChangeMB2U(0xFEFA, 0x00E3A7);
	ChangeMB2U(0xFEFB, 0x00E3A8);
	ChangeMB2U(0xFEFC, 0x00E3A9);
	ChangeMB2U(0xFEFD, 0x00E3AA);
	ChangeMB2U(0xFEFE, 0x00E3AB);

	ChangeMB2U(0xFF00, 0x00F8F3);

	RemoveU2MB(0x000000);
	RemoveU2MB(0x000080);
	RemoveU2MB(0x000081);
	RemoveU2MB(0x000082);
	RemoveU2MB(0x000083);
	RemoveU2MB(0x000084);
	RemoveU2MB(0x000085);
	RemoveU2MB(0x000086);
	RemoveU2MB(0x000087);
	RemoveU2MB(0x000088);
	RemoveU2MB(0x000089);
	RemoveU2MB(0x00008A);
	RemoveU2MB(0x00008B);
	RemoveU2MB(0x00008C);
	RemoveU2MB(0x00008D);
	RemoveU2MB(0x000090);
	RemoveU2MB(0x000091);
	RemoveU2MB(0x000092);
	RemoveU2MB(0x000093);
	RemoveU2MB(0x000094);
	RemoveU2MB(0x000095);
	RemoveU2MB(0x000096);
	RemoveU2MB(0x000097);
	RemoveU2MB(0x000098);
	RemoveU2MB(0x000099);
	RemoveU2MB(0x00009A);
	RemoveU2MB(0x00009B);
	RemoveU2MB(0x00009C);
	RemoveU2MB(0x00009D);
	RemoveU2MB(0x00009E);
	RemoveU2MB(0x00009F);
	RemoveU2MB(0x002225);
	RemoveU2MB(0x00301D);
	RemoveU2MB(0x00663B);
	RemoveU2MB(0x00FF0D);
	RemoveU2MB(0x00FF5E);
	RemoveU2MB(0x00FFE0);
	RemoveU2MB(0x00FFE1);
	RemoveU2MB(0x00FFE2);
	RemoveU2MB(0x00FFE4);

	ChangeU2MB(0x0000A2, 0xA1F1);
	ChangeU2MB(0x0000A3, 0xA1F2);
	ChangeU2MB(0x0000A6, 0xA243);
	ChangeU2MB(0x0000AC, 0xA2CC);
	ChangeU2MB(0x002016, 0xA1C2);
	ChangeU2MB(0x002170, 0xF321);
	ChangeU2MB(0x002171, 0xF322);
	ChangeU2MB(0x002172, 0xF323);
	ChangeU2MB(0x002173, 0xF324);
	ChangeU2MB(0x002174, 0xF325);
	ChangeU2MB(0x002175, 0xF326);
	ChangeU2MB(0x002176, 0xF327);
	ChangeU2MB(0x002177, 0xF328);
	ChangeU2MB(0x002178, 0xF329);
	ChangeU2MB(0x002179, 0xF32A);
	ChangeU2MB(0x002212, 0xA1DD);
	ChangeU2MB(0x00301C, 0xA1C1);
	ChangeU2MB(0x00301E, 0xADE0);
	ChangeU2MB(0x004E28, 0xB029);
	ChangeU2MB(0x004EE1, 0xB048);
	ChangeU2MB(0x004EFC, 0xF33B);
	ChangeU2MB(0x004F00, 0xB052);
	ChangeU2MB(0x004F03, 0xB054);
	ChangeU2MB(0x004F39, 0xB063);
	ChangeU2MB(0x004F56, 0xB06E);
	ChangeU2MB(0x004F8A, 0xB123);
	ChangeU2MB(0x004F92, 0xB127);
	ChangeU2MB(0x004F94, 0xB129);
	ChangeU2MB(0x004F9A, 0xB12C);
	ChangeU2MB(0x004FC9, 0xB13B);
	ChangeU2MB(0x004FCD, 0xB13E);
	ChangeU2MB(0x004FFF, 0xB148);
	ChangeU2MB(0x00501E, 0xB157);
	ChangeU2MB(0x005022, 0xB158);
	ChangeU2MB(0x005040, 0xB15F);
	ChangeU2MB(0x005042, 0xB161);
	ChangeU2MB(0x005046, 0xB163);
	ChangeU2MB(0x005070, 0xB174);
	ChangeU2MB(0x005094, 0xB223);
	ChangeU2MB(0x0050D8, 0xB23B);
	ChangeU2MB(0x0050F4, 0xF33C);
	ChangeU2MB(0x00514A, 0xB266);
	ChangeU2MB(0x005164, 0xB26D);
	ChangeU2MB(0x00519D, 0xB275);
	ChangeU2MB(0x0051BE, 0xB27C);
	ChangeU2MB(0x0051EC, 0xF33D);
	ChangeU2MB(0x005215, 0xB335);
	ChangeU2MB(0x00529C, 0xB358);
	ChangeU2MB(0x0052A6, 0xB35B);
	ChangeU2MB(0x0052AF, 0xB35D);
	ChangeU2MB(0x0052C0, 0xB365);
	ChangeU2MB(0x0052DB, 0xB36E);
	ChangeU2MB(0x005300, 0xB37B);
	ChangeU2MB(0x005307, 0xF33E);
	ChangeU2MB(0x005324, 0xF33F);
	ChangeU2MB(0x005372, 0xB440);
	ChangeU2MB(0x005393, 0xB447);
	ChangeU2MB(0x0053B2, 0xB450);
	ChangeU2MB(0x0053DD, 0xB45E);
	ChangeU2MB(0x00548A, 0xF341);
	ChangeU2MB(0x00549C, 0xB52A);
	ChangeU2MB(0x0054A9, 0xB52F);
	ChangeU2MB(0x0054FF, 0xB544);
	ChangeU2MB(0x005586, 0xB568);
	ChangeU2MB(0x005759, 0xF342);
	ChangeU2MB(0x005765, 0xB742);
	ChangeU2MB(0x0057AC, 0xB764);
	ChangeU2MB(0x0057C7, 0xB767);
	ChangeU2MB(0x0057C8, 0xB768);
	ChangeU2MB(0x00589E, 0xF345);
	ChangeU2MB(0x0058B2, 0xB84E);
	ChangeU2MB(0x00590B, 0xB861);
	ChangeU2MB(0x005953, 0xB875);
	ChangeU2MB(0x00595B, 0xB877);
	ChangeU2MB(0x00595D, 0xB878);
	ChangeU2MB(0x005963, 0xB87C);
	ChangeU2MB(0x0059A4, 0xB92F);
	ChangeU2MB(0x0059BA, 0xB937);
	ChangeU2MB(0x005B56, 0xBA3E);
	ChangeU2MB(0x005BC0, 0xBA5B);
	ChangeU2MB(0x005BD8, 0xBA61);
	ChangeU2MB(0x005BEC, 0xF346);
	ChangeU2MB(0x005C1E, 0xBA6B);
	ChangeU2MB(0x005CA6, 0xBB33);
	ChangeU2MB(0x005CBA, 0xBB38);
	ChangeU2MB(0x005CF5, 0xF347);
	ChangeU2MB(0x005D27, 0xBB4A);
	ChangeU2MB(0x005D42, 0xBB50);
	ChangeU2MB(0x005D53, 0xF348);
	ChangeU2MB(0x005D6D, 0xBB5E);
	ChangeU2MB(0x005DB8, 0xBB74);
	ChangeU2MB(0x005DB9, 0xBB75);
	ChangeU2MB(0x005DD0, 0xBB79);
	ChangeU2MB(0x005F21, 0xBC64);
	ChangeU2MB(0x005F34, 0xBC6D);
	ChangeU2MB(0x005F45, 0xBC74);
	ChangeU2MB(0x005F67, 0xBC7E);
	ChangeU2MB(0x005FB7, 0xF34A);
	ChangeU2MB(0x005FDE, 0xBD42);
	ChangeU2MB(0x00605D, 0xBD67);
	ChangeU2MB(0x006085, 0xF34B);
	ChangeU2MB(0x00608A, 0xBD70);
	ChangeU2MB(0x0060D5, 0xBE2C);
	ChangeU2MB(0x0060DE, 0xBE30);
	ChangeU2MB(0x0060F2, 0xBE33);
	ChangeU2MB(0x006111, 0xBE3D);
	ChangeU2MB(0x006120, 0xF34C);
	ChangeU2MB(0x006130, 0xBE49);
	ChangeU2MB(0x006137, 0xBE4D);
	ChangeU2MB(0x006198, 0xBE64);
	ChangeU2MB(0x006213, 0xBF28);
	ChangeU2MB(0x0062A6, 0xBF49);
	ChangeU2MB(0x0063F5, 0xC044);
	ChangeU2MB(0x006460, 0xC064);
	ChangeU2MB(0x00649D, 0xC074);
	ChangeU2MB(0x0064CE, 0xC126);
	ChangeU2MB(0x00654E, 0xF34D);
	ChangeU2MB(0x006600, 0xC175);
	ChangeU2MB(0x006609, 0xC178);
	ChangeU2MB(0x006615, 0xC17C);
	ChangeU2MB(0x00661E, 0xC221);
	ChangeU2MB(0x006624, 0xC225);
	ChangeU2MB(0x00662E, 0xC22B);
	ChangeU2MB(0x006631, 0xC22D);
	ChangeU2MB(0x006657, 0xC238);
	ChangeU2MB(0x006659, 0xC23A);
	ChangeU2MB(0x006665, 0xF34F);
	ChangeU2MB(0x006673, 0xC244);
	ChangeU2MB(0x006699, 0xC252);
	ChangeU2MB(0x0066A0, 0xC257);
	ChangeU2MB(0x0066B2, 0xC25B);
	ChangeU2MB(0x0066BF, 0xC25E);
	ChangeU2MB(0x0066FA, 0xC26D);
	ChangeU2MB(0x0066FB, 0xC23F);
	ChangeU2MB(0x00670E, 0xC270);
	ChangeU2MB(0x006766, 0xC321);
	ChangeU2MB(0x0067BB, 0xC335);
	ChangeU2MB(0x0067C0, 0xC339);
	ChangeU2MB(0x006801, 0xF352);
	ChangeU2MB(0x006844, 0xC358);
	ChangeU2MB(0x006852, 0xC349);
	ChangeU2MB(0x0068C8, 0xC37C);
	ChangeU2MB(0x0068CF, 0xC37E);
	ChangeU2MB(0x006968, 0xC44C);
	ChangeU2MB(0x006998, 0xC459);
	ChangeU2MB(0x0069E2, 0xC46A);
	ChangeU2MB(0x006A30, 0xC47D);
	ChangeU2MB(0x006A46, 0xC527);
	ChangeU2MB(0x006A6B, 0xF355);
	ChangeU2MB(0x006A73, 0xC535);
	ChangeU2MB(0x006A7E, 0xC536);
	ChangeU2MB(0x006AE2, 0xF356);
	ChangeU2MB(0x006AE4, 0xC555);
	ChangeU2MB(0x006BD6, 0xC638);
	ChangeU2MB(0x006C3F, 0xC657);
	ChangeU2MB(0x006C5C, 0xC660);
	ChangeU2MB(0x006C6F, 0xC663);
	ChangeU2MB(0x006C86, 0xC66A);
	ChangeU2MB(0x006CDA, 0xC721);
	ChangeU2MB(0x006D04, 0xC72B);
	ChangeU2MB(0x006D6F, 0xC743);
	ChangeU2MB(0x006D87, 0xC747);
	ChangeU2MB(0x006D96, 0xC74B);
	ChangeU2MB(0x006DAC, 0xC74F);
	ChangeU2MB(0x006DCF, 0xC759);
	ChangeU2MB(0x006DF2, 0xF358);
	ChangeU2MB(0x006DF8, 0xF357);
	ChangeU2MB(0x006DFC, 0xC766);
	ChangeU2MB(0x006E27, 0xC76B);
	ChangeU2MB(0x006E39, 0xC76E);
	ChangeU2MB(0x006E3C, 0xC770);
	ChangeU2MB(0x006E5C, 0xC77C);
	ChangeU2MB(0x006EBF, 0xC831);
	ChangeU2MB(0x006F88, 0xC865);
	ChangeU2MB(0x006FB5, 0xC878);
	ChangeU2MB(0x006FF5, 0xC926);
	ChangeU2MB(0x007005, 0xC92B);
	ChangeU2MB(0x007007, 0xC92D);
	ChangeU2MB(0x007028, 0xF359);
	ChangeU2MB(0x007085, 0xC94A);
	ChangeU2MB(0x0070AB, 0xC953);
	ChangeU2MB(0x0070BB, 0xF33A);
	ChangeU2MB(0x007104, 0xC963);
	ChangeU2MB(0x00710F, 0xC969);
	ChangeU2MB(0x007146, 0xC974);
	ChangeU2MB(0x007147, 0xC975);
	ChangeU2MB(0x00715C, 0xC97C);
	ChangeU2MB(0x0071C1, 0xCA33);
	ChangeU2MB(0x0071FE, 0xCA3D);
	ChangeU2MB(0x0072B1, 0xCA6F);
	ChangeU2MB(0x0072BE, 0xCA71);
	ChangeU2MB(0x007324, 0xCB2E);
	ChangeU2MB(0x007377, 0xCB4A);
	ChangeU2MB(0x0073BD, 0xCB66);
	ChangeU2MB(0x0073C9, 0xCB6A);
	ChangeU2MB(0x0073D2, 0xCB6E);
	ChangeU2MB(0x0073D6, 0xCB70);
	ChangeU2MB(0x0073E3, 0xCB74);
	ChangeU2MB(0x0073F5, 0xCB79);
	ChangeU2MB(0x007407, 0xCC25);
	ChangeU2MB(0x007426, 0xCC2B);
	ChangeU2MB(0x007429, 0xCC2D);
	ChangeU2MB(0x00742A, 0xCC2E);
	ChangeU2MB(0x00742E, 0xCC32);
	ChangeU2MB(0x007462, 0xCC42);
	ChangeU2MB(0x007489, 0xCC50);
	ChangeU2MB(0x00749F, 0xCC59);
	ChangeU2MB(0x007501, 0xF35C);
	ChangeU2MB(0x00752F, 0xCD2A);
	ChangeU2MB(0x00756F, 0xCD3B);
	ChangeU2MB(0x007682, 0xF35D);
	ChangeU2MB(0x00769B, 0xCE3A);
	ChangeU2MB(0x00769C, 0xCE3B);
	ChangeU2MB(0x00769E, 0xF35E);
	ChangeU2MB(0x0076A6, 0xCE43);
	ChangeU2MB(0x007746, 0xCE72);
	ChangeU2MB(0x007821, 0xCF55);
	ChangeU2MB(0x00784E, 0xCF62);
	ChangeU2MB(0x007864, 0xCF69);
	ChangeU2MB(0x00787A, 0xCF6D);
	ChangeU2MB(0x007930, 0xF360);
	ChangeU2MB(0x007994, 0xD065);
	ChangeU2MB(0x00799B, 0xD069);
	ChangeU2MB(0x007AD1, 0xD168);
	ChangeU2MB(0x007AE7, 0xF365);
	ChangeU2MB(0x007AEB, 0xD16C);
	ChangeU2MB(0x007B9E, 0xD23B);
	ChangeU2MB(0x007D48, 0xD361);
	ChangeU2MB(0x007D5C, 0xD368);
	ChangeU2MB(0x007DA0, 0xF368);
	ChangeU2MB(0x007DB7, 0xD427);
	ChangeU2MB(0x007DD6, 0xF369);
	ChangeU2MB(0x007E52, 0xD454);
	ChangeU2MB(0x007E8A, 0xD463);
	ChangeU2MB(0x007F47, 0xD472);
	ChangeU2MB(0x007FA1, 0xD52E);
	ChangeU2MB(0x008301, 0xD75E);
	ChangeU2MB(0x008362, 0xF36B);
	ChangeU2MB(0x00837F, 0xD822);
	ChangeU2MB(0x0083C7, 0xD837);
	ChangeU2MB(0x0083F6, 0xD841);
	ChangeU2MB(0x008448, 0xD851);
	ChangeU2MB(0x0084B4, 0xD874);
	ChangeU2MB(0x0084DC, 0xD921);
	ChangeU2MB(0x008553, 0xD946);
	ChangeU2MB(0x008559, 0xD948);
	ChangeU2MB(0x00856B, 0xD951);
	ChangeU2MB(0x0085B0, 0xF36D);
	ChangeU2MB(0x008807, 0xF370);
	ChangeU2MB(0x0088F5, 0xDC53);
	ChangeU2MB(0x00891C, 0xDC5F);
	ChangeU2MB(0x008A12, 0xDD48);
	ChangeU2MB(0x008A37, 0xDD54);
	ChangeU2MB(0x008A79, 0xDD6A);
	ChangeU2MB(0x008AA7, 0xDD7A);
	ChangeU2MB(0x008ABE, 0xDE24);
	ChangeU2MB(0x008ADF, 0xDE30);
	ChangeU2MB(0x008AF6, 0xDE35);
	ChangeU2MB(0x008B53, 0xDE4B);
	ChangeU2MB(0x008B7F, 0xF372);
	ChangeU2MB(0x008CF0, 0xDF39);
	ChangeU2MB(0x008CF4, 0xF373);
	ChangeU2MB(0x008D12, 0xDF43);
	ChangeU2MB(0x008D76, 0xF374);
	ChangeU2MB(0x008ECF, 0xE059);
	ChangeU2MB(0x009067, 0xE162);
	ChangeU2MB(0x0090DE, 0xF378);
	ChangeU2MB(0x009115, 0xF37A);
	ChangeU2MB(0x009127, 0xE247);
	ChangeU2MB(0x0091D7, 0xE326);
	ChangeU2MB(0x0091DA, 0xE328);
	ChangeU2MB(0x0091DE, 0xE329);
	ChangeU2MB(0x0091E4, 0xE32A);
	ChangeU2MB(0x0091E5, 0xE32B);
	ChangeU2MB(0x0091ED, 0xE32F);
	ChangeU2MB(0x0091EE, 0xE330);
	ChangeU2MB(0x009206, 0xE33C);
	ChangeU2MB(0x00920A, 0xE33F);
	ChangeU2MB(0x009210, 0xE341);
	ChangeU2MB(0x009239, 0xE354);
	ChangeU2MB(0x00923A, 0xE355);
	ChangeU2MB(0x00923C, 0xE356);
	ChangeU2MB(0x009240, 0xE358);
	ChangeU2MB(0x00924E, 0xE35F);
	ChangeU2MB(0x009251, 0xE361);
	ChangeU2MB(0x009259, 0xE363);
	ChangeU2MB(0x009267, 0xE369);
	ChangeU2MB(0x009277, 0xE371);
	ChangeU2MB(0x009278, 0xE372);
	ChangeU2MB(0x009288, 0xE378);
	ChangeU2MB(0x0092A7, 0xE426);
	ChangeU2MB(0x0092D0, 0xE43E);
	ChangeU2MB(0x0092D3, 0xE43F);
	ChangeU2MB(0x0092D5, 0xE440);
	ChangeU2MB(0x0092D7, 0xE441);
	ChangeU2MB(0x0092D9, 0xE443);
	ChangeU2MB(0x0092E0, 0xE447);
	ChangeU2MB(0x0092E7, 0xE44B);
	ChangeU2MB(0x0092F9, 0xE450);
	ChangeU2MB(0x0092FB, 0xE451);
	ChangeU2MB(0x0092FF, 0xE452);
	ChangeU2MB(0x009302, 0xE454);
	ChangeU2MB(0x00931D, 0xE45B);
	ChangeU2MB(0x00931E, 0xE45C);
	ChangeU2MB(0x009321, 0xE45E);
	ChangeU2MB(0x009325, 0xE460);
	ChangeU2MB(0x009348, 0xE469);
	ChangeU2MB(0x009357, 0xE46F);
	ChangeU2MB(0x009370, 0xE47A);
	ChangeU2MB(0x0093A4, 0xE533);
	ChangeU2MB(0x0093C6, 0xE53F);
	ChangeU2MB(0x0093DE, 0xE549);
	ChangeU2MB(0x0093F8, 0xE550);
	ChangeU2MB(0x009431, 0xE562);
	ChangeU2MB(0x009445, 0xE56A);
	ChangeU2MB(0x009448, 0xE56B);
	ChangeU2MB(0x009592, 0xF37D);
	ChangeU2MB(0x00969D, 0xE668);
	ChangeU2MB(0x0096AF, 0xE66F);
	ChangeU2MB(0x009733, 0xE72C);
	ChangeU2MB(0x00973B, 0xF422);
	ChangeU2MB(0x009743, 0xE72E);
	ChangeU2MB(0x00974D, 0xF463);
	ChangeU2MB(0x00974F, 0xE731);
	ChangeU2MB(0x009751, 0xF424);
	ChangeU2MB(0x009755, 0xE732);
	ChangeU2MB(0x009857, 0xE831);
	ChangeU2MB(0x009865, 0xE836);
	ChangeU2MB(0x009927, 0xE85D);
	ChangeU2MB(0x00999E, 0xF428);
	ChangeU2MB(0x009A4E, 0xE951);
	ChangeU2MB(0x009AD9, 0xF429);
	ChangeU2MB(0x009ADC, 0xE96D);
	ChangeU2MB(0x009B72, 0xF42A);
	ChangeU2MB(0x009B75, 0xEA4D);
	ChangeU2MB(0x009B8F, 0xEA5B);
	ChangeU2MB(0x009BB1, 0xEA66);
	ChangeU2MB(0x009BBB, 0xEA6A);
	ChangeU2MB(0x009C00, 0xEB25);
	ChangeU2MB(0x009D6B, 0xEB7A);
	ChangeU2MB(0x009D70, 0xEB7B);
	ChangeU2MB(0x009E19, 0xEC56);
	ChangeU2MB(0x009ED1, 0xF42C);
	ChangeU2MB(0x00E000, 0xF5A1);
	ChangeU2MB(0x00E001, 0xF5A2);
	ChangeU2MB(0x00E002, 0xF5A3);
	ChangeU2MB(0x00E003, 0xF5A4);
	ChangeU2MB(0x00E004, 0xF5A5);
	ChangeU2MB(0x00E005, 0xF5A6);
	ChangeU2MB(0x00E006, 0xF5A7);
	ChangeU2MB(0x00E007, 0xF5A8);
	ChangeU2MB(0x00E008, 0xF5A9);
	ChangeU2MB(0x00E009, 0xF5AA);
	ChangeU2MB(0x00E00A, 0xF5AB);
	ChangeU2MB(0x00E00B, 0xF5AC);
	ChangeU2MB(0x00E00C, 0xF5AD);
	ChangeU2MB(0x00E00D, 0xF5AE);
	ChangeU2MB(0x00E00E, 0xF5AF);
	ChangeU2MB(0x00E00F, 0xF5B0);
	ChangeU2MB(0x00E010, 0xF5B1);
	ChangeU2MB(0x00E011, 0xF5B2);
	ChangeU2MB(0x00E012, 0xF5B3);
	ChangeU2MB(0x00E013, 0xF5B4);
	ChangeU2MB(0x00E014, 0xF5B5);
	ChangeU2MB(0x00E015, 0xF5B6);
	ChangeU2MB(0x00E016, 0xF5B7);
	ChangeU2MB(0x00E017, 0xF5B8);
	ChangeU2MB(0x00E018, 0xF5B9);
	ChangeU2MB(0x00E019, 0xF5BA);
	ChangeU2MB(0x00E01A, 0xF5BB);
	ChangeU2MB(0x00E01B, 0xF5BC);
	ChangeU2MB(0x00E01C, 0xF5BD);
	ChangeU2MB(0x00E01D, 0xF5BE);
	ChangeU2MB(0x00E01E, 0xF5BF);
	ChangeU2MB(0x00E01F, 0xF5C0);
	ChangeU2MB(0x00E020, 0xF5C1);
	ChangeU2MB(0x00E021, 0xF5C2);
	ChangeU2MB(0x00E022, 0xF5C3);
	ChangeU2MB(0x00E023, 0xF5C4);
	ChangeU2MB(0x00E024, 0xF5C5);
	ChangeU2MB(0x00E025, 0xF5C6);
	ChangeU2MB(0x00E026, 0xF5C7);
	ChangeU2MB(0x00E027, 0xF5C8);
	ChangeU2MB(0x00E028, 0xF5C9);
	ChangeU2MB(0x00E029, 0xF5CA);
	ChangeU2MB(0x00E02A, 0xF5CB);
	ChangeU2MB(0x00E02B, 0xF5CC);
	ChangeU2MB(0x00E02C, 0xF5CD);
	ChangeU2MB(0x00E02D, 0xF5CE);
	ChangeU2MB(0x00E02E, 0xF5CF);
	ChangeU2MB(0x00E02F, 0xF5D0);
	ChangeU2MB(0x00E030, 0xF5D1);
	ChangeU2MB(0x00E031, 0xF5D2);
	ChangeU2MB(0x00E032, 0xF5D3);
	ChangeU2MB(0x00E033, 0xF5D4);
	ChangeU2MB(0x00E034, 0xF5D5);
	ChangeU2MB(0x00E035, 0xF5D6);
	ChangeU2MB(0x00E036, 0xF5D7);
	ChangeU2MB(0x00E037, 0xF5D8);
	ChangeU2MB(0x00E038, 0xF5D9);
	ChangeU2MB(0x00E039, 0xF5DA);
	ChangeU2MB(0x00E03A, 0xF5DB);
	ChangeU2MB(0x00E03B, 0xF5DC);
	ChangeU2MB(0x00E03C, 0xF5DD);
	ChangeU2MB(0x00E03D, 0xF5DE);
	ChangeU2MB(0x00E03E, 0xF5DF);
	ChangeU2MB(0x00E03F, 0xF5E0);
	ChangeU2MB(0x00E040, 0xF5E1);
	ChangeU2MB(0x00E041, 0xF5E2);
	ChangeU2MB(0x00E042, 0xF5E3);
	ChangeU2MB(0x00E043, 0xF5E4);
	ChangeU2MB(0x00E044, 0xF5E5);
	ChangeU2MB(0x00E045, 0xF5E6);
	ChangeU2MB(0x00E046, 0xF5E7);
	ChangeU2MB(0x00E047, 0xF5E8);
	ChangeU2MB(0x00E048, 0xF5E9);
	ChangeU2MB(0x00E049, 0xF5EA);
	ChangeU2MB(0x00E04A, 0xF5EB);
	ChangeU2MB(0x00E04B, 0xF5EC);
	ChangeU2MB(0x00E04C, 0xF5ED);
	ChangeU2MB(0x00E04D, 0xF5EE);
	ChangeU2MB(0x00E04E, 0xF5EF);
	ChangeU2MB(0x00E04F, 0xF5F0);
	ChangeU2MB(0x00E050, 0xF5F1);
	ChangeU2MB(0x00E051, 0xF5F2);
	ChangeU2MB(0x00E052, 0xF5F3);
	ChangeU2MB(0x00E053, 0xF5F4);
	ChangeU2MB(0x00E054, 0xF5F5);
	ChangeU2MB(0x00E055, 0xF5F6);
	ChangeU2MB(0x00E056, 0xF5F7);
	ChangeU2MB(0x00E057, 0xF5F8);
	ChangeU2MB(0x00E058, 0xF5F9);
	ChangeU2MB(0x00E059, 0xF5FA);
	ChangeU2MB(0x00E05A, 0xF5FB);
	ChangeU2MB(0x00E05B, 0xF5FC);
	ChangeU2MB(0x00E05C, 0xF5FD);
	ChangeU2MB(0x00E05D, 0xF5FE);
	ChangeU2MB(0x00E05E, 0xF6A1);
	ChangeU2MB(0x00E05F, 0xF6A2);
	ChangeU2MB(0x00E060, 0xF6A3);
	ChangeU2MB(0x00E061, 0xF6A4);
	ChangeU2MB(0x00E062, 0xF6A5);
	ChangeU2MB(0x00E063, 0xF6A6);
	ChangeU2MB(0x00E064, 0xF6A7);
	ChangeU2MB(0x00E065, 0xF6A8);
	ChangeU2MB(0x00E066, 0xF6A9);
	ChangeU2MB(0x00E067, 0xF6AA);
	ChangeU2MB(0x00E068, 0xF6AB);
	ChangeU2MB(0x00E069, 0xF6AC);
	ChangeU2MB(0x00E06A, 0xF6AD);
	ChangeU2MB(0x00E06B, 0xF6AE);
	ChangeU2MB(0x00E06C, 0xF6AF);
	ChangeU2MB(0x00E06D, 0xF6B0);
	ChangeU2MB(0x00E06E, 0xF6B1);
	ChangeU2MB(0x00E06F, 0xF6B2);
	ChangeU2MB(0x00E070, 0xF6B3);
	ChangeU2MB(0x00E071, 0xF6B4);
	ChangeU2MB(0x00E072, 0xF6B5);
	ChangeU2MB(0x00E073, 0xF6B6);
	ChangeU2MB(0x00E074, 0xF6B7);
	ChangeU2MB(0x00E075, 0xF6B8);
	ChangeU2MB(0x00E076, 0xF6B9);
	ChangeU2MB(0x00E077, 0xF6BA);
	ChangeU2MB(0x00E078, 0xF6BB);
	ChangeU2MB(0x00E079, 0xF6BC);
	ChangeU2MB(0x00E07A, 0xF6BD);
	ChangeU2MB(0x00E07B, 0xF6BE);
	ChangeU2MB(0x00E07C, 0xF6BF);
	ChangeU2MB(0x00E07D, 0xF6C0);
	ChangeU2MB(0x00E07E, 0xF6C1);
	ChangeU2MB(0x00E07F, 0xF6C2);
	ChangeU2MB(0x00E080, 0xF6C3);
	ChangeU2MB(0x00E081, 0xF6C4);
	ChangeU2MB(0x00E082, 0xF6C5);
	ChangeU2MB(0x00E083, 0xF6C6);
	ChangeU2MB(0x00E084, 0xF6C7);
	ChangeU2MB(0x00E085, 0xF6C8);
	ChangeU2MB(0x00E086, 0xF6C9);
	ChangeU2MB(0x00E087, 0xF6CA);
	ChangeU2MB(0x00E088, 0xF6CB);
	ChangeU2MB(0x00E089, 0xF6CC);
	ChangeU2MB(0x00E08A, 0xF6CD);
	ChangeU2MB(0x00E08B, 0xF6CE);
	ChangeU2MB(0x00E08C, 0xF6CF);
	ChangeU2MB(0x00E08D, 0xF6D0);
	ChangeU2MB(0x00E08E, 0xF6D1);
	ChangeU2MB(0x00E08F, 0xF6D2);
	ChangeU2MB(0x00E090, 0xF6D3);
	ChangeU2MB(0x00E091, 0xF6D4);
	ChangeU2MB(0x00E092, 0xF6D5);
	ChangeU2MB(0x00E093, 0xF6D6);
	ChangeU2MB(0x00E094, 0xF6D7);
	ChangeU2MB(0x00E095, 0xF6D8);
	ChangeU2MB(0x00E096, 0xF6D9);
	ChangeU2MB(0x00E097, 0xF6DA);
	ChangeU2MB(0x00E098, 0xF6DB);
	ChangeU2MB(0x00E099, 0xF6DC);
	ChangeU2MB(0x00E09A, 0xF6DD);
	ChangeU2MB(0x00E09B, 0xF6DE);
	ChangeU2MB(0x00E09C, 0xF6DF);
	ChangeU2MB(0x00E09D, 0xF6E0);
	ChangeU2MB(0x00E09E, 0xF6E1);
	ChangeU2MB(0x00E09F, 0xF6E2);
	ChangeU2MB(0x00E0A0, 0xF6E3);
	ChangeU2MB(0x00E0A1, 0xF6E4);
	ChangeU2MB(0x00E0A2, 0xF6E5);
	ChangeU2MB(0x00E0A3, 0xF6E6);
	ChangeU2MB(0x00E0A4, 0xF6E7);
	ChangeU2MB(0x00E0A5, 0xF6E8);
	ChangeU2MB(0x00E0A6, 0xF6E9);
	ChangeU2MB(0x00E0A7, 0xF6EA);
	ChangeU2MB(0x00E0A8, 0xF6EB);
	ChangeU2MB(0x00E0A9, 0xF6EC);
	ChangeU2MB(0x00E0AA, 0xF6ED);
	ChangeU2MB(0x00E0AB, 0xF6EE);
	ChangeU2MB(0x00E0AC, 0xF6EF);
	ChangeU2MB(0x00E0AD, 0xF6F0);
	ChangeU2MB(0x00E0AE, 0xF6F1);
	ChangeU2MB(0x00E0AF, 0xF6F2);
	ChangeU2MB(0x00E0B0, 0xF6F3);
	ChangeU2MB(0x00E0B1, 0xF6F4);
	ChangeU2MB(0x00E0B2, 0xF6F5);
	ChangeU2MB(0x00E0B3, 0xF6F6);
	ChangeU2MB(0x00E0B4, 0xF6F7);
	ChangeU2MB(0x00E0B5, 0xF6F8);
	ChangeU2MB(0x00E0B6, 0xF6F9);
	ChangeU2MB(0x00E0B7, 0xF6FA);
	ChangeU2MB(0x00E0B8, 0xF6FB);
	ChangeU2MB(0x00E0B9, 0xF6FC);
	ChangeU2MB(0x00E0BA, 0xF6FD);
	ChangeU2MB(0x00E0BB, 0xF6FE);
	ChangeU2MB(0x00E0BC, 0xF7A1);
	ChangeU2MB(0x00E0BD, 0xF7A2);
	ChangeU2MB(0x00E0BE, 0xF7A3);
	ChangeU2MB(0x00E0BF, 0xF7A4);
	ChangeU2MB(0x00E0C0, 0xF7A5);
	ChangeU2MB(0x00E0C1, 0xF7A6);
	ChangeU2MB(0x00E0C2, 0xF7A7);
	ChangeU2MB(0x00E0C3, 0xF7A8);
	ChangeU2MB(0x00E0C4, 0xF7A9);
	ChangeU2MB(0x00E0C5, 0xF7AA);
	ChangeU2MB(0x00E0C6, 0xF7AB);
	ChangeU2MB(0x00E0C7, 0xF7AC);
	ChangeU2MB(0x00E0C8, 0xF7AD);
	ChangeU2MB(0x00E0C9, 0xF7AE);
	ChangeU2MB(0x00E0CA, 0xF7AF);
	ChangeU2MB(0x00E0CB, 0xF7B0);
	ChangeU2MB(0x00E0CC, 0xF7B1);
	ChangeU2MB(0x00E0CD, 0xF7B2);
	ChangeU2MB(0x00E0CE, 0xF7B3);
	ChangeU2MB(0x00E0CF, 0xF7B4);
	ChangeU2MB(0x00E0D0, 0xF7B5);
	ChangeU2MB(0x00E0D1, 0xF7B6);
	ChangeU2MB(0x00E0D2, 0xF7B7);
	ChangeU2MB(0x00E0D3, 0xF7B8);
	ChangeU2MB(0x00E0D4, 0xF7B9);
	ChangeU2MB(0x00E0D5, 0xF7BA);
	ChangeU2MB(0x00E0D6, 0xF7BB);
	ChangeU2MB(0x00E0D7, 0xF7BC);
	ChangeU2MB(0x00E0D8, 0xF7BD);
	ChangeU2MB(0x00E0D9, 0xF7BE);
	ChangeU2MB(0x00E0DA, 0xF7BF);
	ChangeU2MB(0x00E0DB, 0xF7C0);
	ChangeU2MB(0x00E0DC, 0xF7C1);
	ChangeU2MB(0x00E0DD, 0xF7C2);
	ChangeU2MB(0x00E0DE, 0xF7C3);
	ChangeU2MB(0x00E0DF, 0xF7C4);
	ChangeU2MB(0x00E0E0, 0xF7C5);
	ChangeU2MB(0x00E0E1, 0xF7C6);
	ChangeU2MB(0x00E0E2, 0xF7C7);
	ChangeU2MB(0x00E0E3, 0xF7C8);
	ChangeU2MB(0x00E0E4, 0xF7C9);
	ChangeU2MB(0x00E0E5, 0xF7CA);
	ChangeU2MB(0x00E0E6, 0xF7CB);
	ChangeU2MB(0x00E0E7, 0xF7CC);
	ChangeU2MB(0x00E0E8, 0xF7CD);
	ChangeU2MB(0x00E0E9, 0xF7CE);
	ChangeU2MB(0x00E0EA, 0xF7CF);
	ChangeU2MB(0x00E0EB, 0xF7D0);
	ChangeU2MB(0x00E0EC, 0xF7D1);
	ChangeU2MB(0x00E0ED, 0xF7D2);
	ChangeU2MB(0x00E0EE, 0xF7D3);
	ChangeU2MB(0x00E0EF, 0xF7D4);
	ChangeU2MB(0x00E0F0, 0xF7D5);
	ChangeU2MB(0x00E0F1, 0xF7D6);
	ChangeU2MB(0x00E0F2, 0xF7D7);
	ChangeU2MB(0x00E0F3, 0xF7D8);
	ChangeU2MB(0x00E0F4, 0xF7D9);
	ChangeU2MB(0x00E0F5, 0xF7DA);
	ChangeU2MB(0x00E0F6, 0xF7DB);
	ChangeU2MB(0x00E0F7, 0xF7DC);
	ChangeU2MB(0x00E0F8, 0xF7DD);
	ChangeU2MB(0x00E0F9, 0xF7DE);
	ChangeU2MB(0x00E0FA, 0xF7DF);
	ChangeU2MB(0x00E0FB, 0xF7E0);
	ChangeU2MB(0x00E0FC, 0xF7E1);
	ChangeU2MB(0x00E0FD, 0xF7E2);
	ChangeU2MB(0x00E0FE, 0xF7E3);
	ChangeU2MB(0x00E0FF, 0xF7E4);
	ChangeU2MB(0x00E100, 0xF7E5);
	ChangeU2MB(0x00E101, 0xF7E6);
	ChangeU2MB(0x00E102, 0xF7E7);
	ChangeU2MB(0x00E103, 0xF7E8);
	ChangeU2MB(0x00E104, 0xF7E9);
	ChangeU2MB(0x00E105, 0xF7EA);
	ChangeU2MB(0x00E106, 0xF7EB);
	ChangeU2MB(0x00E107, 0xF7EC);
	ChangeU2MB(0x00E108, 0xF7ED);
	ChangeU2MB(0x00E109, 0xF7EE);
	ChangeU2MB(0x00E10A, 0xF7EF);
	ChangeU2MB(0x00E10B, 0xF7F0);
	ChangeU2MB(0x00E10C, 0xF7F1);
	ChangeU2MB(0x00E10D, 0xF7F2);
	ChangeU2MB(0x00E10E, 0xF7F3);
	ChangeU2MB(0x00E10F, 0xF7F4);
	ChangeU2MB(0x00E110, 0xF7F5);
	ChangeU2MB(0x00E111, 0xF7F6);
	ChangeU2MB(0x00E112, 0xF7F7);
	ChangeU2MB(0x00E113, 0xF7F8);
	ChangeU2MB(0x00E114, 0xF7F9);
	ChangeU2MB(0x00E115, 0xF7FA);
	ChangeU2MB(0x00E116, 0xF7FB);
	ChangeU2MB(0x00E117, 0xF7FC);
	ChangeU2MB(0x00E118, 0xF7FD);
	ChangeU2MB(0x00E119, 0xF7FE);
	ChangeU2MB(0x00E11A, 0xF8A1);
	ChangeU2MB(0x00E11B, 0xF8A2);
	ChangeU2MB(0x00E11C, 0xF8A3);
	ChangeU2MB(0x00E11D, 0xF8A4);
	ChangeU2MB(0x00E11E, 0xF8A5);
	ChangeU2MB(0x00E11F, 0xF8A6);
	ChangeU2MB(0x00E120, 0xF8A7);
	ChangeU2MB(0x00E121, 0xF8A8);
	ChangeU2MB(0x00E122, 0xF8A9);
	ChangeU2MB(0x00E123, 0xF8AA);
	ChangeU2MB(0x00E124, 0xF8AB);
	ChangeU2MB(0x00E125, 0xF8AC);
	ChangeU2MB(0x00E126, 0xF8AD);
	ChangeU2MB(0x00E127, 0xF8AE);
	ChangeU2MB(0x00E128, 0xF8AF);
	ChangeU2MB(0x00E129, 0xF8B0);
	ChangeU2MB(0x00E12A, 0xF8B1);
	ChangeU2MB(0x00E12B, 0xF8B2);
	ChangeU2MB(0x00E12C, 0xF8B3);
	ChangeU2MB(0x00E12D, 0xF8B4);
	ChangeU2MB(0x00E12E, 0xF8B5);
	ChangeU2MB(0x00E12F, 0xF8B6);
	ChangeU2MB(0x00E130, 0xF8B7);
	ChangeU2MB(0x00E131, 0xF8B8);
	ChangeU2MB(0x00E132, 0xF8B9);
	ChangeU2MB(0x00E133, 0xF8BA);
	ChangeU2MB(0x00E134, 0xF8BB);
	ChangeU2MB(0x00E135, 0xF8BC);
	ChangeU2MB(0x00E136, 0xF8BD);
	ChangeU2MB(0x00E137, 0xF8BE);
	ChangeU2MB(0x00E138, 0xF8BF);
	ChangeU2MB(0x00E139, 0xF8C0);
	ChangeU2MB(0x00E13A, 0xF8C1);
	ChangeU2MB(0x00E13B, 0xF8C2);
	ChangeU2MB(0x00E13C, 0xF8C3);
	ChangeU2MB(0x00E13D, 0xF8C4);
	ChangeU2MB(0x00E13E, 0xF8C5);
	ChangeU2MB(0x00E13F, 0xF8C6);
	ChangeU2MB(0x00E140, 0xF8C7);
	ChangeU2MB(0x00E141, 0xF8C8);
	ChangeU2MB(0x00E142, 0xF8C9);
	ChangeU2MB(0x00E143, 0xF8CA);
	ChangeU2MB(0x00E144, 0xF8CB);
	ChangeU2MB(0x00E145, 0xF8CC);
	ChangeU2MB(0x00E146, 0xF8CD);
	ChangeU2MB(0x00E147, 0xF8CE);
	ChangeU2MB(0x00E148, 0xF8CF);
	ChangeU2MB(0x00E149, 0xF8D0);
	ChangeU2MB(0x00E14A, 0xF8D1);
	ChangeU2MB(0x00E14B, 0xF8D2);
	ChangeU2MB(0x00E14C, 0xF8D3);
	ChangeU2MB(0x00E14D, 0xF8D4);
	ChangeU2MB(0x00E14E, 0xF8D5);
	ChangeU2MB(0x00E14F, 0xF8D6);
	ChangeU2MB(0x00E150, 0xF8D7);
	ChangeU2MB(0x00E151, 0xF8D8);
	ChangeU2MB(0x00E152, 0xF8D9);
	ChangeU2MB(0x00E153, 0xF8DA);
	ChangeU2MB(0x00E154, 0xF8DB);
	ChangeU2MB(0x00E155, 0xF8DC);
	ChangeU2MB(0x00E156, 0xF8DD);
	ChangeU2MB(0x00E157, 0xF8DE);
	ChangeU2MB(0x00E158, 0xF8DF);
	ChangeU2MB(0x00E159, 0xF8E0);
	ChangeU2MB(0x00E15A, 0xF8E1);
	ChangeU2MB(0x00E15B, 0xF8E2);
	ChangeU2MB(0x00E15C, 0xF8E3);
	ChangeU2MB(0x00E15D, 0xF8E4);
	ChangeU2MB(0x00E15E, 0xF8E5);
	ChangeU2MB(0x00E15F, 0xF8E6);
	ChangeU2MB(0x00E160, 0xF8E7);
	ChangeU2MB(0x00E161, 0xF8E8);
	ChangeU2MB(0x00E162, 0xF8E9);
	ChangeU2MB(0x00E163, 0xF8EA);
	ChangeU2MB(0x00E164, 0xF8EB);
	ChangeU2MB(0x00E165, 0xF8EC);
	ChangeU2MB(0x00E166, 0xF8ED);
	ChangeU2MB(0x00E167, 0xF8EE);
	ChangeU2MB(0x00E168, 0xF8EF);
	ChangeU2MB(0x00E169, 0xF8F0);
	ChangeU2MB(0x00E16A, 0xF8F1);
	ChangeU2MB(0x00E16B, 0xF8F2);
	ChangeU2MB(0x00E16C, 0xF8F3);
	ChangeU2MB(0x00E16D, 0xF8F4);
	ChangeU2MB(0x00E16E, 0xF8F5);
	ChangeU2MB(0x00E16F, 0xF8F6);
	ChangeU2MB(0x00E170, 0xF8F7);
	ChangeU2MB(0x00E171, 0xF8F8);
	ChangeU2MB(0x00E172, 0xF8F9);
	ChangeU2MB(0x00E173, 0xF8FA);
	ChangeU2MB(0x00E174, 0xF8FB);
	ChangeU2MB(0x00E175, 0xF8FC);
	ChangeU2MB(0x00E176, 0xF8FD);
	ChangeU2MB(0x00E177, 0xF8FE);
	ChangeU2MB(0x00E178, 0xF9A1);
	ChangeU2MB(0x00E179, 0xF9A2);
	ChangeU2MB(0x00E17A, 0xF9A3);
	ChangeU2MB(0x00E17B, 0xF9A4);
	ChangeU2MB(0x00E17C, 0xF9A5);
	ChangeU2MB(0x00E17D, 0xF9A6);
	ChangeU2MB(0x00E17E, 0xF9A7);
	ChangeU2MB(0x00E17F, 0xF9A8);
	ChangeU2MB(0x00E180, 0xF9A9);
	ChangeU2MB(0x00E181, 0xF9AA);
	ChangeU2MB(0x00E182, 0xF9AB);
	ChangeU2MB(0x00E183, 0xF9AC);
	ChangeU2MB(0x00E184, 0xF9AD);
	ChangeU2MB(0x00E185, 0xF9AE);
	ChangeU2MB(0x00E186, 0xF9AF);
	ChangeU2MB(0x00E187, 0xF9B0);
	ChangeU2MB(0x00E188, 0xF9B1);
	ChangeU2MB(0x00E189, 0xF9B2);
	ChangeU2MB(0x00E18A, 0xF9B3);
	ChangeU2MB(0x00E18B, 0xF9B4);
	ChangeU2MB(0x00E18C, 0xF9B5);
	ChangeU2MB(0x00E18D, 0xF9B6);
	ChangeU2MB(0x00E18E, 0xF9B7);
	ChangeU2MB(0x00E18F, 0xF9B8);
	ChangeU2MB(0x00E190, 0xF9B9);
	ChangeU2MB(0x00E191, 0xF9BA);
	ChangeU2MB(0x00E192, 0xF9BB);
	ChangeU2MB(0x00E193, 0xF9BC);
	ChangeU2MB(0x00E194, 0xF9BD);
	ChangeU2MB(0x00E195, 0xF9BE);
	ChangeU2MB(0x00E196, 0xF9BF);
	ChangeU2MB(0x00E197, 0xF9C0);
	ChangeU2MB(0x00E198, 0xF9C1);
	ChangeU2MB(0x00E199, 0xF9C2);
	ChangeU2MB(0x00E19A, 0xF9C3);
	ChangeU2MB(0x00E19B, 0xF9C4);
	ChangeU2MB(0x00E19C, 0xF9C5);
	ChangeU2MB(0x00E19D, 0xF9C6);
	ChangeU2MB(0x00E19E, 0xF9C7);
	ChangeU2MB(0x00E19F, 0xF9C8);
	ChangeU2MB(0x00E1A0, 0xF9C9);
	ChangeU2MB(0x00E1A1, 0xF9CA);
	ChangeU2MB(0x00E1A2, 0xF9CB);
	ChangeU2MB(0x00E1A3, 0xF9CC);
	ChangeU2MB(0x00E1A4, 0xF9CD);
	ChangeU2MB(0x00E1A5, 0xF9CE);
	ChangeU2MB(0x00E1A6, 0xF9CF);
	ChangeU2MB(0x00E1A7, 0xF9D0);
	ChangeU2MB(0x00E1A8, 0xF9D1);
	ChangeU2MB(0x00E1A9, 0xF9D2);
	ChangeU2MB(0x00E1AA, 0xF9D3);
	ChangeU2MB(0x00E1AB, 0xF9D4);
	ChangeU2MB(0x00E1AC, 0xF9D5);
	ChangeU2MB(0x00E1AD, 0xF9D6);
	ChangeU2MB(0x00E1AE, 0xF9D7);
	ChangeU2MB(0x00E1AF, 0xF9D8);
	ChangeU2MB(0x00E1B0, 0xF9D9);
	ChangeU2MB(0x00E1B1, 0xF9DA);
	ChangeU2MB(0x00E1B2, 0xF9DB);
	ChangeU2MB(0x00E1B3, 0xF9DC);
	ChangeU2MB(0x00E1B4, 0xF9DD);
	ChangeU2MB(0x00E1B5, 0xF9DE);
	ChangeU2MB(0x00E1B6, 0xF9DF);
	ChangeU2MB(0x00E1B7, 0xF9E0);
	ChangeU2MB(0x00E1B8, 0xF9E1);
	ChangeU2MB(0x00E1B9, 0xF9E2);
	ChangeU2MB(0x00E1BA, 0xF9E3);
	ChangeU2MB(0x00E1BB, 0xF9E4);
	ChangeU2MB(0x00E1BC, 0xF9E5);
	ChangeU2MB(0x00E1BD, 0xF9E6);
	ChangeU2MB(0x00E1BE, 0xF9E7);
	ChangeU2MB(0x00E1BF, 0xF9E8);
	ChangeU2MB(0x00E1C0, 0xF9E9);
	ChangeU2MB(0x00E1C1, 0xF9EA);
	ChangeU2MB(0x00E1C2, 0xF9EB);
	ChangeU2MB(0x00E1C3, 0xF9EC);
	ChangeU2MB(0x00E1C4, 0xF9ED);
	ChangeU2MB(0x00E1C5, 0xF9EE);
	ChangeU2MB(0x00E1C6, 0xF9EF);
	ChangeU2MB(0x00E1C7, 0xF9F0);
	ChangeU2MB(0x00E1C8, 0xF9F1);
	ChangeU2MB(0x00E1C9, 0xF9F2);
	ChangeU2MB(0x00E1CA, 0xF9F3);
	ChangeU2MB(0x00E1CB, 0xF9F4);
	ChangeU2MB(0x00E1CC, 0xF9F5);
	ChangeU2MB(0x00E1CD, 0xF9F6);
	ChangeU2MB(0x00E1CE, 0xF9F7);
	ChangeU2MB(0x00E1CF, 0xF9F8);
	ChangeU2MB(0x00E1D0, 0xF9F9);
	ChangeU2MB(0x00E1D1, 0xF9FA);
	ChangeU2MB(0x00E1D2, 0xF9FB);
	ChangeU2MB(0x00E1D3, 0xF9FC);
	ChangeU2MB(0x00E1D4, 0xF9FD);
	ChangeU2MB(0x00E1D5, 0xF9FE);
	ChangeU2MB(0x00E1D6, 0xFAA1);
	ChangeU2MB(0x00E1D7, 0xFAA2);
	ChangeU2MB(0x00E1D8, 0xFAA3);
	ChangeU2MB(0x00E1D9, 0xFAA4);
	ChangeU2MB(0x00E1DA, 0xFAA5);
	ChangeU2MB(0x00E1DB, 0xFAA6);
	ChangeU2MB(0x00E1DC, 0xFAA7);
	ChangeU2MB(0x00E1DD, 0xFAA8);
	ChangeU2MB(0x00E1DE, 0xFAA9);
	ChangeU2MB(0x00E1DF, 0xFAAA);
	ChangeU2MB(0x00E1E0, 0xFAAB);
	ChangeU2MB(0x00E1E1, 0xFAAC);
	ChangeU2MB(0x00E1E2, 0xFAAD);
	ChangeU2MB(0x00E1E3, 0xFAAE);
	ChangeU2MB(0x00E1E4, 0xFAAF);
	ChangeU2MB(0x00E1E5, 0xFAB0);
	ChangeU2MB(0x00E1E6, 0xFAB1);
	ChangeU2MB(0x00E1E7, 0xFAB2);
	ChangeU2MB(0x00E1E8, 0xFAB3);
	ChangeU2MB(0x00E1E9, 0xFAB4);
	ChangeU2MB(0x00E1EA, 0xFAB5);
	ChangeU2MB(0x00E1EB, 0xFAB6);
	ChangeU2MB(0x00E1EC, 0xFAB7);
	ChangeU2MB(0x00E1ED, 0xFAB8);
	ChangeU2MB(0x00E1EE, 0xFAB9);
	ChangeU2MB(0x00E1EF, 0xFABA);
	ChangeU2MB(0x00E1F0, 0xFABB);
	ChangeU2MB(0x00E1F1, 0xFABC);
	ChangeU2MB(0x00E1F2, 0xFABD);
	ChangeU2MB(0x00E1F3, 0xFABE);
	ChangeU2MB(0x00E1F4, 0xFABF);
	ChangeU2MB(0x00E1F5, 0xFAC0);
	ChangeU2MB(0x00E1F6, 0xFAC1);
	ChangeU2MB(0x00E1F7, 0xFAC2);
	ChangeU2MB(0x00E1F8, 0xFAC3);
	ChangeU2MB(0x00E1F9, 0xFAC4);
	ChangeU2MB(0x00E1FA, 0xFAC5);
	ChangeU2MB(0x00E1FB, 0xFAC6);
	ChangeU2MB(0x00E1FC, 0xFAC7);
	ChangeU2MB(0x00E1FD, 0xFAC8);
	ChangeU2MB(0x00E1FE, 0xFAC9);
	ChangeU2MB(0x00E1FF, 0xFACA);
	ChangeU2MB(0x00E200, 0xFACB);
	ChangeU2MB(0x00E201, 0xFACC);
	ChangeU2MB(0x00E202, 0xFACD);
	ChangeU2MB(0x00E203, 0xFACE);
	ChangeU2MB(0x00E204, 0xFACF);
	ChangeU2MB(0x00E205, 0xFAD0);
	ChangeU2MB(0x00E206, 0xFAD1);
	ChangeU2MB(0x00E207, 0xFAD2);
	ChangeU2MB(0x00E208, 0xFAD3);
	ChangeU2MB(0x00E209, 0xFAD4);
	ChangeU2MB(0x00E20A, 0xFAD5);
	ChangeU2MB(0x00E20B, 0xFAD6);
	ChangeU2MB(0x00E20C, 0xFAD7);
	ChangeU2MB(0x00E20D, 0xFAD8);
	ChangeU2MB(0x00E20E, 0xFAD9);
	ChangeU2MB(0x00E20F, 0xFADA);
	ChangeU2MB(0x00E210, 0xFADB);
	ChangeU2MB(0x00E211, 0xFADC);
	ChangeU2MB(0x00E212, 0xFADD);
	ChangeU2MB(0x00E213, 0xFADE);
	ChangeU2MB(0x00E214, 0xFADF);
	ChangeU2MB(0x00E215, 0xFAE0);
	ChangeU2MB(0x00E216, 0xFAE1);
	ChangeU2MB(0x00E217, 0xFAE2);
	ChangeU2MB(0x00E218, 0xFAE3);
	ChangeU2MB(0x00E219, 0xFAE4);
	ChangeU2MB(0x00E21A, 0xFAE5);
	ChangeU2MB(0x00E21B, 0xFAE6);
	ChangeU2MB(0x00E21C, 0xFAE7);
	ChangeU2MB(0x00E21D, 0xFAE8);
	ChangeU2MB(0x00E21E, 0xFAE9);
	ChangeU2MB(0x00E21F, 0xFAEA);
	ChangeU2MB(0x00E220, 0xFAEB);
	ChangeU2MB(0x00E221, 0xFAEC);
	ChangeU2MB(0x00E222, 0xFAED);
	ChangeU2MB(0x00E223, 0xFAEE);
	ChangeU2MB(0x00E224, 0xFAEF);
	ChangeU2MB(0x00E225, 0xFAF0);
	ChangeU2MB(0x00E226, 0xFAF1);
	ChangeU2MB(0x00E227, 0xFAF2);
	ChangeU2MB(0x00E228, 0xFAF3);
	ChangeU2MB(0x00E229, 0xFAF4);
	ChangeU2MB(0x00E22A, 0xFAF5);
	ChangeU2MB(0x00E22B, 0xFAF6);
	ChangeU2MB(0x00E22C, 0xFAF7);
	ChangeU2MB(0x00E22D, 0xFAF8);
	ChangeU2MB(0x00E22E, 0xFAF9);
	ChangeU2MB(0x00E22F, 0xFAFA);
	ChangeU2MB(0x00E230, 0xFAFB);
	ChangeU2MB(0x00E231, 0xFAFC);
	ChangeU2MB(0x00E232, 0xFAFD);
	ChangeU2MB(0x00E233, 0xFAFE);
	ChangeU2MB(0x00E234, 0xFBA1);
	ChangeU2MB(0x00E235, 0xFBA2);
	ChangeU2MB(0x00E236, 0xFBA3);
	ChangeU2MB(0x00E237, 0xFBA4);
	ChangeU2MB(0x00E238, 0xFBA5);
	ChangeU2MB(0x00E239, 0xFBA6);
	ChangeU2MB(0x00E23A, 0xFBA7);
	ChangeU2MB(0x00E23B, 0xFBA8);
	ChangeU2MB(0x00E23C, 0xFBA9);
	ChangeU2MB(0x00E23D, 0xFBAA);
	ChangeU2MB(0x00E23E, 0xFBAB);
	ChangeU2MB(0x00E23F, 0xFBAC);
	ChangeU2MB(0x00E240, 0xFBAD);
	ChangeU2MB(0x00E241, 0xFBAE);
	ChangeU2MB(0x00E242, 0xFBAF);
	ChangeU2MB(0x00E243, 0xFBB0);
	ChangeU2MB(0x00E244, 0xFBB1);
	ChangeU2MB(0x00E245, 0xFBB2);
	ChangeU2MB(0x00E246, 0xFBB3);
	ChangeU2MB(0x00E247, 0xFBB4);
	ChangeU2MB(0x00E248, 0xFBB5);
	ChangeU2MB(0x00E249, 0xFBB6);
	ChangeU2MB(0x00E24A, 0xFBB7);
	ChangeU2MB(0x00E24B, 0xFBB8);
	ChangeU2MB(0x00E24C, 0xFBB9);
	ChangeU2MB(0x00E24D, 0xFBBA);
	ChangeU2MB(0x00E24E, 0xFBBB);
	ChangeU2MB(0x00E24F, 0xFBBC);
	ChangeU2MB(0x00E250, 0xFBBD);
	ChangeU2MB(0x00E251, 0xFBBE);
	ChangeU2MB(0x00E252, 0xFBBF);
	ChangeU2MB(0x00E253, 0xFBC0);
	ChangeU2MB(0x00E254, 0xFBC1);
	ChangeU2MB(0x00E255, 0xFBC2);
	ChangeU2MB(0x00E256, 0xFBC3);
	ChangeU2MB(0x00E257, 0xFBC4);
	ChangeU2MB(0x00E258, 0xFBC5);
	ChangeU2MB(0x00E259, 0xFBC6);
	ChangeU2MB(0x00E25A, 0xFBC7);
	ChangeU2MB(0x00E25B, 0xFBC8);
	ChangeU2MB(0x00E25C, 0xFBC9);
	ChangeU2MB(0x00E25D, 0xFBCA);
	ChangeU2MB(0x00E25E, 0xFBCB);
	ChangeU2MB(0x00E25F, 0xFBCC);
	ChangeU2MB(0x00E260, 0xFBCD);
	ChangeU2MB(0x00E261, 0xFBCE);
	ChangeU2MB(0x00E262, 0xFBCF);
	ChangeU2MB(0x00E263, 0xFBD0);
	ChangeU2MB(0x00E264, 0xFBD1);
	ChangeU2MB(0x00E265, 0xFBD2);
	ChangeU2MB(0x00E266, 0xFBD3);
	ChangeU2MB(0x00E267, 0xFBD4);
	ChangeU2MB(0x00E268, 0xFBD5);
	ChangeU2MB(0x00E269, 0xFBD6);
	ChangeU2MB(0x00E26A, 0xFBD7);
	ChangeU2MB(0x00E26B, 0xFBD8);
	ChangeU2MB(0x00E26C, 0xFBD9);
	ChangeU2MB(0x00E26D, 0xFBDA);
	ChangeU2MB(0x00E26E, 0xFBDB);
	ChangeU2MB(0x00E26F, 0xFBDC);
	ChangeU2MB(0x00E270, 0xFBDD);
	ChangeU2MB(0x00E271, 0xFBDE);
	ChangeU2MB(0x00E272, 0xFBDF);
	ChangeU2MB(0x00E273, 0xFBE0);
	ChangeU2MB(0x00E274, 0xFBE1);
	ChangeU2MB(0x00E275, 0xFBE2);
	ChangeU2MB(0x00E276, 0xFBE3);
	ChangeU2MB(0x00E277, 0xFBE4);
	ChangeU2MB(0x00E278, 0xFBE5);
	ChangeU2MB(0x00E279, 0xFBE6);
	ChangeU2MB(0x00E27A, 0xFBE7);
	ChangeU2MB(0x00E27B, 0xFBE8);
	ChangeU2MB(0x00E27C, 0xFBE9);
	ChangeU2MB(0x00E27D, 0xFBEA);
	ChangeU2MB(0x00E27E, 0xFBEB);
	ChangeU2MB(0x00E27F, 0xFBEC);
	ChangeU2MB(0x00E280, 0xFBED);
	ChangeU2MB(0x00E281, 0xFBEE);
	ChangeU2MB(0x00E282, 0xFBEF);
	ChangeU2MB(0x00E283, 0xFBF0);
	ChangeU2MB(0x00E284, 0xFBF1);
	ChangeU2MB(0x00E285, 0xFBF2);
	ChangeU2MB(0x00E286, 0xFBF3);
	ChangeU2MB(0x00E287, 0xFBF4);
	ChangeU2MB(0x00E288, 0xFBF5);
	ChangeU2MB(0x00E289, 0xFBF6);
	ChangeU2MB(0x00E28A, 0xFBF7);
	ChangeU2MB(0x00E28B, 0xFBF8);
	ChangeU2MB(0x00E28C, 0xFBF9);
	ChangeU2MB(0x00E28D, 0xFBFA);
	ChangeU2MB(0x00E28E, 0xFBFB);
	ChangeU2MB(0x00E28F, 0xFBFC);
	ChangeU2MB(0x00E290, 0xFBFD);
	ChangeU2MB(0x00E291, 0xFBFE);
	ChangeU2MB(0x00E292, 0xFCA1);
	ChangeU2MB(0x00E293, 0xFCA2);
	ChangeU2MB(0x00E294, 0xFCA3);
	ChangeU2MB(0x00E295, 0xFCA4);
	ChangeU2MB(0x00E296, 0xFCA5);
	ChangeU2MB(0x00E297, 0xFCA6);
	ChangeU2MB(0x00E298, 0xFCA7);
	ChangeU2MB(0x00E299, 0xFCA8);
	ChangeU2MB(0x00E29A, 0xFCA9);
	ChangeU2MB(0x00E29B, 0xFCAA);
	ChangeU2MB(0x00E29C, 0xFCAB);
	ChangeU2MB(0x00E29D, 0xFCAC);
	ChangeU2MB(0x00E29E, 0xFCAD);
	ChangeU2MB(0x00E29F, 0xFCAE);
	ChangeU2MB(0x00E2A0, 0xFCAF);
	ChangeU2MB(0x00E2A1, 0xFCB0);
	ChangeU2MB(0x00E2A2, 0xFCB1);
	ChangeU2MB(0x00E2A3, 0xFCB2);
	ChangeU2MB(0x00E2A4, 0xFCB3);
	ChangeU2MB(0x00E2A5, 0xFCB4);
	ChangeU2MB(0x00E2A6, 0xFCB5);
	ChangeU2MB(0x00E2A7, 0xFCB6);
	ChangeU2MB(0x00E2A8, 0xFCB7);
	ChangeU2MB(0x00E2A9, 0xFCB8);
	ChangeU2MB(0x00E2AA, 0xFCB9);
	ChangeU2MB(0x00E2AB, 0xFCBA);
	ChangeU2MB(0x00E2AC, 0xFCBB);
	ChangeU2MB(0x00E2AD, 0xFCBC);
	ChangeU2MB(0x00E2AE, 0xFCBD);
	ChangeU2MB(0x00E2AF, 0xFCBE);
	ChangeU2MB(0x00E2B0, 0xFCBF);
	ChangeU2MB(0x00E2B1, 0xFCC0);
	ChangeU2MB(0x00E2B2, 0xFCC1);
	ChangeU2MB(0x00E2B3, 0xFCC2);
	ChangeU2MB(0x00E2B4, 0xFCC3);
	ChangeU2MB(0x00E2B5, 0xFCC4);
	ChangeU2MB(0x00E2B6, 0xFCC5);
	ChangeU2MB(0x00E2B7, 0xFCC6);
	ChangeU2MB(0x00E2B8, 0xFCC7);
	ChangeU2MB(0x00E2B9, 0xFCC8);
	ChangeU2MB(0x00E2BA, 0xFCC9);
	ChangeU2MB(0x00E2BB, 0xFCCA);
	ChangeU2MB(0x00E2BC, 0xFCCB);
	ChangeU2MB(0x00E2BD, 0xFCCC);
	ChangeU2MB(0x00E2BE, 0xFCCD);
	ChangeU2MB(0x00E2BF, 0xFCCE);
	ChangeU2MB(0x00E2C0, 0xFCCF);
	ChangeU2MB(0x00E2C1, 0xFCD0);
	ChangeU2MB(0x00E2C2, 0xFCD1);
	ChangeU2MB(0x00E2C3, 0xFCD2);
	ChangeU2MB(0x00E2C4, 0xFCD3);
	ChangeU2MB(0x00E2C5, 0xFCD4);
	ChangeU2MB(0x00E2C6, 0xFCD5);
	ChangeU2MB(0x00E2C7, 0xFCD6);
	ChangeU2MB(0x00E2C8, 0xFCD7);
	ChangeU2MB(0x00E2C9, 0xFCD8);
	ChangeU2MB(0x00E2CA, 0xFCD9);
	ChangeU2MB(0x00E2CB, 0xFCDA);
	ChangeU2MB(0x00E2CC, 0xFCDB);
	ChangeU2MB(0x00E2CD, 0xFCDC);
	ChangeU2MB(0x00E2CE, 0xFCDD);
	ChangeU2MB(0x00E2CF, 0xFCDE);
	ChangeU2MB(0x00E2D0, 0xFCDF);
	ChangeU2MB(0x00E2D1, 0xFCE0);
	ChangeU2MB(0x00E2D2, 0xFCE1);
	ChangeU2MB(0x00E2D3, 0xFCE2);
	ChangeU2MB(0x00E2D4, 0xFCE3);
	ChangeU2MB(0x00E2D5, 0xFCE4);
	ChangeU2MB(0x00E2D6, 0xFCE5);
	ChangeU2MB(0x00E2D7, 0xFCE6);
	ChangeU2MB(0x00E2D8, 0xFCE7);
	ChangeU2MB(0x00E2D9, 0xFCE8);
	ChangeU2MB(0x00E2DA, 0xFCE9);
	ChangeU2MB(0x00E2DB, 0xFCEA);
	ChangeU2MB(0x00E2DC, 0xFCEB);
	ChangeU2MB(0x00E2DD, 0xFCEC);
	ChangeU2MB(0x00E2DE, 0xFCED);
	ChangeU2MB(0x00E2DF, 0xFCEE);
	ChangeU2MB(0x00E2E0, 0xFCEF);
	ChangeU2MB(0x00E2E1, 0xFCF0);
	ChangeU2MB(0x00E2E2, 0xFCF1);
	ChangeU2MB(0x00E2E3, 0xFCF2);
	ChangeU2MB(0x00E2E4, 0xFCF3);
	ChangeU2MB(0x00E2E5, 0xFCF4);
	ChangeU2MB(0x00E2E6, 0xFCF5);
	ChangeU2MB(0x00E2E7, 0xFCF6);
	ChangeU2MB(0x00E2E8, 0xFCF7);
	ChangeU2MB(0x00E2E9, 0xFCF8);
	ChangeU2MB(0x00E2EA, 0xFCF9);
	ChangeU2MB(0x00E2EB, 0xFCFA);
	ChangeU2MB(0x00E2EC, 0xFCFB);
	ChangeU2MB(0x00E2ED, 0xFCFC);
	ChangeU2MB(0x00E2EE, 0xFCFD);
	ChangeU2MB(0x00E2EF, 0xFCFE);
	ChangeU2MB(0x00E2F0, 0xFDA1);
	ChangeU2MB(0x00E2F1, 0xFDA2);
	ChangeU2MB(0x00E2F2, 0xFDA3);
	ChangeU2MB(0x00E2F3, 0xFDA4);
	ChangeU2MB(0x00E2F4, 0xFDA5);
	ChangeU2MB(0x00E2F5, 0xFDA6);
	ChangeU2MB(0x00E2F6, 0xFDA7);
	ChangeU2MB(0x00E2F7, 0xFDA8);
	ChangeU2MB(0x00E2F8, 0xFDA9);
	ChangeU2MB(0x00E2F9, 0xFDAA);
	ChangeU2MB(0x00E2FA, 0xFDAB);
	ChangeU2MB(0x00E2FB, 0xFDAC);
	ChangeU2MB(0x00E2FC, 0xFDAD);
	ChangeU2MB(0x00E2FD, 0xFDAE);
	ChangeU2MB(0x00E2FE, 0xFDAF);
	ChangeU2MB(0x00E2FF, 0xFDB0);
	ChangeU2MB(0x00E300, 0xFDB1);
	ChangeU2MB(0x00E301, 0xFDB2);
	ChangeU2MB(0x00E302, 0xFDB3);
	ChangeU2MB(0x00E303, 0xFDB4);
	ChangeU2MB(0x00E304, 0xFDB5);
	ChangeU2MB(0x00E305, 0xFDB6);
	ChangeU2MB(0x00E306, 0xFDB7);
	ChangeU2MB(0x00E307, 0xFDB8);
	ChangeU2MB(0x00E308, 0xFDB9);
	ChangeU2MB(0x00E309, 0xFDBA);
	ChangeU2MB(0x00E30A, 0xFDBB);
	ChangeU2MB(0x00E30B, 0xFDBC);
	ChangeU2MB(0x00E30C, 0xFDBD);
	ChangeU2MB(0x00E30D, 0xFDBE);
	ChangeU2MB(0x00E30E, 0xFDBF);
	ChangeU2MB(0x00E30F, 0xFDC0);
	ChangeU2MB(0x00E310, 0xFDC1);
	ChangeU2MB(0x00E311, 0xFDC2);
	ChangeU2MB(0x00E312, 0xFDC3);
	ChangeU2MB(0x00E313, 0xFDC4);
	ChangeU2MB(0x00E314, 0xFDC5);
	ChangeU2MB(0x00E315, 0xFDC6);
	ChangeU2MB(0x00E316, 0xFDC7);
	ChangeU2MB(0x00E317, 0xFDC8);
	ChangeU2MB(0x00E318, 0xFDC9);
	ChangeU2MB(0x00E319, 0xFDCA);
	ChangeU2MB(0x00E31A, 0xFDCB);
	ChangeU2MB(0x00E31B, 0xFDCC);
	ChangeU2MB(0x00E31C, 0xFDCD);
	ChangeU2MB(0x00E31D, 0xFDCE);
	ChangeU2MB(0x00E31E, 0xFDCF);
	ChangeU2MB(0x00E31F, 0xFDD0);
	ChangeU2MB(0x00E320, 0xFDD1);
	ChangeU2MB(0x00E321, 0xFDD2);
	ChangeU2MB(0x00E322, 0xFDD3);
	ChangeU2MB(0x00E323, 0xFDD4);
	ChangeU2MB(0x00E324, 0xFDD5);
	ChangeU2MB(0x00E325, 0xFDD6);
	ChangeU2MB(0x00E326, 0xFDD7);
	ChangeU2MB(0x00E327, 0xFDD8);
	ChangeU2MB(0x00E328, 0xFDD9);
	ChangeU2MB(0x00E329, 0xFDDA);
	ChangeU2MB(0x00E32A, 0xFDDB);
	ChangeU2MB(0x00E32B, 0xFDDC);
	ChangeU2MB(0x00E32C, 0xFDDD);
	ChangeU2MB(0x00E32D, 0xFDDE);
	ChangeU2MB(0x00E32E, 0xFDDF);
	ChangeU2MB(0x00E32F, 0xFDE0);
	ChangeU2MB(0x00E330, 0xFDE1);
	ChangeU2MB(0x00E331, 0xFDE2);
	ChangeU2MB(0x00E332, 0xFDE3);
	ChangeU2MB(0x00E333, 0xFDE4);
	ChangeU2MB(0x00E334, 0xFDE5);
	ChangeU2MB(0x00E335, 0xFDE6);
	ChangeU2MB(0x00E336, 0xFDE7);
	ChangeU2MB(0x00E337, 0xFDE8);
	ChangeU2MB(0x00E338, 0xFDE9);
	ChangeU2MB(0x00E339, 0xFDEA);
	ChangeU2MB(0x00E33A, 0xFDEB);
	ChangeU2MB(0x00E33B, 0xFDEC);
	ChangeU2MB(0x00E33C, 0xFDED);
	ChangeU2MB(0x00E33D, 0xFDEE);
	ChangeU2MB(0x00E33E, 0xFDEF);
	ChangeU2MB(0x00E33F, 0xFDF0);
	ChangeU2MB(0x00E340, 0xFDF1);
	ChangeU2MB(0x00E341, 0xFDF2);
	ChangeU2MB(0x00E342, 0xFDF3);
	ChangeU2MB(0x00E343, 0xFDF4);
	ChangeU2MB(0x00E344, 0xFDF5);
	ChangeU2MB(0x00E345, 0xFDF6);
	ChangeU2MB(0x00E346, 0xFDF7);
	ChangeU2MB(0x00E347, 0xFDF8);
	ChangeU2MB(0x00E348, 0xFDF9);
	ChangeU2MB(0x00E349, 0xFDFA);
	ChangeU2MB(0x00E34A, 0xFDFB);
	ChangeU2MB(0x00E34B, 0xFDFC);
	ChangeU2MB(0x00E34C, 0xFDFD);
	ChangeU2MB(0x00E34D, 0xFDFE);
	ChangeU2MB(0x00E34E, 0xFEA1);
	ChangeU2MB(0x00E34F, 0xFEA2);
	ChangeU2MB(0x00E350, 0xFEA3);
	ChangeU2MB(0x00E351, 0xFEA4);
	ChangeU2MB(0x00E352, 0xFEA5);
	ChangeU2MB(0x00E353, 0xFEA6);
	ChangeU2MB(0x00E354, 0xFEA7);
	ChangeU2MB(0x00E355, 0xFEA8);
	ChangeU2MB(0x00E356, 0xFEA9);
	ChangeU2MB(0x00E357, 0xFEAA);
	ChangeU2MB(0x00E358, 0xFEAB);
	ChangeU2MB(0x00E359, 0xFEAC);
	ChangeU2MB(0x00E35A, 0xFEAD);
	ChangeU2MB(0x00E35B, 0xFEAE);
	ChangeU2MB(0x00E35C, 0xFEAF);
	ChangeU2MB(0x00E35D, 0xFEB0);
	ChangeU2MB(0x00E35E, 0xFEB1);
	ChangeU2MB(0x00E35F, 0xFEB2);
	ChangeU2MB(0x00E360, 0xFEB3);
	ChangeU2MB(0x00E361, 0xFEB4);
	ChangeU2MB(0x00E362, 0xFEB5);
	ChangeU2MB(0x00E363, 0xFEB6);
	ChangeU2MB(0x00E364, 0xFEB7);
	ChangeU2MB(0x00E365, 0xFEB8);
	ChangeU2MB(0x00E366, 0xFEB9);
	ChangeU2MB(0x00E367, 0xFEBA);
	ChangeU2MB(0x00E368, 0xFEBB);
	ChangeU2MB(0x00E369, 0xFEBC);
	ChangeU2MB(0x00E36A, 0xFEBD);
	ChangeU2MB(0x00E36B, 0xFEBE);
	ChangeU2MB(0x00E36C, 0xFEBF);
	ChangeU2MB(0x00E36D, 0xFEC0);
	ChangeU2MB(0x00E36E, 0xFEC1);
	ChangeU2MB(0x00E36F, 0xFEC2);
	ChangeU2MB(0x00E370, 0xFEC3);
	ChangeU2MB(0x00E371, 0xFEC4);
	ChangeU2MB(0x00E372, 0xFEC5);
	ChangeU2MB(0x00E373, 0xFEC6);
	ChangeU2MB(0x00E374, 0xFEC7);
	ChangeU2MB(0x00E375, 0xFEC8);
	ChangeU2MB(0x00E376, 0xFEC9);
	ChangeU2MB(0x00E377, 0xFECA);
	ChangeU2MB(0x00E378, 0xFECB);
	ChangeU2MB(0x00E379, 0xFECC);
	ChangeU2MB(0x00E37A, 0xFECD);
	ChangeU2MB(0x00E37B, 0xFECE);
	ChangeU2MB(0x00E37C, 0xFECF);
	ChangeU2MB(0x00E37D, 0xFED0);
	ChangeU2MB(0x00E37E, 0xFED1);
	ChangeU2MB(0x00E37F, 0xFED2);
	ChangeU2MB(0x00E380, 0xFED3);
	ChangeU2MB(0x00E381, 0xFED4);
	ChangeU2MB(0x00E382, 0xFED5);
	ChangeU2MB(0x00E383, 0xFED6);
	ChangeU2MB(0x00E384, 0xFED7);
	ChangeU2MB(0x00E385, 0xFED8);
	ChangeU2MB(0x00E386, 0xFED9);
	ChangeU2MB(0x00E387, 0xFEDA);
	ChangeU2MB(0x00E388, 0xFEDB);
	ChangeU2MB(0x00E389, 0xFEDC);
	ChangeU2MB(0x00E38A, 0xFEDD);
	ChangeU2MB(0x00E38B, 0xFEDE);
	ChangeU2MB(0x00E38C, 0xFEDF);
	ChangeU2MB(0x00E38D, 0xFEE0);
	ChangeU2MB(0x00E38E, 0xFEE1);
	ChangeU2MB(0x00E38F, 0xFEE2);
	ChangeU2MB(0x00E390, 0xFEE3);
	ChangeU2MB(0x00E391, 0xFEE4);
	ChangeU2MB(0x00E392, 0xFEE5);
	ChangeU2MB(0x00E393, 0xFEE6);
	ChangeU2MB(0x00E394, 0xFEE7);
	ChangeU2MB(0x00E395, 0xFEE8);
	ChangeU2MB(0x00E396, 0xFEE9);
	ChangeU2MB(0x00E397, 0xFEEA);
	ChangeU2MB(0x00E398, 0xFEEB);
	ChangeU2MB(0x00E399, 0xFEEC);
	ChangeU2MB(0x00E39A, 0xFEED);
	ChangeU2MB(0x00E39B, 0xFEEE);
	ChangeU2MB(0x00E39C, 0xFEEF);
	ChangeU2MB(0x00E39D, 0xFEF0);
	ChangeU2MB(0x00E39E, 0xFEF1);
	ChangeU2MB(0x00E39F, 0xFEF2);
	ChangeU2MB(0x00E3A0, 0xFEF3);
	ChangeU2MB(0x00E3A1, 0xFEF4);
	ChangeU2MB(0x00E3A2, 0xFEF5);
	ChangeU2MB(0x00E3A3, 0xFEF6);
	ChangeU2MB(0x00E3A4, 0xFEF7);
	ChangeU2MB(0x00E3A5, 0xFEF8);
	ChangeU2MB(0x00E3A6, 0xFEF9);
	ChangeU2MB(0x00E3A7, 0xFEFA);
	ChangeU2MB(0x00E3A8, 0xFEFB);
	ChangeU2MB(0x00E3A9, 0xFEFC);
	ChangeU2MB(0x00E3AA, 0xFEFD);
	ChangeU2MB(0x00E3AB, 0xFEFE);
	ChangeU2MB(0x00E3AC, 0xF521);
	ChangeU2MB(0x00E3AD, 0xF522);
	ChangeU2MB(0x00E3AE, 0xF523);
	ChangeU2MB(0x00E3AF, 0xF524);
	ChangeU2MB(0x00E3B0, 0xF525);
	ChangeU2MB(0x00E3B1, 0xF526);
	ChangeU2MB(0x00E3B2, 0xF527);
	ChangeU2MB(0x00E3B3, 0xF528);
	ChangeU2MB(0x00E3B4, 0xF529);
	ChangeU2MB(0x00E3B5, 0xF52A);
	ChangeU2MB(0x00E3B6, 0xF52B);
	ChangeU2MB(0x00E3B7, 0xF52C);
	ChangeU2MB(0x00E3B8, 0xF52D);
	ChangeU2MB(0x00E3B9, 0xF52E);
	ChangeU2MB(0x00E3BA, 0xF52F);
	ChangeU2MB(0x00E3BB, 0xF530);
	ChangeU2MB(0x00E3BC, 0xF531);
	ChangeU2MB(0x00E3BD, 0xF532);
	ChangeU2MB(0x00E3BE, 0xF533);
	ChangeU2MB(0x00E3BF, 0xF534);
	ChangeU2MB(0x00E3C0, 0xF535);
	ChangeU2MB(0x00E3C1, 0xF536);
	ChangeU2MB(0x00E3C2, 0xF537);
	ChangeU2MB(0x00E3C3, 0xF538);
	ChangeU2MB(0x00E3C4, 0xF539);
	ChangeU2MB(0x00E3C5, 0xF53A);
	ChangeU2MB(0x00E3C6, 0xF53B);
	ChangeU2MB(0x00E3C7, 0xF53C);
	ChangeU2MB(0x00E3C8, 0xF53D);
	ChangeU2MB(0x00E3C9, 0xF53E);
	ChangeU2MB(0x00E3CA, 0xF53F);
	ChangeU2MB(0x00E3CB, 0xF540);
	ChangeU2MB(0x00E3CC, 0xF541);
	ChangeU2MB(0x00E3CD, 0xF542);
	ChangeU2MB(0x00E3CE, 0xF543);
	ChangeU2MB(0x00E3CF, 0xF544);
	ChangeU2MB(0x00E3D0, 0xF545);
	ChangeU2MB(0x00E3D1, 0xF546);
	ChangeU2MB(0x00E3D2, 0xF547);
	ChangeU2MB(0x00E3D3, 0xF548);
	ChangeU2MB(0x00E3D4, 0xF549);
	ChangeU2MB(0x00E3D5, 0xF54A);
	ChangeU2MB(0x00E3D6, 0xF54B);
	ChangeU2MB(0x00E3D7, 0xF54C);
	ChangeU2MB(0x00E3D8, 0xF54D);
	ChangeU2MB(0x00E3D9, 0xF54E);
	ChangeU2MB(0x00E3DA, 0xF54F);
	ChangeU2MB(0x00E3DB, 0xF550);
	ChangeU2MB(0x00E3DC, 0xF551);
	ChangeU2MB(0x00E3DD, 0xF552);
	ChangeU2MB(0x00E3DE, 0xF553);
	ChangeU2MB(0x00E3DF, 0xF554);
	ChangeU2MB(0x00E3E0, 0xF555);
	ChangeU2MB(0x00E3E1, 0xF556);
	ChangeU2MB(0x00E3E2, 0xF557);
	ChangeU2MB(0x00E3E3, 0xF558);
	ChangeU2MB(0x00E3E4, 0xF559);
	ChangeU2MB(0x00E3E5, 0xF55A);
	ChangeU2MB(0x00E3E6, 0xF55B);
	ChangeU2MB(0x00E3E7, 0xF55C);
	ChangeU2MB(0x00E3E8, 0xF55D);
	ChangeU2MB(0x00E3E9, 0xF55E);
	ChangeU2MB(0x00E3EA, 0xF55F);
	ChangeU2MB(0x00E3EB, 0xF560);
	ChangeU2MB(0x00E3EC, 0xF561);
	ChangeU2MB(0x00E3ED, 0xF562);
	ChangeU2MB(0x00E3EE, 0xF563);
	ChangeU2MB(0x00E3EF, 0xF564);
	ChangeU2MB(0x00E3F0, 0xF565);
	ChangeU2MB(0x00E3F1, 0xF566);
	ChangeU2MB(0x00E3F2, 0xF567);
	ChangeU2MB(0x00E3F3, 0xF568);
	ChangeU2MB(0x00E3F4, 0xF569);
	ChangeU2MB(0x00E3F5, 0xF56A);
	ChangeU2MB(0x00E3F6, 0xF56B);
	ChangeU2MB(0x00E3F7, 0xF56C);
	ChangeU2MB(0x00E3F8, 0xF56D);
	ChangeU2MB(0x00E3F9, 0xF56E);
	ChangeU2MB(0x00E3FA, 0xF56F);
	ChangeU2MB(0x00E3FB, 0xF570);
	ChangeU2MB(0x00E3FC, 0xF571);
	ChangeU2MB(0x00E3FD, 0xF572);
	ChangeU2MB(0x00E3FE, 0xF573);
	ChangeU2MB(0x00E3FF, 0xF574);
	ChangeU2MB(0x00E400, 0xF575);
	ChangeU2MB(0x00E401, 0xF576);
	ChangeU2MB(0x00E402, 0xF577);
	ChangeU2MB(0x00E403, 0xF578);
	ChangeU2MB(0x00E404, 0xF579);
	ChangeU2MB(0x00E405, 0xF57A);
	ChangeU2MB(0x00E406, 0xF57B);
	ChangeU2MB(0x00E407, 0xF57C);
	ChangeU2MB(0x00E408, 0xF57D);
	ChangeU2MB(0x00E409, 0xF57E);
	ChangeU2MB(0x00E40A, 0xF621);
	ChangeU2MB(0x00E40B, 0xF622);
	ChangeU2MB(0x00E40C, 0xF623);
	ChangeU2MB(0x00E40D, 0xF624);
	ChangeU2MB(0x00E40E, 0xF625);
	ChangeU2MB(0x00E40F, 0xF626);
	ChangeU2MB(0x00E410, 0xF627);
	ChangeU2MB(0x00E411, 0xF628);
	ChangeU2MB(0x00E412, 0xF629);
	ChangeU2MB(0x00E413, 0xF62A);
	ChangeU2MB(0x00E414, 0xF62B);
	ChangeU2MB(0x00E415, 0xF62C);
	ChangeU2MB(0x00E416, 0xF62D);
	ChangeU2MB(0x00E417, 0xF62E);
	ChangeU2MB(0x00E418, 0xF62F);
	ChangeU2MB(0x00E419, 0xF630);
	ChangeU2MB(0x00E41A, 0xF631);
	ChangeU2MB(0x00E41B, 0xF632);
	ChangeU2MB(0x00E41C, 0xF633);
	ChangeU2MB(0x00E41D, 0xF634);
	ChangeU2MB(0x00E41E, 0xF635);
	ChangeU2MB(0x00E41F, 0xF636);
	ChangeU2MB(0x00E420, 0xF637);
	ChangeU2MB(0x00E421, 0xF638);
	ChangeU2MB(0x00E422, 0xF639);
	ChangeU2MB(0x00E423, 0xF63A);
	ChangeU2MB(0x00E424, 0xF63B);
	ChangeU2MB(0x00E425, 0xF63C);
	ChangeU2MB(0x00E426, 0xF63D);
	ChangeU2MB(0x00E427, 0xF63E);
	ChangeU2MB(0x00E428, 0xF63F);
	ChangeU2MB(0x00E429, 0xF640);
	ChangeU2MB(0x00E42A, 0xF641);
	ChangeU2MB(0x00E42B, 0xF642);
	ChangeU2MB(0x00E42C, 0xF643);
	ChangeU2MB(0x00E42D, 0xF644);
	ChangeU2MB(0x00E42E, 0xF645);
	ChangeU2MB(0x00E42F, 0xF646);
	ChangeU2MB(0x00E430, 0xF647);
	ChangeU2MB(0x00E431, 0xF648);
	ChangeU2MB(0x00E432, 0xF649);
	ChangeU2MB(0x00E433, 0xF64A);
	ChangeU2MB(0x00E434, 0xF64B);
	ChangeU2MB(0x00E435, 0xF64C);
	ChangeU2MB(0x00E436, 0xF64D);
	ChangeU2MB(0x00E437, 0xF64E);
	ChangeU2MB(0x00E438, 0xF64F);
	ChangeU2MB(0x00E439, 0xF650);
	ChangeU2MB(0x00E43A, 0xF651);
	ChangeU2MB(0x00E43B, 0xF652);
	ChangeU2MB(0x00E43C, 0xF653);
	ChangeU2MB(0x00E43D, 0xF654);
	ChangeU2MB(0x00E43E, 0xF655);
	ChangeU2MB(0x00E43F, 0xF656);
	ChangeU2MB(0x00E440, 0xF657);
	ChangeU2MB(0x00E441, 0xF658);
	ChangeU2MB(0x00E442, 0xF659);
	ChangeU2MB(0x00E443, 0xF65A);
	ChangeU2MB(0x00E444, 0xF65B);
	ChangeU2MB(0x00E445, 0xF65C);
	ChangeU2MB(0x00E446, 0xF65D);
	ChangeU2MB(0x00E447, 0xF65E);
	ChangeU2MB(0x00E448, 0xF65F);
	ChangeU2MB(0x00E449, 0xF660);
	ChangeU2MB(0x00E44A, 0xF661);
	ChangeU2MB(0x00E44B, 0xF662);
	ChangeU2MB(0x00E44C, 0xF663);
	ChangeU2MB(0x00E44D, 0xF664);
	ChangeU2MB(0x00E44E, 0xF665);
	ChangeU2MB(0x00E44F, 0xF666);
	ChangeU2MB(0x00E450, 0xF667);
	ChangeU2MB(0x00E451, 0xF668);
	ChangeU2MB(0x00E452, 0xF669);
	ChangeU2MB(0x00E453, 0xF66A);
	ChangeU2MB(0x00E454, 0xF66B);
	ChangeU2MB(0x00E455, 0xF66C);
	ChangeU2MB(0x00E456, 0xF66D);
	ChangeU2MB(0x00E457, 0xF66E);
	ChangeU2MB(0x00E458, 0xF66F);
	ChangeU2MB(0x00E459, 0xF670);
	ChangeU2MB(0x00E45A, 0xF671);
	ChangeU2MB(0x00E45B, 0xF672);
	ChangeU2MB(0x00E45C, 0xF673);
	ChangeU2MB(0x00E45D, 0xF674);
	ChangeU2MB(0x00E45E, 0xF675);
	ChangeU2MB(0x00E45F, 0xF676);
	ChangeU2MB(0x00E460, 0xF677);
	ChangeU2MB(0x00E461, 0xF678);
	ChangeU2MB(0x00E462, 0xF679);
	ChangeU2MB(0x00E463, 0xF67A);
	ChangeU2MB(0x00E464, 0xF67B);
	ChangeU2MB(0x00E465, 0xF67C);
	ChangeU2MB(0x00E466, 0xF67D);
	ChangeU2MB(0x00E467, 0xF67E);
	ChangeU2MB(0x00E468, 0xF721);
	ChangeU2MB(0x00E469, 0xF722);
	ChangeU2MB(0x00E46A, 0xF723);
	ChangeU2MB(0x00E46B, 0xF724);
	ChangeU2MB(0x00E46C, 0xF725);
	ChangeU2MB(0x00E46D, 0xF726);
	ChangeU2MB(0x00E46E, 0xF727);
	ChangeU2MB(0x00E46F, 0xF728);
	ChangeU2MB(0x00E470, 0xF729);
	ChangeU2MB(0x00E471, 0xF72A);
	ChangeU2MB(0x00E472, 0xF72B);
	ChangeU2MB(0x00E473, 0xF72C);
	ChangeU2MB(0x00E474, 0xF72D);
	ChangeU2MB(0x00E475, 0xF72E);
	ChangeU2MB(0x00E476, 0xF72F);
	ChangeU2MB(0x00E477, 0xF730);
	ChangeU2MB(0x00E478, 0xF731);
	ChangeU2MB(0x00E479, 0xF732);
	ChangeU2MB(0x00E47A, 0xF733);
	ChangeU2MB(0x00E47B, 0xF734);
	ChangeU2MB(0x00E47C, 0xF735);
	ChangeU2MB(0x00E47D, 0xF736);
	ChangeU2MB(0x00E47E, 0xF737);
	ChangeU2MB(0x00E47F, 0xF738);
	ChangeU2MB(0x00E480, 0xF739);
	ChangeU2MB(0x00E481, 0xF73A);
	ChangeU2MB(0x00E482, 0xF73B);
	ChangeU2MB(0x00E483, 0xF73C);
	ChangeU2MB(0x00E484, 0xF73D);
	ChangeU2MB(0x00E485, 0xF73E);
	ChangeU2MB(0x00E486, 0xF73F);
	ChangeU2MB(0x00E487, 0xF740);
	ChangeU2MB(0x00E488, 0xF741);
	ChangeU2MB(0x00E489, 0xF742);
	ChangeU2MB(0x00E48A, 0xF743);
	ChangeU2MB(0x00E48B, 0xF744);
	ChangeU2MB(0x00E48C, 0xF745);
	ChangeU2MB(0x00E48D, 0xF746);
	ChangeU2MB(0x00E48E, 0xF747);
	ChangeU2MB(0x00E48F, 0xF748);
	ChangeU2MB(0x00E490, 0xF749);
	ChangeU2MB(0x00E491, 0xF74A);
	ChangeU2MB(0x00E492, 0xF74B);
	ChangeU2MB(0x00E493, 0xF74C);
	ChangeU2MB(0x00E494, 0xF74D);
	ChangeU2MB(0x00E495, 0xF74E);
	ChangeU2MB(0x00E496, 0xF74F);
	ChangeU2MB(0x00E497, 0xF750);
	ChangeU2MB(0x00E498, 0xF751);
	ChangeU2MB(0x00E499, 0xF752);
	ChangeU2MB(0x00E49A, 0xF753);
	ChangeU2MB(0x00E49B, 0xF754);
	ChangeU2MB(0x00E49C, 0xF755);
	ChangeU2MB(0x00E49D, 0xF756);
	ChangeU2MB(0x00E49E, 0xF757);
	ChangeU2MB(0x00E49F, 0xF758);
	ChangeU2MB(0x00E4A0, 0xF759);
	ChangeU2MB(0x00E4A1, 0xF75A);
	ChangeU2MB(0x00E4A2, 0xF75B);
	ChangeU2MB(0x00E4A3, 0xF75C);
	ChangeU2MB(0x00E4A4, 0xF75D);
	ChangeU2MB(0x00E4A5, 0xF75E);
	ChangeU2MB(0x00E4A6, 0xF75F);
	ChangeU2MB(0x00E4A7, 0xF760);
	ChangeU2MB(0x00E4A8, 0xF761);
	ChangeU2MB(0x00E4A9, 0xF762);
	ChangeU2MB(0x00E4AA, 0xF763);
	ChangeU2MB(0x00E4AB, 0xF764);
	ChangeU2MB(0x00E4AC, 0xF765);
	ChangeU2MB(0x00E4AD, 0xF766);
	ChangeU2MB(0x00E4AE, 0xF767);
	ChangeU2MB(0x00E4AF, 0xF768);
	ChangeU2MB(0x00E4B0, 0xF769);
	ChangeU2MB(0x00E4B1, 0xF76A);
	ChangeU2MB(0x00E4B2, 0xF76B);
	ChangeU2MB(0x00E4B3, 0xF76C);
	ChangeU2MB(0x00E4B4, 0xF76D);
	ChangeU2MB(0x00E4B5, 0xF76E);
	ChangeU2MB(0x00E4B6, 0xF76F);
	ChangeU2MB(0x00E4B7, 0xF770);
	ChangeU2MB(0x00E4B8, 0xF771);
	ChangeU2MB(0x00E4B9, 0xF772);
	ChangeU2MB(0x00E4BA, 0xF773);
	ChangeU2MB(0x00E4BB, 0xF774);
	ChangeU2MB(0x00E4BC, 0xF775);
	ChangeU2MB(0x00E4BD, 0xF776);
	ChangeU2MB(0x00E4BE, 0xF777);
	ChangeU2MB(0x00E4BF, 0xF778);
	ChangeU2MB(0x00E4C0, 0xF779);
	ChangeU2MB(0x00E4C1, 0xF77A);
	ChangeU2MB(0x00E4C2, 0xF77B);
	ChangeU2MB(0x00E4C3, 0xF77C);
	ChangeU2MB(0x00E4C4, 0xF77D);
	ChangeU2MB(0x00E4C5, 0xF77E);
	ChangeU2MB(0x00E4C6, 0xF821);
	ChangeU2MB(0x00E4C7, 0xF822);
	ChangeU2MB(0x00E4C8, 0xF823);
	ChangeU2MB(0x00E4C9, 0xF824);
	ChangeU2MB(0x00E4CA, 0xF825);
	ChangeU2MB(0x00E4CB, 0xF826);
	ChangeU2MB(0x00E4CC, 0xF827);
	ChangeU2MB(0x00E4CD, 0xF828);
	ChangeU2MB(0x00E4CE, 0xF829);
	ChangeU2MB(0x00E4CF, 0xF82A);
	ChangeU2MB(0x00E4D0, 0xF82B);
	ChangeU2MB(0x00E4D1, 0xF82C);
	ChangeU2MB(0x00E4D2, 0xF82D);
	ChangeU2MB(0x00E4D3, 0xF82E);
	ChangeU2MB(0x00E4D4, 0xF82F);
	ChangeU2MB(0x00E4D5, 0xF830);
	ChangeU2MB(0x00E4D6, 0xF831);
	ChangeU2MB(0x00E4D7, 0xF832);
	ChangeU2MB(0x00E4D8, 0xF833);
	ChangeU2MB(0x00E4D9, 0xF834);
	ChangeU2MB(0x00E4DA, 0xF835);
	ChangeU2MB(0x00E4DB, 0xF836);
	ChangeU2MB(0x00E4DC, 0xF837);
	ChangeU2MB(0x00E4DD, 0xF838);
	ChangeU2MB(0x00E4DE, 0xF839);
	ChangeU2MB(0x00E4DF, 0xF83A);
	ChangeU2MB(0x00E4E0, 0xF83B);
	ChangeU2MB(0x00E4E1, 0xF83C);
	ChangeU2MB(0x00E4E2, 0xF83D);
	ChangeU2MB(0x00E4E3, 0xF83E);
	ChangeU2MB(0x00E4E4, 0xF83F);
	ChangeU2MB(0x00E4E5, 0xF840);
	ChangeU2MB(0x00E4E6, 0xF841);
	ChangeU2MB(0x00E4E7, 0xF842);
	ChangeU2MB(0x00E4E8, 0xF843);
	ChangeU2MB(0x00E4E9, 0xF844);
	ChangeU2MB(0x00E4EA, 0xF845);
	ChangeU2MB(0x00E4EB, 0xF846);
	ChangeU2MB(0x00E4EC, 0xF847);
	ChangeU2MB(0x00E4ED, 0xF848);
	ChangeU2MB(0x00E4EE, 0xF849);
	ChangeU2MB(0x00E4EF, 0xF84A);
	ChangeU2MB(0x00E4F0, 0xF84B);
	ChangeU2MB(0x00E4F1, 0xF84C);
	ChangeU2MB(0x00E4F2, 0xF84D);
	ChangeU2MB(0x00E4F3, 0xF84E);
	ChangeU2MB(0x00E4F4, 0xF84F);
	ChangeU2MB(0x00E4F5, 0xF850);
	ChangeU2MB(0x00E4F6, 0xF851);
	ChangeU2MB(0x00E4F7, 0xF852);
	ChangeU2MB(0x00E4F8, 0xF853);
	ChangeU2MB(0x00E4F9, 0xF854);
	ChangeU2MB(0x00E4FA, 0xF855);
	ChangeU2MB(0x00E4FB, 0xF856);
	ChangeU2MB(0x00E4FC, 0xF857);
	ChangeU2MB(0x00E4FD, 0xF858);
	ChangeU2MB(0x00E4FE, 0xF859);
	ChangeU2MB(0x00E4FF, 0xF85A);
	ChangeU2MB(0x00E500, 0xF85B);
	ChangeU2MB(0x00E501, 0xF85C);
	ChangeU2MB(0x00E502, 0xF85D);
	ChangeU2MB(0x00E503, 0xF85E);
	ChangeU2MB(0x00E504, 0xF85F);
	ChangeU2MB(0x00E505, 0xF860);
	ChangeU2MB(0x00E506, 0xF861);
	ChangeU2MB(0x00E507, 0xF862);
	ChangeU2MB(0x00E508, 0xF863);
	ChangeU2MB(0x00E509, 0xF864);
	ChangeU2MB(0x00E50A, 0xF865);
	ChangeU2MB(0x00E50B, 0xF866);
	ChangeU2MB(0x00E50C, 0xF867);
	ChangeU2MB(0x00E50D, 0xF868);
	ChangeU2MB(0x00E50E, 0xF869);
	ChangeU2MB(0x00E50F, 0xF86A);
	ChangeU2MB(0x00E510, 0xF86B);
	ChangeU2MB(0x00E511, 0xF86C);
	ChangeU2MB(0x00E512, 0xF86D);
	ChangeU2MB(0x00E513, 0xF86E);
	ChangeU2MB(0x00E514, 0xF86F);
	ChangeU2MB(0x00E515, 0xF870);
	ChangeU2MB(0x00E516, 0xF871);
	ChangeU2MB(0x00E517, 0xF872);
	ChangeU2MB(0x00E518, 0xF873);
	ChangeU2MB(0x00E519, 0xF874);
	ChangeU2MB(0x00E51A, 0xF875);
	ChangeU2MB(0x00E51B, 0xF876);
	ChangeU2MB(0x00E51C, 0xF877);
	ChangeU2MB(0x00E51D, 0xF878);
	ChangeU2MB(0x00E51E, 0xF879);
	ChangeU2MB(0x00E51F, 0xF87A);
	ChangeU2MB(0x00E520, 0xF87B);
	ChangeU2MB(0x00E521, 0xF87C);
	ChangeU2MB(0x00E522, 0xF87D);
	ChangeU2MB(0x00E523, 0xF87E);
	ChangeU2MB(0x00E524, 0xF921);
	ChangeU2MB(0x00E525, 0xF922);
	ChangeU2MB(0x00E526, 0xF923);
	ChangeU2MB(0x00E527, 0xF924);
	ChangeU2MB(0x00E528, 0xF925);
	ChangeU2MB(0x00E529, 0xF926);
	ChangeU2MB(0x00E52A, 0xF927);
	ChangeU2MB(0x00E52B, 0xF928);
	ChangeU2MB(0x00E52C, 0xF929);
	ChangeU2MB(0x00E52D, 0xF92A);
	ChangeU2MB(0x00E52E, 0xF92B);
	ChangeU2MB(0x00E52F, 0xF92C);
	ChangeU2MB(0x00E530, 0xF92D);
	ChangeU2MB(0x00E531, 0xF92E);
	ChangeU2MB(0x00E532, 0xF92F);
	ChangeU2MB(0x00E533, 0xF930);
	ChangeU2MB(0x00E534, 0xF931);
	ChangeU2MB(0x00E535, 0xF932);
	ChangeU2MB(0x00E536, 0xF933);
	ChangeU2MB(0x00E537, 0xF934);
	ChangeU2MB(0x00E538, 0xF935);
	ChangeU2MB(0x00E539, 0xF936);
	ChangeU2MB(0x00E53A, 0xF937);
	ChangeU2MB(0x00E53B, 0xF938);
	ChangeU2MB(0x00E53C, 0xF939);
	ChangeU2MB(0x00E53D, 0xF93A);
	ChangeU2MB(0x00E53E, 0xF93B);
	ChangeU2MB(0x00E53F, 0xF93C);
	ChangeU2MB(0x00E540, 0xF93D);
	ChangeU2MB(0x00E541, 0xF93E);
	ChangeU2MB(0x00E542, 0xF93F);
	ChangeU2MB(0x00E543, 0xF940);
	ChangeU2MB(0x00E544, 0xF941);
	ChangeU2MB(0x00E545, 0xF942);
	ChangeU2MB(0x00E546, 0xF943);
	ChangeU2MB(0x00E547, 0xF944);
	ChangeU2MB(0x00E548, 0xF945);
	ChangeU2MB(0x00E549, 0xF946);
	ChangeU2MB(0x00E54A, 0xF947);
	ChangeU2MB(0x00E54B, 0xF948);
	ChangeU2MB(0x00E54C, 0xF949);
	ChangeU2MB(0x00E54D, 0xF94A);
	ChangeU2MB(0x00E54E, 0xF94B);
	ChangeU2MB(0x00E54F, 0xF94C);
	ChangeU2MB(0x00E550, 0xF94D);
	ChangeU2MB(0x00E551, 0xF94E);
	ChangeU2MB(0x00E552, 0xF94F);
	ChangeU2MB(0x00E553, 0xF950);
	ChangeU2MB(0x00E554, 0xF951);
	ChangeU2MB(0x00E555, 0xF952);
	ChangeU2MB(0x00E556, 0xF953);
	ChangeU2MB(0x00E557, 0xF954);
	ChangeU2MB(0x00E558, 0xF955);
	ChangeU2MB(0x00E559, 0xF956);
	ChangeU2MB(0x00E55A, 0xF957);
	ChangeU2MB(0x00E55B, 0xF958);
	ChangeU2MB(0x00E55C, 0xF959);
	ChangeU2MB(0x00E55D, 0xF95A);
	ChangeU2MB(0x00E55E, 0xF95B);
	ChangeU2MB(0x00E55F, 0xF95C);
	ChangeU2MB(0x00E560, 0xF95D);
	ChangeU2MB(0x00E561, 0xF95E);
	ChangeU2MB(0x00E562, 0xF95F);
	ChangeU2MB(0x00E563, 0xF960);
	ChangeU2MB(0x00E564, 0xF961);
	ChangeU2MB(0x00E565, 0xF962);
	ChangeU2MB(0x00E566, 0xF963);
	ChangeU2MB(0x00E567, 0xF964);
	ChangeU2MB(0x00E568, 0xF965);
	ChangeU2MB(0x00E569, 0xF966);
	ChangeU2MB(0x00E56A, 0xF967);
	ChangeU2MB(0x00E56B, 0xF968);
	ChangeU2MB(0x00E56C, 0xF969);
	ChangeU2MB(0x00E56D, 0xF96A);
	ChangeU2MB(0x00E56E, 0xF96B);
	ChangeU2MB(0x00E56F, 0xF96C);
	ChangeU2MB(0x00E570, 0xF96D);
	ChangeU2MB(0x00E571, 0xF96E);
	ChangeU2MB(0x00E572, 0xF96F);
	ChangeU2MB(0x00E573, 0xF970);
	ChangeU2MB(0x00E574, 0xF971);
	ChangeU2MB(0x00E575, 0xF972);
	ChangeU2MB(0x00E576, 0xF973);
	ChangeU2MB(0x00E577, 0xF974);
	ChangeU2MB(0x00E578, 0xF975);
	ChangeU2MB(0x00E579, 0xF976);
	ChangeU2MB(0x00E57A, 0xF977);
	ChangeU2MB(0x00E57B, 0xF978);
	ChangeU2MB(0x00E57C, 0xF979);
	ChangeU2MB(0x00E57D, 0xF97A);
	ChangeU2MB(0x00E57E, 0xF97B);
	ChangeU2MB(0x00E57F, 0xF97C);
	ChangeU2MB(0x00E580, 0xF97D);
	ChangeU2MB(0x00E581, 0xF97E);
	ChangeU2MB(0x00E582, 0xFA21);
	ChangeU2MB(0x00E583, 0xFA22);
	ChangeU2MB(0x00E584, 0xFA23);
	ChangeU2MB(0x00E585, 0xFA24);
	ChangeU2MB(0x00E586, 0xFA25);
	ChangeU2MB(0x00E587, 0xFA26);
	ChangeU2MB(0x00E588, 0xFA27);
	ChangeU2MB(0x00E589, 0xFA28);
	ChangeU2MB(0x00E58A, 0xFA29);
	ChangeU2MB(0x00E58B, 0xFA2A);
	ChangeU2MB(0x00E58C, 0xFA2B);
	ChangeU2MB(0x00E58D, 0xFA2C);
	ChangeU2MB(0x00E58E, 0xFA2D);
	ChangeU2MB(0x00E58F, 0xFA2E);
	ChangeU2MB(0x00E590, 0xFA2F);
	ChangeU2MB(0x00E591, 0xFA30);
	ChangeU2MB(0x00E592, 0xFA31);
	ChangeU2MB(0x00E593, 0xFA32);
	ChangeU2MB(0x00E594, 0xFA33);
	ChangeU2MB(0x00E595, 0xFA34);
	ChangeU2MB(0x00E596, 0xFA35);
	ChangeU2MB(0x00E597, 0xFA36);
	ChangeU2MB(0x00E598, 0xFA37);
	ChangeU2MB(0x00E599, 0xFA38);
	ChangeU2MB(0x00E59A, 0xFA39);
	ChangeU2MB(0x00E59B, 0xFA3A);
	ChangeU2MB(0x00E59C, 0xFA3B);
	ChangeU2MB(0x00E59D, 0xFA3C);
	ChangeU2MB(0x00E59E, 0xFA3D);
	ChangeU2MB(0x00E59F, 0xFA3E);
	ChangeU2MB(0x00E5A0, 0xFA3F);
	ChangeU2MB(0x00E5A1, 0xFA40);
	ChangeU2MB(0x00E5A2, 0xFA41);
	ChangeU2MB(0x00E5A3, 0xFA42);
	ChangeU2MB(0x00E5A4, 0xFA43);
	ChangeU2MB(0x00E5A5, 0xFA44);
	ChangeU2MB(0x00E5A6, 0xFA45);
	ChangeU2MB(0x00E5A7, 0xFA46);
	ChangeU2MB(0x00E5A8, 0xFA47);
	ChangeU2MB(0x00E5A9, 0xFA48);
	ChangeU2MB(0x00E5AA, 0xFA49);
	ChangeU2MB(0x00E5AB, 0xFA4A);
	ChangeU2MB(0x00E5AC, 0xFA4B);
	ChangeU2MB(0x00E5AD, 0xFA4C);
	ChangeU2MB(0x00E5AE, 0xFA4D);
	ChangeU2MB(0x00E5AF, 0xFA4E);
	ChangeU2MB(0x00E5B0, 0xFA4F);
	ChangeU2MB(0x00E5B1, 0xFA50);
	ChangeU2MB(0x00E5B2, 0xFA51);
	ChangeU2MB(0x00E5B3, 0xFA52);
	ChangeU2MB(0x00E5B4, 0xFA53);
	ChangeU2MB(0x00E5B5, 0xFA54);
	ChangeU2MB(0x00E5B6, 0xFA55);
	ChangeU2MB(0x00E5B7, 0xFA56);
	ChangeU2MB(0x00E5B8, 0xFA57);
	ChangeU2MB(0x00E5B9, 0xFA58);
	ChangeU2MB(0x00E5BA, 0xFA59);
	ChangeU2MB(0x00E5BB, 0xFA5A);
	ChangeU2MB(0x00E5BC, 0xFA5B);
	ChangeU2MB(0x00E5BD, 0xFA5C);
	ChangeU2MB(0x00E5BE, 0xFA5D);
	ChangeU2MB(0x00E5BF, 0xFA5E);
	ChangeU2MB(0x00E5C0, 0xFA5F);
	ChangeU2MB(0x00E5C1, 0xFA60);
	ChangeU2MB(0x00E5C2, 0xFA61);
	ChangeU2MB(0x00E5C3, 0xFA62);
	ChangeU2MB(0x00E5C4, 0xFA63);
	ChangeU2MB(0x00E5C5, 0xFA64);
	ChangeU2MB(0x00E5C6, 0xFA65);
	ChangeU2MB(0x00E5C7, 0xFA66);
	ChangeU2MB(0x00E5C8, 0xFA67);
	ChangeU2MB(0x00E5C9, 0xFA68);
	ChangeU2MB(0x00E5CA, 0xFA69);
	ChangeU2MB(0x00E5CB, 0xFA6A);
	ChangeU2MB(0x00E5CC, 0xFA6B);
	ChangeU2MB(0x00E5CD, 0xFA6C);
	ChangeU2MB(0x00E5CE, 0xFA6D);
	ChangeU2MB(0x00E5CF, 0xFA6E);
	ChangeU2MB(0x00E5D0, 0xFA6F);
	ChangeU2MB(0x00E5D1, 0xFA70);
	ChangeU2MB(0x00E5D2, 0xFA71);
	ChangeU2MB(0x00E5D3, 0xFA72);
	ChangeU2MB(0x00E5D4, 0xFA73);
	ChangeU2MB(0x00E5D5, 0xFA74);
	ChangeU2MB(0x00E5D6, 0xFA75);
	ChangeU2MB(0x00E5D7, 0xFA76);
	ChangeU2MB(0x00E5D8, 0xFA77);
	ChangeU2MB(0x00E5D9, 0xFA78);
	ChangeU2MB(0x00E5DA, 0xFA79);
	ChangeU2MB(0x00E5DB, 0xFA7A);
	ChangeU2MB(0x00E5DC, 0xFA7B);
	ChangeU2MB(0x00E5DD, 0xFA7C);
	ChangeU2MB(0x00E5DE, 0xFA7D);
	ChangeU2MB(0x00E5DF, 0xFA7E);
	ChangeU2MB(0x00E5E0, 0xFB21);
	ChangeU2MB(0x00E5E1, 0xFB22);
	ChangeU2MB(0x00E5E2, 0xFB23);
	ChangeU2MB(0x00E5E3, 0xFB24);
	ChangeU2MB(0x00E5E4, 0xFB25);
	ChangeU2MB(0x00E5E5, 0xFB26);
	ChangeU2MB(0x00E5E6, 0xFB27);
	ChangeU2MB(0x00E5E7, 0xFB28);
	ChangeU2MB(0x00E5E8, 0xFB29);
	ChangeU2MB(0x00E5E9, 0xFB2A);
	ChangeU2MB(0x00E5EA, 0xFB2B);
	ChangeU2MB(0x00E5EB, 0xFB2C);
	ChangeU2MB(0x00E5EC, 0xFB2D);
	ChangeU2MB(0x00E5ED, 0xFB2E);
	ChangeU2MB(0x00E5EE, 0xFB2F);
	ChangeU2MB(0x00E5EF, 0xFB30);
	ChangeU2MB(0x00E5F0, 0xFB31);
	ChangeU2MB(0x00E5F1, 0xFB32);
	ChangeU2MB(0x00E5F2, 0xFB33);
	ChangeU2MB(0x00E5F3, 0xFB34);
	ChangeU2MB(0x00E5F4, 0xFB35);
	ChangeU2MB(0x00E5F5, 0xFB36);
	ChangeU2MB(0x00E5F6, 0xFB37);
	ChangeU2MB(0x00E5F7, 0xFB38);
	ChangeU2MB(0x00E5F8, 0xFB39);
	ChangeU2MB(0x00E5F9, 0xFB3A);
	ChangeU2MB(0x00E5FA, 0xFB3B);
	ChangeU2MB(0x00E5FB, 0xFB3C);
	ChangeU2MB(0x00E5FC, 0xFB3D);
	ChangeU2MB(0x00E5FD, 0xFB3E);
	ChangeU2MB(0x00E5FE, 0xFB3F);
	ChangeU2MB(0x00E5FF, 0xFB40);
	ChangeU2MB(0x00E600, 0xFB41);
	ChangeU2MB(0x00E601, 0xFB42);
	ChangeU2MB(0x00E602, 0xFB43);
	ChangeU2MB(0x00E603, 0xFB44);
	ChangeU2MB(0x00E604, 0xFB45);
	ChangeU2MB(0x00E605, 0xFB46);
	ChangeU2MB(0x00E606, 0xFB47);
	ChangeU2MB(0x00E607, 0xFB48);
	ChangeU2MB(0x00E608, 0xFB49);
	ChangeU2MB(0x00E609, 0xFB4A);
	ChangeU2MB(0x00E60A, 0xFB4B);
	ChangeU2MB(0x00E60B, 0xFB4C);
	ChangeU2MB(0x00E60C, 0xFB4D);
	ChangeU2MB(0x00E60D, 0xFB4E);
	ChangeU2MB(0x00E60E, 0xFB4F);
	ChangeU2MB(0x00E60F, 0xFB50);
	ChangeU2MB(0x00E610, 0xFB51);
	ChangeU2MB(0x00E611, 0xFB52);
	ChangeU2MB(0x00E612, 0xFB53);
	ChangeU2MB(0x00E613, 0xFB54);
	ChangeU2MB(0x00E614, 0xFB55);
	ChangeU2MB(0x00E615, 0xFB56);
	ChangeU2MB(0x00E616, 0xFB57);
	ChangeU2MB(0x00E617, 0xFB58);
	ChangeU2MB(0x00E618, 0xFB59);
	ChangeU2MB(0x00E619, 0xFB5A);
	ChangeU2MB(0x00E61A, 0xFB5B);
	ChangeU2MB(0x00E61B, 0xFB5C);
	ChangeU2MB(0x00E61C, 0xFB5D);
	ChangeU2MB(0x00E61D, 0xFB5E);
	ChangeU2MB(0x00E61E, 0xFB5F);
	ChangeU2MB(0x00E61F, 0xFB60);
	ChangeU2MB(0x00E620, 0xFB61);
	ChangeU2MB(0x00E621, 0xFB62);
	ChangeU2MB(0x00E622, 0xFB63);
	ChangeU2MB(0x00E623, 0xFB64);
	ChangeU2MB(0x00E624, 0xFB65);
	ChangeU2MB(0x00E625, 0xFB66);
	ChangeU2MB(0x00E626, 0xFB67);
	ChangeU2MB(0x00E627, 0xFB68);
	ChangeU2MB(0x00E628, 0xFB69);
	ChangeU2MB(0x00E629, 0xFB6A);
	ChangeU2MB(0x00E62A, 0xFB6B);
	ChangeU2MB(0x00E62B, 0xFB6C);
	ChangeU2MB(0x00E62C, 0xFB6D);
	ChangeU2MB(0x00E62D, 0xFB6E);
	ChangeU2MB(0x00E62E, 0xFB6F);
	ChangeU2MB(0x00E62F, 0xFB70);
	ChangeU2MB(0x00E630, 0xFB71);
	ChangeU2MB(0x00E631, 0xFB72);
	ChangeU2MB(0x00E632, 0xFB73);
	ChangeU2MB(0x00E633, 0xFB74);
	ChangeU2MB(0x00E634, 0xFB75);
	ChangeU2MB(0x00E635, 0xFB76);
	ChangeU2MB(0x00E636, 0xFB77);
	ChangeU2MB(0x00E637, 0xFB78);
	ChangeU2MB(0x00E638, 0xFB79);
	ChangeU2MB(0x00E639, 0xFB7A);
	ChangeU2MB(0x00E63A, 0xFB7B);
	ChangeU2MB(0x00E63B, 0xFB7C);
	ChangeU2MB(0x00E63C, 0xFB7D);
	ChangeU2MB(0x00E63D, 0xFB7E);
	ChangeU2MB(0x00E63E, 0xFC21);
	ChangeU2MB(0x00E63F, 0xFC22);
	ChangeU2MB(0x00E640, 0xFC23);
	ChangeU2MB(0x00E641, 0xFC24);
	ChangeU2MB(0x00E642, 0xFC25);
	ChangeU2MB(0x00E643, 0xFC26);
	ChangeU2MB(0x00E644, 0xFC27);
	ChangeU2MB(0x00E645, 0xFC28);
	ChangeU2MB(0x00E646, 0xFC29);
	ChangeU2MB(0x00E647, 0xFC2A);
	ChangeU2MB(0x00E648, 0xFC2B);
	ChangeU2MB(0x00E649, 0xFC2C);
	ChangeU2MB(0x00E64A, 0xFC2D);
	ChangeU2MB(0x00E64B, 0xFC2E);
	ChangeU2MB(0x00E64C, 0xFC2F);
	ChangeU2MB(0x00E64D, 0xFC30);
	ChangeU2MB(0x00E64E, 0xFC31);
	ChangeU2MB(0x00E64F, 0xFC32);
	ChangeU2MB(0x00E650, 0xFC33);
	ChangeU2MB(0x00E651, 0xFC34);
	ChangeU2MB(0x00E652, 0xFC35);
	ChangeU2MB(0x00E653, 0xFC36);
	ChangeU2MB(0x00E654, 0xFC37);
	ChangeU2MB(0x00E655, 0xFC38);
	ChangeU2MB(0x00E656, 0xFC39);
	ChangeU2MB(0x00E657, 0xFC3A);
	ChangeU2MB(0x00E658, 0xFC3B);
	ChangeU2MB(0x00E659, 0xFC3C);
	ChangeU2MB(0x00E65A, 0xFC3D);
	ChangeU2MB(0x00E65B, 0xFC3E);
	ChangeU2MB(0x00E65C, 0xFC3F);
	ChangeU2MB(0x00E65D, 0xFC40);
	ChangeU2MB(0x00E65E, 0xFC41);
	ChangeU2MB(0x00E65F, 0xFC42);
	ChangeU2MB(0x00E660, 0xFC43);
	ChangeU2MB(0x00E661, 0xFC44);
	ChangeU2MB(0x00E662, 0xFC45);
	ChangeU2MB(0x00E663, 0xFC46);
	ChangeU2MB(0x00E664, 0xFC47);
	ChangeU2MB(0x00E665, 0xFC48);
	ChangeU2MB(0x00E666, 0xFC49);
	ChangeU2MB(0x00E667, 0xFC4A);
	ChangeU2MB(0x00E668, 0xFC4B);
	ChangeU2MB(0x00E669, 0xFC4C);
	ChangeU2MB(0x00E66A, 0xFC4D);
	ChangeU2MB(0x00E66B, 0xFC4E);
	ChangeU2MB(0x00E66C, 0xFC4F);
	ChangeU2MB(0x00E66D, 0xFC50);
	ChangeU2MB(0x00E66E, 0xFC51);
	ChangeU2MB(0x00E66F, 0xFC52);
	ChangeU2MB(0x00E670, 0xFC53);
	ChangeU2MB(0x00E671, 0xFC54);
	ChangeU2MB(0x00E672, 0xFC55);
	ChangeU2MB(0x00E673, 0xFC56);
	ChangeU2MB(0x00E674, 0xFC57);
	ChangeU2MB(0x00E675, 0xFC58);
	ChangeU2MB(0x00E676, 0xFC59);
	ChangeU2MB(0x00E677, 0xFC5A);
	ChangeU2MB(0x00E678, 0xFC5B);
	ChangeU2MB(0x00E679, 0xFC5C);
	ChangeU2MB(0x00E67A, 0xFC5D);
	ChangeU2MB(0x00E67B, 0xFC5E);
	ChangeU2MB(0x00E67C, 0xFC5F);
	ChangeU2MB(0x00E67D, 0xFC60);
	ChangeU2MB(0x00E67E, 0xFC61);
	ChangeU2MB(0x00E67F, 0xFC62);
	ChangeU2MB(0x00E680, 0xFC63);
	ChangeU2MB(0x00E681, 0xFC64);
	ChangeU2MB(0x00E682, 0xFC65);
	ChangeU2MB(0x00E683, 0xFC66);
	ChangeU2MB(0x00E684, 0xFC67);
	ChangeU2MB(0x00E685, 0xFC68);
	ChangeU2MB(0x00E686, 0xFC69);
	ChangeU2MB(0x00E687, 0xFC6A);
	ChangeU2MB(0x00E688, 0xFC6B);
	ChangeU2MB(0x00E689, 0xFC6C);
	ChangeU2MB(0x00E68A, 0xFC6D);
	ChangeU2MB(0x00E68B, 0xFC6E);
	ChangeU2MB(0x00E68C, 0xFC6F);
	ChangeU2MB(0x00E68D, 0xFC70);
	ChangeU2MB(0x00E68E, 0xFC71);
	ChangeU2MB(0x00E68F, 0xFC72);
	ChangeU2MB(0x00E690, 0xFC73);
	ChangeU2MB(0x00E691, 0xFC74);
	ChangeU2MB(0x00E692, 0xFC75);
	ChangeU2MB(0x00E693, 0xFC76);
	ChangeU2MB(0x00E694, 0xFC77);
	ChangeU2MB(0x00E695, 0xFC78);
	ChangeU2MB(0x00E696, 0xFC79);
	ChangeU2MB(0x00E697, 0xFC7A);
	ChangeU2MB(0x00E698, 0xFC7B);
	ChangeU2MB(0x00E699, 0xFC7C);
	ChangeU2MB(0x00E69A, 0xFC7D);
	ChangeU2MB(0x00E69B, 0xFC7E);
	ChangeU2MB(0x00E69C, 0xFD21);
	ChangeU2MB(0x00E69D, 0xFD22);
	ChangeU2MB(0x00E69E, 0xFD23);
	ChangeU2MB(0x00E69F, 0xFD24);
	ChangeU2MB(0x00E6A0, 0xFD25);
	ChangeU2MB(0x00E6A1, 0xFD26);
	ChangeU2MB(0x00E6A2, 0xFD27);
	ChangeU2MB(0x00E6A3, 0xFD28);
	ChangeU2MB(0x00E6A4, 0xFD29);
	ChangeU2MB(0x00E6A5, 0xFD2A);
	ChangeU2MB(0x00E6A6, 0xFD2B);
	ChangeU2MB(0x00E6A7, 0xFD2C);
	ChangeU2MB(0x00E6A8, 0xFD2D);
	ChangeU2MB(0x00E6A9, 0xFD2E);
	ChangeU2MB(0x00E6AA, 0xFD2F);
	ChangeU2MB(0x00E6AB, 0xFD30);
	ChangeU2MB(0x00E6AC, 0xFD31);
	ChangeU2MB(0x00E6AD, 0xFD32);
	ChangeU2MB(0x00E6AE, 0xFD33);
	ChangeU2MB(0x00E6AF, 0xFD34);
	ChangeU2MB(0x00E6B0, 0xFD35);
	ChangeU2MB(0x00E6B1, 0xFD36);
	ChangeU2MB(0x00E6B2, 0xFD37);
	ChangeU2MB(0x00E6B3, 0xFD38);
	ChangeU2MB(0x00E6B4, 0xFD39);
	ChangeU2MB(0x00E6B5, 0xFD3A);
	ChangeU2MB(0x00E6B6, 0xFD3B);
	ChangeU2MB(0x00E6B7, 0xFD3C);
	ChangeU2MB(0x00E6B8, 0xFD3D);
	ChangeU2MB(0x00E6B9, 0xFD3E);
	ChangeU2MB(0x00E6BA, 0xFD3F);
	ChangeU2MB(0x00E6BB, 0xFD40);
	ChangeU2MB(0x00E6BC, 0xFD41);
	ChangeU2MB(0x00E6BD, 0xFD42);
	ChangeU2MB(0x00E6BE, 0xFD43);
	ChangeU2MB(0x00E6BF, 0xFD44);
	ChangeU2MB(0x00E6C0, 0xFD45);
	ChangeU2MB(0x00E6C1, 0xFD46);
	ChangeU2MB(0x00E6C2, 0xFD47);
	ChangeU2MB(0x00E6C3, 0xFD48);
	ChangeU2MB(0x00E6C4, 0xFD49);
	ChangeU2MB(0x00E6C5, 0xFD4A);
	ChangeU2MB(0x00E6C6, 0xFD4B);
	ChangeU2MB(0x00E6C7, 0xFD4C);
	ChangeU2MB(0x00E6C8, 0xFD4D);
	ChangeU2MB(0x00E6C9, 0xFD4E);
	ChangeU2MB(0x00E6CA, 0xFD4F);
	ChangeU2MB(0x00E6CB, 0xFD50);
	ChangeU2MB(0x00E6CC, 0xFD51);
	ChangeU2MB(0x00E6CD, 0xFD52);
	ChangeU2MB(0x00E6CE, 0xFD53);
	ChangeU2MB(0x00E6CF, 0xFD54);
	ChangeU2MB(0x00E6D0, 0xFD55);
	ChangeU2MB(0x00E6D1, 0xFD56);
	ChangeU2MB(0x00E6D2, 0xFD57);
	ChangeU2MB(0x00E6D3, 0xFD58);
	ChangeU2MB(0x00E6D4, 0xFD59);
	ChangeU2MB(0x00E6D5, 0xFD5A);
	ChangeU2MB(0x00E6D6, 0xFD5B);
	ChangeU2MB(0x00E6D7, 0xFD5C);
	ChangeU2MB(0x00E6D8, 0xFD5D);
	ChangeU2MB(0x00E6D9, 0xFD5E);
	ChangeU2MB(0x00E6DA, 0xFD5F);
	ChangeU2MB(0x00E6DB, 0xFD60);
	ChangeU2MB(0x00E6DC, 0xFD61);
	ChangeU2MB(0x00E6DD, 0xFD62);
	ChangeU2MB(0x00E6DE, 0xFD63);
	ChangeU2MB(0x00E6DF, 0xFD64);
	ChangeU2MB(0x00E6E0, 0xFD65);
	ChangeU2MB(0x00E6E1, 0xFD66);
	ChangeU2MB(0x00E6E2, 0xFD67);
	ChangeU2MB(0x00E6E3, 0xFD68);
	ChangeU2MB(0x00E6E4, 0xFD69);
	ChangeU2MB(0x00E6E5, 0xFD6A);
	ChangeU2MB(0x00E6E6, 0xFD6B);
	ChangeU2MB(0x00E6E7, 0xFD6C);
	ChangeU2MB(0x00E6E8, 0xFD6D);
	ChangeU2MB(0x00E6E9, 0xFD6E);
	ChangeU2MB(0x00E6EA, 0xFD6F);
	ChangeU2MB(0x00E6EB, 0xFD70);
	ChangeU2MB(0x00E6EC, 0xFD71);
	ChangeU2MB(0x00E6ED, 0xFD72);
	ChangeU2MB(0x00E6EE, 0xFD73);
	ChangeU2MB(0x00E6EF, 0xFD74);
	ChangeU2MB(0x00E6F0, 0xFD75);
	ChangeU2MB(0x00E6F1, 0xFD76);
	ChangeU2MB(0x00E6F2, 0xFD77);
	ChangeU2MB(0x00E6F3, 0xFD78);
	ChangeU2MB(0x00E6F4, 0xFD79);
	ChangeU2MB(0x00E6F5, 0xFD7A);
	ChangeU2MB(0x00E6F6, 0xFD7B);
	ChangeU2MB(0x00E6F7, 0xFD7C);
	ChangeU2MB(0x00E6F8, 0xFD7D);
	ChangeU2MB(0x00E6F9, 0xFD7E);
	ChangeU2MB(0x00E6FA, 0xFE21);
	ChangeU2MB(0x00E6FB, 0xFE22);
	ChangeU2MB(0x00E6FC, 0xFE23);
	ChangeU2MB(0x00E6FD, 0xFE24);
	ChangeU2MB(0x00E6FE, 0xFE25);
	ChangeU2MB(0x00E6FF, 0xFE26);
	ChangeU2MB(0x00E700, 0xFE27);
	ChangeU2MB(0x00E701, 0xFE28);
	ChangeU2MB(0x00E702, 0xFE29);
	ChangeU2MB(0x00E703, 0xFE2A);
	ChangeU2MB(0x00E704, 0xFE2B);
	ChangeU2MB(0x00E705, 0xFE2C);
	ChangeU2MB(0x00E706, 0xFE2D);
	ChangeU2MB(0x00E707, 0xFE2E);
	ChangeU2MB(0x00E708, 0xFE2F);
	ChangeU2MB(0x00E709, 0xFE30);
	ChangeU2MB(0x00E70A, 0xFE31);
	ChangeU2MB(0x00E70B, 0xFE32);
	ChangeU2MB(0x00E70C, 0xFE33);
	ChangeU2MB(0x00E70D, 0xFE34);
	ChangeU2MB(0x00E70E, 0xFE35);
	ChangeU2MB(0x00E70F, 0xFE36);
	ChangeU2MB(0x00E710, 0xFE37);
	ChangeU2MB(0x00E711, 0xFE38);
	ChangeU2MB(0x00E712, 0xFE39);
	ChangeU2MB(0x00E713, 0xFE3A);
	ChangeU2MB(0x00E714, 0xFE3B);
	ChangeU2MB(0x00E715, 0xFE3C);
	ChangeU2MB(0x00E716, 0xFE3D);
	ChangeU2MB(0x00E717, 0xFE3E);
	ChangeU2MB(0x00E718, 0xFE3F);
	ChangeU2MB(0x00E719, 0xFE40);
	ChangeU2MB(0x00E71A, 0xFE41);
	ChangeU2MB(0x00E71B, 0xFE42);
	ChangeU2MB(0x00E71C, 0xFE43);
	ChangeU2MB(0x00E71D, 0xFE44);
	ChangeU2MB(0x00E71E, 0xFE45);
	ChangeU2MB(0x00E71F, 0xFE46);
	ChangeU2MB(0x00E720, 0xFE47);
	ChangeU2MB(0x00E721, 0xFE48);
	ChangeU2MB(0x00E722, 0xFE49);
	ChangeU2MB(0x00E723, 0xFE4A);
	ChangeU2MB(0x00E724, 0xFE4B);
	ChangeU2MB(0x00E725, 0xFE4C);
	ChangeU2MB(0x00E726, 0xFE4D);
	ChangeU2MB(0x00E727, 0xFE4E);
	ChangeU2MB(0x00E728, 0xFE4F);
	ChangeU2MB(0x00E729, 0xFE50);
	ChangeU2MB(0x00E72A, 0xFE51);
	ChangeU2MB(0x00E72B, 0xFE52);
	ChangeU2MB(0x00E72C, 0xFE53);
	ChangeU2MB(0x00E72D, 0xFE54);
	ChangeU2MB(0x00E72E, 0xFE55);
	ChangeU2MB(0x00E72F, 0xFE56);
	ChangeU2MB(0x00E730, 0xFE57);
	ChangeU2MB(0x00E731, 0xFE58);
	ChangeU2MB(0x00E732, 0xFE59);
	ChangeU2MB(0x00E733, 0xFE5A);
	ChangeU2MB(0x00E734, 0xFE5B);
	ChangeU2MB(0x00E735, 0xFE5C);
	ChangeU2MB(0x00E736, 0xFE5D);
	ChangeU2MB(0x00E737, 0xFE5E);
	ChangeU2MB(0x00E738, 0xFE5F);
	ChangeU2MB(0x00E739, 0xFE60);
	ChangeU2MB(0x00E73A, 0xFE61);
	ChangeU2MB(0x00E73B, 0xFE62);
	ChangeU2MB(0x00E73C, 0xFE63);
	ChangeU2MB(0x00E73D, 0xFE64);
	ChangeU2MB(0x00E73E, 0xFE65);
	ChangeU2MB(0x00E73F, 0xFE66);
	ChangeU2MB(0x00E740, 0xFE67);
	ChangeU2MB(0x00E741, 0xFE68);
	ChangeU2MB(0x00E742, 0xFE69);
	ChangeU2MB(0x00E743, 0xFE6A);
	ChangeU2MB(0x00E744, 0xFE6B);
	ChangeU2MB(0x00E745, 0xFE6C);
	ChangeU2MB(0x00E746, 0xFE6D);
	ChangeU2MB(0x00E747, 0xFE6E);
	ChangeU2MB(0x00E748, 0xFE6F);
	ChangeU2MB(0x00E749, 0xFE70);
	ChangeU2MB(0x00E74A, 0xFE71);
	ChangeU2MB(0x00E74B, 0xFE72);
	ChangeU2MB(0x00E74C, 0xFE73);
	ChangeU2MB(0x00E74D, 0xFE74);
	ChangeU2MB(0x00E74E, 0xFE75);
	ChangeU2MB(0x00E74F, 0xFE76);
	ChangeU2MB(0x00E750, 0xFE77);
	ChangeU2MB(0x00E751, 0xFE78);
	ChangeU2MB(0x00E752, 0xFE79);
	ChangeU2MB(0x00E753, 0xFE7A);
	ChangeU2MB(0x00E754, 0xFE7B);
	ChangeU2MB(0x00E755, 0xFE7C);
	ChangeU2MB(0x00E756, 0xFE7D);
	ChangeU2MB(0x00E757, 0xFE7E);
	ChangeU2MB(0x00F929, 0xF351);
	ChangeU2MB(0x00F9DC, 0xF37E);
	ChangeU2MB(0x00FA0E, 0xF340);
	ChangeU2MB(0x00FA0F, 0xF343);
	ChangeU2MB(0x00FA10, 0xF344);
	ChangeU2MB(0x00FA11, 0xF349);
	ChangeU2MB(0x00FA12, 0xF350);
	ChangeU2MB(0x00FA13, 0xF353);
	ChangeU2MB(0x00FA14, 0xF354);
	ChangeU2MB(0x00FA15, 0xF35A);
	ChangeU2MB(0x00FA16, 0xF35B);
	ChangeU2MB(0x00FA17, 0xF35F);
	ChangeU2MB(0x00FA18, 0xF361);
	ChangeU2MB(0x00FA19, 0xF362);
	ChangeU2MB(0x00FA1A, 0xF363);
	ChangeU2MB(0x00FA1B, 0xF364);
	ChangeU2MB(0x00FA1C, 0xF366);
	ChangeU2MB(0x00FA1D, 0xF367);
	ChangeU2MB(0x00FA1E, 0xF36A);
	ChangeU2MB(0x00FA1F, 0xF36C);
	ChangeU2MB(0x00FA20, 0xF36E);
	ChangeU2MB(0x00FA21, 0xF36F);
	ChangeU2MB(0x00FA22, 0xF371);
	ChangeU2MB(0x00FA23, 0xF375);
	ChangeU2MB(0x00FA24, 0xF376);
	ChangeU2MB(0x00FA25, 0xF377);
	ChangeU2MB(0x00FA26, 0xF379);
	ChangeU2MB(0x00FA27, 0xF37B);
	ChangeU2MB(0x00FA28, 0xF37C);
	ChangeU2MB(0x00FA29, 0xF421);
	ChangeU2MB(0x00FA2A, 0xF425);
	ChangeU2MB(0x00FA2B, 0xF426);
	ChangeU2MB(0x00FA2C, 0xF427);
	ChangeU2MB(0x00FA2D, 0xF42B);
	ChangeU2MB(0x00FF02, 0xF336);
	ChangeU2MB(0x00FF07, 0xF335);

	icu42fix();
}

void CP20932TableFixer::icu42fix()
{
#if (U_ICU_VERSION_MAJOR_NUM < 51)
	RemoveMB2U(0xA900);
	RemoveMB2U(0xAA00);
	RemoveMB2U(0xAB00);
	RemoveMB2U(0xAD00);

	RemoveMB2U(0x8EE3);
	RemoveMB2U(0x8EE4);
	RemoveMB2U(0xF338);
	RemoveMB2U(0xF339);
	RemoveMB2U(0xF423);
	RemoveMB2U(0xF42E);
	RemoveMB2U(0xF42F);
	RemoveMB2U(0xF430);
	RemoveMB2U(0xF431);
	RemoveMB2U(0xF432);
	RemoveMB2U(0xF433);
	RemoveMB2U(0xF434);
	RemoveMB2U(0xF435);
	RemoveMB2U(0xF436);
	RemoveMB2U(0xF437);
	RemoveMB2U(0xF438);
	RemoveMB2U(0xF439);
	RemoveMB2U(0xF43A);
	RemoveMB2U(0xF43B);
	RemoveMB2U(0xF43C);
	RemoveMB2U(0xF43D);
	RemoveMB2U(0xF43E);
	RemoveMB2U(0xF43F);
	RemoveMB2U(0xF440);
	RemoveMB2U(0xF442);
	RemoveMB2U(0xF444);
	RemoveMB2U(0xF445);
	RemoveMB2U(0xF446);
	RemoveMB2U(0xF447);
	RemoveMB2U(0xF448);
	RemoveMB2U(0xF449);
	RemoveMB2U(0xF44B);
	RemoveMB2U(0xF44C);
	RemoveMB2U(0xF44D);
	RemoveMB2U(0xF44E);
	RemoveMB2U(0xF44F);
	RemoveMB2U(0xF450);
	RemoveMB2U(0xF451);
	RemoveMB2U(0xF453);
	RemoveMB2U(0xF454);
	RemoveMB2U(0xF455);
	RemoveMB2U(0xF456);
	RemoveMB2U(0xF457);
	RemoveMB2U(0xF458);
	RemoveMB2U(0xF459);
	RemoveMB2U(0xF45A);
	RemoveMB2U(0xF45B);
	RemoveMB2U(0xF45C);
	RemoveMB2U(0xF45E);
	RemoveMB2U(0xF45F);
	RemoveMB2U(0xF460);
	RemoveMB2U(0xF461);
	RemoveMB2U(0xF462);
	RemoveMB2U(0xF464);
	RemoveMB2U(0xF465);
	RemoveMB2U(0xF466);
	RemoveMB2U(0xF467);
	RemoveMB2U(0xF468);
	RemoveMB2U(0xF469);
	RemoveMB2U(0xF46A);
	RemoveMB2U(0xF46B);
	RemoveMB2U(0xF46C);
	RemoveMB2U(0xF46D);
	RemoveMB2U(0xF46E);
	RemoveMB2U(0xF46F);
	RemoveMB2U(0xF471);
	RemoveMB2U(0xF472);
	RemoveMB2U(0xF473);
	RemoveMB2U(0xF474);
	RemoveMB2U(0xF475);
	RemoveMB2U(0xF476);
	RemoveMB2U(0xF477);
	RemoveMB2U(0xF478);

	ChangeMB2U(0xA22F, 0x0002D8);
	ChangeMB2U(0xA230, 0x0002C7);
	ChangeMB2U(0xA231, 0x0000B8);
	ChangeMB2U(0xA232, 0x0002D9);
	ChangeMB2U(0xA233, 0x0002DD);
	ChangeMB2U(0xA234, 0x0000AF);
	ChangeMB2U(0xA235, 0x0002DB);
	ChangeMB2U(0xA236, 0x0002DA);
	ChangeMB2U(0xA238, 0x000384);
	ChangeMB2U(0xA239, 0x000385);
	ChangeMB2U(0xA242, 0x0000A1);
	ChangeMB2U(0xA243, 0x0000A6);
	ChangeMB2U(0xA244, 0x0000BF);
	ChangeMB2U(0xA26B, 0x0000BA);
	ChangeMB2U(0xA26C, 0x0000AA);
	ChangeMB2U(0xA26D, 0x0000A9);
	ChangeMB2U(0xA26E, 0x0000AE);
	ChangeMB2U(0xA26F, 0x002122);
	ChangeMB2U(0xA270, 0x0000A4);
	ChangeMB2U(0xA661, 0x000386);
	ChangeMB2U(0xA662, 0x000388);
	ChangeMB2U(0xA663, 0x000389);
	ChangeMB2U(0xA664, 0x00038A);
	ChangeMB2U(0xA665, 0x0003AA);
	ChangeMB2U(0xA667, 0x00038C);
	ChangeMB2U(0xA669, 0x00038E);
	ChangeMB2U(0xA66A, 0x0003AB);
	ChangeMB2U(0xA66C, 0x00038F);
	ChangeMB2U(0xA671, 0x0003AC);
	ChangeMB2U(0xA672, 0x0003AD);
	ChangeMB2U(0xA673, 0x0003AE);
	ChangeMB2U(0xA674, 0x0003AF);
	ChangeMB2U(0xA675, 0x0003CA);
	ChangeMB2U(0xA676, 0x000390);
	ChangeMB2U(0xA677, 0x0003CC);
	ChangeMB2U(0xA678, 0x0003C2);
	ChangeMB2U(0xA679, 0x0003CD);
	ChangeMB2U(0xA67A, 0x0003CB);
	ChangeMB2U(0xA67B, 0x0003B0);
	ChangeMB2U(0xA67C, 0x0003CE);
	ChangeMB2U(0xA742, 0x000402);
	ChangeMB2U(0xA743, 0x000403);
	ChangeMB2U(0xA744, 0x000404);
	ChangeMB2U(0xA745, 0x000405);
	ChangeMB2U(0xA746, 0x000406);
	ChangeMB2U(0xA747, 0x000407);
	ChangeMB2U(0xA748, 0x000408);
	ChangeMB2U(0xA749, 0x000409);
	ChangeMB2U(0xA74A, 0x00040A);
	ChangeMB2U(0xA74B, 0x00040B);
	ChangeMB2U(0xA74C, 0x00040C);
	ChangeMB2U(0xA74D, 0x00040E);
	ChangeMB2U(0xA74E, 0x00040F);
	ChangeMB2U(0xA772, 0x000452);
	ChangeMB2U(0xA773, 0x000453);
	ChangeMB2U(0xA774, 0x000454);
	ChangeMB2U(0xA775, 0x000455);
	ChangeMB2U(0xA776, 0x000456);
	ChangeMB2U(0xA777, 0x000457);
	ChangeMB2U(0xA778, 0x000458);
	ChangeMB2U(0xA779, 0x000459);
	ChangeMB2U(0xA77A, 0x00045A);
	ChangeMB2U(0xA77B, 0x00045B);
	ChangeMB2U(0xA77C, 0x00045C);
	ChangeMB2U(0xA77D, 0x00045E);
	ChangeMB2U(0xA77E, 0x00045F);
	ChangeMB2U(0xADB5, 0x002160);
	ChangeMB2U(0xADB6, 0x002161);
	ChangeMB2U(0xADB7, 0x002162);
	ChangeMB2U(0xADB8, 0x002163);
	ChangeMB2U(0xADB9, 0x002164);
	ChangeMB2U(0xADBA, 0x002165);
	ChangeMB2U(0xADBB, 0x002166);
	ChangeMB2U(0xADBC, 0x002167);
	ChangeMB2U(0xADBD, 0x002168);
	ChangeMB2U(0xADBE, 0x002169);
	ChangeMB2U(0xADE2, 0x002116);
	ChangeMB2U(0xADE4, 0x002121);
	ChangeMB2U(0xADEA, 0x003231);
	ChangeMB2U(0xB021, 0x004E02);
	ChangeMB2U(0xB022, 0x004E04);
	ChangeMB2U(0xB023, 0x004E05);
	ChangeMB2U(0xB024, 0x004E0C);
	ChangeMB2U(0xB025, 0x004E12);
	ChangeMB2U(0xB026, 0x004E1F);
	ChangeMB2U(0xB027, 0x004E23);
	ChangeMB2U(0xB028, 0x004E24);
	ChangeMB2U(0xB02A, 0x004E2B);
	ChangeMB2U(0xB02B, 0x004E2E);
	ChangeMB2U(0xB02C, 0x004E2F);
	ChangeMB2U(0xB02D, 0x004E30);
	ChangeMB2U(0xB02E, 0x004E35);
	ChangeMB2U(0xB02F, 0x004E40);
	ChangeMB2U(0xB030, 0x004E41);
	ChangeMB2U(0xB031, 0x004E44);
	ChangeMB2U(0xB032, 0x004E47);
	ChangeMB2U(0xB033, 0x004E51);
	ChangeMB2U(0xB034, 0x004E5A);
	ChangeMB2U(0xB035, 0x004E5C);
	ChangeMB2U(0xB036, 0x004E63);
	ChangeMB2U(0xB037, 0x004E68);
	ChangeMB2U(0xB038, 0x004E69);
	ChangeMB2U(0xB039, 0x004E74);
	ChangeMB2U(0xB03A, 0x004E75);
	ChangeMB2U(0xB03B, 0x004E79);
	ChangeMB2U(0xB03C, 0x004E7F);
	ChangeMB2U(0xB03D, 0x004E8D);
	ChangeMB2U(0xB03E, 0x004E96);
	ChangeMB2U(0xB03F, 0x004E97);
	ChangeMB2U(0xB040, 0x004E9D);
	ChangeMB2U(0xB041, 0x004EAF);
	ChangeMB2U(0xB042, 0x004EB9);
	ChangeMB2U(0xB043, 0x004EC3);
	ChangeMB2U(0xB044, 0x004ED0);
	ChangeMB2U(0xB045, 0x004EDA);
	ChangeMB2U(0xB046, 0x004EDB);
	ChangeMB2U(0xB047, 0x004EE0);
	ChangeMB2U(0xB049, 0x004EE2);
	ChangeMB2U(0xB04A, 0x004EE8);
	ChangeMB2U(0xB04B, 0x004EEF);
	ChangeMB2U(0xB04C, 0x004EF1);
	ChangeMB2U(0xB04D, 0x004EF3);
	ChangeMB2U(0xB04E, 0x004EF5);
	ChangeMB2U(0xB04F, 0x004EFD);
	ChangeMB2U(0xB050, 0x004EFE);
	ChangeMB2U(0xB051, 0x004EFF);
	ChangeMB2U(0xB053, 0x004F02);
	ChangeMB2U(0xB055, 0x004F08);
	ChangeMB2U(0xB056, 0x004F0B);
	ChangeMB2U(0xB057, 0x004F0C);
	ChangeMB2U(0xB058, 0x004F12);
	ChangeMB2U(0xB059, 0x004F15);
	ChangeMB2U(0xB05A, 0x004F16);
	ChangeMB2U(0xB05B, 0x004F17);
	ChangeMB2U(0xB05C, 0x004F19);
	ChangeMB2U(0xB05D, 0x004F2E);
	ChangeMB2U(0xB05E, 0x004F31);
	ChangeMB2U(0xB05F, 0x004F60);
	ChangeMB2U(0xB060, 0x004F33);
	ChangeMB2U(0xB061, 0x004F35);
	ChangeMB2U(0xB062, 0x004F37);
	ChangeMB2U(0xB064, 0x004F3B);
	ChangeMB2U(0xB065, 0x004F3E);
	ChangeMB2U(0xB066, 0x004F40);
	ChangeMB2U(0xB067, 0x004F42);
	ChangeMB2U(0xB068, 0x004F48);
	ChangeMB2U(0xB069, 0x004F49);
	ChangeMB2U(0xB06A, 0x004F4B);
	ChangeMB2U(0xB06B, 0x004F4C);
	ChangeMB2U(0xB06C, 0x004F52);
	ChangeMB2U(0xB06D, 0x004F54);
	ChangeMB2U(0xB06F, 0x004F58);
	ChangeMB2U(0xB070, 0x004F5F);
	ChangeMB2U(0xB071, 0x004F63);
	ChangeMB2U(0xB072, 0x004F6A);
	ChangeMB2U(0xB073, 0x004F6C);
	ChangeMB2U(0xB074, 0x004F6E);
	ChangeMB2U(0xB075, 0x004F71);
	ChangeMB2U(0xB076, 0x004F77);
	ChangeMB2U(0xB077, 0x004F78);
	ChangeMB2U(0xB078, 0x004F79);
	ChangeMB2U(0xB079, 0x004F7A);
	ChangeMB2U(0xB07A, 0x004F7D);
	ChangeMB2U(0xB07B, 0x004F7E);
	ChangeMB2U(0xB07C, 0x004F81);
	ChangeMB2U(0xB07D, 0x004F82);
	ChangeMB2U(0xB07E, 0x004F84);
	ChangeMB2U(0xB121, 0x004F85);
	ChangeMB2U(0xB122, 0x004F89);
	ChangeMB2U(0xB124, 0x004F8C);
	ChangeMB2U(0xB125, 0x004F8E);
	ChangeMB2U(0xB126, 0x004F90);
	ChangeMB2U(0xB128, 0x004F93);
	ChangeMB2U(0xB12A, 0x004F97);
	ChangeMB2U(0xB12B, 0x004F99);
	ChangeMB2U(0xB12D, 0x004F9E);
	ChangeMB2U(0xB12E, 0x004F9F);
	ChangeMB2U(0xB12F, 0x004FB2);
	ChangeMB2U(0xB130, 0x004FB7);
	ChangeMB2U(0xB131, 0x004FB9);
	ChangeMB2U(0xB132, 0x004FBB);
	ChangeMB2U(0xB133, 0x004FBC);
	ChangeMB2U(0xB134, 0x004FBD);
	ChangeMB2U(0xB135, 0x004FBE);
	ChangeMB2U(0xB136, 0x004FC0);
	ChangeMB2U(0xB137, 0x004FC1);
	ChangeMB2U(0xB138, 0x004FC5);
	ChangeMB2U(0xB139, 0x004FC6);
	ChangeMB2U(0xB13A, 0x004FC8);
	ChangeMB2U(0xB13C, 0x004FCB);
	ChangeMB2U(0xB13D, 0x004FCC);
	ChangeMB2U(0xB13F, 0x004FCF);
	ChangeMB2U(0xB140, 0x004FD2);
	ChangeMB2U(0xB141, 0x004FDC);
	ChangeMB2U(0xB142, 0x004FE0);
	ChangeMB2U(0xB143, 0x004FE2);
	ChangeMB2U(0xB144, 0x004FF0);
	ChangeMB2U(0xB145, 0x004FF2);
	ChangeMB2U(0xB146, 0x004FFC);
	ChangeMB2U(0xB147, 0x004FFD);
	ChangeMB2U(0xB149, 0x005000);
	ChangeMB2U(0xB14A, 0x005001);
	ChangeMB2U(0xB14B, 0x005004);
	ChangeMB2U(0xB14C, 0x005007);
	ChangeMB2U(0xB14D, 0x00500A);
	ChangeMB2U(0xB14E, 0x00500C);
	ChangeMB2U(0xB14F, 0x00500E);
	ChangeMB2U(0xB150, 0x005010);
	ChangeMB2U(0xB151, 0x005013);
	ChangeMB2U(0xB152, 0x005017);
	ChangeMB2U(0xB153, 0x005018);
	ChangeMB2U(0xB154, 0x00501B);
	ChangeMB2U(0xB155, 0x00501C);
	ChangeMB2U(0xB156, 0x00501D);
	ChangeMB2U(0xB159, 0x005027);
	ChangeMB2U(0xB15A, 0x00502E);
	ChangeMB2U(0xB15B, 0x005030);
	ChangeMB2U(0xB15C, 0x005032);
	ChangeMB2U(0xB15D, 0x005033);
	ChangeMB2U(0xB15E, 0x005035);
	ChangeMB2U(0xB160, 0x005041);
	ChangeMB2U(0xB162, 0x005045);
	ChangeMB2U(0xB164, 0x00504A);
	ChangeMB2U(0xB165, 0x00504C);
	ChangeMB2U(0xB166, 0x00504E);
	ChangeMB2U(0xB167, 0x005051);
	ChangeMB2U(0xB168, 0x005052);
	ChangeMB2U(0xB169, 0x005053);
	ChangeMB2U(0xB16A, 0x005057);
	ChangeMB2U(0xB16B, 0x005059);
	ChangeMB2U(0xB16C, 0x00505F);
	ChangeMB2U(0xB16D, 0x005060);
	ChangeMB2U(0xB16E, 0x005062);
	ChangeMB2U(0xB16F, 0x005063);
	ChangeMB2U(0xB170, 0x005066);
	ChangeMB2U(0xB171, 0x005067);
	ChangeMB2U(0xB172, 0x00506A);
	ChangeMB2U(0xB173, 0x00506D);
	ChangeMB2U(0xB175, 0x005071);
	ChangeMB2U(0xB176, 0x00503B);
	ChangeMB2U(0xB177, 0x005081);
	ChangeMB2U(0xB178, 0x005083);
	ChangeMB2U(0xB179, 0x005084);
	ChangeMB2U(0xB17A, 0x005086);
	ChangeMB2U(0xB17B, 0x00508A);
	ChangeMB2U(0xB17C, 0x00508E);
	ChangeMB2U(0xB17D, 0x00508F);
	ChangeMB2U(0xB17E, 0x005090);
	ChangeMB2U(0xB221, 0x005092);
	ChangeMB2U(0xB222, 0x005093);
	ChangeMB2U(0xB224, 0x005096);
	ChangeMB2U(0xB225, 0x00509B);
	ChangeMB2U(0xB226, 0x00509C);
	ChangeMB2U(0xB227, 0x00509E);
	ChangeMB2U(0xB228, 0x00509F);
	ChangeMB2U(0xB229, 0x0050A0);
	ChangeMB2U(0xB22A, 0x0050A1);
	ChangeMB2U(0xB22B, 0x0050A2);
	ChangeMB2U(0xB22C, 0x0050AA);
	ChangeMB2U(0xB22D, 0x0050AF);
	ChangeMB2U(0xB22E, 0x0050B0);
	ChangeMB2U(0xB22F, 0x0050B9);
	ChangeMB2U(0xB230, 0x0050BA);
	ChangeMB2U(0xB231, 0x0050BD);
	ChangeMB2U(0xB232, 0x0050C0);
	ChangeMB2U(0xB233, 0x0050C3);
	ChangeMB2U(0xB234, 0x0050C4);
	ChangeMB2U(0xB235, 0x0050C7);
	ChangeMB2U(0xB236, 0x0050CC);
	ChangeMB2U(0xB237, 0x0050CE);
	ChangeMB2U(0xB238, 0x0050D0);
	ChangeMB2U(0xB239, 0x0050D3);
	ChangeMB2U(0xB23A, 0x0050D4);
	ChangeMB2U(0xB23C, 0x0050DC);
	ChangeMB2U(0xB23D, 0x0050DD);
	ChangeMB2U(0xB23E, 0x0050DF);
	ChangeMB2U(0xB23F, 0x0050E2);
	ChangeMB2U(0xB240, 0x0050E4);
	ChangeMB2U(0xB241, 0x0050E6);
	ChangeMB2U(0xB242, 0x0050E8);
	ChangeMB2U(0xB243, 0x0050E9);
	ChangeMB2U(0xB244, 0x0050EF);
	ChangeMB2U(0xB245, 0x0050F1);
	ChangeMB2U(0xB246, 0x0050F6);
	ChangeMB2U(0xB247, 0x0050FA);
	ChangeMB2U(0xB248, 0x0050FE);
	ChangeMB2U(0xB249, 0x005103);
	ChangeMB2U(0xB24A, 0x005106);
	ChangeMB2U(0xB24B, 0x005107);
	ChangeMB2U(0xB24C, 0x005108);
	ChangeMB2U(0xB24D, 0x00510B);
	ChangeMB2U(0xB24E, 0x00510C);
	ChangeMB2U(0xB24F, 0x00510D);
	ChangeMB2U(0xB250, 0x00510E);
	ChangeMB2U(0xB251, 0x0050F2);
	ChangeMB2U(0xB252, 0x005110);
	ChangeMB2U(0xB253, 0x005117);
	ChangeMB2U(0xB254, 0x005119);
	ChangeMB2U(0xB255, 0x00511B);
	ChangeMB2U(0xB256, 0x00511C);
	ChangeMB2U(0xB257, 0x00511D);
	ChangeMB2U(0xB258, 0x00511E);
	ChangeMB2U(0xB259, 0x005123);
	ChangeMB2U(0xB25A, 0x005127);
	ChangeMB2U(0xB25B, 0x005128);
	ChangeMB2U(0xB25C, 0x00512C);
	ChangeMB2U(0xB25D, 0x00512D);
	ChangeMB2U(0xB25E, 0x00512F);
	ChangeMB2U(0xB25F, 0x005131);
	ChangeMB2U(0xB260, 0x005133);
	ChangeMB2U(0xB261, 0x005134);
	ChangeMB2U(0xB262, 0x005135);
	ChangeMB2U(0xB263, 0x005138);
	ChangeMB2U(0xB264, 0x005139);
	ChangeMB2U(0xB265, 0x005142);
	ChangeMB2U(0xB267, 0x00514F);
	ChangeMB2U(0xB268, 0x005153);
	ChangeMB2U(0xB269, 0x005155);
	ChangeMB2U(0xB26A, 0x005157);
	ChangeMB2U(0xB26B, 0x005158);
	ChangeMB2U(0xB26C, 0x00515F);
	ChangeMB2U(0xB26E, 0x005166);
	ChangeMB2U(0xB26F, 0x00517E);
	ChangeMB2U(0xB270, 0x005183);
	ChangeMB2U(0xB271, 0x005184);
	ChangeMB2U(0xB272, 0x00518B);
	ChangeMB2U(0xB273, 0x00518E);
	ChangeMB2U(0xB274, 0x005198);
	ChangeMB2U(0xB276, 0x0051A1);
	ChangeMB2U(0xB277, 0x0051A3);
	ChangeMB2U(0xB278, 0x0051AD);
	ChangeMB2U(0xB279, 0x0051B8);
	ChangeMB2U(0xB27A, 0x0051BA);
	ChangeMB2U(0xB27B, 0x0051BC);
	ChangeMB2U(0xB27D, 0x0051BF);
	ChangeMB2U(0xB27E, 0x0051C2);
	ChangeMB2U(0xB321, 0x0051C8);
	ChangeMB2U(0xB322, 0x0051CF);
	ChangeMB2U(0xB323, 0x0051D1);
	ChangeMB2U(0xB324, 0x0051D2);
	ChangeMB2U(0xB325, 0x0051D3);
	ChangeMB2U(0xB326, 0x0051D5);
	ChangeMB2U(0xB327, 0x0051D8);
	ChangeMB2U(0xB328, 0x0051DE);
	ChangeMB2U(0xB329, 0x0051E2);
	ChangeMB2U(0xB32A, 0x0051E5);
	ChangeMB2U(0xB32B, 0x0051EE);
	ChangeMB2U(0xB32C, 0x0051F2);
	ChangeMB2U(0xB32D, 0x0051F3);
	ChangeMB2U(0xB32E, 0x0051F4);
	ChangeMB2U(0xB32F, 0x0051F7);
	ChangeMB2U(0xB330, 0x005201);
	ChangeMB2U(0xB331, 0x005202);
	ChangeMB2U(0xB332, 0x005205);
	ChangeMB2U(0xB333, 0x005212);
	ChangeMB2U(0xB334, 0x005213);
	ChangeMB2U(0xB336, 0x005216);
	ChangeMB2U(0xB337, 0x005218);
	ChangeMB2U(0xB338, 0x005222);
	ChangeMB2U(0xB339, 0x005228);
	ChangeMB2U(0xB33A, 0x005231);
	ChangeMB2U(0xB33B, 0x005232);
	ChangeMB2U(0xB33C, 0x005235);
	ChangeMB2U(0xB33D, 0x00523C);
	ChangeMB2U(0xB33E, 0x005245);
	ChangeMB2U(0xB33F, 0x005249);
	ChangeMB2U(0xB340, 0x005255);
	ChangeMB2U(0xB341, 0x005257);
	ChangeMB2U(0xB342, 0x005258);
	ChangeMB2U(0xB343, 0x00525A);
	ChangeMB2U(0xB344, 0x00525C);
	ChangeMB2U(0xB345, 0x00525F);
	ChangeMB2U(0xB346, 0x005260);
	ChangeMB2U(0xB347, 0x005261);
	ChangeMB2U(0xB348, 0x005266);
	ChangeMB2U(0xB349, 0x00526E);
	ChangeMB2U(0xB34A, 0x005277);
	ChangeMB2U(0xB34B, 0x005278);
	ChangeMB2U(0xB34C, 0x005279);
	ChangeMB2U(0xB34D, 0x005280);
	ChangeMB2U(0xB34E, 0x005282);
	ChangeMB2U(0xB34F, 0x005285);
	ChangeMB2U(0xB350, 0x00528A);
	ChangeMB2U(0xB351, 0x00528C);
	ChangeMB2U(0xB352, 0x005293);
	ChangeMB2U(0xB353, 0x005295);
	ChangeMB2U(0xB354, 0x005296);
	ChangeMB2U(0xB355, 0x005297);
	ChangeMB2U(0xB356, 0x005298);
	ChangeMB2U(0xB357, 0x00529A);
	ChangeMB2U(0xB359, 0x0052A4);
	ChangeMB2U(0xB35A, 0x0052A5);
	ChangeMB2U(0xB35C, 0x0052A7);
	ChangeMB2U(0xB35E, 0x0052B0);
	ChangeMB2U(0xB35F, 0x0052B6);
	ChangeMB2U(0xB360, 0x0052B7);
	ChangeMB2U(0xB361, 0x0052B8);
	ChangeMB2U(0xB362, 0x0052BA);
	ChangeMB2U(0xB363, 0x0052BB);
	ChangeMB2U(0xB364, 0x0052BD);
	ChangeMB2U(0xB366, 0x0052C4);
	ChangeMB2U(0xB367, 0x0052C6);
	ChangeMB2U(0xB368, 0x0052C8);
	ChangeMB2U(0xB369, 0x0052CC);
	ChangeMB2U(0xB36A, 0x0052CF);
	ChangeMB2U(0xB36B, 0x0052D1);
	ChangeMB2U(0xB36C, 0x0052D4);
	ChangeMB2U(0xB36D, 0x0052D6);
	ChangeMB2U(0xB36F, 0x0052DC);
	ChangeMB2U(0xB370, 0x0052E1);
	ChangeMB2U(0xB371, 0x0052E5);
	ChangeMB2U(0xB372, 0x0052E8);
	ChangeMB2U(0xB373, 0x0052E9);
	ChangeMB2U(0xB374, 0x0052EA);
	ChangeMB2U(0xB375, 0x0052EC);
	ChangeMB2U(0xB376, 0x0052F0);
	ChangeMB2U(0xB377, 0x0052F1);
	ChangeMB2U(0xB378, 0x0052F4);
	ChangeMB2U(0xB379, 0x0052F6);
	ChangeMB2U(0xB37A, 0x0052F7);
	ChangeMB2U(0xB37C, 0x005303);
	ChangeMB2U(0xB37D, 0x00530A);
	ChangeMB2U(0xB37E, 0x00530B);
	ChangeMB2U(0xB421, 0x00530C);
	ChangeMB2U(0xB422, 0x005311);
	ChangeMB2U(0xB423, 0x005313);
	ChangeMB2U(0xB424, 0x005318);
	ChangeMB2U(0xB425, 0x00531B);
	ChangeMB2U(0xB426, 0x00531C);
	ChangeMB2U(0xB427, 0x00531E);
	ChangeMB2U(0xB428, 0x00531F);
	ChangeMB2U(0xB429, 0x005325);
	ChangeMB2U(0xB42A, 0x005327);
	ChangeMB2U(0xB42B, 0x005328);
	ChangeMB2U(0xB42C, 0x005329);
	ChangeMB2U(0xB42D, 0x00532B);
	ChangeMB2U(0xB42E, 0x00532C);
	ChangeMB2U(0xB42F, 0x00532D);
	ChangeMB2U(0xB430, 0x005330);
	ChangeMB2U(0xB431, 0x005332);
	ChangeMB2U(0xB432, 0x005335);
	ChangeMB2U(0xB433, 0x00533C);
	ChangeMB2U(0xB434, 0x00533D);
	ChangeMB2U(0xB435, 0x00533E);
	ChangeMB2U(0xB436, 0x005342);
	ChangeMB2U(0xB437, 0x00534C);
	ChangeMB2U(0xB438, 0x00534B);
	ChangeMB2U(0xB439, 0x005359);
	ChangeMB2U(0xB43A, 0x00535B);
	ChangeMB2U(0xB43B, 0x005361);
	ChangeMB2U(0xB43C, 0x005363);
	ChangeMB2U(0xB43D, 0x005365);
	ChangeMB2U(0xB43E, 0x00536C);
	ChangeMB2U(0xB43F, 0x00536D);
	ChangeMB2U(0xB441, 0x005379);
	ChangeMB2U(0xB442, 0x00537E);
	ChangeMB2U(0xB443, 0x005383);
	ChangeMB2U(0xB444, 0x005387);
	ChangeMB2U(0xB445, 0x005388);
	ChangeMB2U(0xB446, 0x00538E);
	ChangeMB2U(0xB448, 0x005394);
	ChangeMB2U(0xB449, 0x005399);
	ChangeMB2U(0xB44A, 0x00539D);
	ChangeMB2U(0xB44B, 0x0053A1);
	ChangeMB2U(0xB44C, 0x0053A4);
	ChangeMB2U(0xB44D, 0x0053AA);
	ChangeMB2U(0xB44E, 0x0053AB);
	ChangeMB2U(0xB44F, 0x0053AF);
	ChangeMB2U(0xB451, 0x0053B4);
	ChangeMB2U(0xB452, 0x0053B5);
	ChangeMB2U(0xB453, 0x0053B7);
	ChangeMB2U(0xB454, 0x0053B8);
	ChangeMB2U(0xB455, 0x0053BA);
	ChangeMB2U(0xB456, 0x0053BD);
	ChangeMB2U(0xB457, 0x0053C0);
	ChangeMB2U(0xB458, 0x0053C5);
	ChangeMB2U(0xB459, 0x0053CF);
	ChangeMB2U(0xB45A, 0x0053D2);
	ChangeMB2U(0xB45B, 0x0053D3);
	ChangeMB2U(0xB45C, 0x0053D5);
	ChangeMB2U(0xB45D, 0x0053DA);
	ChangeMB2U(0xB45F, 0x0053DE);
	ChangeMB2U(0xB460, 0x0053E0);
	ChangeMB2U(0xB461, 0x0053E6);
	ChangeMB2U(0xB462, 0x0053E7);
	ChangeMB2U(0xB463, 0x0053F5);
	ChangeMB2U(0xB464, 0x005402);
	ChangeMB2U(0xB465, 0x005413);
	ChangeMB2U(0xB466, 0x00541A);
	ChangeMB2U(0xB467, 0x005421);
	ChangeMB2U(0xB468, 0x005427);
	ChangeMB2U(0xB469, 0x005428);
	ChangeMB2U(0xB46A, 0x00542A);
	ChangeMB2U(0xB46B, 0x00542F);
	ChangeMB2U(0xB46C, 0x005431);
	ChangeMB2U(0xB46D, 0x005434);
	ChangeMB2U(0xB46E, 0x005435);
	ChangeMB2U(0xB46F, 0x005443);
	ChangeMB2U(0xB470, 0x005444);
	ChangeMB2U(0xB471, 0x005447);
	ChangeMB2U(0xB472, 0x00544D);
	ChangeMB2U(0xB473, 0x00544F);
	ChangeMB2U(0xB474, 0x00545E);
	ChangeMB2U(0xB475, 0x005462);
	ChangeMB2U(0xB476, 0x005464);
	ChangeMB2U(0xB477, 0x005466);
	ChangeMB2U(0xB478, 0x005467);
	ChangeMB2U(0xB479, 0x005469);
	ChangeMB2U(0xB47A, 0x00546B);
	ChangeMB2U(0xB47B, 0x00546D);
	ChangeMB2U(0xB47C, 0x00546E);
	ChangeMB2U(0xB47D, 0x005474);
	ChangeMB2U(0xB47E, 0x00547F);
	ChangeMB2U(0xB521, 0x005481);
	ChangeMB2U(0xB522, 0x005483);
	ChangeMB2U(0xB523, 0x005485);
	ChangeMB2U(0xB524, 0x005488);
	ChangeMB2U(0xB525, 0x005489);
	ChangeMB2U(0xB526, 0x00548D);
	ChangeMB2U(0xB527, 0x005491);
	ChangeMB2U(0xB528, 0x005495);
	ChangeMB2U(0xB529, 0x005496);
	ChangeMB2U(0xB52B, 0x00549F);
	ChangeMB2U(0xB52C, 0x0054A1);
	ChangeMB2U(0xB52D, 0x0054A6);
	ChangeMB2U(0xB52E, 0x0054A7);
	ChangeMB2U(0xB530, 0x0054AA);
	ChangeMB2U(0xB531, 0x0054AD);
	ChangeMB2U(0xB532, 0x0054AE);
	ChangeMB2U(0xB533, 0x0054B1);
	ChangeMB2U(0xB534, 0x0054B7);
	ChangeMB2U(0xB535, 0x0054B9);
	ChangeMB2U(0xB536, 0x0054BA);
	ChangeMB2U(0xB537, 0x0054BB);
	ChangeMB2U(0xB538, 0x0054BF);
	ChangeMB2U(0xB539, 0x0054C6);
	ChangeMB2U(0xB53A, 0x0054CA);
	ChangeMB2U(0xB53B, 0x0054CD);
	ChangeMB2U(0xB53C, 0x0054CE);
	ChangeMB2U(0xB53D, 0x0054E0);
	ChangeMB2U(0xB53E, 0x0054EA);
	ChangeMB2U(0xB53F, 0x0054EC);
	ChangeMB2U(0xB540, 0x0054EF);
	ChangeMB2U(0xB541, 0x0054F6);
	ChangeMB2U(0xB542, 0x0054FC);
	ChangeMB2U(0xB543, 0x0054FE);
	ChangeMB2U(0xB545, 0x005500);
	ChangeMB2U(0xB546, 0x005501);
	ChangeMB2U(0xB547, 0x005505);
	ChangeMB2U(0xB548, 0x005508);
	ChangeMB2U(0xB549, 0x005509);
	ChangeMB2U(0xB54A, 0x00550C);
	ChangeMB2U(0xB54B, 0x00550D);
	ChangeMB2U(0xB54C, 0x00550E);
	ChangeMB2U(0xB54D, 0x005515);
	ChangeMB2U(0xB54E, 0x00552A);
	ChangeMB2U(0xB54F, 0x00552B);
	ChangeMB2U(0xB550, 0x005532);
	ChangeMB2U(0xB551, 0x005535);
	ChangeMB2U(0xB552, 0x005536);
	ChangeMB2U(0xB553, 0x00553B);
	ChangeMB2U(0xB554, 0x00553C);
	ChangeMB2U(0xB555, 0x00553D);
	ChangeMB2U(0xB556, 0x005541);
	ChangeMB2U(0xB557, 0x005547);
	ChangeMB2U(0xB558, 0x005549);
	ChangeMB2U(0xB559, 0x00554A);
	ChangeMB2U(0xB55A, 0x00554D);
	ChangeMB2U(0xB55B, 0x005550);
	ChangeMB2U(0xB55C, 0x005551);
	ChangeMB2U(0xB55D, 0x005558);
	ChangeMB2U(0xB55E, 0x00555A);
	ChangeMB2U(0xB55F, 0x00555B);
	ChangeMB2U(0xB560, 0x00555E);
	ChangeMB2U(0xB561, 0x005560);
	ChangeMB2U(0xB562, 0x005561);
	ChangeMB2U(0xB563, 0x005564);
	ChangeMB2U(0xB564, 0x005566);
	ChangeMB2U(0xB565, 0x00557F);
	ChangeMB2U(0xB566, 0x005581);
	ChangeMB2U(0xB567, 0x005582);
	ChangeMB2U(0xB569, 0x005588);
	ChangeMB2U(0xB56A, 0x00558E);
	ChangeMB2U(0xB56B, 0x00558F);
	ChangeMB2U(0xB56C, 0x005591);
	ChangeMB2U(0xB56D, 0x005592);
	ChangeMB2U(0xB56E, 0x005593);
	ChangeMB2U(0xB56F, 0x005594);
	ChangeMB2U(0xB570, 0x005597);
	ChangeMB2U(0xB571, 0x0055A3);
	ChangeMB2U(0xB572, 0x0055A4);
	ChangeMB2U(0xB573, 0x0055AD);
	ChangeMB2U(0xB574, 0x0055B2);
	ChangeMB2U(0xB575, 0x0055BF);
	ChangeMB2U(0xB576, 0x0055C1);
	ChangeMB2U(0xB577, 0x0055C3);
	ChangeMB2U(0xB578, 0x0055C6);
	ChangeMB2U(0xB579, 0x0055C9);
	ChangeMB2U(0xB57A, 0x0055CB);
	ChangeMB2U(0xB57B, 0x0055CC);
	ChangeMB2U(0xB57C, 0x0055CE);
	ChangeMB2U(0xB57D, 0x0055D1);
	ChangeMB2U(0xB57E, 0x0055D2);
	ChangeMB2U(0xB621, 0x0055D3);
	ChangeMB2U(0xB622, 0x0055D7);
	ChangeMB2U(0xB623, 0x0055D8);
	ChangeMB2U(0xB624, 0x0055DB);
	ChangeMB2U(0xB625, 0x0055DE);
	ChangeMB2U(0xB626, 0x0055E2);
	ChangeMB2U(0xB627, 0x0055E9);
	ChangeMB2U(0xB628, 0x0055F6);
	ChangeMB2U(0xB629, 0x0055FF);
	ChangeMB2U(0xB62A, 0x005605);
	ChangeMB2U(0xB62B, 0x005608);
	ChangeMB2U(0xB62C, 0x00560A);
	ChangeMB2U(0xB62D, 0x00560D);
	ChangeMB2U(0xB62E, 0x00560E);
	ChangeMB2U(0xB62F, 0x00560F);
	ChangeMB2U(0xB630, 0x005610);
	ChangeMB2U(0xB631, 0x005611);
	ChangeMB2U(0xB632, 0x005612);
	ChangeMB2U(0xB633, 0x005619);
	ChangeMB2U(0xB634, 0x00562C);
	ChangeMB2U(0xB635, 0x005630);
	ChangeMB2U(0xB636, 0x005633);
	ChangeMB2U(0xB637, 0x005635);
	ChangeMB2U(0xB638, 0x005637);
	ChangeMB2U(0xB639, 0x005639);
	ChangeMB2U(0xB63A, 0x00563B);
	ChangeMB2U(0xB63B, 0x00563C);
	ChangeMB2U(0xB63C, 0x00563D);
	ChangeMB2U(0xB63D, 0x00563F);
	ChangeMB2U(0xB63E, 0x005640);
	ChangeMB2U(0xB63F, 0x005641);
	ChangeMB2U(0xB640, 0x005643);
	ChangeMB2U(0xB641, 0x005644);
	ChangeMB2U(0xB642, 0x005646);
	ChangeMB2U(0xB643, 0x005649);
	ChangeMB2U(0xB644, 0x00564B);
	ChangeMB2U(0xB645, 0x00564D);
	ChangeMB2U(0xB646, 0x00564F);
	ChangeMB2U(0xB647, 0x005654);
	ChangeMB2U(0xB648, 0x00565E);
	ChangeMB2U(0xB649, 0x005660);
	ChangeMB2U(0xB64A, 0x005661);
	ChangeMB2U(0xB64B, 0x005662);
	ChangeMB2U(0xB64C, 0x005663);
	ChangeMB2U(0xB64D, 0x005666);
	ChangeMB2U(0xB64E, 0x005669);
	ChangeMB2U(0xB64F, 0x00566D);
	ChangeMB2U(0xB650, 0x00566F);
	ChangeMB2U(0xB651, 0x005671);
	ChangeMB2U(0xB652, 0x005672);
	ChangeMB2U(0xB653, 0x005675);
	ChangeMB2U(0xB654, 0x005684);
	ChangeMB2U(0xB655, 0x005685);
	ChangeMB2U(0xB656, 0x005688);
	ChangeMB2U(0xB657, 0x00568B);
	ChangeMB2U(0xB658, 0x00568C);
	ChangeMB2U(0xB659, 0x005695);
	ChangeMB2U(0xB65A, 0x005699);
	ChangeMB2U(0xB65B, 0x00569A);
	ChangeMB2U(0xB65C, 0x00569D);
	ChangeMB2U(0xB65D, 0x00569E);
	ChangeMB2U(0xB65E, 0x00569F);
	ChangeMB2U(0xB65F, 0x0056A6);
	ChangeMB2U(0xB660, 0x0056A7);
	ChangeMB2U(0xB661, 0x0056A8);
	ChangeMB2U(0xB662, 0x0056A9);
	ChangeMB2U(0xB663, 0x0056AB);
	ChangeMB2U(0xB664, 0x0056AC);
	ChangeMB2U(0xB665, 0x0056AD);
	ChangeMB2U(0xB666, 0x0056B1);
	ChangeMB2U(0xB667, 0x0056B3);
	ChangeMB2U(0xB668, 0x0056B7);
	ChangeMB2U(0xB669, 0x0056BE);
	ChangeMB2U(0xB66A, 0x0056C5);
	ChangeMB2U(0xB66B, 0x0056C9);
	ChangeMB2U(0xB66C, 0x0056CA);
	ChangeMB2U(0xB66D, 0x0056CB);
	ChangeMB2U(0xB66E, 0x0056CF);
	ChangeMB2U(0xB66F, 0x0056D0);
	ChangeMB2U(0xB670, 0x0056CC);
	ChangeMB2U(0xB671, 0x0056CD);
	ChangeMB2U(0xB672, 0x0056D9);
	ChangeMB2U(0xB673, 0x0056DC);
	ChangeMB2U(0xB674, 0x0056DD);
	ChangeMB2U(0xB675, 0x0056DF);
	ChangeMB2U(0xB676, 0x0056E1);
	ChangeMB2U(0xB677, 0x0056E4);
	ChangeMB2U(0xB678, 0x0056E5);
	ChangeMB2U(0xB679, 0x0056E6);
	ChangeMB2U(0xB67A, 0x0056E7);
	ChangeMB2U(0xB67B, 0x0056E8);
	ChangeMB2U(0xB67C, 0x0056F1);
	ChangeMB2U(0xB67D, 0x0056EB);
	ChangeMB2U(0xB67E, 0x0056ED);
	ChangeMB2U(0xB721, 0x0056F6);
	ChangeMB2U(0xB722, 0x0056F7);
	ChangeMB2U(0xB723, 0x005701);
	ChangeMB2U(0xB724, 0x005702);
	ChangeMB2U(0xB725, 0x005707);
	ChangeMB2U(0xB726, 0x00570A);
	ChangeMB2U(0xB727, 0x00570C);
	ChangeMB2U(0xB728, 0x005711);
	ChangeMB2U(0xB729, 0x005715);
	ChangeMB2U(0xB72A, 0x00571A);
	ChangeMB2U(0xB72B, 0x00571B);
	ChangeMB2U(0xB72C, 0x00571D);
	ChangeMB2U(0xB72D, 0x005720);
	ChangeMB2U(0xB72E, 0x005722);
	ChangeMB2U(0xB72F, 0x005723);
	ChangeMB2U(0xB730, 0x005724);
	ChangeMB2U(0xB731, 0x005725);
	ChangeMB2U(0xB732, 0x005729);
	ChangeMB2U(0xB733, 0x00572A);
	ChangeMB2U(0xB734, 0x00572C);
	ChangeMB2U(0xB735, 0x00572E);
	ChangeMB2U(0xB736, 0x00572F);
	ChangeMB2U(0xB737, 0x005733);
	ChangeMB2U(0xB738, 0x005734);
	ChangeMB2U(0xB739, 0x00573D);
	ChangeMB2U(0xB73A, 0x00573E);
	ChangeMB2U(0xB73B, 0x00573F);
	ChangeMB2U(0xB73C, 0x005745);
	ChangeMB2U(0xB73D, 0x005746);
	ChangeMB2U(0xB73E, 0x00574C);
	ChangeMB2U(0xB73F, 0x00574D);
	ChangeMB2U(0xB740, 0x005752);
	ChangeMB2U(0xB741, 0x005762);
	ChangeMB2U(0xB743, 0x005767);
	ChangeMB2U(0xB744, 0x005768);
	ChangeMB2U(0xB745, 0x00576B);
	ChangeMB2U(0xB746, 0x00576D);
	ChangeMB2U(0xB747, 0x00576E);
	ChangeMB2U(0xB748, 0x00576F);
	ChangeMB2U(0xB749, 0x005770);
	ChangeMB2U(0xB74A, 0x005771);
	ChangeMB2U(0xB74B, 0x005773);
	ChangeMB2U(0xB74C, 0x005774);
	ChangeMB2U(0xB74D, 0x005775);
	ChangeMB2U(0xB74E, 0x005777);
	ChangeMB2U(0xB74F, 0x005779);
	ChangeMB2U(0xB750, 0x00577A);
	ChangeMB2U(0xB751, 0x00577B);
	ChangeMB2U(0xB752, 0x00577C);
	ChangeMB2U(0xB753, 0x00577E);
	ChangeMB2U(0xB754, 0x005781);
	ChangeMB2U(0xB755, 0x005783);
	ChangeMB2U(0xB756, 0x00578C);
	ChangeMB2U(0xB757, 0x005794);
	ChangeMB2U(0xB758, 0x005797);
	ChangeMB2U(0xB759, 0x005799);
	ChangeMB2U(0xB75A, 0x00579A);
	ChangeMB2U(0xB75B, 0x00579C);
	ChangeMB2U(0xB75C, 0x00579D);
	ChangeMB2U(0xB75D, 0x00579E);
	ChangeMB2U(0xB75E, 0x00579F);
	ChangeMB2U(0xB75F, 0x0057A1);
	ChangeMB2U(0xB760, 0x005795);
	ChangeMB2U(0xB761, 0x0057A7);
	ChangeMB2U(0xB762, 0x0057A8);
	ChangeMB2U(0xB763, 0x0057A9);
	ChangeMB2U(0xB765, 0x0057B8);
	ChangeMB2U(0xB766, 0x0057BD);
	ChangeMB2U(0xB769, 0x0057CC);
	ChangeMB2U(0xB76A, 0x0057CF);
	ChangeMB2U(0xB76B, 0x0057D5);
	ChangeMB2U(0xB76C, 0x0057DD);
	ChangeMB2U(0xB76D, 0x0057DE);
	ChangeMB2U(0xB76E, 0x0057E4);
	ChangeMB2U(0xB76F, 0x0057E6);
	ChangeMB2U(0xB770, 0x0057E7);
	ChangeMB2U(0xB771, 0x0057E9);
	ChangeMB2U(0xB772, 0x0057ED);
	ChangeMB2U(0xB773, 0x0057F0);
	ChangeMB2U(0xB774, 0x0057F5);
	ChangeMB2U(0xB775, 0x0057F6);
	ChangeMB2U(0xB776, 0x0057F8);
	ChangeMB2U(0xB777, 0x0057FD);
	ChangeMB2U(0xB778, 0x0057FE);
	ChangeMB2U(0xB779, 0x0057FF);
	ChangeMB2U(0xB77A, 0x005803);
	ChangeMB2U(0xB77B, 0x005804);
	ChangeMB2U(0xB77C, 0x005808);
	ChangeMB2U(0xB77D, 0x005809);
	ChangeMB2U(0xB77E, 0x0057E1);
	ChangeMB2U(0xB821, 0x00580C);
	ChangeMB2U(0xB822, 0x00580D);
	ChangeMB2U(0xB823, 0x00581B);
	ChangeMB2U(0xB824, 0x00581E);
	ChangeMB2U(0xB825, 0x00581F);
	ChangeMB2U(0xB826, 0x005820);
	ChangeMB2U(0xB827, 0x005826);
	ChangeMB2U(0xB828, 0x005827);
	ChangeMB2U(0xB829, 0x00582D);
	ChangeMB2U(0xB82A, 0x005832);
	ChangeMB2U(0xB82B, 0x005839);
	ChangeMB2U(0xB82C, 0x00583F);
	ChangeMB2U(0xB82D, 0x005849);
	ChangeMB2U(0xB82E, 0x00584C);
	ChangeMB2U(0xB82F, 0x00584D);
	ChangeMB2U(0xB830, 0x00584F);
	ChangeMB2U(0xB831, 0x005850);
	ChangeMB2U(0xB832, 0x005855);
	ChangeMB2U(0xB833, 0x00585F);
	ChangeMB2U(0xB834, 0x005861);
	ChangeMB2U(0xB835, 0x005864);
	ChangeMB2U(0xB836, 0x005867);
	ChangeMB2U(0xB837, 0x005868);
	ChangeMB2U(0xB838, 0x005878);
	ChangeMB2U(0xB839, 0x00587C);
	ChangeMB2U(0xB83A, 0x00587F);
	ChangeMB2U(0xB83B, 0x005880);
	ChangeMB2U(0xB83C, 0x005881);
	ChangeMB2U(0xB83D, 0x005887);
	ChangeMB2U(0xB83E, 0x005888);
	ChangeMB2U(0xB83F, 0x005889);
	ChangeMB2U(0xB840, 0x00588A);
	ChangeMB2U(0xB841, 0x00588C);
	ChangeMB2U(0xB842, 0x00588D);
	ChangeMB2U(0xB843, 0x00588F);
	ChangeMB2U(0xB844, 0x005890);
	ChangeMB2U(0xB845, 0x005894);
	ChangeMB2U(0xB846, 0x005896);
	ChangeMB2U(0xB847, 0x00589D);
	ChangeMB2U(0xB848, 0x0058A0);
	ChangeMB2U(0xB849, 0x0058A1);
	ChangeMB2U(0xB84A, 0x0058A2);
	ChangeMB2U(0xB84B, 0x0058A6);
	ChangeMB2U(0xB84C, 0x0058A9);
	ChangeMB2U(0xB84D, 0x0058B1);
	ChangeMB2U(0xB84F, 0x0058C4);
	ChangeMB2U(0xB850, 0x0058BC);
	ChangeMB2U(0xB851, 0x0058C2);
	ChangeMB2U(0xB852, 0x0058C8);
	ChangeMB2U(0xB853, 0x0058CD);
	ChangeMB2U(0xB854, 0x0058CE);
	ChangeMB2U(0xB855, 0x0058D0);
	ChangeMB2U(0xB856, 0x0058D2);
	ChangeMB2U(0xB857, 0x0058D4);
	ChangeMB2U(0xB858, 0x0058D6);
	ChangeMB2U(0xB859, 0x0058DA);
	ChangeMB2U(0xB85A, 0x0058DD);
	ChangeMB2U(0xB85B, 0x0058E1);
	ChangeMB2U(0xB85C, 0x0058E2);
	ChangeMB2U(0xB85D, 0x0058E9);
	ChangeMB2U(0xB85E, 0x0058F3);
	ChangeMB2U(0xB85F, 0x005905);
	ChangeMB2U(0xB860, 0x005906);
	ChangeMB2U(0xB862, 0x00590C);
	ChangeMB2U(0xB863, 0x005912);
	ChangeMB2U(0xB864, 0x005913);
	ChangeMB2U(0xB865, 0x005914);
	ChangeMB2U(0xB866, 0x008641);
	ChangeMB2U(0xB867, 0x00591D);
	ChangeMB2U(0xB868, 0x005921);
	ChangeMB2U(0xB869, 0x005923);
	ChangeMB2U(0xB86A, 0x005924);
	ChangeMB2U(0xB86B, 0x005928);
	ChangeMB2U(0xB86C, 0x00592F);
	ChangeMB2U(0xB86D, 0x005930);
	ChangeMB2U(0xB86E, 0x005933);
	ChangeMB2U(0xB86F, 0x005935);
	ChangeMB2U(0xB870, 0x005936);
	ChangeMB2U(0xB871, 0x00593F);
	ChangeMB2U(0xB872, 0x005943);
	ChangeMB2U(0xB873, 0x005946);
	ChangeMB2U(0xB874, 0x005952);
	ChangeMB2U(0xB876, 0x005959);
	ChangeMB2U(0xB879, 0x00595E);
	ChangeMB2U(0xB87A, 0x00595F);
	ChangeMB2U(0xB87B, 0x005961);
	ChangeMB2U(0xB87D, 0x00596B);
	ChangeMB2U(0xB87E, 0x00596D);
	ChangeMB2U(0xB921, 0x00596F);
	ChangeMB2U(0xB922, 0x005972);
	ChangeMB2U(0xB923, 0x005975);
	ChangeMB2U(0xB924, 0x005976);
	ChangeMB2U(0xB925, 0x005979);
	ChangeMB2U(0xB926, 0x00597B);
	ChangeMB2U(0xB927, 0x00597C);
	ChangeMB2U(0xB928, 0x00598B);
	ChangeMB2U(0xB929, 0x00598C);
	ChangeMB2U(0xB92A, 0x00598E);
	ChangeMB2U(0xB92B, 0x005992);
	ChangeMB2U(0xB92C, 0x005995);
	ChangeMB2U(0xB92D, 0x005997);
	ChangeMB2U(0xB92E, 0x00599F);
	ChangeMB2U(0xB930, 0x0059A7);
	ChangeMB2U(0xB931, 0x0059AD);
	ChangeMB2U(0xB932, 0x0059AE);
	ChangeMB2U(0xB933, 0x0059AF);
	ChangeMB2U(0xB934, 0x0059B0);
	ChangeMB2U(0xB935, 0x0059B3);
	ChangeMB2U(0xB936, 0x0059B7);
	ChangeMB2U(0xB938, 0x0059BC);
	ChangeMB2U(0xB939, 0x0059C1);
	ChangeMB2U(0xB93A, 0x0059C3);
	ChangeMB2U(0xB93B, 0x0059C4);
	ChangeMB2U(0xB93C, 0x0059C8);
	ChangeMB2U(0xB93D, 0x0059CA);
	ChangeMB2U(0xB93E, 0x0059CD);
	ChangeMB2U(0xB93F, 0x0059D2);
	ChangeMB2U(0xB940, 0x0059DD);
	ChangeMB2U(0xB941, 0x0059DE);
	ChangeMB2U(0xB942, 0x0059DF);
	ChangeMB2U(0xB943, 0x0059E3);
	ChangeMB2U(0xB944, 0x0059E4);
	ChangeMB2U(0xB945, 0x0059E7);
	ChangeMB2U(0xB946, 0x0059EE);
	ChangeMB2U(0xB947, 0x0059EF);
	ChangeMB2U(0xB948, 0x0059F1);
	ChangeMB2U(0xB949, 0x0059F2);
	ChangeMB2U(0xB94A, 0x0059F4);
	ChangeMB2U(0xB94B, 0x0059F7);
	ChangeMB2U(0xB94C, 0x005A00);
	ChangeMB2U(0xB94D, 0x005A04);
	ChangeMB2U(0xB94E, 0x005A0C);
	ChangeMB2U(0xB94F, 0x005A0D);
	ChangeMB2U(0xB950, 0x005A0E);
	ChangeMB2U(0xB951, 0x005A12);
	ChangeMB2U(0xB952, 0x005A13);
	ChangeMB2U(0xB953, 0x005A1E);
	ChangeMB2U(0xB954, 0x005A23);
	ChangeMB2U(0xB955, 0x005A24);
	ChangeMB2U(0xB956, 0x005A27);
	ChangeMB2U(0xB957, 0x005A28);
	ChangeMB2U(0xB958, 0x005A2A);
	ChangeMB2U(0xB959, 0x005A2D);
	ChangeMB2U(0xB95A, 0x005A30);
	ChangeMB2U(0xB95B, 0x005A44);
	ChangeMB2U(0xB95C, 0x005A45);
	ChangeMB2U(0xB95D, 0x005A47);
	ChangeMB2U(0xB95E, 0x005A48);
	ChangeMB2U(0xB95F, 0x005A4C);
	ChangeMB2U(0xB960, 0x005A50);
	ChangeMB2U(0xB961, 0x005A55);
	ChangeMB2U(0xB962, 0x005A5E);
	ChangeMB2U(0xB963, 0x005A63);
	ChangeMB2U(0xB964, 0x005A65);
	ChangeMB2U(0xB965, 0x005A67);
	ChangeMB2U(0xB966, 0x005A6D);
	ChangeMB2U(0xB967, 0x005A77);
	ChangeMB2U(0xB968, 0x005A7A);
	ChangeMB2U(0xB969, 0x005A7B);
	ChangeMB2U(0xB96A, 0x005A7E);
	ChangeMB2U(0xB96B, 0x005A8B);
	ChangeMB2U(0xB96C, 0x005A90);
	ChangeMB2U(0xB96D, 0x005A93);
	ChangeMB2U(0xB96E, 0x005A96);
	ChangeMB2U(0xB96F, 0x005A99);
	ChangeMB2U(0xB970, 0x005A9C);
	ChangeMB2U(0xB971, 0x005A9E);
	ChangeMB2U(0xB972, 0x005A9F);
	ChangeMB2U(0xB973, 0x005AA0);
	ChangeMB2U(0xB974, 0x005AA2);
	ChangeMB2U(0xB975, 0x005AA7);
	ChangeMB2U(0xB976, 0x005AAC);
	ChangeMB2U(0xB977, 0x005AB1);
	ChangeMB2U(0xB978, 0x005AB2);
	ChangeMB2U(0xB979, 0x005AB3);
	ChangeMB2U(0xB97A, 0x005AB5);
	ChangeMB2U(0xB97B, 0x005AB8);
	ChangeMB2U(0xB97C, 0x005ABA);
	ChangeMB2U(0xB97D, 0x005ABB);
	ChangeMB2U(0xB97E, 0x005ABF);
	ChangeMB2U(0xBA21, 0x005AC4);
	ChangeMB2U(0xBA22, 0x005AC6);
	ChangeMB2U(0xBA23, 0x005AC8);
	ChangeMB2U(0xBA24, 0x005ACF);
	ChangeMB2U(0xBA25, 0x005ADA);
	ChangeMB2U(0xBA26, 0x005ADC);
	ChangeMB2U(0xBA27, 0x005AE0);
	ChangeMB2U(0xBA28, 0x005AE5);
	ChangeMB2U(0xBA29, 0x005AEA);
	ChangeMB2U(0xBA2A, 0x005AEE);
	ChangeMB2U(0xBA2B, 0x005AF5);
	ChangeMB2U(0xBA2C, 0x005AF6);
	ChangeMB2U(0xBA2D, 0x005AFD);
	ChangeMB2U(0xBA2E, 0x005B00);
	ChangeMB2U(0xBA2F, 0x005B01);
	ChangeMB2U(0xBA30, 0x005B08);
	ChangeMB2U(0xBA31, 0x005B17);
	ChangeMB2U(0xBA32, 0x005B34);
	ChangeMB2U(0xBA33, 0x005B19);
	ChangeMB2U(0xBA34, 0x005B1B);
	ChangeMB2U(0xBA35, 0x005B1D);
	ChangeMB2U(0xBA36, 0x005B21);
	ChangeMB2U(0xBA37, 0x005B25);
	ChangeMB2U(0xBA38, 0x005B2D);
	ChangeMB2U(0xBA39, 0x005B38);
	ChangeMB2U(0xBA3A, 0x005B41);
	ChangeMB2U(0xBA3B, 0x005B4B);
	ChangeMB2U(0xBA3C, 0x005B4C);
	ChangeMB2U(0xBA3D, 0x005B52);
	ChangeMB2U(0xBA3F, 0x005B5E);
	ChangeMB2U(0xBA40, 0x005B68);
	ChangeMB2U(0xBA41, 0x005B6E);
	ChangeMB2U(0xBA42, 0x005B6F);
	ChangeMB2U(0xBA43, 0x005B7C);
	ChangeMB2U(0xBA44, 0x005B7D);
	ChangeMB2U(0xBA45, 0x005B7E);
	ChangeMB2U(0xBA46, 0x005B7F);
	ChangeMB2U(0xBA47, 0x005B81);
	ChangeMB2U(0xBA48, 0x005B84);
	ChangeMB2U(0xBA49, 0x005B86);
	ChangeMB2U(0xBA4A, 0x005B8A);
	ChangeMB2U(0xBA4B, 0x005B8E);
	ChangeMB2U(0xBA4C, 0x005B90);
	ChangeMB2U(0xBA4D, 0x005B91);
	ChangeMB2U(0xBA4E, 0x005B93);
	ChangeMB2U(0xBA4F, 0x005B94);
	ChangeMB2U(0xBA50, 0x005B96);
	ChangeMB2U(0xBA51, 0x005BA8);
	ChangeMB2U(0xBA52, 0x005BA9);
	ChangeMB2U(0xBA53, 0x005BAC);
	ChangeMB2U(0xBA54, 0x005BAD);
	ChangeMB2U(0xBA55, 0x005BAF);
	ChangeMB2U(0xBA56, 0x005BB1);
	ChangeMB2U(0xBA57, 0x005BB2);
	ChangeMB2U(0xBA58, 0x005BB7);
	ChangeMB2U(0xBA59, 0x005BBA);
	ChangeMB2U(0xBA5A, 0x005BBC);
	ChangeMB2U(0xBA5C, 0x005BC1);
	ChangeMB2U(0xBA5D, 0x005BCD);
	ChangeMB2U(0xBA5E, 0x005BCF);
	ChangeMB2U(0xBA5F, 0x005BD6);
	ChangeMB2U(0xBA60, 0x005BD7);
	ChangeMB2U(0xBA62, 0x005BD9);
	ChangeMB2U(0xBA63, 0x005BDA);
	ChangeMB2U(0xBA64, 0x005BE0);
	ChangeMB2U(0xBA65, 0x005BEF);
	ChangeMB2U(0xBA66, 0x005BF1);
	ChangeMB2U(0xBA67, 0x005BF4);
	ChangeMB2U(0xBA68, 0x005BFD);
	ChangeMB2U(0xBA69, 0x005C0C);
	ChangeMB2U(0xBA6A, 0x005C17);
	ChangeMB2U(0xBA6C, 0x005C1F);
	ChangeMB2U(0xBA6D, 0x005C23);
	ChangeMB2U(0xBA6E, 0x005C26);
	ChangeMB2U(0xBA6F, 0x005C29);
	ChangeMB2U(0xBA70, 0x005C2B);
	ChangeMB2U(0xBA71, 0x005C2C);
	ChangeMB2U(0xBA72, 0x005C2E);
	ChangeMB2U(0xBA73, 0x005C30);
	ChangeMB2U(0xBA74, 0x005C32);
	ChangeMB2U(0xBA75, 0x005C35);
	ChangeMB2U(0xBA76, 0x005C36);
	ChangeMB2U(0xBA77, 0x005C59);
	ChangeMB2U(0xBA78, 0x005C5A);
	ChangeMB2U(0xBA79, 0x005C5C);
	ChangeMB2U(0xBA7A, 0x005C62);
	ChangeMB2U(0xBA7B, 0x005C63);
	ChangeMB2U(0xBA7C, 0x005C67);
	ChangeMB2U(0xBA7D, 0x005C68);
	ChangeMB2U(0xBA7E, 0x005C69);
	ChangeMB2U(0xBB21, 0x005C6D);
	ChangeMB2U(0xBB22, 0x005C70);
	ChangeMB2U(0xBB23, 0x005C74);
	ChangeMB2U(0xBB24, 0x005C75);
	ChangeMB2U(0xBB25, 0x005C7A);
	ChangeMB2U(0xBB26, 0x005C7B);
	ChangeMB2U(0xBB27, 0x005C7C);
	ChangeMB2U(0xBB28, 0x005C7D);
	ChangeMB2U(0xBB29, 0x005C87);
	ChangeMB2U(0xBB2A, 0x005C88);
	ChangeMB2U(0xBB2B, 0x005C8A);
	ChangeMB2U(0xBB2C, 0x005C8F);
	ChangeMB2U(0xBB2D, 0x005C92);
	ChangeMB2U(0xBB2E, 0x005C9D);
	ChangeMB2U(0xBB2F, 0x005C9F);
	ChangeMB2U(0xBB30, 0x005CA0);
	ChangeMB2U(0xBB31, 0x005CA2);
	ChangeMB2U(0xBB32, 0x005CA3);
	ChangeMB2U(0xBB34, 0x005CAA);
	ChangeMB2U(0xBB35, 0x005CB2);
	ChangeMB2U(0xBB36, 0x005CB4);
	ChangeMB2U(0xBB37, 0x005CB5);
	ChangeMB2U(0xBB39, 0x005CC9);
	ChangeMB2U(0xBB3A, 0x005CCB);
	ChangeMB2U(0xBB3B, 0x005CD2);
	ChangeMB2U(0xBB3C, 0x005CDD);
	ChangeMB2U(0xBB3D, 0x005CD7);
	ChangeMB2U(0xBB3E, 0x005CEE);
	ChangeMB2U(0xBB3F, 0x005CF1);
	ChangeMB2U(0xBB40, 0x005CF2);
	ChangeMB2U(0xBB41, 0x005CF4);
	ChangeMB2U(0xBB42, 0x005D01);
	ChangeMB2U(0xBB43, 0x005D06);
	ChangeMB2U(0xBB44, 0x005D0D);
	ChangeMB2U(0xBB45, 0x005D12);
	ChangeMB2U(0xBB46, 0x005D2B);
	ChangeMB2U(0xBB47, 0x005D23);
	ChangeMB2U(0xBB48, 0x005D24);
	ChangeMB2U(0xBB49, 0x005D26);
	ChangeMB2U(0xBB4B, 0x005D31);
	ChangeMB2U(0xBB4C, 0x005D34);
	ChangeMB2U(0xBB4D, 0x005D39);
	ChangeMB2U(0xBB4E, 0x005D3D);
	ChangeMB2U(0xBB4F, 0x005D3F);
	ChangeMB2U(0xBB51, 0x005D43);
	ChangeMB2U(0xBB52, 0x005D46);
	ChangeMB2U(0xBB53, 0x005D48);
	ChangeMB2U(0xBB54, 0x005D55);
	ChangeMB2U(0xBB55, 0x005D51);
	ChangeMB2U(0xBB56, 0x005D59);
	ChangeMB2U(0xBB57, 0x005D4A);
	ChangeMB2U(0xBB58, 0x005D5F);
	ChangeMB2U(0xBB59, 0x005D60);
	ChangeMB2U(0xBB5A, 0x005D61);
	ChangeMB2U(0xBB5B, 0x005D62);
	ChangeMB2U(0xBB5C, 0x005D64);
	ChangeMB2U(0xBB5D, 0x005D6A);
	ChangeMB2U(0xBB5F, 0x005D70);
	ChangeMB2U(0xBB60, 0x005D79);
	ChangeMB2U(0xBB61, 0x005D7A);
	ChangeMB2U(0xBB62, 0x005D7E);
	ChangeMB2U(0xBB63, 0x005D7F);
	ChangeMB2U(0xBB64, 0x005D81);
	ChangeMB2U(0xBB65, 0x005D83);
	ChangeMB2U(0xBB66, 0x005D88);
	ChangeMB2U(0xBB67, 0x005D8A);
	ChangeMB2U(0xBB68, 0x005D92);
	ChangeMB2U(0xBB69, 0x005D93);
	ChangeMB2U(0xBB6A, 0x005D94);
	ChangeMB2U(0xBB6B, 0x005D95);
	ChangeMB2U(0xBB6C, 0x005D99);
	ChangeMB2U(0xBB6D, 0x005D9B);
	ChangeMB2U(0xBB6E, 0x005D9F);
	ChangeMB2U(0xBB6F, 0x005DA0);
	ChangeMB2U(0xBB70, 0x005DA7);
	ChangeMB2U(0xBB71, 0x005DAB);
	ChangeMB2U(0xBB72, 0x005DB0);
	ChangeMB2U(0xBB73, 0x005DB4);
	ChangeMB2U(0xBB76, 0x005DC3);
	ChangeMB2U(0xBB77, 0x005DC7);
	ChangeMB2U(0xBB78, 0x005DCB);
	ChangeMB2U(0xBB7A, 0x005DCE);
	ChangeMB2U(0xBB7B, 0x005DD8);
	ChangeMB2U(0xBB7C, 0x005DD9);
	ChangeMB2U(0xBB7D, 0x005DE0);
	ChangeMB2U(0xBB7E, 0x005DE4);
	ChangeMB2U(0xBC21, 0x005DE9);
	ChangeMB2U(0xBC22, 0x005DF8);
	ChangeMB2U(0xBC23, 0x005DF9);
	ChangeMB2U(0xBC24, 0x005E00);
	ChangeMB2U(0xBC25, 0x005E07);
	ChangeMB2U(0xBC26, 0x005E0D);
	ChangeMB2U(0xBC27, 0x005E12);
	ChangeMB2U(0xBC28, 0x005E14);
	ChangeMB2U(0xBC29, 0x005E15);
	ChangeMB2U(0xBC2A, 0x005E18);
	ChangeMB2U(0xBC2B, 0x005E1F);
	ChangeMB2U(0xBC2C, 0x005E20);
	ChangeMB2U(0xBC2D, 0x005E2E);
	ChangeMB2U(0xBC2E, 0x005E28);
	ChangeMB2U(0xBC2F, 0x005E32);
	ChangeMB2U(0xBC30, 0x005E35);
	ChangeMB2U(0xBC31, 0x005E3E);
	ChangeMB2U(0xBC32, 0x005E4B);
	ChangeMB2U(0xBC33, 0x005E50);
	ChangeMB2U(0xBC34, 0x005E49);
	ChangeMB2U(0xBC35, 0x005E51);
	ChangeMB2U(0xBC36, 0x005E56);
	ChangeMB2U(0xBC37, 0x005E58);
	ChangeMB2U(0xBC38, 0x005E5B);
	ChangeMB2U(0xBC39, 0x005E5C);
	ChangeMB2U(0xBC3A, 0x005E5E);
	ChangeMB2U(0xBC3B, 0x005E68);
	ChangeMB2U(0xBC3C, 0x005E6A);
	ChangeMB2U(0xBC3D, 0x005E6B);
	ChangeMB2U(0xBC3E, 0x005E6C);
	ChangeMB2U(0xBC3F, 0x005E6D);
	ChangeMB2U(0xBC40, 0x005E6E);
	ChangeMB2U(0xBC41, 0x005E70);
	ChangeMB2U(0xBC42, 0x005E80);
	ChangeMB2U(0xBC43, 0x005E8B);
	ChangeMB2U(0xBC44, 0x005E8E);
	ChangeMB2U(0xBC45, 0x005EA2);
	ChangeMB2U(0xBC46, 0x005EA4);
	ChangeMB2U(0xBC47, 0x005EA5);
	ChangeMB2U(0xBC48, 0x005EA8);
	ChangeMB2U(0xBC49, 0x005EAA);
	ChangeMB2U(0xBC4A, 0x005EAC);
	ChangeMB2U(0xBC4B, 0x005EB1);
	ChangeMB2U(0xBC4C, 0x005EB3);
	ChangeMB2U(0xBC4D, 0x005EBD);
	ChangeMB2U(0xBC4E, 0x005EBE);
	ChangeMB2U(0xBC4F, 0x005EBF);
	ChangeMB2U(0xBC50, 0x005EC6);
	ChangeMB2U(0xBC51, 0x005ECC);
	ChangeMB2U(0xBC52, 0x005ECB);
	ChangeMB2U(0xBC53, 0x005ECE);
	ChangeMB2U(0xBC54, 0x005ED1);
	ChangeMB2U(0xBC55, 0x005ED2);
	ChangeMB2U(0xBC56, 0x005ED4);
	ChangeMB2U(0xBC57, 0x005ED5);
	ChangeMB2U(0xBC58, 0x005EDC);
	ChangeMB2U(0xBC59, 0x005EDE);
	ChangeMB2U(0xBC5A, 0x005EE5);
	ChangeMB2U(0xBC5B, 0x005EEB);
	ChangeMB2U(0xBC5C, 0x005F02);
	ChangeMB2U(0xBC5D, 0x005F06);
	ChangeMB2U(0xBC5E, 0x005F07);
	ChangeMB2U(0xBC5F, 0x005F08);
	ChangeMB2U(0xBC60, 0x005F0E);
	ChangeMB2U(0xBC61, 0x005F19);
	ChangeMB2U(0xBC62, 0x005F1C);
	ChangeMB2U(0xBC63, 0x005F1D);
	ChangeMB2U(0xBC65, 0x005F22);
	ChangeMB2U(0xBC66, 0x005F23);
	ChangeMB2U(0xBC67, 0x005F24);
	ChangeMB2U(0xBC68, 0x005F28);
	ChangeMB2U(0xBC69, 0x005F2B);
	ChangeMB2U(0xBC6A, 0x005F2C);
	ChangeMB2U(0xBC6B, 0x005F2E);
	ChangeMB2U(0xBC6C, 0x005F30);
	ChangeMB2U(0xBC6E, 0x005F36);
	ChangeMB2U(0xBC6F, 0x005F3B);
	ChangeMB2U(0xBC70, 0x005F3D);
	ChangeMB2U(0xBC71, 0x005F3F);
	ChangeMB2U(0xBC72, 0x005F40);
	ChangeMB2U(0xBC73, 0x005F44);
	ChangeMB2U(0xBC75, 0x005F47);
	ChangeMB2U(0xBC76, 0x005F4D);
	ChangeMB2U(0xBC77, 0x005F50);
	ChangeMB2U(0xBC78, 0x005F54);
	ChangeMB2U(0xBC79, 0x005F58);
	ChangeMB2U(0xBC7A, 0x005F5B);
	ChangeMB2U(0xBC7B, 0x005F60);
	ChangeMB2U(0xBC7C, 0x005F63);
	ChangeMB2U(0xBC7D, 0x005F64);
	ChangeMB2U(0xBD21, 0x005F6F);
	ChangeMB2U(0xBD22, 0x005F72);
	ChangeMB2U(0xBD23, 0x005F74);
	ChangeMB2U(0xBD24, 0x005F75);
	ChangeMB2U(0xBD25, 0x005F78);
	ChangeMB2U(0xBD26, 0x005F7A);
	ChangeMB2U(0xBD27, 0x005F7D);
	ChangeMB2U(0xBD28, 0x005F7E);
	ChangeMB2U(0xBD29, 0x005F89);
	ChangeMB2U(0xBD2A, 0x005F8D);
	ChangeMB2U(0xBD2B, 0x005F8F);
	ChangeMB2U(0xBD2C, 0x005F96);
	ChangeMB2U(0xBD2D, 0x005F9C);
	ChangeMB2U(0xBD2E, 0x005F9D);
	ChangeMB2U(0xBD2F, 0x005FA2);
	ChangeMB2U(0xBD30, 0x005FA7);
	ChangeMB2U(0xBD31, 0x005FAB);
	ChangeMB2U(0xBD32, 0x005FA4);
	ChangeMB2U(0xBD33, 0x005FAC);
	ChangeMB2U(0xBD34, 0x005FAF);
	ChangeMB2U(0xBD35, 0x005FB0);
	ChangeMB2U(0xBD36, 0x005FB1);
	ChangeMB2U(0xBD37, 0x005FB8);
	ChangeMB2U(0xBD38, 0x005FC4);
	ChangeMB2U(0xBD39, 0x005FC7);
	ChangeMB2U(0xBD3A, 0x005FC8);
	ChangeMB2U(0xBD3B, 0x005FC9);
	ChangeMB2U(0xBD3C, 0x005FCB);
	ChangeMB2U(0xBD3D, 0x005FD0);
	ChangeMB2U(0xBD3E, 0x005FD1);
	ChangeMB2U(0xBD3F, 0x005FD2);
	ChangeMB2U(0xBD40, 0x005FD3);
	ChangeMB2U(0xBD41, 0x005FD4);
	ChangeMB2U(0xBD43, 0x005FE1);
	ChangeMB2U(0xBD44, 0x005FE2);
	ChangeMB2U(0xBD45, 0x005FE8);
	ChangeMB2U(0xBD46, 0x005FE9);
	ChangeMB2U(0xBD47, 0x005FEA);
	ChangeMB2U(0xBD48, 0x005FEC);
	ChangeMB2U(0xBD49, 0x005FED);
	ChangeMB2U(0xBD4A, 0x005FEE);
	ChangeMB2U(0xBD4B, 0x005FEF);
	ChangeMB2U(0xBD4C, 0x005FF2);
	ChangeMB2U(0xBD4D, 0x005FF3);
	ChangeMB2U(0xBD4E, 0x005FF6);
	ChangeMB2U(0xBD4F, 0x005FFA);
	ChangeMB2U(0xBD50, 0x005FFC);
	ChangeMB2U(0xBD51, 0x006007);
	ChangeMB2U(0xBD52, 0x00600A);
	ChangeMB2U(0xBD53, 0x00600D);
	ChangeMB2U(0xBD54, 0x006013);
	ChangeMB2U(0xBD55, 0x006014);
	ChangeMB2U(0xBD56, 0x006017);
	ChangeMB2U(0xBD57, 0x006018);
	ChangeMB2U(0xBD58, 0x00601A);
	ChangeMB2U(0xBD59, 0x00601F);
	ChangeMB2U(0xBD5A, 0x006024);
	ChangeMB2U(0xBD5B, 0x00602D);
	ChangeMB2U(0xBD5C, 0x006033);
	ChangeMB2U(0xBD5D, 0x006035);
	ChangeMB2U(0xBD5E, 0x006040);
	ChangeMB2U(0xBD5F, 0x006047);
	ChangeMB2U(0xBD60, 0x006048);
	ChangeMB2U(0xBD61, 0x006049);
	ChangeMB2U(0xBD62, 0x00604C);
	ChangeMB2U(0xBD63, 0x006051);
	ChangeMB2U(0xBD64, 0x006054);
	ChangeMB2U(0xBD65, 0x006056);
	ChangeMB2U(0xBD66, 0x006057);
	ChangeMB2U(0xBD68, 0x006061);
	ChangeMB2U(0xBD69, 0x006067);
	ChangeMB2U(0xBD6A, 0x006071);
	ChangeMB2U(0xBD6B, 0x00607E);
	ChangeMB2U(0xBD6C, 0x00607F);
	ChangeMB2U(0xBD6D, 0x006082);
	ChangeMB2U(0xBD6E, 0x006086);
	ChangeMB2U(0xBD6F, 0x006088);
	ChangeMB2U(0xBD71, 0x00608E);
	ChangeMB2U(0xBD72, 0x006091);
	ChangeMB2U(0xBD73, 0x006093);
	ChangeMB2U(0xBD74, 0x006095);
	ChangeMB2U(0xBD75, 0x006098);
	ChangeMB2U(0xBD76, 0x00609D);
	ChangeMB2U(0xBD77, 0x00609E);
	ChangeMB2U(0xBD78, 0x0060A2);
	ChangeMB2U(0xBD79, 0x0060A4);
	ChangeMB2U(0xBD7A, 0x0060A5);
	ChangeMB2U(0xBD7B, 0x0060A8);
	ChangeMB2U(0xBD7C, 0x0060B0);
	ChangeMB2U(0xBD7D, 0x0060B1);
	ChangeMB2U(0xBD7E, 0x0060B7);
	ChangeMB2U(0xBE21, 0x0060BB);
	ChangeMB2U(0xBE22, 0x0060BE);
	ChangeMB2U(0xBE23, 0x0060C2);
	ChangeMB2U(0xBE24, 0x0060C4);
	ChangeMB2U(0xBE25, 0x0060C8);
	ChangeMB2U(0xBE26, 0x0060C9);
	ChangeMB2U(0xBE27, 0x0060CA);
	ChangeMB2U(0xBE28, 0x0060CB);
	ChangeMB2U(0xBE29, 0x0060CE);
	ChangeMB2U(0xBE2A, 0x0060CF);
	ChangeMB2U(0xBE2B, 0x0060D4);
	ChangeMB2U(0xBE2D, 0x0060D9);
	ChangeMB2U(0xBE2E, 0x0060DB);
	ChangeMB2U(0xBE2F, 0x0060DD);
	ChangeMB2U(0xBE31, 0x0060E2);
	ChangeMB2U(0xBE32, 0x0060E5);
	ChangeMB2U(0xBE34, 0x0060F5);
	ChangeMB2U(0xBE35, 0x0060F8);
	ChangeMB2U(0xBE36, 0x0060FC);
	ChangeMB2U(0xBE37, 0x0060FD);
	ChangeMB2U(0xBE38, 0x006102);
	ChangeMB2U(0xBE39, 0x006107);
	ChangeMB2U(0xBE3A, 0x00610A);
	ChangeMB2U(0xBE3B, 0x00610C);
	ChangeMB2U(0xBE3C, 0x006110);
	ChangeMB2U(0xBE3E, 0x006112);
	ChangeMB2U(0xBE3F, 0x006113);
	ChangeMB2U(0xBE40, 0x006114);
	ChangeMB2U(0xBE41, 0x006116);
	ChangeMB2U(0xBE42, 0x006117);
	ChangeMB2U(0xBE43, 0x006119);
	ChangeMB2U(0xBE44, 0x00611C);
	ChangeMB2U(0xBE45, 0x00611E);
	ChangeMB2U(0xBE46, 0x006122);
	ChangeMB2U(0xBE47, 0x00612A);
	ChangeMB2U(0xBE48, 0x00612B);
	ChangeMB2U(0xBE4A, 0x006131);
	ChangeMB2U(0xBE4B, 0x006135);
	ChangeMB2U(0xBE4C, 0x006136);
	ChangeMB2U(0xBE4E, 0x006139);
	ChangeMB2U(0xBE4F, 0x006141);
	ChangeMB2U(0xBE50, 0x006145);
	ChangeMB2U(0xBE51, 0x006146);
	ChangeMB2U(0xBE52, 0x006149);
	ChangeMB2U(0xBE53, 0x00615E);
	ChangeMB2U(0xBE54, 0x006160);
	ChangeMB2U(0xBE55, 0x00616C);
	ChangeMB2U(0xBE56, 0x006172);
	ChangeMB2U(0xBE57, 0x006178);
	ChangeMB2U(0xBE58, 0x00617B);
	ChangeMB2U(0xBE59, 0x00617C);
	ChangeMB2U(0xBE5A, 0x00617F);
	ChangeMB2U(0xBE5B, 0x006180);
	ChangeMB2U(0xBE5C, 0x006181);
	ChangeMB2U(0xBE5D, 0x006183);
	ChangeMB2U(0xBE5E, 0x006184);
	ChangeMB2U(0xBE5F, 0x00618B);
	ChangeMB2U(0xBE60, 0x00618D);
	ChangeMB2U(0xBE61, 0x006192);
	ChangeMB2U(0xBE62, 0x006193);
	ChangeMB2U(0xBE63, 0x006197);
	ChangeMB2U(0xBE65, 0x00619C);
	ChangeMB2U(0xBE66, 0x00619D);
	ChangeMB2U(0xBE67, 0x00619F);
	ChangeMB2U(0xBE68, 0x0061A0);
	ChangeMB2U(0xBE69, 0x0061A5);
	ChangeMB2U(0xBE6A, 0x0061A8);
	ChangeMB2U(0xBE6B, 0x0061AA);
	ChangeMB2U(0xBE6C, 0x0061AD);
	ChangeMB2U(0xBE6D, 0x0061B8);
	ChangeMB2U(0xBE6E, 0x0061B9);
	ChangeMB2U(0xBE6F, 0x0061BC);
	ChangeMB2U(0xBE70, 0x0061C0);
	ChangeMB2U(0xBE71, 0x0061C1);
	ChangeMB2U(0xBE72, 0x0061C2);
	ChangeMB2U(0xBE73, 0x0061CE);
	ChangeMB2U(0xBE74, 0x0061CF);
	ChangeMB2U(0xBE75, 0x0061D5);
	ChangeMB2U(0xBE76, 0x0061DC);
	ChangeMB2U(0xBE77, 0x0061DD);
	ChangeMB2U(0xBE78, 0x0061DE);
	ChangeMB2U(0xBE79, 0x0061DF);
	ChangeMB2U(0xBE7A, 0x0061E1);
	ChangeMB2U(0xBE7B, 0x0061E2);
	ChangeMB2U(0xBE7C, 0x0061E7);
	ChangeMB2U(0xBE7D, 0x0061E9);
	ChangeMB2U(0xBE7E, 0x0061E5);
	ChangeMB2U(0xBF21, 0x0061EC);
	ChangeMB2U(0xBF22, 0x0061ED);
	ChangeMB2U(0xBF23, 0x0061EF);
	ChangeMB2U(0xBF24, 0x006201);
	ChangeMB2U(0xBF25, 0x006203);
	ChangeMB2U(0xBF26, 0x006204);
	ChangeMB2U(0xBF27, 0x006207);
	ChangeMB2U(0xBF29, 0x006215);
	ChangeMB2U(0xBF2A, 0x00621C);
	ChangeMB2U(0xBF2B, 0x006220);
	ChangeMB2U(0xBF2C, 0x006222);
	ChangeMB2U(0xBF2D, 0x006223);
	ChangeMB2U(0xBF2E, 0x006227);
	ChangeMB2U(0xBF2F, 0x006229);
	ChangeMB2U(0xBF30, 0x00622B);
	ChangeMB2U(0xBF31, 0x006239);
	ChangeMB2U(0xBF32, 0x00623D);
	ChangeMB2U(0xBF33, 0x006242);
	ChangeMB2U(0xBF34, 0x006243);
	ChangeMB2U(0xBF35, 0x006244);
	ChangeMB2U(0xBF36, 0x006246);
	ChangeMB2U(0xBF37, 0x00624C);
	ChangeMB2U(0xBF38, 0x006250);
	ChangeMB2U(0xBF39, 0x006251);
	ChangeMB2U(0xBF3A, 0x006252);
	ChangeMB2U(0xBF3B, 0x006254);
	ChangeMB2U(0xBF3C, 0x006256);
	ChangeMB2U(0xBF3D, 0x00625A);
	ChangeMB2U(0xBF3E, 0x00625C);
	ChangeMB2U(0xBF3F, 0x006264);
	ChangeMB2U(0xBF40, 0x00626D);
	ChangeMB2U(0xBF41, 0x00626F);
	ChangeMB2U(0xBF42, 0x006273);
	ChangeMB2U(0xBF43, 0x00627A);
	ChangeMB2U(0xBF44, 0x00627D);
	ChangeMB2U(0xBF45, 0x00628D);
	ChangeMB2U(0xBF46, 0x00628E);
	ChangeMB2U(0xBF47, 0x00628F);
	ChangeMB2U(0xBF48, 0x006290);
	ChangeMB2U(0xBF4A, 0x0062A8);
	ChangeMB2U(0xBF4B, 0x0062B3);
	ChangeMB2U(0xBF4C, 0x0062B6);
	ChangeMB2U(0xBF4D, 0x0062B7);
	ChangeMB2U(0xBF4E, 0x0062BA);
	ChangeMB2U(0xBF4F, 0x0062BE);
	ChangeMB2U(0xBF50, 0x0062BF);
	ChangeMB2U(0xBF51, 0x0062C4);
	ChangeMB2U(0xBF52, 0x0062CE);
	ChangeMB2U(0xBF53, 0x0062D5);
	ChangeMB2U(0xBF54, 0x0062D6);
	ChangeMB2U(0xBF55, 0x0062DA);
	ChangeMB2U(0xBF56, 0x0062EA);
	ChangeMB2U(0xBF57, 0x0062F2);
	ChangeMB2U(0xBF58, 0x0062F4);
	ChangeMB2U(0xBF59, 0x0062FC);
	ChangeMB2U(0xBF5A, 0x0062FD);
	ChangeMB2U(0xBF5B, 0x006303);
	ChangeMB2U(0xBF5C, 0x006304);
	ChangeMB2U(0xBF5D, 0x00630A);
	ChangeMB2U(0xBF5E, 0x00630B);
	ChangeMB2U(0xBF5F, 0x00630D);
	ChangeMB2U(0xBF60, 0x006310);
	ChangeMB2U(0xBF61, 0x006313);
	ChangeMB2U(0xBF62, 0x006316);
	ChangeMB2U(0xBF63, 0x006318);
	ChangeMB2U(0xBF64, 0x006329);
	ChangeMB2U(0xBF65, 0x00632A);
	ChangeMB2U(0xBF66, 0x00632D);
	ChangeMB2U(0xBF67, 0x006335);
	ChangeMB2U(0xBF68, 0x006336);
	ChangeMB2U(0xBF69, 0x006339);
	ChangeMB2U(0xBF6A, 0x00633C);
	ChangeMB2U(0xBF6B, 0x006341);
	ChangeMB2U(0xBF6C, 0x006342);
	ChangeMB2U(0xBF6D, 0x006343);
	ChangeMB2U(0xBF6E, 0x006344);
	ChangeMB2U(0xBF6F, 0x006346);
	ChangeMB2U(0xBF70, 0x00634A);
	ChangeMB2U(0xBF71, 0x00634B);
	ChangeMB2U(0xBF72, 0x00634E);
	ChangeMB2U(0xBF73, 0x006352);
	ChangeMB2U(0xBF74, 0x006353);
	ChangeMB2U(0xBF75, 0x006354);
	ChangeMB2U(0xBF76, 0x006358);
	ChangeMB2U(0xBF77, 0x00635B);
	ChangeMB2U(0xBF78, 0x006365);
	ChangeMB2U(0xBF79, 0x006366);
	ChangeMB2U(0xBF7A, 0x00636C);
	ChangeMB2U(0xBF7B, 0x00636D);
	ChangeMB2U(0xBF7C, 0x006371);
	ChangeMB2U(0xBF7D, 0x006374);
	ChangeMB2U(0xBF7E, 0x006375);
	ChangeMB2U(0xC021, 0x006378);
	ChangeMB2U(0xC022, 0x00637C);
	ChangeMB2U(0xC023, 0x00637D);
	ChangeMB2U(0xC024, 0x00637F);
	ChangeMB2U(0xC025, 0x006382);
	ChangeMB2U(0xC026, 0x006384);
	ChangeMB2U(0xC027, 0x006387);
	ChangeMB2U(0xC028, 0x00638A);
	ChangeMB2U(0xC029, 0x006390);
	ChangeMB2U(0xC02A, 0x006394);
	ChangeMB2U(0xC02B, 0x006395);
	ChangeMB2U(0xC02C, 0x006399);
	ChangeMB2U(0xC02D, 0x00639A);
	ChangeMB2U(0xC02E, 0x00639E);
	ChangeMB2U(0xC02F, 0x0063A4);
	ChangeMB2U(0xC030, 0x0063A6);
	ChangeMB2U(0xC031, 0x0063AD);
	ChangeMB2U(0xC032, 0x0063AE);
	ChangeMB2U(0xC033, 0x0063AF);
	ChangeMB2U(0xC034, 0x0063BD);
	ChangeMB2U(0xC035, 0x0063C1);
	ChangeMB2U(0xC036, 0x0063C5);
	ChangeMB2U(0xC037, 0x0063C8);
	ChangeMB2U(0xC038, 0x0063CE);
	ChangeMB2U(0xC039, 0x0063D1);
	ChangeMB2U(0xC03A, 0x0063D3);
	ChangeMB2U(0xC03B, 0x0063D4);
	ChangeMB2U(0xC03C, 0x0063D5);
	ChangeMB2U(0xC03D, 0x0063DC);
	ChangeMB2U(0xC03E, 0x0063E0);
	ChangeMB2U(0xC03F, 0x0063E5);
	ChangeMB2U(0xC040, 0x0063EA);
	ChangeMB2U(0xC041, 0x0063EC);
	ChangeMB2U(0xC042, 0x0063F2);
	ChangeMB2U(0xC043, 0x0063F3);
	ChangeMB2U(0xC045, 0x0063F8);
	ChangeMB2U(0xC046, 0x0063F9);
	ChangeMB2U(0xC047, 0x006409);
	ChangeMB2U(0xC048, 0x00640A);
	ChangeMB2U(0xC049, 0x006410);
	ChangeMB2U(0xC04A, 0x006412);
	ChangeMB2U(0xC04B, 0x006414);
	ChangeMB2U(0xC04C, 0x006418);
	ChangeMB2U(0xC04D, 0x00641E);
	ChangeMB2U(0xC04E, 0x006420);
	ChangeMB2U(0xC04F, 0x006422);
	ChangeMB2U(0xC050, 0x006424);
	ChangeMB2U(0xC051, 0x006425);
	ChangeMB2U(0xC052, 0x006429);
	ChangeMB2U(0xC053, 0x00642A);
	ChangeMB2U(0xC054, 0x00642F);
	ChangeMB2U(0xC055, 0x006430);
	ChangeMB2U(0xC056, 0x006435);
	ChangeMB2U(0xC057, 0x00643D);
	ChangeMB2U(0xC058, 0x00643F);
	ChangeMB2U(0xC059, 0x00644B);
	ChangeMB2U(0xC05A, 0x00644F);
	ChangeMB2U(0xC05B, 0x006451);
	ChangeMB2U(0xC05C, 0x006452);
	ChangeMB2U(0xC05D, 0x006453);
	ChangeMB2U(0xC05E, 0x006454);
	ChangeMB2U(0xC05F, 0x00645A);
	ChangeMB2U(0xC060, 0x00645B);
	ChangeMB2U(0xC061, 0x00645C);
	ChangeMB2U(0xC062, 0x00645D);
	ChangeMB2U(0xC063, 0x00645F);
	ChangeMB2U(0xC065, 0x006461);
	ChangeMB2U(0xC066, 0x006463);
	ChangeMB2U(0xC067, 0x00646D);
	ChangeMB2U(0xC068, 0x006473);
	ChangeMB2U(0xC069, 0x006474);
	ChangeMB2U(0xC06A, 0x00647B);
	ChangeMB2U(0xC06B, 0x00647D);
	ChangeMB2U(0xC06C, 0x006485);
	ChangeMB2U(0xC06D, 0x006487);
	ChangeMB2U(0xC06E, 0x00648F);
	ChangeMB2U(0xC06F, 0x006490);
	ChangeMB2U(0xC070, 0x006491);
	ChangeMB2U(0xC071, 0x006498);
	ChangeMB2U(0xC072, 0x006499);
	ChangeMB2U(0xC073, 0x00649B);
	ChangeMB2U(0xC075, 0x00649F);
	ChangeMB2U(0xC076, 0x0064A1);
	ChangeMB2U(0xC077, 0x0064A3);
	ChangeMB2U(0xC078, 0x0064A6);
	ChangeMB2U(0xC079, 0x0064A8);
	ChangeMB2U(0xC07A, 0x0064AC);
	ChangeMB2U(0xC07B, 0x0064B3);
	ChangeMB2U(0xC07C, 0x0064BD);
	ChangeMB2U(0xC07D, 0x0064BE);
	ChangeMB2U(0xC07E, 0x0064BF);
	ChangeMB2U(0xC121, 0x0064C4);
	ChangeMB2U(0xC122, 0x0064C9);
	ChangeMB2U(0xC123, 0x0064CA);
	ChangeMB2U(0xC124, 0x0064CB);
	ChangeMB2U(0xC125, 0x0064CC);
	ChangeMB2U(0xC127, 0x0064D0);
	ChangeMB2U(0xC128, 0x0064D1);
	ChangeMB2U(0xC129, 0x0064D5);
	ChangeMB2U(0xC12A, 0x0064D7);
	ChangeMB2U(0xC12B, 0x0064E4);
	ChangeMB2U(0xC12C, 0x0064E5);
	ChangeMB2U(0xC12D, 0x0064E9);
	ChangeMB2U(0xC12E, 0x0064EA);
	ChangeMB2U(0xC12F, 0x0064ED);
	ChangeMB2U(0xC130, 0x0064F0);
	ChangeMB2U(0xC131, 0x0064F5);
	ChangeMB2U(0xC132, 0x0064F7);
	ChangeMB2U(0xC133, 0x0064FB);
	ChangeMB2U(0xC134, 0x0064FF);
	ChangeMB2U(0xC135, 0x006501);
	ChangeMB2U(0xC136, 0x006504);
	ChangeMB2U(0xC137, 0x006508);
	ChangeMB2U(0xC138, 0x006509);
	ChangeMB2U(0xC139, 0x00650A);
	ChangeMB2U(0xC13A, 0x00650F);
	ChangeMB2U(0xC13B, 0x006513);
	ChangeMB2U(0xC13C, 0x006514);
	ChangeMB2U(0xC13D, 0x006516);
	ChangeMB2U(0xC13E, 0x006519);
	ChangeMB2U(0xC13F, 0x00651B);
	ChangeMB2U(0xC140, 0x00651E);
	ChangeMB2U(0xC141, 0x00651F);
	ChangeMB2U(0xC142, 0x006522);
	ChangeMB2U(0xC143, 0x006526);
	ChangeMB2U(0xC144, 0x006529);
	ChangeMB2U(0xC145, 0x00652E);
	ChangeMB2U(0xC146, 0x006531);
	ChangeMB2U(0xC147, 0x00653A);
	ChangeMB2U(0xC148, 0x00653C);
	ChangeMB2U(0xC149, 0x00653D);
	ChangeMB2U(0xC14A, 0x006543);
	ChangeMB2U(0xC14B, 0x006547);
	ChangeMB2U(0xC14C, 0x006549);
	ChangeMB2U(0xC14D, 0x006550);
	ChangeMB2U(0xC14E, 0x006552);
	ChangeMB2U(0xC14F, 0x006554);
	ChangeMB2U(0xC150, 0x00655F);
	ChangeMB2U(0xC151, 0x006560);
	ChangeMB2U(0xC152, 0x006567);
	ChangeMB2U(0xC153, 0x00656B);
	ChangeMB2U(0xC154, 0x00657A);
	ChangeMB2U(0xC155, 0x00657D);
	ChangeMB2U(0xC156, 0x006581);
	ChangeMB2U(0xC157, 0x006585);
	ChangeMB2U(0xC158, 0x00658A);
	ChangeMB2U(0xC159, 0x006592);
	ChangeMB2U(0xC15A, 0x006595);
	ChangeMB2U(0xC15B, 0x006598);
	ChangeMB2U(0xC15C, 0x00659D);
	ChangeMB2U(0xC15D, 0x0065A0);
	ChangeMB2U(0xC15E, 0x0065A3);
	ChangeMB2U(0xC15F, 0x0065A6);
	ChangeMB2U(0xC160, 0x0065AE);
	ChangeMB2U(0xC161, 0x0065B2);
	ChangeMB2U(0xC162, 0x0065B3);
	ChangeMB2U(0xC163, 0x0065B4);
	ChangeMB2U(0xC164, 0x0065BF);
	ChangeMB2U(0xC165, 0x0065C2);
	ChangeMB2U(0xC166, 0x0065C8);
	ChangeMB2U(0xC167, 0x0065C9);
	ChangeMB2U(0xC168, 0x0065CE);
	ChangeMB2U(0xC169, 0x0065D0);
	ChangeMB2U(0xC16A, 0x0065D4);
	ChangeMB2U(0xC16B, 0x0065D6);
	ChangeMB2U(0xC16C, 0x0065D8);
	ChangeMB2U(0xC16D, 0x0065DF);
	ChangeMB2U(0xC16E, 0x0065F0);
	ChangeMB2U(0xC16F, 0x0065F2);
	ChangeMB2U(0xC170, 0x0065F4);
	ChangeMB2U(0xC171, 0x0065F5);
	ChangeMB2U(0xC172, 0x0065F9);
	ChangeMB2U(0xC173, 0x0065FE);
	ChangeMB2U(0xC174, 0x0065FF);
	ChangeMB2U(0xC176, 0x006604);
	ChangeMB2U(0xC177, 0x006608);
	ChangeMB2U(0xC179, 0x00660D);
	ChangeMB2U(0xC17A, 0x006611);
	ChangeMB2U(0xC17B, 0x006612);
	ChangeMB2U(0xC17D, 0x006616);
	ChangeMB2U(0xC17E, 0x00661D);
	ChangeMB2U(0xC222, 0x006621);
	ChangeMB2U(0xC223, 0x006622);
	ChangeMB2U(0xC224, 0x006623);
	ChangeMB2U(0xC226, 0x006626);
	ChangeMB2U(0xC227, 0x006629);
	ChangeMB2U(0xC228, 0x00662A);
	ChangeMB2U(0xC229, 0x00662B);
	ChangeMB2U(0xC22A, 0x00662C);
	ChangeMB2U(0xC22C, 0x006630);
	ChangeMB2U(0xC22E, 0x006633);
	ChangeMB2U(0xC22F, 0x006639);
	ChangeMB2U(0xC230, 0x006637);
	ChangeMB2U(0xC231, 0x006640);
	ChangeMB2U(0xC232, 0x006645);
	ChangeMB2U(0xC233, 0x006646);
	ChangeMB2U(0xC234, 0x00664A);
	ChangeMB2U(0xC235, 0x00664C);
	ChangeMB2U(0xC236, 0x006651);
	ChangeMB2U(0xC237, 0x00664E);
	ChangeMB2U(0xC239, 0x006658);
	ChangeMB2U(0xC23B, 0x00665B);
	ChangeMB2U(0xC23C, 0x00665C);
	ChangeMB2U(0xC23D, 0x006660);
	ChangeMB2U(0xC23E, 0x006661);
	ChangeMB2U(0xC240, 0x00666A);
	ChangeMB2U(0xC241, 0x00666B);
	ChangeMB2U(0xC242, 0x00666C);
	ChangeMB2U(0xC243, 0x00667E);
	ChangeMB2U(0xC245, 0x006675);
	ChangeMB2U(0xC246, 0x00667F);
	ChangeMB2U(0xC247, 0x006677);
	ChangeMB2U(0xC248, 0x006678);
	ChangeMB2U(0xC249, 0x006679);
	ChangeMB2U(0xC24A, 0x00667B);
	ChangeMB2U(0xC24B, 0x006680);
	ChangeMB2U(0xC24C, 0x00667C);
	ChangeMB2U(0xC24D, 0x00668B);
	ChangeMB2U(0xC24E, 0x00668C);
	ChangeMB2U(0xC24F, 0x00668D);
	ChangeMB2U(0xC250, 0x006690);
	ChangeMB2U(0xC251, 0x006692);
	ChangeMB2U(0xC253, 0x00669A);
	ChangeMB2U(0xC254, 0x00669B);
	ChangeMB2U(0xC255, 0x00669C);
	ChangeMB2U(0xC256, 0x00669F);
	ChangeMB2U(0xC258, 0x0066A4);
	ChangeMB2U(0xC259, 0x0066AD);
	ChangeMB2U(0xC25A, 0x0066B1);
	ChangeMB2U(0xC25C, 0x0066B5);
	ChangeMB2U(0xC25D, 0x0066BB);
	ChangeMB2U(0xC25F, 0x0066C0);
	ChangeMB2U(0xC260, 0x0066C2);
	ChangeMB2U(0xC261, 0x0066C3);
	ChangeMB2U(0xC262, 0x0066C8);
	ChangeMB2U(0xC263, 0x0066CC);
	ChangeMB2U(0xC264, 0x0066CE);
	ChangeMB2U(0xC265, 0x0066CF);
	ChangeMB2U(0xC266, 0x0066D4);
	ChangeMB2U(0xC267, 0x0066DB);
	ChangeMB2U(0xC268, 0x0066DF);
	ChangeMB2U(0xC269, 0x0066E8);
	ChangeMB2U(0xC26A, 0x0066EB);
	ChangeMB2U(0xC26B, 0x0066EC);
	ChangeMB2U(0xC26C, 0x0066EE);
	ChangeMB2U(0xC26E, 0x006705);
	ChangeMB2U(0xC26F, 0x006707);
	ChangeMB2U(0xC271, 0x006713);
	ChangeMB2U(0xC272, 0x006719);
	ChangeMB2U(0xC273, 0x00671C);
	ChangeMB2U(0xC274, 0x006720);
	ChangeMB2U(0xC275, 0x006722);
	ChangeMB2U(0xC276, 0x006733);
	ChangeMB2U(0xC277, 0x00673E);
	ChangeMB2U(0xC278, 0x006745);
	ChangeMB2U(0xC279, 0x006747);
	ChangeMB2U(0xC27A, 0x006748);
	ChangeMB2U(0xC27B, 0x00674C);
	ChangeMB2U(0xC27C, 0x006754);
	ChangeMB2U(0xC27D, 0x006755);
	ChangeMB2U(0xC27E, 0x00675D);
	ChangeMB2U(0xC322, 0x00676C);
	ChangeMB2U(0xC323, 0x00676E);
	ChangeMB2U(0xC324, 0x006774);
	ChangeMB2U(0xC325, 0x006776);
	ChangeMB2U(0xC326, 0x00677B);
	ChangeMB2U(0xC327, 0x006781);
	ChangeMB2U(0xC328, 0x006784);
	ChangeMB2U(0xC329, 0x00678E);
	ChangeMB2U(0xC32A, 0x00678F);
	ChangeMB2U(0xC32B, 0x006791);
	ChangeMB2U(0xC32C, 0x006793);
	ChangeMB2U(0xC32D, 0x006796);
	ChangeMB2U(0xC32E, 0x006798);
	ChangeMB2U(0xC32F, 0x006799);
	ChangeMB2U(0xC330, 0x00679B);
	ChangeMB2U(0xC331, 0x0067B0);
	ChangeMB2U(0xC332, 0x0067B1);
	ChangeMB2U(0xC333, 0x0067B2);
	ChangeMB2U(0xC334, 0x0067B5);
	ChangeMB2U(0xC336, 0x0067BC);
	ChangeMB2U(0xC337, 0x0067BD);
	ChangeMB2U(0xC338, 0x0067F9);
	ChangeMB2U(0xC33A, 0x0067C2);
	ChangeMB2U(0xC33B, 0x0067C3);
	ChangeMB2U(0xC33C, 0x0067C5);
	ChangeMB2U(0xC33D, 0x0067C8);
	ChangeMB2U(0xC33E, 0x0067C9);
	ChangeMB2U(0xC33F, 0x0067D2);
	ChangeMB2U(0xC340, 0x0067D7);
	ChangeMB2U(0xC341, 0x0067D9);
	ChangeMB2U(0xC342, 0x0067DC);
	ChangeMB2U(0xC343, 0x0067E1);
	ChangeMB2U(0xC344, 0x0067E6);
	ChangeMB2U(0xC345, 0x0067F0);
	ChangeMB2U(0xC346, 0x0067F2);
	ChangeMB2U(0xC347, 0x0067F6);
	ChangeMB2U(0xC348, 0x0067F7);
	ChangeMB2U(0xC34A, 0x006814);
	ChangeMB2U(0xC34B, 0x006819);
	ChangeMB2U(0xC34C, 0x00681D);
	ChangeMB2U(0xC34D, 0x00681F);
	ChangeMB2U(0xC34E, 0x006828);
	ChangeMB2U(0xC34F, 0x006827);
	ChangeMB2U(0xC350, 0x00682C);
	ChangeMB2U(0xC351, 0x00682D);
	ChangeMB2U(0xC352, 0x00682F);
	ChangeMB2U(0xC353, 0x006830);
	ChangeMB2U(0xC354, 0x006831);
	ChangeMB2U(0xC355, 0x006833);
	ChangeMB2U(0xC356, 0x00683B);
	ChangeMB2U(0xC357, 0x00683F);
	ChangeMB2U(0xC359, 0x006845);
	ChangeMB2U(0xC35A, 0x00684A);
	ChangeMB2U(0xC35B, 0x00684C);
	ChangeMB2U(0xC35C, 0x006855);
	ChangeMB2U(0xC35D, 0x006857);
	ChangeMB2U(0xC35E, 0x006858);
	ChangeMB2U(0xC35F, 0x00685B);
	ChangeMB2U(0xC360, 0x00686B);
	ChangeMB2U(0xC361, 0x00686E);
	ChangeMB2U(0xC362, 0x00686F);
	ChangeMB2U(0xC363, 0x006870);
	ChangeMB2U(0xC364, 0x006871);
	ChangeMB2U(0xC365, 0x006872);
	ChangeMB2U(0xC366, 0x006875);
	ChangeMB2U(0xC367, 0x006879);
	ChangeMB2U(0xC368, 0x00687A);
	ChangeMB2U(0xC369, 0x00687B);
	ChangeMB2U(0xC36A, 0x00687C);
	ChangeMB2U(0xC36B, 0x006882);
	ChangeMB2U(0xC36C, 0x006884);
	ChangeMB2U(0xC36D, 0x006886);
	ChangeMB2U(0xC36E, 0x006888);
	ChangeMB2U(0xC36F, 0x006896);
	ChangeMB2U(0xC370, 0x006898);
	ChangeMB2U(0xC371, 0x00689A);
	ChangeMB2U(0xC372, 0x00689C);
	ChangeMB2U(0xC373, 0x0068A1);
	ChangeMB2U(0xC374, 0x0068A3);
	ChangeMB2U(0xC375, 0x0068A5);
	ChangeMB2U(0xC376, 0x0068A9);
	ChangeMB2U(0xC377, 0x0068AA);
	ChangeMB2U(0xC378, 0x0068AE);
	ChangeMB2U(0xC379, 0x0068B2);
	ChangeMB2U(0xC37A, 0x0068BB);
	ChangeMB2U(0xC37B, 0x0068C5);
	ChangeMB2U(0xC37D, 0x0068CC);
	ChangeMB2U(0xC421, 0x0068D0);
	ChangeMB2U(0xC422, 0x0068D1);
	ChangeMB2U(0xC423, 0x0068D3);
	ChangeMB2U(0xC424, 0x0068D6);
	ChangeMB2U(0xC425, 0x0068D9);
	ChangeMB2U(0xC426, 0x0068DC);
	ChangeMB2U(0xC427, 0x0068DD);
	ChangeMB2U(0xC428, 0x0068E5);
	ChangeMB2U(0xC429, 0x0068E8);
	ChangeMB2U(0xC42A, 0x0068EA);
	ChangeMB2U(0xC42B, 0x0068EB);
	ChangeMB2U(0xC42C, 0x0068EC);
	ChangeMB2U(0xC42D, 0x0068ED);
	ChangeMB2U(0xC42E, 0x0068F0);
	ChangeMB2U(0xC42F, 0x0068F1);
	ChangeMB2U(0xC430, 0x0068F5);
	ChangeMB2U(0xC431, 0x0068F6);
	ChangeMB2U(0xC432, 0x0068FB);
	ChangeMB2U(0xC433, 0x0068FC);
	ChangeMB2U(0xC434, 0x0068FD);
	ChangeMB2U(0xC435, 0x006906);
	ChangeMB2U(0xC436, 0x006909);
	ChangeMB2U(0xC437, 0x00690A);
	ChangeMB2U(0xC438, 0x006910);
	ChangeMB2U(0xC439, 0x006911);
	ChangeMB2U(0xC43A, 0x006913);
	ChangeMB2U(0xC43B, 0x006916);
	ChangeMB2U(0xC43C, 0x006917);
	ChangeMB2U(0xC43D, 0x006931);
	ChangeMB2U(0xC43E, 0x006933);
	ChangeMB2U(0xC43F, 0x006935);
	ChangeMB2U(0xC440, 0x006938);
	ChangeMB2U(0xC441, 0x00693B);
	ChangeMB2U(0xC442, 0x006942);
	ChangeMB2U(0xC443, 0x006945);
	ChangeMB2U(0xC444, 0x006949);
	ChangeMB2U(0xC445, 0x00694E);
	ChangeMB2U(0xC446, 0x006957);
	ChangeMB2U(0xC447, 0x00695B);
	ChangeMB2U(0xC448, 0x006963);
	ChangeMB2U(0xC449, 0x006964);
	ChangeMB2U(0xC44A, 0x006965);
	ChangeMB2U(0xC44B, 0x006966);
	ChangeMB2U(0xC44D, 0x006969);
	ChangeMB2U(0xC44E, 0x00696C);
	ChangeMB2U(0xC44F, 0x006970);
	ChangeMB2U(0xC450, 0x006971);
	ChangeMB2U(0xC451, 0x006972);
	ChangeMB2U(0xC452, 0x00697A);
	ChangeMB2U(0xC453, 0x00697B);
	ChangeMB2U(0xC454, 0x00697F);
	ChangeMB2U(0xC455, 0x006980);
	ChangeMB2U(0xC456, 0x00698D);
	ChangeMB2U(0xC457, 0x006992);
	ChangeMB2U(0xC458, 0x006996);
	ChangeMB2U(0xC45A, 0x0069A1);
	ChangeMB2U(0xC45B, 0x0069A5);
	ChangeMB2U(0xC45C, 0x0069A6);
	ChangeMB2U(0xC45D, 0x0069A8);
	ChangeMB2U(0xC45E, 0x0069AB);
	ChangeMB2U(0xC45F, 0x0069AD);
	ChangeMB2U(0xC460, 0x0069AF);
	ChangeMB2U(0xC461, 0x0069B7);
	ChangeMB2U(0xC462, 0x0069B8);
	ChangeMB2U(0xC463, 0x0069BA);
	ChangeMB2U(0xC464, 0x0069BC);
	ChangeMB2U(0xC465, 0x0069C5);
	ChangeMB2U(0xC466, 0x0069C8);
	ChangeMB2U(0xC467, 0x0069D1);
	ChangeMB2U(0xC468, 0x0069D6);
	ChangeMB2U(0xC469, 0x0069D7);
	ChangeMB2U(0xC46B, 0x0069E5);
	ChangeMB2U(0xC46C, 0x0069EE);
	ChangeMB2U(0xC46D, 0x0069EF);
	ChangeMB2U(0xC46E, 0x0069F1);
	ChangeMB2U(0xC46F, 0x0069F3);
	ChangeMB2U(0xC470, 0x0069F5);
	ChangeMB2U(0xC471, 0x0069FE);
	ChangeMB2U(0xC472, 0x006A00);
	ChangeMB2U(0xC473, 0x006A01);
	ChangeMB2U(0xC474, 0x006A03);
	ChangeMB2U(0xC475, 0x006A0F);
	ChangeMB2U(0xC476, 0x006A11);
	ChangeMB2U(0xC477, 0x006A15);
	ChangeMB2U(0xC478, 0x006A1A);
	ChangeMB2U(0xC479, 0x006A1D);
	ChangeMB2U(0xC47A, 0x006A20);
	ChangeMB2U(0xC47B, 0x006A24);
	ChangeMB2U(0xC47C, 0x006A28);
	ChangeMB2U(0xC47E, 0x006A32);
	ChangeMB2U(0xC521, 0x006A34);
	ChangeMB2U(0xC522, 0x006A37);
	ChangeMB2U(0xC523, 0x006A3B);
	ChangeMB2U(0xC524, 0x006A3E);
	ChangeMB2U(0xC525, 0x006A3F);
	ChangeMB2U(0xC526, 0x006A45);
	ChangeMB2U(0xC528, 0x006A49);
	ChangeMB2U(0xC529, 0x006A4A);
	ChangeMB2U(0xC52A, 0x006A4E);
	ChangeMB2U(0xC52B, 0x006A50);
	ChangeMB2U(0xC52C, 0x006A51);
	ChangeMB2U(0xC52D, 0x006A52);
	ChangeMB2U(0xC52E, 0x006A55);
	ChangeMB2U(0xC52F, 0x006A56);
	ChangeMB2U(0xC530, 0x006A5B);
	ChangeMB2U(0xC531, 0x006A64);
	ChangeMB2U(0xC532, 0x006A67);
	ChangeMB2U(0xC533, 0x006A6A);
	ChangeMB2U(0xC534, 0x006A71);
	ChangeMB2U(0xC537, 0x006A81);
	ChangeMB2U(0xC538, 0x006A83);
	ChangeMB2U(0xC539, 0x006A86);
	ChangeMB2U(0xC53A, 0x006A87);
	ChangeMB2U(0xC53B, 0x006A89);
	ChangeMB2U(0xC53C, 0x006A8B);
	ChangeMB2U(0xC53D, 0x006A91);
	ChangeMB2U(0xC53E, 0x006A9B);
	ChangeMB2U(0xC53F, 0x006A9D);
	ChangeMB2U(0xC540, 0x006A9E);
	ChangeMB2U(0xC541, 0x006A9F);
	ChangeMB2U(0xC542, 0x006AA5);
	ChangeMB2U(0xC543, 0x006AAB);
	ChangeMB2U(0xC544, 0x006AAF);
	ChangeMB2U(0xC545, 0x006AB0);
	ChangeMB2U(0xC546, 0x006AB1);
	ChangeMB2U(0xC547, 0x006AB4);
	ChangeMB2U(0xC548, 0x006ABD);
	ChangeMB2U(0xC549, 0x006ABE);
	ChangeMB2U(0xC54A, 0x006ABF);
	ChangeMB2U(0xC54B, 0x006AC6);
	ChangeMB2U(0xC54C, 0x006AC9);
	ChangeMB2U(0xC54D, 0x006AC8);
	ChangeMB2U(0xC54E, 0x006ACC);
	ChangeMB2U(0xC54F, 0x006AD0);
	ChangeMB2U(0xC550, 0x006AD4);
	ChangeMB2U(0xC551, 0x006AD5);
	ChangeMB2U(0xC552, 0x006AD6);
	ChangeMB2U(0xC553, 0x006ADC);
	ChangeMB2U(0xC554, 0x006ADD);
	ChangeMB2U(0xC556, 0x006AE7);
	ChangeMB2U(0xC557, 0x006AEC);
	ChangeMB2U(0xC558, 0x006AF0);
	ChangeMB2U(0xC559, 0x006AF1);
	ChangeMB2U(0xC55A, 0x006AF2);
	ChangeMB2U(0xC55B, 0x006AFC);
	ChangeMB2U(0xC55C, 0x006AFD);
	ChangeMB2U(0xC55D, 0x006B02);
	ChangeMB2U(0xC55E, 0x006B03);
	ChangeMB2U(0xC55F, 0x006B06);
	ChangeMB2U(0xC560, 0x006B07);
	ChangeMB2U(0xC561, 0x006B09);
	ChangeMB2U(0xC562, 0x006B0F);
	ChangeMB2U(0xC563, 0x006B10);
	ChangeMB2U(0xC564, 0x006B11);
	ChangeMB2U(0xC565, 0x006B17);
	ChangeMB2U(0xC566, 0x006B1B);
	ChangeMB2U(0xC567, 0x006B1E);
	ChangeMB2U(0xC568, 0x006B24);
	ChangeMB2U(0xC569, 0x006B28);
	ChangeMB2U(0xC56A, 0x006B2B);
	ChangeMB2U(0xC56B, 0x006B2C);
	ChangeMB2U(0xC56C, 0x006B2F);
	ChangeMB2U(0xC56D, 0x006B35);
	ChangeMB2U(0xC56E, 0x006B36);
	ChangeMB2U(0xC56F, 0x006B3B);
	ChangeMB2U(0xC570, 0x006B3F);
	ChangeMB2U(0xC571, 0x006B46);
	ChangeMB2U(0xC572, 0x006B4A);
	ChangeMB2U(0xC573, 0x006B4D);
	ChangeMB2U(0xC574, 0x006B52);
	ChangeMB2U(0xC575, 0x006B56);
	ChangeMB2U(0xC576, 0x006B58);
	ChangeMB2U(0xC577, 0x006B5D);
	ChangeMB2U(0xC578, 0x006B60);
	ChangeMB2U(0xC579, 0x006B67);
	ChangeMB2U(0xC57A, 0x006B6B);
	ChangeMB2U(0xC57B, 0x006B6E);
	ChangeMB2U(0xC57C, 0x006B70);
	ChangeMB2U(0xC57D, 0x006B75);
	ChangeMB2U(0xC57E, 0x006B7D);
	ChangeMB2U(0xC621, 0x006B7E);
	ChangeMB2U(0xC622, 0x006B82);
	ChangeMB2U(0xC623, 0x006B85);
	ChangeMB2U(0xC624, 0x006B97);
	ChangeMB2U(0xC625, 0x006B9B);
	ChangeMB2U(0xC626, 0x006B9F);
	ChangeMB2U(0xC627, 0x006BA0);
	ChangeMB2U(0xC628, 0x006BA2);
	ChangeMB2U(0xC629, 0x006BA3);
	ChangeMB2U(0xC62A, 0x006BA8);
	ChangeMB2U(0xC62B, 0x006BA9);
	ChangeMB2U(0xC62C, 0x006BAC);
	ChangeMB2U(0xC62D, 0x006BAD);
	ChangeMB2U(0xC62E, 0x006BAE);
	ChangeMB2U(0xC62F, 0x006BB0);
	ChangeMB2U(0xC630, 0x006BB8);
	ChangeMB2U(0xC631, 0x006BB9);
	ChangeMB2U(0xC632, 0x006BBD);
	ChangeMB2U(0xC633, 0x006BBE);
	ChangeMB2U(0xC634, 0x006BC3);
	ChangeMB2U(0xC635, 0x006BC4);
	ChangeMB2U(0xC636, 0x006BC9);
	ChangeMB2U(0xC637, 0x006BCC);
	ChangeMB2U(0xC639, 0x006BDA);
	ChangeMB2U(0xC63A, 0x006BE1);
	ChangeMB2U(0xC63B, 0x006BE3);
	ChangeMB2U(0xC63C, 0x006BE6);
	ChangeMB2U(0xC63D, 0x006BE7);
	ChangeMB2U(0xC63E, 0x006BEE);
	ChangeMB2U(0xC63F, 0x006BF1);
	ChangeMB2U(0xC640, 0x006BF7);
	ChangeMB2U(0xC641, 0x006BF9);
	ChangeMB2U(0xC642, 0x006BFF);
	ChangeMB2U(0xC643, 0x006C02);
	ChangeMB2U(0xC644, 0x006C04);
	ChangeMB2U(0xC645, 0x006C05);
	ChangeMB2U(0xC646, 0x006C09);
	ChangeMB2U(0xC647, 0x006C0D);
	ChangeMB2U(0xC648, 0x006C0E);
	ChangeMB2U(0xC649, 0x006C10);
	ChangeMB2U(0xC64A, 0x006C12);
	ChangeMB2U(0xC64B, 0x006C19);
	ChangeMB2U(0xC64C, 0x006C1F);
	ChangeMB2U(0xC64D, 0x006C26);
	ChangeMB2U(0xC64E, 0x006C27);
	ChangeMB2U(0xC64F, 0x006C28);
	ChangeMB2U(0xC650, 0x006C2C);
	ChangeMB2U(0xC651, 0x006C2E);
	ChangeMB2U(0xC652, 0x006C33);
	ChangeMB2U(0xC653, 0x006C35);
	ChangeMB2U(0xC654, 0x006C36);
	ChangeMB2U(0xC655, 0x006C3A);
	ChangeMB2U(0xC656, 0x006C3B);
	ChangeMB2U(0xC658, 0x006C4A);
	ChangeMB2U(0xC659, 0x006C4B);
	ChangeMB2U(0xC65A, 0x006C4D);
	ChangeMB2U(0xC65B, 0x006C4F);
	ChangeMB2U(0xC65C, 0x006C52);
	ChangeMB2U(0xC65D, 0x006C54);
	ChangeMB2U(0xC65E, 0x006C59);
	ChangeMB2U(0xC65F, 0x006C5B);
	ChangeMB2U(0xC661, 0x006C6B);
	ChangeMB2U(0xC662, 0x006C6D);
	ChangeMB2U(0xC664, 0x006C74);
	ChangeMB2U(0xC665, 0x006C76);
	ChangeMB2U(0xC666, 0x006C78);
	ChangeMB2U(0xC667, 0x006C79);
	ChangeMB2U(0xC668, 0x006C7B);
	ChangeMB2U(0xC669, 0x006C85);
	ChangeMB2U(0xC66B, 0x006C87);
	ChangeMB2U(0xC66C, 0x006C89);
	ChangeMB2U(0xC66D, 0x006C94);
	ChangeMB2U(0xC66E, 0x006C95);
	ChangeMB2U(0xC66F, 0x006C97);
	ChangeMB2U(0xC670, 0x006C98);
	ChangeMB2U(0xC671, 0x006C9C);
	ChangeMB2U(0xC672, 0x006C9F);
	ChangeMB2U(0xC673, 0x006CB0);
	ChangeMB2U(0xC674, 0x006CB2);
	ChangeMB2U(0xC675, 0x006CB4);
	ChangeMB2U(0xC676, 0x006CC2);
	ChangeMB2U(0xC677, 0x006CC6);
	ChangeMB2U(0xC678, 0x006CCD);
	ChangeMB2U(0xC679, 0x006CCF);
	ChangeMB2U(0xC67A, 0x006CD0);
	ChangeMB2U(0xC67B, 0x006CD1);
	ChangeMB2U(0xC67C, 0x006CD2);
	ChangeMB2U(0xC67D, 0x006CD4);
	ChangeMB2U(0xC67E, 0x006CD6);
	ChangeMB2U(0xC722, 0x006CDC);
	ChangeMB2U(0xC723, 0x006CE0);
	ChangeMB2U(0xC724, 0x006CE7);
	ChangeMB2U(0xC725, 0x006CE9);
	ChangeMB2U(0xC726, 0x006CEB);
	ChangeMB2U(0xC727, 0x006CEC);
	ChangeMB2U(0xC728, 0x006CEE);
	ChangeMB2U(0xC729, 0x006CF2);
	ChangeMB2U(0xC72A, 0x006CF4);
	ChangeMB2U(0xC72C, 0x006D07);
	ChangeMB2U(0xC72D, 0x006D0A);
	ChangeMB2U(0xC72E, 0x006D0E);
	ChangeMB2U(0xC72F, 0x006D0F);
	ChangeMB2U(0xC730, 0x006D11);
	ChangeMB2U(0xC731, 0x006D13);
	ChangeMB2U(0xC732, 0x006D1A);
	ChangeMB2U(0xC733, 0x006D26);
	ChangeMB2U(0xC734, 0x006D27);
	ChangeMB2U(0xC735, 0x006D28);
	ChangeMB2U(0xC736, 0x006C67);
	ChangeMB2U(0xC737, 0x006D2E);
	ChangeMB2U(0xC738, 0x006D2F);
	ChangeMB2U(0xC739, 0x006D31);
	ChangeMB2U(0xC73A, 0x006D39);
	ChangeMB2U(0xC73B, 0x006D3C);
	ChangeMB2U(0xC73C, 0x006D3F);
	ChangeMB2U(0xC73D, 0x006D57);
	ChangeMB2U(0xC73E, 0x006D5E);
	ChangeMB2U(0xC73F, 0x006D5F);
	ChangeMB2U(0xC740, 0x006D61);
	ChangeMB2U(0xC741, 0x006D65);
	ChangeMB2U(0xC742, 0x006D67);
	ChangeMB2U(0xC744, 0x006D70);
	ChangeMB2U(0xC745, 0x006D7C);
	ChangeMB2U(0xC746, 0x006D82);
	ChangeMB2U(0xC748, 0x006D91);
	ChangeMB2U(0xC749, 0x006D92);
	ChangeMB2U(0xC74A, 0x006D94);
	ChangeMB2U(0xC74C, 0x006D97);
	ChangeMB2U(0xC74D, 0x006D98);
	ChangeMB2U(0xC74E, 0x006DAA);
	ChangeMB2U(0xC750, 0x006DB4);
	ChangeMB2U(0xC751, 0x006DB7);
	ChangeMB2U(0xC752, 0x006DB9);
	ChangeMB2U(0xC753, 0x006DBD);
	ChangeMB2U(0xC754, 0x006DBF);
	ChangeMB2U(0xC755, 0x006DC4);
	ChangeMB2U(0xC756, 0x006DC8);
	ChangeMB2U(0xC757, 0x006DCA);
	ChangeMB2U(0xC758, 0x006DCE);
	ChangeMB2U(0xC75A, 0x006DD6);
	ChangeMB2U(0xC75B, 0x006DDB);
	ChangeMB2U(0xC75C, 0x006DDD);
	ChangeMB2U(0xC75D, 0x006DDF);
	ChangeMB2U(0xC75E, 0x006DE0);
	ChangeMB2U(0xC75F, 0x006DE2);
	ChangeMB2U(0xC760, 0x006DE5);
	ChangeMB2U(0xC761, 0x006DE9);
	ChangeMB2U(0xC762, 0x006DEF);
	ChangeMB2U(0xC763, 0x006DF0);
	ChangeMB2U(0xC764, 0x006DF4);
	ChangeMB2U(0xC765, 0x006DF6);
	ChangeMB2U(0xC767, 0x006E00);
	ChangeMB2U(0xC768, 0x006E04);
	ChangeMB2U(0xC769, 0x006E1E);
	ChangeMB2U(0xC76A, 0x006E22);
	ChangeMB2U(0xC76C, 0x006E32);
	ChangeMB2U(0xC76D, 0x006E36);
	ChangeMB2U(0xC76F, 0x006E3B);
	ChangeMB2U(0xC771, 0x006E44);
	ChangeMB2U(0xC772, 0x006E45);
	ChangeMB2U(0xC773, 0x006E48);
	ChangeMB2U(0xC774, 0x006E49);
	ChangeMB2U(0xC775, 0x006E4B);
	ChangeMB2U(0xC776, 0x006E4F);
	ChangeMB2U(0xC777, 0x006E51);
	ChangeMB2U(0xC778, 0x006E52);
	ChangeMB2U(0xC779, 0x006E53);
	ChangeMB2U(0xC77A, 0x006E54);
	ChangeMB2U(0xC77B, 0x006E57);
	ChangeMB2U(0xC77D, 0x006E5D);
	ChangeMB2U(0xC77E, 0x006E5E);
	ChangeMB2U(0xC821, 0x006E62);
	ChangeMB2U(0xC822, 0x006E63);
	ChangeMB2U(0xC823, 0x006E68);
	ChangeMB2U(0xC824, 0x006E73);
	ChangeMB2U(0xC825, 0x006E7B);
	ChangeMB2U(0xC826, 0x006E7D);
	ChangeMB2U(0xC827, 0x006E8D);
	ChangeMB2U(0xC828, 0x006E93);
	ChangeMB2U(0xC829, 0x006E99);
	ChangeMB2U(0xC82A, 0x006EA0);
	ChangeMB2U(0xC82B, 0x006EA7);
	ChangeMB2U(0xC82C, 0x006EAD);
	ChangeMB2U(0xC82D, 0x006EAE);
	ChangeMB2U(0xC82E, 0x006EB1);
	ChangeMB2U(0xC82F, 0x006EB3);
	ChangeMB2U(0xC830, 0x006EBB);
	ChangeMB2U(0xC832, 0x006EC0);
	ChangeMB2U(0xC833, 0x006EC1);
	ChangeMB2U(0xC834, 0x006EC3);
	ChangeMB2U(0xC835, 0x006EC7);
	ChangeMB2U(0xC836, 0x006EC8);
	ChangeMB2U(0xC837, 0x006ECA);
	ChangeMB2U(0xC838, 0x006ECD);
	ChangeMB2U(0xC839, 0x006ECE);
	ChangeMB2U(0xC83A, 0x006ECF);
	ChangeMB2U(0xC83B, 0x006EEB);
	ChangeMB2U(0xC83C, 0x006EED);
	ChangeMB2U(0xC83D, 0x006EEE);
	ChangeMB2U(0xC83E, 0x006EF9);
	ChangeMB2U(0xC83F, 0x006EFB);
	ChangeMB2U(0xC840, 0x006EFD);
	ChangeMB2U(0xC841, 0x006F04);
	ChangeMB2U(0xC842, 0x006F08);
	ChangeMB2U(0xC843, 0x006F0A);
	ChangeMB2U(0xC844, 0x006F0C);
	ChangeMB2U(0xC845, 0x006F0D);
	ChangeMB2U(0xC846, 0x006F16);
	ChangeMB2U(0xC847, 0x006F18);
	ChangeMB2U(0xC848, 0x006F1A);
	ChangeMB2U(0xC849, 0x006F1B);
	ChangeMB2U(0xC84A, 0x006F26);
	ChangeMB2U(0xC84B, 0x006F29);
	ChangeMB2U(0xC84C, 0x006F2A);
	ChangeMB2U(0xC84D, 0x006F2F);
	ChangeMB2U(0xC84E, 0x006F30);
	ChangeMB2U(0xC84F, 0x006F33);
	ChangeMB2U(0xC850, 0x006F36);
	ChangeMB2U(0xC851, 0x006F3B);
	ChangeMB2U(0xC852, 0x006F3C);
	ChangeMB2U(0xC853, 0x006F2D);
	ChangeMB2U(0xC854, 0x006F4F);
	ChangeMB2U(0xC855, 0x006F51);
	ChangeMB2U(0xC856, 0x006F52);
	ChangeMB2U(0xC857, 0x006F53);
	ChangeMB2U(0xC858, 0x006F57);
	ChangeMB2U(0xC859, 0x006F59);
	ChangeMB2U(0xC85A, 0x006F5A);
	ChangeMB2U(0xC85B, 0x006F5D);
	ChangeMB2U(0xC85C, 0x006F5E);
	ChangeMB2U(0xC85D, 0x006F61);
	ChangeMB2U(0xC85E, 0x006F62);
	ChangeMB2U(0xC85F, 0x006F68);
	ChangeMB2U(0xC860, 0x006F6C);
	ChangeMB2U(0xC861, 0x006F7D);
	ChangeMB2U(0xC862, 0x006F7E);
	ChangeMB2U(0xC863, 0x006F83);
	ChangeMB2U(0xC864, 0x006F87);
	ChangeMB2U(0xC866, 0x006F8B);
	ChangeMB2U(0xC867, 0x006F8C);
	ChangeMB2U(0xC868, 0x006F8D);
	ChangeMB2U(0xC869, 0x006F90);
	ChangeMB2U(0xC86A, 0x006F92);
	ChangeMB2U(0xC86B, 0x006F93);
	ChangeMB2U(0xC86C, 0x006F94);
	ChangeMB2U(0xC86D, 0x006F96);
	ChangeMB2U(0xC86E, 0x006F9A);
	ChangeMB2U(0xC86F, 0x006F9F);
	ChangeMB2U(0xC870, 0x006FA0);
	ChangeMB2U(0xC871, 0x006FA5);
	ChangeMB2U(0xC872, 0x006FA6);
	ChangeMB2U(0xC873, 0x006FA7);
	ChangeMB2U(0xC874, 0x006FA8);
	ChangeMB2U(0xC875, 0x006FAE);
	ChangeMB2U(0xC876, 0x006FAF);
	ChangeMB2U(0xC877, 0x006FB0);
	ChangeMB2U(0xC879, 0x006FB6);
	ChangeMB2U(0xC87A, 0x006FBC);
	ChangeMB2U(0xC87B, 0x006FC5);
	ChangeMB2U(0xC87C, 0x006FC7);
	ChangeMB2U(0xC87D, 0x006FC8);
	ChangeMB2U(0xC87E, 0x006FCA);
	ChangeMB2U(0xC921, 0x006FDA);
	ChangeMB2U(0xC922, 0x006FDE);
	ChangeMB2U(0xC923, 0x006FE8);
	ChangeMB2U(0xC924, 0x006FE9);
	ChangeMB2U(0xC925, 0x006FF0);
	ChangeMB2U(0xC927, 0x006FF9);
	ChangeMB2U(0xC928, 0x006FFC);
	ChangeMB2U(0xC929, 0x006FFD);
	ChangeMB2U(0xC92A, 0x007000);
	ChangeMB2U(0xC92C, 0x007006);
	ChangeMB2U(0xC92E, 0x00700D);
	ChangeMB2U(0xC92F, 0x007017);
	ChangeMB2U(0xC930, 0x007020);
	ChangeMB2U(0xC931, 0x007023);
	ChangeMB2U(0xC932, 0x00702F);
	ChangeMB2U(0xC933, 0x007034);
	ChangeMB2U(0xC934, 0x007037);
	ChangeMB2U(0xC935, 0x007039);
	ChangeMB2U(0xC936, 0x00703C);
	ChangeMB2U(0xC937, 0x007043);
	ChangeMB2U(0xC938, 0x007044);
	ChangeMB2U(0xC939, 0x007048);
	ChangeMB2U(0xC93A, 0x007049);
	ChangeMB2U(0xC93B, 0x00704A);
	ChangeMB2U(0xC93C, 0x00704B);
	ChangeMB2U(0xC93D, 0x007054);
	ChangeMB2U(0xC93E, 0x007055);
	ChangeMB2U(0xC93F, 0x00705D);
	ChangeMB2U(0xC940, 0x00705E);
	ChangeMB2U(0xC941, 0x00704E);
	ChangeMB2U(0xC942, 0x007064);
	ChangeMB2U(0xC943, 0x007065);
	ChangeMB2U(0xC944, 0x00706C);
	ChangeMB2U(0xC945, 0x00706E);
	ChangeMB2U(0xC946, 0x007075);
	ChangeMB2U(0xC947, 0x007076);
	ChangeMB2U(0xC948, 0x00707E);
	ChangeMB2U(0xC949, 0x007081);
	ChangeMB2U(0xC94B, 0x007086);
	ChangeMB2U(0xC94C, 0x007094);
	ChangeMB2U(0xC94D, 0x007095);
	ChangeMB2U(0xC94E, 0x007096);
	ChangeMB2U(0xC94F, 0x007097);
	ChangeMB2U(0xC950, 0x007098);
	ChangeMB2U(0xC951, 0x00709B);
	ChangeMB2U(0xC952, 0x0070A4);
	ChangeMB2U(0xC954, 0x0070B0);
	ChangeMB2U(0xC955, 0x0070B1);
	ChangeMB2U(0xC956, 0x0070B4);
	ChangeMB2U(0xC957, 0x0070B7);
	ChangeMB2U(0xC958, 0x0070CA);
	ChangeMB2U(0xC959, 0x0070D1);
	ChangeMB2U(0xC95A, 0x0070D3);
	ChangeMB2U(0xC95B, 0x0070D4);
	ChangeMB2U(0xC95C, 0x0070D5);
	ChangeMB2U(0xC95D, 0x0070D6);
	ChangeMB2U(0xC95E, 0x0070D8);
	ChangeMB2U(0xC95F, 0x0070DC);
	ChangeMB2U(0xC960, 0x0070E4);
	ChangeMB2U(0xC961, 0x0070FA);
	ChangeMB2U(0xC962, 0x007103);
	ChangeMB2U(0xC964, 0x007105);
	ChangeMB2U(0xC965, 0x007106);
	ChangeMB2U(0xC966, 0x007107);
	ChangeMB2U(0xC967, 0x00710B);
	ChangeMB2U(0xC968, 0x00710C);
	ChangeMB2U(0xC96A, 0x00711E);
	ChangeMB2U(0xC96B, 0x007120);
	ChangeMB2U(0xC96C, 0x00712B);
	ChangeMB2U(0xC96D, 0x00712D);
	ChangeMB2U(0xC96E, 0x00712F);
	ChangeMB2U(0xC96F, 0x007130);
	ChangeMB2U(0xC970, 0x007131);
	ChangeMB2U(0xC971, 0x007138);
	ChangeMB2U(0xC972, 0x007141);
	ChangeMB2U(0xC973, 0x007145);
	ChangeMB2U(0xC976, 0x00714A);
	ChangeMB2U(0xC977, 0x00714B);
	ChangeMB2U(0xC978, 0x007150);
	ChangeMB2U(0xC979, 0x007152);
	ChangeMB2U(0xC97A, 0x007157);
	ChangeMB2U(0xC97B, 0x00715A);
	ChangeMB2U(0xC97D, 0x00715E);
	ChangeMB2U(0xC97E, 0x007160);
	ChangeMB2U(0xCA21, 0x007168);
	ChangeMB2U(0xCA22, 0x007179);
	ChangeMB2U(0xCA23, 0x007180);
	ChangeMB2U(0xCA24, 0x007185);
	ChangeMB2U(0xCA25, 0x007187);
	ChangeMB2U(0xCA26, 0x00718C);
	ChangeMB2U(0xCA27, 0x007192);
	ChangeMB2U(0xCA28, 0x00719A);
	ChangeMB2U(0xCA29, 0x00719B);
	ChangeMB2U(0xCA2A, 0x0071A0);
	ChangeMB2U(0xCA2B, 0x0071A2);
	ChangeMB2U(0xCA2C, 0x0071AF);
	ChangeMB2U(0xCA2D, 0x0071B0);
	ChangeMB2U(0xCA2E, 0x0071B2);
	ChangeMB2U(0xCA2F, 0x0071B3);
	ChangeMB2U(0xCA30, 0x0071BA);
	ChangeMB2U(0xCA31, 0x0071BF);
	ChangeMB2U(0xCA32, 0x0071C0);
	ChangeMB2U(0xCA34, 0x0071C4);
	ChangeMB2U(0xCA35, 0x0071CB);
	ChangeMB2U(0xCA36, 0x0071CC);
	ChangeMB2U(0xCA37, 0x0071D3);
	ChangeMB2U(0xCA38, 0x0071D6);
	ChangeMB2U(0xCA39, 0x0071D9);
	ChangeMB2U(0xCA3A, 0x0071DA);
	ChangeMB2U(0xCA3B, 0x0071DC);
	ChangeMB2U(0xCA3C, 0x0071F8);
	ChangeMB2U(0xCA3E, 0x007200);
	ChangeMB2U(0xCA3F, 0x007207);
	ChangeMB2U(0xCA40, 0x007208);
	ChangeMB2U(0xCA41, 0x007209);
	ChangeMB2U(0xCA42, 0x007213);
	ChangeMB2U(0xCA43, 0x007217);
	ChangeMB2U(0xCA44, 0x00721A);
	ChangeMB2U(0xCA45, 0x00721D);
	ChangeMB2U(0xCA46, 0x00721F);
	ChangeMB2U(0xCA47, 0x007224);
	ChangeMB2U(0xCA48, 0x00722B);
	ChangeMB2U(0xCA49, 0x00722F);
	ChangeMB2U(0xCA4A, 0x007234);
	ChangeMB2U(0xCA4B, 0x007238);
	ChangeMB2U(0xCA4C, 0x007239);
	ChangeMB2U(0xCA4D, 0x007241);
	ChangeMB2U(0xCA4E, 0x007242);
	ChangeMB2U(0xCA4F, 0x007243);
	ChangeMB2U(0xCA50, 0x007245);
	ChangeMB2U(0xCA51, 0x00724E);
	ChangeMB2U(0xCA52, 0x00724F);
	ChangeMB2U(0xCA53, 0x007250);
	ChangeMB2U(0xCA54, 0x007253);
	ChangeMB2U(0xCA55, 0x007255);
	ChangeMB2U(0xCA56, 0x007256);
	ChangeMB2U(0xCA57, 0x00725A);
	ChangeMB2U(0xCA58, 0x00725C);
	ChangeMB2U(0xCA59, 0x00725E);
	ChangeMB2U(0xCA5A, 0x007260);
	ChangeMB2U(0xCA5B, 0x007263);
	ChangeMB2U(0xCA5C, 0x007268);
	ChangeMB2U(0xCA5D, 0x00726B);
	ChangeMB2U(0xCA5E, 0x00726E);
	ChangeMB2U(0xCA5F, 0x00726F);
	ChangeMB2U(0xCA60, 0x007271);
	ChangeMB2U(0xCA61, 0x007277);
	ChangeMB2U(0xCA62, 0x007278);
	ChangeMB2U(0xCA63, 0x00727B);
	ChangeMB2U(0xCA64, 0x00727C);
	ChangeMB2U(0xCA65, 0x00727F);
	ChangeMB2U(0xCA66, 0x007284);
	ChangeMB2U(0xCA67, 0x007289);
	ChangeMB2U(0xCA68, 0x00728D);
	ChangeMB2U(0xCA69, 0x00728E);
	ChangeMB2U(0xCA6A, 0x007293);
	ChangeMB2U(0xCA6B, 0x00729B);
	ChangeMB2U(0xCA6C, 0x0072A8);
	ChangeMB2U(0xCA6D, 0x0072AD);
	ChangeMB2U(0xCA6E, 0x0072AE);
	ChangeMB2U(0xCA70, 0x0072B4);
	ChangeMB2U(0xCA72, 0x0072C1);
	ChangeMB2U(0xCA73, 0x0072C7);
	ChangeMB2U(0xCA74, 0x0072C9);
	ChangeMB2U(0xCA75, 0x0072CC);
	ChangeMB2U(0xCA76, 0x0072D5);
	ChangeMB2U(0xCA77, 0x0072D6);
	ChangeMB2U(0xCA78, 0x0072D8);
	ChangeMB2U(0xCA79, 0x0072DF);
	ChangeMB2U(0xCA7A, 0x0072E5);
	ChangeMB2U(0xCA7B, 0x0072F3);
	ChangeMB2U(0xCA7C, 0x0072F4);
	ChangeMB2U(0xCA7D, 0x0072FA);
	ChangeMB2U(0xCA7E, 0x0072FB);
	ChangeMB2U(0xCB21, 0x0072FE);
	ChangeMB2U(0xCB22, 0x007302);
	ChangeMB2U(0xCB23, 0x007304);
	ChangeMB2U(0xCB24, 0x007305);
	ChangeMB2U(0xCB25, 0x007307);
	ChangeMB2U(0xCB26, 0x00730B);
	ChangeMB2U(0xCB27, 0x00730D);
	ChangeMB2U(0xCB28, 0x007312);
	ChangeMB2U(0xCB29, 0x007313);
	ChangeMB2U(0xCB2A, 0x007318);
	ChangeMB2U(0xCB2B, 0x007319);
	ChangeMB2U(0xCB2C, 0x00731E);
	ChangeMB2U(0xCB2D, 0x007322);
	ChangeMB2U(0xCB2F, 0x007327);
	ChangeMB2U(0xCB30, 0x007328);
	ChangeMB2U(0xCB31, 0x00732C);
	ChangeMB2U(0xCB32, 0x007331);
	ChangeMB2U(0xCB33, 0x007332);
	ChangeMB2U(0xCB34, 0x007335);
	ChangeMB2U(0xCB35, 0x00733A);
	ChangeMB2U(0xCB36, 0x00733B);
	ChangeMB2U(0xCB37, 0x00733D);
	ChangeMB2U(0xCB38, 0x007343);
	ChangeMB2U(0xCB39, 0x00734D);
	ChangeMB2U(0xCB3A, 0x007350);
	ChangeMB2U(0xCB3B, 0x007352);
	ChangeMB2U(0xCB3C, 0x007356);
	ChangeMB2U(0xCB3D, 0x007358);
	ChangeMB2U(0xCB3E, 0x00735D);
	ChangeMB2U(0xCB3F, 0x00735E);
	ChangeMB2U(0xCB40, 0x00735F);
	ChangeMB2U(0xCB41, 0x007360);
	ChangeMB2U(0xCB42, 0x007366);
	ChangeMB2U(0xCB43, 0x007367);
	ChangeMB2U(0xCB44, 0x007369);
	ChangeMB2U(0xCB45, 0x00736B);
	ChangeMB2U(0xCB46, 0x00736C);
	ChangeMB2U(0xCB47, 0x00736E);
	ChangeMB2U(0xCB48, 0x00736F);
	ChangeMB2U(0xCB49, 0x007371);
	ChangeMB2U(0xCB4B, 0x007379);
	ChangeMB2U(0xCB4C, 0x00737C);
	ChangeMB2U(0xCB4D, 0x007380);
	ChangeMB2U(0xCB4E, 0x007381);
	ChangeMB2U(0xCB4F, 0x007383);
	ChangeMB2U(0xCB50, 0x007385);
	ChangeMB2U(0xCB51, 0x007386);
	ChangeMB2U(0xCB52, 0x00738E);
	ChangeMB2U(0xCB53, 0x007390);
	ChangeMB2U(0xCB54, 0x007393);
	ChangeMB2U(0xCB55, 0x007395);
	ChangeMB2U(0xCB56, 0x007397);
	ChangeMB2U(0xCB57, 0x007398);
	ChangeMB2U(0xCB58, 0x00739C);
	ChangeMB2U(0xCB59, 0x00739E);
	ChangeMB2U(0xCB5A, 0x00739F);
	ChangeMB2U(0xCB5B, 0x0073A0);
	ChangeMB2U(0xCB5C, 0x0073A2);
	ChangeMB2U(0xCB5D, 0x0073A5);
	ChangeMB2U(0xCB5E, 0x0073A6);
	ChangeMB2U(0xCB5F, 0x0073AA);
	ChangeMB2U(0xCB60, 0x0073AB);
	ChangeMB2U(0xCB61, 0x0073AD);
	ChangeMB2U(0xCB62, 0x0073B5);
	ChangeMB2U(0xCB63, 0x0073B7);
	ChangeMB2U(0xCB64, 0x0073B9);
	ChangeMB2U(0xCB65, 0x0073BC);
	ChangeMB2U(0xCB67, 0x0073BF);
	ChangeMB2U(0xCB68, 0x0073C5);
	ChangeMB2U(0xCB69, 0x0073C6);
	ChangeMB2U(0xCB6B, 0x0073CB);
	ChangeMB2U(0xCB6C, 0x0073CC);
	ChangeMB2U(0xCB6D, 0x0073CF);
	ChangeMB2U(0xCB6F, 0x0073D3);
	ChangeMB2U(0xCB71, 0x0073D9);
	ChangeMB2U(0xCB72, 0x0073DD);
	ChangeMB2U(0xCB73, 0x0073E1);
	ChangeMB2U(0xCB75, 0x0073E6);
	ChangeMB2U(0xCB76, 0x0073E7);
	ChangeMB2U(0xCB77, 0x0073E9);
	ChangeMB2U(0xCB78, 0x0073F4);
	ChangeMB2U(0xCB7A, 0x0073F7);
	ChangeMB2U(0xCB7B, 0x0073F9);
	ChangeMB2U(0xCB7C, 0x0073FA);
	ChangeMB2U(0xCB7D, 0x0073FB);
	ChangeMB2U(0xCB7E, 0x0073FD);
	ChangeMB2U(0xCC21, 0x0073FF);
	ChangeMB2U(0xCC22, 0x007400);
	ChangeMB2U(0xCC23, 0x007401);
	ChangeMB2U(0xCC24, 0x007404);
	ChangeMB2U(0xCC26, 0x00740A);
	ChangeMB2U(0xCC27, 0x007411);
	ChangeMB2U(0xCC28, 0x00741A);
	ChangeMB2U(0xCC29, 0x00741B);
	ChangeMB2U(0xCC2A, 0x007424);
	ChangeMB2U(0xCC2C, 0x007428);
	ChangeMB2U(0xCC2F, 0x00742B);
	ChangeMB2U(0xCC30, 0x00742C);
	ChangeMB2U(0xCC31, 0x00742D);
	ChangeMB2U(0xCC33, 0x00742F);
	ChangeMB2U(0xCC34, 0x007430);
	ChangeMB2U(0xCC35, 0x007431);
	ChangeMB2U(0xCC36, 0x007439);
	ChangeMB2U(0xCC37, 0x007440);
	ChangeMB2U(0xCC38, 0x007443);
	ChangeMB2U(0xCC39, 0x007444);
	ChangeMB2U(0xCC3A, 0x007446);
	ChangeMB2U(0xCC3B, 0x007447);
	ChangeMB2U(0xCC3C, 0x00744B);
	ChangeMB2U(0xCC3D, 0x00744D);
	ChangeMB2U(0xCC3E, 0x007451);
	ChangeMB2U(0xCC3F, 0x007452);
	ChangeMB2U(0xCC40, 0x007457);
	ChangeMB2U(0xCC41, 0x00745D);
	ChangeMB2U(0xCC43, 0x007466);
	ChangeMB2U(0xCC44, 0x007467);
	ChangeMB2U(0xCC45, 0x007468);
	ChangeMB2U(0xCC46, 0x00746B);
	ChangeMB2U(0xCC47, 0x00746D);
	ChangeMB2U(0xCC48, 0x00746E);
	ChangeMB2U(0xCC49, 0x007471);
	ChangeMB2U(0xCC4A, 0x007472);
	ChangeMB2U(0xCC4B, 0x007480);
	ChangeMB2U(0xCC4C, 0x007481);
	ChangeMB2U(0xCC4D, 0x007485);
	ChangeMB2U(0xCC4E, 0x007486);
	ChangeMB2U(0xCC4F, 0x007487);
	ChangeMB2U(0xCC51, 0x00748F);
	ChangeMB2U(0xCC52, 0x007490);
	ChangeMB2U(0xCC53, 0x007491);
	ChangeMB2U(0xCC54, 0x007492);
	ChangeMB2U(0xCC55, 0x007498);
	ChangeMB2U(0xCC56, 0x007499);
	ChangeMB2U(0xCC57, 0x00749A);
	ChangeMB2U(0xCC58, 0x00749C);
	ChangeMB2U(0xCC5A, 0x0074A0);
	ChangeMB2U(0xCC5B, 0x0074A1);
	ChangeMB2U(0xCC5C, 0x0074A3);
	ChangeMB2U(0xCC5D, 0x0074A6);
	ChangeMB2U(0xCC5E, 0x0074A8);
	ChangeMB2U(0xCC5F, 0x0074A9);
	ChangeMB2U(0xCC60, 0x0074AA);
	ChangeMB2U(0xCC61, 0x0074AB);
	ChangeMB2U(0xCC62, 0x0074AE);
	ChangeMB2U(0xCC63, 0x0074AF);
	ChangeMB2U(0xCC64, 0x0074B1);
	ChangeMB2U(0xCC65, 0x0074B2);
	ChangeMB2U(0xCC66, 0x0074B5);
	ChangeMB2U(0xCC67, 0x0074B9);
	ChangeMB2U(0xCC68, 0x0074BB);
	ChangeMB2U(0xCC69, 0x0074BF);
	ChangeMB2U(0xCC6A, 0x0074C8);
	ChangeMB2U(0xCC6B, 0x0074C9);
	ChangeMB2U(0xCC6C, 0x0074CC);
	ChangeMB2U(0xCC6D, 0x0074D0);
	ChangeMB2U(0xCC6E, 0x0074D3);
	ChangeMB2U(0xCC6F, 0x0074D8);
	ChangeMB2U(0xCC70, 0x0074DA);
	ChangeMB2U(0xCC71, 0x0074DB);
	ChangeMB2U(0xCC72, 0x0074DE);
	ChangeMB2U(0xCC73, 0x0074DF);
	ChangeMB2U(0xCC74, 0x0074E4);
	ChangeMB2U(0xCC75, 0x0074E8);
	ChangeMB2U(0xCC76, 0x0074EA);
	ChangeMB2U(0xCC77, 0x0074EB);
	ChangeMB2U(0xCC78, 0x0074EF);
	ChangeMB2U(0xCC79, 0x0074F4);
	ChangeMB2U(0xCC7A, 0x0074FA);
	ChangeMB2U(0xCC7B, 0x0074FB);
	ChangeMB2U(0xCC7C, 0x0074FC);
	ChangeMB2U(0xCC7D, 0x0074FF);
	ChangeMB2U(0xCC7E, 0x007506);
	ChangeMB2U(0xCD21, 0x007512);
	ChangeMB2U(0xCD22, 0x007516);
	ChangeMB2U(0xCD23, 0x007517);
	ChangeMB2U(0xCD24, 0x007520);
	ChangeMB2U(0xCD25, 0x007521);
	ChangeMB2U(0xCD26, 0x007524);
	ChangeMB2U(0xCD27, 0x007527);
	ChangeMB2U(0xCD28, 0x007529);
	ChangeMB2U(0xCD29, 0x00752A);
	ChangeMB2U(0xCD2B, 0x007536);
	ChangeMB2U(0xCD2C, 0x007539);
	ChangeMB2U(0xCD2D, 0x00753D);
	ChangeMB2U(0xCD2E, 0x00753E);
	ChangeMB2U(0xCD2F, 0x00753F);
	ChangeMB2U(0xCD30, 0x007540);
	ChangeMB2U(0xCD31, 0x007543);
	ChangeMB2U(0xCD32, 0x007547);
	ChangeMB2U(0xCD33, 0x007548);
	ChangeMB2U(0xCD34, 0x00754E);
	ChangeMB2U(0xCD35, 0x007550);
	ChangeMB2U(0xCD36, 0x007552);
	ChangeMB2U(0xCD37, 0x007557);
	ChangeMB2U(0xCD38, 0x00755E);
	ChangeMB2U(0xCD39, 0x00755F);
	ChangeMB2U(0xCD3A, 0x007561);
	ChangeMB2U(0xCD3C, 0x007571);
	ChangeMB2U(0xCD3D, 0x007579);
	ChangeMB2U(0xCD3E, 0x00757A);
	ChangeMB2U(0xCD3F, 0x00757B);
	ChangeMB2U(0xCD40, 0x00757C);
	ChangeMB2U(0xCD41, 0x00757D);
	ChangeMB2U(0xCD42, 0x00757E);
	ChangeMB2U(0xCD43, 0x007581);
	ChangeMB2U(0xCD44, 0x007585);
	ChangeMB2U(0xCD45, 0x007590);
	ChangeMB2U(0xCD46, 0x007592);
	ChangeMB2U(0xCD47, 0x007593);
	ChangeMB2U(0xCD48, 0x007595);
	ChangeMB2U(0xCD49, 0x007599);
	ChangeMB2U(0xCD4A, 0x00759C);
	ChangeMB2U(0xCD4B, 0x0075A2);
	ChangeMB2U(0xCD4C, 0x0075A4);
	ChangeMB2U(0xCD4D, 0x0075B4);
	ChangeMB2U(0xCD4E, 0x0075BA);
	ChangeMB2U(0xCD4F, 0x0075BF);
	ChangeMB2U(0xCD50, 0x0075C0);
	ChangeMB2U(0xCD51, 0x0075C1);
	ChangeMB2U(0xCD52, 0x0075C4);
	ChangeMB2U(0xCD53, 0x0075C6);
	ChangeMB2U(0xCD54, 0x0075CC);
	ChangeMB2U(0xCD55, 0x0075CE);
	ChangeMB2U(0xCD56, 0x0075CF);
	ChangeMB2U(0xCD57, 0x0075D7);
	ChangeMB2U(0xCD58, 0x0075DC);
	ChangeMB2U(0xCD59, 0x0075DF);
	ChangeMB2U(0xCD5A, 0x0075E0);
	ChangeMB2U(0xCD5B, 0x0075E1);
	ChangeMB2U(0xCD5C, 0x0075E4);
	ChangeMB2U(0xCD5D, 0x0075E7);
	ChangeMB2U(0xCD5E, 0x0075EC);
	ChangeMB2U(0xCD5F, 0x0075EE);
	ChangeMB2U(0xCD60, 0x0075EF);
	ChangeMB2U(0xCD61, 0x0075F1);
	ChangeMB2U(0xCD62, 0x0075F9);
	ChangeMB2U(0xCD63, 0x007600);
	ChangeMB2U(0xCD64, 0x007602);
	ChangeMB2U(0xCD65, 0x007603);
	ChangeMB2U(0xCD66, 0x007604);
	ChangeMB2U(0xCD67, 0x007607);
	ChangeMB2U(0xCD68, 0x007608);
	ChangeMB2U(0xCD69, 0x00760A);
	ChangeMB2U(0xCD6A, 0x00760C);
	ChangeMB2U(0xCD6B, 0x00760F);
	ChangeMB2U(0xCD6C, 0x007612);
	ChangeMB2U(0xCD6D, 0x007613);
	ChangeMB2U(0xCD6E, 0x007615);
	ChangeMB2U(0xCD6F, 0x007616);
	ChangeMB2U(0xCD70, 0x007619);
	ChangeMB2U(0xCD71, 0x00761B);
	ChangeMB2U(0xCD72, 0x00761C);
	ChangeMB2U(0xCD73, 0x00761D);
	ChangeMB2U(0xCD74, 0x00761E);
	ChangeMB2U(0xCD75, 0x007623);
	ChangeMB2U(0xCD76, 0x007625);
	ChangeMB2U(0xCD77, 0x007626);
	ChangeMB2U(0xCD78, 0x007629);
	ChangeMB2U(0xCD79, 0x00762D);
	ChangeMB2U(0xCD7A, 0x007632);
	ChangeMB2U(0xCD7B, 0x007633);
	ChangeMB2U(0xCD7C, 0x007635);
	ChangeMB2U(0xCD7D, 0x007638);
	ChangeMB2U(0xCD7E, 0x007639);
	ChangeMB2U(0xCE21, 0x00763A);
	ChangeMB2U(0xCE22, 0x00763C);
	ChangeMB2U(0xCE23, 0x00764A);
	ChangeMB2U(0xCE24, 0x007640);
	ChangeMB2U(0xCE25, 0x007641);
	ChangeMB2U(0xCE26, 0x007643);
	ChangeMB2U(0xCE27, 0x007644);
	ChangeMB2U(0xCE28, 0x007645);
	ChangeMB2U(0xCE29, 0x007649);
	ChangeMB2U(0xCE2A, 0x00764B);
	ChangeMB2U(0xCE2B, 0x007655);
	ChangeMB2U(0xCE2C, 0x007659);
	ChangeMB2U(0xCE2D, 0x00765F);
	ChangeMB2U(0xCE2E, 0x007664);
	ChangeMB2U(0xCE2F, 0x007665);
	ChangeMB2U(0xCE30, 0x00766D);
	ChangeMB2U(0xCE31, 0x00766E);
	ChangeMB2U(0xCE32, 0x00766F);
	ChangeMB2U(0xCE33, 0x007671);
	ChangeMB2U(0xCE34, 0x007674);
	ChangeMB2U(0xCE35, 0x007681);
	ChangeMB2U(0xCE36, 0x007685);
	ChangeMB2U(0xCE37, 0x00768C);
	ChangeMB2U(0xCE38, 0x00768D);
	ChangeMB2U(0xCE39, 0x007695);
	ChangeMB2U(0xCE3C, 0x00769D);
	ChangeMB2U(0xCE3D, 0x00769F);
	ChangeMB2U(0xCE3E, 0x0076A0);
	ChangeMB2U(0xCE3F, 0x0076A2);
	ChangeMB2U(0xCE40, 0x0076A3);
	ChangeMB2U(0xCE41, 0x0076A4);
	ChangeMB2U(0xCE42, 0x0076A5);
	ChangeMB2U(0xCE44, 0x0076A7);
	ChangeMB2U(0xCE45, 0x0076A8);
	ChangeMB2U(0xCE46, 0x0076AA);
	ChangeMB2U(0xCE47, 0x0076AD);
	ChangeMB2U(0xCE48, 0x0076BD);
	ChangeMB2U(0xCE49, 0x0076C1);
	ChangeMB2U(0xCE4A, 0x0076C5);
	ChangeMB2U(0xCE4B, 0x0076C9);
	ChangeMB2U(0xCE4C, 0x0076CB);
	ChangeMB2U(0xCE4D, 0x0076CC);
	ChangeMB2U(0xCE4E, 0x0076CE);
	ChangeMB2U(0xCE4F, 0x0076D4);
	ChangeMB2U(0xCE50, 0x0076D9);
	ChangeMB2U(0xCE51, 0x0076E0);
	ChangeMB2U(0xCE52, 0x0076E6);
	ChangeMB2U(0xCE53, 0x0076E8);
	ChangeMB2U(0xCE54, 0x0076EC);
	ChangeMB2U(0xCE55, 0x0076F0);
	ChangeMB2U(0xCE56, 0x0076F1);
	ChangeMB2U(0xCE57, 0x0076F6);
	ChangeMB2U(0xCE58, 0x0076F9);
	ChangeMB2U(0xCE59, 0x0076FC);
	ChangeMB2U(0xCE5A, 0x007700);
	ChangeMB2U(0xCE5B, 0x007706);
	ChangeMB2U(0xCE5C, 0x00770A);
	ChangeMB2U(0xCE5D, 0x00770E);
	ChangeMB2U(0xCE5E, 0x007712);
	ChangeMB2U(0xCE5F, 0x007714);
	ChangeMB2U(0xCE60, 0x007715);
	ChangeMB2U(0xCE61, 0x007717);
	ChangeMB2U(0xCE62, 0x007719);
	ChangeMB2U(0xCE63, 0x00771A);
	ChangeMB2U(0xCE64, 0x00771C);
	ChangeMB2U(0xCE65, 0x007722);
	ChangeMB2U(0xCE66, 0x007728);
	ChangeMB2U(0xCE67, 0x00772D);
	ChangeMB2U(0xCE68, 0x00772E);
	ChangeMB2U(0xCE69, 0x00772F);
	ChangeMB2U(0xCE6A, 0x007734);
	ChangeMB2U(0xCE6B, 0x007735);
	ChangeMB2U(0xCE6C, 0x007736);
	ChangeMB2U(0xCE6D, 0x007739);
	ChangeMB2U(0xCE6E, 0x00773D);
	ChangeMB2U(0xCE6F, 0x00773E);
	ChangeMB2U(0xCE70, 0x007742);
	ChangeMB2U(0xCE71, 0x007745);
	ChangeMB2U(0xCE73, 0x00774A);
	ChangeMB2U(0xCE74, 0x00774D);
	ChangeMB2U(0xCE75, 0x00774E);
	ChangeMB2U(0xCE76, 0x00774F);
	ChangeMB2U(0xCE77, 0x007752);
	ChangeMB2U(0xCE78, 0x007756);
	ChangeMB2U(0xCE79, 0x007757);
	ChangeMB2U(0xCE7A, 0x00775C);
	ChangeMB2U(0xCE7B, 0x00775E);
	ChangeMB2U(0xCE7C, 0x00775F);
	ChangeMB2U(0xCE7D, 0x007760);
	ChangeMB2U(0xCE7E, 0x007762);
	ChangeMB2U(0xCF21, 0x007764);
	ChangeMB2U(0xCF22, 0x007767);
	ChangeMB2U(0xCF23, 0x00776A);
	ChangeMB2U(0xCF24, 0x00776C);
	ChangeMB2U(0xCF25, 0x007770);
	ChangeMB2U(0xCF26, 0x007772);
	ChangeMB2U(0xCF27, 0x007773);
	ChangeMB2U(0xCF28, 0x007774);
	ChangeMB2U(0xCF29, 0x00777A);
	ChangeMB2U(0xCF2A, 0x00777D);
	ChangeMB2U(0xCF2B, 0x007780);
	ChangeMB2U(0xCF2C, 0x007784);
	ChangeMB2U(0xCF2D, 0x00778C);
	ChangeMB2U(0xCF2E, 0x00778D);
	ChangeMB2U(0xCF2F, 0x007794);
	ChangeMB2U(0xCF30, 0x007795);
	ChangeMB2U(0xCF31, 0x007796);
	ChangeMB2U(0xCF32, 0x00779A);
	ChangeMB2U(0xCF33, 0x00779F);
	ChangeMB2U(0xCF34, 0x0077A2);
	ChangeMB2U(0xCF35, 0x0077A7);
	ChangeMB2U(0xCF36, 0x0077AA);
	ChangeMB2U(0xCF37, 0x0077AE);
	ChangeMB2U(0xCF38, 0x0077AF);
	ChangeMB2U(0xCF39, 0x0077B1);
	ChangeMB2U(0xCF3A, 0x0077B5);
	ChangeMB2U(0xCF3B, 0x0077BE);
	ChangeMB2U(0xCF3C, 0x0077C3);
	ChangeMB2U(0xCF3D, 0x0077C9);
	ChangeMB2U(0xCF3E, 0x0077D1);
	ChangeMB2U(0xCF3F, 0x0077D2);
	ChangeMB2U(0xCF40, 0x0077D5);
	ChangeMB2U(0xCF41, 0x0077D9);
	ChangeMB2U(0xCF42, 0x0077DE);
	ChangeMB2U(0xCF43, 0x0077DF);
	ChangeMB2U(0xCF44, 0x0077E0);
	ChangeMB2U(0xCF45, 0x0077E4);
	ChangeMB2U(0xCF46, 0x0077E6);
	ChangeMB2U(0xCF47, 0x0077EA);
	ChangeMB2U(0xCF48, 0x0077EC);
	ChangeMB2U(0xCF49, 0x0077F0);
	ChangeMB2U(0xCF4A, 0x0077F1);
	ChangeMB2U(0xCF4B, 0x0077F4);
	ChangeMB2U(0xCF4C, 0x0077F8);
	ChangeMB2U(0xCF4D, 0x0077FB);
	ChangeMB2U(0xCF4E, 0x007805);
	ChangeMB2U(0xCF4F, 0x007806);
	ChangeMB2U(0xCF50, 0x007809);
	ChangeMB2U(0xCF51, 0x00780D);
	ChangeMB2U(0xCF52, 0x00780E);
	ChangeMB2U(0xCF53, 0x007811);
	ChangeMB2U(0xCF54, 0x00781D);
	ChangeMB2U(0xCF56, 0x007822);
	ChangeMB2U(0xCF57, 0x007823);
	ChangeMB2U(0xCF58, 0x00782D);
	ChangeMB2U(0xCF59, 0x00782E);
	ChangeMB2U(0xCF5A, 0x007830);
	ChangeMB2U(0xCF5B, 0x007835);
	ChangeMB2U(0xCF5C, 0x007837);
	ChangeMB2U(0xCF5D, 0x007843);
	ChangeMB2U(0xCF5E, 0x007844);
	ChangeMB2U(0xCF5F, 0x007847);
	ChangeMB2U(0xCF60, 0x007848);
	ChangeMB2U(0xCF61, 0x00784C);
	ChangeMB2U(0xCF63, 0x007852);
	ChangeMB2U(0xCF64, 0x00785C);
	ChangeMB2U(0xCF65, 0x00785E);
	ChangeMB2U(0xCF66, 0x007860);
	ChangeMB2U(0xCF67, 0x007861);
	ChangeMB2U(0xCF68, 0x007863);
	ChangeMB2U(0xCF6A, 0x007868);
	ChangeMB2U(0xCF6B, 0x00786A);
	ChangeMB2U(0xCF6C, 0x00786E);
	ChangeMB2U(0xCF6E, 0x00787E);
	ChangeMB2U(0xCF6F, 0x00788A);
	ChangeMB2U(0xCF70, 0x00788F);
	ChangeMB2U(0xCF71, 0x007894);
	ChangeMB2U(0xCF72, 0x007898);
	ChangeMB2U(0xCF73, 0x0078A1);
	ChangeMB2U(0xCF74, 0x00789D);
	ChangeMB2U(0xCF75, 0x00789E);
	ChangeMB2U(0xCF76, 0x00789F);
	ChangeMB2U(0xCF77, 0x0078A4);
	ChangeMB2U(0xCF78, 0x0078A8);
	ChangeMB2U(0xCF79, 0x0078AC);
	ChangeMB2U(0xCF7A, 0x0078AD);
	ChangeMB2U(0xCF7B, 0x0078B0);
	ChangeMB2U(0xCF7C, 0x0078B1);
	ChangeMB2U(0xCF7D, 0x0078B2);
	ChangeMB2U(0xCF7E, 0x0078B3);
	ChangeMB2U(0xD021, 0x0078BB);
	ChangeMB2U(0xD022, 0x0078BD);
	ChangeMB2U(0xD023, 0x0078BF);
	ChangeMB2U(0xD024, 0x0078C7);
	ChangeMB2U(0xD025, 0x0078C8);
	ChangeMB2U(0xD026, 0x0078C9);
	ChangeMB2U(0xD027, 0x0078CC);
	ChangeMB2U(0xD028, 0x0078CE);
	ChangeMB2U(0xD029, 0x0078D2);
	ChangeMB2U(0xD02A, 0x0078D3);
	ChangeMB2U(0xD02B, 0x0078D5);
	ChangeMB2U(0xD02C, 0x0078D6);
	ChangeMB2U(0xD02D, 0x0078E4);
	ChangeMB2U(0xD02E, 0x0078DB);
	ChangeMB2U(0xD02F, 0x0078DF);
	ChangeMB2U(0xD030, 0x0078E0);
	ChangeMB2U(0xD031, 0x0078E1);
	ChangeMB2U(0xD032, 0x0078E6);
	ChangeMB2U(0xD033, 0x0078EA);
	ChangeMB2U(0xD034, 0x0078F2);
	ChangeMB2U(0xD035, 0x0078F3);
	ChangeMB2U(0xD036, 0x007900);
	ChangeMB2U(0xD037, 0x0078F6);
	ChangeMB2U(0xD038, 0x0078F7);
	ChangeMB2U(0xD039, 0x0078FA);
	ChangeMB2U(0xD03A, 0x0078FB);
	ChangeMB2U(0xD03B, 0x0078FF);
	ChangeMB2U(0xD03C, 0x007906);
	ChangeMB2U(0xD03D, 0x00790C);
	ChangeMB2U(0xD03E, 0x007910);
	ChangeMB2U(0xD03F, 0x00791A);
	ChangeMB2U(0xD040, 0x00791C);
	ChangeMB2U(0xD041, 0x00791E);
	ChangeMB2U(0xD042, 0x00791F);
	ChangeMB2U(0xD043, 0x007920);
	ChangeMB2U(0xD044, 0x007925);
	ChangeMB2U(0xD045, 0x007927);
	ChangeMB2U(0xD046, 0x007929);
	ChangeMB2U(0xD047, 0x00792D);
	ChangeMB2U(0xD048, 0x007931);
	ChangeMB2U(0xD049, 0x007934);
	ChangeMB2U(0xD04A, 0x007935);
	ChangeMB2U(0xD04B, 0x00793B);
	ChangeMB2U(0xD04C, 0x00793D);
	ChangeMB2U(0xD04D, 0x00793F);
	ChangeMB2U(0xD04E, 0x007944);
	ChangeMB2U(0xD04F, 0x007945);
	ChangeMB2U(0xD050, 0x007946);
	ChangeMB2U(0xD051, 0x00794A);
	ChangeMB2U(0xD052, 0x00794B);
	ChangeMB2U(0xD053, 0x00794F);
	ChangeMB2U(0xD054, 0x007951);
	ChangeMB2U(0xD055, 0x007954);
	ChangeMB2U(0xD056, 0x007958);
	ChangeMB2U(0xD057, 0x00795B);
	ChangeMB2U(0xD058, 0x00795C);
	ChangeMB2U(0xD059, 0x007967);
	ChangeMB2U(0xD05A, 0x007969);
	ChangeMB2U(0xD05B, 0x00796B);
	ChangeMB2U(0xD05C, 0x007972);
	ChangeMB2U(0xD05D, 0x007979);
	ChangeMB2U(0xD05E, 0x00797B);
	ChangeMB2U(0xD05F, 0x00797C);
	ChangeMB2U(0xD060, 0x00797E);
	ChangeMB2U(0xD061, 0x00798B);
	ChangeMB2U(0xD062, 0x00798C);
	ChangeMB2U(0xD063, 0x007991);
	ChangeMB2U(0xD064, 0x007993);
	ChangeMB2U(0xD066, 0x007995);
	ChangeMB2U(0xD067, 0x007996);
	ChangeMB2U(0xD068, 0x007998);
	ChangeMB2U(0xD06A, 0x00799C);
	ChangeMB2U(0xD06B, 0x0079A1);
	ChangeMB2U(0xD06C, 0x0079A8);
	ChangeMB2U(0xD06D, 0x0079A9);
	ChangeMB2U(0xD06E, 0x0079AB);
	ChangeMB2U(0xD06F, 0x0079AF);
	ChangeMB2U(0xD070, 0x0079B1);
	ChangeMB2U(0xD071, 0x0079B4);
	ChangeMB2U(0xD072, 0x0079B8);
	ChangeMB2U(0xD073, 0x0079BB);
	ChangeMB2U(0xD074, 0x0079C2);
	ChangeMB2U(0xD075, 0x0079C4);
	ChangeMB2U(0xD076, 0x0079C7);
	ChangeMB2U(0xD077, 0x0079C8);
	ChangeMB2U(0xD078, 0x0079CA);
	ChangeMB2U(0xD079, 0x0079CF);
	ChangeMB2U(0xD07A, 0x0079D4);
	ChangeMB2U(0xD07B, 0x0079D6);
	ChangeMB2U(0xD07C, 0x0079DA);
	ChangeMB2U(0xD07D, 0x0079DD);
	ChangeMB2U(0xD07E, 0x0079DE);
	ChangeMB2U(0xD121, 0x0079E0);
	ChangeMB2U(0xD122, 0x0079E2);
	ChangeMB2U(0xD123, 0x0079E5);
	ChangeMB2U(0xD124, 0x0079EA);
	ChangeMB2U(0xD125, 0x0079EB);
	ChangeMB2U(0xD126, 0x0079ED);
	ChangeMB2U(0xD127, 0x0079F1);
	ChangeMB2U(0xD128, 0x0079F8);
	ChangeMB2U(0xD129, 0x0079FC);
	ChangeMB2U(0xD12A, 0x007A02);
	ChangeMB2U(0xD12B, 0x007A03);
	ChangeMB2U(0xD12C, 0x007A07);
	ChangeMB2U(0xD12D, 0x007A09);
	ChangeMB2U(0xD12E, 0x007A0A);
	ChangeMB2U(0xD12F, 0x007A0C);
	ChangeMB2U(0xD130, 0x007A11);
	ChangeMB2U(0xD131, 0x007A15);
	ChangeMB2U(0xD132, 0x007A1B);
	ChangeMB2U(0xD133, 0x007A1E);
	ChangeMB2U(0xD134, 0x007A21);
	ChangeMB2U(0xD135, 0x007A27);
	ChangeMB2U(0xD136, 0x007A2B);
	ChangeMB2U(0xD137, 0x007A2D);
	ChangeMB2U(0xD138, 0x007A2F);
	ChangeMB2U(0xD139, 0x007A30);
	ChangeMB2U(0xD13A, 0x007A34);
	ChangeMB2U(0xD13B, 0x007A35);
	ChangeMB2U(0xD13C, 0x007A38);
	ChangeMB2U(0xD13D, 0x007A39);
	ChangeMB2U(0xD13E, 0x007A3A);
	ChangeMB2U(0xD13F, 0x007A44);
	ChangeMB2U(0xD140, 0x007A45);
	ChangeMB2U(0xD141, 0x007A47);
	ChangeMB2U(0xD142, 0x007A48);
	ChangeMB2U(0xD143, 0x007A4C);
	ChangeMB2U(0xD144, 0x007A55);
	ChangeMB2U(0xD145, 0x007A56);
	ChangeMB2U(0xD146, 0x007A59);
	ChangeMB2U(0xD147, 0x007A5C);
	ChangeMB2U(0xD148, 0x007A5D);
	ChangeMB2U(0xD149, 0x007A5F);
	ChangeMB2U(0xD14A, 0x007A60);
	ChangeMB2U(0xD14B, 0x007A65);
	ChangeMB2U(0xD14C, 0x007A67);
	ChangeMB2U(0xD14D, 0x007A6A);
	ChangeMB2U(0xD14E, 0x007A6D);
	ChangeMB2U(0xD14F, 0x007A75);
	ChangeMB2U(0xD150, 0x007A78);
	ChangeMB2U(0xD151, 0x007A7E);
	ChangeMB2U(0xD152, 0x007A80);
	ChangeMB2U(0xD153, 0x007A82);
	ChangeMB2U(0xD154, 0x007A85);
	ChangeMB2U(0xD155, 0x007A86);
	ChangeMB2U(0xD156, 0x007A8A);
	ChangeMB2U(0xD157, 0x007A8B);
	ChangeMB2U(0xD158, 0x007A90);
	ChangeMB2U(0xD159, 0x007A91);
	ChangeMB2U(0xD15A, 0x007A94);
	ChangeMB2U(0xD15B, 0x007A9E);
	ChangeMB2U(0xD15C, 0x007AA0);
	ChangeMB2U(0xD15D, 0x007AA3);
	ChangeMB2U(0xD15E, 0x007AAC);
	ChangeMB2U(0xD15F, 0x007AB3);
	ChangeMB2U(0xD160, 0x007AB5);
	ChangeMB2U(0xD161, 0x007AB9);
	ChangeMB2U(0xD162, 0x007ABB);
	ChangeMB2U(0xD163, 0x007ABC);
	ChangeMB2U(0xD164, 0x007AC6);
	ChangeMB2U(0xD165, 0x007AC9);
	ChangeMB2U(0xD166, 0x007ACC);
	ChangeMB2U(0xD167, 0x007ACE);
	ChangeMB2U(0xD169, 0x007ADB);
	ChangeMB2U(0xD16A, 0x007AE8);
	ChangeMB2U(0xD16B, 0x007AE9);
	ChangeMB2U(0xD16D, 0x007AEC);
	ChangeMB2U(0xD16E, 0x007AF1);
	ChangeMB2U(0xD16F, 0x007AF4);
	ChangeMB2U(0xD170, 0x007AFB);
	ChangeMB2U(0xD171, 0x007AFD);
	ChangeMB2U(0xD172, 0x007AFE);
	ChangeMB2U(0xD173, 0x007B07);
	ChangeMB2U(0xD174, 0x007B14);
	ChangeMB2U(0xD175, 0x007B1F);
	ChangeMB2U(0xD176, 0x007B23);
	ChangeMB2U(0xD177, 0x007B27);
	ChangeMB2U(0xD178, 0x007B29);
	ChangeMB2U(0xD179, 0x007B2A);
	ChangeMB2U(0xD17A, 0x007B2B);
	ChangeMB2U(0xD17B, 0x007B2D);
	ChangeMB2U(0xD17C, 0x007B2E);
	ChangeMB2U(0xD17D, 0x007B2F);
	ChangeMB2U(0xD17E, 0x007B30);
	ChangeMB2U(0xD221, 0x007B31);
	ChangeMB2U(0xD222, 0x007B34);
	ChangeMB2U(0xD223, 0x007B3D);
	ChangeMB2U(0xD224, 0x007B3F);
	ChangeMB2U(0xD225, 0x007B40);
	ChangeMB2U(0xD226, 0x007B41);
	ChangeMB2U(0xD227, 0x007B47);
	ChangeMB2U(0xD228, 0x007B4E);
	ChangeMB2U(0xD229, 0x007B55);
	ChangeMB2U(0xD22A, 0x007B60);
	ChangeMB2U(0xD22B, 0x007B64);
	ChangeMB2U(0xD22C, 0x007B66);
	ChangeMB2U(0xD22D, 0x007B69);
	ChangeMB2U(0xD22E, 0x007B6A);
	ChangeMB2U(0xD22F, 0x007B6D);
	ChangeMB2U(0xD230, 0x007B6F);
	ChangeMB2U(0xD231, 0x007B72);
	ChangeMB2U(0xD232, 0x007B73);
	ChangeMB2U(0xD233, 0x007B77);
	ChangeMB2U(0xD234, 0x007B84);
	ChangeMB2U(0xD235, 0x007B89);
	ChangeMB2U(0xD236, 0x007B8E);
	ChangeMB2U(0xD237, 0x007B90);
	ChangeMB2U(0xD238, 0x007B91);
	ChangeMB2U(0xD239, 0x007B96);
	ChangeMB2U(0xD23A, 0x007B9B);
	ChangeMB2U(0xD23C, 0x007BA0);
	ChangeMB2U(0xD23D, 0x007BA5);
	ChangeMB2U(0xD23E, 0x007BAC);
	ChangeMB2U(0xD23F, 0x007BAF);
	ChangeMB2U(0xD240, 0x007BB0);
	ChangeMB2U(0xD241, 0x007BB2);
	ChangeMB2U(0xD242, 0x007BB5);
	ChangeMB2U(0xD243, 0x007BB6);
	ChangeMB2U(0xD244, 0x007BBA);
	ChangeMB2U(0xD245, 0x007BBB);
	ChangeMB2U(0xD246, 0x007BBC);
	ChangeMB2U(0xD247, 0x007BBD);
	ChangeMB2U(0xD248, 0x007BC2);
	ChangeMB2U(0xD249, 0x007BC5);
	ChangeMB2U(0xD24A, 0x007BC8);
	ChangeMB2U(0xD24B, 0x007BCA);
	ChangeMB2U(0xD24C, 0x007BD4);
	ChangeMB2U(0xD24D, 0x007BD6);
	ChangeMB2U(0xD24E, 0x007BD7);
	ChangeMB2U(0xD24F, 0x007BD9);
	ChangeMB2U(0xD250, 0x007BDA);
	ChangeMB2U(0xD251, 0x007BDB);
	ChangeMB2U(0xD252, 0x007BE8);
	ChangeMB2U(0xD253, 0x007BEA);
	ChangeMB2U(0xD254, 0x007BF2);
	ChangeMB2U(0xD255, 0x007BF4);
	ChangeMB2U(0xD256, 0x007BF5);
	ChangeMB2U(0xD257, 0x007BF8);
	ChangeMB2U(0xD258, 0x007BF9);
	ChangeMB2U(0xD259, 0x007BFA);
	ChangeMB2U(0xD25A, 0x007BFC);
	ChangeMB2U(0xD25B, 0x007BFE);
	ChangeMB2U(0xD25C, 0x007C01);
	ChangeMB2U(0xD25D, 0x007C02);
	ChangeMB2U(0xD25E, 0x007C03);
	ChangeMB2U(0xD25F, 0x007C04);
	ChangeMB2U(0xD260, 0x007C06);
	ChangeMB2U(0xD261, 0x007C09);
	ChangeMB2U(0xD262, 0x007C0B);
	ChangeMB2U(0xD263, 0x007C0C);
	ChangeMB2U(0xD264, 0x007C0E);
	ChangeMB2U(0xD265, 0x007C0F);
	ChangeMB2U(0xD266, 0x007C19);
	ChangeMB2U(0xD267, 0x007C1B);
	ChangeMB2U(0xD268, 0x007C20);
	ChangeMB2U(0xD269, 0x007C25);
	ChangeMB2U(0xD26A, 0x007C26);
	ChangeMB2U(0xD26B, 0x007C28);
	ChangeMB2U(0xD26C, 0x007C2C);
	ChangeMB2U(0xD26D, 0x007C31);
	ChangeMB2U(0xD26E, 0x007C33);
	ChangeMB2U(0xD26F, 0x007C34);
	ChangeMB2U(0xD270, 0x007C36);
	ChangeMB2U(0xD271, 0x007C39);
	ChangeMB2U(0xD272, 0x007C3A);
	ChangeMB2U(0xD273, 0x007C46);
	ChangeMB2U(0xD274, 0x007C4A);
	ChangeMB2U(0xD275, 0x007C55);
	ChangeMB2U(0xD276, 0x007C51);
	ChangeMB2U(0xD277, 0x007C52);
	ChangeMB2U(0xD278, 0x007C53);
	ChangeMB2U(0xD279, 0x007C59);
	ChangeMB2U(0xD27A, 0x007C5A);
	ChangeMB2U(0xD27B, 0x007C5B);
	ChangeMB2U(0xD27C, 0x007C5C);
	ChangeMB2U(0xD27D, 0x007C5D);
	ChangeMB2U(0xD27E, 0x007C5E);
	ChangeMB2U(0xD321, 0x007C61);
	ChangeMB2U(0xD322, 0x007C63);
	ChangeMB2U(0xD323, 0x007C67);
	ChangeMB2U(0xD324, 0x007C69);
	ChangeMB2U(0xD325, 0x007C6D);
	ChangeMB2U(0xD326, 0x007C6E);
	ChangeMB2U(0xD327, 0x007C70);
	ChangeMB2U(0xD328, 0x007C72);
	ChangeMB2U(0xD329, 0x007C79);
	ChangeMB2U(0xD32A, 0x007C7C);
	ChangeMB2U(0xD32B, 0x007C7D);
	ChangeMB2U(0xD32C, 0x007C86);
	ChangeMB2U(0xD32D, 0x007C87);
	ChangeMB2U(0xD32E, 0x007C8F);
	ChangeMB2U(0xD32F, 0x007C94);
	ChangeMB2U(0xD330, 0x007C9E);
	ChangeMB2U(0xD331, 0x007CA0);
	ChangeMB2U(0xD332, 0x007CA6);
	ChangeMB2U(0xD333, 0x007CB0);
	ChangeMB2U(0xD334, 0x007CB6);
	ChangeMB2U(0xD335, 0x007CB7);
	ChangeMB2U(0xD336, 0x007CBA);
	ChangeMB2U(0xD337, 0x007CBB);
	ChangeMB2U(0xD338, 0x007CBC);
	ChangeMB2U(0xD339, 0x007CBF);
	ChangeMB2U(0xD33A, 0x007CC4);
	ChangeMB2U(0xD33B, 0x007CC7);
	ChangeMB2U(0xD33C, 0x007CC8);
	ChangeMB2U(0xD33D, 0x007CC9);
	ChangeMB2U(0xD33E, 0x007CCD);
	ChangeMB2U(0xD33F, 0x007CCF);
	ChangeMB2U(0xD340, 0x007CD3);
	ChangeMB2U(0xD341, 0x007CD4);
	ChangeMB2U(0xD342, 0x007CD5);
	ChangeMB2U(0xD343, 0x007CD7);
	ChangeMB2U(0xD344, 0x007CD9);
	ChangeMB2U(0xD345, 0x007CDA);
	ChangeMB2U(0xD346, 0x007CDD);
	ChangeMB2U(0xD347, 0x007CE6);
	ChangeMB2U(0xD348, 0x007CE9);
	ChangeMB2U(0xD349, 0x007CEB);
	ChangeMB2U(0xD34A, 0x007CF5);
	ChangeMB2U(0xD34B, 0x007D03);
	ChangeMB2U(0xD34C, 0x007D07);
	ChangeMB2U(0xD34D, 0x007D08);
	ChangeMB2U(0xD34E, 0x007D09);
	ChangeMB2U(0xD34F, 0x007D0F);
	ChangeMB2U(0xD350, 0x007D11);
	ChangeMB2U(0xD351, 0x007D12);
	ChangeMB2U(0xD352, 0x007D13);
	ChangeMB2U(0xD353, 0x007D16);
	ChangeMB2U(0xD354, 0x007D1D);
	ChangeMB2U(0xD355, 0x007D1E);
	ChangeMB2U(0xD356, 0x007D23);
	ChangeMB2U(0xD357, 0x007D26);
	ChangeMB2U(0xD358, 0x007D2A);
	ChangeMB2U(0xD359, 0x007D2D);
	ChangeMB2U(0xD35A, 0x007D31);
	ChangeMB2U(0xD35B, 0x007D3C);
	ChangeMB2U(0xD35C, 0x007D3D);
	ChangeMB2U(0xD35D, 0x007D3E);
	ChangeMB2U(0xD35E, 0x007D40);
	ChangeMB2U(0xD35F, 0x007D41);
	ChangeMB2U(0xD360, 0x007D47);
	ChangeMB2U(0xD362, 0x007D4D);
	ChangeMB2U(0xD363, 0x007D51);
	ChangeMB2U(0xD364, 0x007D53);
	ChangeMB2U(0xD365, 0x007D57);
	ChangeMB2U(0xD366, 0x007D59);
	ChangeMB2U(0xD367, 0x007D5A);
	ChangeMB2U(0xD369, 0x007D5D);
	ChangeMB2U(0xD36A, 0x007D65);
	ChangeMB2U(0xD36B, 0x007D67);
	ChangeMB2U(0xD36C, 0x007D6A);
	ChangeMB2U(0xD36D, 0x007D70);
	ChangeMB2U(0xD36E, 0x007D78);
	ChangeMB2U(0xD36F, 0x007D7A);
	ChangeMB2U(0xD370, 0x007D7B);
	ChangeMB2U(0xD371, 0x007D7F);
	ChangeMB2U(0xD372, 0x007D81);
	ChangeMB2U(0xD373, 0x007D82);
	ChangeMB2U(0xD374, 0x007D83);
	ChangeMB2U(0xD375, 0x007D85);
	ChangeMB2U(0xD376, 0x007D86);
	ChangeMB2U(0xD377, 0x007D88);
	ChangeMB2U(0xD378, 0x007D8B);
	ChangeMB2U(0xD379, 0x007D8C);
	ChangeMB2U(0xD37A, 0x007D8D);
	ChangeMB2U(0xD37B, 0x007D91);
	ChangeMB2U(0xD37C, 0x007D96);
	ChangeMB2U(0xD37D, 0x007D97);
	ChangeMB2U(0xD37E, 0x007D9D);
	ChangeMB2U(0xD421, 0x007D9E);
	ChangeMB2U(0xD422, 0x007DA6);
	ChangeMB2U(0xD423, 0x007DA7);
	ChangeMB2U(0xD424, 0x007DAA);
	ChangeMB2U(0xD425, 0x007DB3);
	ChangeMB2U(0xD426, 0x007DB6);
	ChangeMB2U(0xD428, 0x007DB9);
	ChangeMB2U(0xD429, 0x007DC2);
	ChangeMB2U(0xD42A, 0x007DC3);
	ChangeMB2U(0xD42B, 0x007DC4);
	ChangeMB2U(0xD42C, 0x007DC5);
	ChangeMB2U(0xD42D, 0x007DC6);
	ChangeMB2U(0xD42E, 0x007DCC);
	ChangeMB2U(0xD42F, 0x007DCD);
	ChangeMB2U(0xD430, 0x007DCE);
	ChangeMB2U(0xD431, 0x007DD7);
	ChangeMB2U(0xD432, 0x007DD9);
	ChangeMB2U(0xD433, 0x007E00);
	ChangeMB2U(0xD434, 0x007DE2);
	ChangeMB2U(0xD435, 0x007DE5);
	ChangeMB2U(0xD436, 0x007DE6);
	ChangeMB2U(0xD437, 0x007DEA);
	ChangeMB2U(0xD438, 0x007DEB);
	ChangeMB2U(0xD439, 0x007DED);
	ChangeMB2U(0xD43A, 0x007DF1);
	ChangeMB2U(0xD43B, 0x007DF5);
	ChangeMB2U(0xD43C, 0x007DF6);
	ChangeMB2U(0xD43D, 0x007DF9);
	ChangeMB2U(0xD43E, 0x007DFA);
	ChangeMB2U(0xD43F, 0x007E08);
	ChangeMB2U(0xD440, 0x007E10);
	ChangeMB2U(0xD441, 0x007E11);
	ChangeMB2U(0xD442, 0x007E15);
	ChangeMB2U(0xD443, 0x007E17);
	ChangeMB2U(0xD444, 0x007E1C);
	ChangeMB2U(0xD445, 0x007E1D);
	ChangeMB2U(0xD446, 0x007E20);
	ChangeMB2U(0xD447, 0x007E27);
	ChangeMB2U(0xD448, 0x007E28);
	ChangeMB2U(0xD449, 0x007E2C);
	ChangeMB2U(0xD44A, 0x007E2D);
	ChangeMB2U(0xD44B, 0x007E2F);
	ChangeMB2U(0xD44C, 0x007E33);
	ChangeMB2U(0xD44D, 0x007E36);
	ChangeMB2U(0xD44E, 0x007E3F);
	ChangeMB2U(0xD44F, 0x007E44);
	ChangeMB2U(0xD450, 0x007E45);
	ChangeMB2U(0xD451, 0x007E47);
	ChangeMB2U(0xD452, 0x007E4E);
	ChangeMB2U(0xD453, 0x007E50);
	ChangeMB2U(0xD455, 0x007E58);
	ChangeMB2U(0xD456, 0x007E5F);
	ChangeMB2U(0xD457, 0x007E61);
	ChangeMB2U(0xD458, 0x007E62);
	ChangeMB2U(0xD459, 0x007E65);
	ChangeMB2U(0xD45A, 0x007E6B);
	ChangeMB2U(0xD45B, 0x007E6E);
	ChangeMB2U(0xD45C, 0x007E6F);
	ChangeMB2U(0xD45D, 0x007E73);
	ChangeMB2U(0xD45E, 0x007E78);
	ChangeMB2U(0xD45F, 0x007E7E);
	ChangeMB2U(0xD460, 0x007E81);
	ChangeMB2U(0xD461, 0x007E86);
	ChangeMB2U(0xD462, 0x007E87);
	ChangeMB2U(0xD464, 0x007E8D);
	ChangeMB2U(0xD465, 0x007E91);
	ChangeMB2U(0xD466, 0x007E95);
	ChangeMB2U(0xD467, 0x007E98);
	ChangeMB2U(0xD468, 0x007E9A);
	ChangeMB2U(0xD469, 0x007E9D);
	ChangeMB2U(0xD46A, 0x007E9E);
	ChangeMB2U(0xD46B, 0x007F3C);
	ChangeMB2U(0xD46C, 0x007F3B);
	ChangeMB2U(0xD46D, 0x007F3D);
	ChangeMB2U(0xD46E, 0x007F3E);
	ChangeMB2U(0xD46F, 0x007F3F);
	ChangeMB2U(0xD470, 0x007F43);
	ChangeMB2U(0xD471, 0x007F44);
	ChangeMB2U(0xD473, 0x007F4F);
	ChangeMB2U(0xD474, 0x007F52);
	ChangeMB2U(0xD475, 0x007F53);
	ChangeMB2U(0xD476, 0x007F5B);
	ChangeMB2U(0xD477, 0x007F5C);
	ChangeMB2U(0xD478, 0x007F5D);
	ChangeMB2U(0xD479, 0x007F61);
	ChangeMB2U(0xD47A, 0x007F63);
	ChangeMB2U(0xD47B, 0x007F64);
	ChangeMB2U(0xD47C, 0x007F65);
	ChangeMB2U(0xD47D, 0x007F66);
	ChangeMB2U(0xD47E, 0x007F6D);
	ChangeMB2U(0xD521, 0x007F71);
	ChangeMB2U(0xD522, 0x007F7D);
	ChangeMB2U(0xD523, 0x007F7E);
	ChangeMB2U(0xD524, 0x007F7F);
	ChangeMB2U(0xD525, 0x007F80);
	ChangeMB2U(0xD526, 0x007F8B);
	ChangeMB2U(0xD527, 0x007F8D);
	ChangeMB2U(0xD528, 0x007F8F);
	ChangeMB2U(0xD529, 0x007F90);
	ChangeMB2U(0xD52A, 0x007F91);
	ChangeMB2U(0xD52B, 0x007F96);
	ChangeMB2U(0xD52C, 0x007F97);
	ChangeMB2U(0xD52D, 0x007F9C);
	ChangeMB2U(0xD52F, 0x007FA2);
	ChangeMB2U(0xD530, 0x007FA6);
	ChangeMB2U(0xD531, 0x007FAA);
	ChangeMB2U(0xD532, 0x007FAD);
	ChangeMB2U(0xD533, 0x007FB4);
	ChangeMB2U(0xD534, 0x007FBC);
	ChangeMB2U(0xD535, 0x007FBF);
	ChangeMB2U(0xD536, 0x007FC0);
	ChangeMB2U(0xD537, 0x007FC3);
	ChangeMB2U(0xD538, 0x007FC8);
	ChangeMB2U(0xD539, 0x007FCE);
	ChangeMB2U(0xD53A, 0x007FCF);
	ChangeMB2U(0xD53B, 0x007FDB);
	ChangeMB2U(0xD53C, 0x007FDF);
	ChangeMB2U(0xD53D, 0x007FE3);
	ChangeMB2U(0xD53E, 0x007FE5);
	ChangeMB2U(0xD53F, 0x007FE8);
	ChangeMB2U(0xD540, 0x007FEC);
	ChangeMB2U(0xD541, 0x007FEE);
	ChangeMB2U(0xD542, 0x007FEF);
	ChangeMB2U(0xD543, 0x007FF2);
	ChangeMB2U(0xD544, 0x007FFA);
	ChangeMB2U(0xD545, 0x007FFD);
	ChangeMB2U(0xD546, 0x007FFE);
	ChangeMB2U(0xD547, 0x007FFF);
	ChangeMB2U(0xD548, 0x008007);
	ChangeMB2U(0xD549, 0x008008);
	ChangeMB2U(0xD54A, 0x00800A);
	ChangeMB2U(0xD54B, 0x00800D);
	ChangeMB2U(0xD54C, 0x00800E);
	ChangeMB2U(0xD54D, 0x00800F);
	ChangeMB2U(0xD54E, 0x008011);
	ChangeMB2U(0xD54F, 0x008013);
	ChangeMB2U(0xD550, 0x008014);
	ChangeMB2U(0xD551, 0x008016);
	ChangeMB2U(0xD552, 0x00801D);
	ChangeMB2U(0xD553, 0x00801E);
	ChangeMB2U(0xD554, 0x00801F);
	ChangeMB2U(0xD555, 0x008020);
	ChangeMB2U(0xD556, 0x008024);
	ChangeMB2U(0xD557, 0x008026);
	ChangeMB2U(0xD558, 0x00802C);
	ChangeMB2U(0xD559, 0x00802E);
	ChangeMB2U(0xD55A, 0x008030);
	ChangeMB2U(0xD55B, 0x008034);
	ChangeMB2U(0xD55C, 0x008035);
	ChangeMB2U(0xD55D, 0x008037);
	ChangeMB2U(0xD55E, 0x008039);
	ChangeMB2U(0xD55F, 0x00803A);
	ChangeMB2U(0xD560, 0x00803C);
	ChangeMB2U(0xD561, 0x00803E);
	ChangeMB2U(0xD562, 0x008040);
	ChangeMB2U(0xD563, 0x008044);
	ChangeMB2U(0xD564, 0x008060);
	ChangeMB2U(0xD565, 0x008064);
	ChangeMB2U(0xD566, 0x008066);
	ChangeMB2U(0xD567, 0x00806D);
	ChangeMB2U(0xD568, 0x008071);
	ChangeMB2U(0xD569, 0x008075);
	ChangeMB2U(0xD56A, 0x008081);
	ChangeMB2U(0xD56B, 0x008088);
	ChangeMB2U(0xD56C, 0x00808E);
	ChangeMB2U(0xD56D, 0x00809C);
	ChangeMB2U(0xD56E, 0x00809E);
	ChangeMB2U(0xD56F, 0x0080A6);
	ChangeMB2U(0xD570, 0x0080A7);
	ChangeMB2U(0xD571, 0x0080AB);
	ChangeMB2U(0xD572, 0x0080B8);
	ChangeMB2U(0xD573, 0x0080B9);
	ChangeMB2U(0xD574, 0x0080C8);
	ChangeMB2U(0xD575, 0x0080CD);
	ChangeMB2U(0xD576, 0x0080CF);
	ChangeMB2U(0xD577, 0x0080D2);
	ChangeMB2U(0xD578, 0x0080D4);
	ChangeMB2U(0xD579, 0x0080D5);
	ChangeMB2U(0xD57A, 0x0080D7);
	ChangeMB2U(0xD57B, 0x0080D8);
	ChangeMB2U(0xD57C, 0x0080E0);
	ChangeMB2U(0xD57D, 0x0080ED);
	ChangeMB2U(0xD57E, 0x0080EE);
	ChangeMB2U(0xD621, 0x0080F0);
	ChangeMB2U(0xD622, 0x0080F2);
	ChangeMB2U(0xD623, 0x0080F3);
	ChangeMB2U(0xD624, 0x0080F6);
	ChangeMB2U(0xD625, 0x0080F9);
	ChangeMB2U(0xD626, 0x0080FA);
	ChangeMB2U(0xD627, 0x0080FE);
	ChangeMB2U(0xD628, 0x008103);
	ChangeMB2U(0xD629, 0x00810B);
	ChangeMB2U(0xD62A, 0x008116);
	ChangeMB2U(0xD62B, 0x008117);
	ChangeMB2U(0xD62C, 0x008118);
	ChangeMB2U(0xD62D, 0x00811C);
	ChangeMB2U(0xD62E, 0x00811E);
	ChangeMB2U(0xD62F, 0x008120);
	ChangeMB2U(0xD630, 0x008124);
	ChangeMB2U(0xD631, 0x008127);
	ChangeMB2U(0xD632, 0x00812C);
	ChangeMB2U(0xD633, 0x008130);
	ChangeMB2U(0xD634, 0x008135);
	ChangeMB2U(0xD635, 0x00813A);
	ChangeMB2U(0xD636, 0x00813C);
	ChangeMB2U(0xD637, 0x008145);
	ChangeMB2U(0xD638, 0x008147);
	ChangeMB2U(0xD639, 0x00814A);
	ChangeMB2U(0xD63A, 0x00814C);
	ChangeMB2U(0xD63B, 0x008152);
	ChangeMB2U(0xD63C, 0x008157);
	ChangeMB2U(0xD63D, 0x008160);
	ChangeMB2U(0xD63E, 0x008161);
	ChangeMB2U(0xD63F, 0x008167);
	ChangeMB2U(0xD640, 0x008168);
	ChangeMB2U(0xD641, 0x008169);
	ChangeMB2U(0xD642, 0x00816D);
	ChangeMB2U(0xD643, 0x00816F);
	ChangeMB2U(0xD644, 0x008177);
	ChangeMB2U(0xD645, 0x008181);
	ChangeMB2U(0xD646, 0x008190);
	ChangeMB2U(0xD647, 0x008184);
	ChangeMB2U(0xD648, 0x008185);
	ChangeMB2U(0xD649, 0x008186);
	ChangeMB2U(0xD64A, 0x00818B);
	ChangeMB2U(0xD64B, 0x00818E);
	ChangeMB2U(0xD64C, 0x008196);
	ChangeMB2U(0xD64D, 0x008198);
	ChangeMB2U(0xD64E, 0x00819B);
	ChangeMB2U(0xD64F, 0x00819E);
	ChangeMB2U(0xD650, 0x0081A2);
	ChangeMB2U(0xD651, 0x0081AE);
	ChangeMB2U(0xD652, 0x0081B2);
	ChangeMB2U(0xD653, 0x0081B4);
	ChangeMB2U(0xD654, 0x0081BB);
	ChangeMB2U(0xD655, 0x0081CB);
	ChangeMB2U(0xD656, 0x0081C3);
	ChangeMB2U(0xD657, 0x0081C5);
	ChangeMB2U(0xD658, 0x0081CA);
	ChangeMB2U(0xD659, 0x0081CE);
	ChangeMB2U(0xD65A, 0x0081CF);
	ChangeMB2U(0xD65B, 0x0081D5);
	ChangeMB2U(0xD65C, 0x0081D7);
	ChangeMB2U(0xD65D, 0x0081DB);
	ChangeMB2U(0xD65E, 0x0081DD);
	ChangeMB2U(0xD65F, 0x0081DE);
	ChangeMB2U(0xD660, 0x0081E1);
	ChangeMB2U(0xD661, 0x0081E4);
	ChangeMB2U(0xD662, 0x0081EB);
	ChangeMB2U(0xD663, 0x0081EC);
	ChangeMB2U(0xD664, 0x0081F0);
	ChangeMB2U(0xD665, 0x0081F1);
	ChangeMB2U(0xD666, 0x0081F2);
	ChangeMB2U(0xD667, 0x0081F5);
	ChangeMB2U(0xD668, 0x0081F6);
	ChangeMB2U(0xD669, 0x0081F8);
	ChangeMB2U(0xD66A, 0x0081F9);
	ChangeMB2U(0xD66B, 0x0081FD);
	ChangeMB2U(0xD66C, 0x0081FF);
	ChangeMB2U(0xD66D, 0x008200);
	ChangeMB2U(0xD66E, 0x008203);
	ChangeMB2U(0xD66F, 0x00820F);
	ChangeMB2U(0xD670, 0x008213);
	ChangeMB2U(0xD671, 0x008214);
	ChangeMB2U(0xD672, 0x008219);
	ChangeMB2U(0xD673, 0x00821A);
	ChangeMB2U(0xD674, 0x00821D);
	ChangeMB2U(0xD675, 0x008221);
	ChangeMB2U(0xD676, 0x008222);
	ChangeMB2U(0xD677, 0x008228);
	ChangeMB2U(0xD678, 0x008232);
	ChangeMB2U(0xD679, 0x008234);
	ChangeMB2U(0xD67A, 0x00823A);
	ChangeMB2U(0xD67B, 0x008243);
	ChangeMB2U(0xD67C, 0x008244);
	ChangeMB2U(0xD67D, 0x008245);
	ChangeMB2U(0xD67E, 0x008246);
	ChangeMB2U(0xD721, 0x00824B);
	ChangeMB2U(0xD722, 0x00824E);
	ChangeMB2U(0xD723, 0x00824F);
	ChangeMB2U(0xD724, 0x008251);
	ChangeMB2U(0xD725, 0x008256);
	ChangeMB2U(0xD726, 0x00825C);
	ChangeMB2U(0xD727, 0x008260);
	ChangeMB2U(0xD728, 0x008263);
	ChangeMB2U(0xD729, 0x008267);
	ChangeMB2U(0xD72A, 0x00826D);
	ChangeMB2U(0xD72B, 0x008274);
	ChangeMB2U(0xD72C, 0x00827B);
	ChangeMB2U(0xD72D, 0x00827D);
	ChangeMB2U(0xD72E, 0x00827F);
	ChangeMB2U(0xD72F, 0x008280);
	ChangeMB2U(0xD730, 0x008281);
	ChangeMB2U(0xD731, 0x008283);
	ChangeMB2U(0xD732, 0x008284);
	ChangeMB2U(0xD733, 0x008287);
	ChangeMB2U(0xD734, 0x008289);
	ChangeMB2U(0xD735, 0x00828A);
	ChangeMB2U(0xD736, 0x00828E);
	ChangeMB2U(0xD737, 0x008291);
	ChangeMB2U(0xD738, 0x008294);
	ChangeMB2U(0xD739, 0x008296);
	ChangeMB2U(0xD73A, 0x008298);
	ChangeMB2U(0xD73B, 0x00829A);
	ChangeMB2U(0xD73C, 0x00829B);
	ChangeMB2U(0xD73D, 0x0082A0);
	ChangeMB2U(0xD73E, 0x0082A1);
	ChangeMB2U(0xD73F, 0x0082A3);
	ChangeMB2U(0xD740, 0x0082A4);
	ChangeMB2U(0xD741, 0x0082A7);
	ChangeMB2U(0xD742, 0x0082A8);
	ChangeMB2U(0xD743, 0x0082A9);
	ChangeMB2U(0xD744, 0x0082AA);
	ChangeMB2U(0xD745, 0x0082AE);
	ChangeMB2U(0xD746, 0x0082B0);
	ChangeMB2U(0xD747, 0x0082B2);
	ChangeMB2U(0xD748, 0x0082B4);
	ChangeMB2U(0xD749, 0x0082B7);
	ChangeMB2U(0xD74A, 0x0082BA);
	ChangeMB2U(0xD74B, 0x0082BC);
	ChangeMB2U(0xD74C, 0x0082BE);
	ChangeMB2U(0xD74D, 0x0082BF);
	ChangeMB2U(0xD74E, 0x0082C6);
	ChangeMB2U(0xD74F, 0x0082D0);
	ChangeMB2U(0xD750, 0x0082D5);
	ChangeMB2U(0xD751, 0x0082DA);
	ChangeMB2U(0xD752, 0x0082E0);
	ChangeMB2U(0xD753, 0x0082E2);
	ChangeMB2U(0xD754, 0x0082E4);
	ChangeMB2U(0xD755, 0x0082E8);
	ChangeMB2U(0xD756, 0x0082EA);
	ChangeMB2U(0xD757, 0x0082ED);
	ChangeMB2U(0xD758, 0x0082EF);
	ChangeMB2U(0xD759, 0x0082F6);
	ChangeMB2U(0xD75A, 0x0082F7);
	ChangeMB2U(0xD75B, 0x0082FD);
	ChangeMB2U(0xD75C, 0x0082FE);
	ChangeMB2U(0xD75D, 0x008300);
	ChangeMB2U(0xD75F, 0x008307);
	ChangeMB2U(0xD760, 0x008308);
	ChangeMB2U(0xD761, 0x00830A);
	ChangeMB2U(0xD762, 0x00830B);
	ChangeMB2U(0xD763, 0x008354);
	ChangeMB2U(0xD764, 0x00831B);
	ChangeMB2U(0xD765, 0x00831D);
	ChangeMB2U(0xD766, 0x00831E);
	ChangeMB2U(0xD767, 0x00831F);
	ChangeMB2U(0xD768, 0x008321);
	ChangeMB2U(0xD769, 0x008322);
	ChangeMB2U(0xD76A, 0x00832C);
	ChangeMB2U(0xD76B, 0x00832D);
	ChangeMB2U(0xD76C, 0x00832E);
	ChangeMB2U(0xD76D, 0x008330);
	ChangeMB2U(0xD76E, 0x008333);
	ChangeMB2U(0xD76F, 0x008337);
	ChangeMB2U(0xD770, 0x00833A);
	ChangeMB2U(0xD771, 0x00833C);
	ChangeMB2U(0xD772, 0x00833D);
	ChangeMB2U(0xD773, 0x008342);
	ChangeMB2U(0xD774, 0x008343);
	ChangeMB2U(0xD775, 0x008344);
	ChangeMB2U(0xD776, 0x008347);
	ChangeMB2U(0xD777, 0x00834D);
	ChangeMB2U(0xD778, 0x00834E);
	ChangeMB2U(0xD779, 0x008351);
	ChangeMB2U(0xD77A, 0x008355);
	ChangeMB2U(0xD77B, 0x008356);
	ChangeMB2U(0xD77C, 0x008357);
	ChangeMB2U(0xD77D, 0x008370);
	ChangeMB2U(0xD77E, 0x008378);
	ChangeMB2U(0xD821, 0x00837D);
	ChangeMB2U(0xD823, 0x008380);
	ChangeMB2U(0xD824, 0x008382);
	ChangeMB2U(0xD825, 0x008384);
	ChangeMB2U(0xD826, 0x008386);
	ChangeMB2U(0xD827, 0x00838D);
	ChangeMB2U(0xD828, 0x008392);
	ChangeMB2U(0xD829, 0x008394);
	ChangeMB2U(0xD82A, 0x008395);
	ChangeMB2U(0xD82B, 0x008398);
	ChangeMB2U(0xD82C, 0x008399);
	ChangeMB2U(0xD82D, 0x00839B);
	ChangeMB2U(0xD82E, 0x00839C);
	ChangeMB2U(0xD82F, 0x00839D);
	ChangeMB2U(0xD830, 0x0083A6);
	ChangeMB2U(0xD831, 0x0083A7);
	ChangeMB2U(0xD832, 0x0083A9);
	ChangeMB2U(0xD833, 0x0083AC);
	ChangeMB2U(0xD834, 0x0083BE);
	ChangeMB2U(0xD835, 0x0083BF);
	ChangeMB2U(0xD836, 0x0083C0);
	ChangeMB2U(0xD838, 0x0083C9);
	ChangeMB2U(0xD839, 0x0083CF);
	ChangeMB2U(0xD83A, 0x0083D0);
	ChangeMB2U(0xD83B, 0x0083D1);
	ChangeMB2U(0xD83C, 0x0083D4);
	ChangeMB2U(0xD83D, 0x0083DD);
	ChangeMB2U(0xD83E, 0x008353);
	ChangeMB2U(0xD83F, 0x0083E8);
	ChangeMB2U(0xD840, 0x0083EA);
	ChangeMB2U(0xD842, 0x0083F8);
	ChangeMB2U(0xD843, 0x0083F9);
	ChangeMB2U(0xD844, 0x0083FC);
	ChangeMB2U(0xD845, 0x008401);
	ChangeMB2U(0xD846, 0x008406);
	ChangeMB2U(0xD847, 0x00840A);
	ChangeMB2U(0xD848, 0x00840F);
	ChangeMB2U(0xD849, 0x008411);
	ChangeMB2U(0xD84A, 0x008415);
	ChangeMB2U(0xD84B, 0x008419);
	ChangeMB2U(0xD84C, 0x0083AD);
	ChangeMB2U(0xD84D, 0x00842F);
	ChangeMB2U(0xD84E, 0x008439);
	ChangeMB2U(0xD84F, 0x008445);
	ChangeMB2U(0xD850, 0x008447);
	ChangeMB2U(0xD852, 0x00844A);
	ChangeMB2U(0xD853, 0x00844D);
	ChangeMB2U(0xD854, 0x00844F);
	ChangeMB2U(0xD855, 0x008451);
	ChangeMB2U(0xD856, 0x008452);
	ChangeMB2U(0xD857, 0x008456);
	ChangeMB2U(0xD858, 0x008458);
	ChangeMB2U(0xD859, 0x008459);
	ChangeMB2U(0xD85A, 0x00845A);
	ChangeMB2U(0xD85B, 0x00845C);
	ChangeMB2U(0xD85C, 0x008460);
	ChangeMB2U(0xD85D, 0x008464);
	ChangeMB2U(0xD85E, 0x008465);
	ChangeMB2U(0xD85F, 0x008467);
	ChangeMB2U(0xD860, 0x00846A);
	ChangeMB2U(0xD861, 0x008470);
	ChangeMB2U(0xD862, 0x008473);
	ChangeMB2U(0xD863, 0x008474);
	ChangeMB2U(0xD864, 0x008476);
	ChangeMB2U(0xD865, 0x008478);
	ChangeMB2U(0xD866, 0x00847C);
	ChangeMB2U(0xD867, 0x00847D);
	ChangeMB2U(0xD868, 0x008481);
	ChangeMB2U(0xD869, 0x008485);
	ChangeMB2U(0xD86A, 0x008492);
	ChangeMB2U(0xD86B, 0x008493);
	ChangeMB2U(0xD86C, 0x008495);
	ChangeMB2U(0xD86D, 0x00849E);
	ChangeMB2U(0xD86E, 0x0084A6);
	ChangeMB2U(0xD86F, 0x0084A8);
	ChangeMB2U(0xD870, 0x0084A9);
	ChangeMB2U(0xD871, 0x0084AA);
	ChangeMB2U(0xD872, 0x0084AF);
	ChangeMB2U(0xD873, 0x0084B1);
	ChangeMB2U(0xD875, 0x0084BA);
	ChangeMB2U(0xD876, 0x0084BD);
	ChangeMB2U(0xD877, 0x0084BE);
	ChangeMB2U(0xD878, 0x0084C0);
	ChangeMB2U(0xD879, 0x0084C2);
	ChangeMB2U(0xD87A, 0x0084C7);
	ChangeMB2U(0xD87B, 0x0084C8);
	ChangeMB2U(0xD87C, 0x0084CC);
	ChangeMB2U(0xD87D, 0x0084CF);
	ChangeMB2U(0xD87E, 0x0084D3);
	ChangeMB2U(0xD922, 0x0084E7);
	ChangeMB2U(0xD923, 0x0084EA);
	ChangeMB2U(0xD924, 0x0084EF);
	ChangeMB2U(0xD925, 0x0084F0);
	ChangeMB2U(0xD926, 0x0084F1);
	ChangeMB2U(0xD927, 0x0084F2);
	ChangeMB2U(0xD928, 0x0084F7);
	ChangeMB2U(0xD929, 0x008532);
	ChangeMB2U(0xD92A, 0x0084FA);
	ChangeMB2U(0xD92B, 0x0084FB);
	ChangeMB2U(0xD92C, 0x0084FD);
	ChangeMB2U(0xD92D, 0x008502);
	ChangeMB2U(0xD92E, 0x008503);
	ChangeMB2U(0xD92F, 0x008507);
	ChangeMB2U(0xD930, 0x00850C);
	ChangeMB2U(0xD931, 0x00850E);
	ChangeMB2U(0xD932, 0x008510);
	ChangeMB2U(0xD933, 0x00851C);
	ChangeMB2U(0xD934, 0x00851E);
	ChangeMB2U(0xD935, 0x008522);
	ChangeMB2U(0xD936, 0x008523);
	ChangeMB2U(0xD937, 0x008524);
	ChangeMB2U(0xD938, 0x008525);
	ChangeMB2U(0xD939, 0x008527);
	ChangeMB2U(0xD93A, 0x00852A);
	ChangeMB2U(0xD93B, 0x00852B);
	ChangeMB2U(0xD93C, 0x00852F);
	ChangeMB2U(0xD93D, 0x008533);
	ChangeMB2U(0xD93E, 0x008534);
	ChangeMB2U(0xD93F, 0x008536);
	ChangeMB2U(0xD940, 0x00853F);
	ChangeMB2U(0xD941, 0x008546);
	ChangeMB2U(0xD942, 0x00854F);
	ChangeMB2U(0xD943, 0x008550);
	ChangeMB2U(0xD944, 0x008551);
	ChangeMB2U(0xD945, 0x008552);
	ChangeMB2U(0xD947, 0x008556);
	ChangeMB2U(0xD949, 0x00855C);
	ChangeMB2U(0xD94A, 0x00855D);
	ChangeMB2U(0xD94B, 0x00855E);
	ChangeMB2U(0xD94C, 0x00855F);
	ChangeMB2U(0xD94D, 0x008560);
	ChangeMB2U(0xD94E, 0x008561);
	ChangeMB2U(0xD94F, 0x008562);
	ChangeMB2U(0xD950, 0x008564);
	ChangeMB2U(0xD952, 0x00856F);
	ChangeMB2U(0xD953, 0x008579);
	ChangeMB2U(0xD954, 0x00857A);
	ChangeMB2U(0xD955, 0x00857B);
	ChangeMB2U(0xD956, 0x00857D);
	ChangeMB2U(0xD957, 0x00857F);
	ChangeMB2U(0xD958, 0x008581);
	ChangeMB2U(0xD959, 0x008585);
	ChangeMB2U(0xD95A, 0x008586);
	ChangeMB2U(0xD95B, 0x008589);
	ChangeMB2U(0xD95C, 0x00858B);
	ChangeMB2U(0xD95D, 0x00858C);
	ChangeMB2U(0xD95E, 0x00858F);
	ChangeMB2U(0xD95F, 0x008593);
	ChangeMB2U(0xD960, 0x008598);
	ChangeMB2U(0xD961, 0x00859D);
	ChangeMB2U(0xD962, 0x00859F);
	ChangeMB2U(0xD963, 0x0085A0);
	ChangeMB2U(0xD964, 0x0085A2);
	ChangeMB2U(0xD965, 0x0085A5);
	ChangeMB2U(0xD966, 0x0085A7);
	ChangeMB2U(0xD967, 0x0085B4);
	ChangeMB2U(0xD968, 0x0085B6);
	ChangeMB2U(0xD969, 0x0085B7);
	ChangeMB2U(0xD96A, 0x0085B8);
	ChangeMB2U(0xD96B, 0x0085BC);
	ChangeMB2U(0xD96C, 0x0085BD);
	ChangeMB2U(0xD96D, 0x0085BE);
	ChangeMB2U(0xD96E, 0x0085BF);
	ChangeMB2U(0xD96F, 0x0085C2);
	ChangeMB2U(0xD970, 0x0085C7);
	ChangeMB2U(0xD971, 0x0085CA);
	ChangeMB2U(0xD972, 0x0085CB);
	ChangeMB2U(0xD973, 0x0085CE);
	ChangeMB2U(0xD974, 0x0085AD);
	ChangeMB2U(0xD975, 0x0085D8);
	ChangeMB2U(0xD976, 0x0085DA);
	ChangeMB2U(0xD977, 0x0085DF);
	ChangeMB2U(0xD978, 0x0085E0);
	ChangeMB2U(0xD979, 0x0085E6);
	ChangeMB2U(0xD97A, 0x0085E8);
	ChangeMB2U(0xD97B, 0x0085ED);
	ChangeMB2U(0xD97C, 0x0085F3);
	ChangeMB2U(0xD97D, 0x0085F6);
	ChangeMB2U(0xD97E, 0x0085FC);
	ChangeMB2U(0xDA21, 0x0085FF);
	ChangeMB2U(0xDA22, 0x008600);
	ChangeMB2U(0xDA23, 0x008604);
	ChangeMB2U(0xDA24, 0x008605);
	ChangeMB2U(0xDA25, 0x00860D);
	ChangeMB2U(0xDA26, 0x00860E);
	ChangeMB2U(0xDA27, 0x008610);
	ChangeMB2U(0xDA28, 0x008611);
	ChangeMB2U(0xDA29, 0x008612);
	ChangeMB2U(0xDA2A, 0x008618);
	ChangeMB2U(0xDA2B, 0x008619);
	ChangeMB2U(0xDA2C, 0x00861B);
	ChangeMB2U(0xDA2D, 0x00861E);
	ChangeMB2U(0xDA2E, 0x008621);
	ChangeMB2U(0xDA2F, 0x008627);
	ChangeMB2U(0xDA30, 0x008629);
	ChangeMB2U(0xDA31, 0x008636);
	ChangeMB2U(0xDA32, 0x008638);
	ChangeMB2U(0xDA33, 0x00863A);
	ChangeMB2U(0xDA34, 0x00863C);
	ChangeMB2U(0xDA35, 0x00863D);
	ChangeMB2U(0xDA36, 0x008640);
	ChangeMB2U(0xDA37, 0x008642);
	ChangeMB2U(0xDA38, 0x008646);
	ChangeMB2U(0xDA39, 0x008652);
	ChangeMB2U(0xDA3A, 0x008653);
	ChangeMB2U(0xDA3B, 0x008656);
	ChangeMB2U(0xDA3C, 0x008657);
	ChangeMB2U(0xDA3D, 0x008658);
	ChangeMB2U(0xDA3E, 0x008659);
	ChangeMB2U(0xDA3F, 0x00865D);
	ChangeMB2U(0xDA40, 0x008660);
	ChangeMB2U(0xDA41, 0x008661);
	ChangeMB2U(0xDA42, 0x008662);
	ChangeMB2U(0xDA43, 0x008663);
	ChangeMB2U(0xDA44, 0x008664);
	ChangeMB2U(0xDA45, 0x008669);
	ChangeMB2U(0xDA46, 0x00866C);
	ChangeMB2U(0xDA47, 0x00866F);
	ChangeMB2U(0xDA48, 0x008675);
	ChangeMB2U(0xDA49, 0x008676);
	ChangeMB2U(0xDA4A, 0x008677);
	ChangeMB2U(0xDA4B, 0x00867A);
	ChangeMB2U(0xDA4C, 0x00868D);
	ChangeMB2U(0xDA4D, 0x008691);
	ChangeMB2U(0xDA4E, 0x008696);
	ChangeMB2U(0xDA4F, 0x008698);
	ChangeMB2U(0xDA50, 0x00869A);
	ChangeMB2U(0xDA51, 0x00869C);
	ChangeMB2U(0xDA52, 0x0086A1);
	ChangeMB2U(0xDA53, 0x0086A6);
	ChangeMB2U(0xDA54, 0x0086A7);
	ChangeMB2U(0xDA55, 0x0086A8);
	ChangeMB2U(0xDA56, 0x0086AD);
	ChangeMB2U(0xDA57, 0x0086B1);
	ChangeMB2U(0xDA58, 0x0086B3);
	ChangeMB2U(0xDA59, 0x0086B4);
	ChangeMB2U(0xDA5A, 0x0086B5);
	ChangeMB2U(0xDA5B, 0x0086B7);
	ChangeMB2U(0xDA5C, 0x0086B8);
	ChangeMB2U(0xDA5D, 0x0086B9);
	ChangeMB2U(0xDA5E, 0x0086BF);
	ChangeMB2U(0xDA5F, 0x0086C0);
	ChangeMB2U(0xDA60, 0x0086C1);
	ChangeMB2U(0xDA61, 0x0086C3);
	ChangeMB2U(0xDA62, 0x0086C5);
	ChangeMB2U(0xDA63, 0x0086D1);
	ChangeMB2U(0xDA64, 0x0086D2);
	ChangeMB2U(0xDA65, 0x0086D5);
	ChangeMB2U(0xDA66, 0x0086D7);
	ChangeMB2U(0xDA67, 0x0086DA);
	ChangeMB2U(0xDA68, 0x0086DC);
	ChangeMB2U(0xDA69, 0x0086E0);
	ChangeMB2U(0xDA6A, 0x0086E3);
	ChangeMB2U(0xDA6B, 0x0086E5);
	ChangeMB2U(0xDA6C, 0x0086E7);
	ChangeMB2U(0xDA6D, 0x008688);
	ChangeMB2U(0xDA6E, 0x0086FA);
	ChangeMB2U(0xDA6F, 0x0086FC);
	ChangeMB2U(0xDA70, 0x0086FD);
	ChangeMB2U(0xDA71, 0x008704);
	ChangeMB2U(0xDA72, 0x008705);
	ChangeMB2U(0xDA73, 0x008707);
	ChangeMB2U(0xDA74, 0x00870B);
	ChangeMB2U(0xDA75, 0x00870E);
	ChangeMB2U(0xDA76, 0x00870F);
	ChangeMB2U(0xDA77, 0x008710);
	ChangeMB2U(0xDA78, 0x008713);
	ChangeMB2U(0xDA79, 0x008714);
	ChangeMB2U(0xDA7A, 0x008719);
	ChangeMB2U(0xDA7B, 0x00871E);
	ChangeMB2U(0xDA7C, 0x00871F);
	ChangeMB2U(0xDA7D, 0x008721);
	ChangeMB2U(0xDA7E, 0x008723);
	ChangeMB2U(0xDB21, 0x008728);
	ChangeMB2U(0xDB22, 0x00872E);
	ChangeMB2U(0xDB23, 0x00872F);
	ChangeMB2U(0xDB24, 0x008731);
	ChangeMB2U(0xDB25, 0x008732);
	ChangeMB2U(0xDB26, 0x008739);
	ChangeMB2U(0xDB27, 0x00873A);
	ChangeMB2U(0xDB28, 0x00873C);
	ChangeMB2U(0xDB29, 0x00873D);
	ChangeMB2U(0xDB2A, 0x00873E);
	ChangeMB2U(0xDB2B, 0x008740);
	ChangeMB2U(0xDB2C, 0x008743);
	ChangeMB2U(0xDB2D, 0x008745);
	ChangeMB2U(0xDB2E, 0x00874D);
	ChangeMB2U(0xDB2F, 0x008758);
	ChangeMB2U(0xDB30, 0x00875D);
	ChangeMB2U(0xDB31, 0x008761);
	ChangeMB2U(0xDB32, 0x008764);
	ChangeMB2U(0xDB33, 0x008765);
	ChangeMB2U(0xDB34, 0x00876F);
	ChangeMB2U(0xDB35, 0x008771);
	ChangeMB2U(0xDB36, 0x008772);
	ChangeMB2U(0xDB37, 0x00877B);
	ChangeMB2U(0xDB38, 0x008783);
	ChangeMB2U(0xDB39, 0x008784);
	ChangeMB2U(0xDB3A, 0x008785);
	ChangeMB2U(0xDB3B, 0x008786);
	ChangeMB2U(0xDB3C, 0x008787);
	ChangeMB2U(0xDB3D, 0x008788);
	ChangeMB2U(0xDB3E, 0x008789);
	ChangeMB2U(0xDB3F, 0x00878B);
	ChangeMB2U(0xDB40, 0x00878C);
	ChangeMB2U(0xDB41, 0x008790);
	ChangeMB2U(0xDB42, 0x008793);
	ChangeMB2U(0xDB43, 0x008795);
	ChangeMB2U(0xDB44, 0x008797);
	ChangeMB2U(0xDB45, 0x008798);
	ChangeMB2U(0xDB46, 0x008799);
	ChangeMB2U(0xDB47, 0x00879E);
	ChangeMB2U(0xDB48, 0x0087A0);
	ChangeMB2U(0xDB49, 0x0087A3);
	ChangeMB2U(0xDB4A, 0x0087A7);
	ChangeMB2U(0xDB4B, 0x0087AC);
	ChangeMB2U(0xDB4C, 0x0087AD);
	ChangeMB2U(0xDB4D, 0x0087AE);
	ChangeMB2U(0xDB4E, 0x0087B1);
	ChangeMB2U(0xDB4F, 0x0087B5);
	ChangeMB2U(0xDB50, 0x0087BE);
	ChangeMB2U(0xDB51, 0x0087BF);
	ChangeMB2U(0xDB52, 0x0087C1);
	ChangeMB2U(0xDB53, 0x0087C8);
	ChangeMB2U(0xDB54, 0x0087C9);
	ChangeMB2U(0xDB55, 0x0087CA);
	ChangeMB2U(0xDB56, 0x0087CE);
	ChangeMB2U(0xDB57, 0x0087D5);
	ChangeMB2U(0xDB58, 0x0087D6);
	ChangeMB2U(0xDB59, 0x0087D9);
	ChangeMB2U(0xDB5A, 0x0087DA);
	ChangeMB2U(0xDB5B, 0x0087DC);
	ChangeMB2U(0xDB5C, 0x0087DF);
	ChangeMB2U(0xDB5D, 0x0087E2);
	ChangeMB2U(0xDB5E, 0x0087E3);
	ChangeMB2U(0xDB5F, 0x0087E4);
	ChangeMB2U(0xDB60, 0x0087EA);
	ChangeMB2U(0xDB61, 0x0087EB);
	ChangeMB2U(0xDB62, 0x0087ED);
	ChangeMB2U(0xDB63, 0x0087F1);
	ChangeMB2U(0xDB64, 0x0087F3);
	ChangeMB2U(0xDB65, 0x0087F8);
	ChangeMB2U(0xDB66, 0x0087FA);
	ChangeMB2U(0xDB67, 0x0087FF);
	ChangeMB2U(0xDB68, 0x008801);
	ChangeMB2U(0xDB69, 0x008803);
	ChangeMB2U(0xDB6A, 0x008806);
	ChangeMB2U(0xDB6B, 0x008809);
	ChangeMB2U(0xDB6C, 0x00880A);
	ChangeMB2U(0xDB6D, 0x00880B);
	ChangeMB2U(0xDB6E, 0x008810);
	ChangeMB2U(0xDB6F, 0x008819);
	ChangeMB2U(0xDB70, 0x008812);
	ChangeMB2U(0xDB71, 0x008813);
	ChangeMB2U(0xDB72, 0x008814);
	ChangeMB2U(0xDB73, 0x008818);
	ChangeMB2U(0xDB74, 0x00881A);
	ChangeMB2U(0xDB75, 0x00881B);
	ChangeMB2U(0xDB76, 0x00881C);
	ChangeMB2U(0xDB77, 0x00881E);
	ChangeMB2U(0xDB78, 0x00881F);
	ChangeMB2U(0xDB79, 0x008828);
	ChangeMB2U(0xDB7A, 0x00882D);
	ChangeMB2U(0xDB7B, 0x00882E);
	ChangeMB2U(0xDB7C, 0x008830);
	ChangeMB2U(0xDB7D, 0x008832);
	ChangeMB2U(0xDB7E, 0x008835);
	ChangeMB2U(0xDC21, 0x00883A);
	ChangeMB2U(0xDC22, 0x00883C);
	ChangeMB2U(0xDC23, 0x008841);
	ChangeMB2U(0xDC24, 0x008843);
	ChangeMB2U(0xDC25, 0x008845);
	ChangeMB2U(0xDC26, 0x008848);
	ChangeMB2U(0xDC27, 0x008849);
	ChangeMB2U(0xDC28, 0x00884A);
	ChangeMB2U(0xDC29, 0x00884B);
	ChangeMB2U(0xDC2A, 0x00884E);
	ChangeMB2U(0xDC2B, 0x008851);
	ChangeMB2U(0xDC2C, 0x008855);
	ChangeMB2U(0xDC2D, 0x008856);
	ChangeMB2U(0xDC2E, 0x008858);
	ChangeMB2U(0xDC2F, 0x00885A);
	ChangeMB2U(0xDC30, 0x00885C);
	ChangeMB2U(0xDC31, 0x00885F);
	ChangeMB2U(0xDC32, 0x008860);
	ChangeMB2U(0xDC33, 0x008864);
	ChangeMB2U(0xDC34, 0x008869);
	ChangeMB2U(0xDC35, 0x008871);
	ChangeMB2U(0xDC36, 0x008879);
	ChangeMB2U(0xDC37, 0x00887B);
	ChangeMB2U(0xDC38, 0x008880);
	ChangeMB2U(0xDC39, 0x008898);
	ChangeMB2U(0xDC3A, 0x00889A);
	ChangeMB2U(0xDC3B, 0x00889B);
	ChangeMB2U(0xDC3C, 0x00889C);
	ChangeMB2U(0xDC3D, 0x00889F);
	ChangeMB2U(0xDC3E, 0x0088A0);
	ChangeMB2U(0xDC3F, 0x0088A8);
	ChangeMB2U(0xDC40, 0x0088AA);
	ChangeMB2U(0xDC41, 0x0088BA);
	ChangeMB2U(0xDC42, 0x0088BD);
	ChangeMB2U(0xDC43, 0x0088BE);
	ChangeMB2U(0xDC44, 0x0088C0);
	ChangeMB2U(0xDC45, 0x0088CA);
	ChangeMB2U(0xDC46, 0x0088CB);
	ChangeMB2U(0xDC47, 0x0088CC);
	ChangeMB2U(0xDC48, 0x0088CD);
	ChangeMB2U(0xDC49, 0x0088CE);
	ChangeMB2U(0xDC4A, 0x0088D1);
	ChangeMB2U(0xDC4B, 0x0088D2);
	ChangeMB2U(0xDC4C, 0x0088D3);
	ChangeMB2U(0xDC4D, 0x0088DB);
	ChangeMB2U(0xDC4E, 0x0088DE);
	ChangeMB2U(0xDC4F, 0x0088E7);
	ChangeMB2U(0xDC50, 0x0088EF);
	ChangeMB2U(0xDC51, 0x0088F0);
	ChangeMB2U(0xDC52, 0x0088F1);
	ChangeMB2U(0xDC54, 0x0088F7);
	ChangeMB2U(0xDC55, 0x008901);
	ChangeMB2U(0xDC56, 0x008906);
	ChangeMB2U(0xDC57, 0x00890D);
	ChangeMB2U(0xDC58, 0x00890E);
	ChangeMB2U(0xDC59, 0x00890F);
	ChangeMB2U(0xDC5A, 0x008915);
	ChangeMB2U(0xDC5B, 0x008916);
	ChangeMB2U(0xDC5C, 0x008918);
	ChangeMB2U(0xDC5D, 0x008919);
	ChangeMB2U(0xDC5E, 0x00891A);
	ChangeMB2U(0xDC60, 0x008920);
	ChangeMB2U(0xDC61, 0x008926);
	ChangeMB2U(0xDC62, 0x008927);
	ChangeMB2U(0xDC63, 0x008928);
	ChangeMB2U(0xDC64, 0x008930);
	ChangeMB2U(0xDC65, 0x008931);
	ChangeMB2U(0xDC66, 0x008932);
	ChangeMB2U(0xDC67, 0x008935);
	ChangeMB2U(0xDC68, 0x008939);
	ChangeMB2U(0xDC69, 0x00893A);
	ChangeMB2U(0xDC6A, 0x00893E);
	ChangeMB2U(0xDC6B, 0x008940);
	ChangeMB2U(0xDC6C, 0x008942);
	ChangeMB2U(0xDC6D, 0x008945);
	ChangeMB2U(0xDC6E, 0x008946);
	ChangeMB2U(0xDC6F, 0x008949);
	ChangeMB2U(0xDC70, 0x00894F);
	ChangeMB2U(0xDC71, 0x008952);
	ChangeMB2U(0xDC72, 0x008957);
	ChangeMB2U(0xDC73, 0x00895A);
	ChangeMB2U(0xDC74, 0x00895B);
	ChangeMB2U(0xDC75, 0x00895C);
	ChangeMB2U(0xDC76, 0x008961);
	ChangeMB2U(0xDC77, 0x008962);
	ChangeMB2U(0xDC78, 0x008963);
	ChangeMB2U(0xDC79, 0x00896B);
	ChangeMB2U(0xDC7A, 0x00896E);
	ChangeMB2U(0xDC7B, 0x008970);
	ChangeMB2U(0xDC7C, 0x008973);
	ChangeMB2U(0xDC7D, 0x008975);
	ChangeMB2U(0xDC7E, 0x00897A);
	ChangeMB2U(0xDD21, 0x00897B);
	ChangeMB2U(0xDD22, 0x00897C);
	ChangeMB2U(0xDD23, 0x00897D);
	ChangeMB2U(0xDD24, 0x008989);
	ChangeMB2U(0xDD25, 0x00898D);
	ChangeMB2U(0xDD26, 0x008990);
	ChangeMB2U(0xDD27, 0x008994);
	ChangeMB2U(0xDD28, 0x008995);
	ChangeMB2U(0xDD29, 0x00899B);
	ChangeMB2U(0xDD2A, 0x00899C);
	ChangeMB2U(0xDD2B, 0x00899F);
	ChangeMB2U(0xDD2C, 0x0089A0);
	ChangeMB2U(0xDD2D, 0x0089A5);
	ChangeMB2U(0xDD2E, 0x0089B0);
	ChangeMB2U(0xDD2F, 0x0089B4);
	ChangeMB2U(0xDD30, 0x0089B5);
	ChangeMB2U(0xDD31, 0x0089B6);
	ChangeMB2U(0xDD32, 0x0089B7);
	ChangeMB2U(0xDD33, 0x0089BC);
	ChangeMB2U(0xDD34, 0x0089D4);
	ChangeMB2U(0xDD35, 0x0089D5);
	ChangeMB2U(0xDD36, 0x0089D6);
	ChangeMB2U(0xDD37, 0x0089D7);
	ChangeMB2U(0xDD38, 0x0089D8);
	ChangeMB2U(0xDD39, 0x0089E5);
	ChangeMB2U(0xDD3A, 0x0089E9);
	ChangeMB2U(0xDD3B, 0x0089EB);
	ChangeMB2U(0xDD3C, 0x0089ED);
	ChangeMB2U(0xDD3D, 0x0089F1);
	ChangeMB2U(0xDD3E, 0x0089F3);
	ChangeMB2U(0xDD3F, 0x0089F6);
	ChangeMB2U(0xDD40, 0x0089F9);
	ChangeMB2U(0xDD41, 0x0089FD);
	ChangeMB2U(0xDD42, 0x0089FF);
	ChangeMB2U(0xDD43, 0x008A04);
	ChangeMB2U(0xDD44, 0x008A05);
	ChangeMB2U(0xDD45, 0x008A07);
	ChangeMB2U(0xDD46, 0x008A0F);
	ChangeMB2U(0xDD47, 0x008A11);
	ChangeMB2U(0xDD49, 0x008A14);
	ChangeMB2U(0xDD4A, 0x008A15);
	ChangeMB2U(0xDD4B, 0x008A1E);
	ChangeMB2U(0xDD4C, 0x008A20);
	ChangeMB2U(0xDD4D, 0x008A22);
	ChangeMB2U(0xDD4E, 0x008A24);
	ChangeMB2U(0xDD4F, 0x008A26);
	ChangeMB2U(0xDD50, 0x008A2B);
	ChangeMB2U(0xDD51, 0x008A2C);
	ChangeMB2U(0xDD52, 0x008A2F);
	ChangeMB2U(0xDD53, 0x008A35);
	ChangeMB2U(0xDD55, 0x008A3D);
	ChangeMB2U(0xDD56, 0x008A3E);
	ChangeMB2U(0xDD57, 0x008A40);
	ChangeMB2U(0xDD58, 0x008A43);
	ChangeMB2U(0xDD59, 0x008A45);
	ChangeMB2U(0xDD5A, 0x008A47);
	ChangeMB2U(0xDD5B, 0x008A49);
	ChangeMB2U(0xDD5C, 0x008A4D);
	ChangeMB2U(0xDD5D, 0x008A4E);
	ChangeMB2U(0xDD5E, 0x008A53);
	ChangeMB2U(0xDD5F, 0x008A56);
	ChangeMB2U(0xDD60, 0x008A57);
	ChangeMB2U(0xDD61, 0x008A58);
	ChangeMB2U(0xDD62, 0x008A5C);
	ChangeMB2U(0xDD63, 0x008A5D);
	ChangeMB2U(0xDD64, 0x008A61);
	ChangeMB2U(0xDD65, 0x008A65);
	ChangeMB2U(0xDD66, 0x008A67);
	ChangeMB2U(0xDD67, 0x008A75);
	ChangeMB2U(0xDD68, 0x008A76);
	ChangeMB2U(0xDD69, 0x008A77);
	ChangeMB2U(0xDD6B, 0x008A7A);
	ChangeMB2U(0xDD6C, 0x008A7B);
	ChangeMB2U(0xDD6D, 0x008A7E);
	ChangeMB2U(0xDD6E, 0x008A7F);
	ChangeMB2U(0xDD6F, 0x008A80);
	ChangeMB2U(0xDD70, 0x008A83);
	ChangeMB2U(0xDD71, 0x008A86);
	ChangeMB2U(0xDD72, 0x008A8B);
	ChangeMB2U(0xDD73, 0x008A8F);
	ChangeMB2U(0xDD74, 0x008A90);
	ChangeMB2U(0xDD75, 0x008A92);
	ChangeMB2U(0xDD76, 0x008A96);
	ChangeMB2U(0xDD77, 0x008A97);
	ChangeMB2U(0xDD78, 0x008A99);
	ChangeMB2U(0xDD79, 0x008A9F);
	ChangeMB2U(0xDD7B, 0x008AA9);
	ChangeMB2U(0xDD7C, 0x008AAE);
	ChangeMB2U(0xDD7D, 0x008AAF);
	ChangeMB2U(0xDD7E, 0x008AB3);
	ChangeMB2U(0xDE21, 0x008AB6);
	ChangeMB2U(0xDE22, 0x008AB7);
	ChangeMB2U(0xDE23, 0x008ABB);
	ChangeMB2U(0xDE25, 0x008AC3);
	ChangeMB2U(0xDE26, 0x008AC6);
	ChangeMB2U(0xDE27, 0x008AC8);
	ChangeMB2U(0xDE28, 0x008AC9);
	ChangeMB2U(0xDE29, 0x008ACA);
	ChangeMB2U(0xDE2A, 0x008AD1);
	ChangeMB2U(0xDE2B, 0x008AD3);
	ChangeMB2U(0xDE2C, 0x008AD4);
	ChangeMB2U(0xDE2D, 0x008AD5);
	ChangeMB2U(0xDE2E, 0x008AD7);
	ChangeMB2U(0xDE2F, 0x008ADD);
	ChangeMB2U(0xDE31, 0x008AEC);
	ChangeMB2U(0xDE32, 0x008AF0);
	ChangeMB2U(0xDE33, 0x008AF4);
	ChangeMB2U(0xDE34, 0x008AF5);
	ChangeMB2U(0xDE36, 0x008AFC);
	ChangeMB2U(0xDE37, 0x008AFF);
	ChangeMB2U(0xDE38, 0x008B05);
	ChangeMB2U(0xDE39, 0x008B06);
	ChangeMB2U(0xDE3A, 0x008B0B);
	ChangeMB2U(0xDE3B, 0x008B11);
	ChangeMB2U(0xDE3C, 0x008B1C);
	ChangeMB2U(0xDE3D, 0x008B1E);
	ChangeMB2U(0xDE3E, 0x008B1F);
	ChangeMB2U(0xDE3F, 0x008B0A);
	ChangeMB2U(0xDE40, 0x008B2D);
	ChangeMB2U(0xDE41, 0x008B30);
	ChangeMB2U(0xDE42, 0x008B37);
	ChangeMB2U(0xDE43, 0x008B3C);
	ChangeMB2U(0xDE44, 0x008B42);
	ChangeMB2U(0xDE45, 0x008B43);
	ChangeMB2U(0xDE46, 0x008B44);
	ChangeMB2U(0xDE47, 0x008B45);
	ChangeMB2U(0xDE48, 0x008B46);
	ChangeMB2U(0xDE49, 0x008B48);
	ChangeMB2U(0xDE4A, 0x008B52);
	ChangeMB2U(0xDE4C, 0x008B54);
	ChangeMB2U(0xDE4D, 0x008B59);
	ChangeMB2U(0xDE4E, 0x008B4D);
	ChangeMB2U(0xDE4F, 0x008B5E);
	ChangeMB2U(0xDE50, 0x008B63);
	ChangeMB2U(0xDE51, 0x008B6D);
	ChangeMB2U(0xDE52, 0x008B76);
	ChangeMB2U(0xDE53, 0x008B78);
	ChangeMB2U(0xDE54, 0x008B79);
	ChangeMB2U(0xDE55, 0x008B7C);
	ChangeMB2U(0xDE56, 0x008B7E);
	ChangeMB2U(0xDE57, 0x008B81);
	ChangeMB2U(0xDE58, 0x008B84);
	ChangeMB2U(0xDE59, 0x008B85);
	ChangeMB2U(0xDE5A, 0x008B8B);
	ChangeMB2U(0xDE5B, 0x008B8D);
	ChangeMB2U(0xDE5C, 0x008B8F);
	ChangeMB2U(0xDE5D, 0x008B94);
	ChangeMB2U(0xDE5E, 0x008B95);
	ChangeMB2U(0xDE5F, 0x008B9C);
	ChangeMB2U(0xDE60, 0x008B9E);
	ChangeMB2U(0xDE61, 0x008B9F);
	ChangeMB2U(0xDE62, 0x008C38);
	ChangeMB2U(0xDE63, 0x008C39);
	ChangeMB2U(0xDE64, 0x008C3D);
	ChangeMB2U(0xDE65, 0x008C3E);
	ChangeMB2U(0xDE66, 0x008C45);
	ChangeMB2U(0xDE67, 0x008C47);
	ChangeMB2U(0xDE68, 0x008C49);
	ChangeMB2U(0xDE69, 0x008C4B);
	ChangeMB2U(0xDE6A, 0x008C4F);
	ChangeMB2U(0xDE6B, 0x008C51);
	ChangeMB2U(0xDE6C, 0x008C53);
	ChangeMB2U(0xDE6D, 0x008C54);
	ChangeMB2U(0xDE6E, 0x008C57);
	ChangeMB2U(0xDE6F, 0x008C58);
	ChangeMB2U(0xDE70, 0x008C5B);
	ChangeMB2U(0xDE71, 0x008C5D);
	ChangeMB2U(0xDE72, 0x008C59);
	ChangeMB2U(0xDE73, 0x008C63);
	ChangeMB2U(0xDE74, 0x008C64);
	ChangeMB2U(0xDE75, 0x008C66);
	ChangeMB2U(0xDE76, 0x008C68);
	ChangeMB2U(0xDE77, 0x008C69);
	ChangeMB2U(0xDE78, 0x008C6D);
	ChangeMB2U(0xDE79, 0x008C73);
	ChangeMB2U(0xDE7A, 0x008C75);
	ChangeMB2U(0xDE7B, 0x008C76);
	ChangeMB2U(0xDE7C, 0x008C7B);
	ChangeMB2U(0xDE7D, 0x008C7E);
	ChangeMB2U(0xDE7E, 0x008C86);
	ChangeMB2U(0xDF21, 0x008C87);
	ChangeMB2U(0xDF22, 0x008C8B);
	ChangeMB2U(0xDF23, 0x008C90);
	ChangeMB2U(0xDF24, 0x008C92);
	ChangeMB2U(0xDF25, 0x008C93);
	ChangeMB2U(0xDF26, 0x008C99);
	ChangeMB2U(0xDF27, 0x008C9B);
	ChangeMB2U(0xDF28, 0x008C9C);
	ChangeMB2U(0xDF29, 0x008CA4);
	ChangeMB2U(0xDF2A, 0x008CB9);
	ChangeMB2U(0xDF2B, 0x008CBA);
	ChangeMB2U(0xDF2C, 0x008CC5);
	ChangeMB2U(0xDF2D, 0x008CC6);
	ChangeMB2U(0xDF2E, 0x008CC9);
	ChangeMB2U(0xDF2F, 0x008CCB);
	ChangeMB2U(0xDF30, 0x008CCF);
	ChangeMB2U(0xDF31, 0x008CD6);
	ChangeMB2U(0xDF32, 0x008CD5);
	ChangeMB2U(0xDF33, 0x008CD9);
	ChangeMB2U(0xDF34, 0x008CDD);
	ChangeMB2U(0xDF35, 0x008CE1);
	ChangeMB2U(0xDF36, 0x008CE8);
	ChangeMB2U(0xDF37, 0x008CEC);
	ChangeMB2U(0xDF38, 0x008CEF);
	ChangeMB2U(0xDF3A, 0x008CF2);
	ChangeMB2U(0xDF3B, 0x008CF5);
	ChangeMB2U(0xDF3C, 0x008CF7);
	ChangeMB2U(0xDF3D, 0x008CF8);
	ChangeMB2U(0xDF3E, 0x008CFE);
	ChangeMB2U(0xDF3F, 0x008CFF);
	ChangeMB2U(0xDF40, 0x008D01);
	ChangeMB2U(0xDF41, 0x008D03);
	ChangeMB2U(0xDF42, 0x008D09);
	ChangeMB2U(0xDF44, 0x008D17);
	ChangeMB2U(0xDF45, 0x008D1B);
	ChangeMB2U(0xDF46, 0x008D65);
	ChangeMB2U(0xDF47, 0x008D69);
	ChangeMB2U(0xDF48, 0x008D6C);
	ChangeMB2U(0xDF49, 0x008D6E);
	ChangeMB2U(0xDF4A, 0x008D7F);
	ChangeMB2U(0xDF4B, 0x008D82);
	ChangeMB2U(0xDF4C, 0x008D84);
	ChangeMB2U(0xDF4D, 0x008D88);
	ChangeMB2U(0xDF4E, 0x008D8D);
	ChangeMB2U(0xDF4F, 0x008D90);
	ChangeMB2U(0xDF50, 0x008D91);
	ChangeMB2U(0xDF51, 0x008D95);
	ChangeMB2U(0xDF52, 0x008D9E);
	ChangeMB2U(0xDF53, 0x008D9F);
	ChangeMB2U(0xDF54, 0x008DA0);
	ChangeMB2U(0xDF55, 0x008DA6);
	ChangeMB2U(0xDF56, 0x008DAB);
	ChangeMB2U(0xDF57, 0x008DAC);
	ChangeMB2U(0xDF58, 0x008DAF);
	ChangeMB2U(0xDF59, 0x008DB2);
	ChangeMB2U(0xDF5A, 0x008DB5);
	ChangeMB2U(0xDF5B, 0x008DB7);
	ChangeMB2U(0xDF5C, 0x008DB9);
	ChangeMB2U(0xDF5D, 0x008DBB);
	ChangeMB2U(0xDF5E, 0x008DC0);
	ChangeMB2U(0xDF5F, 0x008DC5);
	ChangeMB2U(0xDF60, 0x008DC6);
	ChangeMB2U(0xDF61, 0x008DC7);
	ChangeMB2U(0xDF62, 0x008DC8);
	ChangeMB2U(0xDF63, 0x008DCA);
	ChangeMB2U(0xDF64, 0x008DCE);
	ChangeMB2U(0xDF65, 0x008DD1);
	ChangeMB2U(0xDF66, 0x008DD4);
	ChangeMB2U(0xDF67, 0x008DD5);
	ChangeMB2U(0xDF68, 0x008DD7);
	ChangeMB2U(0xDF69, 0x008DD9);
	ChangeMB2U(0xDF6A, 0x008DE4);
	ChangeMB2U(0xDF6B, 0x008DE5);
	ChangeMB2U(0xDF6C, 0x008DE7);
	ChangeMB2U(0xDF6D, 0x008DEC);
	ChangeMB2U(0xDF6E, 0x008DF0);
	ChangeMB2U(0xDF6F, 0x008DBC);
	ChangeMB2U(0xDF70, 0x008DF1);
	ChangeMB2U(0xDF71, 0x008DF2);
	ChangeMB2U(0xDF72, 0x008DF4);
	ChangeMB2U(0xDF73, 0x008DFD);
	ChangeMB2U(0xDF74, 0x008E01);
	ChangeMB2U(0xDF75, 0x008E04);
	ChangeMB2U(0xDF76, 0x008E05);
	ChangeMB2U(0xDF77, 0x008E06);
	ChangeMB2U(0xDF78, 0x008E0B);
	ChangeMB2U(0xDF79, 0x008E11);
	ChangeMB2U(0xDF7A, 0x008E14);
	ChangeMB2U(0xDF7B, 0x008E16);
	ChangeMB2U(0xDF7C, 0x008E20);
	ChangeMB2U(0xDF7D, 0x008E21);
	ChangeMB2U(0xDF7E, 0x008E22);
	ChangeMB2U(0xE021, 0x008E23);
	ChangeMB2U(0xE022, 0x008E26);
	ChangeMB2U(0xE023, 0x008E27);
	ChangeMB2U(0xE024, 0x008E31);
	ChangeMB2U(0xE025, 0x008E33);
	ChangeMB2U(0xE026, 0x008E36);
	ChangeMB2U(0xE027, 0x008E37);
	ChangeMB2U(0xE028, 0x008E38);
	ChangeMB2U(0xE029, 0x008E39);
	ChangeMB2U(0xE02A, 0x008E3D);
	ChangeMB2U(0xE02B, 0x008E40);
	ChangeMB2U(0xE02C, 0x008E41);
	ChangeMB2U(0xE02D, 0x008E4B);
	ChangeMB2U(0xE02E, 0x008E4D);
	ChangeMB2U(0xE02F, 0x008E4E);
	ChangeMB2U(0xE030, 0x008E4F);
	ChangeMB2U(0xE031, 0x008E54);
	ChangeMB2U(0xE032, 0x008E5B);
	ChangeMB2U(0xE033, 0x008E5C);
	ChangeMB2U(0xE034, 0x008E5D);
	ChangeMB2U(0xE035, 0x008E5E);
	ChangeMB2U(0xE036, 0x008E61);
	ChangeMB2U(0xE037, 0x008E62);
	ChangeMB2U(0xE038, 0x008E69);
	ChangeMB2U(0xE039, 0x008E6C);
	ChangeMB2U(0xE03A, 0x008E6D);
	ChangeMB2U(0xE03B, 0x008E6F);
	ChangeMB2U(0xE03C, 0x008E70);
	ChangeMB2U(0xE03D, 0x008E71);
	ChangeMB2U(0xE03E, 0x008E79);
	ChangeMB2U(0xE03F, 0x008E7A);
	ChangeMB2U(0xE040, 0x008E7B);
	ChangeMB2U(0xE041, 0x008E82);
	ChangeMB2U(0xE042, 0x008E83);
	ChangeMB2U(0xE043, 0x008E89);
	ChangeMB2U(0xE044, 0x008E90);
	ChangeMB2U(0xE045, 0x008E92);
	ChangeMB2U(0xE046, 0x008E95);
	ChangeMB2U(0xE047, 0x008E9A);
	ChangeMB2U(0xE048, 0x008E9B);
	ChangeMB2U(0xE049, 0x008E9D);
	ChangeMB2U(0xE04A, 0x008E9E);
	ChangeMB2U(0xE04B, 0x008EA2);
	ChangeMB2U(0xE04C, 0x008EA7);
	ChangeMB2U(0xE04D, 0x008EA9);
	ChangeMB2U(0xE04E, 0x008EAD);
	ChangeMB2U(0xE04F, 0x008EAE);
	ChangeMB2U(0xE050, 0x008EB3);
	ChangeMB2U(0xE051, 0x008EB5);
	ChangeMB2U(0xE052, 0x008EBA);
	ChangeMB2U(0xE053, 0x008EBB);
	ChangeMB2U(0xE054, 0x008EC0);
	ChangeMB2U(0xE055, 0x008EC1);
	ChangeMB2U(0xE056, 0x008EC3);
	ChangeMB2U(0xE057, 0x008EC4);
	ChangeMB2U(0xE058, 0x008EC7);
	ChangeMB2U(0xE05A, 0x008ED1);
	ChangeMB2U(0xE05B, 0x008ED4);
	ChangeMB2U(0xE05C, 0x008EDC);
	ChangeMB2U(0xE05D, 0x008EE8);
	ChangeMB2U(0xE05E, 0x008EEE);
	ChangeMB2U(0xE05F, 0x008EF0);
	ChangeMB2U(0xE060, 0x008EF1);
	ChangeMB2U(0xE061, 0x008EF7);
	ChangeMB2U(0xE062, 0x008EF9);
	ChangeMB2U(0xE063, 0x008EFA);
	ChangeMB2U(0xE064, 0x008EED);
	ChangeMB2U(0xE065, 0x008F00);
	ChangeMB2U(0xE066, 0x008F02);
	ChangeMB2U(0xE067, 0x008F07);
	ChangeMB2U(0xE068, 0x008F08);
	ChangeMB2U(0xE069, 0x008F0F);
	ChangeMB2U(0xE06A, 0x008F10);
	ChangeMB2U(0xE06B, 0x008F16);
	ChangeMB2U(0xE06C, 0x008F17);
	ChangeMB2U(0xE06D, 0x008F18);
	ChangeMB2U(0xE06E, 0x008F1E);
	ChangeMB2U(0xE06F, 0x008F20);
	ChangeMB2U(0xE070, 0x008F21);
	ChangeMB2U(0xE071, 0x008F23);
	ChangeMB2U(0xE072, 0x008F25);
	ChangeMB2U(0xE073, 0x008F27);
	ChangeMB2U(0xE074, 0x008F28);
	ChangeMB2U(0xE075, 0x008F2C);
	ChangeMB2U(0xE076, 0x008F2D);
	ChangeMB2U(0xE077, 0x008F2E);
	ChangeMB2U(0xE078, 0x008F34);
	ChangeMB2U(0xE079, 0x008F35);
	ChangeMB2U(0xE07A, 0x008F36);
	ChangeMB2U(0xE07B, 0x008F37);
	ChangeMB2U(0xE07C, 0x008F3A);
	ChangeMB2U(0xE07D, 0x008F40);
	ChangeMB2U(0xE07E, 0x008F41);
	ChangeMB2U(0xE121, 0x008F43);
	ChangeMB2U(0xE122, 0x008F47);
	ChangeMB2U(0xE123, 0x008F4F);
	ChangeMB2U(0xE124, 0x008F51);
	ChangeMB2U(0xE125, 0x008F52);
	ChangeMB2U(0xE126, 0x008F53);
	ChangeMB2U(0xE127, 0x008F54);
	ChangeMB2U(0xE128, 0x008F55);
	ChangeMB2U(0xE129, 0x008F58);
	ChangeMB2U(0xE12A, 0x008F5D);
	ChangeMB2U(0xE12B, 0x008F5E);
	ChangeMB2U(0xE12C, 0x008F65);
	ChangeMB2U(0xE12D, 0x008F9D);
	ChangeMB2U(0xE12E, 0x008FA0);
	ChangeMB2U(0xE12F, 0x008FA1);
	ChangeMB2U(0xE130, 0x008FA4);
	ChangeMB2U(0xE131, 0x008FA5);
	ChangeMB2U(0xE132, 0x008FA6);
	ChangeMB2U(0xE133, 0x008FB5);
	ChangeMB2U(0xE134, 0x008FB6);
	ChangeMB2U(0xE135, 0x008FB8);
	ChangeMB2U(0xE136, 0x008FBE);
	ChangeMB2U(0xE137, 0x008FC0);
	ChangeMB2U(0xE138, 0x008FC1);
	ChangeMB2U(0xE139, 0x008FC6);
	ChangeMB2U(0xE13A, 0x008FCA);
	ChangeMB2U(0xE13B, 0x008FCB);
	ChangeMB2U(0xE13C, 0x008FCD);
	ChangeMB2U(0xE13D, 0x008FD0);
	ChangeMB2U(0xE13E, 0x008FD2);
	ChangeMB2U(0xE13F, 0x008FD3);
	ChangeMB2U(0xE140, 0x008FD5);
	ChangeMB2U(0xE141, 0x008FE0);
	ChangeMB2U(0xE142, 0x008FE3);
	ChangeMB2U(0xE143, 0x008FE4);
	ChangeMB2U(0xE144, 0x008FE8);
	ChangeMB2U(0xE145, 0x008FEE);
	ChangeMB2U(0xE146, 0x008FF1);
	ChangeMB2U(0xE147, 0x008FF5);
	ChangeMB2U(0xE148, 0x008FF6);
	ChangeMB2U(0xE149, 0x008FFB);
	ChangeMB2U(0xE14A, 0x008FFE);
	ChangeMB2U(0xE14B, 0x009002);
	ChangeMB2U(0xE14C, 0x009004);
	ChangeMB2U(0xE14D, 0x009008);
	ChangeMB2U(0xE14E, 0x00900C);
	ChangeMB2U(0xE14F, 0x009018);
	ChangeMB2U(0xE150, 0x00901B);
	ChangeMB2U(0xE151, 0x009028);
	ChangeMB2U(0xE152, 0x009029);
	ChangeMB2U(0xE153, 0x00902F);
	ChangeMB2U(0xE154, 0x00902A);
	ChangeMB2U(0xE155, 0x00902C);
	ChangeMB2U(0xE156, 0x00902D);
	ChangeMB2U(0xE157, 0x009033);
	ChangeMB2U(0xE158, 0x009034);
	ChangeMB2U(0xE159, 0x009037);
	ChangeMB2U(0xE15A, 0x00903F);
	ChangeMB2U(0xE15B, 0x009043);
	ChangeMB2U(0xE15C, 0x009044);
	ChangeMB2U(0xE15D, 0x00904C);
	ChangeMB2U(0xE15E, 0x00905B);
	ChangeMB2U(0xE15F, 0x00905D);
	ChangeMB2U(0xE160, 0x009062);
	ChangeMB2U(0xE161, 0x009066);
	ChangeMB2U(0xE163, 0x00906C);
	ChangeMB2U(0xE164, 0x009070);
	ChangeMB2U(0xE165, 0x009074);
	ChangeMB2U(0xE166, 0x009079);
	ChangeMB2U(0xE167, 0x009085);
	ChangeMB2U(0xE168, 0x009088);
	ChangeMB2U(0xE169, 0x00908B);
	ChangeMB2U(0xE16A, 0x00908C);
	ChangeMB2U(0xE16B, 0x00908E);
	ChangeMB2U(0xE16C, 0x009090);
	ChangeMB2U(0xE16D, 0x009095);
	ChangeMB2U(0xE16E, 0x009097);
	ChangeMB2U(0xE16F, 0x009098);
	ChangeMB2U(0xE170, 0x009099);
	ChangeMB2U(0xE171, 0x00909B);
	ChangeMB2U(0xE172, 0x0090A0);
	ChangeMB2U(0xE173, 0x0090A1);
	ChangeMB2U(0xE174, 0x0090A2);
	ChangeMB2U(0xE175, 0x0090A5);
	ChangeMB2U(0xE176, 0x0090B0);
	ChangeMB2U(0xE177, 0x0090B2);
	ChangeMB2U(0xE178, 0x0090B3);
	ChangeMB2U(0xE179, 0x0090B4);
	ChangeMB2U(0xE17A, 0x0090B6);
	ChangeMB2U(0xE17B, 0x0090BD);
	ChangeMB2U(0xE17C, 0x0090CC);
	ChangeMB2U(0xE17D, 0x0090BE);
	ChangeMB2U(0xE17E, 0x0090C3);
	ChangeMB2U(0xE221, 0x0090C4);
	ChangeMB2U(0xE222, 0x0090C5);
	ChangeMB2U(0xE223, 0x0090C7);
	ChangeMB2U(0xE224, 0x0090C8);
	ChangeMB2U(0xE225, 0x0090D5);
	ChangeMB2U(0xE226, 0x0090D7);
	ChangeMB2U(0xE227, 0x0090D8);
	ChangeMB2U(0xE228, 0x0090D9);
	ChangeMB2U(0xE229, 0x0090DC);
	ChangeMB2U(0xE22A, 0x0090DD);
	ChangeMB2U(0xE22B, 0x0090DF);
	ChangeMB2U(0xE22C, 0x0090E5);
	ChangeMB2U(0xE22D, 0x0090D2);
	ChangeMB2U(0xE22E, 0x0090F6);
	ChangeMB2U(0xE22F, 0x0090EB);
	ChangeMB2U(0xE230, 0x0090EF);
	ChangeMB2U(0xE231, 0x0090F0);
	ChangeMB2U(0xE232, 0x0090F4);
	ChangeMB2U(0xE233, 0x0090FE);
	ChangeMB2U(0xE234, 0x0090FF);
	ChangeMB2U(0xE235, 0x009100);
	ChangeMB2U(0xE236, 0x009104);
	ChangeMB2U(0xE237, 0x009105);
	ChangeMB2U(0xE238, 0x009106);
	ChangeMB2U(0xE239, 0x009108);
	ChangeMB2U(0xE23A, 0x00910D);
	ChangeMB2U(0xE23B, 0x009110);
	ChangeMB2U(0xE23C, 0x009114);
	ChangeMB2U(0xE23D, 0x009116);
	ChangeMB2U(0xE23E, 0x009117);
	ChangeMB2U(0xE23F, 0x009118);
	ChangeMB2U(0xE240, 0x00911A);
	ChangeMB2U(0xE241, 0x00911C);
	ChangeMB2U(0xE242, 0x00911E);
	ChangeMB2U(0xE243, 0x009120);
	ChangeMB2U(0xE244, 0x009125);
	ChangeMB2U(0xE245, 0x009122);
	ChangeMB2U(0xE246, 0x009123);
	ChangeMB2U(0xE248, 0x009129);
	ChangeMB2U(0xE249, 0x00912E);
	ChangeMB2U(0xE24A, 0x00912F);
	ChangeMB2U(0xE24B, 0x009131);
	ChangeMB2U(0xE24C, 0x009134);
	ChangeMB2U(0xE24D, 0x009136);
	ChangeMB2U(0xE24E, 0x009137);
	ChangeMB2U(0xE24F, 0x009139);
	ChangeMB2U(0xE250, 0x00913A);
	ChangeMB2U(0xE251, 0x00913C);
	ChangeMB2U(0xE252, 0x00913D);
	ChangeMB2U(0xE253, 0x009143);
	ChangeMB2U(0xE254, 0x009147);
	ChangeMB2U(0xE255, 0x009148);
	ChangeMB2U(0xE256, 0x00914F);
	ChangeMB2U(0xE257, 0x009153);
	ChangeMB2U(0xE258, 0x009157);
	ChangeMB2U(0xE259, 0x009159);
	ChangeMB2U(0xE25A, 0x00915A);
	ChangeMB2U(0xE25B, 0x00915B);
	ChangeMB2U(0xE25C, 0x009161);
	ChangeMB2U(0xE25D, 0x009164);
	ChangeMB2U(0xE25E, 0x009167);
	ChangeMB2U(0xE25F, 0x00916D);
	ChangeMB2U(0xE260, 0x009174);
	ChangeMB2U(0xE261, 0x009179);
	ChangeMB2U(0xE262, 0x00917A);
	ChangeMB2U(0xE263, 0x00917B);
	ChangeMB2U(0xE264, 0x009181);
	ChangeMB2U(0xE265, 0x009183);
	ChangeMB2U(0xE266, 0x009185);
	ChangeMB2U(0xE267, 0x009186);
	ChangeMB2U(0xE268, 0x00918A);
	ChangeMB2U(0xE269, 0x00918E);
	ChangeMB2U(0xE26A, 0x009191);
	ChangeMB2U(0xE26B, 0x009193);
	ChangeMB2U(0xE26C, 0x009194);
	ChangeMB2U(0xE26D, 0x009195);
	ChangeMB2U(0xE26E, 0x009198);
	ChangeMB2U(0xE26F, 0x00919E);
	ChangeMB2U(0xE270, 0x0091A1);
	ChangeMB2U(0xE271, 0x0091A6);
	ChangeMB2U(0xE272, 0x0091A8);
	ChangeMB2U(0xE273, 0x0091AC);
	ChangeMB2U(0xE274, 0x0091AD);
	ChangeMB2U(0xE275, 0x0091AE);
	ChangeMB2U(0xE276, 0x0091B0);
	ChangeMB2U(0xE277, 0x0091B1);
	ChangeMB2U(0xE278, 0x0091B2);
	ChangeMB2U(0xE279, 0x0091B3);
	ChangeMB2U(0xE27A, 0x0091B6);
	ChangeMB2U(0xE27B, 0x0091BB);
	ChangeMB2U(0xE27C, 0x0091BC);
	ChangeMB2U(0xE27D, 0x0091BD);
	ChangeMB2U(0xE27E, 0x0091BF);
	ChangeMB2U(0xE321, 0x0091C2);
	ChangeMB2U(0xE322, 0x0091C3);
	ChangeMB2U(0xE323, 0x0091C5);
	ChangeMB2U(0xE324, 0x0091D3);
	ChangeMB2U(0xE325, 0x0091D4);
	ChangeMB2U(0xE327, 0x0091D9);
	ChangeMB2U(0xE32C, 0x0091E9);
	ChangeMB2U(0xE32D, 0x0091EA);
	ChangeMB2U(0xE32E, 0x0091EC);
	ChangeMB2U(0xE331, 0x0091EF);
	ChangeMB2U(0xE332, 0x0091F0);
	ChangeMB2U(0xE333, 0x0091F1);
	ChangeMB2U(0xE334, 0x0091F7);
	ChangeMB2U(0xE335, 0x0091F9);
	ChangeMB2U(0xE336, 0x0091FB);
	ChangeMB2U(0xE337, 0x0091FD);
	ChangeMB2U(0xE338, 0x009200);
	ChangeMB2U(0xE339, 0x009201);
	ChangeMB2U(0xE33A, 0x009204);
	ChangeMB2U(0xE33B, 0x009205);
	ChangeMB2U(0xE33D, 0x009207);
	ChangeMB2U(0xE33E, 0x009209);
	ChangeMB2U(0xE340, 0x00920C);
	ChangeMB2U(0xE342, 0x009212);
	ChangeMB2U(0xE343, 0x009213);
	ChangeMB2U(0xE344, 0x009216);
	ChangeMB2U(0xE345, 0x009218);
	ChangeMB2U(0xE346, 0x00921C);
	ChangeMB2U(0xE347, 0x00921D);
	ChangeMB2U(0xE348, 0x009223);
	ChangeMB2U(0xE349, 0x009224);
	ChangeMB2U(0xE34A, 0x009225);
	ChangeMB2U(0xE34B, 0x009226);
	ChangeMB2U(0xE34C, 0x009228);
	ChangeMB2U(0xE34D, 0x00922E);
	ChangeMB2U(0xE34E, 0x00922F);
	ChangeMB2U(0xE34F, 0x009230);
	ChangeMB2U(0xE350, 0x009233);
	ChangeMB2U(0xE351, 0x009235);
	ChangeMB2U(0xE352, 0x009236);
	ChangeMB2U(0xE353, 0x009238);
	ChangeMB2U(0xE357, 0x00923E);
	ChangeMB2U(0xE359, 0x009242);
	ChangeMB2U(0xE35A, 0x009243);
	ChangeMB2U(0xE35B, 0x009246);
	ChangeMB2U(0xE35C, 0x009247);
	ChangeMB2U(0xE35D, 0x00924A);
	ChangeMB2U(0xE35E, 0x00924D);
	ChangeMB2U(0xE360, 0x00924F);
	ChangeMB2U(0xE362, 0x009258);
	ChangeMB2U(0xE364, 0x00925C);
	ChangeMB2U(0xE365, 0x00925D);
	ChangeMB2U(0xE366, 0x009260);
	ChangeMB2U(0xE367, 0x009261);
	ChangeMB2U(0xE368, 0x009265);
	ChangeMB2U(0xE36A, 0x009268);
	ChangeMB2U(0xE36B, 0x009269);
	ChangeMB2U(0xE36C, 0x00926E);
	ChangeMB2U(0xE36D, 0x00926F);
	ChangeMB2U(0xE36E, 0x009270);
	ChangeMB2U(0xE36F, 0x009275);
	ChangeMB2U(0xE370, 0x009276);
	ChangeMB2U(0xE373, 0x009279);
	ChangeMB2U(0xE374, 0x00927B);
	ChangeMB2U(0xE375, 0x00927C);
	ChangeMB2U(0xE376, 0x00927D);
	ChangeMB2U(0xE377, 0x00927F);
	ChangeMB2U(0xE379, 0x009289);
	ChangeMB2U(0xE37A, 0x00928A);
	ChangeMB2U(0xE37B, 0x00928D);
	ChangeMB2U(0xE37C, 0x00928E);
	ChangeMB2U(0xE37D, 0x009292);
	ChangeMB2U(0xE37E, 0x009297);
	ChangeMB2U(0xE421, 0x009299);
	ChangeMB2U(0xE422, 0x00929F);
	ChangeMB2U(0xE423, 0x0092A0);
	ChangeMB2U(0xE424, 0x0092A4);
	ChangeMB2U(0xE425, 0x0092A5);
	ChangeMB2U(0xE427, 0x0092A8);
	ChangeMB2U(0xE428, 0x0092AB);
	ChangeMB2U(0xE429, 0x0092AF);
	ChangeMB2U(0xE42A, 0x0092B2);
	ChangeMB2U(0xE42B, 0x0092B6);
	ChangeMB2U(0xE42C, 0x0092B8);
	ChangeMB2U(0xE42D, 0x0092BA);
	ChangeMB2U(0xE42E, 0x0092BB);
	ChangeMB2U(0xE42F, 0x0092BC);
	ChangeMB2U(0xE430, 0x0092BD);
	ChangeMB2U(0xE431, 0x0092BF);
	ChangeMB2U(0xE432, 0x0092C0);
	ChangeMB2U(0xE433, 0x0092C1);
	ChangeMB2U(0xE434, 0x0092C2);
	ChangeMB2U(0xE435, 0x0092C3);
	ChangeMB2U(0xE436, 0x0092C5);
	ChangeMB2U(0xE437, 0x0092C6);
	ChangeMB2U(0xE438, 0x0092C7);
	ChangeMB2U(0xE439, 0x0092C8);
	ChangeMB2U(0xE43A, 0x0092CB);
	ChangeMB2U(0xE43B, 0x0092CC);
	ChangeMB2U(0xE43C, 0x0092CD);
	ChangeMB2U(0xE43D, 0x0092CE);
	ChangeMB2U(0xE442, 0x0092D8);
	ChangeMB2U(0xE444, 0x0092DC);
	ChangeMB2U(0xE445, 0x0092DD);
	ChangeMB2U(0xE446, 0x0092DF);
	ChangeMB2U(0xE448, 0x0092E1);
	ChangeMB2U(0xE449, 0x0092E3);
	ChangeMB2U(0xE44A, 0x0092E5);
	ChangeMB2U(0xE44C, 0x0092E8);
	ChangeMB2U(0xE44D, 0x0092EC);
	ChangeMB2U(0xE44E, 0x0092EE);
	ChangeMB2U(0xE44F, 0x0092F0);
	ChangeMB2U(0xE453, 0x009300);
	ChangeMB2U(0xE455, 0x009308);
	ChangeMB2U(0xE456, 0x00930D);
	ChangeMB2U(0xE457, 0x009311);
	ChangeMB2U(0xE458, 0x009314);
	ChangeMB2U(0xE459, 0x009315);
	ChangeMB2U(0xE45A, 0x00931C);
	ChangeMB2U(0xE45D, 0x00931F);
	ChangeMB2U(0xE45F, 0x009324);
	ChangeMB2U(0xE461, 0x009327);
	ChangeMB2U(0xE462, 0x009329);
	ChangeMB2U(0xE463, 0x00932A);
	ChangeMB2U(0xE464, 0x009333);
	ChangeMB2U(0xE465, 0x009334);
	ChangeMB2U(0xE466, 0x009336);
	ChangeMB2U(0xE467, 0x009337);
	ChangeMB2U(0xE468, 0x009347);
	ChangeMB2U(0xE46A, 0x009349);
	ChangeMB2U(0xE46B, 0x009350);
	ChangeMB2U(0xE46C, 0x009351);
	ChangeMB2U(0xE46D, 0x009352);
	ChangeMB2U(0xE46E, 0x009355);
	ChangeMB2U(0xE470, 0x009358);
	ChangeMB2U(0xE471, 0x00935A);
	ChangeMB2U(0xE472, 0x00935E);
	ChangeMB2U(0xE473, 0x009364);
	ChangeMB2U(0xE474, 0x009365);
	ChangeMB2U(0xE475, 0x009367);
	ChangeMB2U(0xE476, 0x009369);
	ChangeMB2U(0xE477, 0x00936A);
	ChangeMB2U(0xE478, 0x00936D);
	ChangeMB2U(0xE479, 0x00936F);
	ChangeMB2U(0xE47B, 0x009371);
	ChangeMB2U(0xE47C, 0x009373);
	ChangeMB2U(0xE47D, 0x009374);
	ChangeMB2U(0xE47E, 0x009376);
	ChangeMB2U(0xE521, 0x00937A);
	ChangeMB2U(0xE522, 0x00937D);
	ChangeMB2U(0xE523, 0x00937F);
	ChangeMB2U(0xE524, 0x009380);
	ChangeMB2U(0xE525, 0x009381);
	ChangeMB2U(0xE526, 0x009382);
	ChangeMB2U(0xE527, 0x009388);
	ChangeMB2U(0xE528, 0x00938A);
	ChangeMB2U(0xE529, 0x00938B);
	ChangeMB2U(0xE52A, 0x00938D);
	ChangeMB2U(0xE52B, 0x00938F);
	ChangeMB2U(0xE52C, 0x009392);
	ChangeMB2U(0xE52D, 0x009395);
	ChangeMB2U(0xE52E, 0x009398);
	ChangeMB2U(0xE52F, 0x00939B);
	ChangeMB2U(0xE530, 0x00939E);
	ChangeMB2U(0xE531, 0x0093A1);
	ChangeMB2U(0xE532, 0x0093A3);
	ChangeMB2U(0xE534, 0x0093A6);
	ChangeMB2U(0xE535, 0x0093A8);
	ChangeMB2U(0xE536, 0x0093AB);
	ChangeMB2U(0xE537, 0x0093B4);
	ChangeMB2U(0xE538, 0x0093B5);
	ChangeMB2U(0xE539, 0x0093B6);
	ChangeMB2U(0xE53A, 0x0093BA);
	ChangeMB2U(0xE53B, 0x0093A9);
	ChangeMB2U(0xE53C, 0x0093C1);
	ChangeMB2U(0xE53D, 0x0093C4);
	ChangeMB2U(0xE53E, 0x0093C5);
	ChangeMB2U(0xE540, 0x0093C7);
	ChangeMB2U(0xE541, 0x0093C9);
	ChangeMB2U(0xE542, 0x0093CA);
	ChangeMB2U(0xE543, 0x0093CB);
	ChangeMB2U(0xE544, 0x0093CC);
	ChangeMB2U(0xE545, 0x0093CD);
	ChangeMB2U(0xE546, 0x0093D3);
	ChangeMB2U(0xE547, 0x0093D9);
	ChangeMB2U(0xE548, 0x0093DC);
	ChangeMB2U(0xE54A, 0x0093DF);
	ChangeMB2U(0xE54B, 0x0093E2);
	ChangeMB2U(0xE54C, 0x0093E6);
	ChangeMB2U(0xE54D, 0x0093E7);
	ChangeMB2U(0xE54E, 0x0093F9);
	ChangeMB2U(0xE54F, 0x0093F7);
	ChangeMB2U(0xE551, 0x0093FA);
	ChangeMB2U(0xE552, 0x0093FB);
	ChangeMB2U(0xE553, 0x0093FD);
	ChangeMB2U(0xE554, 0x009401);
	ChangeMB2U(0xE555, 0x009402);
	ChangeMB2U(0xE556, 0x009404);
	ChangeMB2U(0xE557, 0x009408);
	ChangeMB2U(0xE558, 0x009409);
	ChangeMB2U(0xE559, 0x00940D);
	ChangeMB2U(0xE55A, 0x00940E);
	ChangeMB2U(0xE55B, 0x00940F);
	ChangeMB2U(0xE55C, 0x009415);
	ChangeMB2U(0xE55D, 0x009416);
	ChangeMB2U(0xE55E, 0x009417);
	ChangeMB2U(0xE55F, 0x00941F);
	ChangeMB2U(0xE560, 0x00942E);
	ChangeMB2U(0xE561, 0x00942F);
	ChangeMB2U(0xE563, 0x009432);
	ChangeMB2U(0xE564, 0x009433);
	ChangeMB2U(0xE565, 0x009434);
	ChangeMB2U(0xE566, 0x00943B);
	ChangeMB2U(0xE567, 0x00943F);
	ChangeMB2U(0xE568, 0x00943D);
	ChangeMB2U(0xE569, 0x009443);
	ChangeMB2U(0xE56C, 0x00944A);
	ChangeMB2U(0xE56D, 0x00944C);
	ChangeMB2U(0xE56E, 0x009455);
	ChangeMB2U(0xE56F, 0x009459);
	ChangeMB2U(0xE570, 0x00945C);
	ChangeMB2U(0xE571, 0x00945F);
	ChangeMB2U(0xE572, 0x009461);
	ChangeMB2U(0xE573, 0x009463);
	ChangeMB2U(0xE574, 0x009468);
	ChangeMB2U(0xE575, 0x00946B);
	ChangeMB2U(0xE576, 0x00946D);
	ChangeMB2U(0xE577, 0x00946E);
	ChangeMB2U(0xE578, 0x00946F);
	ChangeMB2U(0xE579, 0x009471);
	ChangeMB2U(0xE57A, 0x009472);
	ChangeMB2U(0xE57B, 0x009484);
	ChangeMB2U(0xE57C, 0x009483);
	ChangeMB2U(0xE57D, 0x009578);
	ChangeMB2U(0xE57E, 0x009579);
	ChangeMB2U(0xE621, 0x00957E);
	ChangeMB2U(0xE622, 0x009584);
	ChangeMB2U(0xE623, 0x009588);
	ChangeMB2U(0xE624, 0x00958C);
	ChangeMB2U(0xE625, 0x00958D);
	ChangeMB2U(0xE626, 0x00958E);
	ChangeMB2U(0xE627, 0x00959D);
	ChangeMB2U(0xE628, 0x00959E);
	ChangeMB2U(0xE629, 0x00959F);
	ChangeMB2U(0xE62A, 0x0095A1);
	ChangeMB2U(0xE62B, 0x0095A6);
	ChangeMB2U(0xE62C, 0x0095A9);
	ChangeMB2U(0xE62D, 0x0095AB);
	ChangeMB2U(0xE62E, 0x0095AC);
	ChangeMB2U(0xE62F, 0x0095B4);
	ChangeMB2U(0xE630, 0x0095B6);
	ChangeMB2U(0xE631, 0x0095BA);
	ChangeMB2U(0xE632, 0x0095BD);
	ChangeMB2U(0xE633, 0x0095BF);
	ChangeMB2U(0xE634, 0x0095C6);
	ChangeMB2U(0xE635, 0x0095C8);
	ChangeMB2U(0xE636, 0x0095C9);
	ChangeMB2U(0xE637, 0x0095CB);
	ChangeMB2U(0xE638, 0x0095D0);
	ChangeMB2U(0xE639, 0x0095D1);
	ChangeMB2U(0xE63A, 0x0095D2);
	ChangeMB2U(0xE63B, 0x0095D3);
	ChangeMB2U(0xE63C, 0x0095D9);
	ChangeMB2U(0xE63D, 0x0095DA);
	ChangeMB2U(0xE63E, 0x0095DD);
	ChangeMB2U(0xE63F, 0x0095DE);
	ChangeMB2U(0xE640, 0x0095DF);
	ChangeMB2U(0xE641, 0x0095E0);
	ChangeMB2U(0xE642, 0x0095E4);
	ChangeMB2U(0xE643, 0x0095E6);
	ChangeMB2U(0xE644, 0x00961D);
	ChangeMB2U(0xE645, 0x00961E);
	ChangeMB2U(0xE646, 0x009622);
	ChangeMB2U(0xE647, 0x009624);
	ChangeMB2U(0xE648, 0x009625);
	ChangeMB2U(0xE649, 0x009626);
	ChangeMB2U(0xE64A, 0x00962C);
	ChangeMB2U(0xE64B, 0x009631);
	ChangeMB2U(0xE64C, 0x009633);
	ChangeMB2U(0xE64D, 0x009637);
	ChangeMB2U(0xE64E, 0x009638);
	ChangeMB2U(0xE64F, 0x009639);
	ChangeMB2U(0xE650, 0x00963A);
	ChangeMB2U(0xE651, 0x00963C);
	ChangeMB2U(0xE652, 0x00963D);
	ChangeMB2U(0xE653, 0x009641);
	ChangeMB2U(0xE654, 0x009652);
	ChangeMB2U(0xE655, 0x009654);
	ChangeMB2U(0xE656, 0x009656);
	ChangeMB2U(0xE657, 0x009657);
	ChangeMB2U(0xE658, 0x009658);
	ChangeMB2U(0xE659, 0x009661);
	ChangeMB2U(0xE65A, 0x00966E);
	ChangeMB2U(0xE65B, 0x009674);
	ChangeMB2U(0xE65C, 0x00967B);
	ChangeMB2U(0xE65D, 0x00967C);
	ChangeMB2U(0xE65E, 0x00967E);
	ChangeMB2U(0xE65F, 0x00967F);
	ChangeMB2U(0xE660, 0x009681);
	ChangeMB2U(0xE661, 0x009682);
	ChangeMB2U(0xE662, 0x009683);
	ChangeMB2U(0xE663, 0x009684);
	ChangeMB2U(0xE664, 0x009689);
	ChangeMB2U(0xE665, 0x009691);
	ChangeMB2U(0xE666, 0x009696);
	ChangeMB2U(0xE667, 0x00969A);
	ChangeMB2U(0xE669, 0x00969F);
	ChangeMB2U(0xE66A, 0x0096A4);
	ChangeMB2U(0xE66B, 0x0096A5);
	ChangeMB2U(0xE66C, 0x0096A6);
	ChangeMB2U(0xE66D, 0x0096A9);
	ChangeMB2U(0xE66E, 0x0096AE);
	ChangeMB2U(0xE670, 0x0096B3);
	ChangeMB2U(0xE671, 0x0096BA);
	ChangeMB2U(0xE672, 0x0096CA);
	ChangeMB2U(0xE673, 0x0096D2);
	ChangeMB2U(0xE674, 0x005DB2);
	ChangeMB2U(0xE675, 0x0096D8);
	ChangeMB2U(0xE676, 0x0096DA);
	ChangeMB2U(0xE677, 0x0096DD);
	ChangeMB2U(0xE678, 0x0096DE);
	ChangeMB2U(0xE679, 0x0096DF);
	ChangeMB2U(0xE67A, 0x0096E9);
	ChangeMB2U(0xE67B, 0x0096EF);
	ChangeMB2U(0xE67C, 0x0096F1);
	ChangeMB2U(0xE67D, 0x0096FA);
	ChangeMB2U(0xE67E, 0x009702);
	ChangeMB2U(0xE721, 0x009703);
	ChangeMB2U(0xE722, 0x009705);
	ChangeMB2U(0xE723, 0x009709);
	ChangeMB2U(0xE724, 0x00971A);
	ChangeMB2U(0xE725, 0x00971B);
	ChangeMB2U(0xE726, 0x00971D);
	ChangeMB2U(0xE727, 0x009721);
	ChangeMB2U(0xE728, 0x009722);
	ChangeMB2U(0xE729, 0x009723);
	ChangeMB2U(0xE72A, 0x009728);
	ChangeMB2U(0xE72B, 0x009731);
	ChangeMB2U(0xE72D, 0x009741);
	ChangeMB2U(0xE72F, 0x00974A);
	ChangeMB2U(0xE730, 0x00974E);
	ChangeMB2U(0xE733, 0x009757);
	ChangeMB2U(0xE734, 0x009758);
	ChangeMB2U(0xE735, 0x00975A);
	ChangeMB2U(0xE736, 0x00975B);
	ChangeMB2U(0xE737, 0x009763);
	ChangeMB2U(0xE738, 0x009767);
	ChangeMB2U(0xE739, 0x00976A);
	ChangeMB2U(0xE73A, 0x00976E);
	ChangeMB2U(0xE73B, 0x009773);
	ChangeMB2U(0xE73C, 0x009776);
	ChangeMB2U(0xE73D, 0x009777);
	ChangeMB2U(0xE73E, 0x009778);
	ChangeMB2U(0xE73F, 0x00977B);
	ChangeMB2U(0xE740, 0x00977D);
	ChangeMB2U(0xE741, 0x00977F);
	ChangeMB2U(0xE742, 0x009780);
	ChangeMB2U(0xE743, 0x009789);
	ChangeMB2U(0xE744, 0x009795);
	ChangeMB2U(0xE745, 0x009796);
	ChangeMB2U(0xE746, 0x009797);
	ChangeMB2U(0xE747, 0x009799);
	ChangeMB2U(0xE748, 0x00979A);
	ChangeMB2U(0xE749, 0x00979E);
	ChangeMB2U(0xE74A, 0x00979F);
	ChangeMB2U(0xE74B, 0x0097A2);
	ChangeMB2U(0xE74C, 0x0097AC);
	ChangeMB2U(0xE74D, 0x0097AE);
	ChangeMB2U(0xE74E, 0x0097B1);
	ChangeMB2U(0xE74F, 0x0097B2);
	ChangeMB2U(0xE750, 0x0097B5);
	ChangeMB2U(0xE751, 0x0097B6);
	ChangeMB2U(0xE752, 0x0097B8);
	ChangeMB2U(0xE753, 0x0097B9);
	ChangeMB2U(0xE754, 0x0097BA);
	ChangeMB2U(0xE755, 0x0097BC);
	ChangeMB2U(0xE756, 0x0097BE);
	ChangeMB2U(0xE757, 0x0097BF);
	ChangeMB2U(0xE758, 0x0097C1);
	ChangeMB2U(0xE759, 0x0097C4);
	ChangeMB2U(0xE75A, 0x0097C5);
	ChangeMB2U(0xE75B, 0x0097C7);
	ChangeMB2U(0xE75C, 0x0097C9);
	ChangeMB2U(0xE75D, 0x0097CA);
	ChangeMB2U(0xE75E, 0x0097CC);
	ChangeMB2U(0xE75F, 0x0097CD);
	ChangeMB2U(0xE760, 0x0097CE);
	ChangeMB2U(0xE761, 0x0097D0);
	ChangeMB2U(0xE762, 0x0097D1);
	ChangeMB2U(0xE763, 0x0097D4);
	ChangeMB2U(0xE764, 0x0097D7);
	ChangeMB2U(0xE765, 0x0097D8);
	ChangeMB2U(0xE766, 0x0097D9);
	ChangeMB2U(0xE767, 0x0097DD);
	ChangeMB2U(0xE768, 0x0097DE);
	ChangeMB2U(0xE769, 0x0097E0);
	ChangeMB2U(0xE76A, 0x0097DB);
	ChangeMB2U(0xE76B, 0x0097E1);
	ChangeMB2U(0xE76C, 0x0097E4);
	ChangeMB2U(0xE76D, 0x0097EF);
	ChangeMB2U(0xE76E, 0x0097F1);
	ChangeMB2U(0xE76F, 0x0097F4);
	ChangeMB2U(0xE770, 0x0097F7);
	ChangeMB2U(0xE771, 0x0097F8);
	ChangeMB2U(0xE772, 0x0097FA);
	ChangeMB2U(0xE773, 0x009807);
	ChangeMB2U(0xE774, 0x00980A);
	ChangeMB2U(0xE775, 0x009819);
	ChangeMB2U(0xE776, 0x00980D);
	ChangeMB2U(0xE777, 0x00980E);
	ChangeMB2U(0xE778, 0x009814);
	ChangeMB2U(0xE779, 0x009816);
	ChangeMB2U(0xE77A, 0x00981C);
	ChangeMB2U(0xE77B, 0x00981E);
	ChangeMB2U(0xE77C, 0x009820);
	ChangeMB2U(0xE77D, 0x009823);
	ChangeMB2U(0xE77E, 0x009826);
	ChangeMB2U(0xE821, 0x00982B);
	ChangeMB2U(0xE822, 0x00982E);
	ChangeMB2U(0xE823, 0x00982F);
	ChangeMB2U(0xE824, 0x009830);
	ChangeMB2U(0xE825, 0x009832);
	ChangeMB2U(0xE826, 0x009833);
	ChangeMB2U(0xE827, 0x009835);
	ChangeMB2U(0xE828, 0x009825);
	ChangeMB2U(0xE829, 0x00983E);
	ChangeMB2U(0xE82A, 0x009844);
	ChangeMB2U(0xE82B, 0x009847);
	ChangeMB2U(0xE82C, 0x00984A);
	ChangeMB2U(0xE82D, 0x009851);
	ChangeMB2U(0xE82E, 0x009852);
	ChangeMB2U(0xE82F, 0x009853);
	ChangeMB2U(0xE830, 0x009856);
	ChangeMB2U(0xE832, 0x009859);
	ChangeMB2U(0xE833, 0x00985A);
	ChangeMB2U(0xE834, 0x009862);
	ChangeMB2U(0xE835, 0x009863);
	ChangeMB2U(0xE837, 0x009866);
	ChangeMB2U(0xE838, 0x00986A);
	ChangeMB2U(0xE839, 0x00986C);
	ChangeMB2U(0xE83A, 0x0098AB);
	ChangeMB2U(0xE83B, 0x0098AD);
	ChangeMB2U(0xE83C, 0x0098AE);
	ChangeMB2U(0xE83D, 0x0098B0);
	ChangeMB2U(0xE83E, 0x0098B4);
	ChangeMB2U(0xE83F, 0x0098B7);
	ChangeMB2U(0xE840, 0x0098B8);
	ChangeMB2U(0xE841, 0x0098BA);
	ChangeMB2U(0xE842, 0x0098BB);
	ChangeMB2U(0xE843, 0x0098BF);
	ChangeMB2U(0xE844, 0x0098C2);
	ChangeMB2U(0xE845, 0x0098C5);
	ChangeMB2U(0xE846, 0x0098C8);
	ChangeMB2U(0xE847, 0x0098CC);
	ChangeMB2U(0xE848, 0x0098E1);
	ChangeMB2U(0xE849, 0x0098E3);
	ChangeMB2U(0xE84A, 0x0098E5);
	ChangeMB2U(0xE84B, 0x0098E6);
	ChangeMB2U(0xE84C, 0x0098E7);
	ChangeMB2U(0xE84D, 0x0098EA);
	ChangeMB2U(0xE84E, 0x0098F3);
	ChangeMB2U(0xE84F, 0x0098F6);
	ChangeMB2U(0xE850, 0x009902);
	ChangeMB2U(0xE851, 0x009907);
	ChangeMB2U(0xE852, 0x009908);
	ChangeMB2U(0xE853, 0x009911);
	ChangeMB2U(0xE854, 0x009915);
	ChangeMB2U(0xE855, 0x009916);
	ChangeMB2U(0xE856, 0x009917);
	ChangeMB2U(0xE857, 0x00991A);
	ChangeMB2U(0xE858, 0x00991B);
	ChangeMB2U(0xE859, 0x00991C);
	ChangeMB2U(0xE85A, 0x00991F);
	ChangeMB2U(0xE85B, 0x009922);
	ChangeMB2U(0xE85C, 0x009926);
	ChangeMB2U(0xE85E, 0x00992B);
	ChangeMB2U(0xE85F, 0x009931);
	ChangeMB2U(0xE860, 0x009932);
	ChangeMB2U(0xE861, 0x009933);
	ChangeMB2U(0xE862, 0x009934);
	ChangeMB2U(0xE863, 0x009935);
	ChangeMB2U(0xE864, 0x009939);
	ChangeMB2U(0xE865, 0x00993A);
	ChangeMB2U(0xE866, 0x00993B);
	ChangeMB2U(0xE867, 0x00993C);
	ChangeMB2U(0xE868, 0x009940);
	ChangeMB2U(0xE869, 0x009941);
	ChangeMB2U(0xE86A, 0x009946);
	ChangeMB2U(0xE86B, 0x009947);
	ChangeMB2U(0xE86C, 0x009948);
	ChangeMB2U(0xE86D, 0x00994D);
	ChangeMB2U(0xE86E, 0x00994E);
	ChangeMB2U(0xE86F, 0x009954);
	ChangeMB2U(0xE870, 0x009958);
	ChangeMB2U(0xE871, 0x009959);
	ChangeMB2U(0xE872, 0x00995B);
	ChangeMB2U(0xE873, 0x00995C);
	ChangeMB2U(0xE874, 0x00995E);
	ChangeMB2U(0xE875, 0x00995F);
	ChangeMB2U(0xE876, 0x009960);
	ChangeMB2U(0xE877, 0x00999B);
	ChangeMB2U(0xE878, 0x00999D);
	ChangeMB2U(0xE879, 0x00999F);
	ChangeMB2U(0xE87A, 0x0099A6);
	ChangeMB2U(0xE87B, 0x0099B0);
	ChangeMB2U(0xE87C, 0x0099B1);
	ChangeMB2U(0xE87D, 0x0099B2);
	ChangeMB2U(0xE87E, 0x0099B5);
	ChangeMB2U(0xE921, 0x0099B9);
	ChangeMB2U(0xE922, 0x0099BA);
	ChangeMB2U(0xE923, 0x0099BD);
	ChangeMB2U(0xE924, 0x0099BF);
	ChangeMB2U(0xE925, 0x0099C3);
	ChangeMB2U(0xE926, 0x0099C9);
	ChangeMB2U(0xE927, 0x0099D3);
	ChangeMB2U(0xE928, 0x0099D4);
	ChangeMB2U(0xE929, 0x0099D9);
	ChangeMB2U(0xE92A, 0x0099DA);
	ChangeMB2U(0xE92B, 0x0099DC);
	ChangeMB2U(0xE92C, 0x0099DE);
	ChangeMB2U(0xE92D, 0x0099E7);
	ChangeMB2U(0xE92E, 0x0099EA);
	ChangeMB2U(0xE92F, 0x0099EB);
	ChangeMB2U(0xE930, 0x0099EC);
	ChangeMB2U(0xE931, 0x0099F0);
	ChangeMB2U(0xE932, 0x0099F4);
	ChangeMB2U(0xE933, 0x0099F5);
	ChangeMB2U(0xE934, 0x0099F9);
	ChangeMB2U(0xE935, 0x0099FD);
	ChangeMB2U(0xE936, 0x0099FE);
	ChangeMB2U(0xE937, 0x009A02);
	ChangeMB2U(0xE938, 0x009A03);
	ChangeMB2U(0xE939, 0x009A04);
	ChangeMB2U(0xE93A, 0x009A0B);
	ChangeMB2U(0xE93B, 0x009A0C);
	ChangeMB2U(0xE93C, 0x009A10);
	ChangeMB2U(0xE93D, 0x009A11);
	ChangeMB2U(0xE93E, 0x009A16);
	ChangeMB2U(0xE93F, 0x009A1E);
	ChangeMB2U(0xE940, 0x009A20);
	ChangeMB2U(0xE941, 0x009A22);
	ChangeMB2U(0xE942, 0x009A23);
	ChangeMB2U(0xE943, 0x009A24);
	ChangeMB2U(0xE944, 0x009A27);
	ChangeMB2U(0xE945, 0x009A2D);
	ChangeMB2U(0xE946, 0x009A2E);
	ChangeMB2U(0xE947, 0x009A33);
	ChangeMB2U(0xE948, 0x009A35);
	ChangeMB2U(0xE949, 0x009A36);
	ChangeMB2U(0xE94A, 0x009A38);
	ChangeMB2U(0xE94B, 0x009A47);
	ChangeMB2U(0xE94C, 0x009A41);
	ChangeMB2U(0xE94D, 0x009A44);
	ChangeMB2U(0xE94E, 0x009A4A);
	ChangeMB2U(0xE94F, 0x009A4B);
	ChangeMB2U(0xE950, 0x009A4C);
	ChangeMB2U(0xE952, 0x009A51);
	ChangeMB2U(0xE953, 0x009A54);
	ChangeMB2U(0xE954, 0x009A56);
	ChangeMB2U(0xE955, 0x009A5D);
	ChangeMB2U(0xE956, 0x009AAA);
	ChangeMB2U(0xE957, 0x009AAC);
	ChangeMB2U(0xE958, 0x009AAE);
	ChangeMB2U(0xE959, 0x009AAF);
	ChangeMB2U(0xE95A, 0x009AB2);
	ChangeMB2U(0xE95B, 0x009AB4);
	ChangeMB2U(0xE95C, 0x009AB5);
	ChangeMB2U(0xE95D, 0x009AB6);
	ChangeMB2U(0xE95E, 0x009AB9);
	ChangeMB2U(0xE95F, 0x009ABB);
	ChangeMB2U(0xE960, 0x009ABE);
	ChangeMB2U(0xE961, 0x009ABF);
	ChangeMB2U(0xE962, 0x009AC1);
	ChangeMB2U(0xE963, 0x009AC3);
	ChangeMB2U(0xE964, 0x009AC6);
	ChangeMB2U(0xE965, 0x009AC8);
	ChangeMB2U(0xE966, 0x009ACE);
	ChangeMB2U(0xE967, 0x009AD0);
	ChangeMB2U(0xE968, 0x009AD2);
	ChangeMB2U(0xE969, 0x009AD5);
	ChangeMB2U(0xE96A, 0x009AD6);
	ChangeMB2U(0xE96B, 0x009AD7);
	ChangeMB2U(0xE96C, 0x009ADB);
	ChangeMB2U(0xE96E, 0x009AE0);
	ChangeMB2U(0xE96F, 0x009AE4);
	ChangeMB2U(0xE970, 0x009AE5);
	ChangeMB2U(0xE971, 0x009AE7);
	ChangeMB2U(0xE972, 0x009AE9);
	ChangeMB2U(0xE973, 0x009AEC);
	ChangeMB2U(0xE974, 0x009AF2);
	ChangeMB2U(0xE975, 0x009AF3);
	ChangeMB2U(0xE976, 0x009AF5);
	ChangeMB2U(0xE977, 0x009AF9);
	ChangeMB2U(0xE978, 0x009AFA);
	ChangeMB2U(0xE979, 0x009AFD);
	ChangeMB2U(0xE97A, 0x009AFF);
	ChangeMB2U(0xE97B, 0x009B00);
	ChangeMB2U(0xE97C, 0x009B01);
	ChangeMB2U(0xE97D, 0x009B02);
	ChangeMB2U(0xE97E, 0x009B03);
	ChangeMB2U(0xEA21, 0x009B04);
	ChangeMB2U(0xEA22, 0x009B05);
	ChangeMB2U(0xEA23, 0x009B08);
	ChangeMB2U(0xEA24, 0x009B09);
	ChangeMB2U(0xEA25, 0x009B0B);
	ChangeMB2U(0xEA26, 0x009B0C);
	ChangeMB2U(0xEA27, 0x009B0D);
	ChangeMB2U(0xEA28, 0x009B0E);
	ChangeMB2U(0xEA29, 0x009B10);
	ChangeMB2U(0xEA2A, 0x009B12);
	ChangeMB2U(0xEA2B, 0x009B16);
	ChangeMB2U(0xEA2C, 0x009B19);
	ChangeMB2U(0xEA2D, 0x009B1B);
	ChangeMB2U(0xEA2E, 0x009B1C);
	ChangeMB2U(0xEA2F, 0x009B20);
	ChangeMB2U(0xEA30, 0x009B26);
	ChangeMB2U(0xEA31, 0x009B2B);
	ChangeMB2U(0xEA32, 0x009B2D);
	ChangeMB2U(0xEA33, 0x009B33);
	ChangeMB2U(0xEA34, 0x009B34);
	ChangeMB2U(0xEA35, 0x009B35);
	ChangeMB2U(0xEA36, 0x009B37);
	ChangeMB2U(0xEA37, 0x009B39);
	ChangeMB2U(0xEA38, 0x009B3A);
	ChangeMB2U(0xEA39, 0x009B3D);
	ChangeMB2U(0xEA3A, 0x009B48);
	ChangeMB2U(0xEA3B, 0x009B4B);
	ChangeMB2U(0xEA3C, 0x009B4C);
	ChangeMB2U(0xEA3D, 0x009B55);
	ChangeMB2U(0xEA3E, 0x009B56);
	ChangeMB2U(0xEA3F, 0x009B57);
	ChangeMB2U(0xEA40, 0x009B5B);
	ChangeMB2U(0xEA41, 0x009B5E);
	ChangeMB2U(0xEA42, 0x009B61);
	ChangeMB2U(0xEA43, 0x009B63);
	ChangeMB2U(0xEA44, 0x009B65);
	ChangeMB2U(0xEA45, 0x009B66);
	ChangeMB2U(0xEA46, 0x009B68);
	ChangeMB2U(0xEA47, 0x009B6A);
	ChangeMB2U(0xEA48, 0x009B6B);
	ChangeMB2U(0xEA49, 0x009B6C);
	ChangeMB2U(0xEA4A, 0x009B6D);
	ChangeMB2U(0xEA4B, 0x009B6E);
	ChangeMB2U(0xEA4C, 0x009B73);
	ChangeMB2U(0xEA4E, 0x009B77);
	ChangeMB2U(0xEA4F, 0x009B78);
	ChangeMB2U(0xEA50, 0x009B79);
	ChangeMB2U(0xEA51, 0x009B7F);
	ChangeMB2U(0xEA52, 0x009B80);
	ChangeMB2U(0xEA53, 0x009B84);
	ChangeMB2U(0xEA54, 0x009B85);
	ChangeMB2U(0xEA55, 0x009B86);
	ChangeMB2U(0xEA56, 0x009B87);
	ChangeMB2U(0xEA57, 0x009B89);
	ChangeMB2U(0xEA58, 0x009B8A);
	ChangeMB2U(0xEA59, 0x009B8B);
	ChangeMB2U(0xEA5A, 0x009B8D);
	ChangeMB2U(0xEA5C, 0x009B90);
	ChangeMB2U(0xEA5D, 0x009B94);
	ChangeMB2U(0xEA5E, 0x009B9A);
	ChangeMB2U(0xEA5F, 0x009B9D);
	ChangeMB2U(0xEA60, 0x009B9E);
	ChangeMB2U(0xEA61, 0x009BA6);
	ChangeMB2U(0xEA62, 0x009BA7);
	ChangeMB2U(0xEA63, 0x009BA9);
	ChangeMB2U(0xEA64, 0x009BAC);
	ChangeMB2U(0xEA65, 0x009BB0);
	ChangeMB2U(0xEA67, 0x009BB2);
	ChangeMB2U(0xEA68, 0x009BB7);
	ChangeMB2U(0xEA69, 0x009BB8);
	ChangeMB2U(0xEA6B, 0x009BBC);
	ChangeMB2U(0xEA6C, 0x009BBE);
	ChangeMB2U(0xEA6D, 0x009BBF);
	ChangeMB2U(0xEA6E, 0x009BC1);
	ChangeMB2U(0xEA6F, 0x009BC7);
	ChangeMB2U(0xEA70, 0x009BC8);
	ChangeMB2U(0xEA71, 0x009BCE);
	ChangeMB2U(0xEA72, 0x009BD0);
	ChangeMB2U(0xEA73, 0x009BD7);
	ChangeMB2U(0xEA74, 0x009BD8);
	ChangeMB2U(0xEA75, 0x009BDD);
	ChangeMB2U(0xEA76, 0x009BDF);
	ChangeMB2U(0xEA77, 0x009BE5);
	ChangeMB2U(0xEA78, 0x009BE7);
	ChangeMB2U(0xEA79, 0x009BEA);
	ChangeMB2U(0xEA7A, 0x009BEB);
	ChangeMB2U(0xEA7B, 0x009BEF);
	ChangeMB2U(0xEA7C, 0x009BF3);
	ChangeMB2U(0xEA7D, 0x009BF7);
	ChangeMB2U(0xEA7E, 0x009BF8);
	ChangeMB2U(0xEB21, 0x009BF9);
	ChangeMB2U(0xEB22, 0x009BFA);
	ChangeMB2U(0xEB23, 0x009BFD);
	ChangeMB2U(0xEB24, 0x009BFF);
	ChangeMB2U(0xEB26, 0x009C02);
	ChangeMB2U(0xEB27, 0x009C0B);
	ChangeMB2U(0xEB28, 0x009C0F);
	ChangeMB2U(0xEB29, 0x009C11);
	ChangeMB2U(0xEB2A, 0x009C16);
	ChangeMB2U(0xEB2B, 0x009C18);
	ChangeMB2U(0xEB2C, 0x009C19);
	ChangeMB2U(0xEB2D, 0x009C1A);
	ChangeMB2U(0xEB2E, 0x009C1C);
	ChangeMB2U(0xEB2F, 0x009C1E);
	ChangeMB2U(0xEB30, 0x009C22);
	ChangeMB2U(0xEB31, 0x009C23);
	ChangeMB2U(0xEB32, 0x009C26);
	ChangeMB2U(0xEB33, 0x009C27);
	ChangeMB2U(0xEB34, 0x009C28);
	ChangeMB2U(0xEB35, 0x009C29);
	ChangeMB2U(0xEB36, 0x009C2A);
	ChangeMB2U(0xEB37, 0x009C31);
	ChangeMB2U(0xEB38, 0x009C35);
	ChangeMB2U(0xEB39, 0x009C36);
	ChangeMB2U(0xEB3A, 0x009C37);
	ChangeMB2U(0xEB3B, 0x009C3D);
	ChangeMB2U(0xEB3C, 0x009C41);
	ChangeMB2U(0xEB3D, 0x009C43);
	ChangeMB2U(0xEB3E, 0x009C44);
	ChangeMB2U(0xEB3F, 0x009C45);
	ChangeMB2U(0xEB40, 0x009C49);
	ChangeMB2U(0xEB41, 0x009C4A);
	ChangeMB2U(0xEB42, 0x009C4E);
	ChangeMB2U(0xEB43, 0x009C4F);
	ChangeMB2U(0xEB44, 0x009C50);
	ChangeMB2U(0xEB45, 0x009C53);
	ChangeMB2U(0xEB46, 0x009C54);
	ChangeMB2U(0xEB47, 0x009C56);
	ChangeMB2U(0xEB48, 0x009C58);
	ChangeMB2U(0xEB49, 0x009C5B);
	ChangeMB2U(0xEB4A, 0x009C5D);
	ChangeMB2U(0xEB4B, 0x009C5E);
	ChangeMB2U(0xEB4C, 0x009C5F);
	ChangeMB2U(0xEB4D, 0x009C63);
	ChangeMB2U(0xEB4E, 0x009C69);
	ChangeMB2U(0xEB4F, 0x009C6A);
	ChangeMB2U(0xEB50, 0x009C5C);
	ChangeMB2U(0xEB51, 0x009C6B);
	ChangeMB2U(0xEB52, 0x009C68);
	ChangeMB2U(0xEB53, 0x009C6E);
	ChangeMB2U(0xEB54, 0x009C70);
	ChangeMB2U(0xEB55, 0x009C72);
	ChangeMB2U(0xEB56, 0x009C75);
	ChangeMB2U(0xEB57, 0x009C77);
	ChangeMB2U(0xEB58, 0x009C7B);
	ChangeMB2U(0xEB59, 0x009CE6);
	ChangeMB2U(0xEB5A, 0x009CF2);
	ChangeMB2U(0xEB5B, 0x009CF7);
	ChangeMB2U(0xEB5C, 0x009CF9);
	ChangeMB2U(0xEB5D, 0x009D0B);
	ChangeMB2U(0xEB5E, 0x009D02);
	ChangeMB2U(0xEB5F, 0x009D11);
	ChangeMB2U(0xEB60, 0x009D17);
	ChangeMB2U(0xEB61, 0x009D18);
	ChangeMB2U(0xEB62, 0x009D1C);
	ChangeMB2U(0xEB63, 0x009D1D);
	ChangeMB2U(0xEB64, 0x009D1E);
	ChangeMB2U(0xEB65, 0x009D2F);
	ChangeMB2U(0xEB66, 0x009D30);
	ChangeMB2U(0xEB67, 0x009D32);
	ChangeMB2U(0xEB68, 0x009D33);
	ChangeMB2U(0xEB69, 0x009D34);
	ChangeMB2U(0xEB6A, 0x009D3A);
	ChangeMB2U(0xEB6B, 0x009D3C);
	ChangeMB2U(0xEB6C, 0x009D45);
	ChangeMB2U(0xEB6D, 0x009D3D);
	ChangeMB2U(0xEB6E, 0x009D42);
	ChangeMB2U(0xEB6F, 0x009D43);
	ChangeMB2U(0xEB70, 0x009D47);
	ChangeMB2U(0xEB71, 0x009D4A);
	ChangeMB2U(0xEB72, 0x009D53);
	ChangeMB2U(0xEB73, 0x009D54);
	ChangeMB2U(0xEB74, 0x009D5F);
	ChangeMB2U(0xEB75, 0x009D63);
	ChangeMB2U(0xEB76, 0x009D62);
	ChangeMB2U(0xEB77, 0x009D65);
	ChangeMB2U(0xEB78, 0x009D69);
	ChangeMB2U(0xEB79, 0x009D6A);
	ChangeMB2U(0xEB7C, 0x009D76);
	ChangeMB2U(0xEB7D, 0x009D77);
	ChangeMB2U(0xEB7E, 0x009D7B);
	ChangeMB2U(0xEC21, 0x009D7C);
	ChangeMB2U(0xEC22, 0x009D7E);
	ChangeMB2U(0xEC23, 0x009D83);
	ChangeMB2U(0xEC24, 0x009D84);
	ChangeMB2U(0xEC25, 0x009D86);
	ChangeMB2U(0xEC26, 0x009D8A);
	ChangeMB2U(0xEC27, 0x009D8D);
	ChangeMB2U(0xEC28, 0x009D8E);
	ChangeMB2U(0xEC29, 0x009D92);
	ChangeMB2U(0xEC2A, 0x009D93);
	ChangeMB2U(0xEC2B, 0x009D95);
	ChangeMB2U(0xEC2C, 0x009D96);
	ChangeMB2U(0xEC2D, 0x009D97);
	ChangeMB2U(0xEC2E, 0x009D98);
	ChangeMB2U(0xEC2F, 0x009DA1);
	ChangeMB2U(0xEC30, 0x009DAA);
	ChangeMB2U(0xEC31, 0x009DAC);
	ChangeMB2U(0xEC32, 0x009DAE);
	ChangeMB2U(0xEC33, 0x009DB1);
	ChangeMB2U(0xEC34, 0x009DB5);
	ChangeMB2U(0xEC35, 0x009DB9);
	ChangeMB2U(0xEC36, 0x009DBC);
	ChangeMB2U(0xEC37, 0x009DBF);
	ChangeMB2U(0xEC38, 0x009DC3);
	ChangeMB2U(0xEC39, 0x009DC7);
	ChangeMB2U(0xEC3A, 0x009DC9);
	ChangeMB2U(0xEC3B, 0x009DCA);
	ChangeMB2U(0xEC3C, 0x009DD4);
	ChangeMB2U(0xEC3D, 0x009DD5);
	ChangeMB2U(0xEC3E, 0x009DD6);
	ChangeMB2U(0xEC3F, 0x009DD7);
	ChangeMB2U(0xEC40, 0x009DDA);
	ChangeMB2U(0xEC41, 0x009DDE);
	ChangeMB2U(0xEC42, 0x009DDF);
	ChangeMB2U(0xEC43, 0x009DE0);
	ChangeMB2U(0xEC44, 0x009DE5);
	ChangeMB2U(0xEC45, 0x009DE7);
	ChangeMB2U(0xEC46, 0x009DE9);
	ChangeMB2U(0xEC47, 0x009DEB);
	ChangeMB2U(0xEC48, 0x009DEE);
	ChangeMB2U(0xEC49, 0x009DF0);
	ChangeMB2U(0xEC4A, 0x009DF3);
	ChangeMB2U(0xEC4B, 0x009DF4);
	ChangeMB2U(0xEC4C, 0x009DFE);
	ChangeMB2U(0xEC4D, 0x009E0A);
	ChangeMB2U(0xEC4E, 0x009E02);
	ChangeMB2U(0xEC4F, 0x009E07);
	ChangeMB2U(0xEC50, 0x009E0E);
	ChangeMB2U(0xEC51, 0x009E10);
	ChangeMB2U(0xEC52, 0x009E11);
	ChangeMB2U(0xEC53, 0x009E12);
	ChangeMB2U(0xEC54, 0x009E15);
	ChangeMB2U(0xEC55, 0x009E16);
	ChangeMB2U(0xEC57, 0x009E1C);
	ChangeMB2U(0xEC58, 0x009E1D);
	ChangeMB2U(0xEC59, 0x009E7A);
	ChangeMB2U(0xEC5A, 0x009E7B);
	ChangeMB2U(0xEC5B, 0x009E7C);
	ChangeMB2U(0xEC5C, 0x009E80);
	ChangeMB2U(0xEC5D, 0x009E82);
	ChangeMB2U(0xEC5E, 0x009E83);
	ChangeMB2U(0xEC5F, 0x009E84);
	ChangeMB2U(0xEC60, 0x009E85);
	ChangeMB2U(0xEC61, 0x009E87);
	ChangeMB2U(0xEC62, 0x009E8E);
	ChangeMB2U(0xEC63, 0x009E8F);
	ChangeMB2U(0xEC64, 0x009E96);
	ChangeMB2U(0xEC65, 0x009E98);
	ChangeMB2U(0xEC66, 0x009E9B);
	ChangeMB2U(0xEC67, 0x009E9E);
	ChangeMB2U(0xEC68, 0x009EA4);
	ChangeMB2U(0xEC69, 0x009EA8);
	ChangeMB2U(0xEC6A, 0x009EAC);
	ChangeMB2U(0xEC6B, 0x009EAE);
	ChangeMB2U(0xEC6C, 0x009EAF);
	ChangeMB2U(0xEC6D, 0x009EB0);
	ChangeMB2U(0xEC6E, 0x009EB3);
	ChangeMB2U(0xEC6F, 0x009EB4);
	ChangeMB2U(0xEC70, 0x009EB5);
	ChangeMB2U(0xEC71, 0x009EC6);
	ChangeMB2U(0xEC72, 0x009EC8);
	ChangeMB2U(0xEC73, 0x009ECB);
	ChangeMB2U(0xEC74, 0x009ED5);
	ChangeMB2U(0xEC75, 0x009EDF);
	ChangeMB2U(0xEC76, 0x009EE4);
	ChangeMB2U(0xEC77, 0x009EE7);
	ChangeMB2U(0xEC78, 0x009EEC);
	ChangeMB2U(0xEC79, 0x009EED);
	ChangeMB2U(0xEC7A, 0x009EEE);
	ChangeMB2U(0xEC7B, 0x009EF0);
	ChangeMB2U(0xEC7C, 0x009EF1);
	ChangeMB2U(0xEC7D, 0x009EF2);
	ChangeMB2U(0xEC7E, 0x009EF5);
	ChangeMB2U(0xED21, 0x009EF8);
	ChangeMB2U(0xED22, 0x009EFF);
	ChangeMB2U(0xED23, 0x009F02);
	ChangeMB2U(0xED24, 0x009F03);
	ChangeMB2U(0xED25, 0x009F09);
	ChangeMB2U(0xED26, 0x009F0F);
	ChangeMB2U(0xED27, 0x009F10);
	ChangeMB2U(0xED28, 0x009F11);
	ChangeMB2U(0xED29, 0x009F12);
	ChangeMB2U(0xED2A, 0x009F14);
	ChangeMB2U(0xED2B, 0x009F16);
	ChangeMB2U(0xED2C, 0x009F17);
	ChangeMB2U(0xED2D, 0x009F19);
	ChangeMB2U(0xED2E, 0x009F1A);
	ChangeMB2U(0xED2F, 0x009F1B);
	ChangeMB2U(0xED30, 0x009F1F);
	ChangeMB2U(0xED31, 0x009F22);
	ChangeMB2U(0xED32, 0x009F26);
	ChangeMB2U(0xED33, 0x009F2A);
	ChangeMB2U(0xED34, 0x009F2B);
	ChangeMB2U(0xED35, 0x009F2F);
	ChangeMB2U(0xED36, 0x009F31);
	ChangeMB2U(0xED37, 0x009F32);
	ChangeMB2U(0xED38, 0x009F34);
	ChangeMB2U(0xED39, 0x009F37);
	ChangeMB2U(0xED3A, 0x009F39);
	ChangeMB2U(0xED3B, 0x009F3A);
	ChangeMB2U(0xED3C, 0x009F3C);
	ChangeMB2U(0xED3D, 0x009F3D);
	ChangeMB2U(0xED3E, 0x009F3F);
	ChangeMB2U(0xED3F, 0x009F41);
	ChangeMB2U(0xED40, 0x009F43);
	ChangeMB2U(0xED41, 0x009F44);
	ChangeMB2U(0xED42, 0x009F45);
	ChangeMB2U(0xED43, 0x009F46);
	ChangeMB2U(0xED44, 0x009F47);
	ChangeMB2U(0xED45, 0x009F53);
	ChangeMB2U(0xED46, 0x009F55);
	ChangeMB2U(0xED47, 0x009F56);
	ChangeMB2U(0xED48, 0x009F57);
	ChangeMB2U(0xED49, 0x009F58);
	ChangeMB2U(0xED4A, 0x009F5A);
	ChangeMB2U(0xED4B, 0x009F5D);
	ChangeMB2U(0xED4C, 0x009F5E);
	ChangeMB2U(0xED4D, 0x009F68);
	ChangeMB2U(0xED4E, 0x009F69);
	ChangeMB2U(0xED4F, 0x009F6D);
	ChangeMB2U(0xED50, 0x009F6E);
	ChangeMB2U(0xED51, 0x009F6F);
	ChangeMB2U(0xED52, 0x009F70);
	ChangeMB2U(0xED53, 0x009F71);
	ChangeMB2U(0xED54, 0x009F73);
	ChangeMB2U(0xED55, 0x009F75);
	ChangeMB2U(0xED56, 0x009F7A);
	ChangeMB2U(0xED57, 0x009F7D);
	ChangeMB2U(0xED58, 0x009F8F);
	ChangeMB2U(0xED59, 0x009F90);
	ChangeMB2U(0xED5A, 0x009F91);
	ChangeMB2U(0xED5B, 0x009F92);
	ChangeMB2U(0xED5C, 0x009F94);
	ChangeMB2U(0xED5D, 0x009F96);
	ChangeMB2U(0xED5E, 0x009F97);
	ChangeMB2U(0xED5F, 0x009F9E);
	ChangeMB2U(0xED60, 0x009FA1);
	ChangeMB2U(0xED61, 0x009FA2);
	ChangeMB2U(0xED62, 0x009FA3);
	ChangeMB2U(0xED63, 0x009FA5);

	RemoveU2MB(0x00F86F);

	ChangeU2MB(0x002116, 0xADE2);
	ChangeU2MB(0x002121, 0xADE4);
	ChangeU2MB(0x002160, 0xADB5);
	ChangeU2MB(0x002161, 0xADB6);
	ChangeU2MB(0x002162, 0xADB7);
	ChangeU2MB(0x002163, 0xADB8);
	ChangeU2MB(0x002164, 0xADB9);
	ChangeU2MB(0x002165, 0xADBA);
	ChangeU2MB(0x002166, 0xADBB);
	ChangeU2MB(0x002167, 0xADBC);
	ChangeU2MB(0x002168, 0xADBD);
	ChangeU2MB(0x002169, 0xADBE);
	ChangeU2MB(0x003231, 0xADEA);

	ChangeU2MB(0x0000A1, 0xA242);
	ChangeU2MB(0x0000A4, 0xA270);
	ChangeU2MB(0x0000A9, 0xA26D);
	ChangeU2MB(0x0000AA, 0xA26C);
	ChangeU2MB(0x0000AE, 0xA26E);
	ChangeU2MB(0x0000AF, 0xA234);
	ChangeU2MB(0x0000B8, 0xA231);
	ChangeU2MB(0x0000BA, 0xA26B);
	ChangeU2MB(0x0000BF, 0xA244);
	ChangeU2MB(0x0000C0, 0xAA22);
	ChangeU2MB(0x0000C1, 0xAA21);
	ChangeU2MB(0x0000C2, 0xAA24);
	ChangeU2MB(0x0000C3, 0xAA2A);
	ChangeU2MB(0x0000C4, 0xAA23);
	ChangeU2MB(0x0000C5, 0xAA29);
	ChangeU2MB(0x0000C6, 0xA921);
	ChangeU2MB(0x0000C7, 0xAA2E);
	ChangeU2MB(0x0000C8, 0xAA32);
	ChangeU2MB(0x0000C9, 0xAA31);
	ChangeU2MB(0x0000CA, 0xAA34);
	ChangeU2MB(0x0000CB, 0xAA33);
	ChangeU2MB(0x0000CC, 0xAA40);
	ChangeU2MB(0x0000CD, 0xAA3F);
	ChangeU2MB(0x0000CE, 0xAA42);
	ChangeU2MB(0x0000CF, 0xAA41);
	ChangeU2MB(0x0000D1, 0xAA50);
	ChangeU2MB(0x0000D2, 0xAA52);
	ChangeU2MB(0x0000D3, 0xAA51);
	ChangeU2MB(0x0000D4, 0xAA54);
	ChangeU2MB(0x0000D5, 0xAA58);
	ChangeU2MB(0x0000D6, 0xAA53);
	ChangeU2MB(0x0000D8, 0xA92C);
	ChangeU2MB(0x0000D9, 0xAA63);
	ChangeU2MB(0x0000DA, 0xAA62);
	ChangeU2MB(0x0000DB, 0xAA65);
	ChangeU2MB(0x0000DC, 0xAA64);
	ChangeU2MB(0x0000DD, 0xAA72);
	ChangeU2MB(0x0000DE, 0xA930);
	ChangeU2MB(0x0000DF, 0xA94E);
	ChangeU2MB(0x0000E0, 0xAB22);
	ChangeU2MB(0x0000E1, 0xAB21);
	ChangeU2MB(0x0000E2, 0xAB24);
	ChangeU2MB(0x0000E3, 0xAB2A);
	ChangeU2MB(0x0000E4, 0xAB23);
	ChangeU2MB(0x0000E5, 0xAB29);
	ChangeU2MB(0x0000E6, 0xA941);
	ChangeU2MB(0x0000E7, 0xAB2E);
	ChangeU2MB(0x0000E8, 0xAB32);
	ChangeU2MB(0x0000E9, 0xAB31);
	ChangeU2MB(0x0000EA, 0xAB34);
	ChangeU2MB(0x0000EB, 0xAB33);
	ChangeU2MB(0x0000EC, 0xAB40);
	ChangeU2MB(0x0000ED, 0xAB3F);
	ChangeU2MB(0x0000EE, 0xAB42);
	ChangeU2MB(0x0000EF, 0xAB41);
	ChangeU2MB(0x0000F0, 0xA943);
	ChangeU2MB(0x0000F1, 0xAB50);
	ChangeU2MB(0x0000F2, 0xAB52);
	ChangeU2MB(0x0000F3, 0xAB51);
	ChangeU2MB(0x0000F4, 0xAB54);
	ChangeU2MB(0x0000F5, 0xAB58);
	ChangeU2MB(0x0000F6, 0xAB53);
	ChangeU2MB(0x0000F8, 0xA94C);
	ChangeU2MB(0x0000F9, 0xAB63);
	ChangeU2MB(0x0000FA, 0xAB62);
	ChangeU2MB(0x0000FB, 0xAB65);
	ChangeU2MB(0x0000FC, 0xAB64);
	ChangeU2MB(0x0000FD, 0xAB72);
	ChangeU2MB(0x0000FE, 0xA950);
	ChangeU2MB(0x0000FF, 0xAB73);
	ChangeU2MB(0x000100, 0xAA27);
	ChangeU2MB(0x000101, 0xAB27);
	ChangeU2MB(0x000102, 0xAA25);
	ChangeU2MB(0x000103, 0xAB25);
	ChangeU2MB(0x000104, 0xAA28);
	ChangeU2MB(0x000105, 0xAB28);
	ChangeU2MB(0x000106, 0xAA2B);
	ChangeU2MB(0x000107, 0xAB2B);
	ChangeU2MB(0x000108, 0xAA2C);
	ChangeU2MB(0x000109, 0xAB2C);
	ChangeU2MB(0x00010A, 0xAA2F);
	ChangeU2MB(0x00010B, 0xAB2F);
	ChangeU2MB(0x00010C, 0xAA2D);
	ChangeU2MB(0x00010D, 0xAB2D);
	ChangeU2MB(0x00010E, 0xAA30);
	ChangeU2MB(0x00010F, 0xAB30);
	ChangeU2MB(0x000110, 0xA922);
	ChangeU2MB(0x000111, 0xA942);
	ChangeU2MB(0x000112, 0xAA37);
	ChangeU2MB(0x000113, 0xAB37);
	ChangeU2MB(0x000116, 0xAA36);
	ChangeU2MB(0x000117, 0xAB36);
	ChangeU2MB(0x000118, 0xAA38);
	ChangeU2MB(0x000119, 0xAB38);
	ChangeU2MB(0x00011A, 0xAA35);
	ChangeU2MB(0x00011B, 0xAB35);
	ChangeU2MB(0x00011C, 0xAA3A);
	ChangeU2MB(0x00011D, 0xAB3A);
	ChangeU2MB(0x00011E, 0xAA3B);
	ChangeU2MB(0x00011F, 0xAB3B);
	ChangeU2MB(0x000120, 0xAA3D);
	ChangeU2MB(0x000121, 0xAB3D);
	ChangeU2MB(0x000122, 0xAA3C);
	ChangeU2MB(0x000124, 0xAA3E);
	ChangeU2MB(0x000125, 0xAB3E);
	ChangeU2MB(0x000126, 0xA924);
	ChangeU2MB(0x000127, 0xA944);
	ChangeU2MB(0x000128, 0xAA47);
	ChangeU2MB(0x000129, 0xAB47);
	ChangeU2MB(0x00012A, 0xAA45);
	ChangeU2MB(0x00012B, 0xAB45);
	ChangeU2MB(0x00012E, 0xAA46);
	ChangeU2MB(0x00012F, 0xAB46);
	ChangeU2MB(0x000130, 0xAA44);
	ChangeU2MB(0x000131, 0xA945);
	ChangeU2MB(0x000132, 0xA926);
	ChangeU2MB(0x000133, 0xA946);
	ChangeU2MB(0x000134, 0xAA48);
	ChangeU2MB(0x000135, 0xAB48);
	ChangeU2MB(0x000136, 0xAA49);
	ChangeU2MB(0x000137, 0xAB49);
	ChangeU2MB(0x000138, 0xA947);
	ChangeU2MB(0x000139, 0xAA4A);
	ChangeU2MB(0x00013A, 0xAB4A);
	ChangeU2MB(0x00013B, 0xAA4C);
	ChangeU2MB(0x00013C, 0xAB4C);
	ChangeU2MB(0x00013D, 0xAA4B);
	ChangeU2MB(0x00013E, 0xAB4B);
	ChangeU2MB(0x00013F, 0xA929);
	ChangeU2MB(0x000140, 0xA949);
	ChangeU2MB(0x000141, 0xA928);
	ChangeU2MB(0x000142, 0xA948);
	ChangeU2MB(0x000143, 0xAA4D);
	ChangeU2MB(0x000144, 0xAB4D);
	ChangeU2MB(0x000145, 0xAA4F);
	ChangeU2MB(0x000146, 0xAB4F);
	ChangeU2MB(0x000147, 0xAA4E);
	ChangeU2MB(0x000148, 0xAB4E);
	ChangeU2MB(0x000149, 0xA94A);
	ChangeU2MB(0x00014A, 0xA92B);
	ChangeU2MB(0x00014B, 0xA94B);
	ChangeU2MB(0x00014C, 0xAA57);
	ChangeU2MB(0x00014D, 0xAB57);
	ChangeU2MB(0x000150, 0xAA56);
	ChangeU2MB(0x000151, 0xAB56);
	ChangeU2MB(0x000152, 0xA92D);
	ChangeU2MB(0x000153, 0xA94D);
	ChangeU2MB(0x000154, 0xAA59);
	ChangeU2MB(0x000155, 0xAB59);
	ChangeU2MB(0x000156, 0xAA5B);
	ChangeU2MB(0x000157, 0xAB5B);
	ChangeU2MB(0x000158, 0xAA5A);
	ChangeU2MB(0x000159, 0xAB5A);
	ChangeU2MB(0x00015A, 0xAA5C);
	ChangeU2MB(0x00015B, 0xAB5C);
	ChangeU2MB(0x00015C, 0xAA5D);
	ChangeU2MB(0x00015D, 0xAB5D);
	ChangeU2MB(0x00015E, 0xAA5F);
	ChangeU2MB(0x00015F, 0xAB5F);
	ChangeU2MB(0x000160, 0xAA5E);
	ChangeU2MB(0x000161, 0xAB5E);
	ChangeU2MB(0x000162, 0xAA61);
	ChangeU2MB(0x000163, 0xAB61);
	ChangeU2MB(0x000164, 0xAA60);
	ChangeU2MB(0x000165, 0xAB60);
	ChangeU2MB(0x000166, 0xA92F);
	ChangeU2MB(0x000167, 0xA94F);
	ChangeU2MB(0x000168, 0xAA6C);
	ChangeU2MB(0x000169, 0xAB6C);
	ChangeU2MB(0x00016A, 0xAA69);
	ChangeU2MB(0x00016B, 0xAB69);
	ChangeU2MB(0x00016C, 0xAA66);
	ChangeU2MB(0x00016D, 0xAB66);
	ChangeU2MB(0x00016E, 0xAA6B);
	ChangeU2MB(0x00016F, 0xAB6B);
	ChangeU2MB(0x000170, 0xAA68);
	ChangeU2MB(0x000171, 0xAB68);
	ChangeU2MB(0x000172, 0xAA6A);
	ChangeU2MB(0x000173, 0xAB6A);
	ChangeU2MB(0x000174, 0xAA71);
	ChangeU2MB(0x000175, 0xAB71);
	ChangeU2MB(0x000176, 0xAA74);
	ChangeU2MB(0x000177, 0xAB74);
	ChangeU2MB(0x000178, 0xAA73);
	ChangeU2MB(0x000179, 0xAA75);
	ChangeU2MB(0x00017A, 0xAB75);
	ChangeU2MB(0x00017B, 0xAA77);
	ChangeU2MB(0x00017C, 0xAB77);
	ChangeU2MB(0x00017D, 0xAA76);
	ChangeU2MB(0x00017E, 0xAB76);
	ChangeU2MB(0x0001CD, 0xAA26);
	ChangeU2MB(0x0001CE, 0xAB26);
	ChangeU2MB(0x0001CF, 0xAA43);
	ChangeU2MB(0x0001D0, 0xAB43);
	ChangeU2MB(0x0001D1, 0xAA55);
	ChangeU2MB(0x0001D2, 0xAB55);
	ChangeU2MB(0x0001D3, 0xAA67);
	ChangeU2MB(0x0001D4, 0xAB67);
	ChangeU2MB(0x0001D5, 0xAA70);
	ChangeU2MB(0x0001D6, 0xAB70);
	ChangeU2MB(0x0001D7, 0xAA6D);
	ChangeU2MB(0x0001D8, 0xAB6D);
	ChangeU2MB(0x0001D9, 0xAA6F);
	ChangeU2MB(0x0001DA, 0xAB6F);
	ChangeU2MB(0x0001DB, 0xAA6E);
	ChangeU2MB(0x0001DC, 0xAB6E);
	ChangeU2MB(0x0001F5, 0xAB39);
	ChangeU2MB(0x0002C7, 0xA230);
	ChangeU2MB(0x0002D8, 0xA22F);
	ChangeU2MB(0x0002D9, 0xA232);
	ChangeU2MB(0x0002DA, 0xA236);
	ChangeU2MB(0x0002DB, 0xA235);
	ChangeU2MB(0x0002DD, 0xA233);
	ChangeU2MB(0x000384, 0xA238);
	ChangeU2MB(0x000385, 0xA239);
	ChangeU2MB(0x000386, 0xA661);
	ChangeU2MB(0x000388, 0xA662);
	ChangeU2MB(0x000389, 0xA663);
	ChangeU2MB(0x00038A, 0xA664);
	ChangeU2MB(0x00038C, 0xA667);
	ChangeU2MB(0x00038E, 0xA669);
	ChangeU2MB(0x00038F, 0xA66C);
	ChangeU2MB(0x000390, 0xA676);
	ChangeU2MB(0x0003AA, 0xA665);
	ChangeU2MB(0x0003AB, 0xA66A);
	ChangeU2MB(0x0003AC, 0xA671);
	ChangeU2MB(0x0003AD, 0xA672);
	ChangeU2MB(0x0003AE, 0xA673);
	ChangeU2MB(0x0003AF, 0xA674);
	ChangeU2MB(0x0003B0, 0xA67B);
	ChangeU2MB(0x0003C2, 0xA678);
	ChangeU2MB(0x0003CA, 0xA675);
	ChangeU2MB(0x0003CB, 0xA67A);
	ChangeU2MB(0x0003CC, 0xA677);
	ChangeU2MB(0x0003CD, 0xA679);
	ChangeU2MB(0x0003CE, 0xA67C);
	ChangeU2MB(0x000402, 0xA742);
	ChangeU2MB(0x000403, 0xA743);
	ChangeU2MB(0x000404, 0xA744);
	ChangeU2MB(0x000405, 0xA745);
	ChangeU2MB(0x000406, 0xA746);
	ChangeU2MB(0x000407, 0xA747);
	ChangeU2MB(0x000408, 0xA748);
	ChangeU2MB(0x000409, 0xA749);
	ChangeU2MB(0x00040A, 0xA74A);
	ChangeU2MB(0x00040B, 0xA74B);
	ChangeU2MB(0x00040C, 0xA74C);
	ChangeU2MB(0x00040E, 0xA74D);
	ChangeU2MB(0x00040F, 0xA74E);
	ChangeU2MB(0x000452, 0xA772);
	ChangeU2MB(0x000453, 0xA773);
	ChangeU2MB(0x000454, 0xA774);
	ChangeU2MB(0x000455, 0xA775);
	ChangeU2MB(0x000456, 0xA776);
	ChangeU2MB(0x000457, 0xA777);
	ChangeU2MB(0x000458, 0xA778);
	ChangeU2MB(0x000459, 0xA779);
	ChangeU2MB(0x00045A, 0xA77A);
	ChangeU2MB(0x00045B, 0xA77B);
	ChangeU2MB(0x00045C, 0xA77C);
	ChangeU2MB(0x00045E, 0xA77D);
	ChangeU2MB(0x00045F, 0xA77E);
	ChangeU2MB(0x002122, 0xA26F);
	ChangeU2MB(0x002211, 0xADF4);
	ChangeU2MB(0x00221F, 0xADF8);
	ChangeU2MB(0x00222E, 0xADF3);
	ChangeU2MB(0x0022BF, 0xADF9);
	ChangeU2MB(0x002460, 0xADA1);
	ChangeU2MB(0x002461, 0xADA2);
	ChangeU2MB(0x002462, 0xADA3);
	ChangeU2MB(0x002463, 0xADA4);
	ChangeU2MB(0x002464, 0xADA5);
	ChangeU2MB(0x002465, 0xADA6);
	ChangeU2MB(0x002466, 0xADA7);
	ChangeU2MB(0x002467, 0xADA8);
	ChangeU2MB(0x002468, 0xADA9);
	ChangeU2MB(0x002469, 0xADAA);
	ChangeU2MB(0x00246A, 0xADAB);
	ChangeU2MB(0x00246B, 0xADAC);
	ChangeU2MB(0x00246C, 0xADAD);
	ChangeU2MB(0x00246D, 0xADAE);
	ChangeU2MB(0x00246E, 0xADAF);
	ChangeU2MB(0x00246F, 0xADB0);
	ChangeU2MB(0x002470, 0xADB1);
	ChangeU2MB(0x002471, 0xADB2);
	ChangeU2MB(0x002472, 0xADB3);
	ChangeU2MB(0x002473, 0xADB4);
	ChangeU2MB(0x00301F, 0xADE1);
	ChangeU2MB(0x003232, 0xADEB);
	ChangeU2MB(0x003239, 0xADEC);
	ChangeU2MB(0x0032A4, 0xADE5);
	ChangeU2MB(0x0032A5, 0xADE6);
	ChangeU2MB(0x0032A6, 0xADE7);
	ChangeU2MB(0x0032A7, 0xADE8);
	ChangeU2MB(0x0032A8, 0xADE9);
	ChangeU2MB(0x003303, 0xADC6);
	ChangeU2MB(0x00330D, 0xADCA);
	ChangeU2MB(0x003314, 0xADC1);
	ChangeU2MB(0x003318, 0xADC4);
	ChangeU2MB(0x003322, 0xADC2);
	ChangeU2MB(0x003323, 0xADCC);
	ChangeU2MB(0x003326, 0xADCB);
	ChangeU2MB(0x003327, 0xADC5);
	ChangeU2MB(0x00332B, 0xADCD);
	ChangeU2MB(0x003336, 0xADC7);
	ChangeU2MB(0x00333B, 0xADCF);
	ChangeU2MB(0x003349, 0xADC0);
	ChangeU2MB(0x00334A, 0xADCE);
	ChangeU2MB(0x00334D, 0xADC3);
	ChangeU2MB(0x003351, 0xADC8);
	ChangeU2MB(0x003357, 0xADC9);
	ChangeU2MB(0x00337B, 0xADDF);
	ChangeU2MB(0x00337C, 0xADEF);
	ChangeU2MB(0x00337D, 0xADEE);
	ChangeU2MB(0x00337E, 0xADED);
	ChangeU2MB(0x00338E, 0xADD3);
	ChangeU2MB(0x00338F, 0xADD4);
	ChangeU2MB(0x00339C, 0xADD0);
	ChangeU2MB(0x00339D, 0xADD1);
	ChangeU2MB(0x00339E, 0xADD2);
	ChangeU2MB(0x0033A1, 0xADD6);
	ChangeU2MB(0x0033C4, 0xADD5);
	ChangeU2MB(0x0033CD, 0xADE3);
	ChangeU2MB(0x004E02, 0xB021);
	ChangeU2MB(0x004E04, 0xB022);
	ChangeU2MB(0x004E05, 0xB023);
	ChangeU2MB(0x004E0C, 0xB024);
	ChangeU2MB(0x004E12, 0xB025);
	ChangeU2MB(0x004E1F, 0xB026);
	ChangeU2MB(0x004E23, 0xB027);
	ChangeU2MB(0x004E24, 0xB028);
	ChangeU2MB(0x004E2B, 0xB02A);
	ChangeU2MB(0x004E2E, 0xB02B);
	ChangeU2MB(0x004E2F, 0xB02C);
	ChangeU2MB(0x004E30, 0xB02D);
	ChangeU2MB(0x004E35, 0xB02E);
	ChangeU2MB(0x004E40, 0xB02F);
	ChangeU2MB(0x004E41, 0xB030);
	ChangeU2MB(0x004E44, 0xB031);
	ChangeU2MB(0x004E47, 0xB032);
	ChangeU2MB(0x004E51, 0xB033);
	ChangeU2MB(0x004E5A, 0xB034);
	ChangeU2MB(0x004E5C, 0xB035);
	ChangeU2MB(0x004E63, 0xB036);
	ChangeU2MB(0x004E68, 0xB037);
	ChangeU2MB(0x004E69, 0xB038);
	ChangeU2MB(0x004E74, 0xB039);
	ChangeU2MB(0x004E75, 0xB03A);
	ChangeU2MB(0x004E79, 0xB03B);
	ChangeU2MB(0x004E7F, 0xB03C);
	ChangeU2MB(0x004E8D, 0xB03D);
	ChangeU2MB(0x004E96, 0xB03E);
	ChangeU2MB(0x004E97, 0xB03F);
	ChangeU2MB(0x004E9D, 0xB040);
	ChangeU2MB(0x004EAF, 0xB041);
	ChangeU2MB(0x004EB9, 0xB042);
	ChangeU2MB(0x004EC3, 0xB043);
	ChangeU2MB(0x004ED0, 0xB044);
	ChangeU2MB(0x004EDA, 0xB045);
	ChangeU2MB(0x004EDB, 0xB046);
	ChangeU2MB(0x004EE0, 0xB047);
	ChangeU2MB(0x004EE2, 0xB049);
	ChangeU2MB(0x004EE8, 0xB04A);
	ChangeU2MB(0x004EEF, 0xB04B);
	ChangeU2MB(0x004EF1, 0xB04C);
	ChangeU2MB(0x004EF3, 0xB04D);
	ChangeU2MB(0x004EF5, 0xB04E);
	ChangeU2MB(0x004EFD, 0xB04F);
	ChangeU2MB(0x004EFE, 0xB050);
	ChangeU2MB(0x004EFF, 0xB051);
	ChangeU2MB(0x004F02, 0xB053);
	ChangeU2MB(0x004F08, 0xB055);
	ChangeU2MB(0x004F0B, 0xB056);
	ChangeU2MB(0x004F0C, 0xB057);
	ChangeU2MB(0x004F12, 0xB058);
	ChangeU2MB(0x004F15, 0xB059);
	ChangeU2MB(0x004F16, 0xB05A);
	ChangeU2MB(0x004F17, 0xB05B);
	ChangeU2MB(0x004F19, 0xB05C);
	ChangeU2MB(0x004F2E, 0xB05D);
	ChangeU2MB(0x004F31, 0xB05E);
	ChangeU2MB(0x004F33, 0xB060);
	ChangeU2MB(0x004F35, 0xB061);
	ChangeU2MB(0x004F37, 0xB062);
	ChangeU2MB(0x004F3B, 0xB064);
	ChangeU2MB(0x004F3E, 0xB065);
	ChangeU2MB(0x004F40, 0xB066);
	ChangeU2MB(0x004F42, 0xB067);
	ChangeU2MB(0x004F48, 0xB068);
	ChangeU2MB(0x004F49, 0xB069);
	ChangeU2MB(0x004F4B, 0xB06A);
	ChangeU2MB(0x004F4C, 0xB06B);
	ChangeU2MB(0x004F52, 0xB06C);
	ChangeU2MB(0x004F54, 0xB06D);
	ChangeU2MB(0x004F58, 0xB06F);
	ChangeU2MB(0x004F5F, 0xB070);
	ChangeU2MB(0x004F60, 0xB05F);
	ChangeU2MB(0x004F63, 0xB071);
	ChangeU2MB(0x004F6A, 0xB072);
	ChangeU2MB(0x004F6C, 0xB073);
	ChangeU2MB(0x004F6E, 0xB074);
	ChangeU2MB(0x004F71, 0xB075);
	ChangeU2MB(0x004F77, 0xB076);
	ChangeU2MB(0x004F78, 0xB077);
	ChangeU2MB(0x004F79, 0xB078);
	ChangeU2MB(0x004F7A, 0xB079);
	ChangeU2MB(0x004F7D, 0xB07A);
	ChangeU2MB(0x004F7E, 0xB07B);
	ChangeU2MB(0x004F81, 0xB07C);
	ChangeU2MB(0x004F82, 0xB07D);
	ChangeU2MB(0x004F84, 0xB07E);
	ChangeU2MB(0x004F85, 0xB121);
	ChangeU2MB(0x004F89, 0xB122);
	ChangeU2MB(0x004F8C, 0xB124);
	ChangeU2MB(0x004F8E, 0xB125);
	ChangeU2MB(0x004F90, 0xB126);
	ChangeU2MB(0x004F93, 0xB128);
	ChangeU2MB(0x004F97, 0xB12A);
	ChangeU2MB(0x004F99, 0xB12B);
	ChangeU2MB(0x004F9E, 0xB12D);
	ChangeU2MB(0x004F9F, 0xB12E);
	ChangeU2MB(0x004FB2, 0xB12F);
	ChangeU2MB(0x004FB7, 0xB130);
	ChangeU2MB(0x004FB9, 0xB131);
	ChangeU2MB(0x004FBB, 0xB132);
	ChangeU2MB(0x004FBC, 0xB133);
	ChangeU2MB(0x004FBD, 0xB134);
	ChangeU2MB(0x004FBE, 0xB135);
	ChangeU2MB(0x004FC0, 0xB136);
	ChangeU2MB(0x004FC1, 0xB137);
	ChangeU2MB(0x004FC5, 0xB138);
	ChangeU2MB(0x004FC6, 0xB139);
	ChangeU2MB(0x004FC8, 0xB13A);
	ChangeU2MB(0x004FCB, 0xB13C);
	ChangeU2MB(0x004FCC, 0xB13D);
	ChangeU2MB(0x004FCF, 0xB13F);
	ChangeU2MB(0x004FD2, 0xB140);
	ChangeU2MB(0x004FDC, 0xB141);
	ChangeU2MB(0x004FE0, 0xB142);
	ChangeU2MB(0x004FE2, 0xB143);
	ChangeU2MB(0x004FF0, 0xB144);
	ChangeU2MB(0x004FF2, 0xB145);
	ChangeU2MB(0x004FFC, 0xB146);
	ChangeU2MB(0x004FFD, 0xB147);
	ChangeU2MB(0x005000, 0xB149);
	ChangeU2MB(0x005001, 0xB14A);
	ChangeU2MB(0x005004, 0xB14B);
	ChangeU2MB(0x005007, 0xB14C);
	ChangeU2MB(0x00500A, 0xB14D);
	ChangeU2MB(0x00500C, 0xB14E);
	ChangeU2MB(0x00500E, 0xB14F);
	ChangeU2MB(0x005010, 0xB150);
	ChangeU2MB(0x005013, 0xB151);
	ChangeU2MB(0x005017, 0xB152);
	ChangeU2MB(0x005018, 0xB153);
	ChangeU2MB(0x00501B, 0xB154);
	ChangeU2MB(0x00501C, 0xB155);
	ChangeU2MB(0x00501D, 0xB156);
	ChangeU2MB(0x005027, 0xB159);
	ChangeU2MB(0x00502E, 0xB15A);
	ChangeU2MB(0x005030, 0xB15B);
	ChangeU2MB(0x005032, 0xB15C);
	ChangeU2MB(0x005033, 0xB15D);
	ChangeU2MB(0x005035, 0xB15E);
	ChangeU2MB(0x00503B, 0xB176);
	ChangeU2MB(0x005041, 0xB160);
	ChangeU2MB(0x005045, 0xB162);
	ChangeU2MB(0x00504A, 0xB164);
	ChangeU2MB(0x00504C, 0xB165);
	ChangeU2MB(0x00504E, 0xB166);
	ChangeU2MB(0x005051, 0xB167);
	ChangeU2MB(0x005052, 0xB168);
	ChangeU2MB(0x005053, 0xB169);
	ChangeU2MB(0x005057, 0xB16A);
	ChangeU2MB(0x005059, 0xB16B);
	ChangeU2MB(0x00505F, 0xB16C);
	ChangeU2MB(0x005060, 0xB16D);
	ChangeU2MB(0x005062, 0xB16E);
	ChangeU2MB(0x005063, 0xB16F);
	ChangeU2MB(0x005066, 0xB170);
	ChangeU2MB(0x005067, 0xB171);
	ChangeU2MB(0x00506A, 0xB172);
	ChangeU2MB(0x00506D, 0xB173);
	ChangeU2MB(0x005071, 0xB175);
	ChangeU2MB(0x005081, 0xB177);
	ChangeU2MB(0x005083, 0xB178);
	ChangeU2MB(0x005084, 0xB179);
	ChangeU2MB(0x005086, 0xB17A);
	ChangeU2MB(0x00508A, 0xB17B);
	ChangeU2MB(0x00508E, 0xB17C);
	ChangeU2MB(0x00508F, 0xB17D);
	ChangeU2MB(0x005090, 0xB17E);
	ChangeU2MB(0x005092, 0xB221);
	ChangeU2MB(0x005093, 0xB222);
	ChangeU2MB(0x005096, 0xB224);
	ChangeU2MB(0x00509B, 0xB225);
	ChangeU2MB(0x00509C, 0xB226);
	ChangeU2MB(0x00509E, 0xB227);
	ChangeU2MB(0x00509F, 0xB228);
	ChangeU2MB(0x0050A0, 0xB229);
	ChangeU2MB(0x0050A1, 0xB22A);
	ChangeU2MB(0x0050A2, 0xB22B);
	ChangeU2MB(0x0050AA, 0xB22C);
	ChangeU2MB(0x0050AF, 0xB22D);
	ChangeU2MB(0x0050B0, 0xB22E);
	ChangeU2MB(0x0050B9, 0xB22F);
	ChangeU2MB(0x0050BA, 0xB230);
	ChangeU2MB(0x0050BD, 0xB231);
	ChangeU2MB(0x0050C0, 0xB232);
	ChangeU2MB(0x0050C3, 0xB233);
	ChangeU2MB(0x0050C4, 0xB234);
	ChangeU2MB(0x0050C7, 0xB235);
	ChangeU2MB(0x0050CC, 0xB236);
	ChangeU2MB(0x0050CE, 0xB237);
	ChangeU2MB(0x0050D0, 0xB238);
	ChangeU2MB(0x0050D3, 0xB239);
	ChangeU2MB(0x0050D4, 0xB23A);
	ChangeU2MB(0x0050DC, 0xB23C);
	ChangeU2MB(0x0050DD, 0xB23D);
	ChangeU2MB(0x0050DF, 0xB23E);
	ChangeU2MB(0x0050E2, 0xB23F);
	ChangeU2MB(0x0050E4, 0xB240);
	ChangeU2MB(0x0050E6, 0xB241);
	ChangeU2MB(0x0050E8, 0xB242);
	ChangeU2MB(0x0050E9, 0xB243);
	ChangeU2MB(0x0050EF, 0xB244);
	ChangeU2MB(0x0050F1, 0xB245);
	ChangeU2MB(0x0050F2, 0xB251);
	ChangeU2MB(0x0050F6, 0xB246);
	ChangeU2MB(0x0050FA, 0xB247);
	ChangeU2MB(0x0050FE, 0xB248);
	ChangeU2MB(0x005103, 0xB249);
	ChangeU2MB(0x005106, 0xB24A);
	ChangeU2MB(0x005107, 0xB24B);
	ChangeU2MB(0x005108, 0xB24C);
	ChangeU2MB(0x00510B, 0xB24D);
	ChangeU2MB(0x00510C, 0xB24E);
	ChangeU2MB(0x00510D, 0xB24F);
	ChangeU2MB(0x00510E, 0xB250);
	ChangeU2MB(0x005110, 0xB252);
	ChangeU2MB(0x005117, 0xB253);
	ChangeU2MB(0x005119, 0xB254);
	ChangeU2MB(0x00511B, 0xB255);
	ChangeU2MB(0x00511C, 0xB256);
	ChangeU2MB(0x00511D, 0xB257);
	ChangeU2MB(0x00511E, 0xB258);
	ChangeU2MB(0x005123, 0xB259);
	ChangeU2MB(0x005127, 0xB25A);
	ChangeU2MB(0x005128, 0xB25B);
	ChangeU2MB(0x00512C, 0xB25C);
	ChangeU2MB(0x00512D, 0xB25D);
	ChangeU2MB(0x00512F, 0xB25E);
	ChangeU2MB(0x005131, 0xB25F);
	ChangeU2MB(0x005133, 0xB260);
	ChangeU2MB(0x005134, 0xB261);
	ChangeU2MB(0x005135, 0xB262);
	ChangeU2MB(0x005138, 0xB263);
	ChangeU2MB(0x005139, 0xB264);
	ChangeU2MB(0x005142, 0xB265);
	ChangeU2MB(0x00514F, 0xB267);
	ChangeU2MB(0x005153, 0xB268);
	ChangeU2MB(0x005155, 0xB269);
	ChangeU2MB(0x005157, 0xB26A);
	ChangeU2MB(0x005158, 0xB26B);
	ChangeU2MB(0x00515F, 0xB26C);
	ChangeU2MB(0x005166, 0xB26E);
	ChangeU2MB(0x00517E, 0xB26F);
	ChangeU2MB(0x005183, 0xB270);
	ChangeU2MB(0x005184, 0xB271);
	ChangeU2MB(0x00518B, 0xB272);
	ChangeU2MB(0x00518E, 0xB273);
	ChangeU2MB(0x005198, 0xB274);
	ChangeU2MB(0x0051A1, 0xB276);
	ChangeU2MB(0x0051A3, 0xB277);
	ChangeU2MB(0x0051AD, 0xB278);
	ChangeU2MB(0x0051B8, 0xB279);
	ChangeU2MB(0x0051BA, 0xB27A);
	ChangeU2MB(0x0051BC, 0xB27B);
	ChangeU2MB(0x0051BF, 0xB27D);
	ChangeU2MB(0x0051C2, 0xB27E);
	ChangeU2MB(0x0051C8, 0xB321);
	ChangeU2MB(0x0051CF, 0xB322);
	ChangeU2MB(0x0051D1, 0xB323);
	ChangeU2MB(0x0051D2, 0xB324);
	ChangeU2MB(0x0051D3, 0xB325);
	ChangeU2MB(0x0051D5, 0xB326);
	ChangeU2MB(0x0051D8, 0xB327);
	ChangeU2MB(0x0051DE, 0xB328);
	ChangeU2MB(0x0051E2, 0xB329);
	ChangeU2MB(0x0051E5, 0xB32A);
	ChangeU2MB(0x0051EE, 0xB32B);
	ChangeU2MB(0x0051F2, 0xB32C);
	ChangeU2MB(0x0051F3, 0xB32D);
	ChangeU2MB(0x0051F4, 0xB32E);
	ChangeU2MB(0x0051F7, 0xB32F);
	ChangeU2MB(0x005201, 0xB330);
	ChangeU2MB(0x005202, 0xB331);
	ChangeU2MB(0x005205, 0xB332);
	ChangeU2MB(0x005212, 0xB333);
	ChangeU2MB(0x005213, 0xB334);
	ChangeU2MB(0x005216, 0xB336);
	ChangeU2MB(0x005218, 0xB337);
	ChangeU2MB(0x005222, 0xB338);
	ChangeU2MB(0x005228, 0xB339);
	ChangeU2MB(0x005231, 0xB33A);
	ChangeU2MB(0x005232, 0xB33B);
	ChangeU2MB(0x005235, 0xB33C);
	ChangeU2MB(0x00523C, 0xB33D);
	ChangeU2MB(0x005245, 0xB33E);
	ChangeU2MB(0x005249, 0xB33F);
	ChangeU2MB(0x005255, 0xB340);
	ChangeU2MB(0x005257, 0xB341);
	ChangeU2MB(0x005258, 0xB342);
	ChangeU2MB(0x00525A, 0xB343);
	ChangeU2MB(0x00525C, 0xB344);
	ChangeU2MB(0x00525F, 0xB345);
	ChangeU2MB(0x005260, 0xB346);
	ChangeU2MB(0x005261, 0xB347);
	ChangeU2MB(0x005266, 0xB348);
	ChangeU2MB(0x00526E, 0xB349);
	ChangeU2MB(0x005277, 0xB34A);
	ChangeU2MB(0x005278, 0xB34B);
	ChangeU2MB(0x005279, 0xB34C);
	ChangeU2MB(0x005280, 0xB34D);
	ChangeU2MB(0x005282, 0xB34E);
	ChangeU2MB(0x005285, 0xB34F);
	ChangeU2MB(0x00528A, 0xB350);
	ChangeU2MB(0x00528C, 0xB351);
	ChangeU2MB(0x005293, 0xB352);
	ChangeU2MB(0x005295, 0xB353);
	ChangeU2MB(0x005296, 0xB354);
	ChangeU2MB(0x005297, 0xB355);
	ChangeU2MB(0x005298, 0xB356);
	ChangeU2MB(0x00529A, 0xB357);
	ChangeU2MB(0x0052A4, 0xB359);
	ChangeU2MB(0x0052A5, 0xB35A);
	ChangeU2MB(0x0052A7, 0xB35C);
	ChangeU2MB(0x0052B0, 0xB35E);
	ChangeU2MB(0x0052B6, 0xB35F);
	ChangeU2MB(0x0052B7, 0xB360);
	ChangeU2MB(0x0052B8, 0xB361);
	ChangeU2MB(0x0052BA, 0xB362);
	ChangeU2MB(0x0052BB, 0xB363);
	ChangeU2MB(0x0052BD, 0xB364);
	ChangeU2MB(0x0052C4, 0xB366);
	ChangeU2MB(0x0052C6, 0xB367);
	ChangeU2MB(0x0052C8, 0xB368);
	ChangeU2MB(0x0052CC, 0xB369);
	ChangeU2MB(0x0052CF, 0xB36A);
	ChangeU2MB(0x0052D1, 0xB36B);
	ChangeU2MB(0x0052D4, 0xB36C);
	ChangeU2MB(0x0052D6, 0xB36D);
	ChangeU2MB(0x0052DC, 0xB36F);
	ChangeU2MB(0x0052E1, 0xB370);
	ChangeU2MB(0x0052E5, 0xB371);
	ChangeU2MB(0x0052E8, 0xB372);
	ChangeU2MB(0x0052E9, 0xB373);
	ChangeU2MB(0x0052EA, 0xB374);
	ChangeU2MB(0x0052EC, 0xB375);
	ChangeU2MB(0x0052F0, 0xB376);
	ChangeU2MB(0x0052F1, 0xB377);
	ChangeU2MB(0x0052F4, 0xB378);
	ChangeU2MB(0x0052F6, 0xB379);
	ChangeU2MB(0x0052F7, 0xB37A);
	ChangeU2MB(0x005303, 0xB37C);
	ChangeU2MB(0x00530A, 0xB37D);
	ChangeU2MB(0x00530B, 0xB37E);
	ChangeU2MB(0x00530C, 0xB421);
	ChangeU2MB(0x005311, 0xB422);
	ChangeU2MB(0x005313, 0xB423);
	ChangeU2MB(0x005318, 0xB424);
	ChangeU2MB(0x00531B, 0xB425);
	ChangeU2MB(0x00531C, 0xB426);
	ChangeU2MB(0x00531E, 0xB427);
	ChangeU2MB(0x00531F, 0xB428);
	ChangeU2MB(0x005325, 0xB429);
	ChangeU2MB(0x005327, 0xB42A);
	ChangeU2MB(0x005328, 0xB42B);
	ChangeU2MB(0x005329, 0xB42C);
	ChangeU2MB(0x00532B, 0xB42D);
	ChangeU2MB(0x00532C, 0xB42E);
	ChangeU2MB(0x00532D, 0xB42F);
	ChangeU2MB(0x005330, 0xB430);
	ChangeU2MB(0x005332, 0xB431);
	ChangeU2MB(0x005335, 0xB432);
	ChangeU2MB(0x00533C, 0xB433);
	ChangeU2MB(0x00533D, 0xB434);
	ChangeU2MB(0x00533E, 0xB435);
	ChangeU2MB(0x005342, 0xB436);
	ChangeU2MB(0x00534B, 0xB438);
	ChangeU2MB(0x00534C, 0xB437);
	ChangeU2MB(0x005359, 0xB439);
	ChangeU2MB(0x00535B, 0xB43A);
	ChangeU2MB(0x005361, 0xB43B);
	ChangeU2MB(0x005363, 0xB43C);
	ChangeU2MB(0x005365, 0xB43D);
	ChangeU2MB(0x00536C, 0xB43E);
	ChangeU2MB(0x00536D, 0xB43F);
	ChangeU2MB(0x005379, 0xB441);
	ChangeU2MB(0x00537E, 0xB442);
	ChangeU2MB(0x005383, 0xB443);
	ChangeU2MB(0x005387, 0xB444);
	ChangeU2MB(0x005388, 0xB445);
	ChangeU2MB(0x00538E, 0xB446);
	ChangeU2MB(0x005394, 0xB448);
	ChangeU2MB(0x005399, 0xB449);
	ChangeU2MB(0x00539D, 0xB44A);
	ChangeU2MB(0x0053A1, 0xB44B);
	ChangeU2MB(0x0053A4, 0xB44C);
	ChangeU2MB(0x0053AA, 0xB44D);
	ChangeU2MB(0x0053AB, 0xB44E);
	ChangeU2MB(0x0053AF, 0xB44F);
	ChangeU2MB(0x0053B4, 0xB451);
	ChangeU2MB(0x0053B5, 0xB452);
	ChangeU2MB(0x0053B7, 0xB453);
	ChangeU2MB(0x0053B8, 0xB454);
	ChangeU2MB(0x0053BA, 0xB455);
	ChangeU2MB(0x0053BD, 0xB456);
	ChangeU2MB(0x0053C0, 0xB457);
	ChangeU2MB(0x0053C5, 0xB458);
	ChangeU2MB(0x0053CF, 0xB459);
	ChangeU2MB(0x0053D2, 0xB45A);
	ChangeU2MB(0x0053D3, 0xB45B);
	ChangeU2MB(0x0053D5, 0xB45C);
	ChangeU2MB(0x0053DA, 0xB45D);
	ChangeU2MB(0x0053DE, 0xB45F);
	ChangeU2MB(0x0053E0, 0xB460);
	ChangeU2MB(0x0053E6, 0xB461);
	ChangeU2MB(0x0053E7, 0xB462);
	ChangeU2MB(0x0053F5, 0xB463);
	ChangeU2MB(0x005402, 0xB464);
	ChangeU2MB(0x005413, 0xB465);
	ChangeU2MB(0x00541A, 0xB466);
	ChangeU2MB(0x005421, 0xB467);
	ChangeU2MB(0x005427, 0xB468);
	ChangeU2MB(0x005428, 0xB469);
	ChangeU2MB(0x00542A, 0xB46A);
	ChangeU2MB(0x00542F, 0xB46B);
	ChangeU2MB(0x005431, 0xB46C);
	ChangeU2MB(0x005434, 0xB46D);
	ChangeU2MB(0x005435, 0xB46E);
	ChangeU2MB(0x005443, 0xB46F);
	ChangeU2MB(0x005444, 0xB470);
	ChangeU2MB(0x005447, 0xB471);
	ChangeU2MB(0x00544D, 0xB472);
	ChangeU2MB(0x00544F, 0xB473);
	ChangeU2MB(0x00545E, 0xB474);
	ChangeU2MB(0x005462, 0xB475);
	ChangeU2MB(0x005464, 0xB476);
	ChangeU2MB(0x005466, 0xB477);
	ChangeU2MB(0x005467, 0xB478);
	ChangeU2MB(0x005469, 0xB479);
	ChangeU2MB(0x00546B, 0xB47A);
	ChangeU2MB(0x00546D, 0xB47B);
	ChangeU2MB(0x00546E, 0xB47C);
	ChangeU2MB(0x005474, 0xB47D);
	ChangeU2MB(0x00547F, 0xB47E);
	ChangeU2MB(0x005481, 0xB521);
	ChangeU2MB(0x005483, 0xB522);
	ChangeU2MB(0x005485, 0xB523);
	ChangeU2MB(0x005488, 0xB524);
	ChangeU2MB(0x005489, 0xB525);
	ChangeU2MB(0x00548D, 0xB526);
	ChangeU2MB(0x005491, 0xB527);
	ChangeU2MB(0x005495, 0xB528);
	ChangeU2MB(0x005496, 0xB529);
	ChangeU2MB(0x00549F, 0xB52B);
	ChangeU2MB(0x0054A1, 0xB52C);
	ChangeU2MB(0x0054A6, 0xB52D);
	ChangeU2MB(0x0054A7, 0xB52E);
	ChangeU2MB(0x0054AA, 0xB530);
	ChangeU2MB(0x0054AD, 0xB531);
	ChangeU2MB(0x0054AE, 0xB532);
	ChangeU2MB(0x0054B1, 0xB533);
	ChangeU2MB(0x0054B7, 0xB534);
	ChangeU2MB(0x0054B9, 0xB535);
	ChangeU2MB(0x0054BA, 0xB536);
	ChangeU2MB(0x0054BB, 0xB537);
	ChangeU2MB(0x0054BF, 0xB538);
	ChangeU2MB(0x0054C6, 0xB539);
	ChangeU2MB(0x0054CA, 0xB53A);
	ChangeU2MB(0x0054CD, 0xB53B);
	ChangeU2MB(0x0054CE, 0xB53C);
	ChangeU2MB(0x0054E0, 0xB53D);
	ChangeU2MB(0x0054EA, 0xB53E);
	ChangeU2MB(0x0054EC, 0xB53F);
	ChangeU2MB(0x0054EF, 0xB540);
	ChangeU2MB(0x0054F6, 0xB541);
	ChangeU2MB(0x0054FC, 0xB542);
	ChangeU2MB(0x0054FE, 0xB543);
	ChangeU2MB(0x005500, 0xB545);
	ChangeU2MB(0x005501, 0xB546);
	ChangeU2MB(0x005505, 0xB547);
	ChangeU2MB(0x005508, 0xB548);
	ChangeU2MB(0x005509, 0xB549);
	ChangeU2MB(0x00550C, 0xB54A);
	ChangeU2MB(0x00550D, 0xB54B);
	ChangeU2MB(0x00550E, 0xB54C);
	ChangeU2MB(0x005515, 0xB54D);
	ChangeU2MB(0x00552A, 0xB54E);
	ChangeU2MB(0x00552B, 0xB54F);
	ChangeU2MB(0x005532, 0xB550);
	ChangeU2MB(0x005535, 0xB551);
	ChangeU2MB(0x005536, 0xB552);
	ChangeU2MB(0x00553B, 0xB553);
	ChangeU2MB(0x00553C, 0xB554);
	ChangeU2MB(0x00553D, 0xB555);
	ChangeU2MB(0x005541, 0xB556);
	ChangeU2MB(0x005547, 0xB557);
	ChangeU2MB(0x005549, 0xB558);
	ChangeU2MB(0x00554A, 0xB559);
	ChangeU2MB(0x00554D, 0xB55A);
	ChangeU2MB(0x005550, 0xB55B);
	ChangeU2MB(0x005551, 0xB55C);
	ChangeU2MB(0x005558, 0xB55D);
	ChangeU2MB(0x00555A, 0xB55E);
	ChangeU2MB(0x00555B, 0xB55F);
	ChangeU2MB(0x00555E, 0xB560);
	ChangeU2MB(0x005560, 0xB561);
	ChangeU2MB(0x005561, 0xB562);
	ChangeU2MB(0x005564, 0xB563);
	ChangeU2MB(0x005566, 0xB564);
	ChangeU2MB(0x00557F, 0xB565);
	ChangeU2MB(0x005581, 0xB566);
	ChangeU2MB(0x005582, 0xB567);
	ChangeU2MB(0x005588, 0xB569);
	ChangeU2MB(0x00558E, 0xB56A);
	ChangeU2MB(0x00558F, 0xB56B);
	ChangeU2MB(0x005591, 0xB56C);
	ChangeU2MB(0x005592, 0xB56D);
	ChangeU2MB(0x005593, 0xB56E);
	ChangeU2MB(0x005594, 0xB56F);
	ChangeU2MB(0x005597, 0xB570);
	ChangeU2MB(0x0055A3, 0xB571);
	ChangeU2MB(0x0055A4, 0xB572);
	ChangeU2MB(0x0055AD, 0xB573);
	ChangeU2MB(0x0055B2, 0xB574);
	ChangeU2MB(0x0055BF, 0xB575);
	ChangeU2MB(0x0055C1, 0xB576);
	ChangeU2MB(0x0055C3, 0xB577);
	ChangeU2MB(0x0055C6, 0xB578);
	ChangeU2MB(0x0055C9, 0xB579);
	ChangeU2MB(0x0055CB, 0xB57A);
	ChangeU2MB(0x0055CC, 0xB57B);
	ChangeU2MB(0x0055CE, 0xB57C);
	ChangeU2MB(0x0055D1, 0xB57D);
	ChangeU2MB(0x0055D2, 0xB57E);
	ChangeU2MB(0x0055D3, 0xB621);
	ChangeU2MB(0x0055D7, 0xB622);
	ChangeU2MB(0x0055D8, 0xB623);
	ChangeU2MB(0x0055DB, 0xB624);
	ChangeU2MB(0x0055DE, 0xB625);
	ChangeU2MB(0x0055E2, 0xB626);
	ChangeU2MB(0x0055E9, 0xB627);
	ChangeU2MB(0x0055F6, 0xB628);
	ChangeU2MB(0x0055FF, 0xB629);
	ChangeU2MB(0x005605, 0xB62A);
	ChangeU2MB(0x005608, 0xB62B);
	ChangeU2MB(0x00560A, 0xB62C);
	ChangeU2MB(0x00560D, 0xB62D);
	ChangeU2MB(0x00560E, 0xB62E);
	ChangeU2MB(0x00560F, 0xB62F);
	ChangeU2MB(0x005610, 0xB630);
	ChangeU2MB(0x005611, 0xB631);
	ChangeU2MB(0x005612, 0xB632);
	ChangeU2MB(0x005619, 0xB633);
	ChangeU2MB(0x00562C, 0xB634);
	ChangeU2MB(0x005630, 0xB635);
	ChangeU2MB(0x005633, 0xB636);
	ChangeU2MB(0x005635, 0xB637);
	ChangeU2MB(0x005637, 0xB638);
	ChangeU2MB(0x005639, 0xB639);
	ChangeU2MB(0x00563B, 0xB63A);
	ChangeU2MB(0x00563C, 0xB63B);
	ChangeU2MB(0x00563D, 0xB63C);
	ChangeU2MB(0x00563F, 0xB63D);
	ChangeU2MB(0x005640, 0xB63E);
	ChangeU2MB(0x005641, 0xB63F);
	ChangeU2MB(0x005643, 0xB640);
	ChangeU2MB(0x005644, 0xB641);
	ChangeU2MB(0x005646, 0xB642);
	ChangeU2MB(0x005649, 0xB643);
	ChangeU2MB(0x00564B, 0xB644);
	ChangeU2MB(0x00564D, 0xB645);
	ChangeU2MB(0x00564F, 0xB646);
	ChangeU2MB(0x005654, 0xB647);
	ChangeU2MB(0x00565E, 0xB648);
	ChangeU2MB(0x005660, 0xB649);
	ChangeU2MB(0x005661, 0xB64A);
	ChangeU2MB(0x005662, 0xB64B);
	ChangeU2MB(0x005663, 0xB64C);
	ChangeU2MB(0x005666, 0xB64D);
	ChangeU2MB(0x005669, 0xB64E);
	ChangeU2MB(0x00566D, 0xB64F);
	ChangeU2MB(0x00566F, 0xB650);
	ChangeU2MB(0x005671, 0xB651);
	ChangeU2MB(0x005672, 0xB652);
	ChangeU2MB(0x005675, 0xB653);
	ChangeU2MB(0x005684, 0xB654);
	ChangeU2MB(0x005685, 0xB655);
	ChangeU2MB(0x005688, 0xB656);
	ChangeU2MB(0x00568B, 0xB657);
	ChangeU2MB(0x00568C, 0xB658);
	ChangeU2MB(0x005695, 0xB659);
	ChangeU2MB(0x005699, 0xB65A);
	ChangeU2MB(0x00569A, 0xB65B);
	ChangeU2MB(0x00569D, 0xB65C);
	ChangeU2MB(0x00569E, 0xB65D);
	ChangeU2MB(0x00569F, 0xB65E);
	ChangeU2MB(0x0056A6, 0xB65F);
	ChangeU2MB(0x0056A7, 0xB660);
	ChangeU2MB(0x0056A8, 0xB661);
	ChangeU2MB(0x0056A9, 0xB662);
	ChangeU2MB(0x0056AB, 0xB663);
	ChangeU2MB(0x0056AC, 0xB664);
	ChangeU2MB(0x0056AD, 0xB665);
	ChangeU2MB(0x0056B1, 0xB666);
	ChangeU2MB(0x0056B3, 0xB667);
	ChangeU2MB(0x0056B7, 0xB668);
	ChangeU2MB(0x0056BE, 0xB669);
	ChangeU2MB(0x0056C5, 0xB66A);
	ChangeU2MB(0x0056C9, 0xB66B);
	ChangeU2MB(0x0056CA, 0xB66C);
	ChangeU2MB(0x0056CB, 0xB66D);
	ChangeU2MB(0x0056CC, 0xB670);
	ChangeU2MB(0x0056CD, 0xB671);
	ChangeU2MB(0x0056CF, 0xB66E);
	ChangeU2MB(0x0056D0, 0xB66F);
	ChangeU2MB(0x0056D9, 0xB672);
	ChangeU2MB(0x0056DC, 0xB673);
	ChangeU2MB(0x0056DD, 0xB674);
	ChangeU2MB(0x0056DF, 0xB675);
	ChangeU2MB(0x0056E1, 0xB676);
	ChangeU2MB(0x0056E4, 0xB677);
	ChangeU2MB(0x0056E5, 0xB678);
	ChangeU2MB(0x0056E6, 0xB679);
	ChangeU2MB(0x0056E7, 0xB67A);
	ChangeU2MB(0x0056E8, 0xB67B);
	ChangeU2MB(0x0056EB, 0xB67D);
	ChangeU2MB(0x0056ED, 0xB67E);
	ChangeU2MB(0x0056F1, 0xB67C);
	ChangeU2MB(0x0056F6, 0xB721);
	ChangeU2MB(0x0056F7, 0xB722);
	ChangeU2MB(0x005701, 0xB723);
	ChangeU2MB(0x005702, 0xB724);
	ChangeU2MB(0x005707, 0xB725);
	ChangeU2MB(0x00570A, 0xB726);
	ChangeU2MB(0x00570C, 0xB727);
	ChangeU2MB(0x005711, 0xB728);
	ChangeU2MB(0x005715, 0xB729);
	ChangeU2MB(0x00571A, 0xB72A);
	ChangeU2MB(0x00571B, 0xB72B);
	ChangeU2MB(0x00571D, 0xB72C);
	ChangeU2MB(0x005720, 0xB72D);
	ChangeU2MB(0x005722, 0xB72E);
	ChangeU2MB(0x005723, 0xB72F);
	ChangeU2MB(0x005724, 0xB730);
	ChangeU2MB(0x005725, 0xB731);
	ChangeU2MB(0x005729, 0xB732);
	ChangeU2MB(0x00572A, 0xB733);
	ChangeU2MB(0x00572C, 0xB734);
	ChangeU2MB(0x00572E, 0xB735);
	ChangeU2MB(0x00572F, 0xB736);
	ChangeU2MB(0x005733, 0xB737);
	ChangeU2MB(0x005734, 0xB738);
	ChangeU2MB(0x00573D, 0xB739);
	ChangeU2MB(0x00573E, 0xB73A);
	ChangeU2MB(0x00573F, 0xB73B);
	ChangeU2MB(0x005745, 0xB73C);
	ChangeU2MB(0x005746, 0xB73D);
	ChangeU2MB(0x00574C, 0xB73E);
	ChangeU2MB(0x00574D, 0xB73F);
	ChangeU2MB(0x005752, 0xB740);
	ChangeU2MB(0x005762, 0xB741);
	ChangeU2MB(0x005767, 0xB743);
	ChangeU2MB(0x005768, 0xB744);
	ChangeU2MB(0x00576B, 0xB745);
	ChangeU2MB(0x00576D, 0xB746);
	ChangeU2MB(0x00576E, 0xB747);
	ChangeU2MB(0x00576F, 0xB748);
	ChangeU2MB(0x005770, 0xB749);
	ChangeU2MB(0x005771, 0xB74A);
	ChangeU2MB(0x005773, 0xB74B);
	ChangeU2MB(0x005774, 0xB74C);
	ChangeU2MB(0x005775, 0xB74D);
	ChangeU2MB(0x005777, 0xB74E);
	ChangeU2MB(0x005779, 0xB74F);
	ChangeU2MB(0x00577A, 0xB750);
	ChangeU2MB(0x00577B, 0xB751);
	ChangeU2MB(0x00577C, 0xB752);
	ChangeU2MB(0x00577E, 0xB753);
	ChangeU2MB(0x005781, 0xB754);
	ChangeU2MB(0x005783, 0xB755);
	ChangeU2MB(0x00578C, 0xB756);
	ChangeU2MB(0x005794, 0xB757);
	ChangeU2MB(0x005795, 0xB760);
	ChangeU2MB(0x005797, 0xB758);
	ChangeU2MB(0x005799, 0xB759);
	ChangeU2MB(0x00579A, 0xB75A);
	ChangeU2MB(0x00579C, 0xB75B);
	ChangeU2MB(0x00579D, 0xB75C);
	ChangeU2MB(0x00579E, 0xB75D);
	ChangeU2MB(0x00579F, 0xB75E);
	ChangeU2MB(0x0057A1, 0xB75F);
	ChangeU2MB(0x0057A7, 0xB761);
	ChangeU2MB(0x0057A8, 0xB762);
	ChangeU2MB(0x0057A9, 0xB763);
	ChangeU2MB(0x0057B8, 0xB765);
	ChangeU2MB(0x0057BD, 0xB766);
	ChangeU2MB(0x0057CC, 0xB769);
	ChangeU2MB(0x0057CF, 0xB76A);
	ChangeU2MB(0x0057D5, 0xB76B);
	ChangeU2MB(0x0057DD, 0xB76C);
	ChangeU2MB(0x0057DE, 0xB76D);
	ChangeU2MB(0x0057E1, 0xB77E);
	ChangeU2MB(0x0057E4, 0xB76E);
	ChangeU2MB(0x0057E6, 0xB76F);
	ChangeU2MB(0x0057E7, 0xB770);
	ChangeU2MB(0x0057E9, 0xB771);
	ChangeU2MB(0x0057ED, 0xB772);
	ChangeU2MB(0x0057F0, 0xB773);
	ChangeU2MB(0x0057F5, 0xB774);
	ChangeU2MB(0x0057F6, 0xB775);
	ChangeU2MB(0x0057F8, 0xB776);
	ChangeU2MB(0x0057FD, 0xB777);
	ChangeU2MB(0x0057FE, 0xB778);
	ChangeU2MB(0x0057FF, 0xB779);
	ChangeU2MB(0x005803, 0xB77A);
	ChangeU2MB(0x005804, 0xB77B);
	ChangeU2MB(0x005808, 0xB77C);
	ChangeU2MB(0x005809, 0xB77D);
	ChangeU2MB(0x00580C, 0xB821);
	ChangeU2MB(0x00580D, 0xB822);
	ChangeU2MB(0x00581B, 0xB823);
	ChangeU2MB(0x00581E, 0xB824);
	ChangeU2MB(0x00581F, 0xB825);
	ChangeU2MB(0x005820, 0xB826);
	ChangeU2MB(0x005826, 0xB827);
	ChangeU2MB(0x005827, 0xB828);
	ChangeU2MB(0x00582D, 0xB829);
	ChangeU2MB(0x005832, 0xB82A);
	ChangeU2MB(0x005839, 0xB82B);
	ChangeU2MB(0x00583F, 0xB82C);
	ChangeU2MB(0x005849, 0xB82D);
	ChangeU2MB(0x00584C, 0xB82E);
	ChangeU2MB(0x00584D, 0xB82F);
	ChangeU2MB(0x00584F, 0xB830);
	ChangeU2MB(0x005850, 0xB831);
	ChangeU2MB(0x005855, 0xB832);
	ChangeU2MB(0x00585F, 0xB833);
	ChangeU2MB(0x005861, 0xB834);
	ChangeU2MB(0x005864, 0xB835);
	ChangeU2MB(0x005867, 0xB836);
	ChangeU2MB(0x005868, 0xB837);
	ChangeU2MB(0x005878, 0xB838);
	ChangeU2MB(0x00587C, 0xB839);
	ChangeU2MB(0x00587F, 0xB83A);
	ChangeU2MB(0x005880, 0xB83B);
	ChangeU2MB(0x005881, 0xB83C);
	ChangeU2MB(0x005887, 0xB83D);
	ChangeU2MB(0x005888, 0xB83E);
	ChangeU2MB(0x005889, 0xB83F);
	ChangeU2MB(0x00588A, 0xB840);
	ChangeU2MB(0x00588C, 0xB841);
	ChangeU2MB(0x00588D, 0xB842);
	ChangeU2MB(0x00588F, 0xB843);
	ChangeU2MB(0x005890, 0xB844);
	ChangeU2MB(0x005894, 0xB845);
	ChangeU2MB(0x005896, 0xB846);
	ChangeU2MB(0x00589D, 0xB847);
	ChangeU2MB(0x0058A0, 0xB848);
	ChangeU2MB(0x0058A1, 0xB849);
	ChangeU2MB(0x0058A2, 0xB84A);
	ChangeU2MB(0x0058A6, 0xB84B);
	ChangeU2MB(0x0058A9, 0xB84C);
	ChangeU2MB(0x0058B1, 0xB84D);
	ChangeU2MB(0x0058BC, 0xB850);
	ChangeU2MB(0x0058C2, 0xB851);
	ChangeU2MB(0x0058C4, 0xB84F);
	ChangeU2MB(0x0058C8, 0xB852);
	ChangeU2MB(0x0058CD, 0xB853);
	ChangeU2MB(0x0058CE, 0xB854);
	ChangeU2MB(0x0058D0, 0xB855);
	ChangeU2MB(0x0058D2, 0xB856);
	ChangeU2MB(0x0058D4, 0xB857);
	ChangeU2MB(0x0058D6, 0xB858);
	ChangeU2MB(0x0058DA, 0xB859);
	ChangeU2MB(0x0058DD, 0xB85A);
	ChangeU2MB(0x0058E1, 0xB85B);
	ChangeU2MB(0x0058E2, 0xB85C);
	ChangeU2MB(0x0058E9, 0xB85D);
	ChangeU2MB(0x0058F3, 0xB85E);
	ChangeU2MB(0x005905, 0xB85F);
	ChangeU2MB(0x005906, 0xB860);
	ChangeU2MB(0x00590C, 0xB862);
	ChangeU2MB(0x005912, 0xB863);
	ChangeU2MB(0x005913, 0xB864);
	ChangeU2MB(0x005914, 0xB865);
	ChangeU2MB(0x00591D, 0xB867);
	ChangeU2MB(0x005921, 0xB868);
	ChangeU2MB(0x005923, 0xB869);
	ChangeU2MB(0x005924, 0xB86A);
	ChangeU2MB(0x005928, 0xB86B);
	ChangeU2MB(0x00592F, 0xB86C);
	ChangeU2MB(0x005930, 0xB86D);
	ChangeU2MB(0x005933, 0xB86E);
	ChangeU2MB(0x005935, 0xB86F);
	ChangeU2MB(0x005936, 0xB870);
	ChangeU2MB(0x00593F, 0xB871);
	ChangeU2MB(0x005943, 0xB872);
	ChangeU2MB(0x005946, 0xB873);
	ChangeU2MB(0x005952, 0xB874);
	ChangeU2MB(0x005959, 0xB876);
	ChangeU2MB(0x00595E, 0xB879);
	ChangeU2MB(0x00595F, 0xB87A);
	ChangeU2MB(0x005961, 0xB87B);
	ChangeU2MB(0x00596B, 0xB87D);
	ChangeU2MB(0x00596D, 0xB87E);
	ChangeU2MB(0x00596F, 0xB921);
	ChangeU2MB(0x005972, 0xB922);
	ChangeU2MB(0x005975, 0xB923);
	ChangeU2MB(0x005976, 0xB924);
	ChangeU2MB(0x005979, 0xB925);
	ChangeU2MB(0x00597B, 0xB926);
	ChangeU2MB(0x00597C, 0xB927);
	ChangeU2MB(0x00598B, 0xB928);
	ChangeU2MB(0x00598C, 0xB929);
	ChangeU2MB(0x00598E, 0xB92A);
	ChangeU2MB(0x005992, 0xB92B);
	ChangeU2MB(0x005995, 0xB92C);
	ChangeU2MB(0x005997, 0xB92D);
	ChangeU2MB(0x00599F, 0xB92E);
	ChangeU2MB(0x0059A7, 0xB930);
	ChangeU2MB(0x0059AD, 0xB931);
	ChangeU2MB(0x0059AE, 0xB932);
	ChangeU2MB(0x0059AF, 0xB933);
	ChangeU2MB(0x0059B0, 0xB934);
	ChangeU2MB(0x0059B3, 0xB935);
	ChangeU2MB(0x0059B7, 0xB936);
	ChangeU2MB(0x0059BC, 0xB938);
	ChangeU2MB(0x0059C1, 0xB939);
	ChangeU2MB(0x0059C3, 0xB93A);
	ChangeU2MB(0x0059C4, 0xB93B);
	ChangeU2MB(0x0059C8, 0xB93C);
	ChangeU2MB(0x0059CA, 0xB93D);
	ChangeU2MB(0x0059CD, 0xB93E);
	ChangeU2MB(0x0059D2, 0xB93F);
	ChangeU2MB(0x0059DD, 0xB940);
	ChangeU2MB(0x0059DE, 0xB941);
	ChangeU2MB(0x0059DF, 0xB942);
	ChangeU2MB(0x0059E3, 0xB943);
	ChangeU2MB(0x0059E4, 0xB944);
	ChangeU2MB(0x0059E7, 0xB945);
	ChangeU2MB(0x0059EE, 0xB946);
	ChangeU2MB(0x0059EF, 0xB947);
	ChangeU2MB(0x0059F1, 0xB948);
	ChangeU2MB(0x0059F2, 0xB949);
	ChangeU2MB(0x0059F4, 0xB94A);
	ChangeU2MB(0x0059F7, 0xB94B);
	ChangeU2MB(0x005A00, 0xB94C);
	ChangeU2MB(0x005A04, 0xB94D);
	ChangeU2MB(0x005A0C, 0xB94E);
	ChangeU2MB(0x005A0D, 0xB94F);
	ChangeU2MB(0x005A0E, 0xB950);
	ChangeU2MB(0x005A12, 0xB951);
	ChangeU2MB(0x005A13, 0xB952);
	ChangeU2MB(0x005A1E, 0xB953);
	ChangeU2MB(0x005A23, 0xB954);
	ChangeU2MB(0x005A24, 0xB955);
	ChangeU2MB(0x005A27, 0xB956);
	ChangeU2MB(0x005A28, 0xB957);
	ChangeU2MB(0x005A2A, 0xB958);
	ChangeU2MB(0x005A2D, 0xB959);
	ChangeU2MB(0x005A30, 0xB95A);
	ChangeU2MB(0x005A44, 0xB95B);
	ChangeU2MB(0x005A45, 0xB95C);
	ChangeU2MB(0x005A47, 0xB95D);
	ChangeU2MB(0x005A48, 0xB95E);
	ChangeU2MB(0x005A4C, 0xB95F);
	ChangeU2MB(0x005A50, 0xB960);
	ChangeU2MB(0x005A55, 0xB961);
	ChangeU2MB(0x005A5E, 0xB962);
	ChangeU2MB(0x005A63, 0xB963);
	ChangeU2MB(0x005A65, 0xB964);
	ChangeU2MB(0x005A67, 0xB965);
	ChangeU2MB(0x005A6D, 0xB966);
	ChangeU2MB(0x005A77, 0xB967);
	ChangeU2MB(0x005A7A, 0xB968);
	ChangeU2MB(0x005A7B, 0xB969);
	ChangeU2MB(0x005A7E, 0xB96A);
	ChangeU2MB(0x005A8B, 0xB96B);
	ChangeU2MB(0x005A90, 0xB96C);
	ChangeU2MB(0x005A93, 0xB96D);
	ChangeU2MB(0x005A96, 0xB96E);
	ChangeU2MB(0x005A99, 0xB96F);
	ChangeU2MB(0x005A9C, 0xB970);
	ChangeU2MB(0x005A9E, 0xB971);
	ChangeU2MB(0x005A9F, 0xB972);
	ChangeU2MB(0x005AA0, 0xB973);
	ChangeU2MB(0x005AA2, 0xB974);
	ChangeU2MB(0x005AA7, 0xB975);
	ChangeU2MB(0x005AAC, 0xB976);
	ChangeU2MB(0x005AB1, 0xB977);
	ChangeU2MB(0x005AB2, 0xB978);
	ChangeU2MB(0x005AB3, 0xB979);
	ChangeU2MB(0x005AB5, 0xB97A);
	ChangeU2MB(0x005AB8, 0xB97B);
	ChangeU2MB(0x005ABA, 0xB97C);
	ChangeU2MB(0x005ABB, 0xB97D);
	ChangeU2MB(0x005ABF, 0xB97E);
	ChangeU2MB(0x005AC4, 0xBA21);
	ChangeU2MB(0x005AC6, 0xBA22);
	ChangeU2MB(0x005AC8, 0xBA23);
	ChangeU2MB(0x005ACF, 0xBA24);
	ChangeU2MB(0x005ADA, 0xBA25);
	ChangeU2MB(0x005ADC, 0xBA26);
	ChangeU2MB(0x005AE0, 0xBA27);
	ChangeU2MB(0x005AE5, 0xBA28);
	ChangeU2MB(0x005AEA, 0xBA29);
	ChangeU2MB(0x005AEE, 0xBA2A);
	ChangeU2MB(0x005AF5, 0xBA2B);
	ChangeU2MB(0x005AF6, 0xBA2C);
	ChangeU2MB(0x005AFD, 0xBA2D);
	ChangeU2MB(0x005B00, 0xBA2E);
	ChangeU2MB(0x005B01, 0xBA2F);
	ChangeU2MB(0x005B08, 0xBA30);
	ChangeU2MB(0x005B17, 0xBA31);
	ChangeU2MB(0x005B19, 0xBA33);
	ChangeU2MB(0x005B1B, 0xBA34);
	ChangeU2MB(0x005B1D, 0xBA35);
	ChangeU2MB(0x005B21, 0xBA36);
	ChangeU2MB(0x005B25, 0xBA37);
	ChangeU2MB(0x005B2D, 0xBA38);
	ChangeU2MB(0x005B34, 0xBA32);
	ChangeU2MB(0x005B38, 0xBA39);
	ChangeU2MB(0x005B41, 0xBA3A);
	ChangeU2MB(0x005B4B, 0xBA3B);
	ChangeU2MB(0x005B4C, 0xBA3C);
	ChangeU2MB(0x005B52, 0xBA3D);
	ChangeU2MB(0x005B5E, 0xBA3F);
	ChangeU2MB(0x005B68, 0xBA40);
	ChangeU2MB(0x005B6E, 0xBA41);
	ChangeU2MB(0x005B6F, 0xBA42);
	ChangeU2MB(0x005B7C, 0xBA43);
	ChangeU2MB(0x005B7D, 0xBA44);
	ChangeU2MB(0x005B7E, 0xBA45);
	ChangeU2MB(0x005B7F, 0xBA46);
	ChangeU2MB(0x005B81, 0xBA47);
	ChangeU2MB(0x005B84, 0xBA48);
	ChangeU2MB(0x005B86, 0xBA49);
	ChangeU2MB(0x005B8A, 0xBA4A);
	ChangeU2MB(0x005B8E, 0xBA4B);
	ChangeU2MB(0x005B90, 0xBA4C);
	ChangeU2MB(0x005B91, 0xBA4D);
	ChangeU2MB(0x005B93, 0xBA4E);
	ChangeU2MB(0x005B94, 0xBA4F);
	ChangeU2MB(0x005B96, 0xBA50);
	ChangeU2MB(0x005BA8, 0xBA51);
	ChangeU2MB(0x005BA9, 0xBA52);
	ChangeU2MB(0x005BAC, 0xBA53);
	ChangeU2MB(0x005BAD, 0xBA54);
	ChangeU2MB(0x005BAF, 0xBA55);
	ChangeU2MB(0x005BB1, 0xBA56);
	ChangeU2MB(0x005BB2, 0xBA57);
	ChangeU2MB(0x005BB7, 0xBA58);
	ChangeU2MB(0x005BBA, 0xBA59);
	ChangeU2MB(0x005BBC, 0xBA5A);
	ChangeU2MB(0x005BC1, 0xBA5C);
	ChangeU2MB(0x005BCD, 0xBA5D);
	ChangeU2MB(0x005BCF, 0xBA5E);
	ChangeU2MB(0x005BD6, 0xBA5F);
	ChangeU2MB(0x005BD7, 0xBA60);
	ChangeU2MB(0x005BD9, 0xBA62);
	ChangeU2MB(0x005BDA, 0xBA63);
	ChangeU2MB(0x005BE0, 0xBA64);
	ChangeU2MB(0x005BEF, 0xBA65);
	ChangeU2MB(0x005BF1, 0xBA66);
	ChangeU2MB(0x005BF4, 0xBA67);
	ChangeU2MB(0x005BFD, 0xBA68);
	ChangeU2MB(0x005C0C, 0xBA69);
	ChangeU2MB(0x005C17, 0xBA6A);
	ChangeU2MB(0x005C1F, 0xBA6C);
	ChangeU2MB(0x005C23, 0xBA6D);
	ChangeU2MB(0x005C26, 0xBA6E);
	ChangeU2MB(0x005C29, 0xBA6F);
	ChangeU2MB(0x005C2B, 0xBA70);
	ChangeU2MB(0x005C2C, 0xBA71);
	ChangeU2MB(0x005C2E, 0xBA72);
	ChangeU2MB(0x005C30, 0xBA73);
	ChangeU2MB(0x005C32, 0xBA74);
	ChangeU2MB(0x005C35, 0xBA75);
	ChangeU2MB(0x005C36, 0xBA76);
	ChangeU2MB(0x005C59, 0xBA77);
	ChangeU2MB(0x005C5A, 0xBA78);
	ChangeU2MB(0x005C5C, 0xBA79);
	ChangeU2MB(0x005C62, 0xBA7A);
	ChangeU2MB(0x005C63, 0xBA7B);
	ChangeU2MB(0x005C67, 0xBA7C);
	ChangeU2MB(0x005C68, 0xBA7D);
	ChangeU2MB(0x005C69, 0xBA7E);
	ChangeU2MB(0x005C6D, 0xBB21);
	ChangeU2MB(0x005C70, 0xBB22);
	ChangeU2MB(0x005C74, 0xBB23);
	ChangeU2MB(0x005C75, 0xBB24);
	ChangeU2MB(0x005C7A, 0xBB25);
	ChangeU2MB(0x005C7B, 0xBB26);
	ChangeU2MB(0x005C7C, 0xBB27);
	ChangeU2MB(0x005C7D, 0xBB28);
	ChangeU2MB(0x005C87, 0xBB29);
	ChangeU2MB(0x005C88, 0xBB2A);
	ChangeU2MB(0x005C8A, 0xBB2B);
	ChangeU2MB(0x005C8F, 0xBB2C);
	ChangeU2MB(0x005C92, 0xBB2D);
	ChangeU2MB(0x005C9D, 0xBB2E);
	ChangeU2MB(0x005C9F, 0xBB2F);
	ChangeU2MB(0x005CA0, 0xBB30);
	ChangeU2MB(0x005CA2, 0xBB31);
	ChangeU2MB(0x005CA3, 0xBB32);
	ChangeU2MB(0x005CAA, 0xBB34);
	ChangeU2MB(0x005CB2, 0xBB35);
	ChangeU2MB(0x005CB4, 0xBB36);
	ChangeU2MB(0x005CB5, 0xBB37);
	ChangeU2MB(0x005CC9, 0xBB39);
	ChangeU2MB(0x005CCB, 0xBB3A);
	ChangeU2MB(0x005CD2, 0xBB3B);
	ChangeU2MB(0x005CD7, 0xBB3D);
	ChangeU2MB(0x005CDD, 0xBB3C);
	ChangeU2MB(0x005CEE, 0xBB3E);
	ChangeU2MB(0x005CF1, 0xBB3F);
	ChangeU2MB(0x005CF2, 0xBB40);
	ChangeU2MB(0x005CF4, 0xBB41);
	ChangeU2MB(0x005D01, 0xBB42);
	ChangeU2MB(0x005D06, 0xBB43);
	ChangeU2MB(0x005D0D, 0xBB44);
	ChangeU2MB(0x005D12, 0xBB45);
	ChangeU2MB(0x005D23, 0xBB47);
	ChangeU2MB(0x005D24, 0xBB48);
	ChangeU2MB(0x005D26, 0xBB49);
	ChangeU2MB(0x005D2B, 0xBB46);
	ChangeU2MB(0x005D31, 0xBB4B);
	ChangeU2MB(0x005D34, 0xBB4C);
	ChangeU2MB(0x005D39, 0xBB4D);
	ChangeU2MB(0x005D3D, 0xBB4E);
	ChangeU2MB(0x005D3F, 0xBB4F);
	ChangeU2MB(0x005D43, 0xBB51);
	ChangeU2MB(0x005D46, 0xBB52);
	ChangeU2MB(0x005D48, 0xBB53);
	ChangeU2MB(0x005D4A, 0xBB57);
	ChangeU2MB(0x005D51, 0xBB55);
	ChangeU2MB(0x005D55, 0xBB54);
	ChangeU2MB(0x005D59, 0xBB56);
	ChangeU2MB(0x005D5F, 0xBB58);
	ChangeU2MB(0x005D60, 0xBB59);
	ChangeU2MB(0x005D61, 0xBB5A);
	ChangeU2MB(0x005D62, 0xBB5B);
	ChangeU2MB(0x005D64, 0xBB5C);
	ChangeU2MB(0x005D6A, 0xBB5D);
	ChangeU2MB(0x005D70, 0xBB5F);
	ChangeU2MB(0x005D79, 0xBB60);
	ChangeU2MB(0x005D7A, 0xBB61);
	ChangeU2MB(0x005D7E, 0xBB62);
	ChangeU2MB(0x005D7F, 0xBB63);
	ChangeU2MB(0x005D81, 0xBB64);
	ChangeU2MB(0x005D83, 0xBB65);
	ChangeU2MB(0x005D88, 0xBB66);
	ChangeU2MB(0x005D8A, 0xBB67);
	ChangeU2MB(0x005D92, 0xBB68);
	ChangeU2MB(0x005D93, 0xBB69);
	ChangeU2MB(0x005D94, 0xBB6A);
	ChangeU2MB(0x005D95, 0xBB6B);
	ChangeU2MB(0x005D99, 0xBB6C);
	ChangeU2MB(0x005D9B, 0xBB6D);
	ChangeU2MB(0x005D9F, 0xBB6E);
	ChangeU2MB(0x005DA0, 0xBB6F);
	ChangeU2MB(0x005DA7, 0xBB70);
	ChangeU2MB(0x005DAB, 0xBB71);
	ChangeU2MB(0x005DB0, 0xBB72);
	ChangeU2MB(0x005DB2, 0xE674);
	ChangeU2MB(0x005DB4, 0xBB73);
	ChangeU2MB(0x005DC3, 0xBB76);
	ChangeU2MB(0x005DC7, 0xBB77);
	ChangeU2MB(0x005DCB, 0xBB78);
	ChangeU2MB(0x005DCE, 0xBB7A);
	ChangeU2MB(0x005DD8, 0xBB7B);
	ChangeU2MB(0x005DD9, 0xBB7C);
	ChangeU2MB(0x005DE0, 0xBB7D);
	ChangeU2MB(0x005DE4, 0xBB7E);
	ChangeU2MB(0x005DE9, 0xBC21);
	ChangeU2MB(0x005DF8, 0xBC22);
	ChangeU2MB(0x005DF9, 0xBC23);
	ChangeU2MB(0x005E00, 0xBC24);
	ChangeU2MB(0x005E07, 0xBC25);
	ChangeU2MB(0x005E0D, 0xBC26);
	ChangeU2MB(0x005E12, 0xBC27);
	ChangeU2MB(0x005E14, 0xBC28);
	ChangeU2MB(0x005E15, 0xBC29);
	ChangeU2MB(0x005E18, 0xBC2A);
	ChangeU2MB(0x005E1F, 0xBC2B);
	ChangeU2MB(0x005E20, 0xBC2C);
	ChangeU2MB(0x005E28, 0xBC2E);
	ChangeU2MB(0x005E2E, 0xBC2D);
	ChangeU2MB(0x005E32, 0xBC2F);
	ChangeU2MB(0x005E35, 0xBC30);
	ChangeU2MB(0x005E3E, 0xBC31);
	ChangeU2MB(0x005E49, 0xBC34);
	ChangeU2MB(0x005E4B, 0xBC32);
	ChangeU2MB(0x005E50, 0xBC33);
	ChangeU2MB(0x005E51, 0xBC35);
	ChangeU2MB(0x005E56, 0xBC36);
	ChangeU2MB(0x005E58, 0xBC37);
	ChangeU2MB(0x005E5B, 0xBC38);
	ChangeU2MB(0x005E5C, 0xBC39);
	ChangeU2MB(0x005E5E, 0xBC3A);
	ChangeU2MB(0x005E68, 0xBC3B);
	ChangeU2MB(0x005E6A, 0xBC3C);
	ChangeU2MB(0x005E6B, 0xBC3D);
	ChangeU2MB(0x005E6C, 0xBC3E);
	ChangeU2MB(0x005E6D, 0xBC3F);
	ChangeU2MB(0x005E6E, 0xBC40);
	ChangeU2MB(0x005E70, 0xBC41);
	ChangeU2MB(0x005E80, 0xBC42);
	ChangeU2MB(0x005E8B, 0xBC43);
	ChangeU2MB(0x005E8E, 0xBC44);
	ChangeU2MB(0x005EA2, 0xBC45);
	ChangeU2MB(0x005EA4, 0xBC46);
	ChangeU2MB(0x005EA5, 0xBC47);
	ChangeU2MB(0x005EA8, 0xBC48);
	ChangeU2MB(0x005EAA, 0xBC49);
	ChangeU2MB(0x005EAC, 0xBC4A);
	ChangeU2MB(0x005EB1, 0xBC4B);
	ChangeU2MB(0x005EB3, 0xBC4C);
	ChangeU2MB(0x005EBD, 0xBC4D);
	ChangeU2MB(0x005EBE, 0xBC4E);
	ChangeU2MB(0x005EBF, 0xBC4F);
	ChangeU2MB(0x005EC6, 0xBC50);
	ChangeU2MB(0x005ECB, 0xBC52);
	ChangeU2MB(0x005ECC, 0xBC51);
	ChangeU2MB(0x005ECE, 0xBC53);
	ChangeU2MB(0x005ED1, 0xBC54);
	ChangeU2MB(0x005ED2, 0xBC55);
	ChangeU2MB(0x005ED4, 0xBC56);
	ChangeU2MB(0x005ED5, 0xBC57);
	ChangeU2MB(0x005EDC, 0xBC58);
	ChangeU2MB(0x005EDE, 0xBC59);
	ChangeU2MB(0x005EE5, 0xBC5A);
	ChangeU2MB(0x005EEB, 0xBC5B);
	ChangeU2MB(0x005F02, 0xBC5C);
	ChangeU2MB(0x005F06, 0xBC5D);
	ChangeU2MB(0x005F07, 0xBC5E);
	ChangeU2MB(0x005F08, 0xBC5F);
	ChangeU2MB(0x005F0E, 0xBC60);
	ChangeU2MB(0x005F19, 0xBC61);
	ChangeU2MB(0x005F1C, 0xBC62);
	ChangeU2MB(0x005F1D, 0xBC63);
	ChangeU2MB(0x005F22, 0xBC65);
	ChangeU2MB(0x005F23, 0xBC66);
	ChangeU2MB(0x005F24, 0xBC67);
	ChangeU2MB(0x005F28, 0xBC68);
	ChangeU2MB(0x005F2B, 0xBC69);
	ChangeU2MB(0x005F2C, 0xBC6A);
	ChangeU2MB(0x005F2E, 0xBC6B);
	ChangeU2MB(0x005F30, 0xBC6C);
	ChangeU2MB(0x005F36, 0xBC6E);
	ChangeU2MB(0x005F3B, 0xBC6F);
	ChangeU2MB(0x005F3D, 0xBC70);
	ChangeU2MB(0x005F3F, 0xBC71);
	ChangeU2MB(0x005F40, 0xBC72);
	ChangeU2MB(0x005F44, 0xBC73);
	ChangeU2MB(0x005F47, 0xBC75);
	ChangeU2MB(0x005F4D, 0xBC76);
	ChangeU2MB(0x005F50, 0xBC77);
	ChangeU2MB(0x005F54, 0xBC78);
	ChangeU2MB(0x005F58, 0xBC79);
	ChangeU2MB(0x005F5B, 0xBC7A);
	ChangeU2MB(0x005F60, 0xBC7B);
	ChangeU2MB(0x005F63, 0xBC7C);
	ChangeU2MB(0x005F64, 0xBC7D);
	ChangeU2MB(0x005F6F, 0xBD21);
	ChangeU2MB(0x005F72, 0xBD22);
	ChangeU2MB(0x005F74, 0xBD23);
	ChangeU2MB(0x005F75, 0xBD24);
	ChangeU2MB(0x005F78, 0xBD25);
	ChangeU2MB(0x005F7A, 0xBD26);
	ChangeU2MB(0x005F7D, 0xBD27);
	ChangeU2MB(0x005F7E, 0xBD28);
	ChangeU2MB(0x005F89, 0xBD29);
	ChangeU2MB(0x005F8D, 0xBD2A);
	ChangeU2MB(0x005F8F, 0xBD2B);
	ChangeU2MB(0x005F96, 0xBD2C);
	ChangeU2MB(0x005F9C, 0xBD2D);
	ChangeU2MB(0x005F9D, 0xBD2E);
	ChangeU2MB(0x005FA2, 0xBD2F);
	ChangeU2MB(0x005FA4, 0xBD32);
	ChangeU2MB(0x005FA7, 0xBD30);
	ChangeU2MB(0x005FAB, 0xBD31);
	ChangeU2MB(0x005FAC, 0xBD33);
	ChangeU2MB(0x005FAF, 0xBD34);
	ChangeU2MB(0x005FB0, 0xBD35);
	ChangeU2MB(0x005FB1, 0xBD36);
	ChangeU2MB(0x005FB8, 0xBD37);
	ChangeU2MB(0x005FC4, 0xBD38);
	ChangeU2MB(0x005FC7, 0xBD39);
	ChangeU2MB(0x005FC8, 0xBD3A);
	ChangeU2MB(0x005FC9, 0xBD3B);
	ChangeU2MB(0x005FCB, 0xBD3C);
	ChangeU2MB(0x005FD0, 0xBD3D);
	ChangeU2MB(0x005FD1, 0xBD3E);
	ChangeU2MB(0x005FD2, 0xBD3F);
	ChangeU2MB(0x005FD3, 0xBD40);
	ChangeU2MB(0x005FD4, 0xBD41);
	ChangeU2MB(0x005FE1, 0xBD43);
	ChangeU2MB(0x005FE2, 0xBD44);
	ChangeU2MB(0x005FE8, 0xBD45);
	ChangeU2MB(0x005FE9, 0xBD46);
	ChangeU2MB(0x005FEA, 0xBD47);
	ChangeU2MB(0x005FEC, 0xBD48);
	ChangeU2MB(0x005FED, 0xBD49);
	ChangeU2MB(0x005FEE, 0xBD4A);
	ChangeU2MB(0x005FEF, 0xBD4B);
	ChangeU2MB(0x005FF2, 0xBD4C);
	ChangeU2MB(0x005FF3, 0xBD4D);
	ChangeU2MB(0x005FF6, 0xBD4E);
	ChangeU2MB(0x005FFA, 0xBD4F);
	ChangeU2MB(0x005FFC, 0xBD50);
	ChangeU2MB(0x006007, 0xBD51);
	ChangeU2MB(0x00600A, 0xBD52);
	ChangeU2MB(0x00600D, 0xBD53);
	ChangeU2MB(0x006013, 0xBD54);
	ChangeU2MB(0x006014, 0xBD55);
	ChangeU2MB(0x006017, 0xBD56);
	ChangeU2MB(0x006018, 0xBD57);
	ChangeU2MB(0x00601A, 0xBD58);
	ChangeU2MB(0x00601F, 0xBD59);
	ChangeU2MB(0x006024, 0xBD5A);
	ChangeU2MB(0x00602D, 0xBD5B);
	ChangeU2MB(0x006033, 0xBD5C);
	ChangeU2MB(0x006035, 0xBD5D);
	ChangeU2MB(0x006040, 0xBD5E);
	ChangeU2MB(0x006047, 0xBD5F);
	ChangeU2MB(0x006048, 0xBD60);
	ChangeU2MB(0x006049, 0xBD61);
	ChangeU2MB(0x00604C, 0xBD62);
	ChangeU2MB(0x006051, 0xBD63);
	ChangeU2MB(0x006054, 0xBD64);
	ChangeU2MB(0x006056, 0xBD65);
	ChangeU2MB(0x006057, 0xBD66);
	ChangeU2MB(0x006061, 0xBD68);
	ChangeU2MB(0x006067, 0xBD69);
	ChangeU2MB(0x006071, 0xBD6A);
	ChangeU2MB(0x00607E, 0xBD6B);
	ChangeU2MB(0x00607F, 0xBD6C);
	ChangeU2MB(0x006082, 0xBD6D);
	ChangeU2MB(0x006086, 0xBD6E);
	ChangeU2MB(0x006088, 0xBD6F);
	ChangeU2MB(0x00608E, 0xBD71);
	ChangeU2MB(0x006091, 0xBD72);
	ChangeU2MB(0x006093, 0xBD73);
	ChangeU2MB(0x006095, 0xBD74);
	ChangeU2MB(0x006098, 0xBD75);
	ChangeU2MB(0x00609D, 0xBD76);
	ChangeU2MB(0x00609E, 0xBD77);
	ChangeU2MB(0x0060A2, 0xBD78);
	ChangeU2MB(0x0060A4, 0xBD79);
	ChangeU2MB(0x0060A5, 0xBD7A);
	ChangeU2MB(0x0060A8, 0xBD7B);
	ChangeU2MB(0x0060B0, 0xBD7C);
	ChangeU2MB(0x0060B1, 0xBD7D);
	ChangeU2MB(0x0060B7, 0xBD7E);
	ChangeU2MB(0x0060BB, 0xBE21);
	ChangeU2MB(0x0060BE, 0xBE22);
	ChangeU2MB(0x0060C2, 0xBE23);
	ChangeU2MB(0x0060C4, 0xBE24);
	ChangeU2MB(0x0060C8, 0xBE25);
	ChangeU2MB(0x0060C9, 0xBE26);
	ChangeU2MB(0x0060CA, 0xBE27);
	ChangeU2MB(0x0060CB, 0xBE28);
	ChangeU2MB(0x0060CE, 0xBE29);
	ChangeU2MB(0x0060CF, 0xBE2A);
	ChangeU2MB(0x0060D4, 0xBE2B);
	ChangeU2MB(0x0060D9, 0xBE2D);
	ChangeU2MB(0x0060DB, 0xBE2E);
	ChangeU2MB(0x0060DD, 0xBE2F);
	ChangeU2MB(0x0060E2, 0xBE31);
	ChangeU2MB(0x0060E5, 0xBE32);
	ChangeU2MB(0x0060F5, 0xBE34);
	ChangeU2MB(0x0060F8, 0xBE35);
	ChangeU2MB(0x0060FC, 0xBE36);
	ChangeU2MB(0x0060FD, 0xBE37);
	ChangeU2MB(0x006102, 0xBE38);
	ChangeU2MB(0x006107, 0xBE39);
	ChangeU2MB(0x00610A, 0xBE3A);
	ChangeU2MB(0x00610C, 0xBE3B);
	ChangeU2MB(0x006110, 0xBE3C);
	ChangeU2MB(0x006112, 0xBE3E);
	ChangeU2MB(0x006113, 0xBE3F);
	ChangeU2MB(0x006114, 0xBE40);
	ChangeU2MB(0x006116, 0xBE41);
	ChangeU2MB(0x006117, 0xBE42);
	ChangeU2MB(0x006119, 0xBE43);
	ChangeU2MB(0x00611C, 0xBE44);
	ChangeU2MB(0x00611E, 0xBE45);
	ChangeU2MB(0x006122, 0xBE46);
	ChangeU2MB(0x00612A, 0xBE47);
	ChangeU2MB(0x00612B, 0xBE48);
	ChangeU2MB(0x006131, 0xBE4A);
	ChangeU2MB(0x006135, 0xBE4B);
	ChangeU2MB(0x006136, 0xBE4C);
	ChangeU2MB(0x006139, 0xBE4E);
	ChangeU2MB(0x006141, 0xBE4F);
	ChangeU2MB(0x006145, 0xBE50);
	ChangeU2MB(0x006146, 0xBE51);
	ChangeU2MB(0x006149, 0xBE52);
	ChangeU2MB(0x00615E, 0xBE53);
	ChangeU2MB(0x006160, 0xBE54);
	ChangeU2MB(0x00616C, 0xBE55);
	ChangeU2MB(0x006172, 0xBE56);
	ChangeU2MB(0x006178, 0xBE57);
	ChangeU2MB(0x00617B, 0xBE58);
	ChangeU2MB(0x00617C, 0xBE59);
	ChangeU2MB(0x00617F, 0xBE5A);
	ChangeU2MB(0x006180, 0xBE5B);
	ChangeU2MB(0x006181, 0xBE5C);
	ChangeU2MB(0x006183, 0xBE5D);
	ChangeU2MB(0x006184, 0xBE5E);
	ChangeU2MB(0x00618B, 0xBE5F);
	ChangeU2MB(0x00618D, 0xBE60);
	ChangeU2MB(0x006192, 0xBE61);
	ChangeU2MB(0x006193, 0xBE62);
	ChangeU2MB(0x006197, 0xBE63);
	ChangeU2MB(0x00619C, 0xBE65);
	ChangeU2MB(0x00619D, 0xBE66);
	ChangeU2MB(0x00619F, 0xBE67);
	ChangeU2MB(0x0061A0, 0xBE68);
	ChangeU2MB(0x0061A5, 0xBE69);
	ChangeU2MB(0x0061A8, 0xBE6A);
	ChangeU2MB(0x0061AA, 0xBE6B);
	ChangeU2MB(0x0061AD, 0xBE6C);
	ChangeU2MB(0x0061B8, 0xBE6D);
	ChangeU2MB(0x0061B9, 0xBE6E);
	ChangeU2MB(0x0061BC, 0xBE6F);
	ChangeU2MB(0x0061C0, 0xBE70);
	ChangeU2MB(0x0061C1, 0xBE71);
	ChangeU2MB(0x0061C2, 0xBE72);
	ChangeU2MB(0x0061CE, 0xBE73);
	ChangeU2MB(0x0061CF, 0xBE74);
	ChangeU2MB(0x0061D5, 0xBE75);
	ChangeU2MB(0x0061DC, 0xBE76);
	ChangeU2MB(0x0061DD, 0xBE77);
	ChangeU2MB(0x0061DE, 0xBE78);
	ChangeU2MB(0x0061DF, 0xBE79);
	ChangeU2MB(0x0061E1, 0xBE7A);
	ChangeU2MB(0x0061E2, 0xBE7B);
	ChangeU2MB(0x0061E5, 0xBE7E);
	ChangeU2MB(0x0061E7, 0xBE7C);
	ChangeU2MB(0x0061E9, 0xBE7D);
	ChangeU2MB(0x0061EC, 0xBF21);
	ChangeU2MB(0x0061ED, 0xBF22);
	ChangeU2MB(0x0061EF, 0xBF23);
	ChangeU2MB(0x006201, 0xBF24);
	ChangeU2MB(0x006203, 0xBF25);
	ChangeU2MB(0x006204, 0xBF26);
	ChangeU2MB(0x006207, 0xBF27);
	ChangeU2MB(0x006215, 0xBF29);
	ChangeU2MB(0x00621C, 0xBF2A);
	ChangeU2MB(0x006220, 0xBF2B);
	ChangeU2MB(0x006222, 0xBF2C);
	ChangeU2MB(0x006223, 0xBF2D);
	ChangeU2MB(0x006227, 0xBF2E);
	ChangeU2MB(0x006229, 0xBF2F);
	ChangeU2MB(0x00622B, 0xBF30);
	ChangeU2MB(0x006239, 0xBF31);
	ChangeU2MB(0x00623D, 0xBF32);
	ChangeU2MB(0x006242, 0xBF33);
	ChangeU2MB(0x006243, 0xBF34);
	ChangeU2MB(0x006244, 0xBF35);
	ChangeU2MB(0x006246, 0xBF36);
	ChangeU2MB(0x00624C, 0xBF37);
	ChangeU2MB(0x006250, 0xBF38);
	ChangeU2MB(0x006251, 0xBF39);
	ChangeU2MB(0x006252, 0xBF3A);
	ChangeU2MB(0x006254, 0xBF3B);
	ChangeU2MB(0x006256, 0xBF3C);
	ChangeU2MB(0x00625A, 0xBF3D);
	ChangeU2MB(0x00625C, 0xBF3E);
	ChangeU2MB(0x006264, 0xBF3F);
	ChangeU2MB(0x00626D, 0xBF40);
	ChangeU2MB(0x00626F, 0xBF41);
	ChangeU2MB(0x006273, 0xBF42);
	ChangeU2MB(0x00627A, 0xBF43);
	ChangeU2MB(0x00627D, 0xBF44);
	ChangeU2MB(0x00628D, 0xBF45);
	ChangeU2MB(0x00628E, 0xBF46);
	ChangeU2MB(0x00628F, 0xBF47);
	ChangeU2MB(0x006290, 0xBF48);
	ChangeU2MB(0x0062A8, 0xBF4A);
	ChangeU2MB(0x0062B3, 0xBF4B);
	ChangeU2MB(0x0062B6, 0xBF4C);
	ChangeU2MB(0x0062B7, 0xBF4D);
	ChangeU2MB(0x0062BA, 0xBF4E);
	ChangeU2MB(0x0062BE, 0xBF4F);
	ChangeU2MB(0x0062BF, 0xBF50);
	ChangeU2MB(0x0062C4, 0xBF51);
	ChangeU2MB(0x0062CE, 0xBF52);
	ChangeU2MB(0x0062D5, 0xBF53);
	ChangeU2MB(0x0062D6, 0xBF54);
	ChangeU2MB(0x0062DA, 0xBF55);
	ChangeU2MB(0x0062EA, 0xBF56);
	ChangeU2MB(0x0062F2, 0xBF57);
	ChangeU2MB(0x0062F4, 0xBF58);
	ChangeU2MB(0x0062FC, 0xBF59);
	ChangeU2MB(0x0062FD, 0xBF5A);
	ChangeU2MB(0x006303, 0xBF5B);
	ChangeU2MB(0x006304, 0xBF5C);
	ChangeU2MB(0x00630A, 0xBF5D);
	ChangeU2MB(0x00630B, 0xBF5E);
	ChangeU2MB(0x00630D, 0xBF5F);
	ChangeU2MB(0x006310, 0xBF60);
	ChangeU2MB(0x006313, 0xBF61);
	ChangeU2MB(0x006316, 0xBF62);
	ChangeU2MB(0x006318, 0xBF63);
	ChangeU2MB(0x006329, 0xBF64);
	ChangeU2MB(0x00632A, 0xBF65);
	ChangeU2MB(0x00632D, 0xBF66);
	ChangeU2MB(0x006335, 0xBF67);
	ChangeU2MB(0x006336, 0xBF68);
	ChangeU2MB(0x006339, 0xBF69);
	ChangeU2MB(0x00633C, 0xBF6A);
	ChangeU2MB(0x006341, 0xBF6B);
	ChangeU2MB(0x006342, 0xBF6C);
	ChangeU2MB(0x006343, 0xBF6D);
	ChangeU2MB(0x006344, 0xBF6E);
	ChangeU2MB(0x006346, 0xBF6F);
	ChangeU2MB(0x00634A, 0xBF70);
	ChangeU2MB(0x00634B, 0xBF71);
	ChangeU2MB(0x00634E, 0xBF72);
	ChangeU2MB(0x006352, 0xBF73);
	ChangeU2MB(0x006353, 0xBF74);
	ChangeU2MB(0x006354, 0xBF75);
	ChangeU2MB(0x006358, 0xBF76);
	ChangeU2MB(0x00635B, 0xBF77);
	ChangeU2MB(0x006365, 0xBF78);
	ChangeU2MB(0x006366, 0xBF79);
	ChangeU2MB(0x00636C, 0xBF7A);
	ChangeU2MB(0x00636D, 0xBF7B);
	ChangeU2MB(0x006371, 0xBF7C);
	ChangeU2MB(0x006374, 0xBF7D);
	ChangeU2MB(0x006375, 0xBF7E);
	ChangeU2MB(0x006378, 0xC021);
	ChangeU2MB(0x00637C, 0xC022);
	ChangeU2MB(0x00637D, 0xC023);
	ChangeU2MB(0x00637F, 0xC024);
	ChangeU2MB(0x006382, 0xC025);
	ChangeU2MB(0x006384, 0xC026);
	ChangeU2MB(0x006387, 0xC027);
	ChangeU2MB(0x00638A, 0xC028);
	ChangeU2MB(0x006390, 0xC029);
	ChangeU2MB(0x006394, 0xC02A);
	ChangeU2MB(0x006395, 0xC02B);
	ChangeU2MB(0x006399, 0xC02C);
	ChangeU2MB(0x00639A, 0xC02D);
	ChangeU2MB(0x00639E, 0xC02E);
	ChangeU2MB(0x0063A4, 0xC02F);
	ChangeU2MB(0x0063A6, 0xC030);
	ChangeU2MB(0x0063AD, 0xC031);
	ChangeU2MB(0x0063AE, 0xC032);
	ChangeU2MB(0x0063AF, 0xC033);
	ChangeU2MB(0x0063BD, 0xC034);
	ChangeU2MB(0x0063C1, 0xC035);
	ChangeU2MB(0x0063C5, 0xC036);
	ChangeU2MB(0x0063C8, 0xC037);
	ChangeU2MB(0x0063CE, 0xC038);
	ChangeU2MB(0x0063D1, 0xC039);
	ChangeU2MB(0x0063D3, 0xC03A);
	ChangeU2MB(0x0063D4, 0xC03B);
	ChangeU2MB(0x0063D5, 0xC03C);
	ChangeU2MB(0x0063DC, 0xC03D);
	ChangeU2MB(0x0063E0, 0xC03E);
	ChangeU2MB(0x0063E5, 0xC03F);
	ChangeU2MB(0x0063EA, 0xC040);
	ChangeU2MB(0x0063EC, 0xC041);
	ChangeU2MB(0x0063F2, 0xC042);
	ChangeU2MB(0x0063F3, 0xC043);
	ChangeU2MB(0x0063F8, 0xC045);
	ChangeU2MB(0x0063F9, 0xC046);
	ChangeU2MB(0x006409, 0xC047);
	ChangeU2MB(0x00640A, 0xC048);
	ChangeU2MB(0x006410, 0xC049);
	ChangeU2MB(0x006412, 0xC04A);
	ChangeU2MB(0x006414, 0xC04B);
	ChangeU2MB(0x006418, 0xC04C);
	ChangeU2MB(0x00641E, 0xC04D);
	ChangeU2MB(0x006420, 0xC04E);
	ChangeU2MB(0x006422, 0xC04F);
	ChangeU2MB(0x006424, 0xC050);
	ChangeU2MB(0x006425, 0xC051);
	ChangeU2MB(0x006429, 0xC052);
	ChangeU2MB(0x00642A, 0xC053);
	ChangeU2MB(0x00642F, 0xC054);
	ChangeU2MB(0x006430, 0xC055);
	ChangeU2MB(0x006435, 0xC056);
	ChangeU2MB(0x00643D, 0xC057);
	ChangeU2MB(0x00643F, 0xC058);
	ChangeU2MB(0x00644B, 0xC059);
	ChangeU2MB(0x00644F, 0xC05A);
	ChangeU2MB(0x006451, 0xC05B);
	ChangeU2MB(0x006452, 0xC05C);
	ChangeU2MB(0x006453, 0xC05D);
	ChangeU2MB(0x006454, 0xC05E);
	ChangeU2MB(0x00645A, 0xC05F);
	ChangeU2MB(0x00645B, 0xC060);
	ChangeU2MB(0x00645C, 0xC061);
	ChangeU2MB(0x00645D, 0xC062);
	ChangeU2MB(0x00645F, 0xC063);
	ChangeU2MB(0x006461, 0xC065);
	ChangeU2MB(0x006463, 0xC066);
	ChangeU2MB(0x00646D, 0xC067);
	ChangeU2MB(0x006473, 0xC068);
	ChangeU2MB(0x006474, 0xC069);
	ChangeU2MB(0x00647B, 0xC06A);
	ChangeU2MB(0x00647D, 0xC06B);
	ChangeU2MB(0x006485, 0xC06C);
	ChangeU2MB(0x006487, 0xC06D);
	ChangeU2MB(0x00648F, 0xC06E);
	ChangeU2MB(0x006490, 0xC06F);
	ChangeU2MB(0x006491, 0xC070);
	ChangeU2MB(0x006498, 0xC071);
	ChangeU2MB(0x006499, 0xC072);
	ChangeU2MB(0x00649B, 0xC073);
	ChangeU2MB(0x00649F, 0xC075);
	ChangeU2MB(0x0064A1, 0xC076);
	ChangeU2MB(0x0064A3, 0xC077);
	ChangeU2MB(0x0064A6, 0xC078);
	ChangeU2MB(0x0064A8, 0xC079);
	ChangeU2MB(0x0064AC, 0xC07A);
	ChangeU2MB(0x0064B3, 0xC07B);
	ChangeU2MB(0x0064BD, 0xC07C);
	ChangeU2MB(0x0064BE, 0xC07D);
	ChangeU2MB(0x0064BF, 0xC07E);
	ChangeU2MB(0x0064C4, 0xC121);
	ChangeU2MB(0x0064C9, 0xC122);
	ChangeU2MB(0x0064CA, 0xC123);
	ChangeU2MB(0x0064CB, 0xC124);
	ChangeU2MB(0x0064CC, 0xC125);
	ChangeU2MB(0x0064D0, 0xC127);
	ChangeU2MB(0x0064D1, 0xC128);
	ChangeU2MB(0x0064D5, 0xC129);
	ChangeU2MB(0x0064D7, 0xC12A);
	ChangeU2MB(0x0064E4, 0xC12B);
	ChangeU2MB(0x0064E5, 0xC12C);
	ChangeU2MB(0x0064E9, 0xC12D);
	ChangeU2MB(0x0064EA, 0xC12E);
	ChangeU2MB(0x0064ED, 0xC12F);
	ChangeU2MB(0x0064F0, 0xC130);
	ChangeU2MB(0x0064F5, 0xC131);
	ChangeU2MB(0x0064F7, 0xC132);
	ChangeU2MB(0x0064FB, 0xC133);
	ChangeU2MB(0x0064FF, 0xC134);
	ChangeU2MB(0x006501, 0xC135);
	ChangeU2MB(0x006504, 0xC136);
	ChangeU2MB(0x006508, 0xC137);
	ChangeU2MB(0x006509, 0xC138);
	ChangeU2MB(0x00650A, 0xC139);
	ChangeU2MB(0x00650F, 0xC13A);
	ChangeU2MB(0x006513, 0xC13B);
	ChangeU2MB(0x006514, 0xC13C);
	ChangeU2MB(0x006516, 0xC13D);
	ChangeU2MB(0x006519, 0xC13E);
	ChangeU2MB(0x00651B, 0xC13F);
	ChangeU2MB(0x00651E, 0xC140);
	ChangeU2MB(0x00651F, 0xC141);
	ChangeU2MB(0x006522, 0xC142);
	ChangeU2MB(0x006526, 0xC143);
	ChangeU2MB(0x006529, 0xC144);
	ChangeU2MB(0x00652E, 0xC145);
	ChangeU2MB(0x006531, 0xC146);
	ChangeU2MB(0x00653A, 0xC147);
	ChangeU2MB(0x00653C, 0xC148);
	ChangeU2MB(0x00653D, 0xC149);
	ChangeU2MB(0x006543, 0xC14A);
	ChangeU2MB(0x006547, 0xC14B);
	ChangeU2MB(0x006549, 0xC14C);
	ChangeU2MB(0x006550, 0xC14D);
	ChangeU2MB(0x006552, 0xC14E);
	ChangeU2MB(0x006554, 0xC14F);
	ChangeU2MB(0x00655F, 0xC150);
	ChangeU2MB(0x006560, 0xC151);
	ChangeU2MB(0x006567, 0xC152);
	ChangeU2MB(0x00656B, 0xC153);
	ChangeU2MB(0x00657A, 0xC154);
	ChangeU2MB(0x00657D, 0xC155);
	ChangeU2MB(0x006581, 0xC156);
	ChangeU2MB(0x006585, 0xC157);
	ChangeU2MB(0x00658A, 0xC158);
	ChangeU2MB(0x006592, 0xC159);
	ChangeU2MB(0x006595, 0xC15A);
	ChangeU2MB(0x006598, 0xC15B);
	ChangeU2MB(0x00659D, 0xC15C);
	ChangeU2MB(0x0065A0, 0xC15D);
	ChangeU2MB(0x0065A3, 0xC15E);
	ChangeU2MB(0x0065A6, 0xC15F);
	ChangeU2MB(0x0065AE, 0xC160);
	ChangeU2MB(0x0065B2, 0xC161);
	ChangeU2MB(0x0065B3, 0xC162);
	ChangeU2MB(0x0065B4, 0xC163);
	ChangeU2MB(0x0065BF, 0xC164);
	ChangeU2MB(0x0065C2, 0xC165);
	ChangeU2MB(0x0065C8, 0xC166);
	ChangeU2MB(0x0065C9, 0xC167);
	ChangeU2MB(0x0065CE, 0xC168);
	ChangeU2MB(0x0065D0, 0xC169);
	ChangeU2MB(0x0065D4, 0xC16A);
	ChangeU2MB(0x0065D6, 0xC16B);
	ChangeU2MB(0x0065D8, 0xC16C);
	ChangeU2MB(0x0065DF, 0xC16D);
	ChangeU2MB(0x0065F0, 0xC16E);
	ChangeU2MB(0x0065F2, 0xC16F);
	ChangeU2MB(0x0065F4, 0xC170);
	ChangeU2MB(0x0065F5, 0xC171);
	ChangeU2MB(0x0065F9, 0xC172);
	ChangeU2MB(0x0065FE, 0xC173);
	ChangeU2MB(0x0065FF, 0xC174);
	ChangeU2MB(0x006604, 0xC176);
	ChangeU2MB(0x006608, 0xC177);
	ChangeU2MB(0x00660D, 0xC179);
	ChangeU2MB(0x006611, 0xC17A);
	ChangeU2MB(0x006612, 0xC17B);
	ChangeU2MB(0x006616, 0xC17D);
	ChangeU2MB(0x00661D, 0xC17E);
	ChangeU2MB(0x006621, 0xC222);
	ChangeU2MB(0x006622, 0xC223);
	ChangeU2MB(0x006623, 0xC224);
	ChangeU2MB(0x006626, 0xC226);
	ChangeU2MB(0x006629, 0xC227);
	ChangeU2MB(0x00662A, 0xC228);
	ChangeU2MB(0x00662B, 0xC229);
	ChangeU2MB(0x00662C, 0xC22A);
	ChangeU2MB(0x006630, 0xC22C);
	ChangeU2MB(0x006633, 0xC22E);
	ChangeU2MB(0x006637, 0xC230);
	ChangeU2MB(0x006639, 0xC22F);
	ChangeU2MB(0x006640, 0xC231);
	ChangeU2MB(0x006645, 0xC232);
	ChangeU2MB(0x006646, 0xC233);
	ChangeU2MB(0x00664A, 0xC234);
	ChangeU2MB(0x00664C, 0xC235);
	ChangeU2MB(0x00664E, 0xC237);
	ChangeU2MB(0x006651, 0xC236);
	ChangeU2MB(0x006658, 0xC239);
	ChangeU2MB(0x00665B, 0xC23B);
	ChangeU2MB(0x00665C, 0xC23C);
	ChangeU2MB(0x006660, 0xC23D);
	ChangeU2MB(0x006661, 0xC23E);
	ChangeU2MB(0x00666A, 0xC240);
	ChangeU2MB(0x00666B, 0xC241);
	ChangeU2MB(0x00666C, 0xC242);
	ChangeU2MB(0x006675, 0xC245);
	ChangeU2MB(0x006677, 0xC247);
	ChangeU2MB(0x006678, 0xC248);
	ChangeU2MB(0x006679, 0xC249);
	ChangeU2MB(0x00667B, 0xC24A);
	ChangeU2MB(0x00667C, 0xC24C);
	ChangeU2MB(0x00667E, 0xC243);
	ChangeU2MB(0x00667F, 0xC246);
	ChangeU2MB(0x006680, 0xC24B);
	ChangeU2MB(0x00668B, 0xC24D);
	ChangeU2MB(0x00668C, 0xC24E);
	ChangeU2MB(0x00668D, 0xC24F);
	ChangeU2MB(0x006690, 0xC250);
	ChangeU2MB(0x006692, 0xC251);
	ChangeU2MB(0x00669A, 0xC253);
	ChangeU2MB(0x00669B, 0xC254);
	ChangeU2MB(0x00669C, 0xC255);
	ChangeU2MB(0x00669F, 0xC256);
	ChangeU2MB(0x0066A4, 0xC258);
	ChangeU2MB(0x0066AD, 0xC259);
	ChangeU2MB(0x0066B1, 0xC25A);
	ChangeU2MB(0x0066B5, 0xC25C);
	ChangeU2MB(0x0066BB, 0xC25D);
	ChangeU2MB(0x0066C0, 0xC25F);
	ChangeU2MB(0x0066C2, 0xC260);
	ChangeU2MB(0x0066C3, 0xC261);
	ChangeU2MB(0x0066C8, 0xC262);
	ChangeU2MB(0x0066CC, 0xC263);
	ChangeU2MB(0x0066CE, 0xC264);
	ChangeU2MB(0x0066CF, 0xC265);
	ChangeU2MB(0x0066D4, 0xC266);
	ChangeU2MB(0x0066DB, 0xC267);
	ChangeU2MB(0x0066DF, 0xC268);
	ChangeU2MB(0x0066E8, 0xC269);
	ChangeU2MB(0x0066EB, 0xC26A);
	ChangeU2MB(0x0066EC, 0xC26B);
	ChangeU2MB(0x0066EE, 0xC26C);
	ChangeU2MB(0x006705, 0xC26E);
	ChangeU2MB(0x006707, 0xC26F);
	ChangeU2MB(0x006713, 0xC271);
	ChangeU2MB(0x006719, 0xC272);
	ChangeU2MB(0x00671C, 0xC273);
	ChangeU2MB(0x006720, 0xC274);
	ChangeU2MB(0x006722, 0xC275);
	ChangeU2MB(0x006733, 0xC276);
	ChangeU2MB(0x00673E, 0xC277);
	ChangeU2MB(0x006745, 0xC278);
	ChangeU2MB(0x006747, 0xC279);
	ChangeU2MB(0x006748, 0xC27A);
	ChangeU2MB(0x00674C, 0xC27B);
	ChangeU2MB(0x006754, 0xC27C);
	ChangeU2MB(0x006755, 0xC27D);
	ChangeU2MB(0x00675D, 0xC27E);
	ChangeU2MB(0x00676C, 0xC322);
	ChangeU2MB(0x00676E, 0xC323);
	ChangeU2MB(0x006774, 0xC324);
	ChangeU2MB(0x006776, 0xC325);
	ChangeU2MB(0x00677B, 0xC326);
	ChangeU2MB(0x006781, 0xC327);
	ChangeU2MB(0x006784, 0xC328);
	ChangeU2MB(0x00678E, 0xC329);
	ChangeU2MB(0x00678F, 0xC32A);
	ChangeU2MB(0x006791, 0xC32B);
	ChangeU2MB(0x006793, 0xC32C);
	ChangeU2MB(0x006796, 0xC32D);
	ChangeU2MB(0x006798, 0xC32E);
	ChangeU2MB(0x006799, 0xC32F);
	ChangeU2MB(0x00679B, 0xC330);
	ChangeU2MB(0x0067B0, 0xC331);
	ChangeU2MB(0x0067B1, 0xC332);
	ChangeU2MB(0x0067B2, 0xC333);
	ChangeU2MB(0x0067B5, 0xC334);
	ChangeU2MB(0x0067BC, 0xC336);
	ChangeU2MB(0x0067BD, 0xC337);
	ChangeU2MB(0x0067C2, 0xC33A);
	ChangeU2MB(0x0067C3, 0xC33B);
	ChangeU2MB(0x0067C5, 0xC33C);
	ChangeU2MB(0x0067C8, 0xC33D);
	ChangeU2MB(0x0067C9, 0xC33E);
	ChangeU2MB(0x0067D2, 0xC33F);
	ChangeU2MB(0x0067D7, 0xC340);
	ChangeU2MB(0x0067D9, 0xC341);
	ChangeU2MB(0x0067DC, 0xC342);
	ChangeU2MB(0x0067E1, 0xC343);
	ChangeU2MB(0x0067E6, 0xC344);
	ChangeU2MB(0x0067F0, 0xC345);
	ChangeU2MB(0x0067F2, 0xC346);
	ChangeU2MB(0x0067F6, 0xC347);
	ChangeU2MB(0x0067F7, 0xC348);
	ChangeU2MB(0x0067F9, 0xC338);
	ChangeU2MB(0x006814, 0xC34A);
	ChangeU2MB(0x006819, 0xC34B);
	ChangeU2MB(0x00681D, 0xC34C);
	ChangeU2MB(0x00681F, 0xC34D);
	ChangeU2MB(0x006827, 0xC34F);
	ChangeU2MB(0x006828, 0xC34E);
	ChangeU2MB(0x00682C, 0xC350);
	ChangeU2MB(0x00682D, 0xC351);
	ChangeU2MB(0x00682F, 0xC352);
	ChangeU2MB(0x006830, 0xC353);
	ChangeU2MB(0x006831, 0xC354);
	ChangeU2MB(0x006833, 0xC355);
	ChangeU2MB(0x00683B, 0xC356);
	ChangeU2MB(0x00683F, 0xC357);
	ChangeU2MB(0x006845, 0xC359);
	ChangeU2MB(0x00684A, 0xC35A);
	ChangeU2MB(0x00684C, 0xC35B);
	ChangeU2MB(0x006855, 0xC35C);
	ChangeU2MB(0x006857, 0xC35D);
	ChangeU2MB(0x006858, 0xC35E);
	ChangeU2MB(0x00685B, 0xC35F);
	ChangeU2MB(0x00686B, 0xC360);
	ChangeU2MB(0x00686E, 0xC361);
	ChangeU2MB(0x00686F, 0xC362);
	ChangeU2MB(0x006870, 0xC363);
	ChangeU2MB(0x006871, 0xC364);
	ChangeU2MB(0x006872, 0xC365);
	ChangeU2MB(0x006875, 0xC366);
	ChangeU2MB(0x006879, 0xC367);
	ChangeU2MB(0x00687A, 0xC368);
	ChangeU2MB(0x00687B, 0xC369);
	ChangeU2MB(0x00687C, 0xC36A);
	ChangeU2MB(0x006882, 0xC36B);
	ChangeU2MB(0x006884, 0xC36C);
	ChangeU2MB(0x006886, 0xC36D);
	ChangeU2MB(0x006888, 0xC36E);
	ChangeU2MB(0x006896, 0xC36F);
	ChangeU2MB(0x006898, 0xC370);
	ChangeU2MB(0x00689A, 0xC371);
	ChangeU2MB(0x00689C, 0xC372);
	ChangeU2MB(0x0068A1, 0xC373);
	ChangeU2MB(0x0068A3, 0xC374);
	ChangeU2MB(0x0068A5, 0xC375);
	ChangeU2MB(0x0068A9, 0xC376);
	ChangeU2MB(0x0068AA, 0xC377);
	ChangeU2MB(0x0068AE, 0xC378);
	ChangeU2MB(0x0068B2, 0xC379);
	ChangeU2MB(0x0068BB, 0xC37A);
	ChangeU2MB(0x0068C5, 0xC37B);
	ChangeU2MB(0x0068CC, 0xC37D);
	ChangeU2MB(0x0068D0, 0xC421);
	ChangeU2MB(0x0068D1, 0xC422);
	ChangeU2MB(0x0068D3, 0xC423);
	ChangeU2MB(0x0068D6, 0xC424);
	ChangeU2MB(0x0068D9, 0xC425);
	ChangeU2MB(0x0068DC, 0xC426);
	ChangeU2MB(0x0068DD, 0xC427);
	ChangeU2MB(0x0068E5, 0xC428);
	ChangeU2MB(0x0068E8, 0xC429);
	ChangeU2MB(0x0068EA, 0xC42A);
	ChangeU2MB(0x0068EB, 0xC42B);
	ChangeU2MB(0x0068EC, 0xC42C);
	ChangeU2MB(0x0068ED, 0xC42D);
	ChangeU2MB(0x0068F0, 0xC42E);
	ChangeU2MB(0x0068F1, 0xC42F);
	ChangeU2MB(0x0068F5, 0xC430);
	ChangeU2MB(0x0068F6, 0xC431);
	ChangeU2MB(0x0068FB, 0xC432);
	ChangeU2MB(0x0068FC, 0xC433);
	ChangeU2MB(0x0068FD, 0xC434);
	ChangeU2MB(0x006906, 0xC435);
	ChangeU2MB(0x006909, 0xC436);
	ChangeU2MB(0x00690A, 0xC437);
	ChangeU2MB(0x006910, 0xC438);
	ChangeU2MB(0x006911, 0xC439);
	ChangeU2MB(0x006913, 0xC43A);
	ChangeU2MB(0x006916, 0xC43B);
	ChangeU2MB(0x006917, 0xC43C);
	ChangeU2MB(0x006931, 0xC43D);
	ChangeU2MB(0x006933, 0xC43E);
	ChangeU2MB(0x006935, 0xC43F);
	ChangeU2MB(0x006938, 0xC440);
	ChangeU2MB(0x00693B, 0xC441);
	ChangeU2MB(0x006942, 0xC442);
	ChangeU2MB(0x006945, 0xC443);
	ChangeU2MB(0x006949, 0xC444);
	ChangeU2MB(0x00694E, 0xC445);
	ChangeU2MB(0x006957, 0xC446);
	ChangeU2MB(0x00695B, 0xC447);
	ChangeU2MB(0x006963, 0xC448);
	ChangeU2MB(0x006964, 0xC449);
	ChangeU2MB(0x006965, 0xC44A);
	ChangeU2MB(0x006966, 0xC44B);
	ChangeU2MB(0x006969, 0xC44D);
	ChangeU2MB(0x00696C, 0xC44E);
	ChangeU2MB(0x006970, 0xC44F);
	ChangeU2MB(0x006971, 0xC450);
	ChangeU2MB(0x006972, 0xC451);
	ChangeU2MB(0x00697A, 0xC452);
	ChangeU2MB(0x00697B, 0xC453);
	ChangeU2MB(0x00697F, 0xC454);
	ChangeU2MB(0x006980, 0xC455);
	ChangeU2MB(0x00698D, 0xC456);
	ChangeU2MB(0x006992, 0xC457);
	ChangeU2MB(0x006996, 0xC458);
	ChangeU2MB(0x0069A1, 0xC45A);
	ChangeU2MB(0x0069A5, 0xC45B);
	ChangeU2MB(0x0069A6, 0xC45C);
	ChangeU2MB(0x0069A8, 0xC45D);
	ChangeU2MB(0x0069AB, 0xC45E);
	ChangeU2MB(0x0069AD, 0xC45F);
	ChangeU2MB(0x0069AF, 0xC460);
	ChangeU2MB(0x0069B7, 0xC461);
	ChangeU2MB(0x0069B8, 0xC462);
	ChangeU2MB(0x0069BA, 0xC463);
	ChangeU2MB(0x0069BC, 0xC464);
	ChangeU2MB(0x0069C5, 0xC465);
	ChangeU2MB(0x0069C8, 0xC466);
	ChangeU2MB(0x0069D1, 0xC467);
	ChangeU2MB(0x0069D6, 0xC468);
	ChangeU2MB(0x0069D7, 0xC469);
	ChangeU2MB(0x0069E5, 0xC46B);
	ChangeU2MB(0x0069EE, 0xC46C);
	ChangeU2MB(0x0069EF, 0xC46D);
	ChangeU2MB(0x0069F1, 0xC46E);
	ChangeU2MB(0x0069F3, 0xC46F);
	ChangeU2MB(0x0069F5, 0xC470);
	ChangeU2MB(0x0069FE, 0xC471);
	ChangeU2MB(0x006A00, 0xC472);
	ChangeU2MB(0x006A01, 0xC473);
	ChangeU2MB(0x006A03, 0xC474);
	ChangeU2MB(0x006A0F, 0xC475);
	ChangeU2MB(0x006A11, 0xC476);
	ChangeU2MB(0x006A15, 0xC477);
	ChangeU2MB(0x006A1A, 0xC478);
	ChangeU2MB(0x006A1D, 0xC479);
	ChangeU2MB(0x006A20, 0xC47A);
	ChangeU2MB(0x006A24, 0xC47B);
	ChangeU2MB(0x006A28, 0xC47C);
	ChangeU2MB(0x006A32, 0xC47E);
	ChangeU2MB(0x006A34, 0xC521);
	ChangeU2MB(0x006A37, 0xC522);
	ChangeU2MB(0x006A3B, 0xC523);
	ChangeU2MB(0x006A3E, 0xC524);
	ChangeU2MB(0x006A3F, 0xC525);
	ChangeU2MB(0x006A45, 0xC526);
	ChangeU2MB(0x006A49, 0xC528);
	ChangeU2MB(0x006A4A, 0xC529);
	ChangeU2MB(0x006A4E, 0xC52A);
	ChangeU2MB(0x006A50, 0xC52B);
	ChangeU2MB(0x006A51, 0xC52C);
	ChangeU2MB(0x006A52, 0xC52D);
	ChangeU2MB(0x006A55, 0xC52E);
	ChangeU2MB(0x006A56, 0xC52F);
	ChangeU2MB(0x006A5B, 0xC530);
	ChangeU2MB(0x006A64, 0xC531);
	ChangeU2MB(0x006A67, 0xC532);
	ChangeU2MB(0x006A6A, 0xC533);
	ChangeU2MB(0x006A71, 0xC534);
	ChangeU2MB(0x006A81, 0xC537);
	ChangeU2MB(0x006A83, 0xC538);
	ChangeU2MB(0x006A86, 0xC539);
	ChangeU2MB(0x006A87, 0xC53A);
	ChangeU2MB(0x006A89, 0xC53B);
	ChangeU2MB(0x006A8B, 0xC53C);
	ChangeU2MB(0x006A91, 0xC53D);
	ChangeU2MB(0x006A9B, 0xC53E);
	ChangeU2MB(0x006A9D, 0xC53F);
	ChangeU2MB(0x006A9E, 0xC540);
	ChangeU2MB(0x006A9F, 0xC541);
	ChangeU2MB(0x006AA5, 0xC542);
	ChangeU2MB(0x006AAB, 0xC543);
	ChangeU2MB(0x006AAF, 0xC544);
	ChangeU2MB(0x006AB0, 0xC545);
	ChangeU2MB(0x006AB1, 0xC546);
	ChangeU2MB(0x006AB4, 0xC547);
	ChangeU2MB(0x006ABD, 0xC548);
	ChangeU2MB(0x006ABE, 0xC549);
	ChangeU2MB(0x006ABF, 0xC54A);
	ChangeU2MB(0x006AC6, 0xC54B);
	ChangeU2MB(0x006AC8, 0xC54D);
	ChangeU2MB(0x006AC9, 0xC54C);
	ChangeU2MB(0x006ACC, 0xC54E);
	ChangeU2MB(0x006AD0, 0xC54F);
	ChangeU2MB(0x006AD4, 0xC550);
	ChangeU2MB(0x006AD5, 0xC551);
	ChangeU2MB(0x006AD6, 0xC552);
	ChangeU2MB(0x006ADC, 0xC553);
	ChangeU2MB(0x006ADD, 0xC554);
	ChangeU2MB(0x006AE7, 0xC556);
	ChangeU2MB(0x006AEC, 0xC557);
	ChangeU2MB(0x006AF0, 0xC558);
	ChangeU2MB(0x006AF1, 0xC559);
	ChangeU2MB(0x006AF2, 0xC55A);
	ChangeU2MB(0x006AFC, 0xC55B);
	ChangeU2MB(0x006AFD, 0xC55C);
	ChangeU2MB(0x006B02, 0xC55D);
	ChangeU2MB(0x006B03, 0xC55E);
	ChangeU2MB(0x006B06, 0xC55F);
	ChangeU2MB(0x006B07, 0xC560);
	ChangeU2MB(0x006B09, 0xC561);
	ChangeU2MB(0x006B0F, 0xC562);
	ChangeU2MB(0x006B10, 0xC563);
	ChangeU2MB(0x006B11, 0xC564);
	ChangeU2MB(0x006B17, 0xC565);
	ChangeU2MB(0x006B1B, 0xC566);
	ChangeU2MB(0x006B1E, 0xC567);
	ChangeU2MB(0x006B24, 0xC568);
	ChangeU2MB(0x006B28, 0xC569);
	ChangeU2MB(0x006B2B, 0xC56A);
	ChangeU2MB(0x006B2C, 0xC56B);
	ChangeU2MB(0x006B2F, 0xC56C);
	ChangeU2MB(0x006B35, 0xC56D);
	ChangeU2MB(0x006B36, 0xC56E);
	ChangeU2MB(0x006B3B, 0xC56F);
	ChangeU2MB(0x006B3F, 0xC570);
	ChangeU2MB(0x006B46, 0xC571);
	ChangeU2MB(0x006B4A, 0xC572);
	ChangeU2MB(0x006B4D, 0xC573);
	ChangeU2MB(0x006B52, 0xC574);
	ChangeU2MB(0x006B56, 0xC575);
	ChangeU2MB(0x006B58, 0xC576);
	ChangeU2MB(0x006B5D, 0xC577);
	ChangeU2MB(0x006B60, 0xC578);
	ChangeU2MB(0x006B67, 0xC579);
	ChangeU2MB(0x006B6B, 0xC57A);
	ChangeU2MB(0x006B6E, 0xC57B);
	ChangeU2MB(0x006B70, 0xC57C);
	ChangeU2MB(0x006B75, 0xC57D);
	ChangeU2MB(0x006B7D, 0xC57E);
	ChangeU2MB(0x006B7E, 0xC621);
	ChangeU2MB(0x006B82, 0xC622);
	ChangeU2MB(0x006B85, 0xC623);
	ChangeU2MB(0x006B97, 0xC624);
	ChangeU2MB(0x006B9B, 0xC625);
	ChangeU2MB(0x006B9F, 0xC626);
	ChangeU2MB(0x006BA0, 0xC627);
	ChangeU2MB(0x006BA2, 0xC628);
	ChangeU2MB(0x006BA3, 0xC629);
	ChangeU2MB(0x006BA8, 0xC62A);
	ChangeU2MB(0x006BA9, 0xC62B);
	ChangeU2MB(0x006BAC, 0xC62C);
	ChangeU2MB(0x006BAD, 0xC62D);
	ChangeU2MB(0x006BAE, 0xC62E);
	ChangeU2MB(0x006BB0, 0xC62F);
	ChangeU2MB(0x006BB8, 0xC630);
	ChangeU2MB(0x006BB9, 0xC631);
	ChangeU2MB(0x006BBD, 0xC632);
	ChangeU2MB(0x006BBE, 0xC633);
	ChangeU2MB(0x006BC3, 0xC634);
	ChangeU2MB(0x006BC4, 0xC635);
	ChangeU2MB(0x006BC9, 0xC636);
	ChangeU2MB(0x006BCC, 0xC637);
	ChangeU2MB(0x006BDA, 0xC639);
	ChangeU2MB(0x006BE1, 0xC63A);
	ChangeU2MB(0x006BE3, 0xC63B);
	ChangeU2MB(0x006BE6, 0xC63C);
	ChangeU2MB(0x006BE7, 0xC63D);
	ChangeU2MB(0x006BEE, 0xC63E);
	ChangeU2MB(0x006BF1, 0xC63F);
	ChangeU2MB(0x006BF7, 0xC640);
	ChangeU2MB(0x006BF9, 0xC641);
	ChangeU2MB(0x006BFF, 0xC642);
	ChangeU2MB(0x006C02, 0xC643);
	ChangeU2MB(0x006C04, 0xC644);
	ChangeU2MB(0x006C05, 0xC645);
	ChangeU2MB(0x006C09, 0xC646);
	ChangeU2MB(0x006C0D, 0xC647);
	ChangeU2MB(0x006C0E, 0xC648);
	ChangeU2MB(0x006C10, 0xC649);
	ChangeU2MB(0x006C12, 0xC64A);
	ChangeU2MB(0x006C19, 0xC64B);
	ChangeU2MB(0x006C1F, 0xC64C);
	ChangeU2MB(0x006C26, 0xC64D);
	ChangeU2MB(0x006C27, 0xC64E);
	ChangeU2MB(0x006C28, 0xC64F);
	ChangeU2MB(0x006C2C, 0xC650);
	ChangeU2MB(0x006C2E, 0xC651);
	ChangeU2MB(0x006C33, 0xC652);
	ChangeU2MB(0x006C35, 0xC653);
	ChangeU2MB(0x006C36, 0xC654);
	ChangeU2MB(0x006C3A, 0xC655);
	ChangeU2MB(0x006C3B, 0xC656);
	ChangeU2MB(0x006C4A, 0xC658);
	ChangeU2MB(0x006C4B, 0xC659);
	ChangeU2MB(0x006C4D, 0xC65A);
	ChangeU2MB(0x006C4F, 0xC65B);
	ChangeU2MB(0x006C52, 0xC65C);
	ChangeU2MB(0x006C54, 0xC65D);
	ChangeU2MB(0x006C59, 0xC65E);
	ChangeU2MB(0x006C5B, 0xC65F);
	ChangeU2MB(0x006C67, 0xC736);
	ChangeU2MB(0x006C6B, 0xC661);
	ChangeU2MB(0x006C6D, 0xC662);
	ChangeU2MB(0x006C74, 0xC664);
	ChangeU2MB(0x006C76, 0xC665);
	ChangeU2MB(0x006C78, 0xC666);
	ChangeU2MB(0x006C79, 0xC667);
	ChangeU2MB(0x006C7B, 0xC668);
	ChangeU2MB(0x006C85, 0xC669);
	ChangeU2MB(0x006C87, 0xC66B);
	ChangeU2MB(0x006C89, 0xC66C);
	ChangeU2MB(0x006C94, 0xC66D);
	ChangeU2MB(0x006C95, 0xC66E);
	ChangeU2MB(0x006C97, 0xC66F);
	ChangeU2MB(0x006C98, 0xC670);
	ChangeU2MB(0x006C9C, 0xC671);
	ChangeU2MB(0x006C9F, 0xC672);
	ChangeU2MB(0x006CB0, 0xC673);
	ChangeU2MB(0x006CB2, 0xC674);
	ChangeU2MB(0x006CB4, 0xC675);
	ChangeU2MB(0x006CC2, 0xC676);
	ChangeU2MB(0x006CC6, 0xC677);
	ChangeU2MB(0x006CCD, 0xC678);
	ChangeU2MB(0x006CCF, 0xC679);
	ChangeU2MB(0x006CD0, 0xC67A);
	ChangeU2MB(0x006CD1, 0xC67B);
	ChangeU2MB(0x006CD2, 0xC67C);
	ChangeU2MB(0x006CD4, 0xC67D);
	ChangeU2MB(0x006CD6, 0xC67E);
	ChangeU2MB(0x006CDC, 0xC722);
	ChangeU2MB(0x006CE0, 0xC723);
	ChangeU2MB(0x006CE7, 0xC724);
	ChangeU2MB(0x006CE9, 0xC725);
	ChangeU2MB(0x006CEB, 0xC726);
	ChangeU2MB(0x006CEC, 0xC727);
	ChangeU2MB(0x006CEE, 0xC728);
	ChangeU2MB(0x006CF2, 0xC729);
	ChangeU2MB(0x006CF4, 0xC72A);
	ChangeU2MB(0x006D07, 0xC72C);
	ChangeU2MB(0x006D0A, 0xC72D);
	ChangeU2MB(0x006D0E, 0xC72E);
	ChangeU2MB(0x006D0F, 0xC72F);
	ChangeU2MB(0x006D11, 0xC730);
	ChangeU2MB(0x006D13, 0xC731);
	ChangeU2MB(0x006D1A, 0xC732);
	ChangeU2MB(0x006D26, 0xC733);
	ChangeU2MB(0x006D27, 0xC734);
	ChangeU2MB(0x006D28, 0xC735);
	ChangeU2MB(0x006D2E, 0xC737);
	ChangeU2MB(0x006D2F, 0xC738);
	ChangeU2MB(0x006D31, 0xC739);
	ChangeU2MB(0x006D39, 0xC73A);
	ChangeU2MB(0x006D3C, 0xC73B);
	ChangeU2MB(0x006D3F, 0xC73C);
	ChangeU2MB(0x006D57, 0xC73D);
	ChangeU2MB(0x006D5E, 0xC73E);
	ChangeU2MB(0x006D5F, 0xC73F);
	ChangeU2MB(0x006D61, 0xC740);
	ChangeU2MB(0x006D65, 0xC741);
	ChangeU2MB(0x006D67, 0xC742);
	ChangeU2MB(0x006D70, 0xC744);
	ChangeU2MB(0x006D7C, 0xC745);
	ChangeU2MB(0x006D82, 0xC746);
	ChangeU2MB(0x006D91, 0xC748);
	ChangeU2MB(0x006D92, 0xC749);
	ChangeU2MB(0x006D94, 0xC74A);
	ChangeU2MB(0x006D97, 0xC74C);
	ChangeU2MB(0x006D98, 0xC74D);
	ChangeU2MB(0x006DAA, 0xC74E);
	ChangeU2MB(0x006DB4, 0xC750);
	ChangeU2MB(0x006DB7, 0xC751);
	ChangeU2MB(0x006DB9, 0xC752);
	ChangeU2MB(0x006DBD, 0xC753);
	ChangeU2MB(0x006DBF, 0xC754);
	ChangeU2MB(0x006DC4, 0xC755);
	ChangeU2MB(0x006DC8, 0xC756);
	ChangeU2MB(0x006DCA, 0xC757);
	ChangeU2MB(0x006DCE, 0xC758);
	ChangeU2MB(0x006DD6, 0xC75A);
	ChangeU2MB(0x006DDB, 0xC75B);
	ChangeU2MB(0x006DDD, 0xC75C);
	ChangeU2MB(0x006DDF, 0xC75D);
	ChangeU2MB(0x006DE0, 0xC75E);
	ChangeU2MB(0x006DE2, 0xC75F);
	ChangeU2MB(0x006DE5, 0xC760);
	ChangeU2MB(0x006DE9, 0xC761);
	ChangeU2MB(0x006DEF, 0xC762);
	ChangeU2MB(0x006DF0, 0xC763);
	ChangeU2MB(0x006DF4, 0xC764);
	ChangeU2MB(0x006DF6, 0xC765);
	ChangeU2MB(0x006E00, 0xC767);
	ChangeU2MB(0x006E04, 0xC768);
	ChangeU2MB(0x006E1E, 0xC769);
	ChangeU2MB(0x006E22, 0xC76A);
	ChangeU2MB(0x006E32, 0xC76C);
	ChangeU2MB(0x006E36, 0xC76D);
	ChangeU2MB(0x006E3B, 0xC76F);
	ChangeU2MB(0x006E44, 0xC771);
	ChangeU2MB(0x006E45, 0xC772);
	ChangeU2MB(0x006E48, 0xC773);
	ChangeU2MB(0x006E49, 0xC774);
	ChangeU2MB(0x006E4B, 0xC775);
	ChangeU2MB(0x006E4F, 0xC776);
	ChangeU2MB(0x006E51, 0xC777);
	ChangeU2MB(0x006E52, 0xC778);
	ChangeU2MB(0x006E53, 0xC779);
	ChangeU2MB(0x006E54, 0xC77A);
	ChangeU2MB(0x006E57, 0xC77B);
	ChangeU2MB(0x006E5D, 0xC77D);
	ChangeU2MB(0x006E5E, 0xC77E);
	ChangeU2MB(0x006E62, 0xC821);
	ChangeU2MB(0x006E63, 0xC822);
	ChangeU2MB(0x006E68, 0xC823);
	ChangeU2MB(0x006E73, 0xC824);
	ChangeU2MB(0x006E7B, 0xC825);
	ChangeU2MB(0x006E7D, 0xC826);
	ChangeU2MB(0x006E8D, 0xC827);
	ChangeU2MB(0x006E93, 0xC828);
	ChangeU2MB(0x006E99, 0xC829);
	ChangeU2MB(0x006EA0, 0xC82A);
	ChangeU2MB(0x006EA7, 0xC82B);
	ChangeU2MB(0x006EAD, 0xC82C);
	ChangeU2MB(0x006EAE, 0xC82D);
	ChangeU2MB(0x006EB1, 0xC82E);
	ChangeU2MB(0x006EB3, 0xC82F);
	ChangeU2MB(0x006EBB, 0xC830);
	ChangeU2MB(0x006EC0, 0xC832);
	ChangeU2MB(0x006EC1, 0xC833);
	ChangeU2MB(0x006EC3, 0xC834);
	ChangeU2MB(0x006EC7, 0xC835);
	ChangeU2MB(0x006EC8, 0xC836);
	ChangeU2MB(0x006ECA, 0xC837);
	ChangeU2MB(0x006ECD, 0xC838);
	ChangeU2MB(0x006ECE, 0xC839);
	ChangeU2MB(0x006ECF, 0xC83A);
	ChangeU2MB(0x006EEB, 0xC83B);
	ChangeU2MB(0x006EED, 0xC83C);
	ChangeU2MB(0x006EEE, 0xC83D);
	ChangeU2MB(0x006EF9, 0xC83E);
	ChangeU2MB(0x006EFB, 0xC83F);
	ChangeU2MB(0x006EFD, 0xC840);
	ChangeU2MB(0x006F04, 0xC841);
	ChangeU2MB(0x006F08, 0xC842);
	ChangeU2MB(0x006F0A, 0xC843);
	ChangeU2MB(0x006F0C, 0xC844);
	ChangeU2MB(0x006F0D, 0xC845);
	ChangeU2MB(0x006F16, 0xC846);
	ChangeU2MB(0x006F18, 0xC847);
	ChangeU2MB(0x006F1A, 0xC848);
	ChangeU2MB(0x006F1B, 0xC849);
	ChangeU2MB(0x006F26, 0xC84A);
	ChangeU2MB(0x006F29, 0xC84B);
	ChangeU2MB(0x006F2A, 0xC84C);
	ChangeU2MB(0x006F2D, 0xC853);
	ChangeU2MB(0x006F2F, 0xC84D);
	ChangeU2MB(0x006F30, 0xC84E);
	ChangeU2MB(0x006F33, 0xC84F);
	ChangeU2MB(0x006F36, 0xC850);
	ChangeU2MB(0x006F3B, 0xC851);
	ChangeU2MB(0x006F3C, 0xC852);
	ChangeU2MB(0x006F4F, 0xC854);
	ChangeU2MB(0x006F51, 0xC855);
	ChangeU2MB(0x006F52, 0xC856);
	ChangeU2MB(0x006F53, 0xC857);
	ChangeU2MB(0x006F57, 0xC858);
	ChangeU2MB(0x006F59, 0xC859);
	ChangeU2MB(0x006F5A, 0xC85A);
	ChangeU2MB(0x006F5D, 0xC85B);
	ChangeU2MB(0x006F5E, 0xC85C);
	ChangeU2MB(0x006F61, 0xC85D);
	ChangeU2MB(0x006F62, 0xC85E);
	ChangeU2MB(0x006F68, 0xC85F);
	ChangeU2MB(0x006F6C, 0xC860);
	ChangeU2MB(0x006F7D, 0xC861);
	ChangeU2MB(0x006F7E, 0xC862);
	ChangeU2MB(0x006F83, 0xC863);
	ChangeU2MB(0x006F87, 0xC864);
	ChangeU2MB(0x006F8B, 0xC866);
	ChangeU2MB(0x006F8C, 0xC867);
	ChangeU2MB(0x006F8D, 0xC868);
	ChangeU2MB(0x006F90, 0xC869);
	ChangeU2MB(0x006F92, 0xC86A);
	ChangeU2MB(0x006F93, 0xC86B);
	ChangeU2MB(0x006F94, 0xC86C);
	ChangeU2MB(0x006F96, 0xC86D);
	ChangeU2MB(0x006F9A, 0xC86E);
	ChangeU2MB(0x006F9F, 0xC86F);
	ChangeU2MB(0x006FA0, 0xC870);
	ChangeU2MB(0x006FA5, 0xC871);
	ChangeU2MB(0x006FA6, 0xC872);
	ChangeU2MB(0x006FA7, 0xC873);
	ChangeU2MB(0x006FA8, 0xC874);
	ChangeU2MB(0x006FAE, 0xC875);
	ChangeU2MB(0x006FAF, 0xC876);
	ChangeU2MB(0x006FB0, 0xC877);
	ChangeU2MB(0x006FB6, 0xC879);
	ChangeU2MB(0x006FBC, 0xC87A);
	ChangeU2MB(0x006FC5, 0xC87B);
	ChangeU2MB(0x006FC7, 0xC87C);
	ChangeU2MB(0x006FC8, 0xC87D);
	ChangeU2MB(0x006FCA, 0xC87E);
	ChangeU2MB(0x006FDA, 0xC921);
	ChangeU2MB(0x006FDE, 0xC922);
	ChangeU2MB(0x006FE8, 0xC923);
	ChangeU2MB(0x006FE9, 0xC924);
	ChangeU2MB(0x006FF0, 0xC925);
	ChangeU2MB(0x006FF9, 0xC927);
	ChangeU2MB(0x006FFC, 0xC928);
	ChangeU2MB(0x006FFD, 0xC929);
	ChangeU2MB(0x007000, 0xC92A);
	ChangeU2MB(0x007006, 0xC92C);
	ChangeU2MB(0x00700D, 0xC92E);
	ChangeU2MB(0x007017, 0xC92F);
	ChangeU2MB(0x007020, 0xC930);
	ChangeU2MB(0x007023, 0xC931);
	ChangeU2MB(0x00702F, 0xC932);
	ChangeU2MB(0x007034, 0xC933);
	ChangeU2MB(0x007037, 0xC934);
	ChangeU2MB(0x007039, 0xC935);
	ChangeU2MB(0x00703C, 0xC936);
	ChangeU2MB(0x007043, 0xC937);
	ChangeU2MB(0x007044, 0xC938);
	ChangeU2MB(0x007048, 0xC939);
	ChangeU2MB(0x007049, 0xC93A);
	ChangeU2MB(0x00704A, 0xC93B);
	ChangeU2MB(0x00704B, 0xC93C);
	ChangeU2MB(0x00704E, 0xC941);
	ChangeU2MB(0x007054, 0xC93D);
	ChangeU2MB(0x007055, 0xC93E);
	ChangeU2MB(0x00705D, 0xC93F);
	ChangeU2MB(0x00705E, 0xC940);
	ChangeU2MB(0x007064, 0xC942);
	ChangeU2MB(0x007065, 0xC943);
	ChangeU2MB(0x00706C, 0xC944);
	ChangeU2MB(0x00706E, 0xC945);
	ChangeU2MB(0x007075, 0xC946);
	ChangeU2MB(0x007076, 0xC947);
	ChangeU2MB(0x00707E, 0xC948);
	ChangeU2MB(0x007081, 0xC949);
	ChangeU2MB(0x007086, 0xC94B);
	ChangeU2MB(0x007094, 0xC94C);
	ChangeU2MB(0x007095, 0xC94D);
	ChangeU2MB(0x007096, 0xC94E);
	ChangeU2MB(0x007097, 0xC94F);
	ChangeU2MB(0x007098, 0xC950);
	ChangeU2MB(0x00709B, 0xC951);
	ChangeU2MB(0x0070A4, 0xC952);
	ChangeU2MB(0x0070B0, 0xC954);
	ChangeU2MB(0x0070B1, 0xC955);
	ChangeU2MB(0x0070B4, 0xC956);
	ChangeU2MB(0x0070B7, 0xC957);
	ChangeU2MB(0x0070CA, 0xC958);
	ChangeU2MB(0x0070D1, 0xC959);
	ChangeU2MB(0x0070D3, 0xC95A);
	ChangeU2MB(0x0070D4, 0xC95B);
	ChangeU2MB(0x0070D5, 0xC95C);
	ChangeU2MB(0x0070D6, 0xC95D);
	ChangeU2MB(0x0070D8, 0xC95E);
	ChangeU2MB(0x0070DC, 0xC95F);
	ChangeU2MB(0x0070E4, 0xC960);
	ChangeU2MB(0x0070FA, 0xC961);
	ChangeU2MB(0x007103, 0xC962);
	ChangeU2MB(0x007105, 0xC964);
	ChangeU2MB(0x007106, 0xC965);
	ChangeU2MB(0x007107, 0xC966);
	ChangeU2MB(0x00710B, 0xC967);
	ChangeU2MB(0x00710C, 0xC968);
	ChangeU2MB(0x00711E, 0xC96A);
	ChangeU2MB(0x007120, 0xC96B);
	ChangeU2MB(0x00712B, 0xC96C);
	ChangeU2MB(0x00712D, 0xC96D);
	ChangeU2MB(0x00712F, 0xC96E);
	ChangeU2MB(0x007130, 0xC96F);
	ChangeU2MB(0x007131, 0xC970);
	ChangeU2MB(0x007138, 0xC971);
	ChangeU2MB(0x007141, 0xC972);
	ChangeU2MB(0x007145, 0xC973);
	ChangeU2MB(0x00714A, 0xC976);
	ChangeU2MB(0x00714B, 0xC977);
	ChangeU2MB(0x007150, 0xC978);
	ChangeU2MB(0x007152, 0xC979);
	ChangeU2MB(0x007157, 0xC97A);
	ChangeU2MB(0x00715A, 0xC97B);
	ChangeU2MB(0x00715E, 0xC97D);
	ChangeU2MB(0x007160, 0xC97E);
	ChangeU2MB(0x007168, 0xCA21);
	ChangeU2MB(0x007179, 0xCA22);
	ChangeU2MB(0x007180, 0xCA23);
	ChangeU2MB(0x007185, 0xCA24);
	ChangeU2MB(0x007187, 0xCA25);
	ChangeU2MB(0x00718C, 0xCA26);
	ChangeU2MB(0x007192, 0xCA27);
	ChangeU2MB(0x00719A, 0xCA28);
	ChangeU2MB(0x00719B, 0xCA29);
	ChangeU2MB(0x0071A0, 0xCA2A);
	ChangeU2MB(0x0071A2, 0xCA2B);
	ChangeU2MB(0x0071AF, 0xCA2C);
	ChangeU2MB(0x0071B0, 0xCA2D);
	ChangeU2MB(0x0071B2, 0xCA2E);
	ChangeU2MB(0x0071B3, 0xCA2F);
	ChangeU2MB(0x0071BA, 0xCA30);
	ChangeU2MB(0x0071BF, 0xCA31);
	ChangeU2MB(0x0071C0, 0xCA32);
	ChangeU2MB(0x0071C4, 0xCA34);
	ChangeU2MB(0x0071CB, 0xCA35);
	ChangeU2MB(0x0071CC, 0xCA36);
	ChangeU2MB(0x0071D3, 0xCA37);
	ChangeU2MB(0x0071D6, 0xCA38);
	ChangeU2MB(0x0071D9, 0xCA39);
	ChangeU2MB(0x0071DA, 0xCA3A);
	ChangeU2MB(0x0071DC, 0xCA3B);
	ChangeU2MB(0x0071F8, 0xCA3C);
	ChangeU2MB(0x007200, 0xCA3E);
	ChangeU2MB(0x007207, 0xCA3F);
	ChangeU2MB(0x007208, 0xCA40);
	ChangeU2MB(0x007209, 0xCA41);
	ChangeU2MB(0x007213, 0xCA42);
	ChangeU2MB(0x007217, 0xCA43);
	ChangeU2MB(0x00721A, 0xCA44);
	ChangeU2MB(0x00721D, 0xCA45);
	ChangeU2MB(0x00721F, 0xCA46);
	ChangeU2MB(0x007224, 0xCA47);
	ChangeU2MB(0x00722B, 0xCA48);
	ChangeU2MB(0x00722F, 0xCA49);
	ChangeU2MB(0x007234, 0xCA4A);
	ChangeU2MB(0x007238, 0xCA4B);
	ChangeU2MB(0x007239, 0xCA4C);
	ChangeU2MB(0x007241, 0xCA4D);
	ChangeU2MB(0x007242, 0xCA4E);
	ChangeU2MB(0x007243, 0xCA4F);
	ChangeU2MB(0x007245, 0xCA50);
	ChangeU2MB(0x00724E, 0xCA51);
	ChangeU2MB(0x00724F, 0xCA52);
	ChangeU2MB(0x007250, 0xCA53);
	ChangeU2MB(0x007253, 0xCA54);
	ChangeU2MB(0x007255, 0xCA55);
	ChangeU2MB(0x007256, 0xCA56);
	ChangeU2MB(0x00725A, 0xCA57);
	ChangeU2MB(0x00725C, 0xCA58);
	ChangeU2MB(0x00725E, 0xCA59);
	ChangeU2MB(0x007260, 0xCA5A);
	ChangeU2MB(0x007263, 0xCA5B);
	ChangeU2MB(0x007268, 0xCA5C);
	ChangeU2MB(0x00726B, 0xCA5D);
	ChangeU2MB(0x00726E, 0xCA5E);
	ChangeU2MB(0x00726F, 0xCA5F);
	ChangeU2MB(0x007271, 0xCA60);
	ChangeU2MB(0x007277, 0xCA61);
	ChangeU2MB(0x007278, 0xCA62);
	ChangeU2MB(0x00727B, 0xCA63);
	ChangeU2MB(0x00727C, 0xCA64);
	ChangeU2MB(0x00727F, 0xCA65);
	ChangeU2MB(0x007284, 0xCA66);
	ChangeU2MB(0x007289, 0xCA67);
	ChangeU2MB(0x00728D, 0xCA68);
	ChangeU2MB(0x00728E, 0xCA69);
	ChangeU2MB(0x007293, 0xCA6A);
	ChangeU2MB(0x00729B, 0xCA6B);
	ChangeU2MB(0x0072A8, 0xCA6C);
	ChangeU2MB(0x0072AD, 0xCA6D);
	ChangeU2MB(0x0072AE, 0xCA6E);
	ChangeU2MB(0x0072B4, 0xCA70);
	ChangeU2MB(0x0072C1, 0xCA72);
	ChangeU2MB(0x0072C7, 0xCA73);
	ChangeU2MB(0x0072C9, 0xCA74);
	ChangeU2MB(0x0072CC, 0xCA75);
	ChangeU2MB(0x0072D5, 0xCA76);
	ChangeU2MB(0x0072D6, 0xCA77);
	ChangeU2MB(0x0072D8, 0xCA78);
	ChangeU2MB(0x0072DF, 0xCA79);
	ChangeU2MB(0x0072E5, 0xCA7A);
	ChangeU2MB(0x0072F3, 0xCA7B);
	ChangeU2MB(0x0072F4, 0xCA7C);
	ChangeU2MB(0x0072FA, 0xCA7D);
	ChangeU2MB(0x0072FB, 0xCA7E);
	ChangeU2MB(0x0072FE, 0xCB21);
	ChangeU2MB(0x007302, 0xCB22);
	ChangeU2MB(0x007304, 0xCB23);
	ChangeU2MB(0x007305, 0xCB24);
	ChangeU2MB(0x007307, 0xCB25);
	ChangeU2MB(0x00730B, 0xCB26);
	ChangeU2MB(0x00730D, 0xCB27);
	ChangeU2MB(0x007312, 0xCB28);
	ChangeU2MB(0x007313, 0xCB29);
	ChangeU2MB(0x007318, 0xCB2A);
	ChangeU2MB(0x007319, 0xCB2B);
	ChangeU2MB(0x00731E, 0xCB2C);
	ChangeU2MB(0x007322, 0xCB2D);
	ChangeU2MB(0x007327, 0xCB2F);
	ChangeU2MB(0x007328, 0xCB30);
	ChangeU2MB(0x00732C, 0xCB31);
	ChangeU2MB(0x007331, 0xCB32);
	ChangeU2MB(0x007332, 0xCB33);
	ChangeU2MB(0x007335, 0xCB34);
	ChangeU2MB(0x00733A, 0xCB35);
	ChangeU2MB(0x00733B, 0xCB36);
	ChangeU2MB(0x00733D, 0xCB37);
	ChangeU2MB(0x007343, 0xCB38);
	ChangeU2MB(0x00734D, 0xCB39);
	ChangeU2MB(0x007350, 0xCB3A);
	ChangeU2MB(0x007352, 0xCB3B);
	ChangeU2MB(0x007356, 0xCB3C);
	ChangeU2MB(0x007358, 0xCB3D);
	ChangeU2MB(0x00735D, 0xCB3E);
	ChangeU2MB(0x00735E, 0xCB3F);
	ChangeU2MB(0x00735F, 0xCB40);
	ChangeU2MB(0x007360, 0xCB41);
	ChangeU2MB(0x007366, 0xCB42);
	ChangeU2MB(0x007367, 0xCB43);
	ChangeU2MB(0x007369, 0xCB44);
	ChangeU2MB(0x00736B, 0xCB45);
	ChangeU2MB(0x00736C, 0xCB46);
	ChangeU2MB(0x00736E, 0xCB47);
	ChangeU2MB(0x00736F, 0xCB48);
	ChangeU2MB(0x007371, 0xCB49);
	ChangeU2MB(0x007379, 0xCB4B);
	ChangeU2MB(0x00737C, 0xCB4C);
	ChangeU2MB(0x007380, 0xCB4D);
	ChangeU2MB(0x007381, 0xCB4E);
	ChangeU2MB(0x007383, 0xCB4F);
	ChangeU2MB(0x007385, 0xCB50);
	ChangeU2MB(0x007386, 0xCB51);
	ChangeU2MB(0x00738E, 0xCB52);
	ChangeU2MB(0x007390, 0xCB53);
	ChangeU2MB(0x007393, 0xCB54);
	ChangeU2MB(0x007395, 0xCB55);
	ChangeU2MB(0x007397, 0xCB56);
	ChangeU2MB(0x007398, 0xCB57);
	ChangeU2MB(0x00739C, 0xCB58);
	ChangeU2MB(0x00739E, 0xCB59);
	ChangeU2MB(0x00739F, 0xCB5A);
	ChangeU2MB(0x0073A0, 0xCB5B);
	ChangeU2MB(0x0073A2, 0xCB5C);
	ChangeU2MB(0x0073A5, 0xCB5D);
	ChangeU2MB(0x0073A6, 0xCB5E);
	ChangeU2MB(0x0073AA, 0xCB5F);
	ChangeU2MB(0x0073AB, 0xCB60);
	ChangeU2MB(0x0073AD, 0xCB61);
	ChangeU2MB(0x0073B5, 0xCB62);
	ChangeU2MB(0x0073B7, 0xCB63);
	ChangeU2MB(0x0073B9, 0xCB64);
	ChangeU2MB(0x0073BC, 0xCB65);
	ChangeU2MB(0x0073BF, 0xCB67);
	ChangeU2MB(0x0073C5, 0xCB68);
	ChangeU2MB(0x0073C6, 0xCB69);
	ChangeU2MB(0x0073CB, 0xCB6B);
	ChangeU2MB(0x0073CC, 0xCB6C);
	ChangeU2MB(0x0073CF, 0xCB6D);
	ChangeU2MB(0x0073D3, 0xCB6F);
	ChangeU2MB(0x0073D9, 0xCB71);
	ChangeU2MB(0x0073DD, 0xCB72);
	ChangeU2MB(0x0073E1, 0xCB73);
	ChangeU2MB(0x0073E6, 0xCB75);
	ChangeU2MB(0x0073E7, 0xCB76);
	ChangeU2MB(0x0073E9, 0xCB77);
	ChangeU2MB(0x0073F4, 0xCB78);
	ChangeU2MB(0x0073F7, 0xCB7A);
	ChangeU2MB(0x0073F9, 0xCB7B);
	ChangeU2MB(0x0073FA, 0xCB7C);
	ChangeU2MB(0x0073FB, 0xCB7D);
	ChangeU2MB(0x0073FD, 0xCB7E);
	ChangeU2MB(0x0073FF, 0xCC21);
	ChangeU2MB(0x007400, 0xCC22);
	ChangeU2MB(0x007401, 0xCC23);
	ChangeU2MB(0x007404, 0xCC24);
	ChangeU2MB(0x00740A, 0xCC26);
	ChangeU2MB(0x007411, 0xCC27);
	ChangeU2MB(0x00741A, 0xCC28);
	ChangeU2MB(0x00741B, 0xCC29);
	ChangeU2MB(0x007424, 0xCC2A);
	ChangeU2MB(0x007428, 0xCC2C);
	ChangeU2MB(0x00742B, 0xCC2F);
	ChangeU2MB(0x00742C, 0xCC30);
	ChangeU2MB(0x00742D, 0xCC31);
	ChangeU2MB(0x00742F, 0xCC33);
	ChangeU2MB(0x007430, 0xCC34);
	ChangeU2MB(0x007431, 0xCC35);
	ChangeU2MB(0x007439, 0xCC36);
	ChangeU2MB(0x007440, 0xCC37);
	ChangeU2MB(0x007443, 0xCC38);
	ChangeU2MB(0x007444, 0xCC39);
	ChangeU2MB(0x007446, 0xCC3A);
	ChangeU2MB(0x007447, 0xCC3B);
	ChangeU2MB(0x00744B, 0xCC3C);
	ChangeU2MB(0x00744D, 0xCC3D);
	ChangeU2MB(0x007451, 0xCC3E);
	ChangeU2MB(0x007452, 0xCC3F);
	ChangeU2MB(0x007457, 0xCC40);
	ChangeU2MB(0x00745D, 0xCC41);
	ChangeU2MB(0x007466, 0xCC43);
	ChangeU2MB(0x007467, 0xCC44);
	ChangeU2MB(0x007468, 0xCC45);
	ChangeU2MB(0x00746B, 0xCC46);
	ChangeU2MB(0x00746D, 0xCC47);
	ChangeU2MB(0x00746E, 0xCC48);
	ChangeU2MB(0x007471, 0xCC49);
	ChangeU2MB(0x007472, 0xCC4A);
	ChangeU2MB(0x007480, 0xCC4B);
	ChangeU2MB(0x007481, 0xCC4C);
	ChangeU2MB(0x007485, 0xCC4D);
	ChangeU2MB(0x007486, 0xCC4E);
	ChangeU2MB(0x007487, 0xCC4F);
	ChangeU2MB(0x00748F, 0xCC51);
	ChangeU2MB(0x007490, 0xCC52);
	ChangeU2MB(0x007491, 0xCC53);
	ChangeU2MB(0x007492, 0xCC54);
	ChangeU2MB(0x007498, 0xCC55);
	ChangeU2MB(0x007499, 0xCC56);
	ChangeU2MB(0x00749A, 0xCC57);
	ChangeU2MB(0x00749C, 0xCC58);
	ChangeU2MB(0x0074A0, 0xCC5A);
	ChangeU2MB(0x0074A1, 0xCC5B);
	ChangeU2MB(0x0074A3, 0xCC5C);
	ChangeU2MB(0x0074A6, 0xCC5D);
	ChangeU2MB(0x0074A8, 0xCC5E);
	ChangeU2MB(0x0074A9, 0xCC5F);
	ChangeU2MB(0x0074AA, 0xCC60);
	ChangeU2MB(0x0074AB, 0xCC61);
	ChangeU2MB(0x0074AE, 0xCC62);
	ChangeU2MB(0x0074AF, 0xCC63);
	ChangeU2MB(0x0074B1, 0xCC64);
	ChangeU2MB(0x0074B2, 0xCC65);
	ChangeU2MB(0x0074B5, 0xCC66);
	ChangeU2MB(0x0074B9, 0xCC67);
	ChangeU2MB(0x0074BB, 0xCC68);
	ChangeU2MB(0x0074BF, 0xCC69);
	ChangeU2MB(0x0074C8, 0xCC6A);
	ChangeU2MB(0x0074C9, 0xCC6B);
	ChangeU2MB(0x0074CC, 0xCC6C);
	ChangeU2MB(0x0074D0, 0xCC6D);
	ChangeU2MB(0x0074D3, 0xCC6E);
	ChangeU2MB(0x0074D8, 0xCC6F);
	ChangeU2MB(0x0074DA, 0xCC70);
	ChangeU2MB(0x0074DB, 0xCC71);
	ChangeU2MB(0x0074DE, 0xCC72);
	ChangeU2MB(0x0074DF, 0xCC73);
	ChangeU2MB(0x0074E4, 0xCC74);
	ChangeU2MB(0x0074E8, 0xCC75);
	ChangeU2MB(0x0074EA, 0xCC76);
	ChangeU2MB(0x0074EB, 0xCC77);
	ChangeU2MB(0x0074EF, 0xCC78);
	ChangeU2MB(0x0074F4, 0xCC79);
	ChangeU2MB(0x0074FA, 0xCC7A);
	ChangeU2MB(0x0074FB, 0xCC7B);
	ChangeU2MB(0x0074FC, 0xCC7C);
	ChangeU2MB(0x0074FF, 0xCC7D);
	ChangeU2MB(0x007506, 0xCC7E);
	ChangeU2MB(0x007512, 0xCD21);
	ChangeU2MB(0x007516, 0xCD22);
	ChangeU2MB(0x007517, 0xCD23);
	ChangeU2MB(0x007520, 0xCD24);
	ChangeU2MB(0x007521, 0xCD25);
	ChangeU2MB(0x007524, 0xCD26);
	ChangeU2MB(0x007527, 0xCD27);
	ChangeU2MB(0x007529, 0xCD28);
	ChangeU2MB(0x00752A, 0xCD29);
	ChangeU2MB(0x007536, 0xCD2B);
	ChangeU2MB(0x007539, 0xCD2C);
	ChangeU2MB(0x00753D, 0xCD2D);
	ChangeU2MB(0x00753E, 0xCD2E);
	ChangeU2MB(0x00753F, 0xCD2F);
	ChangeU2MB(0x007540, 0xCD30);
	ChangeU2MB(0x007543, 0xCD31);
	ChangeU2MB(0x007547, 0xCD32);
	ChangeU2MB(0x007548, 0xCD33);
	ChangeU2MB(0x00754E, 0xCD34);
	ChangeU2MB(0x007550, 0xCD35);
	ChangeU2MB(0x007552, 0xCD36);
	ChangeU2MB(0x007557, 0xCD37);
	ChangeU2MB(0x00755E, 0xCD38);
	ChangeU2MB(0x00755F, 0xCD39);
	ChangeU2MB(0x007561, 0xCD3A);
	ChangeU2MB(0x007571, 0xCD3C);
	ChangeU2MB(0x007579, 0xCD3D);
	ChangeU2MB(0x00757A, 0xCD3E);
	ChangeU2MB(0x00757B, 0xCD3F);
	ChangeU2MB(0x00757C, 0xCD40);
	ChangeU2MB(0x00757D, 0xCD41);
	ChangeU2MB(0x00757E, 0xCD42);
	ChangeU2MB(0x007581, 0xCD43);
	ChangeU2MB(0x007585, 0xCD44);
	ChangeU2MB(0x007590, 0xCD45);
	ChangeU2MB(0x007592, 0xCD46);
	ChangeU2MB(0x007593, 0xCD47);
	ChangeU2MB(0x007595, 0xCD48);
	ChangeU2MB(0x007599, 0xCD49);
	ChangeU2MB(0x00759C, 0xCD4A);
	ChangeU2MB(0x0075A2, 0xCD4B);
	ChangeU2MB(0x0075A4, 0xCD4C);
	ChangeU2MB(0x0075B4, 0xCD4D);
	ChangeU2MB(0x0075BA, 0xCD4E);
	ChangeU2MB(0x0075BF, 0xCD4F);
	ChangeU2MB(0x0075C0, 0xCD50);
	ChangeU2MB(0x0075C1, 0xCD51);
	ChangeU2MB(0x0075C4, 0xCD52);
	ChangeU2MB(0x0075C6, 0xCD53);
	ChangeU2MB(0x0075CC, 0xCD54);
	ChangeU2MB(0x0075CE, 0xCD55);
	ChangeU2MB(0x0075CF, 0xCD56);
	ChangeU2MB(0x0075D7, 0xCD57);
	ChangeU2MB(0x0075DC, 0xCD58);
	ChangeU2MB(0x0075DF, 0xCD59);
	ChangeU2MB(0x0075E0, 0xCD5A);
	ChangeU2MB(0x0075E1, 0xCD5B);
	ChangeU2MB(0x0075E4, 0xCD5C);
	ChangeU2MB(0x0075E7, 0xCD5D);
	ChangeU2MB(0x0075EC, 0xCD5E);
	ChangeU2MB(0x0075EE, 0xCD5F);
	ChangeU2MB(0x0075EF, 0xCD60);
	ChangeU2MB(0x0075F1, 0xCD61);
	ChangeU2MB(0x0075F9, 0xCD62);
	ChangeU2MB(0x007600, 0xCD63);
	ChangeU2MB(0x007602, 0xCD64);
	ChangeU2MB(0x007603, 0xCD65);
	ChangeU2MB(0x007604, 0xCD66);
	ChangeU2MB(0x007607, 0xCD67);
	ChangeU2MB(0x007608, 0xCD68);
	ChangeU2MB(0x00760A, 0xCD69);
	ChangeU2MB(0x00760C, 0xCD6A);
	ChangeU2MB(0x00760F, 0xCD6B);
	ChangeU2MB(0x007612, 0xCD6C);
	ChangeU2MB(0x007613, 0xCD6D);
	ChangeU2MB(0x007615, 0xCD6E);
	ChangeU2MB(0x007616, 0xCD6F);
	ChangeU2MB(0x007619, 0xCD70);
	ChangeU2MB(0x00761B, 0xCD71);
	ChangeU2MB(0x00761C, 0xCD72);
	ChangeU2MB(0x00761D, 0xCD73);
	ChangeU2MB(0x00761E, 0xCD74);
	ChangeU2MB(0x007623, 0xCD75);
	ChangeU2MB(0x007625, 0xCD76);
	ChangeU2MB(0x007626, 0xCD77);
	ChangeU2MB(0x007629, 0xCD78);
	ChangeU2MB(0x00762D, 0xCD79);
	ChangeU2MB(0x007632, 0xCD7A);
	ChangeU2MB(0x007633, 0xCD7B);
	ChangeU2MB(0x007635, 0xCD7C);
	ChangeU2MB(0x007638, 0xCD7D);
	ChangeU2MB(0x007639, 0xCD7E);
	ChangeU2MB(0x00763A, 0xCE21);
	ChangeU2MB(0x00763C, 0xCE22);
	ChangeU2MB(0x007640, 0xCE24);
	ChangeU2MB(0x007641, 0xCE25);
	ChangeU2MB(0x007643, 0xCE26);
	ChangeU2MB(0x007644, 0xCE27);
	ChangeU2MB(0x007645, 0xCE28);
	ChangeU2MB(0x007649, 0xCE29);
	ChangeU2MB(0x00764A, 0xCE23);
	ChangeU2MB(0x00764B, 0xCE2A);
	ChangeU2MB(0x007655, 0xCE2B);
	ChangeU2MB(0x007659, 0xCE2C);
	ChangeU2MB(0x00765F, 0xCE2D);
	ChangeU2MB(0x007664, 0xCE2E);
	ChangeU2MB(0x007665, 0xCE2F);
	ChangeU2MB(0x00766D, 0xCE30);
	ChangeU2MB(0x00766E, 0xCE31);
	ChangeU2MB(0x00766F, 0xCE32);
	ChangeU2MB(0x007671, 0xCE33);
	ChangeU2MB(0x007674, 0xCE34);
	ChangeU2MB(0x007681, 0xCE35);
	ChangeU2MB(0x007685, 0xCE36);
	ChangeU2MB(0x00768C, 0xCE37);
	ChangeU2MB(0x00768D, 0xCE38);
	ChangeU2MB(0x007695, 0xCE39);
	ChangeU2MB(0x00769D, 0xCE3C);
	ChangeU2MB(0x00769F, 0xCE3D);
	ChangeU2MB(0x0076A0, 0xCE3E);
	ChangeU2MB(0x0076A2, 0xCE3F);
	ChangeU2MB(0x0076A3, 0xCE40);
	ChangeU2MB(0x0076A4, 0xCE41);
	ChangeU2MB(0x0076A5, 0xCE42);
	ChangeU2MB(0x0076A7, 0xCE44);
	ChangeU2MB(0x0076A8, 0xCE45);
	ChangeU2MB(0x0076AA, 0xCE46);
	ChangeU2MB(0x0076AD, 0xCE47);
	ChangeU2MB(0x0076BD, 0xCE48);
	ChangeU2MB(0x0076C1, 0xCE49);
	ChangeU2MB(0x0076C5, 0xCE4A);
	ChangeU2MB(0x0076C9, 0xCE4B);
	ChangeU2MB(0x0076CB, 0xCE4C);
	ChangeU2MB(0x0076CC, 0xCE4D);
	ChangeU2MB(0x0076CE, 0xCE4E);
	ChangeU2MB(0x0076D4, 0xCE4F);
	ChangeU2MB(0x0076D9, 0xCE50);
	ChangeU2MB(0x0076E0, 0xCE51);
	ChangeU2MB(0x0076E6, 0xCE52);
	ChangeU2MB(0x0076E8, 0xCE53);
	ChangeU2MB(0x0076EC, 0xCE54);
	ChangeU2MB(0x0076F0, 0xCE55);
	ChangeU2MB(0x0076F1, 0xCE56);
	ChangeU2MB(0x0076F6, 0xCE57);
	ChangeU2MB(0x0076F9, 0xCE58);
	ChangeU2MB(0x0076FC, 0xCE59);
	ChangeU2MB(0x007700, 0xCE5A);
	ChangeU2MB(0x007706, 0xCE5B);
	ChangeU2MB(0x00770A, 0xCE5C);
	ChangeU2MB(0x00770E, 0xCE5D);
	ChangeU2MB(0x007712, 0xCE5E);
	ChangeU2MB(0x007714, 0xCE5F);
	ChangeU2MB(0x007715, 0xCE60);
	ChangeU2MB(0x007717, 0xCE61);
	ChangeU2MB(0x007719, 0xCE62);
	ChangeU2MB(0x00771A, 0xCE63);
	ChangeU2MB(0x00771C, 0xCE64);
	ChangeU2MB(0x007722, 0xCE65);
	ChangeU2MB(0x007728, 0xCE66);
	ChangeU2MB(0x00772D, 0xCE67);
	ChangeU2MB(0x00772E, 0xCE68);
	ChangeU2MB(0x00772F, 0xCE69);
	ChangeU2MB(0x007734, 0xCE6A);
	ChangeU2MB(0x007735, 0xCE6B);
	ChangeU2MB(0x007736, 0xCE6C);
	ChangeU2MB(0x007739, 0xCE6D);
	ChangeU2MB(0x00773D, 0xCE6E);
	ChangeU2MB(0x00773E, 0xCE6F);
	ChangeU2MB(0x007742, 0xCE70);
	ChangeU2MB(0x007745, 0xCE71);
	ChangeU2MB(0x00774A, 0xCE73);
	ChangeU2MB(0x00774D, 0xCE74);
	ChangeU2MB(0x00774E, 0xCE75);
	ChangeU2MB(0x00774F, 0xCE76);
	ChangeU2MB(0x007752, 0xCE77);
	ChangeU2MB(0x007756, 0xCE78);
	ChangeU2MB(0x007757, 0xCE79);
	ChangeU2MB(0x00775C, 0xCE7A);
	ChangeU2MB(0x00775E, 0xCE7B);
	ChangeU2MB(0x00775F, 0xCE7C);
	ChangeU2MB(0x007760, 0xCE7D);
	ChangeU2MB(0x007762, 0xCE7E);
	ChangeU2MB(0x007764, 0xCF21);
	ChangeU2MB(0x007767, 0xCF22);
	ChangeU2MB(0x00776A, 0xCF23);
	ChangeU2MB(0x00776C, 0xCF24);
	ChangeU2MB(0x007770, 0xCF25);
	ChangeU2MB(0x007772, 0xCF26);
	ChangeU2MB(0x007773, 0xCF27);
	ChangeU2MB(0x007774, 0xCF28);
	ChangeU2MB(0x00777A, 0xCF29);
	ChangeU2MB(0x00777D, 0xCF2A);
	ChangeU2MB(0x007780, 0xCF2B);
	ChangeU2MB(0x007784, 0xCF2C);
	ChangeU2MB(0x00778C, 0xCF2D);
	ChangeU2MB(0x00778D, 0xCF2E);
	ChangeU2MB(0x007794, 0xCF2F);
	ChangeU2MB(0x007795, 0xCF30);
	ChangeU2MB(0x007796, 0xCF31);
	ChangeU2MB(0x00779A, 0xCF32);
	ChangeU2MB(0x00779F, 0xCF33);
	ChangeU2MB(0x0077A2, 0xCF34);
	ChangeU2MB(0x0077A7, 0xCF35);
	ChangeU2MB(0x0077AA, 0xCF36);
	ChangeU2MB(0x0077AE, 0xCF37);
	ChangeU2MB(0x0077AF, 0xCF38);
	ChangeU2MB(0x0077B1, 0xCF39);
	ChangeU2MB(0x0077B5, 0xCF3A);
	ChangeU2MB(0x0077BE, 0xCF3B);
	ChangeU2MB(0x0077C3, 0xCF3C);
	ChangeU2MB(0x0077C9, 0xCF3D);
	ChangeU2MB(0x0077D1, 0xCF3E);
	ChangeU2MB(0x0077D2, 0xCF3F);
	ChangeU2MB(0x0077D5, 0xCF40);
	ChangeU2MB(0x0077D9, 0xCF41);
	ChangeU2MB(0x0077DE, 0xCF42);
	ChangeU2MB(0x0077DF, 0xCF43);
	ChangeU2MB(0x0077E0, 0xCF44);
	ChangeU2MB(0x0077E4, 0xCF45);
	ChangeU2MB(0x0077E6, 0xCF46);
	ChangeU2MB(0x0077EA, 0xCF47);
	ChangeU2MB(0x0077EC, 0xCF48);
	ChangeU2MB(0x0077F0, 0xCF49);
	ChangeU2MB(0x0077F1, 0xCF4A);
	ChangeU2MB(0x0077F4, 0xCF4B);
	ChangeU2MB(0x0077F8, 0xCF4C);
	ChangeU2MB(0x0077FB, 0xCF4D);
	ChangeU2MB(0x007805, 0xCF4E);
	ChangeU2MB(0x007806, 0xCF4F);
	ChangeU2MB(0x007809, 0xCF50);
	ChangeU2MB(0x00780D, 0xCF51);
	ChangeU2MB(0x00780E, 0xCF52);
	ChangeU2MB(0x007811, 0xCF53);
	ChangeU2MB(0x00781D, 0xCF54);
	ChangeU2MB(0x007822, 0xCF56);
	ChangeU2MB(0x007823, 0xCF57);
	ChangeU2MB(0x00782D, 0xCF58);
	ChangeU2MB(0x00782E, 0xCF59);
	ChangeU2MB(0x007830, 0xCF5A);
	ChangeU2MB(0x007835, 0xCF5B);
	ChangeU2MB(0x007837, 0xCF5C);
	ChangeU2MB(0x007843, 0xCF5D);
	ChangeU2MB(0x007844, 0xCF5E);
	ChangeU2MB(0x007847, 0xCF5F);
	ChangeU2MB(0x007848, 0xCF60);
	ChangeU2MB(0x00784C, 0xCF61);
	ChangeU2MB(0x007852, 0xCF63);
	ChangeU2MB(0x00785C, 0xCF64);
	ChangeU2MB(0x00785E, 0xCF65);
	ChangeU2MB(0x007860, 0xCF66);
	ChangeU2MB(0x007861, 0xCF67);
	ChangeU2MB(0x007863, 0xCF68);
	ChangeU2MB(0x007868, 0xCF6A);
	ChangeU2MB(0x00786A, 0xCF6B);
	ChangeU2MB(0x00786E, 0xCF6C);
	ChangeU2MB(0x00787E, 0xCF6E);
	ChangeU2MB(0x00788A, 0xCF6F);
	ChangeU2MB(0x00788F, 0xCF70);
	ChangeU2MB(0x007894, 0xCF71);
	ChangeU2MB(0x007898, 0xCF72);
	ChangeU2MB(0x00789D, 0xCF74);
	ChangeU2MB(0x00789E, 0xCF75);
	ChangeU2MB(0x00789F, 0xCF76);
	ChangeU2MB(0x0078A1, 0xCF73);
	ChangeU2MB(0x0078A4, 0xCF77);
	ChangeU2MB(0x0078A8, 0xCF78);
	ChangeU2MB(0x0078AC, 0xCF79);
	ChangeU2MB(0x0078AD, 0xCF7A);
	ChangeU2MB(0x0078B0, 0xCF7B);
	ChangeU2MB(0x0078B1, 0xCF7C);
	ChangeU2MB(0x0078B2, 0xCF7D);
	ChangeU2MB(0x0078B3, 0xCF7E);
	ChangeU2MB(0x0078BB, 0xD021);
	ChangeU2MB(0x0078BD, 0xD022);
	ChangeU2MB(0x0078BF, 0xD023);
	ChangeU2MB(0x0078C7, 0xD024);
	ChangeU2MB(0x0078C8, 0xD025);
	ChangeU2MB(0x0078C9, 0xD026);
	ChangeU2MB(0x0078CC, 0xD027);
	ChangeU2MB(0x0078CE, 0xD028);
	ChangeU2MB(0x0078D2, 0xD029);
	ChangeU2MB(0x0078D3, 0xD02A);
	ChangeU2MB(0x0078D5, 0xD02B);
	ChangeU2MB(0x0078D6, 0xD02C);
	ChangeU2MB(0x0078DB, 0xD02E);
	ChangeU2MB(0x0078DF, 0xD02F);
	ChangeU2MB(0x0078E0, 0xD030);
	ChangeU2MB(0x0078E1, 0xD031);
	ChangeU2MB(0x0078E4, 0xD02D);
	ChangeU2MB(0x0078E6, 0xD032);
	ChangeU2MB(0x0078EA, 0xD033);
	ChangeU2MB(0x0078F2, 0xD034);
	ChangeU2MB(0x0078F3, 0xD035);
	ChangeU2MB(0x0078F6, 0xD037);
	ChangeU2MB(0x0078F7, 0xD038);
	ChangeU2MB(0x0078FA, 0xD039);
	ChangeU2MB(0x0078FB, 0xD03A);
	ChangeU2MB(0x0078FF, 0xD03B);
	ChangeU2MB(0x007900, 0xD036);
	ChangeU2MB(0x007906, 0xD03C);
	ChangeU2MB(0x00790C, 0xD03D);
	ChangeU2MB(0x007910, 0xD03E);
	ChangeU2MB(0x00791A, 0xD03F);
	ChangeU2MB(0x00791C, 0xD040);
	ChangeU2MB(0x00791E, 0xD041);
	ChangeU2MB(0x00791F, 0xD042);
	ChangeU2MB(0x007920, 0xD043);
	ChangeU2MB(0x007925, 0xD044);
	ChangeU2MB(0x007927, 0xD045);
	ChangeU2MB(0x007929, 0xD046);
	ChangeU2MB(0x00792D, 0xD047);
	ChangeU2MB(0x007931, 0xD048);
	ChangeU2MB(0x007934, 0xD049);
	ChangeU2MB(0x007935, 0xD04A);
	ChangeU2MB(0x00793B, 0xD04B);
	ChangeU2MB(0x00793D, 0xD04C);
	ChangeU2MB(0x00793F, 0xD04D);
	ChangeU2MB(0x007944, 0xD04E);
	ChangeU2MB(0x007945, 0xD04F);
	ChangeU2MB(0x007946, 0xD050);
	ChangeU2MB(0x00794A, 0xD051);
	ChangeU2MB(0x00794B, 0xD052);
	ChangeU2MB(0x00794F, 0xD053);
	ChangeU2MB(0x007951, 0xD054);
	ChangeU2MB(0x007954, 0xD055);
	ChangeU2MB(0x007958, 0xD056);
	ChangeU2MB(0x00795B, 0xD057);
	ChangeU2MB(0x00795C, 0xD058);
	ChangeU2MB(0x007967, 0xD059);
	ChangeU2MB(0x007969, 0xD05A);
	ChangeU2MB(0x00796B, 0xD05B);
	ChangeU2MB(0x007972, 0xD05C);
	ChangeU2MB(0x007979, 0xD05D);
	ChangeU2MB(0x00797B, 0xD05E);
	ChangeU2MB(0x00797C, 0xD05F);
	ChangeU2MB(0x00797E, 0xD060);
	ChangeU2MB(0x00798B, 0xD061);
	ChangeU2MB(0x00798C, 0xD062);
	ChangeU2MB(0x007991, 0xD063);
	ChangeU2MB(0x007993, 0xD064);
	ChangeU2MB(0x007995, 0xD066);
	ChangeU2MB(0x007996, 0xD067);
	ChangeU2MB(0x007998, 0xD068);
	ChangeU2MB(0x00799C, 0xD06A);
	ChangeU2MB(0x0079A1, 0xD06B);
	ChangeU2MB(0x0079A8, 0xD06C);
	ChangeU2MB(0x0079A9, 0xD06D);
	ChangeU2MB(0x0079AB, 0xD06E);
	ChangeU2MB(0x0079AF, 0xD06F);
	ChangeU2MB(0x0079B1, 0xD070);
	ChangeU2MB(0x0079B4, 0xD071);
	ChangeU2MB(0x0079B8, 0xD072);
	ChangeU2MB(0x0079BB, 0xD073);
	ChangeU2MB(0x0079C2, 0xD074);
	ChangeU2MB(0x0079C4, 0xD075);
	ChangeU2MB(0x0079C7, 0xD076);
	ChangeU2MB(0x0079C8, 0xD077);
	ChangeU2MB(0x0079CA, 0xD078);
	ChangeU2MB(0x0079CF, 0xD079);
	ChangeU2MB(0x0079D4, 0xD07A);
	ChangeU2MB(0x0079D6, 0xD07B);
	ChangeU2MB(0x0079DA, 0xD07C);
	ChangeU2MB(0x0079DD, 0xD07D);
	ChangeU2MB(0x0079DE, 0xD07E);
	ChangeU2MB(0x0079E0, 0xD121);
	ChangeU2MB(0x0079E2, 0xD122);
	ChangeU2MB(0x0079E5, 0xD123);
	ChangeU2MB(0x0079EA, 0xD124);
	ChangeU2MB(0x0079EB, 0xD125);
	ChangeU2MB(0x0079ED, 0xD126);
	ChangeU2MB(0x0079F1, 0xD127);
	ChangeU2MB(0x0079F8, 0xD128);
	ChangeU2MB(0x0079FC, 0xD129);
	ChangeU2MB(0x007A02, 0xD12A);
	ChangeU2MB(0x007A03, 0xD12B);
	ChangeU2MB(0x007A07, 0xD12C);
	ChangeU2MB(0x007A09, 0xD12D);
	ChangeU2MB(0x007A0A, 0xD12E);
	ChangeU2MB(0x007A0C, 0xD12F);
	ChangeU2MB(0x007A11, 0xD130);
	ChangeU2MB(0x007A15, 0xD131);
	ChangeU2MB(0x007A1B, 0xD132);
	ChangeU2MB(0x007A1E, 0xD133);
	ChangeU2MB(0x007A21, 0xD134);
	ChangeU2MB(0x007A27, 0xD135);
	ChangeU2MB(0x007A2B, 0xD136);
	ChangeU2MB(0x007A2D, 0xD137);
	ChangeU2MB(0x007A2F, 0xD138);
	ChangeU2MB(0x007A30, 0xD139);
	ChangeU2MB(0x007A34, 0xD13A);
	ChangeU2MB(0x007A35, 0xD13B);
	ChangeU2MB(0x007A38, 0xD13C);
	ChangeU2MB(0x007A39, 0xD13D);
	ChangeU2MB(0x007A3A, 0xD13E);
	ChangeU2MB(0x007A44, 0xD13F);
	ChangeU2MB(0x007A45, 0xD140);
	ChangeU2MB(0x007A47, 0xD141);
	ChangeU2MB(0x007A48, 0xD142);
	ChangeU2MB(0x007A4C, 0xD143);
	ChangeU2MB(0x007A55, 0xD144);
	ChangeU2MB(0x007A56, 0xD145);
	ChangeU2MB(0x007A59, 0xD146);
	ChangeU2MB(0x007A5C, 0xD147);
	ChangeU2MB(0x007A5D, 0xD148);
	ChangeU2MB(0x007A5F, 0xD149);
	ChangeU2MB(0x007A60, 0xD14A);
	ChangeU2MB(0x007A65, 0xD14B);
	ChangeU2MB(0x007A67, 0xD14C);
	ChangeU2MB(0x007A6A, 0xD14D);
	ChangeU2MB(0x007A6D, 0xD14E);
	ChangeU2MB(0x007A75, 0xD14F);
	ChangeU2MB(0x007A78, 0xD150);
	ChangeU2MB(0x007A7E, 0xD151);
	ChangeU2MB(0x007A80, 0xD152);
	ChangeU2MB(0x007A82, 0xD153);
	ChangeU2MB(0x007A85, 0xD154);
	ChangeU2MB(0x007A86, 0xD155);
	ChangeU2MB(0x007A8A, 0xD156);
	ChangeU2MB(0x007A8B, 0xD157);
	ChangeU2MB(0x007A90, 0xD158);
	ChangeU2MB(0x007A91, 0xD159);
	ChangeU2MB(0x007A94, 0xD15A);
	ChangeU2MB(0x007A9E, 0xD15B);
	ChangeU2MB(0x007AA0, 0xD15C);
	ChangeU2MB(0x007AA3, 0xD15D);
	ChangeU2MB(0x007AAC, 0xD15E);
	ChangeU2MB(0x007AB3, 0xD15F);
	ChangeU2MB(0x007AB5, 0xD160);
	ChangeU2MB(0x007AB9, 0xD161);
	ChangeU2MB(0x007ABB, 0xD162);
	ChangeU2MB(0x007ABC, 0xD163);
	ChangeU2MB(0x007AC6, 0xD164);
	ChangeU2MB(0x007AC9, 0xD165);
	ChangeU2MB(0x007ACC, 0xD166);
	ChangeU2MB(0x007ACE, 0xD167);
	ChangeU2MB(0x007ADB, 0xD169);
	ChangeU2MB(0x007AE8, 0xD16A);
	ChangeU2MB(0x007AE9, 0xD16B);
	ChangeU2MB(0x007AEC, 0xD16D);
	ChangeU2MB(0x007AF1, 0xD16E);
	ChangeU2MB(0x007AF4, 0xD16F);
	ChangeU2MB(0x007AFB, 0xD170);
	ChangeU2MB(0x007AFD, 0xD171);
	ChangeU2MB(0x007AFE, 0xD172);
	ChangeU2MB(0x007B07, 0xD173);
	ChangeU2MB(0x007B14, 0xD174);
	ChangeU2MB(0x007B1F, 0xD175);
	ChangeU2MB(0x007B23, 0xD176);
	ChangeU2MB(0x007B27, 0xD177);
	ChangeU2MB(0x007B29, 0xD178);
	ChangeU2MB(0x007B2A, 0xD179);
	ChangeU2MB(0x007B2B, 0xD17A);
	ChangeU2MB(0x007B2D, 0xD17B);
	ChangeU2MB(0x007B2E, 0xD17C);
	ChangeU2MB(0x007B2F, 0xD17D);
	ChangeU2MB(0x007B30, 0xD17E);
	ChangeU2MB(0x007B31, 0xD221);
	ChangeU2MB(0x007B34, 0xD222);
	ChangeU2MB(0x007B3D, 0xD223);
	ChangeU2MB(0x007B3F, 0xD224);
	ChangeU2MB(0x007B40, 0xD225);
	ChangeU2MB(0x007B41, 0xD226);
	ChangeU2MB(0x007B47, 0xD227);
	ChangeU2MB(0x007B4E, 0xD228);
	ChangeU2MB(0x007B55, 0xD229);
	ChangeU2MB(0x007B60, 0xD22A);
	ChangeU2MB(0x007B64, 0xD22B);
	ChangeU2MB(0x007B66, 0xD22C);
	ChangeU2MB(0x007B69, 0xD22D);
	ChangeU2MB(0x007B6A, 0xD22E);
	ChangeU2MB(0x007B6D, 0xD22F);
	ChangeU2MB(0x007B6F, 0xD230);
	ChangeU2MB(0x007B72, 0xD231);
	ChangeU2MB(0x007B73, 0xD232);
	ChangeU2MB(0x007B77, 0xD233);
	ChangeU2MB(0x007B84, 0xD234);
	ChangeU2MB(0x007B89, 0xD235);
	ChangeU2MB(0x007B8E, 0xD236);
	ChangeU2MB(0x007B90, 0xD237);
	ChangeU2MB(0x007B91, 0xD238);
	ChangeU2MB(0x007B96, 0xD239);
	ChangeU2MB(0x007B9B, 0xD23A);
	ChangeU2MB(0x007BA0, 0xD23C);
	ChangeU2MB(0x007BA5, 0xD23D);
	ChangeU2MB(0x007BAC, 0xD23E);
	ChangeU2MB(0x007BAF, 0xD23F);
	ChangeU2MB(0x007BB0, 0xD240);
	ChangeU2MB(0x007BB2, 0xD241);
	ChangeU2MB(0x007BB5, 0xD242);
	ChangeU2MB(0x007BB6, 0xD243);
	ChangeU2MB(0x007BBA, 0xD244);
	ChangeU2MB(0x007BBB, 0xD245);
	ChangeU2MB(0x007BBC, 0xD246);
	ChangeU2MB(0x007BBD, 0xD247);
	ChangeU2MB(0x007BC2, 0xD248);
	ChangeU2MB(0x007BC5, 0xD249);
	ChangeU2MB(0x007BC8, 0xD24A);
	ChangeU2MB(0x007BCA, 0xD24B);
	ChangeU2MB(0x007BD4, 0xD24C);
	ChangeU2MB(0x007BD6, 0xD24D);
	ChangeU2MB(0x007BD7, 0xD24E);
	ChangeU2MB(0x007BD9, 0xD24F);
	ChangeU2MB(0x007BDA, 0xD250);
	ChangeU2MB(0x007BDB, 0xD251);
	ChangeU2MB(0x007BE8, 0xD252);
	ChangeU2MB(0x007BEA, 0xD253);
	ChangeU2MB(0x007BF2, 0xD254);
	ChangeU2MB(0x007BF4, 0xD255);
	ChangeU2MB(0x007BF5, 0xD256);
	ChangeU2MB(0x007BF8, 0xD257);
	ChangeU2MB(0x007BF9, 0xD258);
	ChangeU2MB(0x007BFA, 0xD259);
	ChangeU2MB(0x007BFC, 0xD25A);
	ChangeU2MB(0x007BFE, 0xD25B);
	ChangeU2MB(0x007C01, 0xD25C);
	ChangeU2MB(0x007C02, 0xD25D);
	ChangeU2MB(0x007C03, 0xD25E);
	ChangeU2MB(0x007C04, 0xD25F);
	ChangeU2MB(0x007C06, 0xD260);
	ChangeU2MB(0x007C09, 0xD261);
	ChangeU2MB(0x007C0B, 0xD262);
	ChangeU2MB(0x007C0C, 0xD263);
	ChangeU2MB(0x007C0E, 0xD264);
	ChangeU2MB(0x007C0F, 0xD265);
	ChangeU2MB(0x007C19, 0xD266);
	ChangeU2MB(0x007C1B, 0xD267);
	ChangeU2MB(0x007C20, 0xD268);
	ChangeU2MB(0x007C25, 0xD269);
	ChangeU2MB(0x007C26, 0xD26A);
	ChangeU2MB(0x007C28, 0xD26B);
	ChangeU2MB(0x007C2C, 0xD26C);
	ChangeU2MB(0x007C31, 0xD26D);
	ChangeU2MB(0x007C33, 0xD26E);
	ChangeU2MB(0x007C34, 0xD26F);
	ChangeU2MB(0x007C36, 0xD270);
	ChangeU2MB(0x007C39, 0xD271);
	ChangeU2MB(0x007C3A, 0xD272);
	ChangeU2MB(0x007C46, 0xD273);
	ChangeU2MB(0x007C4A, 0xD274);
	ChangeU2MB(0x007C51, 0xD276);
	ChangeU2MB(0x007C52, 0xD277);
	ChangeU2MB(0x007C53, 0xD278);
	ChangeU2MB(0x007C55, 0xD275);
	ChangeU2MB(0x007C59, 0xD279);
	ChangeU2MB(0x007C5A, 0xD27A);
	ChangeU2MB(0x007C5B, 0xD27B);
	ChangeU2MB(0x007C5C, 0xD27C);
	ChangeU2MB(0x007C5D, 0xD27D);
	ChangeU2MB(0x007C5E, 0xD27E);
	ChangeU2MB(0x007C61, 0xD321);
	ChangeU2MB(0x007C63, 0xD322);
	ChangeU2MB(0x007C67, 0xD323);
	ChangeU2MB(0x007C69, 0xD324);
	ChangeU2MB(0x007C6D, 0xD325);
	ChangeU2MB(0x007C6E, 0xD326);
	ChangeU2MB(0x007C70, 0xD327);
	ChangeU2MB(0x007C72, 0xD328);
	ChangeU2MB(0x007C79, 0xD329);
	ChangeU2MB(0x007C7C, 0xD32A);
	ChangeU2MB(0x007C7D, 0xD32B);
	ChangeU2MB(0x007C86, 0xD32C);
	ChangeU2MB(0x007C87, 0xD32D);
	ChangeU2MB(0x007C8F, 0xD32E);
	ChangeU2MB(0x007C94, 0xD32F);
	ChangeU2MB(0x007C9E, 0xD330);
	ChangeU2MB(0x007CA0, 0xD331);
	ChangeU2MB(0x007CA6, 0xD332);
	ChangeU2MB(0x007CB0, 0xD333);
	ChangeU2MB(0x007CB6, 0xD334);
	ChangeU2MB(0x007CB7, 0xD335);
	ChangeU2MB(0x007CBA, 0xD336);
	ChangeU2MB(0x007CBB, 0xD337);
	ChangeU2MB(0x007CBC, 0xD338);
	ChangeU2MB(0x007CBF, 0xD339);
	ChangeU2MB(0x007CC4, 0xD33A);
	ChangeU2MB(0x007CC7, 0xD33B);
	ChangeU2MB(0x007CC8, 0xD33C);
	ChangeU2MB(0x007CC9, 0xD33D);
	ChangeU2MB(0x007CCD, 0xD33E);
	ChangeU2MB(0x007CCF, 0xD33F);
	ChangeU2MB(0x007CD3, 0xD340);
	ChangeU2MB(0x007CD4, 0xD341);
	ChangeU2MB(0x007CD5, 0xD342);
	ChangeU2MB(0x007CD7, 0xD343);
	ChangeU2MB(0x007CD9, 0xD344);
	ChangeU2MB(0x007CDA, 0xD345);
	ChangeU2MB(0x007CDD, 0xD346);
	ChangeU2MB(0x007CE6, 0xD347);
	ChangeU2MB(0x007CE9, 0xD348);
	ChangeU2MB(0x007CEB, 0xD349);
	ChangeU2MB(0x007CF5, 0xD34A);
	ChangeU2MB(0x007D03, 0xD34B);
	ChangeU2MB(0x007D07, 0xD34C);
	ChangeU2MB(0x007D08, 0xD34D);
	ChangeU2MB(0x007D09, 0xD34E);
	ChangeU2MB(0x007D0F, 0xD34F);
	ChangeU2MB(0x007D11, 0xD350);
	ChangeU2MB(0x007D12, 0xD351);
	ChangeU2MB(0x007D13, 0xD352);
	ChangeU2MB(0x007D16, 0xD353);
	ChangeU2MB(0x007D1D, 0xD354);
	ChangeU2MB(0x007D1E, 0xD355);
	ChangeU2MB(0x007D23, 0xD356);
	ChangeU2MB(0x007D26, 0xD357);
	ChangeU2MB(0x007D2A, 0xD358);
	ChangeU2MB(0x007D2D, 0xD359);
	ChangeU2MB(0x007D31, 0xD35A);
	ChangeU2MB(0x007D3C, 0xD35B);
	ChangeU2MB(0x007D3D, 0xD35C);
	ChangeU2MB(0x007D3E, 0xD35D);
	ChangeU2MB(0x007D40, 0xD35E);
	ChangeU2MB(0x007D41, 0xD35F);
	ChangeU2MB(0x007D47, 0xD360);
	ChangeU2MB(0x007D4D, 0xD362);
	ChangeU2MB(0x007D51, 0xD363);
	ChangeU2MB(0x007D53, 0xD364);
	ChangeU2MB(0x007D57, 0xD365);
	ChangeU2MB(0x007D59, 0xD366);
	ChangeU2MB(0x007D5A, 0xD367);
	ChangeU2MB(0x007D5D, 0xD369);
	ChangeU2MB(0x007D65, 0xD36A);
	ChangeU2MB(0x007D67, 0xD36B);
	ChangeU2MB(0x007D6A, 0xD36C);
	ChangeU2MB(0x007D70, 0xD36D);
	ChangeU2MB(0x007D78, 0xD36E);
	ChangeU2MB(0x007D7A, 0xD36F);
	ChangeU2MB(0x007D7B, 0xD370);
	ChangeU2MB(0x007D7F, 0xD371);
	ChangeU2MB(0x007D81, 0xD372);
	ChangeU2MB(0x007D82, 0xD373);
	ChangeU2MB(0x007D83, 0xD374);
	ChangeU2MB(0x007D85, 0xD375);
	ChangeU2MB(0x007D86, 0xD376);
	ChangeU2MB(0x007D88, 0xD377);
	ChangeU2MB(0x007D8B, 0xD378);
	ChangeU2MB(0x007D8C, 0xD379);
	ChangeU2MB(0x007D8D, 0xD37A);
	ChangeU2MB(0x007D91, 0xD37B);
	ChangeU2MB(0x007D96, 0xD37C);
	ChangeU2MB(0x007D97, 0xD37D);
	ChangeU2MB(0x007D9D, 0xD37E);
	ChangeU2MB(0x007D9E, 0xD421);
	ChangeU2MB(0x007DA6, 0xD422);
	ChangeU2MB(0x007DA7, 0xD423);
	ChangeU2MB(0x007DAA, 0xD424);
	ChangeU2MB(0x007DB3, 0xD425);
	ChangeU2MB(0x007DB6, 0xD426);
	ChangeU2MB(0x007DB9, 0xD428);
	ChangeU2MB(0x007DC2, 0xD429);
	ChangeU2MB(0x007DC3, 0xD42A);
	ChangeU2MB(0x007DC4, 0xD42B);
	ChangeU2MB(0x007DC5, 0xD42C);
	ChangeU2MB(0x007DC6, 0xD42D);
	ChangeU2MB(0x007DCC, 0xD42E);
	ChangeU2MB(0x007DCD, 0xD42F);
	ChangeU2MB(0x007DCE, 0xD430);
	ChangeU2MB(0x007DD7, 0xD431);
	ChangeU2MB(0x007DD9, 0xD432);
	ChangeU2MB(0x007DE2, 0xD434);
	ChangeU2MB(0x007DE5, 0xD435);
	ChangeU2MB(0x007DE6, 0xD436);
	ChangeU2MB(0x007DEA, 0xD437);
	ChangeU2MB(0x007DEB, 0xD438);
	ChangeU2MB(0x007DED, 0xD439);
	ChangeU2MB(0x007DF1, 0xD43A);
	ChangeU2MB(0x007DF5, 0xD43B);
	ChangeU2MB(0x007DF6, 0xD43C);
	ChangeU2MB(0x007DF9, 0xD43D);
	ChangeU2MB(0x007DFA, 0xD43E);
	ChangeU2MB(0x007E00, 0xD433);
	ChangeU2MB(0x007E08, 0xD43F);
	ChangeU2MB(0x007E10, 0xD440);
	ChangeU2MB(0x007E11, 0xD441);
	ChangeU2MB(0x007E15, 0xD442);
	ChangeU2MB(0x007E17, 0xD443);
	ChangeU2MB(0x007E1C, 0xD444);
	ChangeU2MB(0x007E1D, 0xD445);
	ChangeU2MB(0x007E20, 0xD446);
	ChangeU2MB(0x007E27, 0xD447);
	ChangeU2MB(0x007E28, 0xD448);
	ChangeU2MB(0x007E2C, 0xD449);
	ChangeU2MB(0x007E2D, 0xD44A);
	ChangeU2MB(0x007E2F, 0xD44B);
	ChangeU2MB(0x007E33, 0xD44C);
	ChangeU2MB(0x007E36, 0xD44D);
	ChangeU2MB(0x007E3F, 0xD44E);
	ChangeU2MB(0x007E44, 0xD44F);
	ChangeU2MB(0x007E45, 0xD450);
	ChangeU2MB(0x007E47, 0xD451);
	ChangeU2MB(0x007E4E, 0xD452);
	ChangeU2MB(0x007E50, 0xD453);
	ChangeU2MB(0x007E58, 0xD455);
	ChangeU2MB(0x007E5F, 0xD456);
	ChangeU2MB(0x007E61, 0xD457);
	ChangeU2MB(0x007E62, 0xD458);
	ChangeU2MB(0x007E65, 0xD459);
	ChangeU2MB(0x007E6B, 0xD45A);
	ChangeU2MB(0x007E6E, 0xD45B);
	ChangeU2MB(0x007E6F, 0xD45C);
	ChangeU2MB(0x007E73, 0xD45D);
	ChangeU2MB(0x007E78, 0xD45E);
	ChangeU2MB(0x007E7E, 0xD45F);
	ChangeU2MB(0x007E81, 0xD460);
	ChangeU2MB(0x007E86, 0xD461);
	ChangeU2MB(0x007E87, 0xD462);
	ChangeU2MB(0x007E8D, 0xD464);
	ChangeU2MB(0x007E91, 0xD465);
	ChangeU2MB(0x007E95, 0xD466);
	ChangeU2MB(0x007E98, 0xD467);
	ChangeU2MB(0x007E9A, 0xD468);
	ChangeU2MB(0x007E9D, 0xD469);
	ChangeU2MB(0x007E9E, 0xD46A);
	ChangeU2MB(0x007F3B, 0xD46C);
	ChangeU2MB(0x007F3C, 0xD46B);
	ChangeU2MB(0x007F3D, 0xD46D);
	ChangeU2MB(0x007F3E, 0xD46E);
	ChangeU2MB(0x007F3F, 0xD46F);
	ChangeU2MB(0x007F43, 0xD470);
	ChangeU2MB(0x007F44, 0xD471);
	ChangeU2MB(0x007F4F, 0xD473);
	ChangeU2MB(0x007F52, 0xD474);
	ChangeU2MB(0x007F53, 0xD475);
	ChangeU2MB(0x007F5B, 0xD476);
	ChangeU2MB(0x007F5C, 0xD477);
	ChangeU2MB(0x007F5D, 0xD478);
	ChangeU2MB(0x007F61, 0xD479);
	ChangeU2MB(0x007F63, 0xD47A);
	ChangeU2MB(0x007F64, 0xD47B);
	ChangeU2MB(0x007F65, 0xD47C);
	ChangeU2MB(0x007F66, 0xD47D);
	ChangeU2MB(0x007F6D, 0xD47E);
	ChangeU2MB(0x007F71, 0xD521);
	ChangeU2MB(0x007F7D, 0xD522);
	ChangeU2MB(0x007F7E, 0xD523);
	ChangeU2MB(0x007F7F, 0xD524);
	ChangeU2MB(0x007F80, 0xD525);
	ChangeU2MB(0x007F8B, 0xD526);
	ChangeU2MB(0x007F8D, 0xD527);
	ChangeU2MB(0x007F8F, 0xD528);
	ChangeU2MB(0x007F90, 0xD529);
	ChangeU2MB(0x007F91, 0xD52A);
	ChangeU2MB(0x007F96, 0xD52B);
	ChangeU2MB(0x007F97, 0xD52C);
	ChangeU2MB(0x007F9C, 0xD52D);
	ChangeU2MB(0x007FA2, 0xD52F);
	ChangeU2MB(0x007FA6, 0xD530);
	ChangeU2MB(0x007FAA, 0xD531);
	ChangeU2MB(0x007FAD, 0xD532);
	ChangeU2MB(0x007FB4, 0xD533);
	ChangeU2MB(0x007FBC, 0xD534);
	ChangeU2MB(0x007FBF, 0xD535);
	ChangeU2MB(0x007FC0, 0xD536);
	ChangeU2MB(0x007FC3, 0xD537);
	ChangeU2MB(0x007FC8, 0xD538);
	ChangeU2MB(0x007FCE, 0xD539);
	ChangeU2MB(0x007FCF, 0xD53A);
	ChangeU2MB(0x007FDB, 0xD53B);
	ChangeU2MB(0x007FDF, 0xD53C);
	ChangeU2MB(0x007FE3, 0xD53D);
	ChangeU2MB(0x007FE5, 0xD53E);
	ChangeU2MB(0x007FE8, 0xD53F);
	ChangeU2MB(0x007FEC, 0xD540);
	ChangeU2MB(0x007FEE, 0xD541);
	ChangeU2MB(0x007FEF, 0xD542);
	ChangeU2MB(0x007FF2, 0xD543);
	ChangeU2MB(0x007FFA, 0xD544);
	ChangeU2MB(0x007FFD, 0xD545);
	ChangeU2MB(0x007FFE, 0xD546);
	ChangeU2MB(0x007FFF, 0xD547);
	ChangeU2MB(0x008007, 0xD548);
	ChangeU2MB(0x008008, 0xD549);
	ChangeU2MB(0x00800A, 0xD54A);
	ChangeU2MB(0x00800D, 0xD54B);
	ChangeU2MB(0x00800E, 0xD54C);
	ChangeU2MB(0x00800F, 0xD54D);
	ChangeU2MB(0x008011, 0xD54E);
	ChangeU2MB(0x008013, 0xD54F);
	ChangeU2MB(0x008014, 0xD550);
	ChangeU2MB(0x008016, 0xD551);
	ChangeU2MB(0x00801D, 0xD552);
	ChangeU2MB(0x00801E, 0xD553);
	ChangeU2MB(0x00801F, 0xD554);
	ChangeU2MB(0x008020, 0xD555);
	ChangeU2MB(0x008024, 0xD556);
	ChangeU2MB(0x008026, 0xD557);
	ChangeU2MB(0x00802C, 0xD558);
	ChangeU2MB(0x00802E, 0xD559);
	ChangeU2MB(0x008030, 0xD55A);
	ChangeU2MB(0x008034, 0xD55B);
	ChangeU2MB(0x008035, 0xD55C);
	ChangeU2MB(0x008037, 0xD55D);
	ChangeU2MB(0x008039, 0xD55E);
	ChangeU2MB(0x00803A, 0xD55F);
	ChangeU2MB(0x00803C, 0xD560);
	ChangeU2MB(0x00803E, 0xD561);
	ChangeU2MB(0x008040, 0xD562);
	ChangeU2MB(0x008044, 0xD563);
	ChangeU2MB(0x008060, 0xD564);
	ChangeU2MB(0x008064, 0xD565);
	ChangeU2MB(0x008066, 0xD566);
	ChangeU2MB(0x00806D, 0xD567);
	ChangeU2MB(0x008071, 0xD568);
	ChangeU2MB(0x008075, 0xD569);
	ChangeU2MB(0x008081, 0xD56A);
	ChangeU2MB(0x008088, 0xD56B);
	ChangeU2MB(0x00808E, 0xD56C);
	ChangeU2MB(0x00809C, 0xD56D);
	ChangeU2MB(0x00809E, 0xD56E);
	ChangeU2MB(0x0080A6, 0xD56F);
	ChangeU2MB(0x0080A7, 0xD570);
	ChangeU2MB(0x0080AB, 0xD571);
	ChangeU2MB(0x0080B8, 0xD572);
	ChangeU2MB(0x0080B9, 0xD573);
	ChangeU2MB(0x0080C8, 0xD574);
	ChangeU2MB(0x0080CD, 0xD575);
	ChangeU2MB(0x0080CF, 0xD576);
	ChangeU2MB(0x0080D2, 0xD577);
	ChangeU2MB(0x0080D4, 0xD578);
	ChangeU2MB(0x0080D5, 0xD579);
	ChangeU2MB(0x0080D7, 0xD57A);
	ChangeU2MB(0x0080D8, 0xD57B);
	ChangeU2MB(0x0080E0, 0xD57C);
	ChangeU2MB(0x0080ED, 0xD57D);
	ChangeU2MB(0x0080EE, 0xD57E);
	ChangeU2MB(0x0080F0, 0xD621);
	ChangeU2MB(0x0080F2, 0xD622);
	ChangeU2MB(0x0080F3, 0xD623);
	ChangeU2MB(0x0080F6, 0xD624);
	ChangeU2MB(0x0080F9, 0xD625);
	ChangeU2MB(0x0080FA, 0xD626);
	ChangeU2MB(0x0080FE, 0xD627);
	ChangeU2MB(0x008103, 0xD628);
	ChangeU2MB(0x00810B, 0xD629);
	ChangeU2MB(0x008116, 0xD62A);
	ChangeU2MB(0x008117, 0xD62B);
	ChangeU2MB(0x008118, 0xD62C);
	ChangeU2MB(0x00811C, 0xD62D);
	ChangeU2MB(0x00811E, 0xD62E);
	ChangeU2MB(0x008120, 0xD62F);
	ChangeU2MB(0x008124, 0xD630);
	ChangeU2MB(0x008127, 0xD631);
	ChangeU2MB(0x00812C, 0xD632);
	ChangeU2MB(0x008130, 0xD633);
	ChangeU2MB(0x008135, 0xD634);
	ChangeU2MB(0x00813A, 0xD635);
	ChangeU2MB(0x00813C, 0xD636);
	ChangeU2MB(0x008145, 0xD637);
	ChangeU2MB(0x008147, 0xD638);
	ChangeU2MB(0x00814A, 0xD639);
	ChangeU2MB(0x00814C, 0xD63A);
	ChangeU2MB(0x008152, 0xD63B);
	ChangeU2MB(0x008157, 0xD63C);
	ChangeU2MB(0x008160, 0xD63D);
	ChangeU2MB(0x008161, 0xD63E);
	ChangeU2MB(0x008167, 0xD63F);
	ChangeU2MB(0x008168, 0xD640);
	ChangeU2MB(0x008169, 0xD641);
	ChangeU2MB(0x00816D, 0xD642);
	ChangeU2MB(0x00816F, 0xD643);
	ChangeU2MB(0x008177, 0xD644);
	ChangeU2MB(0x008181, 0xD645);
	ChangeU2MB(0x008184, 0xD647);
	ChangeU2MB(0x008185, 0xD648);
	ChangeU2MB(0x008186, 0xD649);
	ChangeU2MB(0x00818B, 0xD64A);
	ChangeU2MB(0x00818E, 0xD64B);
	ChangeU2MB(0x008190, 0xD646);
	ChangeU2MB(0x008196, 0xD64C);
	ChangeU2MB(0x008198, 0xD64D);
	ChangeU2MB(0x00819B, 0xD64E);
	ChangeU2MB(0x00819E, 0xD64F);
	ChangeU2MB(0x0081A2, 0xD650);
	ChangeU2MB(0x0081AE, 0xD651);
	ChangeU2MB(0x0081B2, 0xD652);
	ChangeU2MB(0x0081B4, 0xD653);
	ChangeU2MB(0x0081BB, 0xD654);
	ChangeU2MB(0x0081C3, 0xD656);
	ChangeU2MB(0x0081C5, 0xD657);
	ChangeU2MB(0x0081CA, 0xD658);
	ChangeU2MB(0x0081CB, 0xD655);
	ChangeU2MB(0x0081CE, 0xD659);
	ChangeU2MB(0x0081CF, 0xD65A);
	ChangeU2MB(0x0081D5, 0xD65B);
	ChangeU2MB(0x0081D7, 0xD65C);
	ChangeU2MB(0x0081DB, 0xD65D);
	ChangeU2MB(0x0081DD, 0xD65E);
	ChangeU2MB(0x0081DE, 0xD65F);
	ChangeU2MB(0x0081E1, 0xD660);
	ChangeU2MB(0x0081E4, 0xD661);
	ChangeU2MB(0x0081EB, 0xD662);
	ChangeU2MB(0x0081EC, 0xD663);
	ChangeU2MB(0x0081F0, 0xD664);
	ChangeU2MB(0x0081F1, 0xD665);
	ChangeU2MB(0x0081F2, 0xD666);
	ChangeU2MB(0x0081F5, 0xD667);
	ChangeU2MB(0x0081F6, 0xD668);
	ChangeU2MB(0x0081F8, 0xD669);
	ChangeU2MB(0x0081F9, 0xD66A);
	ChangeU2MB(0x0081FD, 0xD66B);
	ChangeU2MB(0x0081FF, 0xD66C);
	ChangeU2MB(0x008200, 0xD66D);
	ChangeU2MB(0x008203, 0xD66E);
	ChangeU2MB(0x00820F, 0xD66F);
	ChangeU2MB(0x008213, 0xD670);
	ChangeU2MB(0x008214, 0xD671);
	ChangeU2MB(0x008219, 0xD672);
	ChangeU2MB(0x00821A, 0xD673);
	ChangeU2MB(0x00821D, 0xD674);
	ChangeU2MB(0x008221, 0xD675);
	ChangeU2MB(0x008222, 0xD676);
	ChangeU2MB(0x008228, 0xD677);
	ChangeU2MB(0x008232, 0xD678);
	ChangeU2MB(0x008234, 0xD679);
	ChangeU2MB(0x00823A, 0xD67A);
	ChangeU2MB(0x008243, 0xD67B);
	ChangeU2MB(0x008244, 0xD67C);
	ChangeU2MB(0x008245, 0xD67D);
	ChangeU2MB(0x008246, 0xD67E);
	ChangeU2MB(0x00824B, 0xD721);
	ChangeU2MB(0x00824E, 0xD722);
	ChangeU2MB(0x00824F, 0xD723);
	ChangeU2MB(0x008251, 0xD724);
	ChangeU2MB(0x008256, 0xD725);
	ChangeU2MB(0x00825C, 0xD726);
	ChangeU2MB(0x008260, 0xD727);
	ChangeU2MB(0x008263, 0xD728);
	ChangeU2MB(0x008267, 0xD729);
	ChangeU2MB(0x00826D, 0xD72A);
	ChangeU2MB(0x008274, 0xD72B);
	ChangeU2MB(0x00827B, 0xD72C);
	ChangeU2MB(0x00827D, 0xD72D);
	ChangeU2MB(0x00827F, 0xD72E);
	ChangeU2MB(0x008280, 0xD72F);
	ChangeU2MB(0x008281, 0xD730);
	ChangeU2MB(0x008283, 0xD731);
	ChangeU2MB(0x008284, 0xD732);
	ChangeU2MB(0x008287, 0xD733);
	ChangeU2MB(0x008289, 0xD734);
	ChangeU2MB(0x00828A, 0xD735);
	ChangeU2MB(0x00828E, 0xD736);
	ChangeU2MB(0x008291, 0xD737);
	ChangeU2MB(0x008294, 0xD738);
	ChangeU2MB(0x008296, 0xD739);
	ChangeU2MB(0x008298, 0xD73A);
	ChangeU2MB(0x00829A, 0xD73B);
	ChangeU2MB(0x00829B, 0xD73C);
	ChangeU2MB(0x0082A0, 0xD73D);
	ChangeU2MB(0x0082A1, 0xD73E);
	ChangeU2MB(0x0082A3, 0xD73F);
	ChangeU2MB(0x0082A4, 0xD740);
	ChangeU2MB(0x0082A7, 0xD741);
	ChangeU2MB(0x0082A8, 0xD742);
	ChangeU2MB(0x0082A9, 0xD743);
	ChangeU2MB(0x0082AA, 0xD744);
	ChangeU2MB(0x0082AE, 0xD745);
	ChangeU2MB(0x0082B0, 0xD746);
	ChangeU2MB(0x0082B2, 0xD747);
	ChangeU2MB(0x0082B4, 0xD748);
	ChangeU2MB(0x0082B7, 0xD749);
	ChangeU2MB(0x0082BA, 0xD74A);
	ChangeU2MB(0x0082BC, 0xD74B);
	ChangeU2MB(0x0082BE, 0xD74C);
	ChangeU2MB(0x0082BF, 0xD74D);
	ChangeU2MB(0x0082C6, 0xD74E);
	ChangeU2MB(0x0082D0, 0xD74F);
	ChangeU2MB(0x0082D5, 0xD750);
	ChangeU2MB(0x0082DA, 0xD751);
	ChangeU2MB(0x0082E0, 0xD752);
	ChangeU2MB(0x0082E2, 0xD753);
	ChangeU2MB(0x0082E4, 0xD754);
	ChangeU2MB(0x0082E8, 0xD755);
	ChangeU2MB(0x0082EA, 0xD756);
	ChangeU2MB(0x0082ED, 0xD757);
	ChangeU2MB(0x0082EF, 0xD758);
	ChangeU2MB(0x0082F6, 0xD759);
	ChangeU2MB(0x0082F7, 0xD75A);
	ChangeU2MB(0x0082FD, 0xD75B);
	ChangeU2MB(0x0082FE, 0xD75C);
	ChangeU2MB(0x008300, 0xD75D);
	ChangeU2MB(0x008307, 0xD75F);
	ChangeU2MB(0x008308, 0xD760);
	ChangeU2MB(0x00830A, 0xD761);
	ChangeU2MB(0x00830B, 0xD762);
	ChangeU2MB(0x00831B, 0xD764);
	ChangeU2MB(0x00831D, 0xD765);
	ChangeU2MB(0x00831E, 0xD766);
	ChangeU2MB(0x00831F, 0xD767);
	ChangeU2MB(0x008321, 0xD768);
	ChangeU2MB(0x008322, 0xD769);
	ChangeU2MB(0x00832C, 0xD76A);
	ChangeU2MB(0x00832D, 0xD76B);
	ChangeU2MB(0x00832E, 0xD76C);
	ChangeU2MB(0x008330, 0xD76D);
	ChangeU2MB(0x008333, 0xD76E);
	ChangeU2MB(0x008337, 0xD76F);
	ChangeU2MB(0x00833A, 0xD770);
	ChangeU2MB(0x00833C, 0xD771);
	ChangeU2MB(0x00833D, 0xD772);
	ChangeU2MB(0x008342, 0xD773);
	ChangeU2MB(0x008343, 0xD774);
	ChangeU2MB(0x008344, 0xD775);
	ChangeU2MB(0x008347, 0xD776);
	ChangeU2MB(0x00834D, 0xD777);
	ChangeU2MB(0x00834E, 0xD778);
	ChangeU2MB(0x008351, 0xD779);
	ChangeU2MB(0x008353, 0xD83E);
	ChangeU2MB(0x008354, 0xD763);
	ChangeU2MB(0x008355, 0xD77A);
	ChangeU2MB(0x008356, 0xD77B);
	ChangeU2MB(0x008357, 0xD77C);
	ChangeU2MB(0x008370, 0xD77D);
	ChangeU2MB(0x008378, 0xD77E);
	ChangeU2MB(0x00837D, 0xD821);
	ChangeU2MB(0x008380, 0xD823);
	ChangeU2MB(0x008382, 0xD824);
	ChangeU2MB(0x008384, 0xD825);
	ChangeU2MB(0x008386, 0xD826);
	ChangeU2MB(0x00838D, 0xD827);
	ChangeU2MB(0x008392, 0xD828);
	ChangeU2MB(0x008394, 0xD829);
	ChangeU2MB(0x008395, 0xD82A);
	ChangeU2MB(0x008398, 0xD82B);
	ChangeU2MB(0x008399, 0xD82C);
	ChangeU2MB(0x00839B, 0xD82D);
	ChangeU2MB(0x00839C, 0xD82E);
	ChangeU2MB(0x00839D, 0xD82F);
	ChangeU2MB(0x0083A6, 0xD830);
	ChangeU2MB(0x0083A7, 0xD831);
	ChangeU2MB(0x0083A9, 0xD832);
	ChangeU2MB(0x0083AC, 0xD833);
	ChangeU2MB(0x0083AD, 0xD84C);
	ChangeU2MB(0x0083BE, 0xD834);
	ChangeU2MB(0x0083BF, 0xD835);
	ChangeU2MB(0x0083C0, 0xD836);
	ChangeU2MB(0x0083C9, 0xD838);
	ChangeU2MB(0x0083CF, 0xD839);
	ChangeU2MB(0x0083D0, 0xD83A);
	ChangeU2MB(0x0083D1, 0xD83B);
	ChangeU2MB(0x0083D4, 0xD83C);
	ChangeU2MB(0x0083DD, 0xD83D);
	ChangeU2MB(0x0083E8, 0xD83F);
	ChangeU2MB(0x0083EA, 0xD840);
	ChangeU2MB(0x0083F8, 0xD842);
	ChangeU2MB(0x0083F9, 0xD843);
	ChangeU2MB(0x0083FC, 0xD844);
	ChangeU2MB(0x008401, 0xD845);
	ChangeU2MB(0x008406, 0xD846);
	ChangeU2MB(0x00840A, 0xD847);
	ChangeU2MB(0x00840F, 0xD848);
	ChangeU2MB(0x008411, 0xD849);
	ChangeU2MB(0x008415, 0xD84A);
	ChangeU2MB(0x008419, 0xD84B);
	ChangeU2MB(0x00842F, 0xD84D);
	ChangeU2MB(0x008439, 0xD84E);
	ChangeU2MB(0x008445, 0xD84F);
	ChangeU2MB(0x008447, 0xD850);
	ChangeU2MB(0x00844A, 0xD852);
	ChangeU2MB(0x00844D, 0xD853);
	ChangeU2MB(0x00844F, 0xD854);
	ChangeU2MB(0x008451, 0xD855);
	ChangeU2MB(0x008452, 0xD856);
	ChangeU2MB(0x008456, 0xD857);
	ChangeU2MB(0x008458, 0xD858);
	ChangeU2MB(0x008459, 0xD859);
	ChangeU2MB(0x00845A, 0xD85A);
	ChangeU2MB(0x00845C, 0xD85B);
	ChangeU2MB(0x008460, 0xD85C);
	ChangeU2MB(0x008464, 0xD85D);
	ChangeU2MB(0x008465, 0xD85E);
	ChangeU2MB(0x008467, 0xD85F);
	ChangeU2MB(0x00846A, 0xD860);
	ChangeU2MB(0x008470, 0xD861);
	ChangeU2MB(0x008473, 0xD862);
	ChangeU2MB(0x008474, 0xD863);
	ChangeU2MB(0x008476, 0xD864);
	ChangeU2MB(0x008478, 0xD865);
	ChangeU2MB(0x00847C, 0xD866);
	ChangeU2MB(0x00847D, 0xD867);
	ChangeU2MB(0x008481, 0xD868);
	ChangeU2MB(0x008485, 0xD869);
	ChangeU2MB(0x008492, 0xD86A);
	ChangeU2MB(0x008493, 0xD86B);
	ChangeU2MB(0x008495, 0xD86C);
	ChangeU2MB(0x00849E, 0xD86D);
	ChangeU2MB(0x0084A6, 0xD86E);
	ChangeU2MB(0x0084A8, 0xD86F);
	ChangeU2MB(0x0084A9, 0xD870);
	ChangeU2MB(0x0084AA, 0xD871);
	ChangeU2MB(0x0084AF, 0xD872);
	ChangeU2MB(0x0084B1, 0xD873);
	ChangeU2MB(0x0084BA, 0xD875);
	ChangeU2MB(0x0084BD, 0xD876);
	ChangeU2MB(0x0084BE, 0xD877);
	ChangeU2MB(0x0084C0, 0xD878);
	ChangeU2MB(0x0084C2, 0xD879);
	ChangeU2MB(0x0084C7, 0xD87A);
	ChangeU2MB(0x0084C8, 0xD87B);
	ChangeU2MB(0x0084CC, 0xD87C);
	ChangeU2MB(0x0084CF, 0xD87D);
	ChangeU2MB(0x0084D3, 0xD87E);
	ChangeU2MB(0x0084E7, 0xD922);
	ChangeU2MB(0x0084EA, 0xD923);
	ChangeU2MB(0x0084EF, 0xD924);
	ChangeU2MB(0x0084F0, 0xD925);
	ChangeU2MB(0x0084F1, 0xD926);
	ChangeU2MB(0x0084F2, 0xD927);
	ChangeU2MB(0x0084F7, 0xD928);
	ChangeU2MB(0x0084FA, 0xD92A);
	ChangeU2MB(0x0084FB, 0xD92B);
	ChangeU2MB(0x0084FD, 0xD92C);
	ChangeU2MB(0x008502, 0xD92D);
	ChangeU2MB(0x008503, 0xD92E);
	ChangeU2MB(0x008507, 0xD92F);
	ChangeU2MB(0x00850C, 0xD930);
	ChangeU2MB(0x00850E, 0xD931);
	ChangeU2MB(0x008510, 0xD932);
	ChangeU2MB(0x00851C, 0xD933);
	ChangeU2MB(0x00851E, 0xD934);
	ChangeU2MB(0x008522, 0xD935);
	ChangeU2MB(0x008523, 0xD936);
	ChangeU2MB(0x008524, 0xD937);
	ChangeU2MB(0x008525, 0xD938);
	ChangeU2MB(0x008527, 0xD939);
	ChangeU2MB(0x00852A, 0xD93A);
	ChangeU2MB(0x00852B, 0xD93B);
	ChangeU2MB(0x00852F, 0xD93C);
	ChangeU2MB(0x008532, 0xD929);
	ChangeU2MB(0x008533, 0xD93D);
	ChangeU2MB(0x008534, 0xD93E);
	ChangeU2MB(0x008536, 0xD93F);
	ChangeU2MB(0x00853F, 0xD940);
	ChangeU2MB(0x008546, 0xD941);
	ChangeU2MB(0x00854F, 0xD942);
	ChangeU2MB(0x008550, 0xD943);
	ChangeU2MB(0x008551, 0xD944);
	ChangeU2MB(0x008552, 0xD945);
	ChangeU2MB(0x008556, 0xD947);
	ChangeU2MB(0x00855C, 0xD949);
	ChangeU2MB(0x00855D, 0xD94A);
	ChangeU2MB(0x00855E, 0xD94B);
	ChangeU2MB(0x00855F, 0xD94C);
	ChangeU2MB(0x008560, 0xD94D);
	ChangeU2MB(0x008561, 0xD94E);
	ChangeU2MB(0x008562, 0xD94F);
	ChangeU2MB(0x008564, 0xD950);
	ChangeU2MB(0x00856F, 0xD952);
	ChangeU2MB(0x008579, 0xD953);
	ChangeU2MB(0x00857A, 0xD954);
	ChangeU2MB(0x00857B, 0xD955);
	ChangeU2MB(0x00857D, 0xD956);
	ChangeU2MB(0x00857F, 0xD957);
	ChangeU2MB(0x008581, 0xD958);
	ChangeU2MB(0x008585, 0xD959);
	ChangeU2MB(0x008586, 0xD95A);
	ChangeU2MB(0x008589, 0xD95B);
	ChangeU2MB(0x00858B, 0xD95C);
	ChangeU2MB(0x00858C, 0xD95D);
	ChangeU2MB(0x00858F, 0xD95E);
	ChangeU2MB(0x008593, 0xD95F);
	ChangeU2MB(0x008598, 0xD960);
	ChangeU2MB(0x00859D, 0xD961);
	ChangeU2MB(0x00859F, 0xD962);
	ChangeU2MB(0x0085A0, 0xD963);
	ChangeU2MB(0x0085A2, 0xD964);
	ChangeU2MB(0x0085A5, 0xD965);
	ChangeU2MB(0x0085A7, 0xD966);
	ChangeU2MB(0x0085AD, 0xD974);
	ChangeU2MB(0x0085B4, 0xD967);
	ChangeU2MB(0x0085B6, 0xD968);
	ChangeU2MB(0x0085B7, 0xD969);
	ChangeU2MB(0x0085B8, 0xD96A);
	ChangeU2MB(0x0085BC, 0xD96B);
	ChangeU2MB(0x0085BD, 0xD96C);
	ChangeU2MB(0x0085BE, 0xD96D);
	ChangeU2MB(0x0085BF, 0xD96E);
	ChangeU2MB(0x0085C2, 0xD96F);
	ChangeU2MB(0x0085C7, 0xD970);
	ChangeU2MB(0x0085CA, 0xD971);
	ChangeU2MB(0x0085CB, 0xD972);
	ChangeU2MB(0x0085CE, 0xD973);
	ChangeU2MB(0x0085D8, 0xD975);
	ChangeU2MB(0x0085DA, 0xD976);
	ChangeU2MB(0x0085DF, 0xD977);
	ChangeU2MB(0x0085E0, 0xD978);
	ChangeU2MB(0x0085E6, 0xD979);
	ChangeU2MB(0x0085E8, 0xD97A);
	ChangeU2MB(0x0085ED, 0xD97B);
	ChangeU2MB(0x0085F3, 0xD97C);
	ChangeU2MB(0x0085F6, 0xD97D);
	ChangeU2MB(0x0085FC, 0xD97E);
	ChangeU2MB(0x0085FF, 0xDA21);
	ChangeU2MB(0x008600, 0xDA22);
	ChangeU2MB(0x008604, 0xDA23);
	ChangeU2MB(0x008605, 0xDA24);
	ChangeU2MB(0x00860D, 0xDA25);
	ChangeU2MB(0x00860E, 0xDA26);
	ChangeU2MB(0x008610, 0xDA27);
	ChangeU2MB(0x008611, 0xDA28);
	ChangeU2MB(0x008612, 0xDA29);
	ChangeU2MB(0x008618, 0xDA2A);
	ChangeU2MB(0x008619, 0xDA2B);
	ChangeU2MB(0x00861B, 0xDA2C);
	ChangeU2MB(0x00861E, 0xDA2D);
	ChangeU2MB(0x008621, 0xDA2E);
	ChangeU2MB(0x008627, 0xDA2F);
	ChangeU2MB(0x008629, 0xDA30);
	ChangeU2MB(0x008636, 0xDA31);
	ChangeU2MB(0x008638, 0xDA32);
	ChangeU2MB(0x00863A, 0xDA33);
	ChangeU2MB(0x00863C, 0xDA34);
	ChangeU2MB(0x00863D, 0xDA35);
	ChangeU2MB(0x008640, 0xDA36);
	ChangeU2MB(0x008641, 0xB866);
	ChangeU2MB(0x008642, 0xDA37);
	ChangeU2MB(0x008646, 0xDA38);
	ChangeU2MB(0x008652, 0xDA39);
	ChangeU2MB(0x008653, 0xDA3A);
	ChangeU2MB(0x008656, 0xDA3B);
	ChangeU2MB(0x008657, 0xDA3C);
	ChangeU2MB(0x008658, 0xDA3D);
	ChangeU2MB(0x008659, 0xDA3E);
	ChangeU2MB(0x00865D, 0xDA3F);
	ChangeU2MB(0x008660, 0xDA40);
	ChangeU2MB(0x008661, 0xDA41);
	ChangeU2MB(0x008662, 0xDA42);
	ChangeU2MB(0x008663, 0xDA43);
	ChangeU2MB(0x008664, 0xDA44);
	ChangeU2MB(0x008669, 0xDA45);
	ChangeU2MB(0x00866C, 0xDA46);
	ChangeU2MB(0x00866F, 0xDA47);
	ChangeU2MB(0x008675, 0xDA48);
	ChangeU2MB(0x008676, 0xDA49);
	ChangeU2MB(0x008677, 0xDA4A);
	ChangeU2MB(0x00867A, 0xDA4B);
	ChangeU2MB(0x008688, 0xDA6D);
	ChangeU2MB(0x00868D, 0xDA4C);
	ChangeU2MB(0x008691, 0xDA4D);
	ChangeU2MB(0x008696, 0xDA4E);
	ChangeU2MB(0x008698, 0xDA4F);
	ChangeU2MB(0x00869A, 0xDA50);
	ChangeU2MB(0x00869C, 0xDA51);
	ChangeU2MB(0x0086A1, 0xDA52);
	ChangeU2MB(0x0086A6, 0xDA53);
	ChangeU2MB(0x0086A7, 0xDA54);
	ChangeU2MB(0x0086A8, 0xDA55);
	ChangeU2MB(0x0086AD, 0xDA56);
	ChangeU2MB(0x0086B1, 0xDA57);
	ChangeU2MB(0x0086B3, 0xDA58);
	ChangeU2MB(0x0086B4, 0xDA59);
	ChangeU2MB(0x0086B5, 0xDA5A);
	ChangeU2MB(0x0086B7, 0xDA5B);
	ChangeU2MB(0x0086B8, 0xDA5C);
	ChangeU2MB(0x0086B9, 0xDA5D);
	ChangeU2MB(0x0086BF, 0xDA5E);
	ChangeU2MB(0x0086C0, 0xDA5F);
	ChangeU2MB(0x0086C1, 0xDA60);
	ChangeU2MB(0x0086C3, 0xDA61);
	ChangeU2MB(0x0086C5, 0xDA62);
	ChangeU2MB(0x0086D1, 0xDA63);
	ChangeU2MB(0x0086D2, 0xDA64);
	ChangeU2MB(0x0086D5, 0xDA65);
	ChangeU2MB(0x0086D7, 0xDA66);
	ChangeU2MB(0x0086DA, 0xDA67);
	ChangeU2MB(0x0086DC, 0xDA68);
	ChangeU2MB(0x0086E0, 0xDA69);
	ChangeU2MB(0x0086E3, 0xDA6A);
	ChangeU2MB(0x0086E5, 0xDA6B);
	ChangeU2MB(0x0086E7, 0xDA6C);
	ChangeU2MB(0x0086FA, 0xDA6E);
	ChangeU2MB(0x0086FC, 0xDA6F);
	ChangeU2MB(0x0086FD, 0xDA70);
	ChangeU2MB(0x008704, 0xDA71);
	ChangeU2MB(0x008705, 0xDA72);
	ChangeU2MB(0x008707, 0xDA73);
	ChangeU2MB(0x00870B, 0xDA74);
	ChangeU2MB(0x00870E, 0xDA75);
	ChangeU2MB(0x00870F, 0xDA76);
	ChangeU2MB(0x008710, 0xDA77);
	ChangeU2MB(0x008713, 0xDA78);
	ChangeU2MB(0x008714, 0xDA79);
	ChangeU2MB(0x008719, 0xDA7A);
	ChangeU2MB(0x00871E, 0xDA7B);
	ChangeU2MB(0x00871F, 0xDA7C);
	ChangeU2MB(0x008721, 0xDA7D);
	ChangeU2MB(0x008723, 0xDA7E);
	ChangeU2MB(0x008728, 0xDB21);
	ChangeU2MB(0x00872E, 0xDB22);
	ChangeU2MB(0x00872F, 0xDB23);
	ChangeU2MB(0x008731, 0xDB24);
	ChangeU2MB(0x008732, 0xDB25);
	ChangeU2MB(0x008739, 0xDB26);
	ChangeU2MB(0x00873A, 0xDB27);
	ChangeU2MB(0x00873C, 0xDB28);
	ChangeU2MB(0x00873D, 0xDB29);
	ChangeU2MB(0x00873E, 0xDB2A);
	ChangeU2MB(0x008740, 0xDB2B);
	ChangeU2MB(0x008743, 0xDB2C);
	ChangeU2MB(0x008745, 0xDB2D);
	ChangeU2MB(0x00874D, 0xDB2E);
	ChangeU2MB(0x008758, 0xDB2F);
	ChangeU2MB(0x00875D, 0xDB30);
	ChangeU2MB(0x008761, 0xDB31);
	ChangeU2MB(0x008764, 0xDB32);
	ChangeU2MB(0x008765, 0xDB33);
	ChangeU2MB(0x00876F, 0xDB34);
	ChangeU2MB(0x008771, 0xDB35);
	ChangeU2MB(0x008772, 0xDB36);
	ChangeU2MB(0x00877B, 0xDB37);
	ChangeU2MB(0x008783, 0xDB38);
	ChangeU2MB(0x008784, 0xDB39);
	ChangeU2MB(0x008785, 0xDB3A);
	ChangeU2MB(0x008786, 0xDB3B);
	ChangeU2MB(0x008787, 0xDB3C);
	ChangeU2MB(0x008788, 0xDB3D);
	ChangeU2MB(0x008789, 0xDB3E);
	ChangeU2MB(0x00878B, 0xDB3F);
	ChangeU2MB(0x00878C, 0xDB40);
	ChangeU2MB(0x008790, 0xDB41);
	ChangeU2MB(0x008793, 0xDB42);
	ChangeU2MB(0x008795, 0xDB43);
	ChangeU2MB(0x008797, 0xDB44);
	ChangeU2MB(0x008798, 0xDB45);
	ChangeU2MB(0x008799, 0xDB46);
	ChangeU2MB(0x00879E, 0xDB47);
	ChangeU2MB(0x0087A0, 0xDB48);
	ChangeU2MB(0x0087A3, 0xDB49);
	ChangeU2MB(0x0087A7, 0xDB4A);
	ChangeU2MB(0x0087AC, 0xDB4B);
	ChangeU2MB(0x0087AD, 0xDB4C);
	ChangeU2MB(0x0087AE, 0xDB4D);
	ChangeU2MB(0x0087B1, 0xDB4E);
	ChangeU2MB(0x0087B5, 0xDB4F);
	ChangeU2MB(0x0087BE, 0xDB50);
	ChangeU2MB(0x0087BF, 0xDB51);
	ChangeU2MB(0x0087C1, 0xDB52);
	ChangeU2MB(0x0087C8, 0xDB53);
	ChangeU2MB(0x0087C9, 0xDB54);
	ChangeU2MB(0x0087CA, 0xDB55);
	ChangeU2MB(0x0087CE, 0xDB56);
	ChangeU2MB(0x0087D5, 0xDB57);
	ChangeU2MB(0x0087D6, 0xDB58);
	ChangeU2MB(0x0087D9, 0xDB59);
	ChangeU2MB(0x0087DA, 0xDB5A);
	ChangeU2MB(0x0087DC, 0xDB5B);
	ChangeU2MB(0x0087DF, 0xDB5C);
	ChangeU2MB(0x0087E2, 0xDB5D);
	ChangeU2MB(0x0087E3, 0xDB5E);
	ChangeU2MB(0x0087E4, 0xDB5F);
	ChangeU2MB(0x0087EA, 0xDB60);
	ChangeU2MB(0x0087EB, 0xDB61);
	ChangeU2MB(0x0087ED, 0xDB62);
	ChangeU2MB(0x0087F1, 0xDB63);
	ChangeU2MB(0x0087F3, 0xDB64);
	ChangeU2MB(0x0087F8, 0xDB65);
	ChangeU2MB(0x0087FA, 0xDB66);
	ChangeU2MB(0x0087FF, 0xDB67);
	ChangeU2MB(0x008801, 0xDB68);
	ChangeU2MB(0x008803, 0xDB69);
	ChangeU2MB(0x008806, 0xDB6A);
	ChangeU2MB(0x008809, 0xDB6B);
	ChangeU2MB(0x00880A, 0xDB6C);
	ChangeU2MB(0x00880B, 0xDB6D);
	ChangeU2MB(0x008810, 0xDB6E);
	ChangeU2MB(0x008812, 0xDB70);
	ChangeU2MB(0x008813, 0xDB71);
	ChangeU2MB(0x008814, 0xDB72);
	ChangeU2MB(0x008818, 0xDB73);
	ChangeU2MB(0x008819, 0xDB6F);
	ChangeU2MB(0x00881A, 0xDB74);
	ChangeU2MB(0x00881B, 0xDB75);
	ChangeU2MB(0x00881C, 0xDB76);
	ChangeU2MB(0x00881E, 0xDB77);
	ChangeU2MB(0x00881F, 0xDB78);
	ChangeU2MB(0x008828, 0xDB79);
	ChangeU2MB(0x00882D, 0xDB7A);
	ChangeU2MB(0x00882E, 0xDB7B);
	ChangeU2MB(0x008830, 0xDB7C);
	ChangeU2MB(0x008832, 0xDB7D);
	ChangeU2MB(0x008835, 0xDB7E);
	ChangeU2MB(0x00883A, 0xDC21);
	ChangeU2MB(0x00883C, 0xDC22);
	ChangeU2MB(0x008841, 0xDC23);
	ChangeU2MB(0x008843, 0xDC24);
	ChangeU2MB(0x008845, 0xDC25);
	ChangeU2MB(0x008848, 0xDC26);
	ChangeU2MB(0x008849, 0xDC27);
	ChangeU2MB(0x00884A, 0xDC28);
	ChangeU2MB(0x00884B, 0xDC29);
	ChangeU2MB(0x00884E, 0xDC2A);
	ChangeU2MB(0x008851, 0xDC2B);
	ChangeU2MB(0x008855, 0xDC2C);
	ChangeU2MB(0x008856, 0xDC2D);
	ChangeU2MB(0x008858, 0xDC2E);
	ChangeU2MB(0x00885A, 0xDC2F);
	ChangeU2MB(0x00885C, 0xDC30);
	ChangeU2MB(0x00885F, 0xDC31);
	ChangeU2MB(0x008860, 0xDC32);
	ChangeU2MB(0x008864, 0xDC33);
	ChangeU2MB(0x008869, 0xDC34);
	ChangeU2MB(0x008871, 0xDC35);
	ChangeU2MB(0x008879, 0xDC36);
	ChangeU2MB(0x00887B, 0xDC37);
	ChangeU2MB(0x008880, 0xDC38);
	ChangeU2MB(0x008898, 0xDC39);
	ChangeU2MB(0x00889A, 0xDC3A);
	ChangeU2MB(0x00889B, 0xDC3B);
	ChangeU2MB(0x00889C, 0xDC3C);
	ChangeU2MB(0x00889F, 0xDC3D);
	ChangeU2MB(0x0088A0, 0xDC3E);
	ChangeU2MB(0x0088A8, 0xDC3F);
	ChangeU2MB(0x0088AA, 0xDC40);
	ChangeU2MB(0x0088BA, 0xDC41);
	ChangeU2MB(0x0088BD, 0xDC42);
	ChangeU2MB(0x0088BE, 0xDC43);
	ChangeU2MB(0x0088C0, 0xDC44);
	ChangeU2MB(0x0088CA, 0xDC45);
	ChangeU2MB(0x0088CB, 0xDC46);
	ChangeU2MB(0x0088CC, 0xDC47);
	ChangeU2MB(0x0088CD, 0xDC48);
	ChangeU2MB(0x0088CE, 0xDC49);
	ChangeU2MB(0x0088D1, 0xDC4A);
	ChangeU2MB(0x0088D2, 0xDC4B);
	ChangeU2MB(0x0088D3, 0xDC4C);
	ChangeU2MB(0x0088DB, 0xDC4D);
	ChangeU2MB(0x0088DE, 0xDC4E);
	ChangeU2MB(0x0088E7, 0xDC4F);
	ChangeU2MB(0x0088EF, 0xDC50);
	ChangeU2MB(0x0088F0, 0xDC51);
	ChangeU2MB(0x0088F1, 0xDC52);
	ChangeU2MB(0x0088F7, 0xDC54);
	ChangeU2MB(0x008901, 0xDC55);
	ChangeU2MB(0x008906, 0xDC56);
	ChangeU2MB(0x00890D, 0xDC57);
	ChangeU2MB(0x00890E, 0xDC58);
	ChangeU2MB(0x00890F, 0xDC59);
	ChangeU2MB(0x008915, 0xDC5A);
	ChangeU2MB(0x008916, 0xDC5B);
	ChangeU2MB(0x008918, 0xDC5C);
	ChangeU2MB(0x008919, 0xDC5D);
	ChangeU2MB(0x00891A, 0xDC5E);
	ChangeU2MB(0x008920, 0xDC60);
	ChangeU2MB(0x008926, 0xDC61);
	ChangeU2MB(0x008927, 0xDC62);
	ChangeU2MB(0x008928, 0xDC63);
	ChangeU2MB(0x008930, 0xDC64);
	ChangeU2MB(0x008931, 0xDC65);
	ChangeU2MB(0x008932, 0xDC66);
	ChangeU2MB(0x008935, 0xDC67);
	ChangeU2MB(0x008939, 0xDC68);
	ChangeU2MB(0x00893A, 0xDC69);
	ChangeU2MB(0x00893E, 0xDC6A);
	ChangeU2MB(0x008940, 0xDC6B);
	ChangeU2MB(0x008942, 0xDC6C);
	ChangeU2MB(0x008945, 0xDC6D);
	ChangeU2MB(0x008946, 0xDC6E);
	ChangeU2MB(0x008949, 0xDC6F);
	ChangeU2MB(0x00894F, 0xDC70);
	ChangeU2MB(0x008952, 0xDC71);
	ChangeU2MB(0x008957, 0xDC72);
	ChangeU2MB(0x00895A, 0xDC73);
	ChangeU2MB(0x00895B, 0xDC74);
	ChangeU2MB(0x00895C, 0xDC75);
	ChangeU2MB(0x008961, 0xDC76);
	ChangeU2MB(0x008962, 0xDC77);
	ChangeU2MB(0x008963, 0xDC78);
	ChangeU2MB(0x00896B, 0xDC79);
	ChangeU2MB(0x00896E, 0xDC7A);
	ChangeU2MB(0x008970, 0xDC7B);
	ChangeU2MB(0x008973, 0xDC7C);
	ChangeU2MB(0x008975, 0xDC7D);
	ChangeU2MB(0x00897A, 0xDC7E);
	ChangeU2MB(0x00897B, 0xDD21);
	ChangeU2MB(0x00897C, 0xDD22);
	ChangeU2MB(0x00897D, 0xDD23);
	ChangeU2MB(0x008989, 0xDD24);
	ChangeU2MB(0x00898D, 0xDD25);
	ChangeU2MB(0x008990, 0xDD26);
	ChangeU2MB(0x008994, 0xDD27);
	ChangeU2MB(0x008995, 0xDD28);
	ChangeU2MB(0x00899B, 0xDD29);
	ChangeU2MB(0x00899C, 0xDD2A);
	ChangeU2MB(0x00899F, 0xDD2B);
	ChangeU2MB(0x0089A0, 0xDD2C);
	ChangeU2MB(0x0089A5, 0xDD2D);
	ChangeU2MB(0x0089B0, 0xDD2E);
	ChangeU2MB(0x0089B4, 0xDD2F);
	ChangeU2MB(0x0089B5, 0xDD30);
	ChangeU2MB(0x0089B6, 0xDD31);
	ChangeU2MB(0x0089B7, 0xDD32);
	ChangeU2MB(0x0089BC, 0xDD33);
	ChangeU2MB(0x0089D4, 0xDD34);
	ChangeU2MB(0x0089D5, 0xDD35);
	ChangeU2MB(0x0089D6, 0xDD36);
	ChangeU2MB(0x0089D7, 0xDD37);
	ChangeU2MB(0x0089D8, 0xDD38);
	ChangeU2MB(0x0089E5, 0xDD39);
	ChangeU2MB(0x0089E9, 0xDD3A);
	ChangeU2MB(0x0089EB, 0xDD3B);
	ChangeU2MB(0x0089ED, 0xDD3C);
	ChangeU2MB(0x0089F1, 0xDD3D);
	ChangeU2MB(0x0089F3, 0xDD3E);
	ChangeU2MB(0x0089F6, 0xDD3F);
	ChangeU2MB(0x0089F9, 0xDD40);
	ChangeU2MB(0x0089FD, 0xDD41);
	ChangeU2MB(0x0089FF, 0xDD42);
	ChangeU2MB(0x008A04, 0xDD43);
	ChangeU2MB(0x008A05, 0xDD44);
	ChangeU2MB(0x008A07, 0xDD45);
	ChangeU2MB(0x008A0F, 0xDD46);
	ChangeU2MB(0x008A11, 0xDD47);
	ChangeU2MB(0x008A14, 0xDD49);
	ChangeU2MB(0x008A15, 0xDD4A);
	ChangeU2MB(0x008A1E, 0xDD4B);
	ChangeU2MB(0x008A20, 0xDD4C);
	ChangeU2MB(0x008A22, 0xDD4D);
	ChangeU2MB(0x008A24, 0xDD4E);
	ChangeU2MB(0x008A26, 0xDD4F);
	ChangeU2MB(0x008A2B, 0xDD50);
	ChangeU2MB(0x008A2C, 0xDD51);
	ChangeU2MB(0x008A2F, 0xDD52);
	ChangeU2MB(0x008A35, 0xDD53);
	ChangeU2MB(0x008A3D, 0xDD55);
	ChangeU2MB(0x008A3E, 0xDD56);
	ChangeU2MB(0x008A40, 0xDD57);
	ChangeU2MB(0x008A43, 0xDD58);
	ChangeU2MB(0x008A45, 0xDD59);
	ChangeU2MB(0x008A47, 0xDD5A);
	ChangeU2MB(0x008A49, 0xDD5B);
	ChangeU2MB(0x008A4D, 0xDD5C);
	ChangeU2MB(0x008A4E, 0xDD5D);
	ChangeU2MB(0x008A53, 0xDD5E);
	ChangeU2MB(0x008A56, 0xDD5F);
	ChangeU2MB(0x008A57, 0xDD60);
	ChangeU2MB(0x008A58, 0xDD61);
	ChangeU2MB(0x008A5C, 0xDD62);
	ChangeU2MB(0x008A5D, 0xDD63);
	ChangeU2MB(0x008A61, 0xDD64);
	ChangeU2MB(0x008A65, 0xDD65);
	ChangeU2MB(0x008A67, 0xDD66);
	ChangeU2MB(0x008A75, 0xDD67);
	ChangeU2MB(0x008A76, 0xDD68);
	ChangeU2MB(0x008A77, 0xDD69);
	ChangeU2MB(0x008A7A, 0xDD6B);
	ChangeU2MB(0x008A7B, 0xDD6C);
	ChangeU2MB(0x008A7E, 0xDD6D);
	ChangeU2MB(0x008A7F, 0xDD6E);
	ChangeU2MB(0x008A80, 0xDD6F);
	ChangeU2MB(0x008A83, 0xDD70);
	ChangeU2MB(0x008A86, 0xDD71);
	ChangeU2MB(0x008A8B, 0xDD72);
	ChangeU2MB(0x008A8F, 0xDD73);
	ChangeU2MB(0x008A90, 0xDD74);
	ChangeU2MB(0x008A92, 0xDD75);
	ChangeU2MB(0x008A96, 0xDD76);
	ChangeU2MB(0x008A97, 0xDD77);
	ChangeU2MB(0x008A99, 0xDD78);
	ChangeU2MB(0x008A9F, 0xDD79);
	ChangeU2MB(0x008AA9, 0xDD7B);
	ChangeU2MB(0x008AAE, 0xDD7C);
	ChangeU2MB(0x008AAF, 0xDD7D);
	ChangeU2MB(0x008AB3, 0xDD7E);
	ChangeU2MB(0x008AB6, 0xDE21);
	ChangeU2MB(0x008AB7, 0xDE22);
	ChangeU2MB(0x008ABB, 0xDE23);
	ChangeU2MB(0x008AC3, 0xDE25);
	ChangeU2MB(0x008AC6, 0xDE26);
	ChangeU2MB(0x008AC8, 0xDE27);
	ChangeU2MB(0x008AC9, 0xDE28);
	ChangeU2MB(0x008ACA, 0xDE29);
	ChangeU2MB(0x008AD1, 0xDE2A);
	ChangeU2MB(0x008AD3, 0xDE2B);
	ChangeU2MB(0x008AD4, 0xDE2C);
	ChangeU2MB(0x008AD5, 0xDE2D);
	ChangeU2MB(0x008AD7, 0xDE2E);
	ChangeU2MB(0x008ADD, 0xDE2F);
	ChangeU2MB(0x008AEC, 0xDE31);
	ChangeU2MB(0x008AF0, 0xDE32);
	ChangeU2MB(0x008AF4, 0xDE33);
	ChangeU2MB(0x008AF5, 0xDE34);
	ChangeU2MB(0x008AFC, 0xDE36);
	ChangeU2MB(0x008AFF, 0xDE37);
	ChangeU2MB(0x008B05, 0xDE38);
	ChangeU2MB(0x008B06, 0xDE39);
	ChangeU2MB(0x008B0A, 0xDE3F);
	ChangeU2MB(0x008B0B, 0xDE3A);
	ChangeU2MB(0x008B11, 0xDE3B);
	ChangeU2MB(0x008B1C, 0xDE3C);
	ChangeU2MB(0x008B1E, 0xDE3D);
	ChangeU2MB(0x008B1F, 0xDE3E);
	ChangeU2MB(0x008B2D, 0xDE40);
	ChangeU2MB(0x008B30, 0xDE41);
	ChangeU2MB(0x008B37, 0xDE42);
	ChangeU2MB(0x008B3C, 0xDE43);
	ChangeU2MB(0x008B42, 0xDE44);
	ChangeU2MB(0x008B43, 0xDE45);
	ChangeU2MB(0x008B44, 0xDE46);
	ChangeU2MB(0x008B45, 0xDE47);
	ChangeU2MB(0x008B46, 0xDE48);
	ChangeU2MB(0x008B48, 0xDE49);
	ChangeU2MB(0x008B4D, 0xDE4E);
	ChangeU2MB(0x008B52, 0xDE4A);
	ChangeU2MB(0x008B54, 0xDE4C);
	ChangeU2MB(0x008B59, 0xDE4D);
	ChangeU2MB(0x008B5E, 0xDE4F);
	ChangeU2MB(0x008B63, 0xDE50);
	ChangeU2MB(0x008B6D, 0xDE51);
	ChangeU2MB(0x008B76, 0xDE52);
	ChangeU2MB(0x008B78, 0xDE53);
	ChangeU2MB(0x008B79, 0xDE54);
	ChangeU2MB(0x008B7C, 0xDE55);
	ChangeU2MB(0x008B7E, 0xDE56);
	ChangeU2MB(0x008B81, 0xDE57);
	ChangeU2MB(0x008B84, 0xDE58);
	ChangeU2MB(0x008B85, 0xDE59);
	ChangeU2MB(0x008B8B, 0xDE5A);
	ChangeU2MB(0x008B8D, 0xDE5B);
	ChangeU2MB(0x008B8F, 0xDE5C);
	ChangeU2MB(0x008B94, 0xDE5D);
	ChangeU2MB(0x008B95, 0xDE5E);
	ChangeU2MB(0x008B9C, 0xDE5F);
	ChangeU2MB(0x008B9E, 0xDE60);
	ChangeU2MB(0x008B9F, 0xDE61);
	ChangeU2MB(0x008C38, 0xDE62);
	ChangeU2MB(0x008C39, 0xDE63);
	ChangeU2MB(0x008C3D, 0xDE64);
	ChangeU2MB(0x008C3E, 0xDE65);
	ChangeU2MB(0x008C45, 0xDE66);
	ChangeU2MB(0x008C47, 0xDE67);
	ChangeU2MB(0x008C49, 0xDE68);
	ChangeU2MB(0x008C4B, 0xDE69);
	ChangeU2MB(0x008C4F, 0xDE6A);
	ChangeU2MB(0x008C51, 0xDE6B);
	ChangeU2MB(0x008C53, 0xDE6C);
	ChangeU2MB(0x008C54, 0xDE6D);
	ChangeU2MB(0x008C57, 0xDE6E);
	ChangeU2MB(0x008C58, 0xDE6F);
	ChangeU2MB(0x008C59, 0xDE72);
	ChangeU2MB(0x008C5B, 0xDE70);
	ChangeU2MB(0x008C5D, 0xDE71);
	ChangeU2MB(0x008C63, 0xDE73);
	ChangeU2MB(0x008C64, 0xDE74);
	ChangeU2MB(0x008C66, 0xDE75);
	ChangeU2MB(0x008C68, 0xDE76);
	ChangeU2MB(0x008C69, 0xDE77);
	ChangeU2MB(0x008C6D, 0xDE78);
	ChangeU2MB(0x008C73, 0xDE79);
	ChangeU2MB(0x008C75, 0xDE7A);
	ChangeU2MB(0x008C76, 0xDE7B);
	ChangeU2MB(0x008C7B, 0xDE7C);
	ChangeU2MB(0x008C7E, 0xDE7D);
	ChangeU2MB(0x008C86, 0xDE7E);
	ChangeU2MB(0x008C87, 0xDF21);
	ChangeU2MB(0x008C8B, 0xDF22);
	ChangeU2MB(0x008C90, 0xDF23);
	ChangeU2MB(0x008C92, 0xDF24);
	ChangeU2MB(0x008C93, 0xDF25);
	ChangeU2MB(0x008C99, 0xDF26);
	ChangeU2MB(0x008C9B, 0xDF27);
	ChangeU2MB(0x008C9C, 0xDF28);
	ChangeU2MB(0x008CA4, 0xDF29);
	ChangeU2MB(0x008CB9, 0xDF2A);
	ChangeU2MB(0x008CBA, 0xDF2B);
	ChangeU2MB(0x008CC5, 0xDF2C);
	ChangeU2MB(0x008CC6, 0xDF2D);
	ChangeU2MB(0x008CC9, 0xDF2E);
	ChangeU2MB(0x008CCB, 0xDF2F);
	ChangeU2MB(0x008CCF, 0xDF30);
	ChangeU2MB(0x008CD5, 0xDF32);
	ChangeU2MB(0x008CD6, 0xDF31);
	ChangeU2MB(0x008CD9, 0xDF33);
	ChangeU2MB(0x008CDD, 0xDF34);
	ChangeU2MB(0x008CE1, 0xDF35);
	ChangeU2MB(0x008CE8, 0xDF36);
	ChangeU2MB(0x008CEC, 0xDF37);
	ChangeU2MB(0x008CEF, 0xDF38);
	ChangeU2MB(0x008CF2, 0xDF3A);
	ChangeU2MB(0x008CF5, 0xDF3B);
	ChangeU2MB(0x008CF7, 0xDF3C);
	ChangeU2MB(0x008CF8, 0xDF3D);
	ChangeU2MB(0x008CFE, 0xDF3E);
	ChangeU2MB(0x008CFF, 0xDF3F);
	ChangeU2MB(0x008D01, 0xDF40);
	ChangeU2MB(0x008D03, 0xDF41);
	ChangeU2MB(0x008D09, 0xDF42);
	ChangeU2MB(0x008D17, 0xDF44);
	ChangeU2MB(0x008D1B, 0xDF45);
	ChangeU2MB(0x008D65, 0xDF46);
	ChangeU2MB(0x008D69, 0xDF47);
	ChangeU2MB(0x008D6C, 0xDF48);
	ChangeU2MB(0x008D6E, 0xDF49);
	ChangeU2MB(0x008D7F, 0xDF4A);
	ChangeU2MB(0x008D82, 0xDF4B);
	ChangeU2MB(0x008D84, 0xDF4C);
	ChangeU2MB(0x008D88, 0xDF4D);
	ChangeU2MB(0x008D8D, 0xDF4E);
	ChangeU2MB(0x008D90, 0xDF4F);
	ChangeU2MB(0x008D91, 0xDF50);
	ChangeU2MB(0x008D95, 0xDF51);
	ChangeU2MB(0x008D9E, 0xDF52);
	ChangeU2MB(0x008D9F, 0xDF53);
	ChangeU2MB(0x008DA0, 0xDF54);
	ChangeU2MB(0x008DA6, 0xDF55);
	ChangeU2MB(0x008DAB, 0xDF56);
	ChangeU2MB(0x008DAC, 0xDF57);
	ChangeU2MB(0x008DAF, 0xDF58);
	ChangeU2MB(0x008DB2, 0xDF59);
	ChangeU2MB(0x008DB5, 0xDF5A);
	ChangeU2MB(0x008DB7, 0xDF5B);
	ChangeU2MB(0x008DB9, 0xDF5C);
	ChangeU2MB(0x008DBB, 0xDF5D);
	ChangeU2MB(0x008DBC, 0xDF6F);
	ChangeU2MB(0x008DC0, 0xDF5E);
	ChangeU2MB(0x008DC5, 0xDF5F);
	ChangeU2MB(0x008DC6, 0xDF60);
	ChangeU2MB(0x008DC7, 0xDF61);
	ChangeU2MB(0x008DC8, 0xDF62);
	ChangeU2MB(0x008DCA, 0xDF63);
	ChangeU2MB(0x008DCE, 0xDF64);
	ChangeU2MB(0x008DD1, 0xDF65);
	ChangeU2MB(0x008DD4, 0xDF66);
	ChangeU2MB(0x008DD5, 0xDF67);
	ChangeU2MB(0x008DD7, 0xDF68);
	ChangeU2MB(0x008DD9, 0xDF69);
	ChangeU2MB(0x008DE4, 0xDF6A);
	ChangeU2MB(0x008DE5, 0xDF6B);
	ChangeU2MB(0x008DE7, 0xDF6C);
	ChangeU2MB(0x008DEC, 0xDF6D);
	ChangeU2MB(0x008DF0, 0xDF6E);
	ChangeU2MB(0x008DF1, 0xDF70);
	ChangeU2MB(0x008DF2, 0xDF71);
	ChangeU2MB(0x008DF4, 0xDF72);
	ChangeU2MB(0x008DFD, 0xDF73);
	ChangeU2MB(0x008E01, 0xDF74);
	ChangeU2MB(0x008E04, 0xDF75);
	ChangeU2MB(0x008E05, 0xDF76);
	ChangeU2MB(0x008E06, 0xDF77);
	ChangeU2MB(0x008E0B, 0xDF78);
	ChangeU2MB(0x008E11, 0xDF79);
	ChangeU2MB(0x008E14, 0xDF7A);
	ChangeU2MB(0x008E16, 0xDF7B);
	ChangeU2MB(0x008E20, 0xDF7C);
	ChangeU2MB(0x008E21, 0xDF7D);
	ChangeU2MB(0x008E22, 0xDF7E);
	ChangeU2MB(0x008E23, 0xE021);
	ChangeU2MB(0x008E26, 0xE022);
	ChangeU2MB(0x008E27, 0xE023);
	ChangeU2MB(0x008E31, 0xE024);
	ChangeU2MB(0x008E33, 0xE025);
	ChangeU2MB(0x008E36, 0xE026);
	ChangeU2MB(0x008E37, 0xE027);
	ChangeU2MB(0x008E38, 0xE028);
	ChangeU2MB(0x008E39, 0xE029);
	ChangeU2MB(0x008E3D, 0xE02A);
	ChangeU2MB(0x008E40, 0xE02B);
	ChangeU2MB(0x008E41, 0xE02C);
	ChangeU2MB(0x008E4B, 0xE02D);
	ChangeU2MB(0x008E4D, 0xE02E);
	ChangeU2MB(0x008E4E, 0xE02F);
	ChangeU2MB(0x008E4F, 0xE030);
	ChangeU2MB(0x008E54, 0xE031);
	ChangeU2MB(0x008E5B, 0xE032);
	ChangeU2MB(0x008E5C, 0xE033);
	ChangeU2MB(0x008E5D, 0xE034);
	ChangeU2MB(0x008E5E, 0xE035);
	ChangeU2MB(0x008E61, 0xE036);
	ChangeU2MB(0x008E62, 0xE037);
	ChangeU2MB(0x008E69, 0xE038);
	ChangeU2MB(0x008E6C, 0xE039);
	ChangeU2MB(0x008E6D, 0xE03A);
	ChangeU2MB(0x008E6F, 0xE03B);
	ChangeU2MB(0x008E70, 0xE03C);
	ChangeU2MB(0x008E71, 0xE03D);
	ChangeU2MB(0x008E79, 0xE03E);
	ChangeU2MB(0x008E7A, 0xE03F);
	ChangeU2MB(0x008E7B, 0xE040);
	ChangeU2MB(0x008E82, 0xE041);
	ChangeU2MB(0x008E83, 0xE042);
	ChangeU2MB(0x008E89, 0xE043);
	ChangeU2MB(0x008E90, 0xE044);
	ChangeU2MB(0x008E92, 0xE045);
	ChangeU2MB(0x008E95, 0xE046);
	ChangeU2MB(0x008E9A, 0xE047);
	ChangeU2MB(0x008E9B, 0xE048);
	ChangeU2MB(0x008E9D, 0xE049);
	ChangeU2MB(0x008E9E, 0xE04A);
	ChangeU2MB(0x008EA2, 0xE04B);
	ChangeU2MB(0x008EA7, 0xE04C);
	ChangeU2MB(0x008EA9, 0xE04D);
	ChangeU2MB(0x008EAD, 0xE04E);
	ChangeU2MB(0x008EAE, 0xE04F);
	ChangeU2MB(0x008EB3, 0xE050);
	ChangeU2MB(0x008EB5, 0xE051);
	ChangeU2MB(0x008EBA, 0xE052);
	ChangeU2MB(0x008EBB, 0xE053);
	ChangeU2MB(0x008EC0, 0xE054);
	ChangeU2MB(0x008EC1, 0xE055);
	ChangeU2MB(0x008EC3, 0xE056);
	ChangeU2MB(0x008EC4, 0xE057);
	ChangeU2MB(0x008EC7, 0xE058);
	ChangeU2MB(0x008ED1, 0xE05A);
	ChangeU2MB(0x008ED4, 0xE05B);
	ChangeU2MB(0x008EDC, 0xE05C);
	ChangeU2MB(0x008EE8, 0xE05D);
	ChangeU2MB(0x008EED, 0xE064);
	ChangeU2MB(0x008EEE, 0xE05E);
	ChangeU2MB(0x008EF0, 0xE05F);
	ChangeU2MB(0x008EF1, 0xE060);
	ChangeU2MB(0x008EF7, 0xE061);
	ChangeU2MB(0x008EF9, 0xE062);
	ChangeU2MB(0x008EFA, 0xE063);
	ChangeU2MB(0x008F00, 0xE065);
	ChangeU2MB(0x008F02, 0xE066);
	ChangeU2MB(0x008F07, 0xE067);
	ChangeU2MB(0x008F08, 0xE068);
	ChangeU2MB(0x008F0F, 0xE069);
	ChangeU2MB(0x008F10, 0xE06A);
	ChangeU2MB(0x008F16, 0xE06B);
	ChangeU2MB(0x008F17, 0xE06C);
	ChangeU2MB(0x008F18, 0xE06D);
	ChangeU2MB(0x008F1E, 0xE06E);
	ChangeU2MB(0x008F20, 0xE06F);
	ChangeU2MB(0x008F21, 0xE070);
	ChangeU2MB(0x008F23, 0xE071);
	ChangeU2MB(0x008F25, 0xE072);
	ChangeU2MB(0x008F27, 0xE073);
	ChangeU2MB(0x008F28, 0xE074);
	ChangeU2MB(0x008F2C, 0xE075);
	ChangeU2MB(0x008F2D, 0xE076);
	ChangeU2MB(0x008F2E, 0xE077);
	ChangeU2MB(0x008F34, 0xE078);
	ChangeU2MB(0x008F35, 0xE079);
	ChangeU2MB(0x008F36, 0xE07A);
	ChangeU2MB(0x008F37, 0xE07B);
	ChangeU2MB(0x008F3A, 0xE07C);
	ChangeU2MB(0x008F40, 0xE07D);
	ChangeU2MB(0x008F41, 0xE07E);
	ChangeU2MB(0x008F43, 0xE121);
	ChangeU2MB(0x008F47, 0xE122);
	ChangeU2MB(0x008F4F, 0xE123);
	ChangeU2MB(0x008F51, 0xE124);
	ChangeU2MB(0x008F52, 0xE125);
	ChangeU2MB(0x008F53, 0xE126);
	ChangeU2MB(0x008F54, 0xE127);
	ChangeU2MB(0x008F55, 0xE128);
	ChangeU2MB(0x008F58, 0xE129);
	ChangeU2MB(0x008F5D, 0xE12A);
	ChangeU2MB(0x008F5E, 0xE12B);
	ChangeU2MB(0x008F65, 0xE12C);
	ChangeU2MB(0x008F9D, 0xE12D);
	ChangeU2MB(0x008FA0, 0xE12E);
	ChangeU2MB(0x008FA1, 0xE12F);
	ChangeU2MB(0x008FA4, 0xE130);
	ChangeU2MB(0x008FA5, 0xE131);
	ChangeU2MB(0x008FA6, 0xE132);
	ChangeU2MB(0x008FB5, 0xE133);
	ChangeU2MB(0x008FB6, 0xE134);
	ChangeU2MB(0x008FB8, 0xE135);
	ChangeU2MB(0x008FBE, 0xE136);
	ChangeU2MB(0x008FC0, 0xE137);
	ChangeU2MB(0x008FC1, 0xE138);
	ChangeU2MB(0x008FC6, 0xE139);
	ChangeU2MB(0x008FCA, 0xE13A);
	ChangeU2MB(0x008FCB, 0xE13B);
	ChangeU2MB(0x008FCD, 0xE13C);
	ChangeU2MB(0x008FD0, 0xE13D);
	ChangeU2MB(0x008FD2, 0xE13E);
	ChangeU2MB(0x008FD3, 0xE13F);
	ChangeU2MB(0x008FD5, 0xE140);
	ChangeU2MB(0x008FE0, 0xE141);
	ChangeU2MB(0x008FE3, 0xE142);
	ChangeU2MB(0x008FE4, 0xE143);
	ChangeU2MB(0x008FE8, 0xE144);
	ChangeU2MB(0x008FEE, 0xE145);
	ChangeU2MB(0x008FF1, 0xE146);
	ChangeU2MB(0x008FF5, 0xE147);
	ChangeU2MB(0x008FF6, 0xE148);
	ChangeU2MB(0x008FFB, 0xE149);
	ChangeU2MB(0x008FFE, 0xE14A);
	ChangeU2MB(0x009002, 0xE14B);
	ChangeU2MB(0x009004, 0xE14C);
	ChangeU2MB(0x009008, 0xE14D);
	ChangeU2MB(0x00900C, 0xE14E);
	ChangeU2MB(0x009018, 0xE14F);
	ChangeU2MB(0x00901B, 0xE150);
	ChangeU2MB(0x009028, 0xE151);
	ChangeU2MB(0x009029, 0xE152);
	ChangeU2MB(0x00902A, 0xE154);
	ChangeU2MB(0x00902C, 0xE155);
	ChangeU2MB(0x00902D, 0xE156);
	ChangeU2MB(0x00902F, 0xE153);
	ChangeU2MB(0x009033, 0xE157);
	ChangeU2MB(0x009034, 0xE158);
	ChangeU2MB(0x009037, 0xE159);
	ChangeU2MB(0x00903F, 0xE15A);
	ChangeU2MB(0x009043, 0xE15B);
	ChangeU2MB(0x009044, 0xE15C);
	ChangeU2MB(0x00904C, 0xE15D);
	ChangeU2MB(0x00905B, 0xE15E);
	ChangeU2MB(0x00905D, 0xE15F);
	ChangeU2MB(0x009062, 0xE160);
	ChangeU2MB(0x009066, 0xE161);
	ChangeU2MB(0x00906C, 0xE163);
	ChangeU2MB(0x009070, 0xE164);
	ChangeU2MB(0x009074, 0xE165);
	ChangeU2MB(0x009079, 0xE166);
	ChangeU2MB(0x009085, 0xE167);
	ChangeU2MB(0x009088, 0xE168);
	ChangeU2MB(0x00908B, 0xE169);
	ChangeU2MB(0x00908C, 0xE16A);
	ChangeU2MB(0x00908E, 0xE16B);
	ChangeU2MB(0x009090, 0xE16C);
	ChangeU2MB(0x009095, 0xE16D);
	ChangeU2MB(0x009097, 0xE16E);
	ChangeU2MB(0x009098, 0xE16F);
	ChangeU2MB(0x009099, 0xE170);
	ChangeU2MB(0x00909B, 0xE171);
	ChangeU2MB(0x0090A0, 0xE172);
	ChangeU2MB(0x0090A1, 0xE173);
	ChangeU2MB(0x0090A2, 0xE174);
	ChangeU2MB(0x0090A5, 0xE175);
	ChangeU2MB(0x0090B0, 0xE176);
	ChangeU2MB(0x0090B2, 0xE177);
	ChangeU2MB(0x0090B3, 0xE178);
	ChangeU2MB(0x0090B4, 0xE179);
	ChangeU2MB(0x0090B6, 0xE17A);
	ChangeU2MB(0x0090BD, 0xE17B);
	ChangeU2MB(0x0090BE, 0xE17D);
	ChangeU2MB(0x0090C3, 0xE17E);
	ChangeU2MB(0x0090C4, 0xE221);
	ChangeU2MB(0x0090C5, 0xE222);
	ChangeU2MB(0x0090C7, 0xE223);
	ChangeU2MB(0x0090C8, 0xE224);
	ChangeU2MB(0x0090CC, 0xE17C);
	ChangeU2MB(0x0090D2, 0xE22D);
	ChangeU2MB(0x0090D5, 0xE225);
	ChangeU2MB(0x0090D7, 0xE226);
	ChangeU2MB(0x0090D8, 0xE227);
	ChangeU2MB(0x0090D9, 0xE228);
	ChangeU2MB(0x0090DC, 0xE229);
	ChangeU2MB(0x0090DD, 0xE22A);
	ChangeU2MB(0x0090DF, 0xE22B);
	ChangeU2MB(0x0090E5, 0xE22C);
	ChangeU2MB(0x0090EB, 0xE22F);
	ChangeU2MB(0x0090EF, 0xE230);
	ChangeU2MB(0x0090F0, 0xE231);
	ChangeU2MB(0x0090F4, 0xE232);
	ChangeU2MB(0x0090F6, 0xE22E);
	ChangeU2MB(0x0090FE, 0xE233);
	ChangeU2MB(0x0090FF, 0xE234);
	ChangeU2MB(0x009100, 0xE235);
	ChangeU2MB(0x009104, 0xE236);
	ChangeU2MB(0x009105, 0xE237);
	ChangeU2MB(0x009106, 0xE238);
	ChangeU2MB(0x009108, 0xE239);
	ChangeU2MB(0x00910D, 0xE23A);
	ChangeU2MB(0x009110, 0xE23B);
	ChangeU2MB(0x009114, 0xE23C);
	ChangeU2MB(0x009116, 0xE23D);
	ChangeU2MB(0x009117, 0xE23E);
	ChangeU2MB(0x009118, 0xE23F);
	ChangeU2MB(0x00911A, 0xE240);
	ChangeU2MB(0x00911C, 0xE241);
	ChangeU2MB(0x00911E, 0xE242);
	ChangeU2MB(0x009120, 0xE243);
	ChangeU2MB(0x009122, 0xE245);
	ChangeU2MB(0x009123, 0xE246);
	ChangeU2MB(0x009125, 0xE244);
	ChangeU2MB(0x009129, 0xE248);
	ChangeU2MB(0x00912E, 0xE249);
	ChangeU2MB(0x00912F, 0xE24A);
	ChangeU2MB(0x009131, 0xE24B);
	ChangeU2MB(0x009134, 0xE24C);
	ChangeU2MB(0x009136, 0xE24D);
	ChangeU2MB(0x009137, 0xE24E);
	ChangeU2MB(0x009139, 0xE24F);
	ChangeU2MB(0x00913A, 0xE250);
	ChangeU2MB(0x00913C, 0xE251);
	ChangeU2MB(0x00913D, 0xE252);
	ChangeU2MB(0x009143, 0xE253);
	ChangeU2MB(0x009147, 0xE254);
	ChangeU2MB(0x009148, 0xE255);
	ChangeU2MB(0x00914F, 0xE256);
	ChangeU2MB(0x009153, 0xE257);
	ChangeU2MB(0x009157, 0xE258);
	ChangeU2MB(0x009159, 0xE259);
	ChangeU2MB(0x00915A, 0xE25A);
	ChangeU2MB(0x00915B, 0xE25B);
	ChangeU2MB(0x009161, 0xE25C);
	ChangeU2MB(0x009164, 0xE25D);
	ChangeU2MB(0x009167, 0xE25E);
	ChangeU2MB(0x00916D, 0xE25F);
	ChangeU2MB(0x009174, 0xE260);
	ChangeU2MB(0x009179, 0xE261);
	ChangeU2MB(0x00917A, 0xE262);
	ChangeU2MB(0x00917B, 0xE263);
	ChangeU2MB(0x009181, 0xE264);
	ChangeU2MB(0x009183, 0xE265);
	ChangeU2MB(0x009185, 0xE266);
	ChangeU2MB(0x009186, 0xE267);
	ChangeU2MB(0x00918A, 0xE268);
	ChangeU2MB(0x00918E, 0xE269);
	ChangeU2MB(0x009191, 0xE26A);
	ChangeU2MB(0x009193, 0xE26B);
	ChangeU2MB(0x009194, 0xE26C);
	ChangeU2MB(0x009195, 0xE26D);
	ChangeU2MB(0x009198, 0xE26E);
	ChangeU2MB(0x00919E, 0xE26F);
	ChangeU2MB(0x0091A1, 0xE270);
	ChangeU2MB(0x0091A6, 0xE271);
	ChangeU2MB(0x0091A8, 0xE272);
	ChangeU2MB(0x0091AC, 0xE273);
	ChangeU2MB(0x0091AD, 0xE274);
	ChangeU2MB(0x0091AE, 0xE275);
	ChangeU2MB(0x0091B0, 0xE276);
	ChangeU2MB(0x0091B1, 0xE277);
	ChangeU2MB(0x0091B2, 0xE278);
	ChangeU2MB(0x0091B3, 0xE279);
	ChangeU2MB(0x0091B6, 0xE27A);
	ChangeU2MB(0x0091BB, 0xE27B);
	ChangeU2MB(0x0091BC, 0xE27C);
	ChangeU2MB(0x0091BD, 0xE27D);
	ChangeU2MB(0x0091BF, 0xE27E);
	ChangeU2MB(0x0091C2, 0xE321);
	ChangeU2MB(0x0091C3, 0xE322);
	ChangeU2MB(0x0091C5, 0xE323);
	ChangeU2MB(0x0091D3, 0xE324);
	ChangeU2MB(0x0091D4, 0xE325);
	ChangeU2MB(0x0091D9, 0xE327);
	ChangeU2MB(0x0091E9, 0xE32C);
	ChangeU2MB(0x0091EA, 0xE32D);
	ChangeU2MB(0x0091EC, 0xE32E);
	ChangeU2MB(0x0091EF, 0xE331);
	ChangeU2MB(0x0091F0, 0xE332);
	ChangeU2MB(0x0091F1, 0xE333);
	ChangeU2MB(0x0091F7, 0xE334);
	ChangeU2MB(0x0091F9, 0xE335);
	ChangeU2MB(0x0091FB, 0xE336);
	ChangeU2MB(0x0091FD, 0xE337);
	ChangeU2MB(0x009200, 0xE338);
	ChangeU2MB(0x009201, 0xE339);
	ChangeU2MB(0x009204, 0xE33A);
	ChangeU2MB(0x009205, 0xE33B);
	ChangeU2MB(0x009207, 0xE33D);
	ChangeU2MB(0x009209, 0xE33E);
	ChangeU2MB(0x00920C, 0xE340);
	ChangeU2MB(0x009212, 0xE342);
	ChangeU2MB(0x009213, 0xE343);
	ChangeU2MB(0x009216, 0xE344);
	ChangeU2MB(0x009218, 0xE345);
	ChangeU2MB(0x00921C, 0xE346);
	ChangeU2MB(0x00921D, 0xE347);
	ChangeU2MB(0x009223, 0xE348);
	ChangeU2MB(0x009224, 0xE349);
	ChangeU2MB(0x009225, 0xE34A);
	ChangeU2MB(0x009226, 0xE34B);
	ChangeU2MB(0x009228, 0xE34C);
	ChangeU2MB(0x00922E, 0xE34D);
	ChangeU2MB(0x00922F, 0xE34E);
	ChangeU2MB(0x009230, 0xE34F);
	ChangeU2MB(0x009233, 0xE350);
	ChangeU2MB(0x009235, 0xE351);
	ChangeU2MB(0x009236, 0xE352);
	ChangeU2MB(0x009238, 0xE353);
	ChangeU2MB(0x00923E, 0xE357);
	ChangeU2MB(0x009242, 0xE359);
	ChangeU2MB(0x009243, 0xE35A);
	ChangeU2MB(0x009246, 0xE35B);
	ChangeU2MB(0x009247, 0xE35C);
	ChangeU2MB(0x00924A, 0xE35D);
	ChangeU2MB(0x00924D, 0xE35E);
	ChangeU2MB(0x00924F, 0xE360);
	ChangeU2MB(0x009258, 0xE362);
	ChangeU2MB(0x00925C, 0xE364);
	ChangeU2MB(0x00925D, 0xE365);
	ChangeU2MB(0x009260, 0xE366);
	ChangeU2MB(0x009261, 0xE367);
	ChangeU2MB(0x009265, 0xE368);
	ChangeU2MB(0x009268, 0xE36A);
	ChangeU2MB(0x009269, 0xE36B);
	ChangeU2MB(0x00926E, 0xE36C);
	ChangeU2MB(0x00926F, 0xE36D);
	ChangeU2MB(0x009270, 0xE36E);
	ChangeU2MB(0x009275, 0xE36F);
	ChangeU2MB(0x009276, 0xE370);
	ChangeU2MB(0x009279, 0xE373);
	ChangeU2MB(0x00927B, 0xE374);
	ChangeU2MB(0x00927C, 0xE375);
	ChangeU2MB(0x00927D, 0xE376);
	ChangeU2MB(0x00927F, 0xE377);
	ChangeU2MB(0x009289, 0xE379);
	ChangeU2MB(0x00928A, 0xE37A);
	ChangeU2MB(0x00928D, 0xE37B);
	ChangeU2MB(0x00928E, 0xE37C);
	ChangeU2MB(0x009292, 0xE37D);
	ChangeU2MB(0x009297, 0xE37E);
	ChangeU2MB(0x009299, 0xE421);
	ChangeU2MB(0x00929F, 0xE422);
	ChangeU2MB(0x0092A0, 0xE423);
	ChangeU2MB(0x0092A4, 0xE424);
	ChangeU2MB(0x0092A5, 0xE425);
	ChangeU2MB(0x0092A8, 0xE427);
	ChangeU2MB(0x0092AB, 0xE428);
	ChangeU2MB(0x0092AF, 0xE429);
	ChangeU2MB(0x0092B2, 0xE42A);
	ChangeU2MB(0x0092B6, 0xE42B);
	ChangeU2MB(0x0092B8, 0xE42C);
	ChangeU2MB(0x0092BA, 0xE42D);
	ChangeU2MB(0x0092BB, 0xE42E);
	ChangeU2MB(0x0092BC, 0xE42F);
	ChangeU2MB(0x0092BD, 0xE430);
	ChangeU2MB(0x0092BF, 0xE431);
	ChangeU2MB(0x0092C0, 0xE432);
	ChangeU2MB(0x0092C1, 0xE433);
	ChangeU2MB(0x0092C2, 0xE434);
	ChangeU2MB(0x0092C3, 0xE435);
	ChangeU2MB(0x0092C5, 0xE436);
	ChangeU2MB(0x0092C6, 0xE437);
	ChangeU2MB(0x0092C7, 0xE438);
	ChangeU2MB(0x0092C8, 0xE439);
	ChangeU2MB(0x0092CB, 0xE43A);
	ChangeU2MB(0x0092CC, 0xE43B);
	ChangeU2MB(0x0092CD, 0xE43C);
	ChangeU2MB(0x0092CE, 0xE43D);
	ChangeU2MB(0x0092D8, 0xE442);
	ChangeU2MB(0x0092DC, 0xE444);
	ChangeU2MB(0x0092DD, 0xE445);
	ChangeU2MB(0x0092DF, 0xE446);
	ChangeU2MB(0x0092E1, 0xE448);
	ChangeU2MB(0x0092E3, 0xE449);
	ChangeU2MB(0x0092E5, 0xE44A);
	ChangeU2MB(0x0092E8, 0xE44C);
	ChangeU2MB(0x0092EC, 0xE44D);
	ChangeU2MB(0x0092EE, 0xE44E);
	ChangeU2MB(0x0092F0, 0xE44F);
	ChangeU2MB(0x009300, 0xE453);
	ChangeU2MB(0x009308, 0xE455);
	ChangeU2MB(0x00930D, 0xE456);
	ChangeU2MB(0x009311, 0xE457);
	ChangeU2MB(0x009314, 0xE458);
	ChangeU2MB(0x009315, 0xE459);
	ChangeU2MB(0x00931C, 0xE45A);
	ChangeU2MB(0x00931F, 0xE45D);
	ChangeU2MB(0x009324, 0xE45F);
	ChangeU2MB(0x009327, 0xE461);
	ChangeU2MB(0x009329, 0xE462);
	ChangeU2MB(0x00932A, 0xE463);
	ChangeU2MB(0x009333, 0xE464);
	ChangeU2MB(0x009334, 0xE465);
	ChangeU2MB(0x009336, 0xE466);
	ChangeU2MB(0x009337, 0xE467);
	ChangeU2MB(0x009347, 0xE468);
	ChangeU2MB(0x009349, 0xE46A);
	ChangeU2MB(0x009350, 0xE46B);
	ChangeU2MB(0x009351, 0xE46C);
	ChangeU2MB(0x009352, 0xE46D);
	ChangeU2MB(0x009355, 0xE46E);
	ChangeU2MB(0x009358, 0xE470);
	ChangeU2MB(0x00935A, 0xE471);
	ChangeU2MB(0x00935E, 0xE472);
	ChangeU2MB(0x009364, 0xE473);
	ChangeU2MB(0x009365, 0xE474);
	ChangeU2MB(0x009367, 0xE475);
	ChangeU2MB(0x009369, 0xE476);
	ChangeU2MB(0x00936A, 0xE477);
	ChangeU2MB(0x00936D, 0xE478);
	ChangeU2MB(0x00936F, 0xE479);
	ChangeU2MB(0x009371, 0xE47B);
	ChangeU2MB(0x009373, 0xE47C);
	ChangeU2MB(0x009374, 0xE47D);
	ChangeU2MB(0x009376, 0xE47E);
	ChangeU2MB(0x00937A, 0xE521);
	ChangeU2MB(0x00937D, 0xE522);
	ChangeU2MB(0x00937F, 0xE523);
	ChangeU2MB(0x009380, 0xE524);
	ChangeU2MB(0x009381, 0xE525);
	ChangeU2MB(0x009382, 0xE526);
	ChangeU2MB(0x009388, 0xE527);
	ChangeU2MB(0x00938A, 0xE528);
	ChangeU2MB(0x00938B, 0xE529);
	ChangeU2MB(0x00938D, 0xE52A);
	ChangeU2MB(0x00938F, 0xE52B);
	ChangeU2MB(0x009392, 0xE52C);
	ChangeU2MB(0x009395, 0xE52D);
	ChangeU2MB(0x009398, 0xE52E);
	ChangeU2MB(0x00939B, 0xE52F);
	ChangeU2MB(0x00939E, 0xE530);
	ChangeU2MB(0x0093A1, 0xE531);
	ChangeU2MB(0x0093A3, 0xE532);
	ChangeU2MB(0x0093A6, 0xE534);
	ChangeU2MB(0x0093A8, 0xE535);
	ChangeU2MB(0x0093A9, 0xE53B);
	ChangeU2MB(0x0093AB, 0xE536);
	ChangeU2MB(0x0093B4, 0xE537);
	ChangeU2MB(0x0093B5, 0xE538);
	ChangeU2MB(0x0093B6, 0xE539);
	ChangeU2MB(0x0093BA, 0xE53A);
	ChangeU2MB(0x0093C1, 0xE53C);
	ChangeU2MB(0x0093C4, 0xE53D);
	ChangeU2MB(0x0093C5, 0xE53E);
	ChangeU2MB(0x0093C7, 0xE540);
	ChangeU2MB(0x0093C9, 0xE541);
	ChangeU2MB(0x0093CA, 0xE542);
	ChangeU2MB(0x0093CB, 0xE543);
	ChangeU2MB(0x0093CC, 0xE544);
	ChangeU2MB(0x0093CD, 0xE545);
	ChangeU2MB(0x0093D3, 0xE546);
	ChangeU2MB(0x0093D9, 0xE547);
	ChangeU2MB(0x0093DC, 0xE548);
	ChangeU2MB(0x0093DF, 0xE54A);
	ChangeU2MB(0x0093E2, 0xE54B);
	ChangeU2MB(0x0093E6, 0xE54C);
	ChangeU2MB(0x0093E7, 0xE54D);
	ChangeU2MB(0x0093F7, 0xE54F);
	ChangeU2MB(0x0093F9, 0xE54E);
	ChangeU2MB(0x0093FA, 0xE551);
	ChangeU2MB(0x0093FB, 0xE552);
	ChangeU2MB(0x0093FD, 0xE553);
	ChangeU2MB(0x009401, 0xE554);
	ChangeU2MB(0x009402, 0xE555);
	ChangeU2MB(0x009404, 0xE556);
	ChangeU2MB(0x009408, 0xE557);
	ChangeU2MB(0x009409, 0xE558);
	ChangeU2MB(0x00940D, 0xE559);
	ChangeU2MB(0x00940E, 0xE55A);
	ChangeU2MB(0x00940F, 0xE55B);
	ChangeU2MB(0x009415, 0xE55C);
	ChangeU2MB(0x009416, 0xE55D);
	ChangeU2MB(0x009417, 0xE55E);
	ChangeU2MB(0x00941F, 0xE55F);
	ChangeU2MB(0x00942E, 0xE560);
	ChangeU2MB(0x00942F, 0xE561);
	ChangeU2MB(0x009432, 0xE563);
	ChangeU2MB(0x009433, 0xE564);
	ChangeU2MB(0x009434, 0xE565);
	ChangeU2MB(0x00943B, 0xE566);
	ChangeU2MB(0x00943D, 0xE568);
	ChangeU2MB(0x00943F, 0xE567);
	ChangeU2MB(0x009443, 0xE569);
	ChangeU2MB(0x00944A, 0xE56C);
	ChangeU2MB(0x00944C, 0xE56D);
	ChangeU2MB(0x009455, 0xE56E);
	ChangeU2MB(0x009459, 0xE56F);
	ChangeU2MB(0x00945C, 0xE570);
	ChangeU2MB(0x00945F, 0xE571);
	ChangeU2MB(0x009461, 0xE572);
	ChangeU2MB(0x009463, 0xE573);
	ChangeU2MB(0x009468, 0xE574);
	ChangeU2MB(0x00946B, 0xE575);
	ChangeU2MB(0x00946D, 0xE576);
	ChangeU2MB(0x00946E, 0xE577);
	ChangeU2MB(0x00946F, 0xE578);
	ChangeU2MB(0x009471, 0xE579);
	ChangeU2MB(0x009472, 0xE57A);
	ChangeU2MB(0x009483, 0xE57C);
	ChangeU2MB(0x009484, 0xE57B);
	ChangeU2MB(0x009578, 0xE57D);
	ChangeU2MB(0x009579, 0xE57E);
	ChangeU2MB(0x00957E, 0xE621);
	ChangeU2MB(0x009584, 0xE622);
	ChangeU2MB(0x009588, 0xE623);
	ChangeU2MB(0x00958C, 0xE624);
	ChangeU2MB(0x00958D, 0xE625);
	ChangeU2MB(0x00958E, 0xE626);
	ChangeU2MB(0x00959D, 0xE627);
	ChangeU2MB(0x00959E, 0xE628);
	ChangeU2MB(0x00959F, 0xE629);
	ChangeU2MB(0x0095A1, 0xE62A);
	ChangeU2MB(0x0095A6, 0xE62B);
	ChangeU2MB(0x0095A9, 0xE62C);
	ChangeU2MB(0x0095AB, 0xE62D);
	ChangeU2MB(0x0095AC, 0xE62E);
	ChangeU2MB(0x0095B4, 0xE62F);
	ChangeU2MB(0x0095B6, 0xE630);
	ChangeU2MB(0x0095BA, 0xE631);
	ChangeU2MB(0x0095BD, 0xE632);
	ChangeU2MB(0x0095BF, 0xE633);
	ChangeU2MB(0x0095C6, 0xE634);
	ChangeU2MB(0x0095C8, 0xE635);
	ChangeU2MB(0x0095C9, 0xE636);
	ChangeU2MB(0x0095CB, 0xE637);
	ChangeU2MB(0x0095D0, 0xE638);
	ChangeU2MB(0x0095D1, 0xE639);
	ChangeU2MB(0x0095D2, 0xE63A);
	ChangeU2MB(0x0095D3, 0xE63B);
	ChangeU2MB(0x0095D9, 0xE63C);
	ChangeU2MB(0x0095DA, 0xE63D);
	ChangeU2MB(0x0095DD, 0xE63E);
	ChangeU2MB(0x0095DE, 0xE63F);
	ChangeU2MB(0x0095DF, 0xE640);
	ChangeU2MB(0x0095E0, 0xE641);
	ChangeU2MB(0x0095E4, 0xE642);
	ChangeU2MB(0x0095E6, 0xE643);
	ChangeU2MB(0x00961D, 0xE644);
	ChangeU2MB(0x00961E, 0xE645);
	ChangeU2MB(0x009622, 0xE646);
	ChangeU2MB(0x009624, 0xE647);
	ChangeU2MB(0x009625, 0xE648);
	ChangeU2MB(0x009626, 0xE649);
	ChangeU2MB(0x00962C, 0xE64A);
	ChangeU2MB(0x009631, 0xE64B);
	ChangeU2MB(0x009633, 0xE64C);
	ChangeU2MB(0x009637, 0xE64D);
	ChangeU2MB(0x009638, 0xE64E);
	ChangeU2MB(0x009639, 0xE64F);
	ChangeU2MB(0x00963A, 0xE650);
	ChangeU2MB(0x00963C, 0xE651);
	ChangeU2MB(0x00963D, 0xE652);
	ChangeU2MB(0x009641, 0xE653);
	ChangeU2MB(0x009652, 0xE654);
	ChangeU2MB(0x009654, 0xE655);
	ChangeU2MB(0x009656, 0xE656);
	ChangeU2MB(0x009657, 0xE657);
	ChangeU2MB(0x009658, 0xE658);
	ChangeU2MB(0x009661, 0xE659);
	ChangeU2MB(0x00966E, 0xE65A);
	ChangeU2MB(0x009674, 0xE65B);
	ChangeU2MB(0x00967B, 0xE65C);
	ChangeU2MB(0x00967C, 0xE65D);
	ChangeU2MB(0x00967E, 0xE65E);
	ChangeU2MB(0x00967F, 0xE65F);
	ChangeU2MB(0x009681, 0xE660);
	ChangeU2MB(0x009682, 0xE661);
	ChangeU2MB(0x009683, 0xE662);
	ChangeU2MB(0x009684, 0xE663);
	ChangeU2MB(0x009689, 0xE664);
	ChangeU2MB(0x009691, 0xE665);
	ChangeU2MB(0x009696, 0xE666);
	ChangeU2MB(0x00969A, 0xE667);
	ChangeU2MB(0x00969F, 0xE669);
	ChangeU2MB(0x0096A4, 0xE66A);
	ChangeU2MB(0x0096A5, 0xE66B);
	ChangeU2MB(0x0096A6, 0xE66C);
	ChangeU2MB(0x0096A9, 0xE66D);
	ChangeU2MB(0x0096AE, 0xE66E);
	ChangeU2MB(0x0096B3, 0xE670);
	ChangeU2MB(0x0096BA, 0xE671);
	ChangeU2MB(0x0096CA, 0xE672);
	ChangeU2MB(0x0096D2, 0xE673);
	ChangeU2MB(0x0096D8, 0xE675);
	ChangeU2MB(0x0096DA, 0xE676);
	ChangeU2MB(0x0096DD, 0xE677);
	ChangeU2MB(0x0096DE, 0xE678);
	ChangeU2MB(0x0096DF, 0xE679);
	ChangeU2MB(0x0096E9, 0xE67A);
	ChangeU2MB(0x0096EF, 0xE67B);
	ChangeU2MB(0x0096F1, 0xE67C);
	ChangeU2MB(0x0096FA, 0xE67D);
	ChangeU2MB(0x009702, 0xE67E);
	ChangeU2MB(0x009703, 0xE721);
	ChangeU2MB(0x009705, 0xE722);
	ChangeU2MB(0x009709, 0xE723);
	ChangeU2MB(0x00971A, 0xE724);
	ChangeU2MB(0x00971B, 0xE725);
	ChangeU2MB(0x00971D, 0xE726);
	ChangeU2MB(0x009721, 0xE727);
	ChangeU2MB(0x009722, 0xE728);
	ChangeU2MB(0x009723, 0xE729);
	ChangeU2MB(0x009728, 0xE72A);
	ChangeU2MB(0x009731, 0xE72B);
	ChangeU2MB(0x009741, 0xE72D);
	ChangeU2MB(0x00974A, 0xE72F);
	ChangeU2MB(0x00974E, 0xE730);
	ChangeU2MB(0x009757, 0xE733);
	ChangeU2MB(0x009758, 0xE734);
	ChangeU2MB(0x00975A, 0xE735);
	ChangeU2MB(0x00975B, 0xE736);
	ChangeU2MB(0x009763, 0xE737);
	ChangeU2MB(0x009767, 0xE738);
	ChangeU2MB(0x00976A, 0xE739);
	ChangeU2MB(0x00976E, 0xE73A);
	ChangeU2MB(0x009773, 0xE73B);
	ChangeU2MB(0x009776, 0xE73C);
	ChangeU2MB(0x009777, 0xE73D);
	ChangeU2MB(0x009778, 0xE73E);
	ChangeU2MB(0x00977B, 0xE73F);
	ChangeU2MB(0x00977D, 0xE740);
	ChangeU2MB(0x00977F, 0xE741);
	ChangeU2MB(0x009780, 0xE742);
	ChangeU2MB(0x009789, 0xE743);
	ChangeU2MB(0x009795, 0xE744);
	ChangeU2MB(0x009796, 0xE745);
	ChangeU2MB(0x009797, 0xE746);
	ChangeU2MB(0x009799, 0xE747);
	ChangeU2MB(0x00979A, 0xE748);
	ChangeU2MB(0x00979E, 0xE749);
	ChangeU2MB(0x00979F, 0xE74A);
	ChangeU2MB(0x0097A2, 0xE74B);
	ChangeU2MB(0x0097AC, 0xE74C);
	ChangeU2MB(0x0097AE, 0xE74D);
	ChangeU2MB(0x0097B1, 0xE74E);
	ChangeU2MB(0x0097B2, 0xE74F);
	ChangeU2MB(0x0097B5, 0xE750);
	ChangeU2MB(0x0097B6, 0xE751);
	ChangeU2MB(0x0097B8, 0xE752);
	ChangeU2MB(0x0097B9, 0xE753);
	ChangeU2MB(0x0097BA, 0xE754);
	ChangeU2MB(0x0097BC, 0xE755);
	ChangeU2MB(0x0097BE, 0xE756);
	ChangeU2MB(0x0097BF, 0xE757);
	ChangeU2MB(0x0097C1, 0xE758);
	ChangeU2MB(0x0097C4, 0xE759);
	ChangeU2MB(0x0097C5, 0xE75A);
	ChangeU2MB(0x0097C7, 0xE75B);
	ChangeU2MB(0x0097C9, 0xE75C);
	ChangeU2MB(0x0097CA, 0xE75D);
	ChangeU2MB(0x0097CC, 0xE75E);
	ChangeU2MB(0x0097CD, 0xE75F);
	ChangeU2MB(0x0097CE, 0xE760);
	ChangeU2MB(0x0097D0, 0xE761);
	ChangeU2MB(0x0097D1, 0xE762);
	ChangeU2MB(0x0097D4, 0xE763);
	ChangeU2MB(0x0097D7, 0xE764);
	ChangeU2MB(0x0097D8, 0xE765);
	ChangeU2MB(0x0097D9, 0xE766);
	ChangeU2MB(0x0097DB, 0xE76A);
	ChangeU2MB(0x0097DD, 0xE767);
	ChangeU2MB(0x0097DE, 0xE768);
	ChangeU2MB(0x0097E0, 0xE769);
	ChangeU2MB(0x0097E1, 0xE76B);
	ChangeU2MB(0x0097E4, 0xE76C);
	ChangeU2MB(0x0097EF, 0xE76D);
	ChangeU2MB(0x0097F1, 0xE76E);
	ChangeU2MB(0x0097F4, 0xE76F);
	ChangeU2MB(0x0097F7, 0xE770);
	ChangeU2MB(0x0097F8, 0xE771);
	ChangeU2MB(0x0097FA, 0xE772);
	ChangeU2MB(0x009807, 0xE773);
	ChangeU2MB(0x00980A, 0xE774);
	ChangeU2MB(0x00980D, 0xE776);
	ChangeU2MB(0x00980E, 0xE777);
	ChangeU2MB(0x009814, 0xE778);
	ChangeU2MB(0x009816, 0xE779);
	ChangeU2MB(0x009819, 0xE775);
	ChangeU2MB(0x00981C, 0xE77A);
	ChangeU2MB(0x00981E, 0xE77B);
	ChangeU2MB(0x009820, 0xE77C);
	ChangeU2MB(0x009823, 0xE77D);
	ChangeU2MB(0x009825, 0xE828);
	ChangeU2MB(0x009826, 0xE77E);
	ChangeU2MB(0x00982B, 0xE821);
	ChangeU2MB(0x00982E, 0xE822);
	ChangeU2MB(0x00982F, 0xE823);
	ChangeU2MB(0x009830, 0xE824);
	ChangeU2MB(0x009832, 0xE825);
	ChangeU2MB(0x009833, 0xE826);
	ChangeU2MB(0x009835, 0xE827);
	ChangeU2MB(0x00983E, 0xE829);
	ChangeU2MB(0x009844, 0xE82A);
	ChangeU2MB(0x009847, 0xE82B);
	ChangeU2MB(0x00984A, 0xE82C);
	ChangeU2MB(0x009851, 0xE82D);
	ChangeU2MB(0x009852, 0xE82E);
	ChangeU2MB(0x009853, 0xE82F);
	ChangeU2MB(0x009856, 0xE830);
	ChangeU2MB(0x009859, 0xE832);
	ChangeU2MB(0x00985A, 0xE833);
	ChangeU2MB(0x009862, 0xE834);
	ChangeU2MB(0x009863, 0xE835);
	ChangeU2MB(0x009866, 0xE837);
	ChangeU2MB(0x00986A, 0xE838);
	ChangeU2MB(0x00986C, 0xE839);
	ChangeU2MB(0x0098AB, 0xE83A);
	ChangeU2MB(0x0098AD, 0xE83B);
	ChangeU2MB(0x0098AE, 0xE83C);
	ChangeU2MB(0x0098B0, 0xE83D);
	ChangeU2MB(0x0098B4, 0xE83E);
	ChangeU2MB(0x0098B7, 0xE83F);
	ChangeU2MB(0x0098B8, 0xE840);
	ChangeU2MB(0x0098BA, 0xE841);
	ChangeU2MB(0x0098BB, 0xE842);
	ChangeU2MB(0x0098BF, 0xE843);
	ChangeU2MB(0x0098C2, 0xE844);
	ChangeU2MB(0x0098C5, 0xE845);
	ChangeU2MB(0x0098C8, 0xE846);
	ChangeU2MB(0x0098CC, 0xE847);
	ChangeU2MB(0x0098E1, 0xE848);
	ChangeU2MB(0x0098E3, 0xE849);
	ChangeU2MB(0x0098E5, 0xE84A);
	ChangeU2MB(0x0098E6, 0xE84B);
	ChangeU2MB(0x0098E7, 0xE84C);
	ChangeU2MB(0x0098EA, 0xE84D);
	ChangeU2MB(0x0098F3, 0xE84E);
	ChangeU2MB(0x0098F6, 0xE84F);
	ChangeU2MB(0x009902, 0xE850);
	ChangeU2MB(0x009907, 0xE851);
	ChangeU2MB(0x009908, 0xE852);
	ChangeU2MB(0x009911, 0xE853);
	ChangeU2MB(0x009915, 0xE854);
	ChangeU2MB(0x009916, 0xE855);
	ChangeU2MB(0x009917, 0xE856);
	ChangeU2MB(0x00991A, 0xE857);
	ChangeU2MB(0x00991B, 0xE858);
	ChangeU2MB(0x00991C, 0xE859);
	ChangeU2MB(0x00991F, 0xE85A);
	ChangeU2MB(0x009922, 0xE85B);
	ChangeU2MB(0x009926, 0xE85C);
	ChangeU2MB(0x00992B, 0xE85E);
	ChangeU2MB(0x009931, 0xE85F);
	ChangeU2MB(0x009932, 0xE860);
	ChangeU2MB(0x009933, 0xE861);
	ChangeU2MB(0x009934, 0xE862);
	ChangeU2MB(0x009935, 0xE863);
	ChangeU2MB(0x009939, 0xE864);
	ChangeU2MB(0x00993A, 0xE865);
	ChangeU2MB(0x00993B, 0xE866);
	ChangeU2MB(0x00993C, 0xE867);
	ChangeU2MB(0x009940, 0xE868);
	ChangeU2MB(0x009941, 0xE869);
	ChangeU2MB(0x009946, 0xE86A);
	ChangeU2MB(0x009947, 0xE86B);
	ChangeU2MB(0x009948, 0xE86C);
	ChangeU2MB(0x00994D, 0xE86D);
	ChangeU2MB(0x00994E, 0xE86E);
	ChangeU2MB(0x009954, 0xE86F);
	ChangeU2MB(0x009958, 0xE870);
	ChangeU2MB(0x009959, 0xE871);
	ChangeU2MB(0x00995B, 0xE872);
	ChangeU2MB(0x00995C, 0xE873);
	ChangeU2MB(0x00995E, 0xE874);
	ChangeU2MB(0x00995F, 0xE875);
	ChangeU2MB(0x009960, 0xE876);
	ChangeU2MB(0x00999B, 0xE877);
	ChangeU2MB(0x00999D, 0xE878);
	ChangeU2MB(0x00999F, 0xE879);
	ChangeU2MB(0x0099A6, 0xE87A);
	ChangeU2MB(0x0099B0, 0xE87B);
	ChangeU2MB(0x0099B1, 0xE87C);
	ChangeU2MB(0x0099B2, 0xE87D);
	ChangeU2MB(0x0099B5, 0xE87E);
	ChangeU2MB(0x0099B9, 0xE921);
	ChangeU2MB(0x0099BA, 0xE922);
	ChangeU2MB(0x0099BD, 0xE923);
	ChangeU2MB(0x0099BF, 0xE924);
	ChangeU2MB(0x0099C3, 0xE925);
	ChangeU2MB(0x0099C9, 0xE926);
	ChangeU2MB(0x0099D3, 0xE927);
	ChangeU2MB(0x0099D4, 0xE928);
	ChangeU2MB(0x0099D9, 0xE929);
	ChangeU2MB(0x0099DA, 0xE92A);
	ChangeU2MB(0x0099DC, 0xE92B);
	ChangeU2MB(0x0099DE, 0xE92C);
	ChangeU2MB(0x0099E7, 0xE92D);
	ChangeU2MB(0x0099EA, 0xE92E);
	ChangeU2MB(0x0099EB, 0xE92F);
	ChangeU2MB(0x0099EC, 0xE930);
	ChangeU2MB(0x0099F0, 0xE931);
	ChangeU2MB(0x0099F4, 0xE932);
	ChangeU2MB(0x0099F5, 0xE933);
	ChangeU2MB(0x0099F9, 0xE934);
	ChangeU2MB(0x0099FD, 0xE935);
	ChangeU2MB(0x0099FE, 0xE936);
	ChangeU2MB(0x009A02, 0xE937);
	ChangeU2MB(0x009A03, 0xE938);
	ChangeU2MB(0x009A04, 0xE939);
	ChangeU2MB(0x009A0B, 0xE93A);
	ChangeU2MB(0x009A0C, 0xE93B);
	ChangeU2MB(0x009A10, 0xE93C);
	ChangeU2MB(0x009A11, 0xE93D);
	ChangeU2MB(0x009A16, 0xE93E);
	ChangeU2MB(0x009A1E, 0xE93F);
	ChangeU2MB(0x009A20, 0xE940);
	ChangeU2MB(0x009A22, 0xE941);
	ChangeU2MB(0x009A23, 0xE942);
	ChangeU2MB(0x009A24, 0xE943);
	ChangeU2MB(0x009A27, 0xE944);
	ChangeU2MB(0x009A2D, 0xE945);
	ChangeU2MB(0x009A2E, 0xE946);
	ChangeU2MB(0x009A33, 0xE947);
	ChangeU2MB(0x009A35, 0xE948);
	ChangeU2MB(0x009A36, 0xE949);
	ChangeU2MB(0x009A38, 0xE94A);
	ChangeU2MB(0x009A41, 0xE94C);
	ChangeU2MB(0x009A44, 0xE94D);
	ChangeU2MB(0x009A47, 0xE94B);
	ChangeU2MB(0x009A4A, 0xE94E);
	ChangeU2MB(0x009A4B, 0xE94F);
	ChangeU2MB(0x009A4C, 0xE950);
	ChangeU2MB(0x009A51, 0xE952);
	ChangeU2MB(0x009A54, 0xE953);
	ChangeU2MB(0x009A56, 0xE954);
	ChangeU2MB(0x009A5D, 0xE955);
	ChangeU2MB(0x009AAA, 0xE956);
	ChangeU2MB(0x009AAC, 0xE957);
	ChangeU2MB(0x009AAE, 0xE958);
	ChangeU2MB(0x009AAF, 0xE959);
	ChangeU2MB(0x009AB2, 0xE95A);
	ChangeU2MB(0x009AB4, 0xE95B);
	ChangeU2MB(0x009AB5, 0xE95C);
	ChangeU2MB(0x009AB6, 0xE95D);
	ChangeU2MB(0x009AB9, 0xE95E);
	ChangeU2MB(0x009ABB, 0xE95F);
	ChangeU2MB(0x009ABE, 0xE960);
	ChangeU2MB(0x009ABF, 0xE961);
	ChangeU2MB(0x009AC1, 0xE962);
	ChangeU2MB(0x009AC3, 0xE963);
	ChangeU2MB(0x009AC6, 0xE964);
	ChangeU2MB(0x009AC8, 0xE965);
	ChangeU2MB(0x009ACE, 0xE966);
	ChangeU2MB(0x009AD0, 0xE967);
	ChangeU2MB(0x009AD2, 0xE968);
	ChangeU2MB(0x009AD5, 0xE969);
	ChangeU2MB(0x009AD6, 0xE96A);
	ChangeU2MB(0x009AD7, 0xE96B);
	ChangeU2MB(0x009ADB, 0xE96C);
	ChangeU2MB(0x009AE0, 0xE96E);
	ChangeU2MB(0x009AE4, 0xE96F);
	ChangeU2MB(0x009AE5, 0xE970);
	ChangeU2MB(0x009AE7, 0xE971);
	ChangeU2MB(0x009AE9, 0xE972);
	ChangeU2MB(0x009AEC, 0xE973);
	ChangeU2MB(0x009AF2, 0xE974);
	ChangeU2MB(0x009AF3, 0xE975);
	ChangeU2MB(0x009AF5, 0xE976);
	ChangeU2MB(0x009AF9, 0xE977);
	ChangeU2MB(0x009AFA, 0xE978);
	ChangeU2MB(0x009AFD, 0xE979);
	ChangeU2MB(0x009AFF, 0xE97A);
	ChangeU2MB(0x009B00, 0xE97B);
	ChangeU2MB(0x009B01, 0xE97C);
	ChangeU2MB(0x009B02, 0xE97D);
	ChangeU2MB(0x009B03, 0xE97E);
	ChangeU2MB(0x009B04, 0xEA21);
	ChangeU2MB(0x009B05, 0xEA22);
	ChangeU2MB(0x009B08, 0xEA23);
	ChangeU2MB(0x009B09, 0xEA24);
	ChangeU2MB(0x009B0B, 0xEA25);
	ChangeU2MB(0x009B0C, 0xEA26);
	ChangeU2MB(0x009B0D, 0xEA27);
	ChangeU2MB(0x009B0E, 0xEA28);
	ChangeU2MB(0x009B10, 0xEA29);
	ChangeU2MB(0x009B12, 0xEA2A);
	ChangeU2MB(0x009B16, 0xEA2B);
	ChangeU2MB(0x009B19, 0xEA2C);
	ChangeU2MB(0x009B1B, 0xEA2D);
	ChangeU2MB(0x009B1C, 0xEA2E);
	ChangeU2MB(0x009B20, 0xEA2F);
	ChangeU2MB(0x009B26, 0xEA30);
	ChangeU2MB(0x009B2B, 0xEA31);
	ChangeU2MB(0x009B2D, 0xEA32);
	ChangeU2MB(0x009B33, 0xEA33);
	ChangeU2MB(0x009B34, 0xEA34);
	ChangeU2MB(0x009B35, 0xEA35);
	ChangeU2MB(0x009B37, 0xEA36);
	ChangeU2MB(0x009B39, 0xEA37);
	ChangeU2MB(0x009B3A, 0xEA38);
	ChangeU2MB(0x009B3D, 0xEA39);
	ChangeU2MB(0x009B48, 0xEA3A);
	ChangeU2MB(0x009B4B, 0xEA3B);
	ChangeU2MB(0x009B4C, 0xEA3C);
	ChangeU2MB(0x009B55, 0xEA3D);
	ChangeU2MB(0x009B56, 0xEA3E);
	ChangeU2MB(0x009B57, 0xEA3F);
	ChangeU2MB(0x009B5B, 0xEA40);
	ChangeU2MB(0x009B5E, 0xEA41);
	ChangeU2MB(0x009B61, 0xEA42);
	ChangeU2MB(0x009B63, 0xEA43);
	ChangeU2MB(0x009B65, 0xEA44);
	ChangeU2MB(0x009B66, 0xEA45);
	ChangeU2MB(0x009B68, 0xEA46);
	ChangeU2MB(0x009B6A, 0xEA47);
	ChangeU2MB(0x009B6B, 0xEA48);
	ChangeU2MB(0x009B6C, 0xEA49);
	ChangeU2MB(0x009B6D, 0xEA4A);
	ChangeU2MB(0x009B6E, 0xEA4B);
	ChangeU2MB(0x009B73, 0xEA4C);
	ChangeU2MB(0x009B77, 0xEA4E);
	ChangeU2MB(0x009B78, 0xEA4F);
	ChangeU2MB(0x009B79, 0xEA50);
	ChangeU2MB(0x009B7F, 0xEA51);
	ChangeU2MB(0x009B80, 0xEA52);
	ChangeU2MB(0x009B84, 0xEA53);
	ChangeU2MB(0x009B85, 0xEA54);
	ChangeU2MB(0x009B86, 0xEA55);
	ChangeU2MB(0x009B87, 0xEA56);
	ChangeU2MB(0x009B89, 0xEA57);
	ChangeU2MB(0x009B8A, 0xEA58);
	ChangeU2MB(0x009B8B, 0xEA59);
	ChangeU2MB(0x009B8D, 0xEA5A);
	ChangeU2MB(0x009B90, 0xEA5C);
	ChangeU2MB(0x009B94, 0xEA5D);
	ChangeU2MB(0x009B9A, 0xEA5E);
	ChangeU2MB(0x009B9D, 0xEA5F);
	ChangeU2MB(0x009B9E, 0xEA60);
	ChangeU2MB(0x009BA6, 0xEA61);
	ChangeU2MB(0x009BA7, 0xEA62);
	ChangeU2MB(0x009BA9, 0xEA63);
	ChangeU2MB(0x009BAC, 0xEA64);
	ChangeU2MB(0x009BB0, 0xEA65);
	ChangeU2MB(0x009BB2, 0xEA67);
	ChangeU2MB(0x009BB7, 0xEA68);
	ChangeU2MB(0x009BB8, 0xEA69);
	ChangeU2MB(0x009BBC, 0xEA6B);
	ChangeU2MB(0x009BBE, 0xEA6C);
	ChangeU2MB(0x009BBF, 0xEA6D);
	ChangeU2MB(0x009BC1, 0xEA6E);
	ChangeU2MB(0x009BC7, 0xEA6F);
	ChangeU2MB(0x009BC8, 0xEA70);
	ChangeU2MB(0x009BCE, 0xEA71);
	ChangeU2MB(0x009BD0, 0xEA72);
	ChangeU2MB(0x009BD7, 0xEA73);
	ChangeU2MB(0x009BD8, 0xEA74);
	ChangeU2MB(0x009BDD, 0xEA75);
	ChangeU2MB(0x009BDF, 0xEA76);
	ChangeU2MB(0x009BE5, 0xEA77);
	ChangeU2MB(0x009BE7, 0xEA78);
	ChangeU2MB(0x009BEA, 0xEA79);
	ChangeU2MB(0x009BEB, 0xEA7A);
	ChangeU2MB(0x009BEF, 0xEA7B);
	ChangeU2MB(0x009BF3, 0xEA7C);
	ChangeU2MB(0x009BF7, 0xEA7D);
	ChangeU2MB(0x009BF8, 0xEA7E);
	ChangeU2MB(0x009BF9, 0xEB21);
	ChangeU2MB(0x009BFA, 0xEB22);
	ChangeU2MB(0x009BFD, 0xEB23);
	ChangeU2MB(0x009BFF, 0xEB24);
	ChangeU2MB(0x009C02, 0xEB26);
	ChangeU2MB(0x009C0B, 0xEB27);
	ChangeU2MB(0x009C0F, 0xEB28);
	ChangeU2MB(0x009C11, 0xEB29);
	ChangeU2MB(0x009C16, 0xEB2A);
	ChangeU2MB(0x009C18, 0xEB2B);
	ChangeU2MB(0x009C19, 0xEB2C);
	ChangeU2MB(0x009C1A, 0xEB2D);
	ChangeU2MB(0x009C1C, 0xEB2E);
	ChangeU2MB(0x009C1E, 0xEB2F);
	ChangeU2MB(0x009C22, 0xEB30);
	ChangeU2MB(0x009C23, 0xEB31);
	ChangeU2MB(0x009C26, 0xEB32);
	ChangeU2MB(0x009C27, 0xEB33);
	ChangeU2MB(0x009C28, 0xEB34);
	ChangeU2MB(0x009C29, 0xEB35);
	ChangeU2MB(0x009C2A, 0xEB36);
	ChangeU2MB(0x009C31, 0xEB37);
	ChangeU2MB(0x009C35, 0xEB38);
	ChangeU2MB(0x009C36, 0xEB39);
	ChangeU2MB(0x009C37, 0xEB3A);
	ChangeU2MB(0x009C3D, 0xEB3B);
	ChangeU2MB(0x009C41, 0xEB3C);
	ChangeU2MB(0x009C43, 0xEB3D);
	ChangeU2MB(0x009C44, 0xEB3E);
	ChangeU2MB(0x009C45, 0xEB3F);
	ChangeU2MB(0x009C49, 0xEB40);
	ChangeU2MB(0x009C4A, 0xEB41);
	ChangeU2MB(0x009C4E, 0xEB42);
	ChangeU2MB(0x009C4F, 0xEB43);
	ChangeU2MB(0x009C50, 0xEB44);
	ChangeU2MB(0x009C53, 0xEB45);
	ChangeU2MB(0x009C54, 0xEB46);
	ChangeU2MB(0x009C56, 0xEB47);
	ChangeU2MB(0x009C58, 0xEB48);
	ChangeU2MB(0x009C5B, 0xEB49);
	ChangeU2MB(0x009C5C, 0xEB50);
	ChangeU2MB(0x009C5D, 0xEB4A);
	ChangeU2MB(0x009C5E, 0xEB4B);
	ChangeU2MB(0x009C5F, 0xEB4C);
	ChangeU2MB(0x009C63, 0xEB4D);
	ChangeU2MB(0x009C68, 0xEB52);
	ChangeU2MB(0x009C69, 0xEB4E);
	ChangeU2MB(0x009C6A, 0xEB4F);
	ChangeU2MB(0x009C6B, 0xEB51);
	ChangeU2MB(0x009C6E, 0xEB53);
	ChangeU2MB(0x009C70, 0xEB54);
	ChangeU2MB(0x009C72, 0xEB55);
	ChangeU2MB(0x009C75, 0xEB56);
	ChangeU2MB(0x009C77, 0xEB57);
	ChangeU2MB(0x009C7B, 0xEB58);
	ChangeU2MB(0x009CE6, 0xEB59);
	ChangeU2MB(0x009CF2, 0xEB5A);
	ChangeU2MB(0x009CF7, 0xEB5B);
	ChangeU2MB(0x009CF9, 0xEB5C);
	ChangeU2MB(0x009D02, 0xEB5E);
	ChangeU2MB(0x009D0B, 0xEB5D);
	ChangeU2MB(0x009D11, 0xEB5F);
	ChangeU2MB(0x009D17, 0xEB60);
	ChangeU2MB(0x009D18, 0xEB61);
	ChangeU2MB(0x009D1C, 0xEB62);
	ChangeU2MB(0x009D1D, 0xEB63);
	ChangeU2MB(0x009D1E, 0xEB64);
	ChangeU2MB(0x009D2F, 0xEB65);
	ChangeU2MB(0x009D30, 0xEB66);
	ChangeU2MB(0x009D32, 0xEB67);
	ChangeU2MB(0x009D33, 0xEB68);
	ChangeU2MB(0x009D34, 0xEB69);
	ChangeU2MB(0x009D3A, 0xEB6A);
	ChangeU2MB(0x009D3C, 0xEB6B);
	ChangeU2MB(0x009D3D, 0xEB6D);
	ChangeU2MB(0x009D42, 0xEB6E);
	ChangeU2MB(0x009D43, 0xEB6F);
	ChangeU2MB(0x009D45, 0xEB6C);
	ChangeU2MB(0x009D47, 0xEB70);
	ChangeU2MB(0x009D4A, 0xEB71);
	ChangeU2MB(0x009D53, 0xEB72);
	ChangeU2MB(0x009D54, 0xEB73);
	ChangeU2MB(0x009D5F, 0xEB74);
	ChangeU2MB(0x009D62, 0xEB76);
	ChangeU2MB(0x009D63, 0xEB75);
	ChangeU2MB(0x009D65, 0xEB77);
	ChangeU2MB(0x009D69, 0xEB78);
	ChangeU2MB(0x009D6A, 0xEB79);
	ChangeU2MB(0x009D76, 0xEB7C);
	ChangeU2MB(0x009D77, 0xEB7D);
	ChangeU2MB(0x009D7B, 0xEB7E);
	ChangeU2MB(0x009D7C, 0xEC21);
	ChangeU2MB(0x009D7E, 0xEC22);
	ChangeU2MB(0x009D83, 0xEC23);
	ChangeU2MB(0x009D84, 0xEC24);
	ChangeU2MB(0x009D86, 0xEC25);
	ChangeU2MB(0x009D8A, 0xEC26);
	ChangeU2MB(0x009D8D, 0xEC27);
	ChangeU2MB(0x009D8E, 0xEC28);
	ChangeU2MB(0x009D92, 0xEC29);
	ChangeU2MB(0x009D93, 0xEC2A);
	ChangeU2MB(0x009D95, 0xEC2B);
	ChangeU2MB(0x009D96, 0xEC2C);
	ChangeU2MB(0x009D97, 0xEC2D);
	ChangeU2MB(0x009D98, 0xEC2E);
	ChangeU2MB(0x009DA1, 0xEC2F);
	ChangeU2MB(0x009DAA, 0xEC30);
	ChangeU2MB(0x009DAC, 0xEC31);
	ChangeU2MB(0x009DAE, 0xEC32);
	ChangeU2MB(0x009DB1, 0xEC33);
	ChangeU2MB(0x009DB5, 0xEC34);
	ChangeU2MB(0x009DB9, 0xEC35);
	ChangeU2MB(0x009DBC, 0xEC36);
	ChangeU2MB(0x009DBF, 0xEC37);
	ChangeU2MB(0x009DC3, 0xEC38);
	ChangeU2MB(0x009DC7, 0xEC39);
	ChangeU2MB(0x009DC9, 0xEC3A);
	ChangeU2MB(0x009DCA, 0xEC3B);
	ChangeU2MB(0x009DD4, 0xEC3C);
	ChangeU2MB(0x009DD5, 0xEC3D);
	ChangeU2MB(0x009DD6, 0xEC3E);
	ChangeU2MB(0x009DD7, 0xEC3F);
	ChangeU2MB(0x009DDA, 0xEC40);
	ChangeU2MB(0x009DDE, 0xEC41);
	ChangeU2MB(0x009DDF, 0xEC42);
	ChangeU2MB(0x009DE0, 0xEC43);
	ChangeU2MB(0x009DE5, 0xEC44);
	ChangeU2MB(0x009DE7, 0xEC45);
	ChangeU2MB(0x009DE9, 0xEC46);
	ChangeU2MB(0x009DEB, 0xEC47);
	ChangeU2MB(0x009DEE, 0xEC48);
	ChangeU2MB(0x009DF0, 0xEC49);
	ChangeU2MB(0x009DF3, 0xEC4A);
	ChangeU2MB(0x009DF4, 0xEC4B);
	ChangeU2MB(0x009DFE, 0xEC4C);
	ChangeU2MB(0x009E02, 0xEC4E);
	ChangeU2MB(0x009E07, 0xEC4F);
	ChangeU2MB(0x009E0A, 0xEC4D);
	ChangeU2MB(0x009E0E, 0xEC50);
	ChangeU2MB(0x009E10, 0xEC51);
	ChangeU2MB(0x009E11, 0xEC52);
	ChangeU2MB(0x009E12, 0xEC53);
	ChangeU2MB(0x009E15, 0xEC54);
	ChangeU2MB(0x009E16, 0xEC55);
	ChangeU2MB(0x009E1C, 0xEC57);
	ChangeU2MB(0x009E1D, 0xEC58);
	ChangeU2MB(0x009E7A, 0xEC59);
	ChangeU2MB(0x009E7B, 0xEC5A);
	ChangeU2MB(0x009E7C, 0xEC5B);
	ChangeU2MB(0x009E80, 0xEC5C);
	ChangeU2MB(0x009E82, 0xEC5D);
	ChangeU2MB(0x009E83, 0xEC5E);
	ChangeU2MB(0x009E84, 0xEC5F);
	ChangeU2MB(0x009E85, 0xEC60);
	ChangeU2MB(0x009E87, 0xEC61);
	ChangeU2MB(0x009E8E, 0xEC62);
	ChangeU2MB(0x009E8F, 0xEC63);
	ChangeU2MB(0x009E96, 0xEC64);
	ChangeU2MB(0x009E98, 0xEC65);
	ChangeU2MB(0x009E9B, 0xEC66);
	ChangeU2MB(0x009E9E, 0xEC67);
	ChangeU2MB(0x009EA4, 0xEC68);
	ChangeU2MB(0x009EA8, 0xEC69);
	ChangeU2MB(0x009EAC, 0xEC6A);
	ChangeU2MB(0x009EAE, 0xEC6B);
	ChangeU2MB(0x009EAF, 0xEC6C);
	ChangeU2MB(0x009EB0, 0xEC6D);
	ChangeU2MB(0x009EB3, 0xEC6E);
	ChangeU2MB(0x009EB4, 0xEC6F);
	ChangeU2MB(0x009EB5, 0xEC70);
	ChangeU2MB(0x009EC6, 0xEC71);
	ChangeU2MB(0x009EC8, 0xEC72);
	ChangeU2MB(0x009ECB, 0xEC73);
	ChangeU2MB(0x009ED5, 0xEC74);
	ChangeU2MB(0x009EDF, 0xEC75);
	ChangeU2MB(0x009EE4, 0xEC76);
	ChangeU2MB(0x009EE7, 0xEC77);
	ChangeU2MB(0x009EEC, 0xEC78);
	ChangeU2MB(0x009EED, 0xEC79);
	ChangeU2MB(0x009EEE, 0xEC7A);
	ChangeU2MB(0x009EF0, 0xEC7B);
	ChangeU2MB(0x009EF1, 0xEC7C);
	ChangeU2MB(0x009EF2, 0xEC7D);
	ChangeU2MB(0x009EF5, 0xEC7E);
	ChangeU2MB(0x009EF8, 0xED21);
	ChangeU2MB(0x009EFF, 0xED22);
	ChangeU2MB(0x009F02, 0xED23);
	ChangeU2MB(0x009F03, 0xED24);
	ChangeU2MB(0x009F09, 0xED25);
	ChangeU2MB(0x009F0F, 0xED26);
	ChangeU2MB(0x009F10, 0xED27);
	ChangeU2MB(0x009F11, 0xED28);
	ChangeU2MB(0x009F12, 0xED29);
	ChangeU2MB(0x009F14, 0xED2A);
	ChangeU2MB(0x009F16, 0xED2B);
	ChangeU2MB(0x009F17, 0xED2C);
	ChangeU2MB(0x009F19, 0xED2D);
	ChangeU2MB(0x009F1A, 0xED2E);
	ChangeU2MB(0x009F1B, 0xED2F);
	ChangeU2MB(0x009F1F, 0xED30);
	ChangeU2MB(0x009F22, 0xED31);
	ChangeU2MB(0x009F26, 0xED32);
	ChangeU2MB(0x009F2A, 0xED33);
	ChangeU2MB(0x009F2B, 0xED34);
	ChangeU2MB(0x009F2F, 0xED35);
	ChangeU2MB(0x009F31, 0xED36);
	ChangeU2MB(0x009F32, 0xED37);
	ChangeU2MB(0x009F34, 0xED38);
	ChangeU2MB(0x009F37, 0xED39);
	ChangeU2MB(0x009F39, 0xED3A);
	ChangeU2MB(0x009F3A, 0xED3B);
	ChangeU2MB(0x009F3C, 0xED3C);
	ChangeU2MB(0x009F3D, 0xED3D);
	ChangeU2MB(0x009F3F, 0xED3E);
	ChangeU2MB(0x009F41, 0xED3F);
	ChangeU2MB(0x009F43, 0xED40);
	ChangeU2MB(0x009F44, 0xED41);
	ChangeU2MB(0x009F45, 0xED42);
	ChangeU2MB(0x009F46, 0xED43);
	ChangeU2MB(0x009F47, 0xED44);
	ChangeU2MB(0x009F53, 0xED45);
	ChangeU2MB(0x009F55, 0xED46);
	ChangeU2MB(0x009F56, 0xED47);
	ChangeU2MB(0x009F57, 0xED48);
	ChangeU2MB(0x009F58, 0xED49);
	ChangeU2MB(0x009F5A, 0xED4A);
	ChangeU2MB(0x009F5D, 0xED4B);
	ChangeU2MB(0x009F5E, 0xED4C);
	ChangeU2MB(0x009F68, 0xED4D);
	ChangeU2MB(0x009F69, 0xED4E);
	ChangeU2MB(0x009F6D, 0xED4F);
	ChangeU2MB(0x009F6E, 0xED50);
	ChangeU2MB(0x009F6F, 0xED51);
	ChangeU2MB(0x009F70, 0xED52);
	ChangeU2MB(0x009F71, 0xED53);
	ChangeU2MB(0x009F73, 0xED54);
	ChangeU2MB(0x009F75, 0xED55);
	ChangeU2MB(0x009F7A, 0xED56);
	ChangeU2MB(0x009F7D, 0xED57);
	ChangeU2MB(0x009F8F, 0xED58);
	ChangeU2MB(0x009F90, 0xED59);
	ChangeU2MB(0x009F91, 0xED5A);
	ChangeU2MB(0x009F92, 0xED5B);
	ChangeU2MB(0x009F94, 0xED5C);
	ChangeU2MB(0x009F96, 0xED5D);
	ChangeU2MB(0x009F97, 0xED5E);
	ChangeU2MB(0x009F9E, 0xED5F);
	ChangeU2MB(0x009FA1, 0xED60);
	ChangeU2MB(0x009FA2, 0xED61);
	ChangeU2MB(0x009FA3, 0xED62);
	ChangeU2MB(0x009FA5, 0xED63);
#endif
}

};// namespace xm
